#!/bin/bash

##########################################################################
# 
# This file is part of FORCE - Framework for Operational Radiometric 
# Correction for Environmental monitoring.
# 
# Copyright (C) 2013-2018 David Frantz
# 
# FORCE is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# FORCE is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with FORCE.  If not, see <http://www.gnu.org/licenses/>.
# 
##########################################################################

# this script is an higher-level batch interface for the FORCE Level-2 processing system

function L2PS(){

  FAIL=0

  if [[ $1 == *"tar.gz"* ]]; then

    BASE=$(basename $1 | sed 's/.tar.gz//')
    TODO=$4/$BASE

    mkdir $TODO
    if [ ! -d $TODO ]; then
      echo "creating temp directory failed"
      exit
    fi



    tar --ignore-command-error -xzf $1 --exclude gap_mask --exclude='*ANG.txt' --exclude='*GCP.txt' --exclude='*VER.jpg' --exclude='*VER.txt' --exclude='*BQA.TIF' --exclude='*.GTF' --exclude='LE07*B8.TIF' --exclude='LE07*B6_VCID_2.TIF' --exclude='LC08*B11.TIF' --exclude='LC08*B8.TIF' -C $TODO &> /dev/null
    if [ ! $? -eq 0 ]; then
      echo "$BASE: tar.gz container is corrupt."
      FAIL=1
    fi

  else

    TODO=$1

  fi

  if [ $FAIL -eq 0 ]; then
    # process the image
    $3/force-l2ps $TODO $2
  fi

  # update queue status; queue must be locked and unlocked
  SEARCH=$(echo $1 | sed 's_/_\\/_g')
  lockfile-create $5
  sed -i.tmp "s/$SEARCH.*/$SEARCH DONE/g" $5
  chmod --reference $5".tmp" $5
  rm $5".tmp"
  lockfile-remove $5

  if [[ $1 == *"tar.gz"* ]]; then
    rm -r $TODO
  fi

}

export -f L2PS


EXPECTED_ARGS=3

# if wrong number of input args, stop
if [ $# -ne $EXPECTED_ARGS ]; then
  echo "Usage: `basename $0` prm(file) ncpu delay"
  exit
fi

PRM=$1
CPU=$2
DELAY=$3
BINDIR=???


# if parameter file doesn't exist, stop
if [ ! -f $PRM ]; then
  echo $PRM "does not exist"; exit
elif [ ! -r $PRM ]; then
  echo $PRM "is not readable"; exit
fi


# get input file list from parameter file
if [ $(grep 'FILE_QUEUE' $PRM | wc -l) -eq 0 ]; then
  echo "FILE_QUEUE is not given in parameter file"; exit
else
  IFILE=$(grep 'FILE_QUEUE' $PRM | sed 's/.* *= *//g')
  if [ ! -f $IFILE ]; then
    echo $IFILE "does not exist"; exit
  elif [ ! -r $IFILE ]; then
    echo $IFILE "is not readable"; exit
  fi
fi


# get output directory from parameter file
if [ $(grep 'DIR_LEVEL2' $PRM | wc -l) -eq 0 ]; then
  echo "DIR_LEVEL2 is not given in parameter file"; exit
else
  OD=$(grep 'DIR_LEVEL2' $PRM | sed 's/.* *= *//g')
  if [ ! -d $OD ]; then
    echo $OD "does not exist"; exit
  elif [ ! -w $OD ]; then
    echo $OD "is not writeable"; exit
  fi
fi


# get temporary directory from parameter file
if [ $(grep 'DIR_TEMP' $PRM | wc -l) -eq 0 ]; then
  echo "DIR_TEMP is not given in parameter file"; exit
else
  TD=$(grep 'DIR_TEMP' $PRM | sed 's/.* *= *//g')
  if [ ! -d $TD ]; then
    echo $TD "does not exist"; exit
  elif [ ! -w $TD ]; then
    echo $TD "is not writeable"; exit
  fi
fi


# cuurent time
TIME=$(date +"%Y%m%d%H%M%S")


# protect against multiple calls
if [ $(ps aux | grep 'L2PS' | wc -l) -gt 1 ]; then
  echo "L2PS is already running. Exit." > $OD/FORCE-L2PS_$TIME.log
  exit
fi


QUEUE=$(grep 'QUEUED' $IFILE | sed 's/QUEUED//g')

# count input
NUM=$(echo $QUEUE | wc -w)
if [ $NUM -eq 0 ]; then
  echo "No images queue in " $IFILE; exit
else
  echo $NUM "images enqueued. Start processing with" $CPU "CPUs"
fi


# test if full paths are given in $IFILE
for f in $QUEUE; do
  if [[ ! $f == /* ]]; then
    echo "relative pathnames are not allowed in FILE_QUEUE"; exit
  fi
done


# test if force-l2ps was compiled in DEBUG mode
$BINDIR/force-l2ps ? ?
if [ $? -eq 1 ]; then
  echo "force-l2ps was compiled in DEBUG mode"
  if [ $NUM -gt 1 ] || [ $CPU -gt 1 ]; then
    echo "  cannot use DEBUG mode in force-level2"
    echo "  solutions:"
    echo "    (1) re-compile with disabled DEBUG mode or"
    echo "    (2) reduce ncpu to 1 and reduce queue length in FILE_QUEUE to 1"
    exit
  fi
fi

CPUFILE=$TD"/cpu-"$TIME
echo $CPU > $CPUFILE

echo $QUEUE | parallel -d ' ' -j$CPUFILE --delay $DELAY --eta L2PS {} $PRM $BINDIR $TD $IFILE > $OD/FORCE-L2PS_$TIME.log

rm $CPUFILE
