/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
This file contains general functions for memory allocation / deallocation
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#include "force.h"


/** Allocate array
+++ This function allocates a block of memory, and initializes it with 0.
--- ptr:    Pointer to the memory block
--- n:      Number of elements to allocate
--- size:   Size of each element
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void alloc(void **ptr, int n, int size){

  *ptr = (void*) calloc(n, size);
  if (ptr == NULL){ printf("unable to allocate memory!\n"); exit(1);}

  return;
}


/** Allocate 2D-array
+++ This function allocates blocks of memory, and initializes them with 0.
--- ptr:    Pointer to the memory block
--- n1:     Number of elements to allocate (1st dimension)
--- n2:     Number of elements to allocate (2nd dimension)
--- size:   Size of each element
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void alloc_2D(void ***ptr, int n1, int n2, int size){
void **tmp;
int i;

  tmp = (void**) malloc(n1*sizeof(void*));
  if (tmp == NULL){ printf("unable to allocate memory!\n"); exit(1);}

  for (i=0; i<n1; i++){
    tmp[i] = (void*) calloc(n2, size);
    if (tmp[i] == NULL){ printf("unable to allocate memory!\n"); exit(1);}
  }

  *ptr = tmp;
  return;
}


/** Allocate 3D-array
+++ This function allocates blocks of memory, and initializes them with 0.
--- ptr:    Pointer to the memory block
--- n1:     Number of elements to allocate (1st dimension)
--- n2:     Number of elements to allocate (2nd dimension)
--- n3:     Number of elements to allocate (3nd dimension)
--- size:   Size of each element
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void alloc_3D(void ****ptr, int n1, int n2, int n3, int size){
void ***tmp;
int i, j;

  tmp = (void***) malloc(n1*sizeof(void**));
  if (tmp == NULL){ printf("unable to allocate memory!\n"); exit(1);}

  for (i=0; i<n1; i++){

    tmp[i] = (void**) malloc(n2*sizeof(void*));
    if (tmp[i] == NULL){ printf("unable to allocate memory!\n"); exit(1);}

    for (j=0; j<n2; j++){
      tmp[i][j] = (void*) calloc(n3, size);
      if (tmp[i][j] == NULL){ printf("unable to allocate memory!\n"); exit(1);}
    }

  }

  *ptr = tmp;
  return;
}


/** Allocate contiguos 2D-array
+++ This function allocates a block of memory, and initializes it with 0.
+++ Unlike alloc_2D, one single block of memory is allocated and pointers
+++ are set for the 2nd dimension.
--- ptr:    Pointer to the memory block
--- n1:     Number of elements to allocate (1st dimension)
--- n2:     Number of elements to allocate (2nd dimension)
--- size:   Size of each element
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void alloc_2DC(void ***ptr, int n1, int n2, int size){
void **tmp;
int i;

  tmp = (void**) malloc(n1*sizeof(void*));
  if (tmp == NULL){ printf("unable to allocate memory!\n"); exit(1);}

  tmp[0] = (void*) calloc(n1*n2, size);
  if (tmp[0] == NULL){ printf("unable to allocate memory!\n"); exit(1);}

  for (i=0; i<n1; i++) tmp[i] = ((char*)tmp + n2*i);

  *ptr = tmp;
  return;
}


/** Free 2D-array
+++ This function deallocates an allocated 2D array.
--- ptr:    Pointer to the memory block
--- n1:     Number of elements to deallocate (1st dimension)
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void free_2D(void **ptr, int n){
int i;

  for (i=0; i<n; i++) free((void*)ptr[i]);
  free((void*)ptr);

  return;
} 


/** Free 3D-array
+++ This function deallocates an allocated 2D array.
--- ptr:    Pointer to the memory block
--- n1:     Number of elements to deallocate (1st dimension)
--- n2:     Number of elements to deallocate (2nd dimension)
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void free_3D(void ***ptr, int n1, int n2){
int i, j;

  for (i=0; i<n1; i++){
    for (j=0; j<n2; j++) free((void*)ptr[i][j]);
    free((void*)ptr[i]);
  }
  free((void*)ptr);

  return;
} 


/** Free contiguos 2D-array
+++ This function deallocates an allocated, contiguous 2D array.
--- ptr:    Pointer to the memory block
--- n1:     Number of elements to deallocate (1st dimension)
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void free_2DC(void **ptr){
  
  free(ptr[0]);
  free(ptr);

  return;
} 

