/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
This file contains functions for handling Aerosol Optical Depth
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#include "force.h"


/** Read external AOD
+++ This function reads externally provided AOD values, e.g. to use an AOD
+++ climatology. The function also sets the global AOD fallback, which is 
+++ used as external value, if no AOD LUTs are provided.
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int aodfileread(){
FILE *fp = NULL;
char  buffer[MAX_STR_LEN] = "\0";
char *tokenptr = NULL;
const char *separator = " ";
char fname[MAX_STR_LEN];
int b; 
float co[3] = { -2.8665, -1.4787, -0.0486 };
coord_t center, site, diff;
float dist, min_dist = LONG_MAX;


  /** global fallback
  +** *******************************************************************/

  // AOD derived from global AERONET data (median AOD)
  for (b=0; b<_NB_; b++) atc.aod_ext[b] = exp(co[0]) * 
                          pow(atc.wvl[b], co[1] + co[2]*atc.lwvl[b]);

  #ifdef DEBUG
  print_fvector(atc.aod_ext,   "\nglobal aod fallback", _NB_, 1, 3);
  #endif

  // if AOD directory not given, continue with global fallback
  if ((strcmp(par.d_aod, "NULL") == 0)) return SUCCESS;


  /** spatio-temporal fallback
  +** *******************************************************************/

  // scene center
  center.x = atc.cg[(atc.ngy/2)*atc.ngx+(atc.ngx/2)].geox;
  center.y = atc.cg[(atc.ngy/2)*atc.ngx+(atc.ngx/2)].geoy;

  // daily LUT
  sprintf(fname, "%s/AOD_%03d.txt", par.d_aod, meta.date.doy);

  // open aod file
  if ((fp = fopen(fname, "r")) == NULL){
      printf("Unable to open aod file!\n"); return FAILURE;}

  // process line by line
  while (fgets(buffer, MAX_STR_LEN, fp) != NULL){

    tokenptr = strtok(buffer, separator);

    site.x = atof(tokenptr); tokenptr = strtok(NULL, separator);
    site.y = atof(tokenptr); tokenptr = strtok(NULL, separator);
    diff.x = center.x-site.x; diff.y = center.y-site.y;

    dist = sqrt(diff.x*diff.x+diff.y*diff.y);
    if (dist < min_dist){
      min_dist = dist;
      co[0] = atof(tokenptr); tokenptr = strtok(NULL, separator);
      co[1] = atof(tokenptr); tokenptr = strtok(NULL, separator);
      co[2] = atof(tokenptr);
    }

  }

  fclose(fp);

  // AOD derived from LUT
  for (b=0; b<_NB_; b++) atc.aod_ext[b] = exp(co[0]) * 
                        pow(atc.wvl[b], co[1] + co[2]*atc.lwvl[b]);

  #ifdef DEBUG
  print_fvector(atc.aod_ext,   "\naod from file", _NB_, 1, 3);
  #endif

  return SUCCESS;
}


/** Compile Dark Object Persistency
+++ This function is intended to compile the Dark Object Persistency layer
+++ of the Dark Object Database.
+++ Note that this functionality is currently unavailable in this version
+++ of FORCE. To be updated.
--- DOP: Dark Object Persistency
+++ Return: FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int dark_object_persistency(int2u **DOP){

  printf("DODB deprecated. ");
  return FAILURE;
}


/** Extract dark targets from image
+++ This function identifies candidate dark targets and extracts object-
+++ related information. For each object, the target location, radius, 
+++ size, reflectance and environment reflectance is tabulated. There are
+++ different methods for identifying water/shadow/vegetation targets. The
+++ function returns the dark target information and number of targets.
--- TOA:    TOA reflectance
--- QAI:    Quality Assurance Information
--- DOP:    Dark Object Persistency
--- TOP:    Topographic Derivatives
--- type:   Target type (water/shadow/vegetation)
--- DOBJ:   Extracted targets (returned)
+++ Return: Number of candidate targets
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int extract_dark_target(int2s **TOA, qai_t QAI, int2u *DOP, top_t TOP, int type, darkobj_t **DOBJ){
int b, bb, i, j, p, o;
bool   valid;
bool  *LAPSE    = NULL; // lapsing spectra
bool  *TARGET   = NULL; // candidate targets
bool  *INVERT   = NULL; // inverse of candidate targets
int2u *DISTANCE = NULL; // distance to "shore" of target
int   *SEGMENT  = NULL; // target segmentation
int num;                // number of candidate targets
double *esum = NULL;    // sum for env. reflectance mean
double *tsum = NULL;    // sum for target reflectance mean
double *csum = NULL;    // sum for illumin. angle mean
double *ssum = NULL;    // sum for sky view mean
double *zsum = NULL;    // sum for elevation mean
double tn = 0, en = 0;  // ctr for env. / target reflectance mean
float *etoa = NULL;     // global env. reflectance
darkobj_t *dobj = NULL; // dark targets


  // shadow type is disabled until further notice
  if (type == AOD_SHD){
    *DOBJ = dobj;
    return 0;
  }
  

  /** allocate memory
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
  
  alloc((void**)&TARGET,   _NC_, sizeof(bool));
  alloc((void**)&INVERT,   _NC_, sizeof(bool));
  alloc((void**)&LAPSE,    _NC_, sizeof(bool));  

  alloc((void**)&SEGMENT,  _NC_, sizeof(int));
  alloc((void**)&DISTANCE, _NC_, sizeof(int2u));
  
  alloc((void**)&tsum,     _NB_, sizeof(double));
  alloc((void**)&esum,     _NB_, sizeof(double));
  alloc((void**)&etoa,     _NB_, sizeof(float));


  /** identify candidate targets
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

  for (p=0; p<_NC_; p++){

    switch (type){
      case AOD_WAT:                               // water target:
        if (DOP[p] > 500 &&                       // persistence > 5%
            TOA[_B_.swir2][p] < 250 &&            // swir2 < 2.5%
            TOA[_B_.red][p] < 2000 &&             // red < 20%
            TOA[_B_.blue][p] > TOA[_B_.nir][p] && // blue > nir
            !QAI.off[p] &&                        // within-image
            TOP.slp[p] < 870){                    // slope < 5°
          TARGET[p] = true;
        } else {
          INVERT[p] = true;
        }
        break;
      case AOD_SHD:                               // shadow target:
        if (DOP[p] > 500 &&                       // persistence > 5%
            TOA[_B_.swir2][p] < 500 &&            // swir2 < 5%
            TOA[_B_.red][p] < 2000 &&             // red < 20%
            TOA[_B_.blue][p] > TOA[_B_.nir][p] && // blue > nir
            !QAI.off[p] &&                        // within-image
            TOP.ill[p] < 5000 &&                  // illumination angle > 60°
            TOP.slp[p] > 870){                    // slope > 5°
          TARGET[p] = true;
        } else {
          INVERT[p] = true;
        }
        break;
      case AOD_VEG:                                    // vegetation target:
        if (DOP[p] > 500 &&                            // persistence > 5%
            TOA[_B_.swir2][p] < 500 &&                 // swir2 < 5%
            TOA[_B_.red][p] < 2000 &&                  // red < 20%
            TOA[_B_.blue][p] < TOA[_B_.nir][p] &&      // blue < nir
            !QAI.off[p] &&                             // within-image
            (TOA[_B_.nir][p]/_RS_-TOA[_B_.red][p]/_RS_)/         // 
            (TOA[_B_.nir][p]/_RS_+TOA[_B_.red][p]/_RS_) > 0.4 && // NDVI > 0.4
            TOP.slp[p] < 2620){                        // slope < 15°
          TARGET[p] = true;
        } else {
          INVERT[p] = true;
        }
        break;
      default:
        printf("unknown dark target type.\n"); exit(1);
    }

  }


  /** target segmentation
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

  // sub-divide the targets. This ensures that a target cannot run through
  // the complete image, e.g. a river
  for (i=0; i<_NY_; i+=atc.gres){
  for (j=0; j<_NX_; j++){
    p = i*_NX_+j;
    TARGET[p] = FALSE;
  }
  }

  for (j=0; j<_NX_; j+=atc.gres){
  for (i=0; i<_NY_; i++){
    p = i*_NX_+j;
    TARGET[p] = FALSE;
  }
  }

  // segmentate targets
  num = connectedcomponents(TARGET, SEGMENT);
  free((void*)TARGET);


  // distance to "shore"
  dist_transform(INVERT, DISTANCE);
  free((void*)INVERT);


  #ifdef DEBUG
  char fname[MAX_STR_LEN]; sprintf(fname, "DO-CCD-TYPE-%d", type);
  db_write_image_1D((void*)SEGMENT, fname, _NX_, _NY_, sizeof(int), 3);
  sprintf(fname, "DO-DIS-TYPE-%d", type);
  db_write_image_1D((void*)DISTANCE, fname, _NX_, _NY_, sizeof(int2u), 12);
  #endif


  /** find lapsing spectra
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

  for (p=0; p<_NC_; p++){

    if (QAI.off[p]) continue;
    if (SEGMENT[p] < 1) continue;
    if ((QAI.shd[p] || QAI.cld[p] > 0) && type != AOD_SHD) continue;
    if (type == AOD_VEG && DISTANCE[p] < 2) continue;
    
    for (b=1, valid=true; b<_NB_; b++){

      if (!atc.aod_bands[b]) continue;
      if (type == AOD_VEG && b == _B_.swir2) continue;

      for (bb=b-1; bb>=0; bb--){
        if (!atc.aod_bands[bb]) continue;
        if (TOA[b][p] > TOA[bb][p]){ valid = false; break;}
      }
      
      if (!valid) break;
    }

    LAPSE[p] = valid;

  }

  #ifdef DEBUG
  sprintf(fname, "DO-LAPSE-TYPE-%d", type);
  db_write_image_1D((void*)LAPSE, fname, _NX_, _NY_, sizeof(bool), 1);
  #endif


  /** scene average: global environment
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

  for (p=0; p<_NC_; p++){

    if (QAI.off[p] || QAI.shd[p] || QAI.cld[p] > 0) continue;

    for (b=0; b<_NB_; b++){
      if (!atc.aod_bands[b]) continue;
      esum[b] += TOA[b][p]/_RS_;
    }

    en++;

  }

  if (en > 0){
    for (b=0; b<_NB_; b++) etoa[b] = (float)(esum[b]/en);
  } else {
    for (b=0; b<_NB_; b++) etoa[b] = 0.0;
  }

  #ifdef DEBUG
  print_fvector(etoa, "scene environment", _NB_, 1, 4);
  #endif


  /** allocate and initialize darkobject struct
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
  alloc((void**)&dobj, num, sizeof(darkobj_t));
  for (o=0; o<num; o++){
    dobj[o].imin = _NY_; dobj[o].jmin = _NX_;
    alloc((void**)&dobj[o].ttoa, _NB_, sizeof(float));
    alloc((void**)&dobj[o].etoa, _NB_, sizeof(float));
    alloc((void**)&dobj[o].aod,  _NB_, sizeof(float));
    alloc((void**)&dobj[o].est,  _NB_, sizeof(float));
  }

  alloc((void**)&zsum, num, sizeof(double));
  alloc((void**)&csum, num, sizeof(double));
  alloc((void**)&ssum, num, sizeof(double));


  /** extract information, 1st pass, loop over image
  +++ determine bounding box, object size, mean illumin. angle, elevation
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
  
  for (i=0, p=0; i<_NY_; i++){
  for (j=0; j<_NX_; j++, p++){

    if (QAI.off[p]) continue;
    if ((o = SEGMENT[p]-1) < 0) continue;
    if ((QAI.shd[p] || QAI.cld[p] > 0) && type != AOD_SHD) continue;
    if (!LAPSE[p]) continue;

    // sum for illumin. angle, sky view and elevation
    csum[o] += TOP.ill[p]/10000.0;
    ssum[o] += TOP.sky[p]/10000.0;
    //zsum[o] += (TOP.dem[p]*0.1+atc.dem.cmin+0.05);
    zsum[o] += (TOP.dem[p]*atc.dem.step+atc.dem.min+atc.dem.step/2);

    // object radius
    if (DISTANCE[p] > dobj[o].r) dobj[o].r = DISTANCE[p];

    // object location
    if (i > dobj[o].imax) dobj[o].imax = i;
    if (j > dobj[o].jmax) dobj[o].jmax = j;
    if (i < dobj[o].imin) dobj[o].imin = i;
    if (j < dobj[o].jmin) dobj[o].jmin = j;

    // pixel number
    dobj[o].k++;

  }
  }


  /** extract information, 2nd pass, loop over objects
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

  for (o=0; o<num; o++){

    // ignore small and narrow objects
    if (dobj[o].k < 10){
      dobj[o].valid = false;
      continue;
    } else if (dobj[o].r < 3){
      dobj[o].valid = false;
      continue;
    } else {
      dobj[o].valid = true;
    }

    en = tn = 0.0;
    for (b=0; b<_NB_; b++) tsum[b] = 0.0;
    for (b=0; b<_NB_; b++) esum[b] = 0.0;


    // use 3x object radius for environment reflectance
    for (i=dobj[o].imin-3*dobj[o].r; i<=dobj[o].imax+3*dobj[o].r; i++){
    for (j=dobj[o].jmin-3*dobj[o].r; j<=dobj[o].jmax+3*dobj[o].r; j++){
      
      if (i < 0 || j < 0 || i >= _NY_ || j >= _NX_) continue;
      p = i*_NX_+j;

      if (QAI.off[p]) continue;
      if ((QAI.shd[p] || QAI.cld[p] > 0) && type != AOD_SHD) continue;

      // no target --> environment reflectance
      if (SEGMENT[p] == 0){

        for (b=0; b<_NB_; b++){
          if (!atc.aod_bands[b]) continue;
          esum[b] += TOA[b][p]/_RS_;
        }
        en++;

      // if target and pixel is lapsing --> target reflectance
      } else if (SEGMENT[p]-1 == o && LAPSE[p]){

        for (b=0; b<_NB_; b++){
          if (!atc.aod_bands[b]) continue;
          tsum[b] += TOA[b][p]/_RS_;
        }
        tn++;

      }

    }
    }

    // if less than 10 pixels were valid targets, skip
    if (tn < 10){
      dobj[o].valid = false;
      continue;
    }

    // if less than 10 pixels were valid environment, use global one
    if (en < 10){
      for (b=0; b<_NB_; b++) esum[b] = etoa[b];
      en = 1;
    }

    // mean illumination angle and elevation
    dobj[o].cosi = (float)(csum[o]/dobj[o].k);
    dobj[o].sky  = (float)(ssum[o]/dobj[o].k);
    dobj[o].z    = (float)(zsum[o]/dobj[o].k);

    // AOD/MOD elevation correction factors
    od_elev_factor(dobj[o].z, &dobj[o].Ha, &dobj[o].Hr);

    // target centroid
    dobj[o].i = (int)((dobj[o].imax+dobj[o].imin)/2);
    dobj[o].j = (int)((dobj[o].jmax+dobj[o].jmin)/2);

    // target centroid in coarse grid
    dobj[o].gx = tocg(dobj[o].j);
    dobj[o].gy = tocg(dobj[o].i);
    dobj[o].g = dobj[o].gy*atc.ngx+dobj[o].gx;

    // target      reflectance
    // environment reflectance
    for (b=0; b<_NB_; b++){
      if (!atc.aod_bands[b]) continue;
      dobj[o].etoa[b] = (float)(esum[b]/en);
      dobj[o].ttoa[b] = (float)(tsum[b]/tn);
    }

  }


  #ifdef DEBUG
  int1u *DARKTARGET;
  alloc((void**)&DARKTARGET, _NC_, sizeof(int1u));
  sprintf(fname, "DO-TARGETS-TYPE-%d", type);
  for (p=0; p<_NC_; p++){
    if (QAI.off[p]) continue;
    if ((o = SEGMENT[p]-1) < 0) continue;
    if (!dobj[o].valid){ DARKTARGET[p] = 10; continue;}
    if (!LAPSE[p]){ DARKTARGET[p] = 100; continue;}
    if ((QAI.shd[p] || QAI.cld[p] > 0) && type != AOD_SHD){ DARKTARGET[p] = 50; continue;}
    DARKTARGET[p] = 255;
  }
  db_write_image_1D((void**)DARKTARGET, fname, _NX_, _NY_, sizeof(int1u), 1);
  free((void*)DARKTARGET);
  #endif

  // clean
  free((void*)LAPSE); 
  free((void*)SEGMENT); free((void*)DISTANCE);
  free((void*)tsum); free((void*)esum); free((void*)etoa);
  free((void*)csum); free((void*)zsum); free((void*)ssum);


  *DOBJ = dobj;
  return num;
}


/** estimate AOD from dark targets
+++ This function attempts to estimates AOD over each valid dark target.
+++ This is done on the object-level. A reference library based method is
+++ used, where AOD is found by comparing the target TOA reflectance with
+++ the atmospherically contaminated reference spectra. The contamination
+++ is increased iteratively. The AOD is propagated through the spectrum
+++ by either fitting a 2nd order polynomial or a linear function to the
+++ logarithm of AOD and wavelengths. The fit with highest R² wins. For 
+++ each coarse grid cell, all successful AOD estimates are averaged using
+++ the square of R² as weight.
--- dobj:   Dark targets (deallocated within)
--- num:    Number of candidate targets
--- cg_aod: AOD map (returned)
--- type:   Target type (water/shadow/vegetation)
+++ Return: Number of successful targets
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int aod_from_target(darkobj_t *dobj, int num, float ***cg_aod, int type){
speclib_t lib;
int b, bb, c, o, g, w, wbest, k = 0;
float **aodest = NULL;
float **aodlog = NULL;
float **aodfit = NULL;
float rsq, rsqbest, mean, vres, vaod;
bool valid;
int fit, fitbest;
float coef[3], coefbest[3];
double **cg_aodsum = NULL;
double  *cg_wsum   = NULL;
float  **cg_aodavg = NULL;
float weight;


  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif
 
  /** allocate memory **/
  alloc_2D((void***)&cg_aodavg, _NB_, atc.ngc, sizeof(float));

  // shadow type is disabled until further notice
  if (type == AOD_SHD){
    *cg_aod = cg_aodavg;
    return 0;
  }

  /** allocate memory **/
  alloc_2D((void***)&cg_aodsum, _NB_, atc.ngc, sizeof(double));
  alloc((void**)&cg_wsum, atc.ngc, sizeof(double));


  /** initialize spectral library **/
  switch (type){
    case AOD_WAT:
      get_water_lib(&lib);
      break;
    case AOD_SHD:
      get_land_lib(&lib);
      for (b=0; b<_NB_; b++) atc.aod_bands[b] = false;
      if (_B_.ultrablue >= 0) atc.aod_bands[_B_.ultrablue] = true;
      atc.aod_bands[_B_.blue] = true;
      atc.aod_bands[_B_.swir2] = true;
      break;
    case AOD_VEG:
      get_veg_lib(&lib);
      for (b=0; b<_NB_; b++) atc.aod_bands[b] = false;
      atc.aod_bands[_B_.blue]  = true;
      atc.aod_bands[_B_.green] = true;
      atc.aod_bands[_B_.red]   = true;
      atc.aod_bands[_B_.swir2] = true;
      break;
  }


  /** allocate memory **/
  alloc_2D((void***)&aodest, lib.n, _NB_, sizeof(float)); // estimated AOD, band-wise
  alloc_2D((void***)&aodlog, lib.n, _NB_, sizeof(float)); // log of estimated AOD
  alloc_2D((void***)&aodfit, lib.n, _NB_, sizeof(float)); // fitted AOD

  #ifdef DEBUG
  printf("estimate aod from type-%d targets\n", type);
  #endif


  /** estimate AOD for every target
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

  for (o=0; o<num; o++){

    if (!dobj[o].valid) continue;

    g = dobj[o].g;
   

    /** estimate band-wise aod that matches best with each pair of target 
    +++ and library reflectance **/
    aod_lib_to_target(dobj[o], lib, type, aodest);

    /** fit aerosol model to each pair of target and library spectrum **/
    for (w=0, wbest=-1, rsqbest=0; w<lib.n; w++){

      /** skip spectrum if any estimated AOD < 0 **/
      for (b=0, valid=true; b<_NB_; b++){
        if (!atc.aod_bands[b]) continue;
        if (aodest[w][b] < 0){ valid = false; break;}
      }
      if (!valid) continue;

      /** if vegetation target and red AOD > blue AOD, skip **/
      if (type == AOD_VEG && aodest[w][_B_.red] > aodest[w][_B_.blue]) continue;

      /** fix SWIR2 to low value (derived from AERONET analysis) **/
      aodest[w][_B_.swir2] = 0.0172;

      /** if blue AOD < assumed SWIR2 AOD, skip **/
      if (aodest[w][_B_.blue] <= aodest[w][_B_.swir2]) continue;

      /** if ultrablue AOD < blue AOD, skip **/
      if (_B_.ultrablue >= 0 && atc.aod_bands[_B_.ultrablue] &&
          aodest[w][_B_.ultrablue] < aodest[w][_B_.blue]) continue;

      /** compute logarithm of aod + mean of AOD over all bands **/
      for (b=0, mean=0; b<_NB_; b++){

        if (!atc.aod_bands[b]) continue;

        // if any AOD < assumed SWIR2 AOD, force all following bands to this value
        if (aodest[w][b] <= aodest[w][_B_.swir2]){
          for (bb=b; bb<_NB_; bb++){
            if (!atc.aod_bands[bb]) continue;
            aodest[w][bb] = aodest[w][_B_.swir2];
          }
        }

        aodlog[w][b] = log(aodest[w][b]);
        mean += aodest[w][b];
      }
      mean /= bsum(atc.aod_bands, _NB_);


      /** fit polynomial AOD model (King and Byrne) **/

      if (type != AOD_VEG){

        fit = 0;
        aod_polynomial_fit(aodlog[w], &coef[0], &coef[1], &coef[2]);
        for (b=0; b<_NB_; b++) aodfit[w][b] = coef[0] * pow(atc.wvl[b], coef[1]+coef[2]*atc.lwvl[b]);

        /** some reasonability checks **/
        valid = true;
        for (b=1, valid = true; b<_NB_; b++){
          if (aodfit[w][b] > aodfit[w][b-1] && 
              aodfit[w][b] > 0.01 &&
              aodfit[w][b-1] > 0.01) valid = false;
          if (aodfit[w][b] < -0.01 &&
              aodfit[w][b-1] < -0.01) valid = false;
        }

        if (coef[0] < 0 || coef[0] > 0.3) valid = false;
        if (coef[1] > 0 || coef[1] < -3)  valid = false;
        if (coef[2] > 0) valid = false;
      
      } else valid = false;


      /** if polynomial fit was not successful or if we have a vegetation 
      +++ target, try to fit linear AOD model (Angstrom) **/

      if (!valid){

        fit = 1; coef[2] = 0;
        aod_linear_fit(aodlog[w], &coef[0], &coef[1]);
        for (b=0; b<_NB_; b++) aodfit[w][b] = coef[0] * pow(atc.wvl[b], coef[1]);

        /** some reasonability checks **/
        valid = true;
        for (b=1, valid = true; b<_NB_; b++){
          if (aodfit[w][b] > aodfit[w][b-1] && 
            aodfit[w][b] > 0.01 &&
            aodfit[w][b-1] > 0.01) valid = false;
          if (aodfit[w][b] < -0.01 &&
            aodfit[w][b-1] < -0.01) valid = false;
        }

        if (coef[0] < 0 || coef[0] > 0.3) valid = false;
        if (coef[1] > 0 || coef[1] < -3)  valid = false;

      }

      /** skip if neither fit was successful **/
      if (!valid) continue;

      /** compute R² of fit, and find best spectrum **/
      for (b=0, vres=0, vaod=0; b<_NB_; b++){
        if (!atc.aod_bands[b]) continue;
        vres += (aodest[w][b]-aodfit[w][b])*(aodest[w][b]-aodfit[w][b]);
        vaod += (aodest[w][b]-mean)*(aodest[w][b]-mean);}
      if ((rsq = 1.0-vres/vaod) > rsqbest){
        rsqbest = rsq;
        wbest = w;
        fitbest = fit;
        for (c=0; c<3; c++) coefbest[c] = coef[c];
      }

    }

    /** if no valid fit, skip **/
    if (wbest < 0 || rsqbest < 0.1){
      dobj[o].valid = false;
      continue;
    }

    /** copy best aod & compute aod at z = scene average **/
    for (b=0; b<_NB_; b++){
      dobj[o].aod[b] = aodfit[wbest][b]/dobj[o].Ha*atc.Ha;
      dobj[o].est[b] = aodest[wbest][b]/dobj[o].Ha*atc.Ha;
    }

    /** copy coefs, rsq, spectrum ID and type of fit **/
    for (c=0; c<3; c++) dobj[o].coef[c] = coefbest[c];
    dobj[o].rsq  = rsqbest;
    dobj[o].lib_id = wbest;
    dobj[o].ang_fit = fitbest;

    /** number of successful targets **/
    k++;


    /** add to coarse grid cell for AOD averaging **/
    weight = dobj[o].rsq*dobj[o].rsq;
    for (b=0; b<_NB_; b++) cg_aodsum[b][g] += dobj[o].aod[b]*weight;
    cg_wsum[g] += weight;


    #ifdef PRINTDARK
    db_print_dark_object(dobj, o, g);
    #endif

  }

  /** compute AOD average for coarse grid cells **/
  for (g=0; g<atc.ngc; g++){
    for (b=0; b<_NB_; b++){
      if (cg_wsum[g] > 0) cg_aodavg[b][g] = (float)(cg_aodsum[b][g]/cg_wsum[g]);
    }
  }


  #ifdef DEBUG
  char fname[MAX_STR_LEN]; sprintf(fname, "CG-AOD-TYPE-%d", type);
  db_write_image_2D((void**)cg_aodavg, fname, _NB_, atc.ngx, atc.ngy, sizeof(float), 4);
  #endif


  /** clean **/
  free_2D((void**)cg_aodsum, _NB_);
  free((void*)cg_wsum);

  free_2D((void**)lib.s,  lib.n);
  free_2D((void**)aodest, lib.n);
  free_2D((void**)aodlog, lib.n);
  free_2D((void**)aodfit, lib.n);

  for (o=0; o<num; o++){
    free((void*)dobj[o].ttoa); free((void*)dobj[o].etoa);
    free((void*)dobj[o].aod);
    free((void*)dobj[o].est);
  }
  free((void*)dobj);

  
  #ifdef CLOCK
  proctime_print("object-based AOD estimation", TIME);
  #endif

  *cg_aod = cg_aodavg;
  return k;
}


/** Compile water library
+++ This function compiles the water library used for estimating AOD
--- lib:    Spectral library (returned)
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void get_water_lib(speclib_t *lib){
int i, b, b_rsr, w;
double v, s;

  lib->n = _AERO_WATERLIB_DIM_[0];
  alloc_2D((void***)&lib->s, lib->n, _NB_, sizeof(float));

  for (i=0; i<lib->n; i++){
    for (b=0; b<_NB_; b++){
      b_rsr = meta.refcal[b].rsr_band;
      for (w=0, v=0, s=0; w<_AERO_WATERLIB_DIM_[1]; w++){
        v += _RSR_[b_rsr][w]*_AERO_WATERLIB_[i][w];
        s += _RSR_[b_rsr][w];
      }
      if (s > 0) lib->s[i][b] = v/s;
    }
  }

  return;
}


/** Compile land library
+++ This function compiles the land library used for estimating AOD
+++ Note that this functionality is currently unavailable in this version
+++ of FORCE. To be updated.
--- lib:    Spectral library (returned)
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void get_land_lib(speclib_t *lib){
int k = 1;

  lib->n = k;
  alloc_2D((void***)&lib->s, lib->n, _NB_, sizeof(float));

  return;
}


/** Compile vegetation library
+++ This function compiles the vegetation library used for estimating AOD
--- lib:    Spectral library (returned)
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void get_veg_lib(speclib_t *lib){
int k;
float rb, g;


  for (rb=0.01, k=0; rb<=0.05; rb+=0.005){
  for (g=0.01; g<=0.05; g+=0.005){
      k++; 
  }
  }

  lib->n = k;
  alloc_2D((void***)&lib->s, lib->n, _NB_, sizeof(float));

  for (rb=0.01, k=0; rb<=0.05; rb+=0.005){
  for (g=0.01; g<=0.05; g+=0.005){
      lib->s[k][_B_.blue]  = rb;
      lib->s[k][_B_.green] = rb+g;
      lib->s[k][_B_.red]   = rb;
      k++;
  }
  }

  return;
}


/** find AOD by matching library spectra to target TOA reflectance
+++ This function returns the AOD, at which the target TOA reflectance and
+++ the atmospherically contaminated reference spectra match best. AOD is
+++ estimated by increasing the contamination iteratively. The function 
+++ processes one target.
--- dobj:   Dark target
--- lib:    Spectral library
--- type:   Target type (water/shadow/vegetation)
--- aodest: Estimated AOD
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int aod_lib_to_target(darkobj_t dobj, speclib_t lib, int type, float **aodest){
int b, g, w;
float aod, mod, od, tmp;
float T, Ts, Tv, tvd, tsd, tvs, tss, s;
float rhop, rhoe, rhow, rhoc, rhob;
float Fr, Fa, F, km;
float match, *bestmatch;
float iso, vol, geo, brdf;
float E0, E0_, Eg, Egc, k;


  g = dobj.g;

  alloc((void**)&bestmatch, lib.n, sizeof(float));

  /** do separately for each band **/
  for (b=0; b<_NB_; b++){

    if (!atc.aod_bands[b]) continue;

    /** BRDF adjustment for vegetation targets **/
    if (type == AOD_VEG){

      if (b == _B_.blue){
        iso = 0.0774; vol = 0.0372; geo = 0.0079;
      } else if (b == _B_.green){ 
        iso = 0.1306; vol = 0.0580; geo = 0.0178;
      } else if (b == _B_.red){ 
        iso = 0.1690; vol = 0.0574; geo = 0.0227;
      }

      brdf = 
        brdf_forward(45*_D2R_CONV_, 0, 0, iso, vol, geo) / 
        brdf_forward(atc.cg[g].szen, atc.cg[g].vzen,
             atc.cg[g].sazi-atc.cg[g].vazi, iso, vol, geo);
    }


    // initilaize best match
    for (w=0; w<lib.n; w++) bestmatch[w] = INT_MAX;

    // exoatmospheric solar irradiance
    E0  = meta.refcal[b].E0;
    E0_ = E0 * atc.cg[g].Tvo[b]*atc.cg[g].Tso[b];

    // elevation scale factor for MOD
    mod = mod_elev_scale(atc.cg[g].mod[b], atc.cg[g].Hr, dobj.Hr);


    /** iteratively increase AOD **/
    for (aod=-0.2; aod<3; aod+=0.001){

      // total optical depth
      od = optical_depth(aod, atc.cg[g].mod[b]);

      // scattering transmittances
      T = scatt_transmitt(aod, mod, od, atc.cg[g].ms, 
        atc.cg[g].mv, &Ts, &Tv, &tsd, &tss, &tvd, &tvs);

      // global irradiance at ground
      Eg = E0_*tsd + E0_*tss;

      // anisotropy index
      k  = (E0_*tsd)/E0;

      // topographic correction factor
      if (dobj.cosi > 0){
        Egc = E0_*tsd*dobj.cosi/atc.cg[g].ms +
              E0_*tss*k*dobj.cosi/atc.cg[g].ms +
              E0_*tss*(1-k)*dobj.sky;
      } else {
        Egc = E0_*tss*(1-k)*dobj.sky;
      }

      // path reflectance
      rhop = path_ref(aod, mod, od, atc.cg[g].Pa, atc.cg[g].Pr, 
        tsd, tvd, atc.cg[g].ms, atc.cg[g].mv);

      // spherical albedo
      s = sphere_albedo(aod, mod, od);

      // reasonability check
      if (dobj.etoa[b]-rhop < -0.05){
        break;
      }

      // simplified reflectance environment
      tmp = (dobj.etoa[b]-rhop)/atc.cg[g].Tg[b];
      rhoe = tmp / (T + s*tmp);

      // environmental weighting function, F(r)
      km = dobj.r*meta.res/1000.0+meta.res/1000.0/2;
      Fr = env_weight_molecular(km);
      Fa = env_weight_aerosol(km);
      F = env_weight(aod, mod, Fa, Fr);


      /** compute TOA reference reflectance and compare to 
      +++ measured TOA reflectance **/

      for (w=0; w<lib.n; w++){

        // reference reflectance including Fresnel
        if (type == AOD_WAT){
          rhow = (lib.s[w][b] + atc.cg[g].rhof*tss)/Ts;
        } else {
          rhow =  lib.s[w][b];
        }

        // reference reflectance including topography and BRDF
        if (type == AOD_SHD) rhow = rhow*Egc/Eg;
        if (type == AOD_VEG) rhow = rhow*Egc/(brdf*Eg);

        // background reflectance
        rhob = rhow*F+(1-F)*rhoe;

        // computed TOA reflectance
        rhoc = atc.cg[g].Tg[b]*(rhop+(Ts*(tvd*rhow+tvs*rhob))/(1-rhob*s));

        // match similarity
        match = fabs(rhoc-dobj.ttoa[b]);
        if (match < bestmatch[w]){
          bestmatch[w] = match;
          aodest[w][b] = aod;
        }

      }


    }

  }

  free((void*)bestmatch);

  return SUCCESS;
}


/** linear fit between ln(aod) and ln(wvl)
+++ This function fits the Angstrom expression to the AOD spectrum and es-
+++ timates the turbidity coefficient (a0) and Angstrom exponent (a1).
--- logaod: logarithm of estimated AOD
--- a0:     turbidity coefficient (returned)
--- a1:     Angstrom exponent     (returned)
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int aod_linear_fit(float *logaod, float *a0, float *a1){
int b, k;
float xsum = 0, ysum = 0, xm, ym;
float csum = 0, vsum = 0, cv, vr;
float slope, inter;

  // mean
  for (b=0, k=0; b<_NB_; b++){
    if (!atc.aod_bands[b]) continue;
    xsum += atc.lwvl[b];
    ysum += logaod[b];
    k++;
  }
  xm = xsum/k;
  ym = ysum/k;

  //covariance and variance
  for (b=0; b<_NB_; b++){
    if (!atc.aod_bands[b]) continue;
    csum += (atc.lwvl[b]-xm)*(logaod[b]-ym);
    vsum += (atc.lwvl[b]-xm)*(atc.lwvl[b]-xm);
  }
  cv = csum/k;
  vr = vsum/k;

  slope = cv/vr;
  inter = ym - slope*xm;

  *a0 = exp(inter);
  *a1 = slope;

  return SUCCESS;
}


/** 2nd order polynomial fit between ln(aod) and ln(wvl)
+++ This function fits a modified Angstrom expression to the AOD spectrum,
+++ which accounts for curvature in the AOD spectrum. The function estima-
+++ tes the turbidity coefficient (a0), Angstrom exponent (a1) and a term
+++ for describing the curvature (a2).
--- logaod: logarithm of estimated AOD
--- a0:     turbidity coefficient (returned)
--- a1:     Angstrom exponent     (returned)
--- a2:     curvature coefficient (returned)
+++ Return: GSL fit status
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int aod_polynomial_fit(float *logaod, float *a0, float *a1, float *a2){
const gsl_multimin_fminimizer_type *T = gsl_multimin_fminimizer_nmsimplex2;
gsl_multimin_fminimizer *s = NULL;
gsl_vector *ss, *x;
gsl_multimin_function minex_func;
int b;
float *param;
size_t iter = 0;
int status;
double size;

  alloc((void**)&param, _NB_, sizeof(float));

  for (b=0; b<_NB_; b++) param[b] = logaod[b];


  /* Starting point */
  x = gsl_vector_alloc (3);
  gsl_vector_set (x, 0, 0);
  gsl_vector_set (x, 1, 0);
  gsl_vector_set (x, 2, 0);

  /* Set initial step sizes */
  ss = gsl_vector_alloc (3);
  gsl_vector_set (ss, 0, 0.01);
  gsl_vector_set (ss, 1, 0.01);
  gsl_vector_set (ss, 2, 0.02);

  /* Initialize method and iterate */
  minex_func.n = 3;
  minex_func.f = aod_polynomial_minimizer;
  minex_func.params = param;

  s = gsl_multimin_fminimizer_alloc(T, 3);
  gsl_multimin_fminimizer_set(s, &minex_func, x, ss);

  do {
    iter++;
    status = gsl_multimin_fminimizer_iterate(s);

    if (status) break;

    size = gsl_multimin_fminimizer_size(s);
    status = gsl_multimin_test_size(size, 1e-2);

  } while (status == GSL_CONTINUE && iter < 100);

  *a0 = exp(gsl_vector_get(s->x, 0));
  *a1 = gsl_vector_get(s->x, 1);
  *a2 = gsl_vector_get(s->x, 2);

  free((void*)param);
  gsl_vector_free(x);
  gsl_vector_free(ss);
  gsl_multimin_fminimizer_free(s);

  return status;
}


/** Minimizer for 2nd order polynomial fit between ln(aod) and ln(wvl)
+++ This function minimizes the RMSE of the polynomial fit.
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
double aod_polynomial_minimizer(const gsl_vector *v, void *params){
float a0, a1, a2;
int b;
float *p = (float *)params;
double y, tmp = 0;

  a0 = gsl_vector_get(v, 0);
  a1 = gsl_vector_get(v, 1);
  a2 = gsl_vector_get(v, 2);

  for (b=0; b<_NB_; b++){
    if (!atc.aod_bands[b]) continue;
    y = a0 + a1*atc.lwvl[b] + a2*atc.lwvl[b]*atc.lwvl[b];
    tmp += (y-p[b])*(y-p[b]);
  }

  return sqrt(tmp/bsum(atc.aod_bands, _NB_));
}


/** Compile AOD map
+++ This function builds the continuous AOD field at low resolution. The 
+++ sparse AOD maps of the individual estimators are averaged per grid 
+++ cell. Interpolation and lowpass filtering is applied to provide a con-
+++ tinuos field.
--- dark:   dark target container
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void aod_map(dark_t dark){
int b, g;
float gsum, gnum;

  // alocate memory
  alloc_2D((void***)&dark.cg_avg, _NB_, atc.ngc, sizeof(float));

  // compute grid averages
  for (b=0; b<_NB_; b++){
    for (g=0; g<atc.ngc; g++){

      gsum = gnum = 0;

      if (dark.cg_wat[b][g] > 0){
        gsum += dark.cg_wat[b][g]; gnum++;}
      if (dark.cg_shd[b][g] > 0){
        gsum += dark.cg_shd[b][g]; gnum++;}
      if (dark.cg_veg[b][g] > 0){
        gsum += dark.cg_veg[b][g]; gnum++;}

      if (gnum > 0){
        dark.cg_avg[b][g] = gsum/gnum;
      } else {
        atc.cg[g].aod_int = true;
      }

    }
  }


  #ifdef DEBUG
  db_write_image_2D((void**)dark.cg_avg, "CG-AOD-AVG", _NB_, atc.ngx, atc.ngy, sizeof(float), 4);
  #endif
  
  // interpolate AOD map
  if (interpolate_cg_aod(dark.cg_avg) != SUCCESS){
    printf("interpolating between AOD failed.\n"); exit(1);}

  free_2D((void**)dark.cg_avg, _NB_);

  return;
}


/** Interpolate sparse AOD map
+++ This function interpolates the average AOD maps that may contain no-
+++ data values. Inverse Distance Weighting is used, this reduces extra-
+++ polation artifacts. Lowpass filtering is applied to the interpolation
+++ to further smooth the map.
--- cg_aod: averaged AOD map
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int interpolate_cg_aod(float **cg_aod){
int b, gy, gx, g, yy, xx;
int ng = 0, k = 0;
GDALGridAlgorithm eAlgorithm = GGA_InverseDistanceToAPower;
void *pOptions = NULL;
GDALDataType eOutputType = GDT_Float64;
double  sum,  num;
double gsum, gnum;
double  *aod_x   = NULL;
double  *aod_y   = NULL;
double **aod_z   = NULL;
double *interpol = NULL;

  GDALAllRegister();
  ParseAlgorithmAndOptions(szAlgNameInvDist, &eAlgorithm, &pOptions);

  // number of cells with AOD estimate
  for (g=0; g<atc.ngc; g++){
    if (cg_aod[_B_.green][g] > 0) ng++;
  }

  // allocate memory
  alloc((void**)&aod_x, ng, sizeof(double));
  alloc((void**)&aod_y, ng, sizeof(double));
  alloc_2D((void***)&aod_z, _NB_, ng, sizeof(double));
  alloc((void**)&interpol, atc.ngc, sizeof(double));


  // copy AOD estimates to array
  for (gy=0, g=0; gy<atc.ngy; gy++){
  for (gx=0; gx<atc.ngx; gx++, g++){
    if (cg_aod[_B_.green][g] > 0){
      for (b=0; b<_NB_; b++){
        aod_z[b][k] = cg_aod[b][g];
      }
      aod_x[k] = gx;
      aod_y[k] = gy;
      k++;
    }
  }
  }


  // interpolate and smooth AOD
  for (b=0; b<_NB_; b++){

    // compute interpolation
    GDALGridCreate(eAlgorithm, pOptions, ng, aod_x, aod_y, aod_z[b],
      0, atc.ngx-1, 0, atc.ngy-1, atc.ngx, atc.ngy, 
      eOutputType, (void*)interpol, NULL, NULL);

    sum = num = 0;

    // lowpass filter
    for (gy=0, g=0; gy<atc.ngy; gy++){
    for (gx=0; gx<atc.ngx; gx++, g++){

      if (atc.cg[g].vzen < 0) continue;

      gsum = gnum = 0;
      for (yy=-1; yy<=1; yy++){
      for (xx=-1; xx<=1; xx++){
        if (gy+yy < 0 || gx+xx < 0 || gy+yy >= atc.ngy || gx+xx >= atc.ngx) continue;
        gsum += interpol[(gy+yy)*atc.ngx + (gx+xx)];
        gnum++;
      }
      }
      
      if (gnum > 0) atc.cg[g].aod[b] = (float)(gsum/gnum); else atc.cg[g].aod[b] = 0.0;
      if (atc.cg[g].aod[b] > 0){
        sum += atc.cg[g].aod[b];
        num++;
      }

    }
    }

    if (num > 0) atc.aod[b] = (float)(sum/num); else atc.aod[b] = 0.0;

  }

  // clean
  CPLFree(pOptions);
  free((void*)aod_x);
  free((void*)aod_y);
  free_2D((void**)aod_z, _NB_);
  free((void*)interpol);

  return SUCCESS;
}


/** Compile a usable set of AOD values
+++ This function is the entry point for AOD estimation. The fallback is
+++ either set to a global constant or to an externally provided spectrum.
+++ The Dark Object Persistance is compiled if demanded; if not, the comp-
+++ lete image is declared persitent. Dark targets are extracted from the
+++ image, and AOD is inferred on a per-object basis. The objects are av-
+++ eraged, and interpolated to produce an AOD map. Finally AOD is logged.
--- TOA:    TOA reflectance
--- QAI:    Quality Assurance Information
--- TOP:    Topographic Derivatives
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void compile_aod(int2s **TOA, qai_t QAI, top_t TOP){
int p, b;
int2u *DOP;
dark_t dark;

  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif


  /** read external AOD file
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
  if (aodfileread() != SUCCESS) exit(1);


  /** estimate AOD from image or use fallback
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
  if (par.doaod){

    /** use all visible bands + swir2 for AOD estimation **/
    for (b=0; b<_NB_; b++){
      if (atc.wvl[b] < 0.7 || b == _B_.swir2) atc.aod_bands[b] = true;
    }


    /** read DOP or declare that everything is persistent
    +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
    par.dododb = false;
    if (par.dododb){
      if (dark_object_persistency(&DOP) == FAILURE){
        printf("Could not compile DOP!\n"); exit(1);}
    } else {
      alloc((void**)&DOP, _NC_, sizeof(int2u));
      for (p=0; p<_NC_; p++) DOP[p] = 10000;
    }


    /** extract dark targets from image and tabulate necessary information
    +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
    dark.kwat = extract_dark_target(TOA, QAI, DOP, TOP, AOD_WAT, &dark.wat);
    dark.kshd = extract_dark_target(TOA, QAI, DOP, TOP, AOD_SHD, &dark.shd);
    dark.kveg = extract_dark_target(TOA, QAI, DOP, TOP, AOD_VEG, &dark.veg);
    free((void*)DOP);


    /** object-based estimation of AOD
    +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
    dark.nwat = aod_from_target(dark.wat, dark.kwat, &dark.cg_wat, AOD_WAT);
    dark.nshd = aod_from_target(dark.shd, dark.kshd, &dark.cg_shd, AOD_SHD);
    dark.nveg = aod_from_target(dark.veg, dark.kveg, &dark.cg_veg, AOD_VEG);


    /** compile AOD map or use fallback
    +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
    if ((dark.nwat+dark.nshd+dark.nveg) > 0){
      aod_map(dark);
      atc.aodmap = true;
    } else {
      for (b=0; b<_NB_; b++) atc.aod[b] = atc.aod_ext[b];
      atc.aodmap = false;
    }

    free_2D((void**)dark.cg_wat, _NB_);
    free_2D((void**)dark.cg_shd, _NB_);
    free_2D((void**)dark.cg_veg, _NB_);

  } else {
    for (b=0; b<_NB_; b++) atc.aod[b] = atc.aod_ext[b];
  }

  /** print for logfile **/
  printf("AOD: %06.4f. # of targets: %d/%d. ", 
    atc.aod[_B_.green], dark.nwat, dark.nveg);

  #ifdef CLOCK
  proctime_print("AOD estimation", TIME);
  #endif

  return;
}

