/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
This file contains functions for atmospheric correction
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#include "force.h"


/** Initialize atmospheric parameters
+++ This function initializes parameters used for atmospheric correction.
+++ This includes precomputing TTHG parameters, environmental weighting,
+++ absorption coefficients, wavelengths and allocating memory.
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void atc_init(){
int r = 1;

    /** initialize Henyey-Greenstein parameters for faster computation
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
  atc.tthg.g1 = 0.801; atc.tthg.g2 = 0.580; atc.tthg.alpha = 0.948;
  atc.tthg.hg1  = (1.0-atc.tthg.g1*atc.tthg.g1)*atc.tthg.alpha;
  atc.tthg.hg2  = (1.0-atc.tthg.g2*atc.tthg.g2)*(1.0-atc.tthg.alpha);
  atc.tthg.hg3  = 1.0+atc.tthg.g1*atc.tthg.g1; atc.tthg.hg4 = 1.0+atc.tthg.g2*atc.tthg.g2;
  atc.tthg.hg5  = 2.0*atc.tthg.g1; atc.tthg.hg6 = 2.0*atc.tthg.g2;
  atc.tthg.sob1 = 3.0*(atc.tthg.alpha*(atc.tthg.g1+atc.tthg.g2)-atc.tthg.g2);
 

  /** initialize environmental weighting functions for MOD/AOD
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
  atc.km = r*(meta.res/1000.0)+0.5*meta.res/1000.0;
  atc.Fr = 1.0-(0.930*exp(-0.8*atc.km)+0.070*exp(-1.10*atc.km));
  atc.Fa = 1.0-(0.375*exp(-0.2*atc.km)+0.625*exp(-1.83*atc.km));


  /** allocate band-dependent variables
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
  alloc((void**)&atc.wvl,       _NB_, sizeof(float)); 
  alloc((void**)&atc.lwvl,      _NB_, sizeof(float));
  alloc((void**)&atc.lwvl2,     _NB_, sizeof(float));
  alloc((void**)&atc.od,        _NB_, sizeof(float));
  alloc((void**)&atc.mod,       _NB_, sizeof(float));
  alloc((void**)&atc.aod,       _NB_, sizeof(float));
  alloc((void**)&atc.aod_ext,   _NB_, sizeof(float));
  alloc((void**)&atc.aod_bands, _NB_, sizeof(bool));


  /** initialize sensor specifics
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
  if (wavelengths() == FAILURE){ printf("error in getting wavelengths!\n"); exit(1);}

  return;
};


/** Deallocate atmospheric parameters
+++ This function deallocates parameters used for atmospheric correction.
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void atc_free(){

  free((void*)atc.wvl); free((void*)atc.lwvl); free((void*)atc.lwvl2); 
  free((void*)atc.od);  free((void*)atc.mod); free((void*)atc.aod); 
  free((void*)atc.aod_ext);  
  free((void*)atc.aod_bands);

  return;  
}


/** Compile coarse grid
+++ This function compiles the coarse grid used for atmospheric correction
--- GRD:    coarse grid (returned)
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void cggrid(int2u **GRD){
int i, j, p, g;
int2u *grd;

  // resolution and number of cells in x- and y-direction
  atc.gres = par.svgrid;
  atc.ngx  = ceil(_NX_/(float)atc.gres);
  atc.ngy  = ceil(_NY_/(float)atc.gres);
  atc.ngc  = atc.ngx*atc.ngy;
  
  #ifdef DEBUG
  printf("\nnumber of coarse grid points: %d, ngx: %d, ngy: %d, res: %d\n", atc.ngc, atc.ngx, atc.ngy, atc.gres);
  #endif

  if (atc.ngc > USHRT_MAX){
    printf("too many coarse grid cells.\n"); exit(1);}

  // allocate atmospheric parameters for coarse grid cells
  alloc((void**)&atc.cg, atc.ngc, sizeof(grid_t));
  for (g=0; g<atc.ngc; g++){
    atc.cg[g].vzen = _NODATA_;
    alloc((void**)&atc.cg[g].mod,   _NB_, sizeof(float));
    alloc((void**)&atc.cg[g].aod,   _NB_, sizeof(float));
    alloc((void**)&atc.cg[g].Tvw,   _NB_, sizeof(float));
    alloc((void**)&atc.cg[g].Tsw,   _NB_, sizeof(float));
    alloc((void**)&atc.cg[g].Tvo,   _NB_, sizeof(float));
    alloc((void**)&atc.cg[g].Tso,   _NB_, sizeof(float));
    alloc((void**)&atc.cg[g].Tg,    _NB_, sizeof(float));
    alloc((void**)&atc.cg[g].brdf,  _NB_, sizeof(float));
  }

  // coarse grid dictionary
  alloc((void**)&grd, _NC_, sizeof(int2u));

  for (i=0, p=0; i<_NY_; i++){
  for (j=0; j<_NX_; j++, p++){
    grd[p] = floor(i/atc.gres)*atc.ngx + floor(j/atc.gres);
  }
  }

  #ifdef DEBUG
  db_write_image_1D((void**)grd, "GRID", _NX_, _NY_, sizeof(int2u), 12);
  #endif

  *GRD = grd;
  return;
}


/** Deallocate atmospheric parameters for coarse grid cells
+++ This function deallocates atmospheric parameters for coarse grid cells
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void cgfree(){
int g;

  for (g=0; g<atc.ngc; g++){
    free((void*)atc.cg[g].mod);
    free((void*)atc.cg[g].aod);
    free((void*)atc.cg[g].Tvw);
    free((void*)atc.cg[g].Tsw);
    free((void*)atc.cg[g].Tvo);
    free((void*)atc.cg[g].Tso);
    free((void*)atc.cg[g].Tg);
    free((void*)atc.cg[g].brdf);
  }
  free((void*)atc.cg);

  return;
}


/** Compute sun positions and view geometry
+++ This function computes sun positions (+cos/sin/tan), and view angles
--- QAI:    Quality Assurance Information
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int sun_target_view(qai_t QAI){
float zen, azi;
int g, gy, gx;


  // average satellite height (m)
  if (meta.mission == MISSION_LANDSAT){
    atc.view.H = 705000.0;
  } else if (meta.mission == MISSION_SENTINEL2){
    atc.view.H = 786000.0;
  } else { printf("unknown mission.\n"); exit(1);}
  atc.view.H2 = atc.view.H*atc.view.H;
  
  // compute approx. view geometry for Landsat
  if (meta.mission == MISSION_LANDSAT) viewgeo(QAI);

  atc.cosszen[0] = atc.cosvzen[0] = INT_MAX;
  atc.cosszen[1] = atc.cosvzen[1] = INT_MIN;

  for (gy=0, g=0; gy<atc.ngy; gy++){
  for (gx=0; gx<atc.ngx; gx++, g++){

    // utm coordinate
    atc.cg[g].mapy = meta.map_ul.y-gy*atc.gres*meta.res;
    atc.cg[g].mapx = meta.map_ul.x+gx*atc.gres*meta.res;

    // geo coordinate
    reproj_point(atc.cg[g].mapx, atc.cg[g].mapy, 
                 &atc.cg[g].geox, &atc.cg[g].geoy,
                 meta.css_map, meta.css_geo);

    // calculate sun angles
    sunpos(atc.cg[g].geoy, atc.cg[g].geox, meta.date.year, meta.date.month, meta.date.day,
      meta.date.hh, meta.date.mm, meta.date.ss, meta.date.tz,
      &zen, &azi);
    atc.cg[g].szen = deg2rad(zen);
    atc.cg[g].sazi = deg2rad(azi);
    atc.cg[g].cosszen = cos(atc.cg[g].szen);
    atc.cg[g].sinszen = sin(atc.cg[g].szen);
    atc.cg[g].tanszen = tan(atc.cg[g].szen);
    atc.cg[g].cossazi = cos(atc.cg[g].sazi);
    atc.cg[g].sinsazi = sin(atc.cg[g].sazi);
    atc.cg[g].tansazi = tan(atc.cg[g].sazi);

    // satellite view geometry
    view_angle(gx, gy, g, QAI);

    if (atc.cg[g].vzen == _NODATA_) continue;

    if (atc.cg[g].cosszen < atc.cosszen[0]) atc.cosszen[0] = atc.cg[g].cosszen;
    if (atc.cg[g].cosvzen < atc.cosvzen[0]) atc.cosvzen[0] = atc.cg[g].cosvzen;
    if (atc.cg[g].cosszen > atc.cosszen[1]) atc.cosszen[1] = atc.cg[g].cosszen;
    if (atc.cg[g].cosvzen > atc.cosvzen[1]) atc.cosvzen[1] = atc.cg[g].cosvzen;

  }
  }

  #ifdef DEBUG
  printf("min/max cosszen: %.2f/%.2f\n", atc.cosszen[0], atc.cosszen[1]);
  printf("min/max cosvzen: %.2f/%.2f\n", atc.cosvzen[0], atc.cosvzen[1]);
  #endif

  return SUCCESS;
}


/** Compute approximate view geometry
+++ This function computes the approximate view geometry for Landsat
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void viewgeo(qai_t QAI){
int i, j, p;
int uly=-1, ury, lly, lry=-1;
int ulx, urx=-1, llx=_NC_, lrx;
int uy, ly, ux, lx;
float muly, mury, mlly, mlry;
float mulx, murx, mllx, mlrx;
float muy, mly, mux, mlx;
float guly, gury, glly, glry;
float gulx, gurx, gllx, glrx;
float guy, gly, gux, glx;
float dx, dy;

  
  // find corners of image
  for (i=0, p=0; i<_NY_; i++){
  for (j=0; j<_NX_; j++, p++){
    if (QAI.off[p]) continue;
    if (uly == -1){ uly = i; ulx = j;}
    if (j > urx){ ury = i; urx = j;}
    if (j <= llx){ lly = i; llx = j;}
    if (i >= lry){ lry = i; lrx = j;}
  }
  }

  // midpoints of top and bottom
  ux = (ulx+urx)/2; uy = (uly+ury)/2;
  lx = (llx+lrx)/2; ly = (lly+lry)/2;
  
  // corners and midpoints in utm coordinates
  mulx = meta.map_ul.x+ulx*meta.res; muly = meta.map_ul.y-uly*meta.res;
  mux  = meta.map_ul.x+ux *meta.res; muy  = meta.map_ul.y-uy *meta.res;
  murx = meta.map_ul.x+urx*meta.res; mury = meta.map_ul.y-ury*meta.res;
  mllx = meta.map_ul.x+llx*meta.res; mlly = meta.map_ul.y-lly*meta.res;
  mlx  = meta.map_ul.x+lx *meta.res; mly  = meta.map_ul.y-ly *meta.res;
  mlrx = meta.map_ul.x+lrx*meta.res; mlry = meta.map_ul.y-lry*meta.res;

  // geo coordinate
  reproj_point(mulx, muly, &gulx, &guly, meta.css_map, meta.css_geo);
  reproj_point(mux,  muy,  &gux,  &guy,  meta.css_map, meta.css_geo);
  reproj_point(murx, mury, &gurx, &gury, meta.css_map, meta.css_geo);
  reproj_point(mllx, mlly, &gllx, &glly, meta.css_map, meta.css_geo);
  reproj_point(mlx,  mly,  &glx,  &gly,  meta.css_map, meta.css_geo);
  reproj_point(mlrx, mlry, &glrx, &glry, meta.css_map, meta.css_geo);

  // difference from top to bottom
  dx = gux-glx; dy = gly-guy;

  // angle of nadir line, measured clockwise from north or south
  atc.view.geo_angle_nadir = atan2(dy,dx) + M_PI/2;


  // cartesian definition of nadir line: ax + by + c = 0
  atc.view.a = uy-ly;
  atc.view.b = lx-ux;
  atc.view.c = ly*ux-lx*uy;

  // the perpendicular distance of a point to the nadir line is given by
  // |ax+by+c|/sqrt(aa+bb)
  atc.view.ab = sqrt(atc.view.a*atc.view.a+atc.view.b*atc.view.b);


  #ifdef DEBUG
  printf("\nimage corners:\n");
  printf("UL (%d/%d), UM (%d/%d), UR (%d/%d)\n", ulx, uly, ux, uy, urx, ury);
  printf("LL (%d/%d), LM (%d/%d), LR (%d/%d)\n", llx, lly, lx, ly, lrx, lry);
  printf("UL (%05.2f/%05.2f), UM (%05.2f/%05.2f), UR (%05.2f/%05.2f)\n", gulx, guly, gux, guy, gurx, gury);
  printf("LL (%05.2f/%05.2f), LM (%05.2f/%05.2f), LR (%05.2f/%05.2f)\n", gllx, glly, glx, gly, glrx, glry);
  printf("\nestimated view geometry in image coordinates:\n");
  printf(" a: %.0f, b: %.0f, c: %.0f\n", atc.view.a, atc.view.b, atc.view.c);
  printf("\nestimated view geometry in geographic coordinates:\n");
  printf(" angle_nadir: %.2f\n", rad2deg(atc.view.geo_angle_nadir));
  printf("\nsensor height parameters:\n");
  printf(" H: %.0f, H2: %.0f\n", atc.view.H, atc.view.H2);
  #endif

  return;
}


/** Compute view angle
+++ This function computes the view angles
+++ Return: void
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void view_angle(int gx, int gy, int g, qai_t QAI){
int i, j, p;
float dist;


  if (meta.mission == MISSION_LANDSAT){

    // distance from nadir line in pixels
    dist = (atc.view.a*gx*atc.gres+atc.view.b*gy*atc.gres+atc.view.c)/atc.view.ab;

    // view zenith
    atc.cg[g].vzen = acos(atc.view.H/sqrt(atc.view.H2+900*dist*dist));
    if (atc.cg[g].vzen != atc.cg[g].vzen) atc.cg[g].vzen = 0; // if vzen is NaN

    // view azimuth
    if (dist > 0){ // left of nadir, i.e. right looking
      atc.cg[g].vazi = atc.view.geo_angle_nadir+M_PI/2;
    } else {       // right of nadir, i.e. left looking
      atc.cg[g].vazi = atc.view.geo_angle_nadir-M_PI/2;
      while (atc.cg[g].vazi < 0) atc.cg[g].vazi += 2*M_PI;
    }

  } else if (meta.mission == MISSION_SENTINEL2){

    if (meta.s2.vzen[gy][gx] == _NODATA_ || meta.s2.vazi[gy][gx] == _NODATA_){

      atc.cg[g].vzen = _NODATA_; atc.cg[g].vazi = _NODATA_;

      for (i=gy*atc.gres; i<((gy+1)*atc.gres); i++){
      for (j=gx*atc.gres; j<((gx+1)*atc.gres); j++){

      if (i > _NY_-1) continue;
      if (j > _NX_-1) continue;
        p = i*_NX_+j;
        QAI.off[p] = true;
      }
      }

    } else {
      atc.cg[g].vzen = deg2rad(meta.s2.vzen[gy][gx]);
      atc.cg[g].vazi = deg2rad(meta.s2.vazi[gy][gx]);
    }

  }
  

  atc.cg[g].cosvzen = cos(atc.cg[g].vzen);
  atc.cg[g].sinvzen = sin(atc.cg[g].vzen);
  atc.cg[g].tanvzen = tan(atc.cg[g].vzen);
  atc.cg[g].cosvazi = cos(atc.cg[g].vazi);
  atc.cg[g].sinvazi = sin(atc.cg[g].vazi);
  atc.cg[g].tanvazi = tan(atc.cg[g].vazi);

  // fix viewing angles
  //*vzen = 0; *cosvzen = 1;
  //*vazi = deg2rad(98.19998);

  return;
}


/** Smooth coarse resolution atmospheric parameters
+++ This function interpolates the coarse resolution atmospheric parame-
+++ ters that are needed to convert TOA to BOA. This is done to avoid 
+++ sharp edges between coarse grid cells.
--- atc_2:  smoothed atmospheric parameters
--- i:      y-coordinate of pixel
--- j:      x-coordinate of pixel
--- z:      binned elevation
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void smooth_atc(zgrid_t *atc_2, int i, int j, int z){
int b;
int   gx, gy;
float cx, cy;
int   gleft, gright, gtop, gdown;
float cleft, cright, ctop, cdown;
int   gul, gur, gll, glr;
float wul, wur, wll, wlr;
float wtop, wdown;
bool vtop = true, vdown = true;

  // coarse grid cell
  gy = tocg(i);
  gx = tocg(j);

  cy = gy*par.svgrid + par.svgrid/2;
  cx = gx*par.svgrid + par.svgrid/2;

  if (j < cx){ gleft = gx-1; gright = gx;} else { gleft = gx; gright = gx+1;}
  if (i < cy){ gtop  = gy-1; gdown  = gy;} else { gtop  = gy; gdown  = gy+1;}

  if (gleft < 0) gleft = gx;
  if (gtop < 0)  gtop  = gy;
  if (gright >= atc.ngx) gright = gx;
  if (gdown >= atc.ngy)  gdown  = gy;

  cleft  = gleft*par.svgrid  + par.svgrid/2;
  cright = gright*par.svgrid + par.svgrid/2;
  ctop   = gtop*par.svgrid   + par.svgrid/2;
  cdown  = gdown*par.svgrid  + par.svgrid/2;

  gul = gtop*atc.ngx  + gleft;
  gur = gtop*atc.ngx  + gright;
  gll = gdown*atc.ngx + gleft;
  glr = gdown*atc.ngx + gright;

  if (gleft == gright){
    wul = 0.5; wur = 0.5;
  } else if (atc.cg[gul].vzen > 0 && atc.cg[gur].vzen > 0){
    wul = (cright-j)/(cright-cleft);
    wur = (j-cleft)/(cright-cleft);
  } else if (atc.cg[gul].vzen > 0 && atc.cg[gur].vzen < 0){
    wul = 1; wur = 0;
  } else if (atc.cg[gul].vzen < 0 && atc.cg[gur].vzen > 0){
    wul = 0; wur = 1;
  } else {
    wul = 0; wur = 0;
    vtop = false;
  }

  if (gleft == gright){
    wll = 0.5; wlr = 0.5;
  } else if (atc.cg[gll].vzen > 0 && atc.cg[glr].vzen > 0){
    wll = (cright-j)/(cright-cleft);
    wlr = (j-cleft)/(cright-cleft);
  } else if (atc.cg[gll].vzen > 0 && atc.cg[glr].vzen < 0){
    wll = 1; wlr = 0;
  } else if (atc.cg[gll].vzen < 0 && atc.cg[glr].vzen > 0){
    wll = 0; wlr = 1;
  } else {
    wll = 0; wlr = 0;
    vdown = false;
  }

  if (gtop == gdown){
    wtop = 0.5; wdown = 0.5;
  } else if (vtop && vdown){
    wtop  = (cdown-i)/(cdown-ctop);
    wdown = (i-ctop)/(cdown-ctop);
  } else if (vtop){
    wtop = 1; wdown = 0;
  } else if (vdown){
    wtop = 0; wdown = 1;
  } else {
    wtop = 0; wdown = 0;
  }


  for (b=0; b<_NB_; b++){

    if (!par.use.refbands[b]) continue;

    // interpolated
    atc_2->aod[b]   = wtop*(wul*atc.cg[gul].z[z].aod[b]   + wur*atc.cg[gur].z[z].aod[b]  ) + wdown*(wll*atc.cg[gll].z[z].aod[b]   + wlr*atc.cg[glr].z[z].aod[b]  );
    atc_2->rho_p[b] = wtop*(wul*atc.cg[gul].z[z].rho_p[b] + wur*atc.cg[gur].z[z].rho_p[b]) + wdown*(wll*atc.cg[gll].z[z].rho_p[b] + wlr*atc.cg[glr].z[z].rho_p[b]);
    atc_2->s[b]     = wtop*(wul*atc.cg[gul].z[z].s[b]     + wur*atc.cg[gur].z[z].s[b]    ) + wdown*(wll*atc.cg[gll].z[z].s[b]     + wlr*atc.cg[glr].z[z].s[b]    );
    atc_2->Ts[b]    = wtop*(wul*atc.cg[gul].z[z].Ts[b]    + wur*atc.cg[gur].z[z].Ts[b]   ) + wdown*(wll*atc.cg[gll].z[z].Ts[b]    + wlr*atc.cg[glr].z[z].Ts[b]   );
    atc_2->tss[b]   = wtop*(wul*atc.cg[gul].z[z].tss[b]   + wur*atc.cg[gur].z[z].tss[b]  ) + wdown*(wll*atc.cg[gll].z[z].tss[b]   + wlr*atc.cg[glr].z[z].tss[b]  );
    atc_2->tsd[b]   = wtop*(wul*atc.cg[gul].z[z].tsd[b]   + wur*atc.cg[gur].z[z].tsd[b]  ) + wdown*(wll*atc.cg[gll].z[z].tsd[b]   + wlr*atc.cg[glr].z[z].tsd[b]  );
    atc_2->tvs[b]   = wtop*(wul*atc.cg[gul].z[z].tvs[b]   + wur*atc.cg[gur].z[z].tvs[b]  ) + wdown*(wll*atc.cg[gll].z[z].tvs[b]   + wlr*atc.cg[glr].z[z].tvs[b]  );
    atc_2->tvd[b]   = wtop*(wul*atc.cg[gul].z[z].tvd[b]   + wur*atc.cg[gur].z[z].tvd[b]  ) + wdown*(wll*atc.cg[gll].z[z].tvd[b]   + wlr*atc.cg[glr].z[z].tvd[b]  );
    atc_2->T[b]     = wtop*(wul*atc.cg[gul].z[z].T[b]     + wur*atc.cg[gur].z[z].T[b]    ) + wdown*(wll*atc.cg[gll].z[z].T[b]     + wlr*atc.cg[glr].z[z].T[b]    );
    atc_2->Tg[b]    = wtop*(wul*atc.cg[gul].Tg[b]         + wur*atc.cg[gur].Tg[b]        ) + wdown*(wll*atc.cg[gll].Tg[b]         + wlr*atc.cg[glr].Tg[b]        );

  }

  return;
}


/** Bottom-of-Atmosphere reflectance
+++ This function eventually converts TOA to BOA reflectance, and gene-
+++ rates the data stack used as input for the geometric module.
--- TOA:    TOA reflectance
--- BCK:    Background reflectance
--- QAI:    Quality Assurance Information (modified)
--- GRD:    coarse resolution grid
--- TOP:    Topographic Derivatives
--- Tg:     Gas transmittance
--- WVP:    Water vapor
--- DAT:    Data stack (returned)
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void toa2boa(int2s **TOA, int2s **BCK, qai_t QAI, int2u *GRD, top_t TOP, int2s **Tg, int2s *WVP, int2s ***DAT){
int i, j, p, ii, jj, ip, jp, g, b, bb, z;
float tmp, tg;
float bck;
float *rhot = NULL;
int2s **dat = NULL;
int2s *DIS = NULL;
int2s wvp;
zgrid_t atc_2;
bool valid;
float *A, *E0, E0_;
float sky, ill, cor, f, f0, h0; 


  #ifdef CLOCK 
  time_t TIME; time(&TIME);
  #endif

  /** allocate and initialize memory for topographic correction **/
  alloc((void**)&A,  _NO_, sizeof(float));
  alloc((void**)&E0, _NO_, sizeof(float));

  for (b=0, bb=0; b<_NB_; b++){
    if (!par.use.refbands[b]) continue;
    A[bb] = 1.0;
    E0[bb]  = meta.refcal[b].E0;
    bb++;
  }


  /** allocate memory for data stack **/
  alloc_2D((void***)&dat, _ND_, _NC_, sizeof(int2s));
  for (b=0; b<_ND_; b++){
  for (p=0; p<_NC_; p++){
      dat[b][p] = _NODATA_;
  }
  }

  /** allocate memory for boa **/
  alloc((void**)&rhot, _NO_, sizeof(float));

  /** allocate memory for smoothed atmospheric parameters **/
  alloc((void**)&atc_2.aod,   _NB_, sizeof(float));
  alloc((void**)&atc_2.rho_p, _NB_, sizeof(float));
  alloc((void**)&atc_2.Ts,    _NB_, sizeof(float));
  alloc((void**)&atc_2.tvd,   _NB_, sizeof(float));
  alloc((void**)&atc_2.tvs,   _NB_, sizeof(float));
  alloc((void**)&atc_2.T,     _NB_, sizeof(float));
  alloc((void**)&atc_2.s,     _NB_, sizeof(float));
  alloc((void**)&atc_2.Tg,    _NB_, sizeof(float));
  alloc((void**)&atc_2.tsd,   _NB_, sizeof(float));
  alloc((void**)&atc_2.tss,   _NB_, sizeof(float));
  
  #ifdef DEBUG
  float **A_; if (par.dotopo) alloc_2D((void***)&A_, _NO_, _NC_, sizeof(float));
  #endif
  

  /** radiometric correction **/
  for (i=0, g=0; i<_NY_; i+=atc.gres){
  for (j=0; j<_NX_; j+=atc.gres, g++){

    for (ii=0; ii<atc.gres; ii++){
    for (jj=0; jj<atc.gres; jj++){

      ip = i+ii; jp = j+jj;
      if (ip > _NY_-1) continue;
      if (jp > _NX_-1) continue;
      p = ip*_NX_+jp;

      if (QAI.off[p]) continue;

      valid = true;

      if (par.doatmo){

        z = TOP.dem[p];
        
        // high sun zenith flag
        if (atc.cg[g].szen > 1.308997) QAI.sun[p] = true;

        // set AOD quality
        if (!atc.aodmap){
          QAI.aod[p] = 3;
        } else if (atc.cg[g].z[z].aod[_B_.green] > 0.6){
          QAI.aod[p] = 2;
        } else if (atc.cg[g].aod_int){
          QAI.aod[p] = 1;
        }

        // smooth aod and other variables
        smooth_atc(&atc_2, ip, jp, z);

        // topographic correction factor
        if (par.dotopo){
          
          sky = TOP.sky[p]/10000.0;
          ill = TOP.ill[p]/10000.0;
          cor = TOP.cor[p]/10000.0;

          for (b=0, bb=0; b<_NB_; b++){

            if (!par.use.refbands[b]) continue;

            if (ill > 0){

              E0_ = E0[bb] * atc.cg[g].Tvo[b]*atc.cg[g].Tso[b];
              f  = E0_*atc_2.tss[b]/(E0_*atc_2.tsd[b]);
              f0 = E0_*atc_2.tss[_B_.swir2]/(E0_*atc_2.tsd[_B_.swir2]);
              h0 = (M_PI+2*atc.cg[g].szen)/_M_PI2_;

              A[bb] = (atc.cg[g].cosszen+cor/f0*f/h0)/(ill+sky*cor/f0*f/h0);
              if (A[bb] < 0) A[bb] = -10000.0;

            } else A[bb] = 1.0;

            #ifdef DEBUG
            A_[bb][p] = A[bb];
            #endif
            bb++;

          }
        }


        // target reflectance
        if (par.doenv){

          for (b=0, bb=0; b<_NB_; b++){

            if (!par.use.refbands[b]) continue;
            
            if (meta.mission == MISSION_LANDSAT){
              tg = atc_2.Tg[b];
            } else if (meta.mission == MISSION_SENTINEL2){
              tg = Tg[bb][p]/10000.0;
            }

            bck = BCK[bb][p]/10000.0;
            tmp = (1-bck*atc_2.s[b]);

            rhot[bb] = A[bb] * atc.cg[g].brdf[b] *
                     (TOA[b][p]/_RS_/tg*tmp - 
                    atc_2.rho_p[b]*tmp - 
                    atc_2.Ts[b]*atc_2.tvs[b]*bck) / 
                     (atc_2.Ts[b]*atc_2.tvd[b]);
            bb++;

          }

        // for a homogeneous target  
        } else {

          for (b=0, bb=0; b<_NB_; b++){

            if (!par.use.refbands[b]) continue;
            
            if (meta.mission == MISSION_LANDSAT){
              tg = atc_2.Tg[b];
            } else if (meta.mission == MISSION_SENTINEL2){
              tg = Tg[bb][p]/10000.0;
            }
            
            tmp = (TOA[b][p]/_RS_-atc_2.rho_p[b])/tg;

            rhot[bb] = A[bb] * atc.cg[g].brdf[b] *
              tmp / (atc_2.T[b] + atc_2.s[b]*tmp);
            bb++;

          }
        }

      // top-of-atmosphere
      } else {

        for (b=0, bb=0; b<_NB_; b++){

          if (!par.use.refbands[b]) continue;

          rhot[bb] = TOA[b][p]/_RS_;
          bb++;

        }

      }


      /** BOA reflectance to data stack **/
      if (par.oref){
        for (b=0, bb=0; b<_NB_; b++){
          if (!par.use.refbands[b]) continue;
          if (rhot[bb] <  0.0) QAI.sub[p] = true;
          if (rhot[bb] < -1.0) valid = false;
          if (rhot[bb] >  1.0) QAI.sat[p] = true;
          if (rhot[bb] >  2.0) valid = false;
          dat[bb][p] = (int2s) (rhot[bb]*10000.0);
          bb++;
        }
        if (!valid){
          for (bb=0; bb<_NO_; bb++) dat[bb][p] = _NODATA_;
          QAI.off[p] = 1;
        }
      }

      /** HOT to data stack **/
      if (par.ohot){
        dat[par.m2d[_B_.hot].start][p] = TOA[_B_.blue][p] - 0.5*TOA[_B_.red][p] - 800;
      }

      /** view zenith to data stack **/
      if (par.ovzn){
        dat[par.m2d[_B_.vzn].start][p] = (int2s) (rad2deg(atc.cg[g].vzen)*100.0);
      }
      
      /** AOD to data stack **/
      if (par.oaod && par.doatmo){
        dat[par.m2d[_B_.aod].start][p] = (int2s) (atc_2.aod[_B_.green]*1000.0);
      }
      
      /** water vapor to data stack **/
      if (par.owvp && par.doatmo){
        if (meta.mission == MISSION_LANDSAT){
          wvp = (int2s) (atc.wvp*1000.0);
        } else if (meta.mission == MISSION_SENTINEL2){
          wvp = WVP[p];
        }
        dat[par.m2d[_B_.wvp].start][p] =  wvp;
      }

    }
    }

  }
  }


  /** cloud/shadow distance to data stack **/
  if (par.ocld){
    cloud_distance(QAI, &DIS);
    for (p=0; p<_NC_; p++) dat[par.m2d[_B_.cld].start][p] = DIS[p];
    free((void*)DIS);
  }


  /** build QAI **/
  if (par.oqai) assemble_qai(QAI, dat, par.m2d[_B_.qai].start);


  #ifdef DEBUG
  if (par.dotopo){
    db_write_image_2D((void**)A_, "TOPO-A-FACTOR", _NO_, _NX_, _NY_, sizeof(float), 4);
    free_2D((void**)A_, _NO_);
  }
  #endif

  /** clean **/
  free((void*)atc_2.aod); free((void*)atc_2.rho_p); free((void*)atc_2.Ts); 
  free((void*)atc_2.tvd); free((void*)atc_2.tvs);   free((void*)atc_2.T); 
  free((void*)atc_2.s);   free((void*)atc_2.Tg);    free((void*)rhot);
  free((void*)atc_2.tsd); free((void*)atc_2.tss); 
  free((void*)E0);        free((void*)A);           free((void*)GRD);


  #ifdef CLOCK
  proctime_print("convert TOA to BOA", TIME);
  #endif

  *DAT = dat;
  return;
}


/** Compute background reflectance
+++ This function computes the background reflectance used for correcting
+++ adjacency effects
--- TOA:    TOA reflectance
--- QAI:    Quality Assurance Information (modified)
--- GRD:    coarse resolution grid
--- TOP:    Topographic Derivatives
--- Tg:     Gas transmittance
--- BCK:    Background reflectance (returned)
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void background_ref(int2s **TOA, qai_t QAI, int2u *GRD, top_t TOP, int2s **Tg, int2s ***BCK){
int b, bb, p, i, j, g, z, k;
float F[3], Fr, Fa, km[3], r[3], dF[3], sF, aod, mod;
float alpha[3], old[3], tmp, tg;
float  *ref = NULL;
float **avg = NULL;
int2s **bck = NULL;
double *imean = NULL;
double *jmean = NULL;
double in, jn;
float w_avg;
#ifdef DEBUG
char fname[MAX_STR_LEN];
#endif

  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif


  if (!par.doenv) return;

  /** allocate memory **/
  alloc((void**)&ref, _NC_, sizeof(float));
  alloc((void**)&imean, _NY_, sizeof(double));
  alloc((void**)&jmean, _NX_, sizeof(double));
  alloc_2D((void***)&avg, 3, _NC_, sizeof(float));
  alloc_2D((void***)&bck, _NO_, _NC_, sizeof(int2s));

  /** radius of rings **/
  km[0] = 0.25;
  km[1] = 1.00;
  km[2] = 4.50;
  for (k=0; k<3; k++) r[k] = km[k]*1000/meta.res;

  /** alpha for exponential moving average **/
  for (k=0; k<3; k++) alpha[k] = 2.0/(r[k]+1.0);

  #ifdef DEBUG
  for (k=0; k<3; k++) printf("alpha for exponential moving average is %0.4f (ring %d)\n", alpha[k], k);
  #endif


  /** estimate for each band **/
  for (b=0, bb=0; b<_NB_; b++){

    if (!par.use.refbands[b]) continue;

    /** BOA reflectance without env. correction **/
    for (p=0; p<_NC_; p++){

      if (QAI.off[p]) continue;

      z = TOP.dem[p];
      g = GRD[p];

      if (meta.mission == MISSION_LANDSAT){
        tg = atc.cg[g].Tg[b];
      } else if (meta.mission == MISSION_SENTINEL2){
        tg = Tg[bb][p]/10000.0;
      }

      tmp = (TOA[b][p]/_RS_-atc.cg[g].z[z].rho_p[b])/tg;
      ref[p] = tmp / (atc.cg[g].z[z].T[b] + atc.cg[g].z[z].s[b]*tmp);

    }

    #ifdef DEBUG
    sprintf(fname, "REF_APP_BAND%02d", b);
    db_write_image_1D((void**)ref, fname, _NX_, _NY_, sizeof(float), 4);
    #endif


    /** row means **/
    for (i=0; i<_NY_; i++){
      imean[i] = in = 0;
      for (j=0, p=i*_NX_; j<_NX_; j++, p++){
        if (QAI.off[p] || QAI.cld[p] > 0) continue;
        imean[i] += ref[p];
        in++;
      }
      if (in > 0) imean[i] = imean[i] /= in;
    }

    /** column means **/
    for (j=0; j<_NX_; j++){
      jmean[j] = jn = 0;
      for (i=0, p=j; i<_NY_; i++, p+=_NX_){
        if (QAI.off[p] || QAI.cld[p] > 0) continue;
        jmean[j] += ref[p];
        jn++;
      }
      if (jn > 0) jmean[j] = jmean[j] /= jn;
    }

    /** exponential moving average, left to right, then right to left **/
    for (i=0; i<_NY_; i++){

      for (k=0; k<3; k++) old[k] = (float)imean[i];

      for (j=0, p=i*_NX_; j<_NX_; j++, p++){
        if (QAI.off[p]) continue;
        for (k=0; k<3; k++){
          if (QAI.cld[p] > 0){
            avg[k][p] = old[k];
          } else {
            avg[k][p] = alpha[k]*ref[p] + (1.0-alpha[k])*old[k];
            old[k] = avg[k][p];
          }
        }
      }
      p--;

      for (j=_NX_-1; j>=0; j--, p--){
        if (QAI.off[p]) continue;
        for (k=0; k<3; k++){
          if (QAI.cld[p] > 0){
            avg[k][p] = old[k];
          } else {
            avg[k][p] = alpha[k]*avg[k][p] + (1.0-alpha[k])*old[k];
            old[k] = avg[k][p];
          }
        }
      }

    }

    /** exponential moving average, top to bottom, then bottom to top **/
    for (j=0; j<_NX_; j++){

      for (k=0; k<3; k++) old[k] = (float)jmean[j];

      for (i=0, p=j; i<_NY_; i++, p+=_NX_){
        if (QAI.off[p]) continue;
        for (k=0; k<3; k++){
          if (QAI.cld[p] > 0){
            avg[k][p] = old[k];
          } else {
            avg[k][p] = alpha[k]*avg[k][p] + (1.0-alpha[k])*old[k];
            old[k] = avg[k][p];
          }
        }
      }
      p-=_NX_;

      for (i=_NY_-1; i>=0; i--, p-=_NX_){
        if (QAI.off[p]) continue;
        for (k=0; k<3; k++){
          if (QAI.cld[p] > 0){
            avg[k][p] = old[k];
          } else {
            avg[k][p] = alpha[k]*avg[k][p] + (1.0-alpha[k])*old[k];
            old[k] = avg[k][p];
          }
        }
      }

    }

    #ifdef DEBUG
    sprintf(fname, "REF_BCK_BAND%02d", b);
    db_write_image_2D((void**)avg, fname, 3, _NX_, _NY_, sizeof(float), 4);
    #endif


    /** relative influence of rings **/
    mod = mod_elev_scale(atc.mod[b], 1, atc.Hr);
    aod = atc.aod[b];

    for (k=0; k<3; k++){
      Fa = env_weight_aerosol(km[k]);
      Fr = env_weight_molecular(km[k]);
      F[k] = env_weight(aod, mod, Fa, Fr);
    }

    dF[0] = F[0]; dF[1] = F[1]-F[0]; dF[2] = F[2]-F[1];
    sF = dF[0]+dF[1]+dF[2];
    
    #ifdef DEBUG
    for (k=0; k<3; k++) printf("relative influence of ring %d: %.2f\n", k, dF[k]/sF);
    #endif
    
    /** background reflectance **/
    for (p=0; p<_NC_; p++){

      if (QAI.off[p]) continue;

      z = TOP.dem[p];
      g = GRD[p];

      w_avg = (dF[0]*avg[0][p] + dF[1]*avg[1][p] + dF[2]*avg[2][p])/sF;
      bck[bb][p] = (int2s) ((ref[p]*atc.cg[g].z[z].F[b] + 
                   (1-atc.cg[g].z[z].F[b])*w_avg) * 10000.0);

    }

    bb++;

  }

  
  free((void*)ref);
  free((void*)imean);
  free((void*)jmean);
  free_2D((void**)avg, 3);


  #ifdef DEBUG
  db_write_image_2D((void**)bck, "REF-BCK",  _NO_, _NX_, _NY_, sizeof(int2s), 2);
  #endif

  #ifdef CLOCK
  proctime_print("compute background reflectance", TIME);
  #endif

  *BCK = bck;
  return;
}


/** Compute angle-dependent atmospheric parameters
+++ This function computes all atmospheric parameters that are a function 
+++ of view or sun angles.
--- TOP:    Topographic Derivatives
--- QAI:    Quality Assurance Information
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void atmo_angledep(top_t TOP, qai_t QAI){
int i, j, g, b;


  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  for (i=0, g=0; i<_NY_; i+=atc.gres){
  for (j=0; j<_NX_; j+=atc.gres, g++){

    // bi-directional correction parameter
    brdf_factor(g);

    // relative air mass
    atc.cg[g].ms = atc.cg[g].cosszen;
    atc.cg[g].mv = atc.cg[g].cosvzen;

    // cosine of backscattering angle
    atc.cg[g].cospsi = backscatter(atc.cg[g].ms, atc.cg[g].mv, atc.cg[g].sazi, atc.cg[g].vazi);

    // phase functions for molecular and aerosol scattering
    atc.cg[g].Pr = phase_molecular(atc.cg[g].cospsi);
    atc.cg[g].Pa = phase_aerosol(atc.cg[g].cospsi);

    atc.cg[g].ozone = ozone_amount(atc.cg[g].geox, atc.cg[g].geoy, meta.date.doy);

    for (b=0; b<_NB_; b++){

    // water vapor transmittance
      atc.cg[g].Tsw[b] = wvp_transmitt(b, atc.wvp, atc.cg[g].ms);
      atc.cg[g].Tvw[b] = wvp_transmitt(b, atc.wvp, atc.cg[g].mv);
      atc.cg[g].Tso[b] = ozone_transmitt(b, atc.cg[g].ozone, atc.cg[g].ms);
      atc.cg[g].Tvo[b] = ozone_transmitt(b, atc.cg[g].ozone, atc.cg[g].mv);

      // gaseous transmittance
      atc.cg[g].Tg[b] = gas_transmitt(atc.cg[g].Tsw[b], atc.cg[g].Tvw[b], atc.cg[g].Tso[b], atc.cg[g].Tvo[b]);

    }

    // average elevation
    atc.cg[g].cdem  = average_elevation_cg(TOP, QAI, atc.gres, j, i);

    // correct rayleigh for elevation
    //od_elev_factor(atc.cg[g].cdem*0.1+atc.dem.cmin+0.05, &atc.cg[g].Ha, &atc.cg[g].Hr);
    od_elev_factor(atc.cg[g].cdem*atc.dem.step+atc.dem.min+atc.dem.step/2, 
      &atc.cg[g].Ha, &atc.cg[g].Hr);
    for (b=0; b<_NB_; b++) atc.cg[g].mod[b] = mod_elev_scale(atc.mod[b], 1, atc.cg[g].Hr);

    // Fresnel reflection for AOD from water retrieval 
    // --> assumed that the water body is flat -> incidence angle = szen
    atc.cg[g].rhof = fresnel_reflection(atc.cg[g].szen);

  }
  }

  #ifdef DEBUG
  print_fvector(atc.cg[(atc.ngy/2)*atc.ngx+(atc.ngx/2)].Tg, "\ngas transmittance", _NB_, 1, 3);
  #endif

  #ifdef CLOCK
  proctime_print("angle-dependent atm. modelling", TIME);
  #endif

  return;
}


/** Compute elevation-dependent atmospheric parameters
+++ This function computes all atmospheric parameters that are a function 
+++ of terrain elevation.
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void atmo_elevdep(){
int i, j, g, b, z;
float zmed;


  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  for (i=0, g=0; i<_NY_; i+=atc.gres){
  for (j=0; j<_NX_; j+=atc.gres, g++){

    alloc((void**)&atc.cg[g].z, atc.dem.cnum, sizeof(zgrid_t));

    // for every possible elevation (100m steps)
    //for (z=0, zmed=(atc.dem.cmin+0.05); z<atc.dem.cnum; z++, zmed+=0.1){
    for (z=0, zmed=(atc.dem.min+atc.dem.step/2); z<atc.dem.cnum; z++, zmed+=atc.dem.step){

      // allocate band-dependent variables
      alloc((void**)&atc.cg[g].z[z].mod,   _NB_, sizeof(float));
      alloc((void**)&atc.cg[g].z[z].aod,   _NB_, sizeof(float));
      alloc((void**)&atc.cg[g].z[z].od,    _NB_, sizeof(float));
      alloc((void**)&atc.cg[g].z[z].rho_p, _NB_, sizeof(float));
      alloc((void**)&atc.cg[g].z[z].Ts,    _NB_, sizeof(float));
      alloc((void**)&atc.cg[g].z[z].tsd,   _NB_, sizeof(float));
      alloc((void**)&atc.cg[g].z[z].tss,   _NB_, sizeof(float));
      alloc((void**)&atc.cg[g].z[z].Tv,    _NB_, sizeof(float));
      alloc((void**)&atc.cg[g].z[z].tvd,   _NB_, sizeof(float));
      alloc((void**)&atc.cg[g].z[z].tvs,   _NB_, sizeof(float));
      alloc((void**)&atc.cg[g].z[z].T,     _NB_, sizeof(float));
      alloc((void**)&atc.cg[g].z[z].s,     _NB_, sizeof(float));
      alloc((void**)&atc.cg[g].z[z].F,     _NB_, sizeof(float));

      // correct mod for elevation
      od_elev_factor(zmed, &atc.cg[g].z[z].Ha, &atc.cg[g].z[z].Hr);


      // down/up-welling scattering transmittances
      for (b=0; b<_NB_; b++){

        // correct mod and aod for elevation
        atc.cg[g].z[z].mod[b] = mod_elev_scale(atc.mod[b], 1, atc.cg[g].z[z].Hr);

        if (atc.aodmap){
          atc.cg[g].z[z].aod[b] = aod_elev_scale(atc.cg[g].aod[b], atc.Ha, atc.cg[g].z[z].Ha);
        } else {
          atc.cg[g].z[z].aod[b] = aod_elev_scale(atc.aod[b], atc.Ha, atc.cg[g].z[z].Ha);
        }

        // optical depth
        atc.cg[g].z[z].od[b] = optical_depth(atc.cg[g].z[z].aod[b], 
          atc.cg[g].z[z].mod[b]);

        // scattering transmittance
        atc.cg[g].z[z].T[b] = scatt_transmitt(atc.cg[g].z[z].aod[b], 
          atc.cg[g].z[z].mod[b], atc.cg[g].z[z].od[b], atc.cg[g].ms, 
          atc.cg[g].mv, &atc.cg[g].z[z].Ts[b], &atc.cg[g].z[z].Tv[b],
          &atc.cg[g].z[z].tsd[b], &atc.cg[g].z[z].tss[b], 
          &atc.cg[g].z[z].tvd[b], &atc.cg[g].z[z].tvs[b]);

        // path reflectance
        atc.cg[g].z[z].rho_p[b] = path_ref(atc.cg[g].z[z].aod[b], 
          atc.cg[g].z[z].mod[b], atc.cg[g].z[z].od[b], atc.cg[g].Pa, 
          atc.cg[g].Pr, atc.cg[g].z[z].tsd[b], atc.cg[g].z[z].tvd[b], 
          atc.cg[g].ms, atc.cg[g].mv);

        // spherical albedo
        atc.cg[g].z[z].s[b] = sphere_albedo(atc.cg[g].z[z].aod[b], atc.cg[g].z[z].mod[b], 
          atc.cg[g].z[z].od[b]);

        // environmental weighting function
        if (par.doenv){
          atc.cg[g].z[z].F[b] = env_weight(atc.cg[g].z[z].aod[b], 
            atc.cg[g].z[z].mod[b], atc.Fa, atc.Fr);
        } else atc.cg[g].z[z].F[b] = 1.0;

      }

    }

  }
  }


  #ifdef CLOCK
  proctime_print("elevation-dependent atm. modelling", TIME);
  #endif

  return;
}


/** Radiometric module, main function
+++ This function is the 2nd core module of FORCE L2PS, and does radiomet-
+++ ric correction. This includes estimating AOD, water vapor (S2), envi-
+++ ronment effect, topographic correction and atmospheric correction.
--- TOA:    TOA reflectance
--- QAI:    Quality Assurance Information (modified)
--- TOP:    Topographic Derivatives
--- GRD:    coarse resolution grid
--- DATA:   data stack (returned)
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int radiometric_module(int2s **TOA, qai_t QAI, top_t TOP, int2u *GRD, int2s ***DATA){
int b, g, z;
int2s **BCK = NULL;
int2s **DAT = NULL;
int2s  *WVP = NULL;
int2s **Tg  = NULL;

  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif


  /** elevation stats (mean, max/min, # of 100m classes)
  +++ rayleigh scattering @ sea level
  +++ elevation correction factors for optical depths
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
  if (par.doatmo){

    for (b=0; b<_NB_; b++) atc.mod[b] = molecular_optical_depth(atc.wvl[b]);
    od_elev_factor(atc.dem.avg, &atc.Ha, &atc.Hr);

    #ifdef DEBUG
    print_fvector(atc.mod, "rayleigh @ sea level", _NB_, 1, 4);
    printf("Hr: %.4f, Ha: %.4f\n", atc.Hr, atc.Ha);
    #endif
  }


  /** precipitable water vapor
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
  if (water_vapor_from_lut() == FAILURE){
    printf("Cannot read wvp from LUT. "); return FAILURE;}


  /** angle-dependent coarse-grid atmospheric modelling
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
  if (par.doatmo) atmo_angledep(TOP, QAI);


  /** compile AOD, use image-based water/shadow targets, refine by DODB, 
  +++ use external values (one or several options are possible)
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
  if (par.doatmo) compile_aod(TOA, QAI, TOP);


  /** elevation-dependent coarse-grid atmospheric modelling
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
  if (par.doatmo) atmo_elevdep();


  /** water vapor and gaseous transmittance estimation
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
  if (wvp_transmitt_lut() != SUCCESS){
      printf("error in water vapor transmittance LUT. "); return FAILURE;}
  if (water_vapor(TOA, QAI, GRD, TOP, &WVP, &Tg) != SUCCESS){
      printf("error in water vapor estimation. "); return FAILURE;}


  /** estimate topographic correction factor
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
  if (topographic_factor(TOA, QAI, GRD, &TOP) != SUCCESS){
      printf("error in topographic correction. "); return FAILURE;}
  if (TOP.slp != NULL) free((void*)TOP.slp); TOP.slp = NULL;
  if (TOP.asp != NULL) free((void*)TOP.asp); TOP.asp = NULL;
  

  #ifdef DEBUG
  if (par.doatmo) db_write_image_CG(atc);
  #endif


  /** environment layer **/
  background_ref(TOA, QAI, GRD, TOP, Tg, &BCK);


  /** surface reflectance **/
  toa2boa(TOA, BCK, QAI, GRD, TOP, Tg, WVP, &DAT);

  #ifdef DEBUG
  db_write_image_2D((void**)DAT, "DATA-STACK", _ND_, _NX_, _NY_, sizeof(int2s), 2);
  #endif


  /** clean
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
  if (TOP.dem != NULL) free((void*)TOP.dem); TOP.dem = NULL;
  if (TOP.ill != NULL) free((void*)TOP.ill); TOP.ill = NULL;
  if (TOP.sky != NULL) free((void*)TOP.sky); TOP.sky = NULL;
  if (TOP.cor != NULL) free((void*)TOP.cor); TOP.cor = NULL;

  if (par.doenv) free_2D((void**)BCK, _NO_);
  free_2D((void**)TOA, _NB_);

  if (par.doatmo){
    for (g=0; g<atc.ngc; g++){
      for (z=0; z<atc.dem.cnum; z++){
        free((void*)atc.cg[g].z[z].mod); free((void*)atc.cg[g].z[z].aod);
        free((void*)atc.cg[g].z[z].od); free((void*)atc.cg[g].z[z].rho_p);
        free((void*)atc.cg[g].z[z].Ts); free((void*)atc.cg[g].z[z].tsd);
        free((void*)atc.cg[g].z[z].tss); free((void*)atc.cg[g].z[z].Tv);
        free((void*)atc.cg[g].z[z].tvd); free((void*)atc.cg[g].z[z].tvs);
        free((void*)atc.cg[g].z[z].T); free((void*)atc.cg[g].z[z].s);
        free((void*)atc.cg[g].z[z].F);
      }
      free((void*)atc.cg[g].z);
    }
  }

  free((void*)meta.refcal);

  if (meta.mission == MISSION_SENTINEL2 && par.doatmo) free_2D((void**)Tg, _NO_);
  if (meta.mission == MISSION_SENTINEL2 && par.doatmo) free((void*)WVP);
  free_qai(QAI);

  #ifdef CLOCK
  proctime_print("radiometric module", TIME);
  #endif

  *DATA = DAT;
  return SUCCESS;
}

