/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
This file contains functions for converting between DN and TOA reflectance
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#include "force.h"


/** Initialize calibration coefficients
+++ This function initializes the calibration coefficients
--- cal:    calibration parameters (modified)
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int calinit(cal_t *cal){
int fill = -32767;

  strcpy(cal->fname, "NULL");
  strcpy(cal->mtl_band, "NULL");
  cal->rsr_band = fill;
  cal->lmax     = fill;
  cal->lmin     = fill;
  cal->qmax     = fill;
  cal->qmin     = fill;
  cal->rmul     = fill;
  cal->radd     = fill;
  cal->E0       = fill;
  cal->wvl      = fill;
  cal->k1       = fill;
  cal->k2       = fill;

  return SUCCESS;
}


/** Exoatmospheric solar irradiance
+++ This function computes band-specific E0 values based on the bands' re-
+++ lative spectral response and E0 spectrum.
+++ Return: SUCCESS / FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int E0(){
int b, b_rsr, w;
double e, s;


  for (b=0; b<_NB_; b++){
    b_rsr = meta.refcal[b].rsr_band;
    for (w=0, e=0, s=0; w<_WVL_DIM_; w++){
      e += _RSR_[b_rsr][w]*_E0_[w];
      s += _RSR_[b_rsr][w];
    }
    meta.refcal[b].E0 = e/s;
  }

  return SUCCESS;
}


/** Wavelengths
+++ This function computes band-specific wavelengths, and logarithm, and
+++ squared logarithm based on the bands' relative spectral response.
+++ Return: SUCCESS / FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int wavelengths(){
int b, b_rsr, w;
double v, s;


  for (b=0; b<_NB_; b++){
    b_rsr = meta.refcal[b].rsr_band;
    for (w=0, v=0, s=0; w<_WVL_DIM_; w++){
      v += _RSR_[b_rsr][w]*_WVL_[w];
      s += _RSR_[b_rsr][w];
    }
    meta.refcal[b].wvl = v/s;
    atc.wvl[b] = v/s;
    atc.lwvl[b]  = log(atc.wvl[b]);
    atc.lwvl2[b] = atc.lwvl[b]*atc.lwvl[b];
  }


  #ifdef DEBUG
  printf("\nwavelengths:\n");
  print_fvector(atc.wvl,   "wvl       ", _NB_, 2, 3);
  print_fvector(atc.lwvl,  "log(wvl)  ", _NB_, 2, 3);
  print_fvector(atc.lwvl2, "log(wvl)^2", _NB_, 2, 3);
  #endif

  return SUCCESS;
}


/** DN to TOA reflectance / brightness temperature
+++ This function converts a DN band to TOA reflectance (type = 0) or bri-
+++ ghtness temperature (type = 1).
--- cal:    calibration parameters (modified)
--- GRD:    coarse processing grid
--- QAI:    Quality Assurance Information
--- DN:     Digital Numbers
--- TOA:    Top of Atmosphere reflectance / temperature (returned)
--- type:   convert to reflectance (0) or brightness temperature (1)
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void dn2toa(cal_t *cal, int2u *GRD, qai_t QAI, int2u *DN, int2s **TOA, int type){
int p, g;
float A, rad, pi_dsun2, tmp;
int fill = -32767;
int2s *toa = NULL;

  alloc((void**)&toa, _NC_, sizeof(int2s));

  if (type == 0){
    cal->dsun = doy2dsun(meta.date.doy);
    pi_dsun2  = M_PI*cal->dsun*cal->dsun;

    #ifdef DEBUG
    printf("%.3fµm ::: d %.4f, E0 %.2f\n", cal-> wvl, cal->dsun, cal->E0);
    #endif
  }

  A = (cal->lmax-cal->lmin)/(cal->qmax-cal->qmin);

  for (p=0; p<_NC_; p++){

    if (QAI.off[p]) continue;

    // dn to radiance
    rad = A * ((float)DN[p]-cal->qmin) + cal->lmin;

    // radiance to reflectance
    if (type == 0){

      g = GRD[p];
      if (cal->rmul == fill){
        tmp = rad*pi_dsun2 / (cal->E0*atc.cg[g].cosszen);
      } else {
        tmp = (cal->radd + cal->rmul*(float)DN[p])/atc.cg[g].cosszen;
      }
      if (tmp < FLT_MIN){
        toa[p] = 0;
        QAI.off[p] = true;
      } else {
        toa[p] = (int2s)(tmp*_RS_);
      }

    // radiance to brightness temperature in kelvin
    } else if (type == 1){

      tmp = cal->k2/log((cal->k1/rad)+1)*_TS_;
      if (tmp < SHRT_MAX){
        toa[p] = (int2s)tmp;
      } else {
        toa[p] = SHRT_MAX;
      }

    } else {
      printf("unknown type.\n"); exit(1);
    }

  }
  
  *TOA = toa;
  return;
}


/** DN to TOA reflectance / brightness temperature wrapper
+++ This function converts all the DN bands to TOA reflectance or bright-
+++ ness Temperature. In case of Sentinel-2, TOA reflectance is first re-
+++ transformed to DNs before it is converted to TOA reflectance again.
+++ This is done to maintain a constant calibration between sensors and
+++ to retain the flexibility to e.g. use another E0 spectrum.
--- GRD:    coarse processing grid
--- QAI:    Quality Assurance Information
--- DN:     Digital Numbers (is deallocated within)
--- TOA:    Top of Atmosphere reflectance and temperature (returned)
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int dn_to_toa(int2u *GRD, qai_t QAI, dn_t DN, toa_t *TOA){
int b;

int p;

  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  if ((TOA->ref = (int2s**)malloc(_NB_*sizeof(int2s*))) == NULL){
    printf("Unable to allocate memory. "); return FAILURE;}

  /** TOA reflectance to TOA reflectance (Sentinel-2) **/
  if (meta.mission == MISSION_SENTINEL2){

    for (b=0; b<_NB_; b++){
      alloc((void**)&TOA->ref[b], _NC_, sizeof(int2s));
      for (p=0; p<_NC_; p++) TOA->ref[b][p] = (int2s)(((float)DN.ref[b][p]/meta.scale)*_RS_);
    }

  /** digital numbers to TOA reflectance and brightness temperature (Landsat) **/
  } else {

    for (b=0; b<_NB_; b++) dn2toa(&meta.refcal[b], GRD, QAI, DN.ref[b], &TOA->ref[b], 0);
    if (par.band_bt)       dn2toa(&meta.btcal,     GRD, QAI, DN.bt, &TOA->bt, 1);

  }

  /** clean **/
  free_2D((void**)DN.ref, _NB_);
  if (par.band_bt) free((void*)DN.bt);
  if (meta.mission == MISSION_SENTINEL2){
    free_2D((void**)meta.s2.szen, meta.s2.ny);
    free_2D((void**)meta.s2.vzen, meta.s2.ny);
    free_2D((void**)meta.s2.sazi, meta.s2.ny);
    free_2D((void**)meta.s2.vazi, meta.s2.ny);
  }

  #ifdef CLOCK
  proctime_print("DN to TOA conversion", TIME);
  #endif

  #ifdef DEBUG
  db_write_image_2D((void**)TOA->ref, "TOA-REF", _NB_, _NX_, _NY_, sizeof(int2s), 2);
  if (par.band_bt) db_write_image_1D((void**)TOA->bt,  "TOA-BT",  _NX_, _NY_, sizeof(int2s), 2);
  #endif

  return SUCCESS;
}


/** Detect extreme values
+++ This function detects extreme values and builds the nodata and satu-
+++ ration masks.
--- DN:     digital numbers
--- QAI:    Quality Assurance Information
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int dn_extreme(dn_t DN, qai_t QAI){
int i, j, p, b;
int sat;
bool tvalid = false;


  if (meta.dtype == 16) sat = 65535; else if (meta.dtype == 8) sat = 255; else {
    printf ("Unknown datatype. "); return FAILURE;}

    
  // do for each cell
  for (i=0, p=0; i<_NY_; i++){
  for (j=0; j<_NX_; j++, p++){

    // boundary or missing value

    // if any layer void --> boundary
    for (b=0; b<_NB_; b++){
      if (DN.ref[b][p] == 0){ QAI.off[p] = true; break;}
    }


    if (par.band_bt){
      if (DN.bt[p] == 0) QAI.off[p] = true;
    }


    // saturation in reflectance
    for (b=0; b<_NB_; b++){
      if (DN.ref[b][p] == sat){ QAI.sat[p] = true; break;}
    }

    if (par.band_bt && !tvalid && DN.bt[p] > 0) tvalid = true;

  }
  }

  // buffer one pixel (pixels are somehow contaminated. due to resampling?)
  buffer(QAI.off, 1);

  if (par.band_bt && !tvalid){
    printf("zero-filled temperature. "); return FAILURE;}

  detect_impulse_noise(DN.ref, QAI);

  #ifdef DEBUG
  db_write_image_1D((void**)QAI.off, "QAI-OFF", _NX_, _NY_, sizeof(bool), 1);
  db_write_image_1D((void**)QAI.sat, "QAI-SAT", _NX_, _NY_, sizeof(bool), 1);
  #endif

  return SUCCESS;
}


/** Detect Impulse Noise
+++ This function attempts to detect and mask Impulse Noise, a phenomenon
+++ observed in 8bit Landsat data. The first three bands (RGB) are used
+++ to detect this. Note that IN is not confined to these bands and small
+++ IN won't be detected.. This function simply identifies the worst occu-
+++ rences.
--- DN:     digital numbers
--- QAI:    Quality Assurance Information
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void detect_impulse_noise(int2u **DN, qai_t QAI){
int i, j, ii, jj, p, q, k, count = 0, b, nb = 3;
double mx[3],varx[3], sd[3];
double  maxsd, max2sd;

  // impact noise was not observed yet in 16bit data
  if (meta.dtype != 8) return;


  for (i=0, p=0; i<_NY_; i++){
    for (j=0; j<_NX_; j++, p++){

      if (QAI.off[p]) continue;

      k = 0;
      for (b=0; b<nb; b++) mx[b] = varx[b] = 0;
      for (ii=-1;ii<=1;ii++){
        for (jj=-1;jj<=1;jj++){

          q = _NX_*(i+ii)+j+jj;
          if (QAI.off[q]) continue;

          k++;

          if (k == 1){
            for (b=0; b<nb; b++) mx[b] = (double) DN[b][q];
          } else {
            for (b=0; b<nb; b++) var_recurrence((double) DN[b][q], &mx[b], &varx[b], k);
          }

        }
      }


      if (k>0){
        for (b=0; b<nb; b++) sd[b] = standdev(varx[b], k);
        max2sd = maxsd = sd[0];
        for (b=1; b<nb; b++){
          if (sd[b] > maxsd){ max2sd = maxsd; maxsd = sd[b];}
        }
        if ((maxsd-max2sd) > 15){ QAI.off[p] = true; count++;}
      }


    }
  }

  return;
}

