/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/** The following code is a modified version of the Fmask code, openly 
+++ shared by Zhe Zhu, Department of Geosciences, Texas Tech University:
+++ https://github.com/prs021/fmask
+++ Fmask Copyright (C) 2012-2015 Zhe Zhu, Curtis Woodcock
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
This file contains functions for cloud and cloud shadow identification
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#include "force.h"


/** Build final cloud classification
+++ This function builds the final cloud and cloud shadow classification.
--- npix:   number of valid image pixels
--- CLOUD:  cloud mask (is deallocated within)
--- SHADOW: shadow mask (is deallocated within)
--- TOA:    TOA reflectance
--- QAI:    Quality Assurance Information (modified)
+++ Return: total cloud cover in %
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
float finalize_cloud(float npix, bool *CLOUD, bool *SHADOW, int2s **TOA, qai_t QAI){
int p, k = 0;
float pct;

  /** set confident cloud, and cirrus **/
  for (p=0; p<_NC_; p++){
    if (QAI.off[p]) continue;
    if (QAI.snw[p]){
      CLOUD[p] = false;
    } else if (CLOUD[p]){
      QAI.cld[p] = 2;
//    } else if (_B_.cir  != -1 && TOA[_B_.cir][p] > 100 && TOA[_B_.blue][p] < 1000 && !QAI.snw[p]){
//      CLOUD[p] = true;
//      QAI.cld[p] = 3;
    }
  }

  /** set 300m around cloud to less confident cloud **/
  buffer(CLOUD,  300/meta.res);

  for (p=0; p<_NC_; p++){
    if (QAI.off[p]) continue;
    if (QAI.snw[p]){
      CLOUD[p] = SHADOW[p] = false;
    } else if (CLOUD[p]){
      if (QAI.cld[p] == 0) QAI.cld[p] = 1;
    } else if (_B_.cir  != -1 && TOA[_B_.cir][p] > 100 && TOA[_B_.blue][p] < 1000 && !QAI.snw[p]){
      QAI.cld[p] = 3;
    } else if (SHADOW[p]){
      QAI.shd[p] = true;
    }
  }

  for (p=0; p<_NC_; p++){
    if (QAI.off[p]) continue;
    if (QAI.cld[p] > 0 || QAI.shd[p]) k++;
  }

  /** total cloud / cloud shadow cover in % **/
  pct = 100.0*k/npix;

  free((void*)CLOUD); free((void*)SHADOW);
  
  #ifdef DEBUG
  db_write_image_1D((void**)QAI.cld, "QAI-CLD",   _NX_, _NY_, sizeof(int1u), 1);
  db_write_image_1D((void**)QAI.shd, "QAI-SHD",   _NX_, _NY_, sizeof(bool),  1);
  #endif

  return pct;
}


/** Cloud / cloud shadow / snow distance
+++ Compute distance to the next cloud, cloud shadow or snow pixel in m
--- QAI:    Quality Assurance Information
--- DIST:   Distance (returned)
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int cloud_distance(qai_t QAI, int2s **DIST){
int p, k=0;
float dist;
bool  *TO_DIST    = NULL;
int2u *DISTANCE_P = NULL;
int2s *DISTANCE_M = NULL;


  /** build binary mask **/
  alloc((void**)&TO_DIST, _NC_, sizeof(bool));
  for (p=0; p<_NC_; p++){
    if (QAI.cld[p] > 0 || QAI.shd[p] || QAI.snw[p]){
      TO_DIST[p] = true; 
      k++;
    }
  }

  /** compute pixel distance **/
  alloc((void**)&DISTANCE_P, _NC_, sizeof(int2u));
  if (k > 0){
    dist_transform(TO_DIST, DISTANCE_P);
  } else {
    for (p=0; p<_NC_; p++) DISTANCE_P[p] = _NY_+_NX_;
  }
  free((void*)TO_DIST);

  /** compute distance in meters, capped by 32767 **/
  alloc((void**)&DISTANCE_M, _NC_, sizeof(int2s));
  for (p=0; p<_NC_; p++){
        if (QAI.off[p]){
      DISTANCE_M[p] = _NODATA_;
    } else if ((dist = DISTANCE_P[p]*meta.res) < SHRT_MAX){
      DISTANCE_M[p] = (int2s) dist;
    } else {
      DISTANCE_M[p] = SHRT_MAX;
    }
  }
  free((void*)DISTANCE_P);

  *DIST = DISTANCE_M;
  return SUCCESS;
}


/** Potential cloud pixels
+++ This function computes the potential cloud pixels.
--- npix:   number of valid image pixels (returned)
--- nclear: number of clear-sky pixels (returned)
--- nland:  number of clear-sky land pixels (returned)
--- TOA:    TOA reflectance
--- QAI:    Quality Assurance Information (modified)
--- SLP:    Topographic slope
--- CLR:    Clear-sky pixels (returned)
--- LND:    Clear-sky land pixels (returned)
--- BRT:    Bright pixels (returned)
--- VPR:    Variability probability (returned)
--- PCP:    Potential Cloud Pixels (returned)
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void potential_cloud(float *npix, int *nclear, int *nland, 
  toa_t TOA, qai_t QAI, int2u *SLP,
  bool **CLR, bool **LND, bool **BRT, float **VPR, bool **PCP){
int p, nsnow;
float bt, ndvi, ndsi, hot, vis, white, tmp, r45;
float blue, green, red, nir, sw1, sw2;
bool basic;
bool  *BRIGHT = NULL;
bool  *CLOUD  = NULL;
bool  *CLEAR  = NULL;
bool  *LAND   = NULL;
float *VPROB  = NULL;


  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  alloc((void**)&BRIGHT, _NC_, sizeof(bool));
  alloc((void**)&CLOUD,  _NC_, sizeof(bool)); 
  alloc((void**)&CLEAR,  _NC_, sizeof(bool));
  alloc((void**)&LAND,   _NC_, sizeof(bool));
  alloc((void**)&VPROB,  _NC_, sizeof(float));


  for (p=0; p<_NC_; p++){

    if (QAI.off[p]) continue;

    blue  = TOA.ref[_B_.blue][p]/_RS_;
    green = TOA.ref[_B_.green][p]/_RS_;
    red   = TOA.ref[_B_.red][p]/_RS_;
    nir   = TOA.ref[_B_.nir][p]/_RS_;
    sw1   = TOA.ref[_B_.swir1][p]/_RS_;
    sw2   = TOA.ref[_B_.swir2][p]/_RS_;
    
    if (par.band_bt) bt = TOA.bt[p]/_TS_;  else bt = 0.0;

    if ((tmp = nir+red) != 0){
      ndvi = (nir-red)/tmp;
    } else ndvi = 0.01;

    if ((tmp = green+sw1) != 0){
      ndsi = (green-sw1)/tmp;
    } else ndsi = 0.01;
    
    // Zhu et al., 2015 update, use 283K instead of 277K
    if (par.band_bt && ndsi > 0.15 && bt < 283 && 
                       nir > 0.11 && green > 0.1){
      QAI.snw[p] = true; 
    } else if (ndsi > 0.4 && sw2 < 0.11 &&
                       nir > 0.11 && green > 0.1){
      QAI.snw[p] = true; 
    }

    if (((ndvi < 0.01 && nir < 0.11) || 
         (ndvi < 0.1 && ndvi > 0 && nir < 0.05)) &&
          SLP[p] < 870){
      QAI.wtr[p] = true;
      QAI.snw[p]  = false;
    }

    if (ndsi < 0.8 && ndvi < 0.8 && sw2 > 0.03 && bt < 300.15){
      basic = true;
    } else basic = false;

    if ((vis = (blue+green+red)/3.0) > 0.15) BRIGHT[p] = true;

    white = (fabs(blue-vis) + fabs(green-vis) + fabs(red-vis))/vis;
    if (white > 1) white = 1.0;

    hot = blue - 0.5*red - 0.08;

    if (sw1 != 0) r45 = nir/sw1; else r45 = 0.0;

    if (QAI.sat[p] && !QAI.snw[p]){
      CLOUD[p]   = true;
    } else if (QAI.snw[p]){
      CLOUD[p] = false;
    } else if (basic && white < 0.7 && BRIGHT[p] && hot > 0 && r45 > 0.75){
      CLOUD[p] = true;
    }

    // MODIFY NDVI & NDSI
    if (QAI.sat[p] && ndvi > 0) ndvi = 0.0; else ndvi = fabs(ndvi);
    if (QAI.sat[p] && ndsi < 0) ndsi = 0.0; else ndsi = fabs(ndsi);

    // variability probability
    if (ndvi >= ndsi){
      if (ndvi >= white) VPROB[p] = 1.0 - ndvi; else VPROB[p] = 1.0 - white;
    } else {
      if (ndsi >= white) VPROB[p] = 1.0 - ndsi; else VPROB[p] = 1.0 - white;
    }

    // PERCENT CLEAR
    if (!CLOUD[p] && !QAI.snw[p])                CLEAR[p] = true;
    if (!CLOUD[p] && !QAI.snw[p] && !QAI.wtr[p]) LAND[p] = true;

  }


   // Zhu et al., 2015 update, buffer snow layer by 1px
  buffer(QAI.snw, 1);
  for (p=0; p<_NC_; p++){
    if (QAI.wtr[p]) QAI.snw[p] = false;
    if (QAI.snw[p]) CLOUD[p] = false;
  }
  nsnow  = bsum(QAI.snw, _NC_);

  *npix   = _NC_ - bsum(QAI.off, _NC_);
  *nclear = bsum(CLEAR, _NC_);
  *nland  = bsum(LAND, _NC_);

  *CLR = CLEAR;
  *LND = LAND;
  *VPR = VPROB;
  *BRT = BRIGHT;
  *PCP = CLOUD;

  printf("sc: %6.2f%%. ", nsnow/(float)*npix);

  #ifdef CLOCK
  proctime_print("potential cloud identification", TIME);
  #endif

  #ifdef DEBUG
  db_write_image_1D((void**)CLOUD,  "FMASK-PCP",    _NX_, _NY_, sizeof(bool),  1);
  db_write_image_1D((void**)CLEAR,  "FMASK-CLEAR",  _NX_, _NY_, sizeof(bool),  1);
  db_write_image_1D((void**)LAND,   "FMASK-LAND",   _NX_, _NY_, sizeof(bool),  1);
  db_write_image_1D((void**)QAI.wtr,  "QAI-WAT",  _NX_, _NY_, sizeof(bool),  1);
  db_write_image_1D((void**)VPROB,  "FMASK-VPROB",  _NX_, _NY_, sizeof(float), 4);
  db_write_image_1D((void**)BRIGHT, "FMASK-BRIGHT", _NX_, _NY_, sizeof(bool),  1);
  db_write_image_1D((void**)QAI.snw, "QAI-SNW",   _NX_, _NY_, sizeof(bool),  1);
  #endif

  return;
}


/** Cloud probability
+++ This function computes the cloud probability.
--- npix:   number of valid image pixels
--- nclear: number of clear-sky pixels
--- nland:  number of clear-sky land pixels
--- ncloud: number of cloud pixels (returned)
--- cc:     revised total cloud cover in % (returned)
--- lowt:   lower  temperature quantile (returned)
--- hight:  higher temperature quantile (returned)
--- CLEAR:   Clear-sky pixels (is deallocated within)
--- LAND:   Clear-sky land pixels
--- TOA:    TOA reflectance
--- QAI:    Quality Assurance Information
--- BRIGHT: Bright pixels (is deallocated within)
--- VPROB:  Variability probability (is deallocated within)
--- PCP:    Potential Cloud Pixels (is deallocated within)
--- OCLOUD: Cloud mask (returned)
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void cloud_probability(float npix, int nclear, int nland, int *ncloud, float *cc, float *lowt, float *hight, bool *CLEAR, bool *LAND, toa_t TOA, qai_t QAI, bool *BRIGHT, float *VPROB, bool *PCP, bool **OCLOUD){
int p, pos;
int nCLEARWTR = 0;
float lclr_max, wclr_max;
float tempprob, brightprob, cirprob;
bool pcp;
bool *CLEARWTR = NULL;
bool *CLOUD = NULL;
float *WATERPROB = NULL, *CLEARWPROB = NULL;
float *LANDPROB = NULL, *CLEARLPROB = NULL;
float *CLEARTEMP = NULL;
float lo = 0.175, hi = 0.825;

float lowtemp = 0.0, hightemp = 0.0, t_wtemp = 0.0;
float A, B;
float bt;

  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  if (par.band_bt){
    // get clear all or clear land temp + temperature quantiles
    if ((100.0*nland/npix) >= 0.1){

      alloc((void**)&CLEARTEMP, nland, sizeof(float));

      for (p=0, pos=0; p<_NC_; p++){
        if (LAND[p]){
          CLEARTEMP[pos++] = TOA.bt[p]/_TS_;
        }
      }

      lowtemp = quantile(CLEARTEMP, nland, lo);
      hightemp = quantile(CLEARTEMP, nland, hi);

    } else {

      alloc((void**)&CLEARTEMP, nclear, sizeof(float));

      for (p=0, pos=0; p<_NC_; p++){
        if (CLEAR[p]){
          CLEARTEMP[pos++] = TOA.bt[p]/_TS_;
        }
      }

      lowtemp = quantile(CLEARTEMP, nclear, lo);
      hightemp = quantile(CLEARTEMP, nclear, hi);

    }

    free((void*)CLEARTEMP);

  }

  free((void*)CLEAR);

  // get clear water + temperature quantiles
  alloc((void**)&CLEARWTR, _NC_, sizeof(bool));

  for (p=0; p<_NC_; p++){
    if (QAI.wtr[p] && TOA.ref[_B_.swir2][p] <= 300){
      CLEARWTR[p] = true;
      nCLEARWTR++;
    }
  }

  // test if there is clear water, if not skip and give water prob = -1
  if (nCLEARWTR > 0){

    if (par.band_bt){
      alloc((void**)&CLEARTEMP, nCLEARWTR, sizeof(float));

      for (p=0, pos=0; p<_NC_; p++){
        if (CLEARWTR[p]){
          CLEARTEMP[pos++] = TOA.bt[p]/_TS_;
        }
      }

      t_wtemp = quantile(CLEARTEMP, nCLEARWTR, hi);
      free((void*)CLEARTEMP);
    }

    alloc((void**)&WATERPROB, _NC_, sizeof(float));

    for (p=0; p<_NC_; p++){
      if (par.band_bt){ 
        tempprob = (t_wtemp-TOA.bt[p]/_TS_)/4.0;
      } else {
        tempprob = 1.0;
      }
      brightprob = (TOA.ref[_B_.swir1][p]/_RS_)/0.11;
      //if (_B_.cir  != -1) cirprob = (TOA.ref[_B_.cir][p]/_RS_)/0.04; else cirprob = 0.0;
      cirprob = 0.0;
      if (brightprob > 1) brightprob = 1;
      WATERPROB[p] = tempprob*brightprob + cirprob;
    }

    alloc((void**)&CLEARWPROB, nCLEARWTR, sizeof(float));

    for (p=0, pos=0; p<_NC_; p++){
      if (CLEARWTR[p]){
        CLEARWPROB[pos] = WATERPROB[p];
        pos++;
      }
    }

    wclr_max = quantile(CLEARWPROB, nCLEARWTR, hi) + par.cldprob;
    free((void*)CLEARWPROB);
    free((void*)CLEARWTR);

  } else {

    free((void*)CLEARWTR);
    alloc((void**)&WATERPROB, _NC_, sizeof(float));
    wclr_max = 1;

  }

  #ifdef DEBUG
  db_write_image_1D((void**)WATERPROB, "WATERPROB", _NX_, _NY_, sizeof(float), 4);
  #endif


  A = hightemp+4.0;
  B = hightemp+4.0-lowtemp+4.0;

  alloc((void**)&LANDPROB, _NC_, sizeof(float));
  #ifdef DEBUG
  float *TEMPPROB = NULL, *CIRPROB = NULL;
  alloc((void**)&TEMPPROB, _NC_, sizeof(float));
  alloc((void**)&CIRPROB,  _NC_, sizeof(float));
  #endif

  for (p=0; p<_NC_; p++){

    if (par.band_bt){
      tempprob = (A-TOA.bt[p]/_TS_)/B;
      if (tempprob < 0) tempprob = 0.0;
    } else {
      tempprob = 1.0;
    }

    //if (_B_.cir  != -1) cirprob = (TOA.ref[_B_.cir][p]/_RS_)/0.04; else cirprob = 0.0;
    cirprob = 0.0;

    #ifdef DEBUG
    TEMPPROB[p] = tempprob;
    CIRPROB[p]  = cirprob;
    #endif

    if (QAI.off[p]){
      LANDPROB[p] = 0;
    } else {
      LANDPROB[p] = tempprob*VPROB[p] + cirprob;
    }

  }

  free((void*)VPROB);

  #ifdef DEBUG
  db_write_image_1D((void**)LANDPROB, "LANDPROB", _NX_, _NY_, sizeof(float), 4);
  db_write_image_1D((void**)TEMPPROB, "TEMPPROB", _NX_, _NY_, sizeof(float), 4);
  db_write_image_1D((void**)CIRPROB,  "CIRPROB",  _NX_, _NY_, sizeof(float), 4);
  free((void*)TEMPPROB); free((void*)CIRPROB);
  #endif


  // dynamic cloud detection threshold over land
  alloc((void**)&CLEARLPROB, nland, sizeof(float));

  for (p=0, pos=0; p<_NC_; p++){
    if (LAND[p]){
      CLEARLPROB[pos] = LANDPROB[p];
      pos++;
    }
  }

  lclr_max = quantile(CLEARLPROB, nland, hi) + par.cldprob;
  free((void*)CLEARLPROB);

  #ifdef DEBUG
  printf("\n\nclr_max land/water/bt-cold: %f/%f/%f\n", lclr_max, wclr_max, lowtemp-35);
  #endif

  alloc((void**)&CLOUD, _NC_, sizeof(bool));

  // final cloud layer = 
    // cloud over land thin OR 
    // cloud over water OR 
    // high prob cloud (land) 
    //OR extremly cold cloud
  for (p=0; p<_NC_; p++){
    pcp = PCP[p];
    if (par.band_bt) bt = TOA.bt[p]/_TS_; else bt = 0.0;

    // Zhu et al., 2015 modification :::
    // Frantz et al., 2015 modification :::
    CLOUD[p] = ((pcp && !QAI.wtr[p] && LANDPROB[p] > lclr_max) ||
          (pcp && QAI.wtr[p] && WATERPROB[p] > wclr_max) ||
          //(landprob > 99 && !QAI.wtr[p] && BRIGHT[p]) ||
          bt < (lowtemp-35) || QAI.sat[p]) && BRIGHT[p];
    //if (_B_.cir  != -1 && TOA.ref[_B_.cir][p]/_RS_ > 0.01) CLOUD[p] = true;

  }


  free((void*)BRIGHT);
  free((void*)LANDPROB);
  free((void*)WATERPROB);
  free((void*)PCP);


  // set clear pixel to cloud if 5 or more cloud pixels in neighborhood + 
  // set boundary to clear
  majorfill(CLOUD, QAI);
  for (p=0; p<_NC_; p++){
    if (QAI.snw[p]) CLOUD[p] = false;
  }
  *ncloud = bsum(CLOUD, _NC_);


  #ifdef CLOCK
  proctime_print("cloud probability computation", TIME);
  #endif

  #ifdef DEBUG
  printf("temperatur quantiles (potential clear): %.2f %.2f\n", lowtemp, hightemp);
  db_write_image_1D((void**)CLOUD, "CLOUD-in-Fmask", _NX_, _NY_, sizeof(bool), 1);
  #endif

  // revised cloud coverage
  *cc = 100.0 * (float)*ncloud/npix;

  *OCLOUD = CLOUD;
  *lowt = lowtemp;
  *hight = hightemp;

  return;
}


/** Fast hybrid greyscale reconstruction algorithm
+++ This function is an efficient implementation of a greyscale flood-fill
+++ procesure used in Fmask to identify potential shadows.
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+++ Vincent, L. (1993). Morphological Grayscale Reconstruction in Image An
+++ alysis: Applications and Efficient Algorithms. IEEE Transactions on Im
+++ age Processing, 2 (2), 176-201.
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
--- MASK:   mask image
--- MARKER: marker image
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void greyscale_reconstruction(float *MASK, float *MARKER){
int i, j, p, ii, jj, np, np1[4], np2[4], k;
float min, max;
queue_t *fifo;

  np1[0] = -_NX_-1; np1[1] = -_NX_; np1[2] = -_NX_+1; np1[3] = -1;
  np2[0] = _NX_+1; np2[1] = _NX_; np2[2] = _NX_-1; np2[3] = 1;


  // initialize queue
  if ((fifo = create_queue()) == NULL){
    printf("failed to create new queue!\n"); exit(1);}


  /* set border of marker image to mask value
     set rest of image to maximum of mask */
  max = SHRT_MIN;  
  for (p=0; p<_NC_; p++){
    if (MASK[p] > max) max = MASK[p];
  }
  
  
  for (i=0, p=0; i<_NY_; i++){
  for (j=0; j<_NX_; j++, p++){
    
    if (i == 0 || j == 0 || i == _NY_-1 || j == _NX_-1){
      MARKER[p] = MASK[p];
    } else {
      MARKER[p] = max;
    }

  }
  }


  // sequential reconstruction in raster order
  for (i=1; i<_NY_-1; i++){
  for (j=1; j<_NX_-1; j++){

    p = i*_NX_+j; 
    min = MARKER[p];
    
    for (k=0; k<4; k++){
      np = p+np1[k];
      if (MARKER[np] < min) min = MARKER[np];
    }

    if (min > MASK[p]) MARKER[p] = min; else MARKER[p] = MASK[p];

  }
  }


  // sequential reconstruction in anti-raster order
  for (i=_NY_-2; i>=1; i--){
  for (j=_NX_-2; j>=1; j--){

    p = i*_NX_+j; 
    min = MARKER[p];

    for (k=0; k<4; k++){
      np = p+np2[k];
      if (MARKER[np] < min) min = MARKER[np];
    }

    if (min > MASK[p]) MARKER[p] = min; else MARKER[p] = MASK[p];

    for (k=0; k<4; k++){
      np = p+np2[k];
      if (MARKER[np] > MARKER[p] && MARKER[np] > MASK[np]){

        if (enqueue(fifo, i, j) == FAILURE){
          printf("Failed to enqueue another coord. pair!\n"); 
          exit(1);
        }
        continue;

      }
    }

  }
  }

  //fprintf("step 2: done. queue length: %d\n", fifo->length);

  // queue propagation
  while (dequeue(fifo, &i, &j) == SUCCESS){

    p = _NX_*i+j;

    for (ii=-1; ii<=1; ii++){
    for (jj=-1; jj<=1; jj++){

      if (ii == 0 && jj == 0) continue;

      np = (i+ii)*_NX_+j+jj;

      if (MARKER[np] > MARKER[p] && MASK[np] != MARKER[np]){

        if (MARKER[p] > MASK[np]) MARKER[np] = MARKER[p]; else MARKER[np] = MASK[np];
        
        if (enqueue(fifo, i+ii, j+jj) == FAILURE){
          printf("Failed to enqueue another coord. pair!\n");
          exit(1);
        }

      }

    }
    }

  }  

  //fprintf("step 3: done. queue length: %d\n", fifo->length);

  // free queue's memory
  destroy_queue(fifo);
  fifo = NULL;

  return;
}


/** Shadow probability and potential shadow pixels
+++ This function computes the shadow probability and return potential 
+++ shadow pixels.
--- nland:  number of clear-sky land pixels
--- LAND:   Clear-sky land pixels (deallocated within)
--- TOA:    TOA reflectance
--- CLOUD:  Cloud mask
--- QAI:    Quality Assurance Information
--- PSP:    Potential Shadow Pixels (returned)
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void shadow_probability(int nland, bool *LAND, int2s **TOA, bool *CLOUD, qai_t QAI, int2u **SPROB){
int i, j, p, pos, g, gx, gy;
float *CLEARB4 = NULL, *CLEARB5 = NULL;
float *TOA4 = NULL, *TOA5 = NULL, *NIR = NULL, *SWIR = NULL;
int2u *DIST;
float maxdist;
float bck4, bck5, shadprob;
int2u *SHADPROB = NULL;
float lo = 0.175;

  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  // Zhu et al., 2015 modification :::
  // use NIR and SWIR

  alloc((void**)&TOA4, _NC_, sizeof(float));
  alloc((void**)&TOA5, _NC_, sizeof(float));

  for (p=0; p<_NC_; p++){
    TOA4[p] = TOA[_B_.nir][p]/_RS_;
    TOA5[p] = TOA[_B_.swir1][p]/_RS_;
  }

  // get background value
  alloc((void**)&CLEARB4, nland, sizeof(float));
  alloc((void**)&CLEARB5, nland, sizeof(float));

  for (p=0, pos=0; p<_NC_; p++){
    if (LAND[p]){
      CLEARB4[pos] = TOA4[p];
      CLEARB5[pos] = TOA5[p];
      pos++;
    }
  }
  free((void*)LAND);

  bck4 = quantile(CLEARB4, nland, lo);
  bck5 = quantile(CLEARB5, nland, lo);

  free((void*)CLEARB4);
  free((void*)CLEARB5);

  // compute shadow probability only for pixels that 
  // are close enough to clouds
  alloc((void**)&DIST, _NC_, sizeof(int2u));
  dist_transform(CLOUD, DIST);
  maxdist = 12000 * tan(acos(atc.cosszen[0]))/meta.res;

  #ifdef DEBUG
  printf("maximum distance for cloud shadows: %.0f", maxdist);
  #endif

  for (i=0, p=0; i<_NY_; i++){
  for (j=0; j<_NX_; j++, p++){
    
    if (i == 0 || i == _NY_-1 || j == 0 || j == _NX_-1 || 
        QAI.off[p] || DIST[p] > maxdist){
      TOA4[p] = bck4;
      TOA5[p] = bck5;
    }

  }
  }
  
  free((void*)DIST);

  alloc((void**)&NIR,  _NC_, sizeof(float));
  alloc((void**)&SWIR, _NC_, sizeof(float));

  // use following lines if compiling with efence...,
  // greyscale_reconstruction doesn't work in efence
  //for (p=0; p<_NC_; p++) NIR[p] = TOA[_B_.nir][p]/_RS_;
  //for (p=0; p<_NC_; p++) SWIR[p] = TOA[_B_.swir1][p]/_RS_;

  greyscale_reconstruction(TOA4, NIR);
  greyscale_reconstruction(TOA5, SWIR);

  #ifdef DEBUG
  db_write_image_1D((void**)NIR,  "FLOODFILL-NIR",  _NX_, _NY_, sizeof(float), 4);
  db_write_image_1D((void**)SWIR, "FLOODFILL-SWIR", _NX_, _NY_, sizeof(float), 4);
  #endif

  for (p=0; p<_NC_; p++){
    NIR[p]  -= TOA4[p];
    SWIR[p] -= TOA5[p];
  }

  #ifdef DEBUG
  db_write_image_1D((void**)NIR,  "FLOODFILL-NIR-PROB",  _NX_, _NY_, sizeof(float), 4);
  db_write_image_1D((void**)SWIR, "FLOODFILL-SWIR-PROB", _NX_, _NY_, sizeof(float), 4);
  #endif

  free((void*)TOA4); free((void*)TOA5);


  alloc((void**)&SHADPROB, _NC_, sizeof(int2u));

  for (i=0, p=0; i<_NY_; i++){
  for (j=0; j<_NX_; j++, p++){
    gx = tocg(j); gy = tocg(i); g = gy*atc.ngx+gx;
    if (NIR[p] < SWIR[p]) shadprob = NIR[p]; else shadprob = SWIR[p];
    if (shadprob > 0) SHADPROB[p] = shadprob*_RS_;
  }
  }

  #ifdef DEBUG
  db_write_image_1D((void**)SHADPROB, "SHADOWPROB",  _NX_, _NY_, sizeof(int2u), 12);
  #endif

  free((void*)NIR); free((void*)SWIR);

  #ifdef CLOCK
  proctime_print("shadow probability computation", TIME);
  #endif

  *SPROB = SHADPROB;
}


/** Sentinel-2 surrogate for cloud probability
+++ This function separates PCPs on the basis of parallax effects in Sen-
+++ tinel-2 images.
--- npix:   number of valid image pixels
--- ncloud: number of cloud pixels (returned)
--- cc:     revised total cloud cover in % (returned)
--- TOA:    TOA reflectance
--- QAI:    Quality Assurance Information
--- BRIGHT: Bright pixels (is deallocated within)
--- PCP:    Potential Cloud Pixels (is deallocated within)
--- CLD:    Cloud mask (returned)
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int cloud_parallax(float npix, int *ncloud, float *cc, int2s **TOA, qai_t QAI, bool *BRIGHT, bool *PCP, bool **CLD){
int i, j, p, ii, jj, ni, nj, np;
int nx, ny, nc;
int b0 = _B_.bnir-1;
int b1 = _B_.bnir;
int b2 = _B_.nir;
int nk = 5;
float **kernel = NULL;
float sum, num;
double mx, my, varx, vary, vx, vy, cv, k;
bool *CLOUD;
float *GAUSS = NULL;
float **RATIO = NULL;
float *CDI = NULL;
bool pcp, valid;
bool *cloud = NULL;
queue_t *fifo;


  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  nx = _NX_/2; ny = _NY_/2; nc = nx*ny;
  

  /** Convolute BNIR with gaussian kernel **/

  if (gauss_kernel(nk, 0.5, &kernel) != SUCCESS){
    printf("Could not generate kernel. "); return FAILURE;}

  alloc((void**)&GAUSS, _NC_, sizeof(float));

  for (i=0, p=0; i<_NY_; i++){
  for (j=0; j<_NX_; j++, p++){
    
    if (QAI.off[p]) continue;

    sum = num = 0;

    for (ii=0; ii<nk; ii++){
    for (jj=0; jj<nk; jj++){

      ni = -(nk-1)/2 + ii + i; nj = -(nk-1)/2 + jj + j;
      if (ni < 0 || ni >= _NY_ || nj < 0 || nj >= _NX_) continue;
      np = ni*_NX_+nj;
      
      if (QAI.off[np]) continue;

      sum += (TOA[b1][np]/_RS_)*kernel[ii][jj];
      num += kernel[ii][jj];

    }
    }

    if (num > 0) GAUSS[p] = sum/num;

  }
  }

  free_2D((void**)kernel, nk);
  
  #ifdef DEBUG
  db_write_image_1D((void**)GAUSS,  "NIR-GAUSS", _NX_, _NY_, sizeof(float), 4);
  #endif


  /** Compute NIR Ratios at reduced spatial resolution.
  +++  This is only done for PCPs. **/

  alloc_2D((void***)&RATIO, 2, nc, sizeof(float));

  for (i=0, p=0; i<_NY_; i+=2){
  for (j=0; j<_NX_; j+=2, p++){
    
    if (QAI.off[i*_NX_+j]) continue;

    sum = num = 0;
    pcp = false;

    for (ii=0; ii<2; ii++){
    for (jj=0; jj<2; jj++){

      ni = i+ii; nj = j+jj;
      if (ni >= _NY_ || nj >= _NX_) continue;
      np = ni*_NX_+nj;

      if (QAI.off[np]) continue;

      sum += GAUSS[np];
      num++;

      if (PCP[np]) pcp = true;

    }
    }

    if (num > 0 && pcp){
      RATIO[0][p] = (TOA[b0][i*_NX_+j]/_RS_)/(TOA[b2][i*_NX_+j]/_RS_);
      RATIO[1][p] = (sum/num)/(TOA[b2][i*_NX_+j]/_RS_);
    }

  }
  }

  free((void*)GAUSS);

  #ifdef DEBUG
  db_write_image_2D((void**)RATIO,  "NIR-RATIOS", 2, nx, ny, sizeof(float), 4);
  #endif


  /** Compute Cloud Displacement Index CDI
  +++ Texture (variance) is obtained from both ratios and a normalied 
  +++ differenced variance ratio (CDI) is computed thereof. **/

  alloc((void**)&CDI, nc, sizeof(float));
  for (p=0; p<nc; p++) CDI[p] = _NODATA_;

  for (i=0, p=0; i<ny; i++){
  for (j=0; j<nx; j++, p++){

    if (RATIO[0][p] == 0) continue;

    mx = my = vx = vy = cv = k = 0;

    for (ii=-3; ii<=3; ii++){
    for (jj=-3; jj<=3; jj++){

      ni = i+ii; nj = j+jj;
      if (ni < 0 || ni >= ny || nj < 0 || nj >= nx) continue;
      np = ni*nx+nj;

      if (RATIO[0][np] == 0) continue;

      if (++k == 1){
        mx = RATIO[0][np]; my = RATIO[1][np];
      } else {
        covar_recurrence(RATIO[0][np], RATIO[1][np],
          &mx, &my, &vx, &vy, &cv, k);
      }

    }
    }

    if (k > 2){
      varx = variance(vx, k);
      vary = variance(vy, k);
      CDI[p] = (varx-vary)/(varx+vary);
    }

  }
  }

  free_2D((void**)RATIO, 2);

  #ifdef DEBUG
  db_write_image_1D((void**)CDI, "CLOUD-DISPLACEMENT-INDEX", nx, ny, sizeof(float), 4);
  #endif


  /** Separate clouds from built-up using CDI
  +++ Clouds are eroded by one pixel **/

  alloc((void**)&cloud, nc, sizeof(bool));
  
  for (i=0, p=0; i<ny; i++){
  for (j=0; j<nx; j++, p++){

    if (CDI[p] != _NODATA_ && CDI[p] <= -0.5){

      valid = true;

      for (ii=-1; ii<=1; ii++){
      for (jj=-1; jj<=1; jj++){

        ni = i+ii; nj = j+jj;
        if (ni < 0 || ni >= ny || nj < 0 || nj >= nx) continue;
        np = ni*nx+nj;

        if (CDI[np] == _NODATA_ || CDI[np] >  -0.5) valid = false;

      }
      }

      if (valid) cloud[p] = true;
      
    }

  }
  }

  #ifdef DEBUG
  db_write_image_1D((void**)cloud, "PCP-cloud-conf", nx, ny, sizeof(bool), 1);
  #endif
  

  /** Region-growing 1. Put all cloud edge pixels, from which
  +++ clouds can be grown in FIFO **/
  
  if ((fifo = create_queue()) == NULL){
    printf("failed to create new queue!\n"); exit(1);}

  for (i=0, p=0; i<ny; i++){
  for (j=0; j<nx; j++, p++){

    if (!cloud[p]) continue;

      for (ii=-1; ii<=1; ii++){
      for (jj=-1; jj<=1; jj++){

        if (ii==0 && jj==0) continue;

        ni = i+ii; nj = j+jj;
        if (ni < 0 || ni >= ny || nj < 0 || nj >= nx) continue;
        np = ni*nx+nj;

        if (!cloud[np] && CDI[np] < -0.25 && CDI[np] >= -1){
          if (enqueue(fifo, i, j) == FAILURE){
            printf("Failed to allocate enqueue memory\n"); exit(1);}
          ii = jj = 10;
        }

      }
      }

  }
  }


  /** Region-growing 2. Use FIFO to grow cloud objects.
  +++ All connected pixels with fairly low CDI become clouds. **/

  while (dequeue(fifo, &i, &j) == SUCCESS){

    p = nx*i+j;

    for (ii=-1; ii<=1; ii++){
    for (jj=-1; jj<=1; jj++){

      if (ii==0 && jj==0) continue;

      ni = i+ii; nj = j+jj;
      if (ni < 0 || ni >= ny || nj < 0 || nj >= nx) continue;
      np = nx*ni+nj;

      if (!cloud[np] && CDI[np] < -0.25 && CDI[np] >= -1){
        cloud[np] = true;
        if (enqueue(fifo, ni, nj) == FAILURE){
          printf("Failed to allocate enqueue memory\n"); exit(1);}
      }

    }
    }

  }

  free((void*)CDI);
  destroy_queue(fifo);
  fifo = NULL;

  #ifdef DEBUG
  db_write_image_1D((void**)cloud, "PCP-cloud-grown", nx, ny, sizeof(bool), 1);
  #endif


  /** Restore original resolution **/

  alloc((void**)&CLOUD, _NC_, sizeof(bool));

  // final cloud layer
  for (i=0, p=0; i<ny; i++){
  for (j=0; j<nx; j++, p++){

    if (cloud[p]){

      for (ii=i*2; ii<=(i*2+1); ii++){
      for (jj=j*2; jj<=(j*2+1); jj++){
          CLOUD[ii*_NX_+jj] = true;
      }
      }
    
    }
    
  }
  }

  free((void*)cloud);


  /** Add saturated, remove dark pixels */

  for (p=0; p<_NC_; p++){
    // Frantz et al., 2015 modification :::
    if (QAI.sat[p]) CLOUD[p] = true;
    if (CLOUD[p] && !BRIGHT[p]) CLOUD[p] = false;
  }

  free((void*)BRIGHT);
  free((void*)PCP);

  // set clear pixel to cloud if 5 or more cloud pixels in neighborhood
  //majorfill(CLOUD, QAI.off);

  #ifdef DEBUG
  db_write_image_1D((void**)CLOUD, "CLOUD-in-Fmask", _NX_, _NY_, sizeof(bool), 1);
  #endif

  // revised cloud coverage
  *ncloud = bsum(CLOUD, _NC_);
  *cc = 100.0 * (float)*ncloud /npix;
  *CLD = CLOUD;

  #ifdef CLOCK
  proctime_print("cloud from motion computation", TIME);
  #endif

  return SUCCESS;
}


/** Identify cloud objects
+++ This function segments the cloud layer. Small clouds (< nmin) are eli-
+++ minated. The segmentation result (unique ID per cloud), and the number
+++ of clouds and cloud pixels is returned.
--- CLOUD:  Cloud mask
--- OBJ:    Cloud segmentation (returned)
--- SIZE:   Cloud sizes (returned)
--- nobj:   Number of clouds (returned)
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void cloud_objects(bool *CLOUD, int **OBJ, int **SIZE, int *nobj){
int *CCL = NULL; // connected component labelling
int *CCS = NULL; // connected component object size
int id, p, nmin, no;

  // delete all objects smaller than 2700m², i.e. 3 Landsat pixels
  //nmin = 2700/(meta.res*meta.res);
  nmin = 3;

  alloc((void**)&CCL, _NC_, sizeof(int));

  // to ensure that 1st cloud gets ID=1
  CLOUD[0] = false;

  if ((no = connectedcomponents(CLOUD, CCL)) > 0){

    alloc((void**)&CCS, no, sizeof(int));

    for (p=0; p<_NC_; p++){
      if (!CLOUD[p]) continue;
      id = CCL[p]; 
      CCS[id-1]++;
    }

    for (p=0; p<_NC_; p++){
      if (!CLOUD[p]) continue;
      id = CCL[p];
      if (CCS[id-1] < nmin) CLOUD[p] = 0;
    }

    for (p=0; p<_NC_; p++) CCL[p] = 0;
    free((void*)CCS);

    if ((no = connectedcomponents(CLOUD, CCL)) > 0){

      alloc((void**)&CCS, no, sizeof(int));

      for (p=0; p<_NC_; p++){
        if (!CLOUD[p]){ CCL[p] = 0; continue;}
        id = CCL[p]; 
        CCS[id-1]++;
      }

    } else free((void*)CCL);

  } else free((void*)CCL);


  #ifdef DEBUG
  printf("number of cloud objs: %d\n", no);
  db_write_image_1D((void**)CCL, "CCL", _NX_, _NY_, sizeof(int), 3);
  #endif

  *nobj = no;
  *OBJ  = CCL;
  *SIZE = CCS;

  return;
}


/** Project a cloud to the ground --> shadow position
+++ Due to the scanning geometry, a cloud is projected to the image plane
+++ in a ~perpendicular angle to the nadir line, i.e. the viewing azimuth.
+++ This is not exactly correct for detectors with multiple lines, but it
+++ should be sufficiently precise for this purpose. Thus, the cloud needs
+++ to be shifted towards the nadir line in across-scan direction in order
+++ to find the 'real' cloud position. The distance depends on the assumed
+++ cloud pixel height and viewing zenith. The shift is larger for large 
+++ zenith angles and high clouds.
+++ Afterwards, the cloud needs to be projected to the image plane in sun
+++ azimuth direction. The distance depends on the assumed cloud pixel 
+++ height and sun zenith. The shift is larger for large zenith angles and
+++ high clouds. The function returns the position of the casted shadow in 
+++ x/y pixel coords.
--- h:      height of object (in m above ground) 
--- x:      location of object in image
--- y:      location of object in image
--- newx:   location of object in real world (returned)
--- newy:   location of object in real world (returned)
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void shadow_position(float h, int x, int y, int *newx, int *newy){
int gx, gy, g;
float dist_across, dist_cast;
float dx_across, dx_cast;
float dy_across, dy_cast;
 
  gx = tocg(x);
  gy = tocg(y);
  g = gy*atc.ngx+gx;

  dist_across = h * atc.cg[g].tanvzen / meta.res;
  dist_cast   = h * atc.cg[g].tanszen / meta.res;

  dx_across = dist_across*atc.cg[g].sinvazi;
  dx_cast   = dist_cast*atc.cg[g].sinsazi;

  dy_across = dist_across*atc.cg[g].cosvazi;
  dy_cast   = dist_cast*atc.cg[g].cossazi;

  *newx = (int) round(x + dx_across - dx_cast);
  *newy = (int) round(y - dy_across + dy_cast);

  return;
}


/** Shadow matching
+++ Knowing the view and solar geometry (and cloud height), we can predict
+++ the location where the shadow is projected to the ground. This projec-
+++ tion is matched with the shadow probability, and if it matches, a sha-
+++ dow is drawn. As the cloud height is not known precisely, the possible
+++ cloud base height range is iterated. The base height range can be nar-
+++ rowed down with a temperature-based prediction. The individual cloud
+++ pixel height is also predicted from temperature, i.e. a cloud DEM is 
+++ built before projecting the cloud to the image plane. If a temperature
+++ band is not available, the full cloud base height range is tested and
+++ the clouds are treated as flat planes.
--- lowtemp:   lower  temperature quantile (returned)
--- hightemp:  higher temperature quantile (returned)
--- QAI:    Quality Assurance Information
--- CLOUD:  Cloud mask
--- SLP:    Topographic slope
--- PSP:    Potential Shadow Pixels (deallocated within)
--- BT:     Brightness Temperature (deallocated within)
--- SHD:    Shadow mask (returned)
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void shadow_matching(float lowtemp, float hightemp, qai_t QAI, bool *CLOUD, int2u *SLP, int2u *SPROB, int2s *BT, bool **SHD){
int p, i, j, k, influence = 8;
int nobj, id, size_max = 0, size;
int *P = NULL, *best_P = NULL;
float *qtemp = NULL;
float radius, core, basetemp;
float base_min, base_max, base_step, base;
float best_match, match, shadow, total;
int x, y;
float height;
float wlapse = 0.0065, dlapse = 0.0098; // wet & dry adiabatic lapse rate kelvin/m
bool *SHADOW = NULL;
int *CCL = NULL, *SIZE = NULL;
cloud_t C;

  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  #ifdef DEV
  printf("shadow match increased to .95, was .85. 2 occurences.\n");
  #endif



  /** make cloud objects and delete small clouds **/
  cloud_objects(CLOUD, &CCL, &SIZE, &nobj);

  alloc((void**)&SHADOW, _NC_, sizeof(bool));

  /** skip shadow matching if there is no cloud left **/
  if (nobj == 0){

    if (par.band_bt) free((void*)BT);
    free((void*)SPROB);
    
    #ifdef CLOCK
    proctime_print("shadow matching", TIME);
    #endif

    #ifdef DEBUG
    printf("\nno remaining cloud object. skip shadow matching.\n");
    #endif

    *SHD = SHADOW;
    return;
  }



  #ifdef DEBUG
  float *SHIFT = NULL; alloc((void**)&SHIFT, _NC_, sizeof(float));
  float *SHD_MATCH = NULL; alloc((void**)&SHD_MATCH, _NC_, sizeof(float));
  #endif





  
  if ((C.X = (int**) malloc(nobj*sizeof(int*))) == NULL){
    printf("unable to allocate memory!\n"); exit(1);}
  for (id=0; id<nobj; id++){
    if ((C.X[id] = (int*) malloc(SIZE[id]*sizeof(int))) == NULL){
      printf("unable to allocate memory!\n"); exit(1);}
  }

  if ((C.Y = (int**) malloc(nobj*sizeof(int*))) == NULL){
    printf("unable to allocate memory!\n"); exit(1);}
  for (id=0; id<nobj; id++){
    if ((C.Y[id] = (int*) malloc(SIZE[id]*sizeof(int))) == NULL){
      printf("unable to allocate memory!\n"); exit(1);}
  }

  if ((C.BT = (float**) malloc(nobj*sizeof(float*))) == NULL){
    printf("unable to allocate memory!\n"); exit(1);}
  for (id=0; id<nobj; id++){
    if ((C.BT[id] = (float*) malloc(SIZE[id]*sizeof(float))) == NULL){
      printf("unable to allocate memory!\n"); exit(1);}
  }

  int *K;
  alloc((void**)&K, nobj, sizeof(int));
  

  // copy relevant info to smaller arrays
  for (i=0, p=0; i<_NY_; i++){
  for (j=0; j<_NX_; j++, p++){
    if ((id = CCL[p]) > 0){
      // only do for each 2nd pixel in case of Sentinel-2
      if (meta.mission == 1 && (i % 2 != 0 || j % 2 != 0)) continue;
      id--;
      C.X[id][K[id]] = j;
      C.Y[id][K[id]] = i;
      if (par.band_bt) C.BT[id][K[id]] = BT[p]/_TS_;
      K[id]++;
    }
  }
  }

  // copy new size
  if (meta.mission == 1){
    for (id=0; id<nobj; id++) SIZE[id] = K[id];
  }
  
  free((void*)K);


  /** Determine the step size for the base height iteration
  +++ Pick height that coincides with a horizontal shift of two pixels. 
  +++ Use the sun zenith at scene center as approximation. **/
  base_step = 2*meta.res / atc.cg[(atc.ngy/2)*atc.ngx+(atc.ngx/2)].tanszen;

  // increase by two in case of Sentinel-2
  if (meta.mission == 1) base_step *= 2;
  
  #ifdef DEBUG
  printf("\nbase height iteration in %.2fm steps\n", base_step);
  #endif

  for (id=0; id<nobj; id++){
    if (SIZE[id] > size_max) size_max = SIZE[id];
  }

  #ifdef DEBUG
  printf("max. object size: %d\n", size_max);
  #endif


  if (par.band_bt) alloc((void**)&qtemp, size_max, sizeof(float));
  alloc((void**)&P,       size_max, sizeof(int));
  alloc((void**)&best_P,  size_max, sizeof(int));


  for (id=0; id<nobj; id++){
//if (id != 2710 && id != 2119 && id != 3076) continue;
    // assume object is round
    size = SIZE[id];
    radius = sqrt(size/M_PI);

    // percent of cloud core area
    if (radius > influence){
      core = (radius-influence)*(radius-influence)/(radius*radius);
    } else {
      core = 0.0;
    }

    for (k=0; k<size; k++){
      if (par.band_bt) qtemp[k] = C.BT[id][k];
    }

    /** Predict min and max cloud base height:
    +++ The cloud temperature and the dry adiabatic lapse rate are used
    +++ to narrow down the base height range for better performance. If 
    +++ there is no temperature band, use the full range. **/
    if (par.band_bt){

      /** edge of the cloud is influenced by the warm surface
      +++ Pixels that are too warm get the value from the cloud core. **/
      basetemp = quantile(qtemp, size, core);
      for (k=0; k<size; k++){
        if (C.BT[id][k] > basetemp) C.BT[id][k] = basetemp;
      }

      /** base heigt is estimated from the warmest cloud temperatures. **/
      if ((base_min = (lowtemp-4-basetemp)/dlapse) < 200) base_min = 200;
      if ((base_max = 1000*(hightemp+4-basetemp)) > 12000) base_max = 12000;
      if (base_min > base_max) continue;
      
    } else {
      base_min = 200;
      base_max = 12000;
    }


    /** Base height iteration:
    +++ Lift the cloud up across the possible base height range and match
    +++ the casted shadow with the potential shadow layer. **/
    for (base=base_min, best_match=0; base<=base_max; base+=base_step){

      for (k=0, shadow=0, total=0; k<size; k++){

        /** Predict cloud pixel height:
        +++ A cloud DEM is used for a more exact calculation of projected
        +++ shadow position. The height is predicted using the temperature
        +++ band and the wet adiabatic lapse rate, plus the cloud base
        +++ height. If there is no temperature band, the cloud is assumed
        +++ to be a flat plate. **/
        if (par.band_bt){
          height = (basetemp-C.BT[id][k])/wlapse + base;
        } else {
          height = base;
        }

        /** Position of projected shadow:
        +++ Copmpute the position of the projected shadow as a function of
        +++ view and sun geometry. **/
        shadow_position(height, C.X[id][k], C.Y[id][k], &x, &y);

        if (y < 0 || y >= _NY_ || x < 0 || x >= _NX_){
          P[k] = -1; continue;}

        P[k] = p = _NX_*y + x; 

        #ifdef DEBUG
        SHIFT[p] = base;
        #endif

        /** Simplified match:
        +++ A shadow is matched if it is a potential shadow, but not a 
        +++ cloud. The shadow matching 'runs' into big clouds if clouds
        +++ are also permitted. The match is measured relative to the
        +++ complete shifted object, excluding the original cloud. **/
        if (SPROB[p]/_RS_ > par.shdprob && !CLOUD[p] && !(SLP[p] == 0 && QAI.wtr[p])) shadow += (SPROB[p]/_RS_)/(0.075);
        if (CCL[p] != id+1) total++;

      }

      /** Matching measure:
      +++ The match is expressed in fractions between 0...1 **/
      if (total > 0) match = shadow/total; else match = 0;


      /** Excellent match:
      +++ If the match is better than 95%, draw a shadow. This is a greedy
      +++ implementation when compared to the original Fmask, but we found
      +++ that the shadow matching terminated too early in many cases. We
      +++ do not terminate the matching once the match decreases again.**/
      if (match >= 0.85){
        for (k=0; k<size; k++){
          if ((p = P[k]) != -1){
            SHADOW[p] = true;
            #ifdef DEBUG
            SHD_MATCH[p] = match;
            #endif
          }
        }
      }

      /** Best match:
      +++ Record the position and value of the best match. **/
      if (match > best_match){
        best_match = match;
        for (k=0; k<size; k++) best_P[k] = P[k];
      }

    }


    /** Best match:
    +++ If best match is less than 85%, draw the shadow anyway, even if
    +++ the match is not very high. This is a greedy implementation when
    +++ compared to the original Fmask, but every cloud has a shadow, so
    +++ we assume that the best match is in the right position. If the
    +++ best match is better than 85%, it was already drawn. **/
    if (best_match < 0.85 && best_match > 0){
      for (k=0; k<size; k++){
        if ((p = best_P[k]) != -1){
          SHADOW[p] = true;
          #ifdef DEBUG
          SHD_MATCH[p] = best_match;
          #endif
        }
      }
    }

  }

  
  buffer(SHADOW, 3);



  for (id=0; id<nobj; id++){
    free((void*)C.X[id]);
    free((void*)C.Y[id]);
    free((void*)C.BT[id]);
  }
  free((void*)C.X);
  free((void*)C.Y);
  free((void*)C.BT);

  free((void*)qtemp);
  free((void*)P);
  free((void*)best_P);

  free((void*)CCL);
  free((void*)SIZE);

  #ifdef DEBUG
  db_write_image_1D((void**)SHIFT, "CLOUDSHIFT", _NX_, _NY_, sizeof(float), 4);
  db_write_image_1D((void**)SHD_MATCH, "SHADOW_MATCH", _NX_, _NY_, sizeof(float), 4);
  db_write_image_1D((void**)SHADOW, "SHADOW", _NX_, _NY_, sizeof(bool), 1);
  free((void*)SHIFT);
  free((void*)SHD_MATCH);
  #endif


  if (par.band_bt) free((void*)BT);
  free((void*)SPROB);
  
  #ifdef CLOCK
  proctime_print("shadow matching", TIME);
  #endif

  *SHD = SHADOW;
  return;
}


/** Cloud module, main function
+++ This function is the 1st core module of FORCE L2PS, and is a modified 
+++ implementation of the Function of Mask code, better known as Fmask. 
+++ Cloud and cloud shadow detection is done in this module.
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+++ Zhu, Z. & Woodcock, C.E. (2012). Object-based cloud and cloud shadow d
+++ etection in Landsat imagery. Remote Sensing of Environment, 118, 83-94
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
--- TOA:    TOA reflectance
--- QAI:    Quality Assurance Information (modified)
--- TOP:    Topographic Derivatives
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int cloud_module(toa_t TOA, qai_t QAI, top_t TOP){
int nclear, nland, ncloud, p;
float npix, lowtemp = -1.0, hightemp = -1.0;
float cc;
float *VPROB = NULL;
int2u *SPROB = NULL;
bool *BRIGHT = NULL;
bool *PCP = NULL;
bool *CLEAR = NULL, *LAND = NULL, *CLOUD = NULL;
bool *SHADOW = NULL;


  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif


  potential_cloud(&npix, &nclear, &nland, 
            TOA, QAI, TOP.slp, &CLEAR, &LAND, &BRIGHT, &VPROB, &PCP);

  // more than 0.1% clear pixels? -> compute probabilities
  if ((100.0*nclear/npix) > 0.1){

    if (meta.mission == MISSION_LANDSAT){
      cloud_probability(npix, nclear, nland, &ncloud, &cc, &lowtemp, &hightemp,
          CLEAR, LAND, TOA, QAI, BRIGHT, VPROB, PCP, &CLOUD);
    } else if (meta.mission == MISSION_SENTINEL2){
      free((void*)CLEAR); free((void*)VPROB);
      cloud_parallax(npix, &ncloud, &cc, TOA.ref, QAI, BRIGHT, PCP, &CLOUD);
    }

    // less than 80% of max. allowable cloud cover? -> shadow matching
    if (cc <= par.maxcc*0.8){

      // if there is no cloud, there is no shadow
      if (ncloud > 0){
        shadow_probability(nland, LAND, TOA.ref, CLOUD, QAI, &SPROB);
        shadow_matching(lowtemp, hightemp, QAI, CLOUD, TOP.slp, SPROB, TOA.bt, &SHADOW);
      } else {
        alloc((void**)&SHADOW, _NC_, sizeof(bool));
        if (par.band_bt) free((void*)TOA.bt);
        free((void*)LAND);
      }

      // create the cloud/shadow mask and calculate distance
      atc.cc = finalize_cloud(npix, CLOUD, SHADOW, TOA.ref, QAI);

    // more than 80% of max. allowable cloud cover? -> everything is cloud or shadow
    } else {

      alloc((void**)&SHADOW, _NC_, sizeof(bool));
      for (p=0; p<_NC_; p++) SHADOW[p] = true;
      atc.cc = finalize_cloud(npix, CLOUD, SHADOW, TOA.ref, QAI);
      atc.cc = 100;

      free((void*)LAND);
      if (par.band_bt) free((void*)TOA.bt);

    }


   // less than 0.1% clear pixels? -> everything is cloud or shadow
  } else {

    alloc((void**)&SHADOW, _NC_, sizeof(bool));
    for (p=0; p<_NC_; p++) SHADOW[p] = true;
    atc.cc = finalize_cloud(npix, PCP, SHADOW, TOA.ref, QAI);
    atc.cc = 100;

    free((void*)CLEAR); free((void*)LAND);
    free((void*)BRIGHT); free((void*)VPROB);
    if (par.band_bt) free((void*)TOA.bt);

  }

  printf("cc: %6.2f%%. ", atc.cc);
  if (atc.cc > par.maxcc){ printf("Skip. "); return FAILURE;}


  #ifdef CLOCK
  proctime_print("cloud module", TIME);
  #endif

  return SUCCESS;
}

