/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
This file contains functions needed for compositing L2 data
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#include "force.h"

  
/** Print compositing stats to stdout
+++ This function summarizes what the compositing algorithm has done.
--- hist:   histogram array
--- tag:    name tag of the statistic
--- off:    offset of the bin/values
--- multi:  gain of the bin/values
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int print_bap_stats(off_t hist[BIG_STR_LEN], char *tag, int off, int multi){
int hbin;
off_t sum = 0;

  printf("\nHistogram of %s in %%:\n", tag);
  
  for (hbin=0; hbin<BIG_STR_LEN; hbin++) sum += hist[hbin];

  if (sum == 0){
    printf("Nothing to report.\n"); return FAILURE;}

  for (hbin=0; hbin<BIG_STR_LEN; hbin++){
    if (hist[hbin] > 0) printf(" %d", (hbin+off)*multi);
  }
  printf("\n");

  for (hbin=0; hbin<BIG_STR_LEN; hbin++){
    if (hist[hbin] > 0) printf(" %.2f", 100.0*hist[hbin]/sum);
  }
  printf("\n");

  return SUCCESS;
}


/** Sigmoid minimizer function
+++ This minimizer function computes the RMSE between computed function 
+++ values and predefined function values (s0, s1, s2).
+++ Return: RMSE between computed and defined function values
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
double fun_logistic_target(const gsl_vector *v, void *params){
float a, b, *p = (float *)params;
double y, tmp;
int f;

  a = gsl_vector_get(v, 0);
  b = gsl_vector_get(v, 1);

  for (f=0, tmp=0; f<3; f++){
    y = p[6]/(1+exp(a*p[f]+b));
    tmp += (p[f+3]-y)*(p[f+3]-y);
  }

  return sqrt(tmp/3);
}


/** Sigmoid function optimization
+++ This estimates the sigmoid function parameters using a simplex fit.
+++ Return: function values for a sigmoid function
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int optimize_logistic_target(float *param, float *a, float *b){
const gsl_multimin_fminimizer_type *T = gsl_multimin_fminimizer_nmsimplex2;
gsl_multimin_fminimizer *s = NULL;
gsl_vector *ss, *x;
gsl_multimin_function minex_func;
size_t iter = 0;
int status;
double size;


  // start value
  x = gsl_vector_alloc (2);
  gsl_vector_set (x, 0, 0);
  gsl_vector_set (x, 1, 0);

  // initial step size of simplex
  ss = gsl_vector_alloc (2);
  gsl_vector_set (ss, 0, 0.01);
  gsl_vector_set (ss, 1, 1.0);

  // initialize method and iterate
  minex_func.n = 2;
  minex_func.f = fun_logistic_target;
  minex_func.params = param;

  s = gsl_multimin_fminimizer_alloc(T, 2);
  gsl_multimin_fminimizer_set(s, &minex_func, x, ss);

  do {

    iter++;
    if ((status = gsl_multimin_fminimizer_iterate(s))) break;

    size = gsl_multimin_fminimizer_size(s);
    status = gsl_multimin_test_size(size, 1e-2);

  }
  while (status == GSL_CONTINUE && iter < 100);

  *a = gsl_vector_get(s->x, 0);
  *b = gsl_vector_get(s->x, 1);

  gsl_vector_free(x);
  gsl_vector_free(ss);
  gsl_multimin_fminimizer_free (s);

  return SUCCESS;
}


/** Estimate parameters for temporal scoring functions
+++ This function estimates the function parameters for the temporal sco-
+++ ring functions, i.e. the sigmas (left and right tail) of the Gaussian
+++ or a and b for the sigmoid functions. Parameters are estimated for all
+++ pixels and years. If the static approach is used, one global set is 
+++ estimated. If the phenology-adative approach is used, parameters are 
+++ estimated for each pixel on the basis of the phenology. 
--- P:      target dates
--- A:      first  parameter (returned, allocated within)
--- B:      second parameter (returned, allocated within)
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int temporal_target_function(float ***P, float ***A, float ***B){
int f, b;
float **a_ = NULL, **b_ = NULL, param[7];

  alloc_2D((void***)&a_, par.Yn, _NC_, sizeof(float));
  alloc_2D((void***)&b_, par.Yn, _NC_, sizeof(float));

  if (par.score_type == SCR_SIG_DES || par.score_type == SCR_SIG_ASC){
    for (f=0; f<3; f++) param[f+3] = par.s[f];
    if (par.score_type == SCR_SIG_DES) param[6] = par.s[0];
    if (par.score_type == SCR_SIG_ASC) param[6] = par.s[2];
  }

  if (par.lsp){

    #pragma omp parallel firstprivate(param) private(b)
    {

      #pragma omp for
      for (f=0; f<_NC_; f++){
      for (b=0; b<par.Yn; b++){

        // center the dates
        param[0] = P[0][b][f]-P[1][b][f];
        param[1] = 0;
        param[2] = P[2][b][f]-P[1][b][f];

        if (par.score_type == SCR_GAUSS){
          a_[b][f] = fabs((param[0])/sqrt(-2*log(par.s[0]/par.s[1])));
          b_[b][f] = fabs((param[2])/sqrt(-2*log(par.s[2]/par.s[1])));
        } else if (par.score_type == SCR_SIG_DES || par.score_type == SCR_SIG_ASC){
          optimize_logistic_target(param, &a_[b][f], &b_[b][f]);
        }
  
      }
      }

    }

  } else {

    for (b=0; b<par.Yn; b++){

      // center the dates
      param[0] = P[0][b][0]-P[1][b][0];
      param[1] = 0;
      param[2] = P[2][b][0]-P[1][b][0];

      if (par.score_type == SCR_GAUSS){
        a_[b][0] = fabs((param[0])/sqrt(-2*log(par.s[0]/par.s[1])));
        b_[b][0] = fabs((param[2])/sqrt(-2*log(par.s[2]/par.s[1])));
      } else if (par.score_type == SCR_SIG_DES || par.score_type == SCR_SIG_ASC){
        optimize_logistic_target(param, &a_[b][0], &b_[b][0]);
      }

      for (f=0; f<_NC_; f++){
        a_[b][f] = a_[b][0];
        b_[b][f] = b_[b][0];
      }

    }
  }


  *A = a_;
  *B = b_;
  return SUCCESS;
}


/** Compile temporal target
+++ This function compiles the temporal target, at which the composites
+++ are anchored. If the phenology-adaptive approach is used, the LSP data
+++ is read and processed, otherwise the static values from the parameter 
+++ file are used. Level-2 data that are not covered by this period are
+++ discarded. Based on the temporal target, the function parameters for 
+++ the Gaussian or sigmoidal scoring functions are estimated.
--- tx:     current tile X-ID
--- ty:     current tile Y-ID
--- TAR:    temporal target (returned, allocated within)
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int compile_temporal_target(int tx, int ty, target_t *TAR){
int f, yr, p;
int ce, ce_min = INT_MAX, ce_max = INT_MIN;
target_t tar;

  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif


  /** read LSP data, return FAILURE if none exist
  +** *******************************************************************/

  if (par.lsp && list_lsp(tx, ty) == FAILURE){
      #ifdef DEBUG
      printf("listing LSP failed. "); 
      #endif
      return FAILURE;
  }


  /** Compile temporal target
  +** *******************************************************************/

  tar.p = (float***) malloc(3*sizeof(float**));

  for (f=0; f<3; f++){

    // read the LSP data
    if (par.lsp){

      if (read_lsp(par.f_lsp[f], &tar.p[f]) != SUCCESS){
        printf("Error reading LSP"); return FAILURE;}

    // use static dates from parameter file
    } else {

      alloc_2D((void***)&tar.p[f], par.Yn, _NC_, sizeof(float));

      for (yr=0; yr<par.Yn; yr++){

        if (f == 0 && par.Dt[f] > par.Dt[f+1]){
          ce = doy2ce(par.Dt[f], par.Yt-par.Yr+yr-1);
        } else if (f == 2 && par.Dt[f] < par.Dt[f-1]){
          ce = doy2ce(par.Dt[f], par.Yt-par.Yr+yr+1);
        } else {
          ce = doy2ce(par.Dt[f], par.Yt-par.Yr+yr);
        }

        for (p=0; p<_NC_; p++) tar.p[f][yr][p] = ce;

      }

    }

  }


  /** Find earliest and latest target date
  +** *******************************************************************/

  if (par.lsp){
    for (p=0, yr=0; p<_NC_; p++){
      if ((int)tar.p[0][yr][p] < ce_min) ce_min = (int)tar.p[0][yr][p];
    }
    for (p=0, yr=par.Yn-1; p<_NC_; p++){
      if ((int)tar.p[2][yr][p] > ce_max) ce_max = (int)tar.p[2][yr][p];
    }
  } else {
     if (par.Dt[0] > par.Dt[1]){
       ce_min = doy2ce(1, par.Yt-par.Yr-1);
     } else {
       ce_min = doy2ce(1, par.Yt-par.Yr);
     }
     if (par.Dt[2] < par.Dt[1]){
       ce_max = doy2ce(365, par.Yt+par.Yr+1);
     } else {
       ce_max = doy2ce(365, par.Yt+par.Yr);
     }
  }

  #ifdef DEBUG
  printf("ce range: %d - %d", ce_min, ce_max);
  #endif


  /** Discard L2 images that are not within the requested time frame,
  +** return FAILURE if none remain
  +** *******************************************************************/

  if (list_level2_filter_ce(ce_min, ce_max) == FAILURE){
    #ifdef DEBUG
    printf("no level-2 chip within ce range. ");
    #endif
    return FAILURE;
  }


  /** Estimate function values for temporal scoring functions
  +** *******************************************************************/

  temporal_target_function(tar.p, &tar.a, &tar.b);


  #ifdef CLOCK
  proctime_print("compiled temporal target", TIME);
  #endif

  *TAR = tar;
  return SUCCESS;
}


/** Correlation matrix
+++ This function computes the correlation matrix between all potential 
+++ observations of a given pixel to devaluate data artifacts or transient
+++ phenomena during compositing. The function exits gracefully if the 
+++ correlation score was disabled, i.e. the correlation score weight was
+++ set to 0. The correlation is set to _NODATA_ for incomplete pairs.
--- L2:     level-2 data
--- p:      current pixel ID
--- cor:    correlation matrix of all observations (modified)
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int corr_matrix(level2_t *L2, qai_t *QAI, int p, double **cor){
int t, l, b;
double xm, ym, xv, yv, cv;


  if (par.w.r == 0) return SUCCESS;

  for (t=0;     t<dir.n; t++){
  for (l=(t+1); l<dir.n; l++){

    if (!use_this_pixel(QAI[t], p) || 
        !use_this_pixel(QAI[l], p)){

      cor[t][l] = cor[l][t] = _NODATA_;

    } else {

      for (b=0, xm=0, ym=0; b<_NB_; b++){
        xm += L2[t].ref[b][p];
        ym += L2[l].ref[b][p];
      }
      xm /= _NB_; ym /= _NB_;

      for (b=0, cv=0, xv=0, yv=0; b<_NB_; b++){
        cv += (L2[t].ref[b][p]-xm)*(L2[l].ref[b][p]-ym);
        xv += (L2[t].ref[b][p]-xm)*(L2[t].ref[b][p]-xm);
        yv += (L2[l].ref[b][p]-ym)*(L2[l].ref[b][p]-ym);
      }

      if (sqrt(xv*yv) == 0){
        cor[t][l] = cor[l][t] = _NODATA_;
      } else {
        cor[t][l] = cor[l][t] = cv/sqrt(xv*yv);
      }

    }

  }
  }

  for (t=0;             t<dir.n; t++){
  for (l=0, xm=0, ym=0; l<dir.n; l++){

    if (t == l || cor[t][l] == _NODATA_) continue;

    xm += fabs(cor[t][l]);
    ym++;
    cor[t][0] = xm/ym;

  }
  }


  return SUCCESS;
}


/** Compositing score
+++ This function computes the compositing scores for each potential ob-
+++ servations of a given pixel, i.e. the total sum (weighted sum of tem-
+++ poral and auxilliary scores), temporal scores (DOY and Year scores),
+++ and auxilliary scores (cloud/shadow, haze, correlation view zenith 
+++ scores).
--- L2:     level-2 data
--- tar:    temporal compositing target
--- p:      current pixel ID
--- cor:    correlation matrix of all observations
--- score:  compositing scores (modified)
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int compositing_scores(level2_t *L2, qai_t *QAI, target_t tar, int p, double **cor, score_t *score){
int b, t, bb;
int slice, dy;
float tmp, tdist;
float vz;


  for (t=0; t<dir.n; t++){

    score[t].t = 0;

    // skip if no data
    if (!use_this_pixel(QAI[t], p)) continue;

    // get suitable year
    tdist = INT_MAX; b = 0;
    for (bb=0; bb<par.Yn; bb++){
      if ((tmp = fabs(L2[t].date.ce-tar.p[1][bb][p])) < tdist){
        tdist = tmp; b = bb;
      }
    }
    tdist = L2[t].date.ce-tar.p[1][b][p];

    // if target is corrupt, skip pixel
    if (tar.p[0][b][p] < 1900*365 || tar.p[1][b][p] < 1900*365 || 
        tar.p[2][b][p] < 1900*365) continue;

    // difference between acquisition and target year
    dy = abs(L2[t].date.year-par.Yt);


    /** Seasonal suitability: DOY score
    /** Annual   suitability: Year score
    +** *****************************************************************/

    if (par.score_type == SCR_GAUSS){

      // use left tail of Gaussian if acquisition before target
      if (L2[t].date.ce < tar.p[1][b][p]){

        // DOY score
        score[t].d = par.s[1] * exp(tdist*tdist/
                  (-2*tar.a[b][p]*tar.a[b][p]));

        // Year score
        slice = (tar.p[1][b][p]-tar.p[0][b][p])/((par.Yr+1)*par.Yf);
        score[t].y = par.s[1] * exp(dy*slice*dy*slice/
                  (-2*tar.a[b][p]*tar.a[b][p]));

      // use right tail of Gaussian if acquisition before target
      } else {

        // DOY score
        score[t].d = par.s[1]*exp(tdist*tdist/
                  (-2*tar.b[b][p]*tar.b[b][p]));

        // Year score
        slice = (tar.p[2][b][p]-tar.p[1][b][p])/((par.Yr+1)*par.Yf);
        score[t].y = par.s[1] * exp(dy*slice*dy*slice/
                  (-2*tar.b[b][p]*tar.b[b][p]));

      }

      // difference between acquisition and target
      score[t].dd = tdist;

    } else if (par.score_type == SCR_SIG_DES){

      // use 0 weight if acquisition before target
      if (L2[t].date.ce < tar.p[0][b][p]){

        score[t].d = score[t].y  = 0;

      // use descending sigmoid if acquisition after target
      } else {

        // DOY score
        score[t].d = par.s[0] / (1 + 
          exp(tar.a[b][p] * tdist + tar.b[b][p]));

        // Year score
        slice = (tar.p[1][b][p]-tar.p[0][b][p])/((par.Yr+1)*par.Yf);
        score[t].y = par.s[0] / (1 + exp(tar.a[b][p] * 
            (tar.p[0][b][p]+dy*slice-tar.p[1][b][p]) + tar.b[b][p]));
      }
      
      // difference between acquisition and target
      score[t].dd = L2[t].date.ce-tar.p[0][b][p];

    } else if (par.score_type == SCR_SIG_ASC){

      // use 0 weight if acquisition after target
      if (L2[t].date.ce > tar.p[2][b][p]){

        score[t].d = score[t].y  = 0;

      // use ascending sigmoid if acquisition before target
      } else {

      // DOY score
        score[t].d = par.s[2] / (1 + 
          exp(tar.a[b][p] * tdist + tar.b[b][p]));

        // Year score
        slice = (tar.p[2][b][p]-tar.p[1][b][p])/((par.Yr+1)*par.Yf);
        score[t].y = par.s[2] / (1 + exp(tar.a[b][p] * 
            (tar.p[2][b][p]-dy*slice-tar.p[1][b][p]) + tar.b[b][p]));

      }

      // difference between acquisition and target
      score[t].dd = L2[t].date.ce-tar.p[2][b][p];

    }


    /** Auxilliary scores
    +** *****************************************************************/

    // doesn't make sense, but to be sure
    if (par.w.d == 0) score[t].d = 0.0;
    if (par.w.y == 0) score[t].y = 0.0;

    // cloud distance score
    if (par.w.c > 0){
      score[t].c = 1.0 / (1.0 + exp((-10.0/par.dreq) * 
                ((float)L2[t].cld[p] - (par.dreq/2.0)) ));
    } else score[t].c = 0.0;

    // haze score
    if (par.w.h > 0){
      score[t].h = 1.0 / (1.0 + exp((10.0/200.0) * 
                ((float)L2[t].hot[p] + 150.0)));
    } else score[t].h = 0.0;

    // correlation score
    if (par.w.r > 0){
      score[t].r = 1.0 / (1.0 + exp((-10.0/(0.5*2.0/3.0)) *
                (cor[t][0] - (0.5*4.0/3.0))));
    } else score[t].r = 0.0;

    // view zenith score
    if (par.w.v > 0){
      vz = ((float)L2[t].vzn[p])*0.01;
      if (vz < 0 || vz > 180){ score[t].v = 0.0;
      } else { score[t].v = 1.0 / (1.0 + exp(10.0/par.vreq * 
                          (vz - (par.vreq/2.0))));
      }
    } else score[t].v = 0.0;


    /** Total score
    +** *****************************************************************/

    score[t].t += (par.w.d * score[t].d);
    score[t].t += (par.w.y * score[t].y);
    score[t].t += (par.w.c * score[t].c);
    score[t].t += (par.w.h * score[t].h);
    score[t].t += (par.w.r * score[t].r);
    score[t].t += (par.w.v * score[t].v);
    score[t].t /= par.w.t;

  }

  return SUCCESS;
}

