/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
This program computes clear-sky observation statistics
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#include "force.h"


int main ( int argc, char *argv[] ){
level2_t *L2 = NULL;
int l, w, m, y, p;
int tx, ty, tc = 0, tk = 0;
int *tiles_x = NULL, *tiles_y = NULL, tiles_k;
char dname[MAX_STR_LEN];
char fname[MAX_STR_LEN];
double geotran[6] = { 0, 0, 0, 0, 0, 0 };
double **avg = NULL, **var = NULL, sd, *k = NULL;
int **min = NULL, **max = NULL;
time_t TIME, TIME_I, TIME_C, TIME_O;
double time_i = 0, time_c = 0, time_o = 0, ctime = 0;
double eta_h, eta_m, eta_s;

  if (argc != 2){ printf("usage: %s parameter-file\n", argv[0]); exit(1);}


  /** INITIALIZING
   +** ******************************************************************/

  time(&TIME);
  printf("Initializing startup"); fflush(stdout);


  par.f_par = argv[1];

  // parse parameter file
  if (parse_param_cso() != SUCCESS){
    printf("Reading parameter file failed!\n"); exit(1);}

  // copy datacube definition
  if (copy_datacube_def(par.d_level2, par.d_cso) == FAILURE){
    printf("Copying datacube definition failed.\n"); exit(1);}

  // get tile file and read it
  if (tilefileread(&tiles_x, &tiles_y, &tiles_k) != SUCCESS){
    printf("Reading tile file failed!\n"); exit(1);}

  // how many tiles to do?
  for (ty=par.tminy; ty<=par.tmaxy; ty++){
  for (tx=par.tminx; tx<=par.tmaxx; tx++){

    // if tile is not whitelisted (if specified), skip
    if (tile_whitelisted(tiles_x, tiles_y, tiles_k, tx, ty) == FAILURE) continue;
    tk++;

  }
  }
   
  mem_to_disc_dict_init();
  mem_to_disc_dict_cso();

  // init overall stats
  alloc_2D((void***)&avg, _NP_, par.nwin, sizeof(double));
  alloc_2D((void***)&var, _NP_, par.nwin, sizeof(double));
  alloc_2D((void***)&min, _NP_, par.nwin, sizeof(int));
  alloc_2D((void***)&max, _NP_, par.nwin, sizeof(int));
  alloc((void**)&k, par.nwin, sizeof(double));

  for (w=0; w<par.nwin; w++) k[w] = 1;

  for (l=0; l<_NP_; l++){
  for (w=0; w<par.nwin; w++){
    min[l][w] = INT_MAX;
    max[l][w] = INT_MIN;
  }
  }

  printf("\rCSO processing started, I/C/O TBD, ETA: TBD");
  fflush(stdout);

  // open threads
  omp_set_num_threads(par.cpu);
  

  /** START PROCESSING; LOOP OVER ALL AVAILABLE TILES
  +** *******************************************************************/

  for (ty=par.tminy; ty<=par.tmaxy; ty++){
  for (tx=par.tminx; tx<=par.tmaxx; tx++){

    time(&TIME_I);

    // current tile
    sprintf(dname, "%s/X%04d_Y%04d", par.d_level2, tx, ty);

    // if tile is not whitelisted (if specified), skip
    if (tile_whitelisted(tiles_x, tiles_y, tiles_k, tx, ty) == FAILURE) continue;

    // get Level2 file listing, skip if tile is empty
    if (list_level2(dname) == FAILURE) continue;

    sprintf(fname, "%s/%s", dname, dir.list[0]);
    image_info(fname, geotran);


    /** READ LEVEL2 DATA
    +** *****************************************************************/
    if (read_level2(dname, &L2, _NX_*par.res, _NY_*par.res) != SUCCESS){
      printf("error reading L2.\n"); exit(1);}

    mem_to_disc_dict_cso();

    time_i += proctime(TIME_I);

    time(&TIME_C);

    /** CLEAR-SKY STATS
    +** *****************************************************************/
    clear_sky_stats(L2);


    /** overall stats **/
    for (l=0; l<_NP_; l++){

      if (!par.m2d[l].save) continue;

      for (w=0; w<par.m2d[l].length; w++){

        for (p=0; p<_NC_; p++, k[w]++){
          if (k[w] == 1){
            avg[l][w] = (double)par.m2d[l].data[w][p];
          } else {
            var_recurrence((double)par.m2d[l].data[w][p], &avg[l][w], &var[l][w], k[w]);
          }
          if (par.m2d[l].data[w][p] < min[l][w]) min[l][w] = par.m2d[l].data[w][p];
          if (par.m2d[l].data[w][p] > max[l][w]) max[l][w] = par.m2d[l].data[w][p];
        }

      }

    }

    time_c += proctime(TIME_C);

    time(&TIME_O);

    /** write output **/
    for (l=0; l<_NP_; l++){
      if (par.m2d[l].save) write_products_cso(tx, ty, par.m2d[l], geotran);
    }


    /** CLEAN
    +** *****************************************************************/
    mem_to_disc_free();
    free_2D((void**)dir.list, dir.N);
    free_2D((void**)dir.LIST, dir.N);
    free_level2(L2);
    
    time_o += proctime(TIME_O);


    /** PROGRESS
    +** *****************************************************************/
    tc++; ctime = time_i + time_c + time_o;
    eta_s = (tk-tc)*(ctime/tc);
    if (eta_s >= 3600){
      eta_h = floor(eta_s/3600); eta_s -= eta_h*3600;
    } else eta_h = 0;
    if (eta_s >= 60){
      eta_m = floor(eta_s/60); eta_s -= eta_m*60;
    } else eta_m = 0;

    printf("\rCSO processing %6.2f%%, I/C/O %03.0f/%03.0f/%03.0f, ETA: %02.0fh %02.0fm %02.0fs", 
      (float)tc/tk*100.0, round(time_i/ctime*100), round(time_c/ctime*100), round(time_o/ctime*100), 
       eta_h, eta_m, eta_s);
    fflush(stdout);

  }
  }

  printf("\n\n");


  /** print overall stats **/
  printf("Overall stats:\n\n");
  mem_to_disc_dict_cso();


  for (l=0; l<_NP_; l++){

    if (!par.m2d[l].save) continue;

    printf("%s\n", par.m2d[l].product);

    for (w=0; w<par.nwin; w++){
      sd = standdev(var[l][w], k[w]);
      printf("%.2f - min: %03d, mean-sd: %06.2f, mean: %06.2f, mean+sd: %06.2f, max: %03d\n",
        par.ywin[w] + (par.mwin[w]-1)/12.0, min[l][w], avg[l][w]-sd, avg[l][w], avg[l][w]+sd, max[l][w]);
    }
    printf("\n");

  }


  /** CLEAN
  +** *****************************************************************/
  free_2D((void**)avg, _NP_); free_2D((void**)var, _NP_);
  free_2D((void**)max, _NP_); free_2D((void**)min, _NP_);
  free((void*)k);
  free((void*)tiles_x); free((void*)tiles_y);
  
  proctime_print("Total runtime", TIME);

  return SUCCESS;
}

