/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/** The date implementation included in this file, is based on code from 
+++ SpliTS, a framework for spline analysis of  time series. SpliTS is 
+++ licensed under the terms of the GNU General Public License (>= 3).
+++ http://sebastian-mader.net/splits/
+++ SpliTS Copyright (C) 2010-2016 Sebastian Mader
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
This file contains functions for converting dates
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#include "force.h"


/** Convert DOY to month and day
+++ This function converts a day-of-year value to month and day. Note that
+++ leap years are ignored, a year has 365 days.
--- doy:    day-of-year
--- m:      month (returned)
--- d:      day   (returned)
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void doy2md(int doy, int* m, int* d){

  if (doy>334){
    *m = 12; *d = doy-334; return;
  } else if (doy>304){
    *m = 11; *d = doy-304; return;
  } else if (doy>273){
    *m = 10; *d = doy-273; return;
  } else if (doy>243){
    *m = 9;  *d = doy-243; return;
  } else if (doy>212){
    *m = 8;  *d = doy-212; return;
  } else if (doy>181){
    *m = 7;  *d = doy-181; return;
  } else if (doy>151){
    *m = 6;  *d = doy-151; return;
  } else if (doy>120){
    *m = 5;  *d = doy-120; return;
  } else if (doy>90){
    *m = 4;  *d = doy-90;  return;
  } else if (doy>59){
    *m = 3;  *d = doy-59;  return;
  } else if (doy>31){
    *m = 2;  *d = doy-31;  return;
  } else {
    *m = 1;  *d = doy;     return;
  }

  return;
}


/** Convert month and day to DOY
+++ This function converts month and day values to day-of-year. Note that 
+++ leap years are ignored, a year has 365 days.
--- m:      month
--- d:      day
+++ Return: day-of-year
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int md2doy(int m, int d){
int doy;

  switch(m){
    case 1:  doy = d;     break;
    case 2:  doy = 31+d;  break;
    case 3:  doy = 59+d;  break;
    case 4:  doy = 90+d;  break;
    case 5:  doy = 120+d; break;
    case 6:  doy = 151+d; break;
    case 7:  doy = 181+d; break;
    case 8:  doy = 212+d; break;
    case 9:  doy = 243+d; break;
    case 10: doy = 273+d; break;
    case 11: doy = 304+d; break;
    case 12: doy = 334+d; break;
  }

  return doy;
}


/** Convert DOY to week
+++ This function converts a day-of-year value to week-of-year.  Note that
+++ the 1st week starts at DOY 1, a year has 52 weeks (The last week con-
+++ a couple more days).
--- doy:    day-of-year
+++ Return: week-of-year
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int doy2week(int doy){
float w;

  if ((w = ceil(doy/7.0)) > 52) return 52; else return (int)w;
}


/** Convert date to days since current era
+++ This function converts a date to days since CE.  Note that leap years 
+++ are ignored, a year has 365 days.
--- m:      month
--- d:      day
--- y:      year
+++ Return: days since CE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int date2ce(int m, int d, int y){

  return y*365+md2doy(m,d);
}


/** Convert day-of-year and year values to days since current era
+++ This function converts day-of-year and year values to days since CE. 
+++ Note that leap years are ignored, a year has 365 days.
--- doy:    day-of-year
--- y:      year
+++ Return: days since CE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int doy2ce(int doy, int y){

  return y*365+doy;
}


/** Convert days since current era to date
+++ This function converts a days since CE value to a a date. Note that 
+++ leap years are ignored, a year has 365 days.
--- ce:     days since CE
--- m:      month (returned)
--- d:      day   (returned)
--- y:      year  (returned)
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void ce2date(int ce, int* m, int* d, int* y){
int year, doy;

  year = (int)floor((double)ce/365.0);
  doy = ce-year*365;

  if (doy==0){
    *y = year-1;
    *m = 12;
    *d = 31;
  } else {
    *y = year;
    doy2md(doy, m, d);
  }

  return;
}


/** Convert days since current era to day-of-year and year
+++ This function converts a days since CE value to day-of-year and year. 
+++ Note that leap years are ignored, a year has 365 days.
--- ce:     days since CE
--- doy:    day-of-year (returned)
--- y:      year        (returned)
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void ce2doy(int ce, int* doy, int* y){
int year, tmp;

  year = (int)floor((double)ce/365.0);
  tmp = ce-year*365;

  if (tmp==0){
    *y = year-1;
    *doy = 365;
  } else {
    *y = year;
    *doy = tmp;
  }

  return;
}


/** Get current time
+++ This function gets the current time and date and returns a date struct
+++ containing year, month, day, day-of-year, week-of-year and days since
+++ current era. 
+++ This function needs a global variable timeinfo (see force.h/global.c).
--- d:      date struct (returned)
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void current_date(date_t *d){
time_t rawtime;

  time(&rawtime);
  timeinfo = localtime(&rawtime);
  d->year  = timeinfo->tm_year+1900;
  d->month = timeinfo->tm_mon+1;
  d->day   = timeinfo->tm_mday;
  d->doy   = md2doy(d->month,d->day);
  d->week  = doy2week(d->doy);
  d->ce    = doy2ce(d->doy,d->year);

  return;
}


/** Set time
+++ This function sets the time to the given date.
+++ This function needs a global variable timeinfo (see force.h/global.c).
+++ date_set needs to be called before using this function.
--- d:      date struct
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void date_set(date_t d){
time_t rawtime;

  time(&rawtime);
  timeinfo = localtime(&rawtime);
  timeinfo->tm_year = d.year-1900;
  timeinfo->tm_mon = d.month-1;
  timeinfo->tm_mday = d.day;
  mktime(timeinfo);

  return;
}


/** Increment time
+++ This function increments the time with one day, returns a date struct
+++ containing year, month, day, day-of-year, week-of-year and days since
+++ current era. 
+++ This function needs a global variable timeinfo (see force.h/global.c).
--- d:      date struct (returned)
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void date_plus(date_t *d){

  date_set(*d);

  timeinfo->tm_mday++;
  mktime(timeinfo);

  d->year  = timeinfo->tm_year+1900;
  d->month = timeinfo->tm_mon+1;
  d->day   = timeinfo->tm_mday;
  d->doy   = md2doy(d->month,d->day);
  d->week  = doy2week(d->doy);
  d->ce    = doy2ce(d->doy,d->year);

  return;
}


/** Decrement time
+++ This function dwcrements the time with one day, returns a date struct
+++ containing year, month, day, day-of-year, week-of-year and days since
+++ current era. 
+++ This function needs a global variable timeinfo (see force.h/global.c).
+++ date_set needs to be called before using this function.
--- d:      date struct (returned)
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void date_minus(date_t *d){

  timeinfo->tm_mday--;
  mktime(timeinfo);
  d->year  = timeinfo->tm_year+1900;
  d->month = timeinfo->tm_mon+1;
  d->day   = timeinfo->tm_mday;
  d->doy   = md2doy(d->month,d->day);
  d->week  = doy2week(d->doy);
  d->ce    = doy2ce(d->doy,d->year);

  return;
}

