/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
This file contains functions for debugging purposes
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#include "force.h"


#ifdef DEBUG

/** Check if DEBUG is OK
+++ This function checks if the debugging directory exists
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void db_check(){

  if (!fileexist(DEBUGPATH)){
    printf("DEBUGPATH does not exist!\n");
    printf(" Solution 1: Turn off DEBUG mode --> comment out!\n");
    printf(" Solution 2: Change DEBUGPATH and/or create this directory!\n");
    printf(" Anyway, DEBUG & DEBUGPATH must be set in force.h and\n");
    printf("  the code must be re-compiled!\n");
    exit(1);
  }

  return;
}


/** Write header of DEBUG images
+++ This function writes a basic ENVI header
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int db_write_header(char *fname, int nx, int ny, int nb, int dt, char *mapinfo){
FILE *fp;

  if ((fp=fopen(fname, "w")) == NULL){
    printf("\nOpen header file %s error!", fname);
    return FAILURE;
  }

  fprintf(fp, "ENVI\n");
  fprintf(fp, "description = {\n"); 
  fprintf(fp, "... }\n");
  fprintf(fp, "samples = %d\n", nx);
  fprintf(fp, "lines = %d\n", ny);
  fprintf(fp, "bands = %d\n", nb);
  fprintf(fp, "header offset = 0\n");
  fprintf(fp, "file type = ENVI Standard\n");
  fprintf(fp, "data type = %d\n", dt);
  fprintf(fp, "interleave = bsq\n");
  fprintf(fp, "sensor type = Unknown\n");
  fprintf(fp, "byte order = 0\n");
  fprintf(fp, "wavelength units = Unknown\n");
  fprintf(fp, "%s\n", mapinfo);

  fclose(fp);

  return SUCCESS;
}


/** Write 1D DEBUG image
+++ This function writes a one-band DEBUG image in raw binary format and
+++ adds a basic ENVI header
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void db_write_image_1D(void *IMG, char *base, int nx, int ny, size_t size, int envidt){
char fname[MAX_STR_LEN];
FILE *fp = NULL;

  sprintf(fname, "%s/%s.dat", DEBUGPATH, base);
  fp = fopen(fname, "wb");
  fwrite((const void*)  IMG, size, nx*ny, fp);
  fclose(fp);
  sprintf(fname, "%s/%s.hdr", DEBUGPATH, base);
  db_write_header(fname, nx, ny, 1, envidt, (char*) "");

  return;
}


/** Write 2D DEBUG image
+++ This function writes a multi-band DEBUG image in raw binary format and
+++ adds a basic ENVI header
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void db_write_image_2D(void **IMG, char *base, int nb, int nx, int ny, size_t size, int envidt){
int b;
char fname[MAX_STR_LEN];
FILE *fp = NULL;

  sprintf(fname, "%s/%s.dat", DEBUGPATH, base);
  fp = fopen(fname, "wb");
  for (b=0; b<nb; b++) fwrite((const void*)  IMG[b], size, nx*ny, fp);
  fclose(fp);
  sprintf(fname, "%s/%s.hdr", DEBUGPATH, base);
  db_write_header(fname, nx, ny, nb, envidt, (char*) "");

  return;
}


/** Write coarse grid DEBUG images
+++ This function writes DEBUG images of radiometric variables that are
+++ represented in the coarse processing grid in raw binary format and
+++ adds a basic ENVI header
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void db_write_image_CG(atc_t atc){
int g, b, z;
char fname[MAX_STR_LEN];
FILE *fp = NULL;

  //z = floor((atc.dem.cmax+atc.dem.cmin)/2);
  z = 128;

  // coordinates
  sprintf(fname, "%s/CG-COORDS.dat", DEBUGPATH);
  fp = fopen(fname, "wb");
  for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].mapx,      sizeof(float), 1, fp);
  for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].mapy,      sizeof(float), 1, fp);
  for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].geox,      sizeof(float), 1, fp);
  for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].geoy,      sizeof(float), 1, fp);
  fclose(fp);
  sprintf(fname, "%s/CG-COORDS.hdr", DEBUGPATH);
  db_write_header(fname, atc.ngx, atc.ngy, 4, 4, (char*) "");

  // angles
  sprintf(fname, "%s/CG-ANGLE.dat", DEBUGPATH);
  fp = fopen(fname, "wb");
  for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].szen,      sizeof(float), 1, fp);
  for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].vzen,      sizeof(float), 1, fp);
  for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].sazi,      sizeof(float), 1, fp);
  for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].vazi,      sizeof(float), 1, fp);
  for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].ms,        sizeof(float), 1, fp);
  for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].mv,        sizeof(float), 1, fp);
  for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].cospsi,    sizeof(float), 1, fp);
  for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].Pr,        sizeof(float), 1, fp);
  for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].Pa,        sizeof(float), 1, fp);
  fclose(fp);
  sprintf(fname, "%s/CG-ANGLE.hdr", DEBUGPATH);
  db_write_header(fname, atc.ngx, atc.ngy, 9, 4, (char*) "");

  // BRDF correction factors
  sprintf(fname, "%s/CG-BRDF.dat", DEBUGPATH);
  fp = fopen(fname, "wb");
  for (b=0; b<_NB_; b++){
    for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].brdf[b], sizeof(float), 1, fp);
  }
  fclose(fp);
  sprintf(fname, "%s/CG-BRDF.hdr", DEBUGPATH);
  db_write_header(fname, atc.ngx, atc.ngy, _NB_, 4, (char*) "");

  // Fresnel reflectance
  sprintf(fname, "%s/CG-RHOF.dat", DEBUGPATH);
  fp = fopen(fname, "wb");
  for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].rhof, sizeof(float), 1, fp);
  fclose(fp);
  sprintf(fname, "%s/CG-RHOF.hdr", DEBUGPATH);
  db_write_header(fname, atc.ngx, atc.ngy, 1, 4, (char*) "");

  // AOD @ scene average elevation
  sprintf(fname, "%s/CG-AOD-0.dat", DEBUGPATH);
  fp = fopen(fname, "wb");
  for (b=0; b<_NB_; b++){
    for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].aod[b], sizeof(float), 1, fp);
  }
  fclose(fp);
  sprintf(fname, "%s/CG-AOD-0.hdr", DEBUGPATH);
  db_write_header(fname, atc.ngx, atc.ngy, _NB_, 4, (char*) "");

  // AOD at actual elevation
  sprintf(fname, "%s/CG-AOD.dat", DEBUGPATH);
  fp = fopen(fname, "wb");
  for (b=0; b<_NB_; b++){
    for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].z[atc.cg[g].cdem].aod[b], sizeof(float), 1, fp);
  }
  fclose(fp);
  sprintf(fname, "%s/CG-AOD.hdr", DEBUGPATH);
  db_write_header(fname, atc.ngx, atc.ngy, _NB_, 4, (char*) "");

  // MOD/AOD elevation scaling factors
  sprintf(fname, "%s/CG-OD-ELEV.dat", DEBUGPATH);
  fp = fopen(fname, "wb");
  for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].z[atc.cg[g].cdem].Hr, sizeof(float), 1, fp);
  for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].z[atc.cg[g].cdem].Ha, sizeof(float), 1, fp);
  fclose(fp);
  sprintf(fname, "%s/CG-OD-ELEV.hdr", DEBUGPATH);
  db_write_header(fname, atc.ngx, atc.ngy, 2, 4, (char*) "");

  // path reflectance
  sprintf(fname, "%s/CG-RHOP.dat", DEBUGPATH);
  fp = fopen(fname, "wb");
  for (b=0; b<_NB_; b++){
    for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].z[atc.cg[g].cdem].rho_p[b], sizeof(float), 1, fp);
  }
  fclose(fp);
  sprintf(fname, "%s/CG-RHOP.hdr", DEBUGPATH);
  db_write_header(fname, atc.ngx, atc.ngy, _NB_, 4, (char*) "");

  // optical depth
  sprintf(fname, "%s/CG-OD.dat", DEBUGPATH);
  fp = fopen(fname, "wb");
  for (b=0; b<_NB_; b++){
    for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].z[atc.cg[g].cdem].od[b], sizeof(float), 1, fp);
  }
  fclose(fp);
  sprintf(fname, "%s/CG-OD.hdr", DEBUGPATH);
  db_write_header(fname, atc.ngx, atc.ngy, _NB_, 4, (char*) "");

  // MOD
  sprintf(fname, "%s/CG-MOD.dat", DEBUGPATH);
  fp = fopen(fname, "wb");
  for (b=0; b<_NB_; b++){
    for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].z[atc.cg[g].cdem].mod[b], sizeof(float), 1, fp);
  }
  fclose(fp);
  sprintf(fname, "%s/CG-MOD.hdr", DEBUGPATH);
  db_write_header(fname, atc.ngx, atc.ngy, _NB_, 4, (char*) "");

  // spherical albedo
  sprintf(fname, "%s/CG-SALB.dat", DEBUGPATH);
  fp = fopen(fname, "wb");
  for (b=0; b<_NB_; b++){
    for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].z[atc.cg[g].cdem].s[b], sizeof(float), 1, fp);
  }
  fclose(fp);
  sprintf(fname, "%s/CG-SALB.hdr", DEBUGPATH);
  db_write_header(fname, atc.ngx, atc.ngy, _NB_, 4, (char*) "");

  // total scattering up
  sprintf(fname, "%s/CG-SCAT-Tv.dat", DEBUGPATH);
  fp = fopen(fname, "wb");
  for (b=0; b<_NB_; b++){
    for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].z[atc.cg[g].cdem].Tv[b], sizeof(float), 1, fp);
  }
  fclose(fp);
  sprintf(fname, "%s/CG-SCAT-Tv.hdr", DEBUGPATH);
  db_write_header(fname, atc.ngx, atc.ngy, _NB_, 4, (char*) "");

  // total scattering down
  sprintf(fname, "%s/CG-SCAT-Ts.dat", DEBUGPATH);
  fp = fopen(fname, "wb");
  for (b=0; b<_NB_; b++){
    for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].z[atc.cg[g].cdem].Ts[b], sizeof(float), 1, fp);
  }
  fclose(fp);
  sprintf(fname, "%s/CG-SCAT-Ts.hdr", DEBUGPATH);
  db_write_header(fname, atc.ngx, atc.ngy, _NB_, 4, (char*) "");

  // direct scattering up
  sprintf(fname, "%s/CG-SCAT-tvd.dat", DEBUGPATH);
  fp = fopen(fname, "wb");
  for (b=0; b<_NB_; b++){
    for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].z[atc.cg[g].cdem].tvd[b], sizeof(float), 1, fp);
  }
  fclose(fp);
  sprintf(fname, "%s/CG-SCAT-tvd.hdr", DEBUGPATH);
  db_write_header(fname, atc.ngx, atc.ngy, _NB_, 4, (char*) "");

  // direct scattering down
  sprintf(fname, "%s/CG-SCAT-tvs.dat", DEBUGPATH);
  fp = fopen(fname, "wb");
  for (b=0; b<_NB_; b++){
    for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].z[atc.cg[g].cdem].tvs[b], sizeof(float), 1, fp);
  }
  fclose(fp);
  sprintf(fname, "%s/CG-SCAT-tvs.hdr", DEBUGPATH);
  db_write_header(fname, atc.ngx, atc.ngy, _NB_, 4, (char*) "");

  // diffuse scattering up
  sprintf(fname, "%s/CG-SCAT-tsd.dat", DEBUGPATH);
  fp = fopen(fname, "wb");
  for (b=0; b<_NB_; b++){
    for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].z[atc.cg[g].cdem].tsd[b], sizeof(float), 1, fp);
  }
  fclose(fp);
  sprintf(fname, "%s/CG-SCAT-tsd.hdr", DEBUGPATH);
  db_write_header(fname, atc.ngx, atc.ngy, _NB_, 4, (char*) "");

  // diffuse scattering down
  sprintf(fname, "%s/CG-SCAT-tss.dat", DEBUGPATH);
  fp = fopen(fname, "wb");
  for (b=0; b<_NB_; b++){
    for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].z[atc.cg[g].cdem].tss[b], sizeof(float), 1, fp);
  }
  fclose(fp);
  sprintf(fname, "%s/CG-SCAT-tss.hdr", DEBUGPATH);
  db_write_header(fname, atc.ngx, atc.ngy, _NB_, 4, (char*) "");

  // water vapor transmittance up
  sprintf(fname, "%s/CG-GAS-Tv.dat", DEBUGPATH);
  fp = fopen(fname, "wb");
  for (b=0; b<_NB_; b++){
    for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].Tvw[b], sizeof(float), 1, fp);
  }
  fclose(fp);
  sprintf(fname, "%s/CG-GAS-Tv.hdr", DEBUGPATH);
  db_write_header(fname, atc.ngx, atc.ngy, _NB_, 4, (char*) "");

  // water vapor transmittance down
  sprintf(fname, "%s/CG-GAS-Ts.dat", DEBUGPATH);
  fp = fopen(fname, "wb");
  for (b=0; b<_NB_; b++){
    for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].Tsw[b], sizeof(float), 1, fp);
  }
  fclose(fp);
  sprintf(fname, "%s/CG-GAS-Ts.hdr", DEBUGPATH);
  db_write_header(fname, atc.ngx, atc.ngy, _NB_, 4, (char*) "");

  // total gas transmittance
  sprintf(fname, "%s/CG-GAS-T.dat", DEBUGPATH);
  fp = fopen(fname, "wb");
  for (b=0; b<_NB_; b++){
    for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].Tg[b], sizeof(float), 1, fp);
  }
  fclose(fp);
  sprintf(fname, "%s/CG-GAS-T.hdr", DEBUGPATH);
  db_write_header(fname, atc.ngx, atc.ngy, _NB_, 4, (char*) "");

  // environment reflectance function
  sprintf(fname, "%s/CG-F.dat", DEBUGPATH);
  fp = fopen(fname, "wb");
  for (b=0; b<_NB_; b++){
    for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].z[atc.cg[g].cdem].F[b], sizeof(float), 1, fp);
  }
  fclose(fp);
  sprintf(fname, "%s/CG-F.hdr", DEBUGPATH);
  db_write_header(fname, atc.ngx, atc.ngy, _NB_, 4, (char*) "");

  // elevation
  sprintf(fname, "%s/CG-CDEM.dat", DEBUGPATH);
  fp = fopen(fname, "wb");
  for (g=0; g<atc.ngc; g++) fwrite((const void*) &atc.cg[g].cdem, sizeof(int), 1, fp);
  fclose(fp);
  sprintf(fname, "%s/CG-CDEM.hdr", DEBUGPATH);
  db_write_header(fname, atc.ngx, atc.ngy, 1, 13, (char*) "");

  return;
}


/** Print Dark Objects
+++ This function prints AOD estimation parameters for a dark object
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void db_print_dark_object(darkobj_t *dobj, int o, int g){

  printf("obj %04d, n %06d, r %02d, CG-X/Y %04d/%04d, COSI %.2f, Z %01.3f\n", 
    o+1, dobj[o].k, dobj[o].r, tocg(dobj[o].j), tocg(dobj[o].i), 
    dobj[o].cosi, atc.cg[g].cdem*atc.dem.step+atc.dem.min+atc.dem.step/2);
  printf(" mv/ms %.3f/%.3f, rel-azi %.3f\n", atc.cg[g].mv, atc.cg[g].ms, atc.cg[g].vazi-atc.cg[g].sazi);
  print_fvector(dobj[o].ttoa, "   target ", _NB_, 1, 3);
  print_fvector(dobj[o].etoa, "   environ", _NB_, 1, 3);
  print_fvector(dobj[o].est,  "   aod img", _NB_, 1, 3);
  print_fvector(dobj[o].aod,  "   aod fit", _NB_, 1, 3);
  print_fvector(dobj[o].coef, "   fitcoef", 3, 3, 2);
  printf("   Rsq %01.2f, spec # %d, fit # %d\n", dobj[o].rsq, dobj[o].lib_id, dobj[o].ang_fit);

  return;
}

#endif

