/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
This file contains functions for reading directories and handling names
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#include "force.h"


/** Test for existing file/directory
+++ Return: true/false
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
bool fileexist(char *fname){

  if (access(fname, F_OK) == 0) return true; else return false;
}


/** Find file matching a pattern
+++ This function searches the given directory for a file, which contains
+++ the given pattern. The first match will be returned. NULL will be re-
+++ turned if no file was found.
--- dir_path: directory
--- pattern:  pattern
+++ Return:   file path
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
char *findfile(char *dir_path, char *pattern){
DIR *dp;
struct dirent *ep;
char *fname;

  dp = opendir(dir_path);
  fname = (char*) malloc(sizeof(char)*1024);

  if (dp != NULL){
    while ((ep = readdir(dp))){
      if (ep->d_name && strstr(ep->d_name, pattern)){
        sprintf(fname, "%s/%s", dir_path, ep->d_name);
        (void) closedir(dp);
        return fname;
      }
    }
    (void) closedir(dp);
  } else perror("Couldn't open the directory");

  free((void*) fname);
  return NULL;
}


/** Find file matching a pattern with respect to a filter
+++ This function searches the given directory for a file, which contains
+++ the given pattern and filter (e.g. ".txt"). The first match will be 
++++ returned. NULL will be returned if no file was found.
--- dir_path: directory
--- pattern:  pattern
--- filter:   filter
+++ Return:   file path
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
char *findfile_filter(char *dir_path, char *pattern, char *filter){
DIR *dp;
struct dirent *ep;
char *fname;

  dp = opendir(dir_path);
  fname = (char*) malloc(sizeof(char)*1024);

  if (dp != NULL){
    while ((ep = readdir(dp))){
      if (ep->d_name && strstr(ep->d_name, pattern) && strstr(ep->d_name, filter)){
        sprintf(fname, "%s/%s", dir_path, ep->d_name);
        (void) closedir(dp);
        return fname;
      }
    }
    (void) closedir(dp);
  } else perror("Couldn't open the directory");

  free((void*) fname);
  return NULL;
}


/** Count files matching a pattern
+++ This function searches the given directory for files, which contains
+++ the given pattern. The number of files will be returned.
--- dir_path: directory
--- pattern:  pattern
+++ Return:   number of files
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int countfile(char *dir_path, char *pattern){
DIR *dp;
struct dirent *ep;
dp = opendir(dir_path);
int k = 0;

  if (dp != NULL){
    while ((ep = readdir(dp))){
      if (ep->d_name && strstr(ep->d_name, pattern)) k++;
    }
    (void) closedir(dp);
  } else perror("Couldn't open the directory");

  return k;
}


/** Create directory
+++ This function creates a directory. The function will return SUCCESS if
+++ the directory was successfully created or if it was already existent.
--- dir_path: directory
+++ Return:   SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int createdir(char *dir_path){
int e, errno;
struct stat sb;

  if ((e = stat(dir_path, &sb)) != 0){
    if (errno == ENOENT){
      if ((e = mkdir(dir_path, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH)) != 0){
        printf("mkdir failed\n"); exit(1);}
    }
  }

  return SUCCESS;
}


/** Basename of file without extension
+++ This function extracts the basename from a file and removes the file
+++ extension.
--- path:   file path
+++ Return: basename
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
char *basename_without_ext(char* path){
int plen, blen;
char *start, *dot, *basename;

  if (path == NULL) return NULL;

  plen = strlen(path);

  // Strip trailing slashes if any
  while (path[plen-1] == '/'){
    path[plen-1] = '\0';
    plen--;
  }

  // Locate the last slash
  start = strrchr(path, '/');
  if (start == NULL){
    start = path;
  } else start++;

  blen = strlen(start);
  basename = (char*)malloc(blen*sizeof(char)+1);
  strncpy(basename, start, blen+1);

  // Locate the first dot
  dot = strchr(basename, '.');
  if (dot != NULL) *dot = '\0';

  return basename;
}


/** Basename of file
+++ This function extracts the basename from a file.
--- path:   file path
+++ Return: basename
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
char *basename_with_ext(char* path){
int plen, blen;
char *start, *basename;

  if (path == NULL) return NULL;

  plen = strlen(path);

  // Strip trailing slashes if any
  while (path[plen-1] == '/'){
    path[plen-1] = '\0';
    plen--;
  }

  // Locate the last slash
  start = strrchr(path, '/');
  if (start == NULL){
    start = path;
  } else start++;

  blen = strlen(start);
  basename = (char*)malloc(blen*sizeof(char)+1);
  strncpy(basename, start, blen+1);

  return basename;
}


/** Directory name of file
+++ This function extracts the directory name of a file.
--- path:   file path
+++ Return: directory name
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
char *directoryname(char* path){
int len;
char *slash, *dirname;

  if (path == NULL) return NULL;

  len = strlen(path);
  dirname = (char*)malloc(len*sizeof(char)+1);
  strcpy(dirname, path);

  // Locate the last slash
  slash = strrchr(dirname, '/');
  if (slash != NULL) *slash = '\0';

  return dirname;
}

