/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
Data type definitions
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#ifndef DTYPE_H
#define DTYPE_H

#ifdef __cplusplus
extern "C" {
#endif


/** software version
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

#define _VERSION_ "2.1"


/** named constants
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

#define SUCCESS  0
#define FAILURE -1

#define MAX_STR_LEN 1024
#define BIG_STR_LEN 8192

#define MAX_WIN_LEN 1024

#define MISSION_LANDSAT   0
#define MISSION_SENTINEL2 1

#define SCR_GAUSS   1
#define SCR_SIG_DES 2
#define SCR_SIG_ASC 3

#define AOD_WAT 1
#define AOD_SHD 2
#define AOD_VEG 3

#define NONE   0
#define LINEAR 1
#define MOVING 2
#define RBF    3

#define REGRESSION 1
#define IMPROPHE 2
#define STARFM 3


/** data types
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

typedef unsigned int int4u;
typedef signed int int4s;
typedef unsigned short int int2u;
typedef signed short int int2s;
typedef unsigned char int1u;
typedef signed char int1s;


/** date struct
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

typedef struct {
  int ce;      // days since current era
  int day;     // day
  int doy;     // day-of-year
  int week;    // week
  int month;   // month
  int year;    // year
  int hh;      // hour
  int mm;      // minute
  int ss;      // second
  int tz;      // timezone
} date_t;


/** coordinate struct
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

typedef struct {
  float x, y;
} coord_t;


/** struct for directory listing
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

typedef struct {
struct dirent **LIST; // files in directory
int N;                // number of files in directory
char **list;          // filtered files
int n;                // number of filtered files
} dir_t;


/** cURL struct used for downloading data
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

typedef struct {
  const char *fname;      // filename
  const char *dname;      // directory name
  char fcat[MAX_STR_LEN]; // concatenated filename
  FILE *fp;               // file stream
  long int size;          // file size
} curl_t;


/** First-In-First-Out queue structs
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

// FIFO queue element
typedef struct node_t {
  int i, j;
  node_t *prev, *next;
} node_t;

// FIFO queue
typedef struct {
  int length;
  node_t *first, *last;
} queue_t;


/** Struct for mapping variables in memory to output
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

typedef struct {
  bool save;               // save product flag
  int nodata;              // nodata value
  int start;               // band ID where product starts
  int stop;                // band ID where product stops
  int length;              // number of bands for product
  char product[4];         // product short name
  char fname[MAX_STR_LEN]; // filename for product
  int2s **data;            // data
  int2s *scale;            // scale factor
  float *ts;               // timestamp
  date_t *date;            // date
  char **bn;               // band names
  char **sid;              // sensor ID
} mem_to_disc_t;


/** Struct for band / product usage
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

typedef struct {
  bool ref, *refbands;
  bool qai;
  bool imp;
  bool tem, cld, aod;
  bool hot, vzn, wvp;
} use_t;


/** Struct for Level 2 band dictionary
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

typedef struct {
  int    n;
  int  **band;
  char **sensor;
  char   target[MAX_STR_LEN];
} level2_sensor_t;


/** Level-1 structs
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

typedef struct {
  int2u **ref; // Digital numbers, reflective bands
  int2u  *bt;  // Digital numbers, brightness temperature
} dn_t;

typedef struct {
  int2s **ref; // TOA, reflective bands
  int2s  *bt;  // TOA, brightness temperature
} toa_t;


/** Level-2 struct
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

typedef struct {
  int2s **ref;  // BOA/TOA reflectance
  int2s  *qai;  // quality assurance information (bit-coding)
  int2s  *cld;  // cloud / cloud shadow distance
  int2s  *aod;  // aerosol optical depth @ 550 nm
  int2s  *hot;  // haze optimized transformation
  int2s  *vzn;  // view zenith angle
  int2s  *wvp;  // water vapor
  int     sid;  // sensor id
  date_t  date; // acquisition date
} level2_t;


/** ImproPhe struct
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

typedef struct {
  float **AVG;
  float **MR;
  float **CR;
  float  *MX;
  float **CX;
  float **PRED;
} imp_t;


/** Quality Assurance Information struct
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

typedef struct {
  bool  *off; // valid data flag
  int1u *cld; // cloud state flag
  bool  *shd; // cloud shadow flag
  bool  *snw; // snow flag
  bool  *wtr; // water flag
  int1u *aod; // aerosol state
  bool  *sub; // subzero flag
  bool  *sat; // saturation flag
  bool  *sun; // high sun zenith flag
  int1u *ill; // illumination state flag
  bool  *slp; // slope flag
  bool  *wvp; // water vapor flag
  // one bin is still empty
} qai_t;


/** Topographic derivatives struct
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

typedef struct {
  int1u *dem; // binned digital elevation model
  int2u *slp; // slope
  int2u *asp; // aspect
  int2s *ill; // illumination angle
  int2u *sky; // sky view factor
  int2u *cor; // C-factor SWIR2
} top_t;


/** Metadata / calibration struct
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

typedef struct {
  int   rsr_band;           // ID in RSR table
  char  mtl_band[5];        // Band ID in original file name
  char  fname[MAX_STR_LEN]; // file name
  float lmax, lmin;         // radiance min/max
  float qmax, qmin;         // quantized DN min/msx
  float rmul, radd;         // reflectance scaling factor
  float E0;                 // solar irradiance
  float dsun;               // distance sun-earth
  float wvl;                // wavelength
  float k1, k2;             // conversion factors brightness temperature
} cal_t;


typedef struct {
  float **szen, **sazi; // sun  zenith / azimuth from metadata
  float **vzen, **vazi; // view zenith / azimuth from metadata
  int nx, ny;           // number of cells in angle grid
} sentinel2_t;


typedef struct {
  int id;                    // Landsat sensor ID
  int mission;               // satellite mission ID
  char satellite[6];         // satellite short name
  int dtype;                 // data type (bytes)
  cal_t *refcal;             // calibration DN->TOA reflectance
  cal_t  btcal;              // calibration DN->TOA brightness temperature
  int scale;                 // scaling factor
  int res;                   // spatial resolution
  coord_t map_ul;            // upper-left coordinate
  char css_map[MAX_STR_LEN]; // map projection
  char css_geo[MAX_STR_LEN]; // geographic projection
  char wrs[MAX_STR_LEN];     // worldwide reference system ID
  date_t date;               // acquisition date
  int tier;                  // tier level
  sentinel2_t s2;            // Sentinel-2 calibration specific
} meta_t;


/** Cloud matching structs
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

typedef struct {
  int **X;
  int **Y;
  float  **BT;
} cloud_t;


/** AOD estimation structs
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

// spectral library
typedef struct {
  int n;     // number of spectra
  float **s; // spectra
} speclib_t;

// single dark target
typedef struct {
  int k;                      // number of valid pixels
  bool valid;                 // valid object?
  int imin, imax, jmin, jmax; // bounding box
  int i, j;                   // location of centroid
  int gy, gx, g;              // coarse grid cell
  float  *ttoa, *etoa;        // target and environment reflectance
  int r;                      // target radius
  float cosi, sky, z;         // target illumination angle, sky view and elevation
  float Ha, Hr;               // elevation correction factors for AOD/MOD
  float  *aod, *est;          // fitted and estimated AOD
  float rsq;                  // R-squared of best fit, and ID of spectrum
  int lib_id;                 // ID of best spectrum
  int ang_fit;                // type of AOD fit; 0: curved, 1: linear
  float coef[3];              // Angstrom coefficients of best fit
} darkobj_t;

// dark target container
typedef struct {
  darkobj_t *wat;       // water targets
  darkobj_t *shd;       // shadow targets
  darkobj_t *veg;       // vegetation targets
  int kwat, kshd, kveg; // number of candidate targets
  int nwat, nshd, nveg; // number of valid targets
  float **cg_wat;       // AOD map water
  float **cg_shd;       // AOD map shadow
  float **cg_veg;       // AOD map vegetation
  float **cg_avg;       // AOD map averaged
} dark_t;


/** atmospheric correction structs
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

// viewing geometry
typedef struct {
  float a, b, c, ab;     // cartesian definition of nadir line
  float geo_angle_nadir; // angle of nadir line
  float H, H2;           // satellite elevation, and square of elevation
} view_t;

// elevation stats/bins
typedef struct {
  float avg, min, max; // average/min/max elevation
  float step;
  //float cmin, cmax;    // min/max of binned elevation
  int cnum;            // number of elevation bins
} dem_t;

// two-term Henyey-Greenstein coefficicents
typedef struct {
  float g1, g2, alpha;
  float hg1, hg2, hg3, hg4, hg5, hg6, sob1;
} tthg_t;

// atmospheric parameters for elevation bin
typedef struct {
  float *od, *mod, *aod; // total, molecular, aerosol optical depth
  float Hr, Ha;          // elevation correction factors for AOD/MOD
  float *rho_p;          // path reflectance
  float *Ts, *tsd, *tss; // scattering transmittance, down
  float *Tv, *tvd, *tvs; // scattering transmittance, up
  float *T;              // total scattering transmittance
  float *s;              // spherical albedo
  float *F;              // environmental reflectance weighting function
  float wvp;             // water vapor concentration
  float *Tg;             // gaseous transmittance
} zgrid_t;

// atmospheric parameters for coarse grid cell
typedef struct {
  float mapx, mapy;       // map coordinate
  float geox, geoy;       // geogr. coordinate
  int   cdem;             // elevation bin
  float vzen, szen;       // view and sun zenith  angle
  float vazi, sazi;       // view and sun azimuth  angle
  float cosvzen, cosszen; // cos. of view and sun zenith  angle
  float cosvazi, cossazi; // cos. of view and sun azimuth angle
  float sinvzen, sinszen; // sin. of view and sun zenith  angle
  float sinvazi, sinsazi; // sin. of view and sun azimuth angle
  float tanvzen, tanszen; // tan. of view and sun zenith  angle
  float tanvazi, tansazi; // tan. of view and sun azimuth angle
  float mv, ms;           // air mass, up & down
  float cospsi;           // backscattering angle
  float *mod, *aod;       // molecular, aerosol optical depth
  float Hr, Ha;           // elevation correction factors for AOD/MOD
  float Pr, Pa;           // phase function, molecular and aerosol scatt.
  bool  aod_int;          // interpolated AOD flag
  float rhof;             // Fresnel reflection
  float wvp, ozone;       // water vapor, ozone concentration
  float *Tvw, *Tsw;       // water vapor transmittance, up & down
  float *Tvo, *Tso;       // ozone transmittance, up & down
  float *Tg;              // gaseous transmittance
  float *brdf;            // BRDF correction factor
  zgrid_t *z;             // atm. parameters for elevation bins
} grid_t;

// atmospheric parameters for image
typedef struct {
  int ngx, ngy, ngc, gres;   // number and resolution of coarse grid cells
  view_t view;               // viewing geometry
  float *wvl, *lwvl, *lwvl2; // wavelengths, log of w, square of log of w
  float cc;                  // cloud cover
  dem_t dem;                 // elevation stats/bins
  grid_t *cg;                // atm. parameters for coarse grid cells
  float cosszen[2];          // min/max cos. of sun zenith
  float cosvzen[2];          // min/max cos. of view zenith
  float *od;                 // total optical depth
  float *mod;                // molecular optical depth
  float *aod;                // aerosol optical depth (from image)
  float *aod_ext;            // aerosol optical depth (from LUT)
  float Hr, Ha;              // elevation correction factors for AOD/MOD
  tthg_t tthg;               // two-term Henyey-Greenstein coefficicents
  bool *aod_bands;           // flag: use band to estimate AOD
  bool aodmap;               // flag: AOD map successful
  float Fr, Fa, km;          // environmental reflectance weighting function
  float wvp;                 // water vapor concentration
  float ***Tw_lut;           // water vapor transmittance LUT
} atc_t;


/** band dictionary struct
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

typedef struct {
  int ultrablue; // band ID for ~ 0.44 µm
  int blue;      // band ID for ~ 0.48 µm
  int green;     // band ID for ~ 0.55 µm
  int red;       // band ID for ~ 0.65 µm
  int rededge1;  // band ID for ~ 0.70 µm
  int rededge2;  // band ID for ~ 0.74 µm
  int rededge3;  // band ID for ~ 0.78 µm
  int nir;       // band ID for ~ 0.86 µm
  int swir1;     // band ID for ~ 1.60 µm
  int swir2;     // band ID for ~ 2.20 µm
  int bnir;      // band ID for high resolution NIR
  int cir;       // band ID for ~ 1.37 µm
  int w0, w1;    // band ID for water vapor reference and measurement
  int ref;       // band ID (output stack) for reflectance
  int qai;       // band ID (output stack) for quality assurance inf.
  int cld;       // band ID (output stack) for cloud/shadow distance
  int hot;       // band ID (output stack) for haze optimized transf.
  int aod;       // band ID (output stack) for AOD @ 550nm
  int wvp;       // band ID (output stack) for water vapor
  int vzn;       // band ID (output stack) for view zenith
  int bap;       // band ID for best available pixel composite
  int inf;       // band ID for compositing information
  int scr;       // band ID for compositing scores
  int num;       // band ID for temporal average
  int avg;       // band ID for temporal average
  int std;       // band ID for temporal standard deviation
  int min;       // band ID for temporal minimum
  int max;       // band ID for temporal maximum
  int rng;       // band ID for temporal range
  int skw;       // band ID for temporal skewness
  int krt;       // band ID for temporal kurtosis
  int q25;       // band ID for temporal 25%-Quantile
  int q50;       // band ID for temporal 50%-Quantile
  int q75;       // band ID for temporal 75%-Quantile
  int iqr;       // band ID for temporal Interquantile range
  int tss;       // band ID for time series stack
  int rms;       // band ID for model residuals
  int sta;       // band ID for basic statistics
  int tsi;       // band ID for time series interpolation
  int fby;       // band ID for fold-by-year time series
  int fbm;       // band ID for fold-by-month time series
  int fbw;       // band ID for fold-by-week time series
  int fbd;       // band ID for fold-by-doy time series
  int trd;       // band ID for trend parameters
  int cat;       // band ID for change and trend parameters  
  int dem, dss, dri, dps, dfi, des;
  int dlm, lts, lgs, vem, vss, vri;
  int vps, vfi, ves, vlm, vbl, vsa;
  int ist, ibl, ibt, igs, rar, raf;
  int rmr, rmf;  // band IDs for LSP metrics
} bands_t;


/** compositing structs
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

// compositing target
typedef struct {
  float ***p;     // target DOY, i.e. DOY triplet p0, p1, p2
  float **a, **b; // fitted parameters for logistic S-curve or Gaussian
} target_t;

// compositing scores
typedef struct {
  float d; // DOY score
  float y; // year score
  float c; // cloud score
  float h; // haze score
  float r; // correlation score
  float v; // view zenith score
  float t; // total score
  int dd;  // difference to target DOY
} score_t;


/** StarFM candidate struct
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

typedef struct {
  int dx, dy;     // distance
  float coarse;   // coarse data
  float fine;     // fine data
  float r_coarse; // relative difference between two coarse obs.
  float r_fine;   // relative difference between coarse and fine obs.
  float C;        // reversed distance
} starfm_candidate_t;


/** processing parameter struct
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

typedef struct {

  /** input/output filenames/directories **/
  char *f_par;                // parameter file
  char *d_img;                // input Level-1 image (directory)
  char *b_img;                // basename of input Level-1 image
  char d_out[MAX_STR_LEN];    // exact output directory
  char b_out[MAX_STR_LEN];    // basename of output file
  char fdem[MAX_STR_LEN];     // Digital Elevation Model
  char f_tile[MAX_STR_LEN];   // tile white-list
  char f_queue[MAX_STR_LEN];  // file queue
  char d_aod[MAX_STR_LEN];    // directory of AOD LUT
  char d_wvp[MAX_STR_LEN];    // directory of water vapor LUT
  char d_dob[MAX_STR_LEN];    // directory of DODB
  char d_level2[MAX_STR_LEN]; // Level-2 directory
  char d_level3[MAX_STR_LEN]; // Level-3 directory
  char d_tsa[MAX_STR_LEN];    // Level-4 TSA directory
  char d_cso[MAX_STR_LEN];    // clear-sky stats directory
  char d_tmp[MAX_STR_LEN];    // temporary directory
  char d_lsp[MAX_STR_LEN];    // phenology directory
  char f_lsp[3][MAX_STR_LEN]; // phenology file names
  char p_lsp[3][MAX_STR_LEN]; // phenology file name patterns
  char d_mask[MAX_STR_LEN];   // mask directory
  char f_endmem[MAX_STR_LEN]; // endmember file

  /** output parameters **/
  int format;                 // output format
  bool oref, voref;           // flag: output reflectance
  bool oqai, voqai;           // flag: output quality assurance information
  bool ocld, vocld;           // flag: output cloud/shadow distance
  bool ovzn, vovzn;           // flag: output view zenith angle
  bool ohot, vohot;           // flag: output haze optimized transformation
  bool oaod, voaod;           // flag: output AOD @ 550nm
  bool owvp, vowvp;           // flag: output water vapor
  bool obap, vobap;           // flag: output best available pixel composite
  bool oinf, voinf;           // flag: output compositing information
  bool oscr, voscr;           // flag: output compositing scores
  bool onum, vonum;           // flag: output nobs
  bool oavg, voavg;           // flag: output average
  bool ostd, vostd;           // flag: output standard deviation
  bool omin, vomin;           // flag: output minimum
  bool omax, vomax;           // flag: output maximum
  bool orng, vorng;           // flag: output range
  bool oskw, voskw;           // flag: output skewness
  bool okrt, vokrt;           // flag: output kurtosis
  bool oq25, voq25;           // flag: output 25% Quantile
  bool oq50, voq50;           // flag: output 50% Quantile
  bool oq75, voq75;           // flag: output 75% Quantile
  bool oiqr, voiqr;           // flag: output Interquantile range
  bool otss, votss;           // flag: output time series stack
  bool orms, vorms;           // flag: output model residuals
  bool osta, vosta;           // flag: output basic statistics
  bool otsi, votsi;           // flag: output time series interpolation
  bool ofby, vofby;           // flag: output fold-by-year time series
  bool ofbm, vofbm;           // flag: output fold-by-month time series
  bool ofbw, vofbw;           // flag: output fold-by-week time series
  bool ofbd, vofbd;           // flag: output fold-by-doy time series
  bool otrd, votrd;           // flag: output trend parameters
  bool ocat, vocat;           // flag: output change and trend parameters
  bool olsp, volsp;           // flag: output LSP metrics

  mem_to_disc_t m2d[MAX_STR_LEN];     // map between memory and output

  /** tile parameters **/
  bool dotile, vtile;         // flag: tile
  float tilesize;             // tile size in desination unit
  int tilesize_px;            // tile size in pixels
  int tminx, tmaxx;           // processing extent min/max tile
  int tminy, tmaxy;           // processing extent min/max tile
  int tnx, tny;               // number of tiles in x-, y-direction

  /** projection parameters **/
  bool doreproj, vreproj;     // flag: reproject
  float res;                  // spatial resolution
  int resample;               // resampling option
  coord_t origin_geo;         // origin of grid, geographic
  coord_t origin_map;         // origin of grid, destination projection
  char css_dst[MAX_STR_LEN];  // destination projection
  
  /** resolution parameters **/
  bool psf, vpsf;             // flag: point spread function
  bool vimp;                  // flag: use ImproPhe'd data
  int  resmerge;              // resolution merge

  /** L2 processing parameters **/
  bool dotopo, vtopo;         // flag: topographic correction
  bool doatmo, vatmo;         // flag: atmospheric correction
  bool dobrdf, vbrdf;         // flag: BRDF reduction
  bool dododb, vdodb;         // flag: use DODB
  bool doaod,  vaod;          // flag: estimate AOD
  bool domulti, vmulti;       // flag: multiscattering
  bool doenv, venv;           // flag: environment reflectance
  bool band_bt;               // flag: brightness temperature available
  float maxcc, maxtc;         // max. allowable cloud cover per scene/tile
  float wvp;                  // water vapor dummy value
  float cldprob, shdprob;     // Fmask thresholds
  int svgrid;                 // resolution of coarse grid
  int tier;                   // tier level

  /** dictionaries **/
  level2_sensor_t l2dict;     // Level-2 sensor dictionary
  use_t use;                  // band/product usage
  
  /** QAI filters **/
  bool vqai;         // QAI parameter set?
  bool qai_off;      // filter out nodata
  bool qai_cld_opq;  // filter out opaque, confident clouds
  bool qai_cld_unc;  // filter out opaque, less confident clouds
  bool qai_cld_cir;  // filter out cirrus clouds
  bool qai_shd;      // filter out cloud shadow
  bool qai_snw;      // filter out snow
  bool qai_wtr;      // filter out water
  bool qai_aod_fill; // filter out filled AOD
  bool qai_aod_high; // filter out high AOD
  bool qai_aod_int;  // filter out interpolated AOD
  bool qai_sub;      // filter out subzero
  bool qai_sat;      // filter out saturated
  bool qai_sun;      // filter out high sun zenith
  bool qai_ill_shd;  // filter out no illumination
  bool qai_ill_poor; // filter out poor illumination
  bool qai_ill_low;  // filter out low illumination
  bool qai_slp;      // filter out sloped
  bool qai_wvp;      // filter out no water vapor estimation

  /** temporal parameters **/
  int doymin, doymax;         // min/max DOY
  int weekmin, weekmax;       // min/max week
  int monthmin, monthmax;     // min/max month
  int yearmin, yearmax;       // min/max year
  int cemin, cemax;       // min/max date since CE
  int Yt;                     // target year
  int Yr;                     // number of bracketing years
  int Yn;                     // number of years
  float Yf;                   // Y-factos
  int Dt[3];                  // target DOYs
  float s[3];                 // function values for target DOYs
  bool voffsea, offsea;       // use off-season data?
  
  /** phenology-adaptive compositing **/
  bool lsp, vlsp;             // flag: phenology-adaptive compositing
  int  lsp_y0;                // 1st year of phenology
  int  lsp_start;             // start of LSP values
  float lsp_nodata;           // nodata of phenology
  float lsp_rmse;             // cutoff threshold for phenology

  /** other compositing parameters **/
  int score_type;             // scoring type
  score_t w;                  // weighting factors for scoring functions
  float dreq, vreq;           // cloud / view zenith scoring  
 
  /** time series parameters **/
  int index;                  // index type
  int trend;                  // trend type
  int fold;                   // folding type
  int tail;                   // tail type
  char index_c[4];            // short name index type
  char trend_c[2];            // short name trend type
  char fold_c[4];             // short name folding type
  char tail_c[2];             // short name tail type
  float conf;                 // confidence level
  bool docenter, vcenter;     // flag: center time series
  bool dostandard, vostandard;// flag: standardize time series
  float despike;              // despiking threshold

  /** interpolation parameters **/
  int interpolate;            // interpolation method
  int int_nstep;              // interpolate each n days
  int int_n;                  // number of interpolated dates  
  int mov_max;                // max temp. dist for moving stats filter
  int rbf_nk;                 // number of kernels for RBF fit
  int rbf_sigma[MAX_STR_LEN]; // sigmas for RBF fit
  float rbf_cutoff;           // cutoff for RBF fit

  /** sma parameters **/
  bool vsma_sto, sma_sto;     // sum-to-one constrained
  bool vsma_pos, sma_pos;     // non-negative constrained
  bool vsma_shn, sma_shn;     // shade normalization
  int sma_emb;                // endmember number

  /** phenology parameters **/
  int lsp_dprev;              // use data from this date (year-1)
  int lsp_dnext;              // until this date (year+1)
  int lsp_hemi;               // southern/mixed/northern hemisphere
  int lsp_nseg;               // number of segments per year

  /** improphe parameters **/
  char f_cr[MAX_STR_LEN][MAX_STR_LEN];
  char d_improphe[MAX_STR_LEN];
  int num_cr;
  int ymin, ymax;
  int ywin[MAX_STR_LEN], dwin[MAX_STR_LEN], mwin[MAX_STR_LEN];
  int cewin[MAX_STR_LEN];
  int bwin, nwin;
  int ystart_cr;
  float nodata_cr;
  int ksize;
  int ksd;
  char imp_prd[MAX_STR_LEN][MAX_STR_LEN];
  char csen_mr[MAX_STR_LEN];
  char csen_cr[MAX_STR_LEN];
  
  int monthstep;

  /** miscellaneous **/
  int cpu;                    // number of CPUs for parallel processing
  bool vowr, owr;             // flag: overwrite output

} par_t;


#ifdef __cplusplus
}
#endif

#endif

