/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
Top-level header file for FORCE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#ifndef FORCE_H
#define FORCE_H


/** libraries
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/** C Standard library **/
#include <float.h>   // macro constants of the floating-point library
#include <limits.h>  // macro constants of the integer types
#include <math.h>    // common mathematical functions
#include <stdbool.h> // boolean data type
#include <stdio.h>   // core input and output functions
#include <stdlib.h>  // standard general utilities library
#include <string.h>  // string handling functions
#include <ctype.h>       // transform individual characters
#include <time.h>    // date and time handling functions

/** C POSIX library **/
#include <dirent.h>      // allows the opening and listing of directories
#include <sys/utsname.h> // system name structure

/** GNU Scientific Library (GSL) **/
#include <gsl/gsl_errno.h>             // error reporting and handling
#include <gsl/gsl_statistics_double.h> // statistical functions
#include <gsl/gsl_roots.h>             // root finding functions
#include <gsl/gsl_multimin.h>          // minimization functions 
#include <gsl/gsl_multifit.h>          // multi-parameter fitting
//#include "gsl/gsl_matrix.h"
#include "gsl/gsl_blas.h"
#include "gsl/gsl_eigen.h"
#include "gsl/gsl_linalg.h"
//#include "gsl/gsl_sort_vector.h"

/** Geospatial Data Abstraction Library (GDAL) **/
#include "cpl_conv.h"       // various convenience functions for CPL
#include "cpl_string.h"     // various convenience functions for strings
#include "gdal.h"           // public (C callable) GDAL entry points
#include "gdalgrid.h"       // GDAL gridder related entry points and defs
#include "gdalwarper.h"     // GDAL warper related entry points and defs
#include "ogr_spatialref.h" // coordinate systems services
#include "ogr_api.h"        // OGR geometry and feature definition

/** OpenMP **/
#include <omp.h> // multi-platform shared memory multiprocessing

/** libcurl **/
#include <curl/curl.h> // multiprotocol file transfer


#ifdef __cplusplus
extern "C" {
#endif


/** FORCE header files
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

#include "table.h"
#include "dtype.h"
#include "global.h"


/** compiler options
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

// compile with SPLITS?
//#define SPLITS

// debugging compiler options:
//#define CLOCK
//#define DEBUG
//#define DEBUGPATH "/home/frantzda/temp"
//#define DEV
//#define PRINTDARK


/** functions in alloc.c
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void alloc(void **ptr, int n, int size);
void alloc_2D(void ***ptr, int n1, int n2, int size);
void alloc_3D(void ****ptr, int n1, int n2, int n3, int size);
void alloc_2DC(void ***ptr, int n1, int n2, int size);
void free_2D(void **ptr, int n);
void free_3D(void ***ptr, int n1, int n2);
void free_2DC(void **ptr);


/** functions in aod.c
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int aodfileread();
int dark_object_persistency(int2u **DOP);
int extract_dark_target(int2s **TOA, qai_t QAI, int2u *DOP, top_t TOP, int type, darkobj_t **DOBJ);
int aod_from_target(darkobj_t *dark, int num, float ***cg_aod, int type);
void get_water_lib(speclib_t *lib);
void get_land_lib(speclib_t *lib);
void get_veg_lib(speclib_t *lib);
int aod_lib_to_target(darkobj_t dobj, speclib_t lib, int type, float **aodest);
int aod_linear_fit(float *logaod, float *angb, float *angn);
int aod_polynomial_fit(float *logaod, float *a0, float *a1, float *a2);
double aod_polynomial_minimizer(const gsl_vector *v, void *params);
void aod_map(dark_t dark);
int interpolate_cg_aod(float **cg_aod);
void compile_aod(int2s **TOA, qai_t QAI, top_t TOP);


/** functions in atmo.c
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void atc_init();
void atc_free();
void cggrid(int2u **GRD);
void cgfree();
int sun_target_view(qai_t QAI);
void viewgeo(qai_t QAI);
void view_angle(int j, int i, int g, qai_t QAI);
void smooth_atc(zgrid_t *atc_2, int i, int j, int z);
void toa2boa(int2s **TOA, int2s **BCK, qai_t QAI, int2u *GRD, top_t TOP, int1u *CDEM, float **Tg, int2s *WVP, int2s ***DAT);
void background_ref(int2s **TOA, qai_t QAI, int2u *GRD, top_t TOP, int2s **Tg, int2s ***BCK);
void atmo_angledep(top_t TOP, qai_t QAI);
void atmo_elevdep();
int radiometric_module(int2s **TOA, qai_t QAI, top_t TOP, int2u *GRD, int2s ***DATA);

/** functions in brdf.c
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void brdf_factor(int g);
float brdf_forward(float ti, float tv, float phi, float iso, float vol, float geo);
void LiKernel(float hbratio, float brratio, float tantv, float tanti, float sinphi, float cosphi, float *result);
void GetPhaang(float cos1, float cos2, float sin1, float sin2, float cos3, float *cosres, float *res,float *sinres);
void GetpAngles(float brratio, float tan1, float *sinp, float *cosp, float *tanp);
void GetDistance(float tan1, float tan2, float cos3,float *res);
void GetOverlap(float hbratio, float distance, float cos1, float cos2, float tan1, float tan2, float sin3, float *overlap, float *temp);


/** functions in calib.c
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int calinit(cal_t *cal);
int E0();
int wavelengths();
void dn2toa(cal_t *cal, int2u *GRD, qai_t QAI, int2u *DN, int2s **TOA, int type);
int dn_to_toa(int2u *GRD, qai_t QAI, dn_t DN, toa_t *TOA);
int dn_extreme(dn_t DN, qai_t QAI);
void detect_impulse_noise(int2u **DN, qai_t QAI);


/** functions in cloud.c
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
float finalize_cloud(float npix, bool *CLOUD, bool *SHADOW, int2s **TOA, qai_t QAI);
int cloud_distance(qai_t QAI, int2s **DIST);
void potential_cloud(float *npix, int *nclear, int *nland, toa_t TOA, qai_t QAI, int2u *SLP, bool **CLR, bool **LND, bool **BRT, float **VPR, bool **PCP);
void cloud_probability(float npix, int nclear, int nland, int *ncloud, float *cc, float *lowt, float *hight, bool *CLEAR, bool *LAND, toa_t TOA, qai_t QAI, bool *BRIGHT, float *VPROB, bool *PCP, bool **OCLOUD);
void greyscale_reconstruction(float *MASK, float *MARKER);
void shadow_probability(int nland, bool *LAND, int2s **TOA, bool *CLOUD, qai_t QAI, int2u **SPROB);
int cloud_parallax(float npix, int *ncloud, float *cc, int2s **TOA, qai_t QAI, bool *BRIGHT, bool *PCP, bool **CLD);
void cloud_objects(bool *CLOUD, int **OBJ, int **SIZE, int *nobj);
void shadow_position(float h, int x, int y, int *newx, int *newy);
void shadow_matching(float lowtemp, float hightemp, qai_t QAI, bool *CLOUD, int2u *SLP, int2u *SPROB, int2s *BT, bool **SHD);
int cloud_module(toa_t TOA, qai_t QAI, top_t TOP);


/** functions in compo.c
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int print_bap_stats(off_t hist[BIG_STR_LEN], char *tag, int off, int multi);
double fun_logistic_target(const gsl_vector *v, void *params);
int optimize_logistic_target(float *param, float *a, float *b);
int temporal_target_function(float ***P, float ***A, float ***B);
int compile_temporal_target(int tx, int ty, target_t *TAR);
int corr_matrix(level2_t *L2, qai_t *QAI, int p, double **cor);
int compositing_scores(level2_t *L2, qai_t *QAI, target_t tar, int p, double **cor, score_t *score);


/** functions in date.c
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void doy2md(int doy, int* m, int* d);
int md2doy(int m, int d);
int doy2week(int doy);
int date2ce(int m, int d, int y);
int doy2ce(int doy, int y);
void ce2date(int ce, int* m, int* d, int* y);
void ce2doy(int ce, int* doy, int* y);
void current_date(date_t *d);
void date_set(date_t d);
void date_plus(date_t *d);
void date_minus(date_t *d);


/** functions in debug.c
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void db_check();
int db_write_header(char *fname, int nx, int ny, int nb, int dt, char *mapinfo);
void db_write_image_1D(void *IMG, char *base, int nx, int ny, size_t size, int envidt);
void db_write_image_2D(void **IMG, char *base, int nb, int nx, int ny, size_t size, int envidt);
void db_write_image_CG(atc_t atc);
void db_print_dark_object(darkobj_t *dark, int o, int g);


/** functions in dir.c
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
bool fileexist(char *fname);
char *findfile(char *dir_path, char *pattern);
char *findfile_filter(char *dir_path, char *pattern, char *filter);
int countfile(char *dir_path, char *pattern);
int createdir(char *dir_path);
char *basename_without_ext(char* path);
char *basename_with_ext(char* path);
char *directoryname(char* path);


/** functions in download.c
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int download_file(char *f_remote, char *f_local);
int download_pattern(char *d_local, char *pattern);
long curl_open_download(struct curl_fileinfo *remote, curl_t *data, int remains);
size_t curl_write_download(void *ptr, size_t size, size_t count, void *data);
long curl_close_download(curl_t *data);


/** functions in imgfuns.c
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int bsum(bool *B, int n);
float find_sigma(float r);
int gauss_kernel(int nk, float sigma, float ***kernel);
int distance_kernel(int nk, float ***kernel);
void dist_transform(bool *B, int2u *DT);
int dt_dfun(int x, int i, int y, int2u *G);
int dt_Sep(int i, int u, int y, int2u *G);
int connectedcomponents(bool *B, int *CCL);
void ccl_tracer(int *cy, int *cx, int *dir, bool *B, int *CCL);
void ccl_contourtrace(int cy, int cx, int label, int dir, bool *B, int *CCL);
int focal_sd(float **DAT, float nodata, int h, int nx, int ny, int nb, int bstart, float **SD);
void buffer(bool *B, int r);
void majorfill(bool *B, qai_t QAI);


/** functions in improphe.c
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int improphe(imp_t IMP, float **KDIST, float nodata, int i, int j, int pred_p, int h, int nx, int ny, int nb_m, int nb_c, int nk, int mink);
int improphe_sentinel2(int2s **TOA, qai_t QAI, int2s **PRED, float **KDIST, int i, int j, int pred_p, int h, int nx, int ny, int nb_m, int nb_c, int *bands_m, int *bands_c, int nk, int mink);
double rescale_weight(double weight, double minweight, double maxweight);
int average_season(level2_t *L2, int n, float **AVG, int x_off, int y_off, int type);
int standardize_float(float *data, float nodata, int nobs);


/** functions in index.c
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int spectral_index(level2_t *L2, bool *MASK);
int common_index(level2_t L2, bool *MASK, int type, int2s *index);
int sma_index(level2_t L2, bool *MASK, float **endmember, int n_endmember, int2s *index, int2s *rmse);
int disturbance_index(level2_t L2, bool *MASK, int2s *index);


/** functions in input.c
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int image_info(char *fname, double *geotran);
int image_info_fullres(char *fname, double *geotran);
int mosaic_info(double *geotran);
int reduce_psf(int2s *hr, int nx, int ny, int nc, int NX, int NY, int NC, int2s *lr);
int read_single_image_bool(char *file, bool **img);
int read_single_image_2u(char *file, int2u **img);
int read_single_image_2s(char *file, int2s **img, int2s nodata, float fnx_, float fny_,  bool psf);
int read_multi_image_2s(char *file, int2s ***img, int2s nodata, int *bands, float fnx_, float fny_, bool psf);
int read_multi_image_fl(char *file, float ***img, int *nb);
int read_level1(dn_t *DN);
int read_level2(char *dname, level2_t **L2, float fnx_, float fny_);
int read_mask(char *dname, bool **MSK);
int buffer_level2(level2_t *L2);
int free_level2(level2_t *L2);
int date_level2(date_t *DATE, char *bname);
int sensor_level2(int *sid, char *bname);
int list_level2(char *dname);
int list_level2_filter_ce(int cemin, int cemax);
int list_level2_filter_multiday();
int band_available(int b, bool *valid);
int level2sensor_dict(char *request, level2_sensor_t *l2s);
int parse_coord_list(char *fcoords, float ***COORD);
int parse_endmember_list(char *fendmem, float ***ENDMEM);


/** functions in lsp.c
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int list_lsp(int tx, int ty);
int offset_lsp(int nb, float **LSP0);
int read_lsp(char *fname, float ***LSP);
int annual_lsp(int nb, float **LSP0, float ***LSP1);


/** functions in meta.c
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int mtlinit();
int mtlcheck();
int mtlprint();
int parseMetadataLandsat();
int parseMetadataSentinel();
void parse_metaband(char *tag, char *value, cal_t *cal, int type);
int parseMission();
void init_bands();
int parseMetadata();


/** functions in output.c
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int mem_to_disc_dict_init();
int mem_to_disc_free();
int mem_to_disc_dict_l2();
int mem_to_disc_dict_l3();
int mem_to_disc_dict_tsa();
int mem_to_disc_dict_cso();
int mem_to_disc_dict_il2();
int init_products_l3(double geotran[6]);
int final_products_l3(double geotran[6]);
int append_envi_header(char *fname, char *desc, char *prd, int level, int nx, int ny, int nb, double geotran[6], char **bn, float *wvl, char *wvu, char *acq, int nodata, float dc, float occ, float bcc, float ccc, float csc, float sc, float wc, int2s *scale, int *zlim, char **zlab, char **sid, float sa, float se);
int write_products_l2(int2s **DAT, int nx, int ny, int xmid, int ymid, double geotran[6]);
int write_products_l3(int tx, int ty, double geotran[6]);
int write_products_tsa(int tx, int ty, mem_to_disc_t m2d, double geotran[6]);
int write_products_cso(int tx, int ty, mem_to_disc_t m2d, double geotran[6]);
int write_products_imp(int tx, int ty, int y, imp_t IMP, double geotran[6]);
int write_products_il2(int tx, int ty, mem_to_disc_t m2d, double geotran[6]);


/** functions in pansharpen.c
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int resolution_merge(int2s **TOA, qai_t QAI);
int resolution_merge_1(int2s **TOA, qai_t QAI);
int resolution_merge_2(int2s **TOA, qai_t QAI);
int resolution_merge_3(int2s **TOA, qai_t QAI);
int starfm(qai_t QAI, float **coarse, float **fine, float **pred, int r);


/** functions in param.c
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
bool parse_logical(char *tag, char *value, bool *param);
bool parse_float(char *tag, char *value, float *param, float multi, float min, float max);
bool parse_int(char *tag, char *value, int *param, int multi, int min, int max);
int parse_char_seq(char *tag, char *str, char param[MAX_STR_LEN][MAX_STR_LEN], int *n);
bool parse_float_seq(char *tag, char *str, float *param, int *n, float multi, float min, float max);
bool parse_int_seq(char *tag, char *str, int *param, int *n, int multi, int min, int max);
int parse_quality(char *request);
void init_param_l2();
void init_param_l3();
void init_param_tsa();
void init_param_cso();
void init_param_imp();
void init_param_il2();
int check_param_l2();
int check_param_l3();
int check_param_tsa();
int check_param_cso();
int check_param_imp();
int check_param_il2();
int log_quality(FILE *fp);
int log_param_l2(FILE *fp);
int log_param_l3(FILE *fp);
int log_param_tsa(FILE *fp);
int log_param_cso(FILE *fp);
int log_param_imp(FILE *fp);
int log_param_il2(FILE *fp);
int parse_param_l2();
int parse_param_l3();
int parse_param_tsa();
int parse_param_cso();
int parse_param_imp();
int parse_param_il2();


/** functions in quality.c
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int alloc_qai(qai_t *QAI);
int free_qai(qai_t QAI);
void int2bit(int integer, bool *bin, int start, int size);
int bit2int(bool *bin, int start, int size);
bool use_this_pixel(qai_t QAI, int p);
int assemble_qai(qai_t QAI, int2s **DAT, int qa_band);
int parse_qai(qai_t QAI, int2s *DAT);


/** functions in queue.c
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
queue_t* create_queue();
void destroy_queue(queue_t* q);
int enqueue(queue_t* q, int i, int j);
int dequeue(queue_t* q, int* i, int* j);


/** functions in radtran.c
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
float molecular_optical_depth(float wvl);
void od_elev_factor(float z, float *Ha, float *Hr);
float mod_elev_scale(float mod, float Hr_is, float Hr);
float aod_elev_scale(float aod, float Ha_is, float Ha);
float optical_depth(float aod, float mod);
float scatt_transmitt(float aod, float mod, float od, float ms, float mv, float *Ts, float *Tv, float *tsd, float *tss, float *tvd, float *tvs);
float path_ref(float aod, float mod, float od, float Pa, float Pr, float tsd, float tvd, float ms, float mv);
float sphere_albedo(float aod, float mod, float od);
float env_weight_aerosol(float r);
float env_weight_molecular(float r);
float env_weight(float aod, float mod, float Fa, float Fr);
float backscatter(float ms, float mv, float sazi, float vazi);
float phase_molecular(float cospsi);
float phase_aerosol(float cospsi);
float wvp_transmitt(int b, float w, float m);
float ozone_transmitt(int b, float o, float m);
float gas_transmitt(float Tsw, float Tvw, float Tso, float Tvo);
float fresnel_reflection(float i);
float illumin(float csz, float ssz, float ctz, float stz, float sa, float ta);
float c_factor_emp(double b, double m);
float c_factor_com(float h0, float f, float ms);
float f_factor(float tss, float tsd);


/** functions in splits.cpp
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int derive_lsp(int p, int year);


/** functions in stats.c
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void covar_recurrence(double   x, double   y, double *mx, double *my, double *vx, double *vy, double *cv, double n);
void cov_recurrence(double   x, double   y, double *mx, double *my, double *cv, double n);
void kurt_recurrence(double   x,    double *mx, double *vx,    double *sx,double *kx, double n);
void skew_recurrence(double   x,    double *mx, double *vx,    double *sx,double n);
void var_recurrence(double   x, double *mx, double *vx, double n);
double kurtosis(double var, double kurt, double n);
double skewness(double var, double skew, double n);
double variance(double var, double n);
double standdev(double var, double n);
double covariance(double cov, double n);
void linreg_slope(double cov, double varx, double *slope);
void linreg_intercept(double slope, double mx, double my, double *intercept);
void linreg_coefs(double mx, double my, double cov, double varx, double *slope, double *intercept);
void linreg_r(double cov, double varx, double vary, double *r);
void linreg_rsquared(double cov, double varx, double vary, double *rsq);
void linreg_predict(double x, double slope, double intercept, double *y);
int slope_significant(float p, int tailtype, int n, float b, float b0, float seb);
int tscore(float p, int df, int tailtype, float *tscore);
float tscore_tail2left(float p, int tail, bool negative);
float tscore_left2twotail(float p, bool negative);
float tscore_tail2twotail(float p, int tail, bool negative);
float tscore_Norm_p(float z);
float tscore_Norm_z(float p);
float tscore_Hills_inv_t(float p, int idf);
float tscore_T_z(float t, int df);
float tscore_T_p(float t, int df);
void quantile_swap(float *x, int from, int to);
float quantile(float *x, int n, float p);
int pca(float **INP, float nodata, int nb, int nc, float minvar, float ***OUT, int *newnb);


/** functions in sun.c
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
float doy2dsun(int doy);
float sunJC(int year, int month, int day, float timenow);
float sunGeomMeanLong(float t);
float sunGeomMeanAnomaly(float t);
float sunEccentricityEarthOrbit(float t);
float sunEqOfCenter(float t);
float sunTrueLong(float t);
float sunTrueAnomaly(float t);
float sunRadVector(float t);
float sunApparentLong(float t);
float sunMeanObliquityOfEcliptic(float t);
float sunObliquityCorrection(float t);
float sunRtAscension(float t);
float sunDeclination(float t);
float sunEquationOfTime(float t);
float sunHourAngle(float time, float longitude, float eqtime);
void sunpos(float latitude, float longitude, int year, int month, int day, int hh, int mm, int ss, int zone, float *zen, float *azi);


/** functions in tile.c
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int tilefileread(int **X, int **Y, int *k);
int tile_whitelisted(int *white_x, int *white_y, int white_k, int x, int y);
int tile_find(float ulx, float uly, float *tilex, float *tiley, int *idx, int *idy);
int tile_count(int nx, int ny, float ulx, float uly, float t_ulx, float t_uly, int *ntx, int *nty);
int tile_image(int2s **DAT, int nx, int ny, float ulx, float uly);
int flush_image(int2s **DAT, int nx, int ny, float ulx, float uly);
int write_datacube_def(char *d_write);
int read_datacube_def(char *d_read);
int copy_datacube_def(char *d_read, char *d_write);
int geometric_module(int2s **DAT0);


/** functions in timeseries.c
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int despike_timeseries(bool *MASK, int thr);
int interpolate_timeseries(bool *MASK);
int phenology_timeseries(bool *MASK);
int basic_timeseries_stats(bool *MASK);
int standardize_timeseries(int2s **TS, bool *MASK, int n);
int trend_timeseries(int2s **TS, float *timestamp, bool *MASK, int n, bool days);
int cat_timeseries(int2s **TS, float *timestamp, bool *MASK, int n, bool days);
int fold_timeseries(bool *MASK, int by);
int clear_sky_stats(level2_t *L2);


/** functions in topo.c
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int compile_topo(top_t *TOP, qai_t QAI, int2u *GRD);
int ocean_topo(float *DEM);
int smooth_topo(float **DEM);
int slope_and_aspect(qai_t QAI, float *DEM, int2u **SLP, int2u **ASP, float nodata);
int elevation_stats(float *DEM, qai_t QAI, int1u **CDEM);
int average_elevation_cg(top_t TOP, qai_t QAI, int gres, int j, int i);
int illumination(top_t *TOP, qai_t QAI, int2u *GRD);
int topographic_factor(int2s **TOA, qai_t QAI, int2u *GRD, top_t *TOP);


/** functions in ts-interpolate.c
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int interpolate_linear(bool *MASK);
int interpolate_moving(bool *MASK);
int interpolate_rbf(bool *MASK);


/** functions in utils.c
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void print_ivector(int    *v, char *name, int n, int big);
void print_fvector(float  *v, char *name, int n, int big, int small);
void print_dvector(double *v, char *name, int n, int big, int small);
float rad2deg(float x);
float deg2rad(float x);
double proctime(time_t start);
void proctime_print(char *string, time_t start);
int tocg(int x);


/** functions in warp.cpp
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int geo_to_map(float srs_x, float srs_y, float *dst_x, float *dst_y, char *dst_wkt);
int reproj_point(float  srs_x, float  srs_y, float *dst_x, float *dst_y, char *src_wkt, char *dst_wkt);
int reproj_topo(int rsm, float **DEM);
int reproj_image(char *fname, int rsm, int b, int nx, int ny, float nodata, double dst_geotran[6], float **IMG);
int warp_image(int2s **SRC, int2s ***DST, float src_ulx, float src_uly, char *src_wkt, float *dst_ulx, float *dst_uly, int *dst_nx, int *dst_ny);
int warp_qai(int2s **SRC, int2s **DST, float src_ulx, float src_uly, char *src_wkt, float dst_ulx, float dst_uly, int dst_nx, int dst_ny);


/** functions in wvp.c
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int wvp_transmitt_lut();
double fun_wvp(const gsl_vector *v, void *params);
int water_vapor(int2s **TOA, qai_t QAI, int2u *GRD, top_t TOP, int2s **WVP, int2s ***Tg);
float ozone_amount(float lon, float lat, int doy);
int water_vapor_from_lut();
int failure_terra(date_t d);
int failure_aqua(date_t d);
int read_modis_geometa(char *fname, char *sen, char ***fid, float ****gr, bool **v, int *nl, int *nv);
int modis_intersect(float lat, float lon, float ***gr, bool *v, int nl, int nv, int **ptr);
int modis_inside(float lat, float lon, float ***gr, int nintersect, int *ptr);
void compile_modis_wvp(char *dir_geo, char *dir_hdf, date_t d_now, char *sen, int nc, float **COO, double **avg, double **count);
void choose_modis_wvp(bool aqua, int nc, float *WVP, char **SEN, double *modavg, double *mydavg, double *modctr, double *mydctr);
void write_wvp_lut(char *fname, int nc, float **COO, float *WVP, char **SEN);
void write_avg_table(char *dir_wvp, int nc, float **COO, double ***AVG);
void read_wvp_lut(char *fname, int nc, float **COO, float *WVP);
void create_wvp_lut(char *dir_geo, char *dir_hdf, char *tablename, date_t d_now, int nc, float **COO, float *WVP, char **SEN);


#ifdef __cplusplus
}
#endif

#endif

