/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
This file contains functions for binary-based image processing methods
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#include "force.h"


/** Number of TRUE cells
+++ This function counts all TRUE cells of a bool array.
--- B:      bool array
--- n:      number of cells
+++ Return: Number of TRUE cells
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int bsum(bool *B, int n){
int y=0, i;

  for (i=0; i<n; i++) y += B[i];
  return y;
}


/** Find width of Gaussian for Gaussian convolution
+++ This function estimates the width of a Gaussian used to perform Gaus-
+++ sian lowpass filtering.
--- r:      ratio of target and actual resolution
+++ Return: width of Gaussian (sigma)
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
float find_sigma(float r){
float r2, amp, sig, sigma, k, diff;

  r2 = (r/2.0)*(r/2.0);

  sigma = 0.0; diff = INT_MAX;

  while (diff > 0){
    
    sigma += 0.01;

    sig = sigma*sigma*2;
    amp = 1.0/(M_PI*sig);
    k = amp*exp(-r2/sig);

    diff = amp/2.0 - k;

  }

  sigma -= 0.01;

  return sigma;
}


/** Gaussian kernel
+++ This function generates a Gaussian kernel to perform Gaussian lowpass
+++ filtering
--- nk:     kernel width
--- sigma:  width of Gaussian
--- kernel: kernel with weighting coefficients (returned)
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int gauss_kernel(int nk, float sigma, float ***kernel){
int i, j;
float hx, hy, sig;
float **k = NULL;

  if ((nk % 2) != 1){
    printf("gaussian kernel width must be odd. "); return FAILURE;}

  #ifdef DEBUG
  printf("gaussian kernel width %d, sigma: %.2f\n", nk, sigma);
  #endif

  sig = sigma*sigma*2;
  alloc_2D((void***)&k, nk, nk, sizeof(float));

  for (i=0; i<nk; i++){
  for (j=0; j<nk; j++){

    hx = -(nk-1)/2 + j;
    hy = -(nk-1)/2 + i;
    k[i][j] = 1/(M_PI*sig) * exp(-(hx*hx+hy*hy)/(sig));

  }
  }

  *kernel = k;
  return SUCCESS;
}


/** Pre-computation of the pixel-distance within a kernel
+++ This function generates a distance kernel for kernel-based methods
--- nk:     kernel width
--- kernel: kernel with distance to central pixel (returned)
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int distance_kernel(int nk, float ***kernel){
int ki, kj, ii, jj, h;
float **k = NULL;

  if ((nk % 2) != 1){
    printf("distance kernel width must be odd. "); return FAILURE;}

  alloc_2D((void***)&k, nk, nk, sizeof(float));

  h = (nk-1)/2;

  // pre-compute kernel distance
  for (ii=-h, ki=0; ii<=h; ii++, ki++){
  for (jj=-h, kj=0; jj<=h; jj++, kj++){
    k[ki][kj] = sqrt(ii*ii+jj*jj);
  }
  }

  *kernel = k;
  return SUCCESS;
}


/** Distance transformation
+++ This function computes the distance transformation, i.e. the pixel 
+++ distance of any FALSE cell to its next TRUE cell.
+++-----------------------------------------------------------------------
+++ Meijster, A., Roerdink, J.B.T.M., Hesselink, W.H. (2006). A general 
+++ algorithm for computing distance transforms in linear time. Computa-
+++ tional Imaging and Vision, 18, 331-340.
+++-----------------------------------------------------------------------
--- B:      Binary image
--- DT:     Distance transformed image
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void dist_transform(bool *B, int2u *DT){
int x, y, q, w, u, *s, *t;
int2u *G;

  alloc((void**)&G, _NC_, sizeof(int2u));

  /** first phase **/
  for (x=0; x<_NX_; x++){

    // scan 1
    if (B[x]) G[x] = 0; else G[x] = _NX_+_NY_;
    
    for (y=1; y<_NY_; y++){
      if (B[y*_NX_+x]){
        G[y*_NX_+x] = 0;
      } else {
        G[y*_NX_+x] = 1 + G[(y-1)*_NX_+x];
      }
    }

    // scan 2    
    for (y=_NY_-2; y>=0; y--){
      if (G[(y+1)*_NX_+x] < G[y*_NX_+x]){
        G[y*_NX_+x] = 1 + G[(y+1)*_NX_+x];
      }
    }

  }

  // allocate s and t
  alloc((void**)&s, _NX_, sizeof(int));
  alloc((void**)&t, _NX_, sizeof(int));

  /** second phase **/
  for (y=0; y<_NY_; y++){
  
    q = 0; s[0] = 0; t[0] = 0;
  
    // scan 3
    for (u=1; u<_NX_; u++){
    
      while (q >= 0 && dt_dfun(t[q], s[q], y, G) > dt_dfun(t[q], u, y, G)) q--;

      if (q < 0){
        q = 0;
        s[0] = u;
      } else {
        w = 1 + dt_Sep(s[q], u, y, G);
        if (w < _NX_){
          q++;
          s[q] = u;
          t[q] = w;
        }
      }
    
    }
  
    // scan 4
    for (u=_NX_-1; u>=0; u--){
      DT[y*_NX_+u] = (int2u) sqrt(dt_dfun(u, s[q], y, G));
      if (u == t[q]) q--;
    }
  
  }

  // free memory
  free((void*)G);
  free((void*)s);
  free((void*)t);

  return;
}


/** Function for euclidian distance
+++ This function is used in the distance transformation (dist_transform)
+++--------------------------------------------------------------------**/
int dt_dfun(int x, int i, int y, int2u *G){
int d;

  d = (x-i)*(x-i) + G[_NX_*y+i]*G[_NX_*y+i];

  return d;
}


/** Sep function for euclidian distance
+++ This function is used in the distance transformation (dist_transform)
+++--------------------------------------------------------------------**/
int dt_Sep(int i, int u, int y, int2u *G){
int sep;

  // note that integer division is used on purpose
  sep = (u*u - i*i + G[_NX_*y+u]*G[_NX_*y+u] - 
                     G[_NX_*y+i]*G[_NX_*y+i]) / (2*(u-i));

  return sep;
}


/** Connected components labelling
+++ This function performs connected components labeling, i.e. it segments
+++ all 8-connected TRUE patches. A unique ID is given to each segment, 
+++ starting at top-left.
+++-----------------------------------------------------------------------
+++ Chang, F., Chen, C.-J., Lu, C.-J. (2004). A linear-time component-la-
+++ beling algorithm using contour tracing technique. Computer Vision and
+++ Image Understanding, 93 (2), 206-220.
+++-----------------------------------------------------------------------
--- B:      Binary image
--- CCL:    Connected components
+++ Return: number of segments
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int connectedcomponents(bool *B, int *CCL){
int i, j, p;
int dir, k = 0, label = 0;

  for (i=0, p=0; i<_NY_; i++){
  for (j=0, label=0; j<_NX_; j++, p++){


    if (B[p]){

      if (label != 0){ // use last label

        CCL[p] = label;

      } else {

        label = CCL[p];

        if (label == 0){

          label = ++k;
          dir = 0;
          ccl_contourtrace(i, j, label, dir, B, CCL); // external contour
          CCL[p] = label;

        }

      }

    } else if (label != 0){ // if last pixel was labeled

      if (CCL[p] == 0){
        dir = 1;
        ccl_contourtrace(i, j-1, label, dir, B, CCL); // internal contour
      }

      label = 0;

    }

  }
  }


  // replace -1 with 0
  for (p=0; p<_NC_; p++){
    if (CCL[p] < 0) CCL[p] = 0;
  }

  return(k);
}


/** Tracer Function for connected components labeling
+++ This function is used in connectedcomponents
+++--------------------------------------------------------------------**/
void ccl_tracer(int *cy, int *cx, int *dir, bool *B, int *CCL){
int i, y, x, tval;
static int neighbor[8][2] = {{0,1},{1,1},{1,0},{1,-1},{0,-1},{-1,-1},{-1,0},{-1,1}};


  for (i=0; i<7; i++){

    y = *cy + neighbor[*dir][0];
    x = *cx + neighbor[*dir][1];

    if (y>=0 && y<_NY_ && x>=0 && x<_NX_){

      tval = B[y*_NX_+x];
      
    } else {

      tval = 0;

    }

    if (tval == 0){

      if (y>=0 && y<_NY_ && x>=0 && x<_NX_) CCL[y*_NX_+x] = -1;
      *dir = (*dir + 1) % 8;

    } else {

      *cy = y;
      *cx = x;
      break;

    }

  }

  return;
}


/** Contour tracing Function for connected components labeling
+++ This function is used in connectedcomponents
+++--------------------------------------------------------------------**/
void ccl_contourtrace(int cy, int cx, int label, int dir, 
                    bool *B, int *CCL){
bool stop = false, search = true;
int fx, fy, sx = cx, sy = cy;

  ccl_tracer(&cy, &cx, &dir, B, CCL);

  if (cx != sx || cy != sy){

    fx = cx;
    fy = cy;

    while (search){

      dir = (dir + 6) % 8;
      CCL[cy*_NX_+cx] = label;
      ccl_tracer(&cy, &cx, &dir, B, CCL);

      if (cx == sx && cy == sy){

        stop = true;

      } else if (stop){

        if (cx == fx && cy == fy){
          search = false;
        } else {
          stop = false;
        }

      }

    }

  }

  return;
}


/** Focal standard deviation filter for multi-/single-band images
+++ This function computes the spatial standard deviation as a measure for
+++ texture. If a multiband image is given, the pixel-based standard devi-
+++ ation of the band with highest standard deviation is returned.
--- DAT:    single-/multi-band image
--- nodata: nodata value of image
--- h:      radius of kernel
--- nx:     number of columns
--- ny:     number of rows
--- nb:     number of bands to be used
--- bstart: first band to be used
--- SD:     standard deviation (allocated within, returned)
+++ Return: SUCCESS/Failure
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int focal_sd(float **DAT, float nodata, int h, int nx, int ny, int nb, int bstart, float **SD){
int i, j, ii, jj, ni, nj, b, nc = nx*ny;
double sd_0, sd_max;
double avg, var, k, dat;
float *sd;


  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  alloc((void**)&sd, nc, sizeof(float));

  #pragma omp parallel private(b, sd_0, sd_max, avg, var, k, dat, ii, jj, ni, nj) shared(bstart,h,nb,nx,ny,sd,DAT,nodata) default(none)
  {

    #pragma omp for collapse(2) schedule(dynamic,1)
    for (i=0; i<ny; i++){
    for (j=0; j<nx; j++){

      for (b=bstart, sd_max=0; b<(bstart+nb); b++){

        avg = var = k = 0.0;

        for (ii=-h; ii<=h; ii++){
        for (jj=-h; jj<=h; jj++){

          ni = i+ii; nj = j+jj;

          if (ni < 0 || nj < 0 || ni >= ny || nj >= nx){
            continue;}

          if ((dat = (double)DAT[b][ni*nx+nj]) == nodata){
            continue;}

          if (++k == 1){
            avg = dat;
          } else {
            var_recurrence(dat, &avg, &var, k);
          }
        }
        }

        if ((sd_0 = standdev(var, k))  > sd_max) sd_max = sd_0;

      }

      sd[nx*i+j] = sd_max;

    }
    }

  }


  #ifdef CLOCK
  proctime_print("computing focal sd", TIME);
  #endif

  *SD = sd;
  return SUCCESS;
}


/** Buffer
+++ This function buffers all TRUE pixels by r pixels using the midpoint 
+++ circle algorithm.
--- B:      Binary image (will be altered)
--- r:      radius
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void buffer(bool *B, int r){
bool *tmp;
int i, j, k, p, d, x, y;

  alloc((void**)&tmp, _NC_, sizeof(bool));

  // do for every pixel, except boundary
  // boundary width is determined by buffer radius
  for (i=r; i<(_NY_-r); i++){
  for (j=r; j<(_NX_-r); j++){

    p = _NX_*i+j;

    if (!B[p]) continue;

    // midpoint circle algorithm
    d = 3 - r;
    x = 0;
    y = r;

    while (x <= y){

      // draw buffer with radius around point
      for (k=i-y; k<=i+y; k++){ tmp[_NX_*k+j+x] = true; tmp[_NX_*k+j-x] = true; }
      for (k=i-x; k<=i+x; k++){ tmp[_NX_*k+j+y] = true; tmp[_NX_*k+j-y] = true; }

      if (d < 0){
        d = d + (4 * x) + 6;
      } else {
        d = d + 4 * (x - y) + 10;
        y -= 1;
      }
      x += 1;

    }
  }
  }

  // copy buffer to input
  for (p=0; p<_NC_; p++){
    if (tmp[p]) B[p] = true;
  }

  free((void*)tmp);

  return;
}


/** Majority filling
+++ This function performs a majority filling, i.e. FALSE cells that are 
+++ surrounded by 5 or more TRUE cells will be set to TRUE. This fills
+++ small holes in a binary image.
--- B:      Binary image (will be altered)
--- QAI:    Quality Assurance Information
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void majorfill(bool *B, qai_t QAI){

int i, j, ii, jj, ni, nj, p, k;

  for (i=0, p=0; i<_NY_; i++){
  for (j=0; j<_NX_; j++, p++){

    if (QAI.off[p] || B[p]) continue;

    k = 0;

    for (ii=-1; ii<=1; ii++){
    for (jj=-1; jj<=1; jj++){

      if (ii == 0 && jj == 0) continue;
      ni = i+ii; nj = j+jj;
      if (ni < 0 || ni >= _NY_ || nj < 0 || nj >= _NX_) continue;

      k += B[ni*_NX_+nj];

    }
    }

    if (k >= 5) B[p] = true;

  }
  }

  return;
}

