/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
This program improves coarser spatial resolution  Level-2 data using
higher spatial resolution data (ImproPhe methodology)
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#include "force.h"


int main ( int argc, char *argv[] ){
level2_t *L2 = NULL; // Level 2 data
int i, j, p, b, d, y, w, f;  // iterators
int tx, ty, tc = 0, tk = 0, tyy, txx; // iterators for tiles
int *tiles_x = NULL, *tiles_y = NULL, tiles_k; // tile white-list
char dname[MAX_STR_LEN]; // directory name of current tile
double geotran_mosaic[6]   = { 0, 0, 0, 0, 0, 0 }; // geo transformation of all tiles
double geotran_tile[6]     = { 0, 0, 0, 0, 0, 0 }; // geo transformation of central tile
float **KDIST = NULL; // kernel distance
int nk, mink, h; // number of kernel pixels, and minimum number of pixels for good prediction, radius
int NX3, NY3, NC3, NPC; // number of pixels of 3x3 mosaic, number of bands after PCA
int XOFF3, YOFF3; // pixel offset in 3x3 mosaic for central tile
imp_t IMP; // ImproPhe struct
qai_t QAI;
char fname[MAX_STR_LEN];
int2s **OUT = NULL; 
int doymin, doymax, yearmin, yearmax, cemin, cemax;
int ce;
int tcen = 4;
int n[9];
float fnx_, fny_;
int doy, year;
bool ok;
int np;
int t, tt;
time_t TIME, TIME_I, TIME_C, TIME_O; // time measurements
double time_i = 0, time_c = 0, time_o = 0, ctime = 0; // cumulated time
double eta_h, eta_m, eta_s; // ETA estimate hour, minute, seconds


  if (argc != 2){ printf("usage: %s parameter-file\n", argv[0]); return FAILURE;}


/** INITIALIZING
+** read parameter file, get raster properties, prepare output files etc.
+** *********************************************************************/

  time(&TIME);
  printf("Initializing startup"); fflush(stdout);

  par.f_par = argv[1];

  // parse parameter file
  if (parse_param_il2() != SUCCESS){
    printf("Reading parameter file failed!\n"); return FAILURE;}

  // get tile file and read it
  if (tilefileread(&tiles_x, &tiles_y, &tiles_k) != SUCCESS){
    printf("Reading tile file failed!\n"); return FAILURE;}
    
  // how many tiles to do?
  for (ty=par.tminy; ty<=par.tmaxy; ty++){
  for (tx=par.tminx; tx<=par.tmaxx; tx++){

    // if tile is not whitelisted (if specified), skip
    if (tile_whitelisted(tiles_x, tiles_y, tiles_k, tx, ty) == FAILURE) continue;
    tk++;

  }
  }

  // find UL coordinate of mosaic + resolution
  if (mosaic_info(geotran_mosaic) != SUCCESS){
    printf("Obtaining mosaic info failed.\n"); return FAILURE;}
  geotran_tile[1] = geotran_mosaic[1];
  geotran_tile[5] = geotran_mosaic[5];

  // number of pixels in 3x3 mosaic + offset of central tile
  NX3 = _NX_*3; NY3 = _NY_*3; NC3 = NX3*NY3;
  XOFF3 = _NX_; YOFF3 = _NY_;
  #ifdef DEBUG
  printf("number of pixels in 3x3 mosaic: %d x %d = %d\n", NX3, NY3, NC3);
  #endif

  mem_to_disc_dict_init();


  // pre-compute kernel distances
  // determine minimum # of neighbors that make a good prediction
  distance_kernel(par.ksize, &KDIST);
  nk = par.ksize*par.ksize;
  h = (par.ksize-1)/2;
  mink = (int)((nk*M_PI/4.0)*0.5/100.0);
  #ifdef DEBUG
  printf("ksize/nkernel/minnum: %d/%d/%d\n", par.ksize, nk, mink);
  #endif

  printf("\rImproPhe L2 processing started, I/C/O TBD, ETA: TBD");
  fflush(stdout);

  // open threads
  omp_set_num_threads(par.cpu);


/** START PROCESSING; LOOP OVER ALL AVAILABLE TILES AND YEARS
+** *********************************************************************/

  for (ty=par.tminy; ty<=par.tmaxy; ty++){
  for (tx=par.tminx; tx<=par.tmaxx; tx++){

    time(&TIME_I);

    // if tile is not whitelisted (if specified), skip
    if (tile_whitelisted(tiles_x, tiles_y, tiles_k, tx, ty) == FAILURE) continue;

    // set geotransformation
    geotran_tile[0] = geotran_mosaic[0]+_NX_*geotran_tile[1]*(tx-par.tminx);
    geotran_tile[3] = geotran_mosaic[3]+_NY_*geotran_tile[5]*(ty-par.tminy);


    /** prepare higher resolution data **/

    // only read higher resolution data, update sensor dictionary
    // note that _NB_ is updated
    if (par.l2dict.n > 0){
      free_2D((void**)par.l2dict.sensor, par.l2dict.n);
      free_2D((void**)par.l2dict.band,   par.l2dict.n);
      par.l2dict.n = 0;
    }
    if (level2sensor_dict(par.csen_mr, &par.l2dict) != SUCCESS) return FAILURE;

    alloc_2D((void***)&IMP.AVG, par.nwin*_NB_, NC3, sizeof(float));
    for (p=0; p<NC3; p++){
      for (b=0; b<par.nwin*_NB_; b++) IMP.AVG[b][p] = _NODATA_;
    }

    for (tyy=-1; tyy<=1; tyy++){
    for (txx=-1; txx<=1; txx++){

      // current tile
      sprintf(dname, "%s/X%04d_Y%04d", par.d_level2, tx+txx, ty+tyy);

      // get Level2 file listing, skip if tile is empty
      if (list_level2(dname) == FAILURE) continue;

      // read Level 2
      if (tyy == 0) fny_ = _NY_*par.res; else fny_ = tyy*h*par.res;
      if (txx == 0) fnx_ = _NX_*par.res; else fnx_ = txx*h*par.res;

      if (read_level2(dname, &L2, fnx_, fny_) != SUCCESS){
        printf("error reading L2.\n"); return FAILURE;}
      buffer_level2(L2);

      // compute average per seasonal window
      average_season(L2, dir.n, IMP.AVG, (txx+1)*XOFF3, (tyy+1)*YOFF3, 0);

      // clean Level 2
      free_level2(L2);
      free_2D((void**)dir.list, dir.N);
      free_2D((void**)dir.LIST, dir.N);

    }
    }

    #ifdef DEBUG
    db_write_image_2D((void**)IMP.AVG, "IMPROPHE_INPUT", par.nwin*_NB_, NX3, NY3, sizeof(float), 4);
    #endif
    
    time_i += proctime(TIME_I);
    time(&TIME_C);

    // reduce MR dimensionality using principal component analysis
    if (pca(IMP.AVG, _NODATA_, par.nwin*_NB_, NC3, 0.975, &IMP.MR, &NPC) != SUCCESS){
      printf("error in PCA.\n"); return FAILURE;}
    free_2D((void**)IMP.AVG, par.nwin*_NB_);

    // normalize PCs
    for (b=0; b<NPC; b++){
      if (standardize_float(IMP.MR[b], _NODATA_, NC3) != SUCCESS){
        printf("error in standardizing PCA.\n"); return FAILURE;}
    }

    #ifdef DEBUG
    db_write_image_2D((void**)IMP.MR, "IMPROPHE_MR", NPC, NX3, NY3, sizeof(float), 4);
    #endif

    // MR texture
    if (focal_sd(IMP.MR, _NODATA_, (par.ksd-1)/2, NX3, NY3, NPC, 0, &IMP.MX) != SUCCESS){
      printf("error in computing med-res texture.\n"); return FAILURE;}

    #ifdef DEBUG
    db_write_image_1D((void**)IMP.MX, "IMPROPHE_MX", NX3, NY3, sizeof(float), 4);
    #endif

    time_c += proctime(TIME_C);
    time(&TIME_I);


    /** prepare lower resolution data / predict **/

    // only read lower resolution data, update sensor dictionary
    // note that _NB_ is updated
    if (par.l2dict.n > 0){
      free_2D((void**)par.l2dict.sensor, par.l2dict.n);
      free_2D((void**)par.l2dict.band,   par.l2dict.n);
      par.l2dict.n = 0;
    }
    if (level2sensor_dict(par.csen_cr, &par.l2dict) != SUCCESS) return FAILURE;

    mem_to_disc_dict_il2();

    alloc_2D((void***)&IMP.CR,   _NB_, NC3,   sizeof(float));
    alloc_2D((void***)&IMP.PRED, _NB_, _NC_, sizeof(float));
    alloc_2D((void***)&par.m2d[_B_.ref].data, _NB_, _NC_, sizeof(int2s));

    time_i += proctime(TIME_I);

    /** attempt prediction for each day in target year **/
    for (ce=par.ymin*365+1; ce<par.ymax*365+1; ce++){

      time(&TIME_I);
      //printf("predicting ce: %d\n", ce);

      for (p=0; p<NC3; p++){
        for (b=0; b<_NB_; b++) IMP.CR[b][p] = _NODATA_;
      }
      
      ok = false;

      for (tyy=-1; tyy<=1; tyy++){
      for (txx=-1; txx<=1; txx++){

        // current tile
        sprintf(dname, "%s/X%04d_Y%04d", par.d_level2, tx+txx, ty+tyy);

        // get Level2 file listing, skip if tile is empty
        if (list_level2(dname) == FAILURE) continue;
        list_level2_filter_ce(ce, ce);
        if (tyy == 0 && txx == 0 && dir.n > 0) ok = true;
        if (dir.n < 1) continue;

        // read Level 2
        if (tyy == 0) fny_ = _NY_*par.res; else fny_ = tyy*h*par.res;
        if (txx == 0) fnx_ = _NX_*par.res; else fnx_ = txx*h*par.res;

        if (read_level2(dname, &L2, fnx_, fny_) != SUCCESS){
          printf("error reading L2.\n"); return FAILURE;}

        // parse Quality Assurance Information
        alloc_qai(&QAI);
        parse_qai(QAI, L2[0].qai);

        // copy to 3x3 mosaick
        for (i=0; i<_NY_; i++){
        for (j=0; j<_NX_; j++){

          // pixel in current tile
          p = i*_NX_+j;

          // pixel in 3x3 mosaic
          np = (i+(tyy+1)*YOFF3)*NX3+(j+(txx+1)*XOFF3);

          // if nonusable, copy original value
          // to output, and set as nodata during prediction (prevents cloud smudging..)
          if (!use_this_pixel(QAI, p)){
            for (b=0; b<_NB_; b++){
              if (tyy == 0 && txx == 0) par.m2d[_B_.ref].data[b][p] = L2[0].ref[b][p];
              IMP.CR[b][np] = _NODATA_;
            }
          } else {
            for (b=0; b<_NB_; b++){
              IMP.CR[b][np] = L2[0].ref[b][p]/_RS_;
            }
          }

        }
        }

        // basename of output
        sprintf(par.b_out, "%04d%02d%02d_LEVEL2_%s_IMP", 
          L2[0].date.year, L2[0].date.month, L2[0].date.day, par.l2dict.sensor[L2[0].sid]);

        // clean Level 2
        free_level2(L2);
        free_qai(QAI);
        free_2D((void**)dir.list, dir.N);
        free_2D((void**)dir.LIST, dir.N);

      }
      }

      #ifdef DEBUG
      db_write_image_2D((void**)IMP.CR, "IMPROPHE_CR", _NB_, NX3, NY3, sizeof(float), 4);
      #endif

      // skip this date if file already exists, if overwrite is disabled
      if (!par.owr){
        sprintf(fname, "%s/X%04d_Y%04d/%s.dat", par.d_level2, tx, ty, par.b_out);
        if (fileexist(fname)) ok = false;
        sprintf(fname, "%s/X%04d_Y%04d/%s.tif", par.d_level2, tx, ty, par.b_out);
        if (fileexist(fname)) ok = false;
      }
      

      if (!ok) continue;

      time_i += proctime(TIME_I);
      time(&TIME_C);

      // CR texture

      if ((IMP.CX = (float**)malloc(_NB_*sizeof(float*))) == NULL){
      printf("unable to allocate memory! "); return FAILURE;}
      
      for (b=0; b<_NB_; b++){
        if (focal_sd(IMP.CR, _NODATA_, (par.ksd-1)/2, NX3, NY3, 1, b, &IMP.CX[b]) != SUCCESS){
          printf("error in computing med-res texture.\n"); return FAILURE;}
      }

      #ifdef DEBUG
      db_write_image_2D((void**)IMP.CX, "IMPROPHE_CX", _NB_, NX3, NY3, sizeof(float), 4);
      #endif


      /** predict **/

      // reset prediction
      for (p=0; p<_NC_; p++){
        for (b=0; b<_NB_; b++) IMP.PRED[b][p] = _NODATA_;
      }


      #pragma omp parallel shared(_NX_,_NY_,XOFF3,YOFF3,IMP,KDIST,NX3,NY3,NPC,_NB_,nk,mink,h) default(none)
      {

        #pragma omp for collapse(2) schedule(guided)
        for (i=YOFF3; i<(YOFF3+_NY_); i++){
        for (j=XOFF3; j<(XOFF3+_NX_); j++){

          improphe(IMP, KDIST, _NODATA_, i, j, (i-YOFF3)*_NX_+(j-XOFF3), 
            h, NX3, NY3, NPC, _NB_, nk, mink);

        }
        }

      }

      #ifdef DEBUG
      db_write_image_2D((void**)IMP.PRED, "IMPROPHE_PRED", _NB_, _NX_, _NY_, sizeof(float), 4);
      #endif

      time_c += proctime(TIME_C);
      time(&TIME_O);

      /** write output **/
      for (p=0; p<_NC_; p++){
        for (b=0; b<_NB_; b++){
          if (IMP.PRED[b][p] != _NODATA_) par.m2d[_B_.ref].data[b][p] = (int2s)(IMP.PRED[b][p]*_RS_);
        }
      }
      write_products_il2(tx, ty, par.m2d[_B_.ref], geotran_tile);


      /** clean **/
      free_2D((void**)IMP.CX, _NB_);

      time_o += proctime(TIME_O);

    }

    time(&TIME_O);
    
    /** clean **/
    free_2D((void**)IMP.MR, NPC);
    free_2D((void**)IMP.CR, _NB_);
    free_2D((void**)IMP.PRED, _NB_);
    free((void*)IMP.MX);
    mem_to_disc_free();

    time_o += proctime(TIME_O);


    /** PROGRESS
    +** *****************************************************************/
    tc++; ctime = time_i + time_c + time_o;
    eta_s = (tk-tc)*(ctime/tc);
    if (eta_s >= 3600){
      eta_h = floor(eta_s/3600); eta_s -= eta_h*3600;
    } else eta_h = 0;
    if (eta_s >= 60){
      eta_m = floor(eta_s/60); eta_s -= eta_m*60;
    } else eta_m = 0;

    printf("\rImproPhe L2 processing %6.2f%%, I/C/O %03.0f/%03.0f/%03.0f, ETA: %02.0fh %02.0fm %02.0fs", 
      (float)tc/tk*100.0, round(time_i/ctime*100), round(time_c/ctime*100), round(time_o/ctime*100), 
       eta_h, eta_m, eta_s);
    fflush(stdout);

  }
  }


  /** CLEAN
  +** *****************************************************************/
  free_2D((void**)KDIST, par.ksize);
  free((void*)tiles_x); free((void*)tiles_y);

  proctime_print("Total runtime", TIME);

  return SUCCESS;
}

