/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
This file contains functions for Improving the spatial resolution of sea-
son-specific coarse continuous fields like Land Surface Phenology
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#include "force.h"


/** Predict medium resolution data (ImproPhe method)
+++ This function is the core method of ImproPhe and predicts the MR data
+++ on the basis of MR reflectance, MR texture, CR texture and the CR da-
+++ ta. Neighboring pixels of the same class are used for the prediction.
--- IMP:    ImproPhe struct
--- KDIST:  kernel distance
--- nodata: nodata of CR
--- i:      Y-coordinate of current pixel (3x3 mosaic)
--- j:      X-coordinate of current pixel (3x3 mosaic)
--- pred_p: current pixel of central tile
--- h:      prediction radius
--- nx:     number of X-pixels (3x3 mosaic)
--- ny:     number of Y-pixels (3x3 mosaic)
--- nb:     number of bands in MR data
--- nk:     number of kernel pixels, and 
--- mink:   minimum number of pixels for good prediction
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int improphe(imp_t IMP, float **KDIST, float nodata, int i, int j, int pred_p, int h, int nx, int ny, int nb_m, int nb_c, int nk, int mink){
int ki, kj; // iterator for KDIST
int ii, jj; // iterator for kernel relative to center
int ni, nj, np, op; // pixel positions
int b, f, s; // iterator for bands, CR parameters, cutoff-threshold
int wn = 0, k; // iterator for neighbours

double D;     // pixel distance
double S, SS; // spectral similarity + rescaled S
double T, TT; // subpixel heterogeneity + rescaled T
double U, UU; // CR heterogeneity + rescaled U
double W;     // pixel weight

// S, T, U, C of all valid neighbours
double *Srecord, *Trecord, **Urecord, **Crecord;
// S, T, U range of all valid neighbours
double Srange[2][5], Trange[2][5], Urange[2][MAX_STR_LEN][5];

double weightxdata[MAX_STR_LEN], weight[MAX_STR_LEN];

double sum;          // temporary variables to calculate S
float Sthr[5] = { 0.025, 0.05, 0.1, 0.2, 0.4 }; // cutoff-threshold for S
int Sn[5] = { 0, 0, 0, 0, 0 };  // number of neighbours for each cutoff-threshold
int ns = 5, Sc, *Sclass;             // index for cutoff-threshold

bool skip;


  // central pixel
  op = i*nx+j;

  // if any CR parameter == nodata, skip
  for (f=0, skip=false; f<nb_c; f++){
    if (IMP.CR[f][op] == nodata){
      skip = true; break;}
  }
  if (skip) return SUCCESS;;


  // allocate
  alloc((void**)&Sclass, nk, sizeof(int));
  alloc((void**)&Srecord, nk, sizeof(double));
  alloc((void**)&Trecord, nk, sizeof(double));
  alloc_2D((void***)&Crecord, nb_c, nk, sizeof(double));
  alloc_2D((void***)&Urecord, nb_c, nk, sizeof(double));


  // initialize
  for (s=0; s<ns; s++){
    Srange[0][s] = Trange[0][s] = INT_MAX; 
    Srange[1][s] = Trange[1][s] = INT_MIN;
    for (f=0; f<nb_c; f++){ 
      Urange[0][f][s] = INT_MAX; 
      Urange[1][f][s] = INT_MIN;
    }
  }
  for (f=0; f<nb_c; f++) weightxdata[f] = weight[f] = 0.0;


  /** for each neighbour **/
  for (ii=-h, ki=0; ii<=h; ii++, ki++){
  for (jj=-h, kj=0; jj<=h; jj++, kj++){

    ni  = i+ii; nj = j+jj;
    if (ni < 0 || nj < 0 || ni >= ny || nj >= nx){
      continue;}
    np = nx*ni+nj;
    
    // if not in circular kernel, skip
    if ((D = KDIST[ki][kj]) > h) continue;
      
    // if MR == nodata, skip
    if (IMP.MR[0][np] == _NODATA_) continue;

    // if any CR parameter == nodata, skip
    for (f=0, skip=false; f<nb_c; f++){
      if (IMP.CR[f][np] == nodata){
        skip = true; break;}
    }
    if (skip) continue;

    // specral distance (MAE)
    for (b=0, sum=0; b<nb_m; b++) sum += fabs(IMP.MR[b][op]-IMP.MR[b][np]);
    S = sum/(float)nb_m;

    // get cutoff-threshold index for S
    for (s=0, Sc=-1; s<ns; s++){
      if (S < Sthr[s]){ Sc = s; break;}
    }
    if (Sc < 0) continue;
    Sclass[wn] = Sc;

    // keep S + determine range
    Srecord[wn] = S;
    if (S > 0){
      for (s=Sc; s<ns; s++){
        if (S > Srange[1][s]) Srange[1][s] = S;
        if (S < Srange[0][s]) Srange[0][s] = S;
      }
    }

    // keep CR data
    for (f=0; f<nb_c; f++) Crecord[f][wn] = IMP.CR[f][np];

    // keep MR texture + determine range
    T = IMP.MX[np];
    Trecord[wn] = T;
    if (T > 0){
      for (s=Sc; s<ns; s++){
        if (T > Trange[1][s]) Trange[1][s] = T;
        if (T < Trange[0][s]) Trange[0][s] = T;
      }
    }

    // keep CR texture + determine range
    for (f=0; f<nb_c; f++){
      U = IMP.CX[f][np];
      Urecord[f][wn] = U;
      if (U > 0){
        for (s=Sc; s<ns; s++){
          if (U > Urange[1][f][s]) Urange[1][f][s] = U;
          if (U < Urange[0][f][s]) Urange[0][f][s] = U;
        }
      }
    }

    wn++; // number of valid neighbours
    for (s=Sc; s<ns; s++) Sn[s]++; // number of valid neighbours in S-class

  }
  }

  // if no valid neighbour... damn.. use CR
  if (wn == 0){
    for (f=0; f<nb_c; f++) IMP.PRED[f][pred_p] = IMP.CR[f][op];
    free((void*)Sclass);
    free((void*)Srecord);
    free((void*)Trecord);
    free_2D((void**)Urecord, nb_c);
    free_2D((void**)Crecord, nb_c);
    return SUCCESS;
  }

  // determine the spectral similarity cutoff threshold
  for (s=0, Sc=-1; s<ns; s++){
    if (Sn[s] >= mink){ Sc = s; break;}
  }
  if (Sc < 0) Sc = ns-1;


  // compute pixel weight
  for (k=0; k<wn; k++){

    if (Sclass[k] > Sc) continue;

    SS = rescale_weight(Srecord[k], Srange[0][Sc], Srange[1][Sc]);
    TT = rescale_weight(Trecord[k], Trange[0][Sc], Trange[1][Sc]);

    for (f=0; f<nb_c; f++){

      UU = rescale_weight(
        Urecord[f][k], Urange[0][f][Sc], Urange[1][f][Sc]);

      W = 1/(SS*TT*UU); 

      weightxdata[f] += W*Crecord[f][k];
      weight[f] += W;

    }
  }

  // prediction -> weighted mean
  for (f=0; f<nb_c; f++) IMP.PRED[f][pred_p] = (float)(weightxdata[f]/weight[f]);


  // tidy up things
  free((void*)Sclass);
  free((void*)Srecord);
  free((void*)Trecord);
  free_2D((void**)Urecord, nb_c);
  free_2D((void**)Crecord, nb_c);

  return SUCCESS;
}


/** Predict Sentinel-2 10m bands (ImproPhe method)
+++ This function is the same as improphe, but deals with Sentinel-2 data
+++ directly. This avoids copying the 16bit TOA reflectance to FLOAT va-
+++ lues, which would significantly increase RAM requirements. Texture is
+++ also disabled as it does not make sense for a 2:1 ratio of spatial re-
+++ solution.
--- TOA:     TOA reflectance
--- QAI:     Quality Assurance Information
--- PRED:    Predicted TOA
--- KDIST:   kernel distance
--- i:       Y-coordinate of current pixel (3x3 mosaic)
--- j:       X-coordinate of current pixel (3x3 mosaic)
--- pred_p:  current pixel of central tile
--- h:       prediction radius
--- nx:      number of X-pixels (3x3 mosaic)
--- ny:      number of Y-pixels (3x3 mosaic)
--- nb_m:    number of bands in MR data
--- nb_c:    number of bands in CR data
--- bands_m: MR bands to use
--- bands_c: CR bands to predict
--- nk:      number of kernel pixels, and 
--- mink:    minimum number of pixels for good prediction
+++ Return:  SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int improphe_sentinel2(int2s **TOA, qai_t QAI, int2s **PRED, float **KDIST, int i, int j, int pred_p, int h, int nx, int ny, int nb_m, int nb_c, int *bands_m, int *bands_c, int nk, int mink){
int ki, kj; // iterator for KDIST
int ii, jj; // iterator for kernel relative to center
int ni, nj, np, op; // pixel positions
int b, f, s; // iterator for bands, CR parameters, cutoff-threshold
int wn = 0, k; // iterator for neighbours

double D;     // pixel distance
double S, SS; // spectral similarity + rescaled S
double W;     // pixel weight

// S, T, U, C of all valid neighbours
double *Srecord, **Crecord;
// S, T, U range of all valid neighbours
double Srange[2][5];

double weightxdata[MAX_STR_LEN], weight;

double sum;          // temporary variables to calculate S
float Sthr[5] = { 0.0025, 0.005, 0.01, 0.02, 0.025 }; // cutoff-threshold for S
int Sn[5] = { 0, 0, 0, 0, 0 };  // number of neighbours for each cutoff-threshold
int ns = 5, Sc, *Sclass;             // index for cutoff-threshold

bool skip;


  // central pixel
  op = i*nx+j;


  // allocate
  alloc((void**)&Sclass, nk, sizeof(int));
  alloc((void**)&Srecord, nk, sizeof(double));
  alloc_2D((void***)&Crecord, nb_c, nk, sizeof(double));


  // initialize
  for (s=0; s<ns; s++){
    Srange[0][s] = INT_MAX; 
    Srange[1][s] = INT_MIN;
  }
  for (f=0; f<nb_c; f++) weightxdata[f] = 0.0;
  weight = 0.0;


  /** for each neighbour **/
  for (ii=-h, ki=0; ii<=h; ii++, ki++){
  for (jj=-h, kj=0; jj<=h; jj++, kj++){

    ni  = i+ii; nj = j+jj;
    if (ni < 0 || nj < 0 || ni >= ny || nj >= nx){
      continue;}
    np = nx*ni+nj;
    
    // if not in circular kernel, skip
    if ((D = KDIST[ki][kj]) > h) continue;

    // if nodata, skip
    if (QAI.off[np] || QAI.cld[np] > 0 || QAI.shd[np]) continue;

    // specral distance (MAE)
    for (b=0, sum=0; b<nb_m; b++) sum += fabs(((float)(TOA[bands_m[b]][op]-TOA[bands_m[b]][np]))/_RS_);
    S = sum/(float)nb_m;

    // get cutoff-threshold index for S
    for (s=0, Sc=-1; s<ns; s++){
      if (S < Sthr[s]){ Sc = s; break;}
    }
    if (Sc < 0) continue;
    Sclass[wn] = Sc;

    // keep S + determine range
    Srecord[wn] = S;
    if (S > 0){
      for (s=Sc; s<ns; s++){
        if (S > Srange[1][s]) Srange[1][s] = S;
        if (S < Srange[0][s]) Srange[0][s] = S;
      }
    }

    // keep CR data
    for (f=0; f<nb_c; f++) Crecord[f][wn] = ((float)TOA[bands_c[f]][np])/_RS_;

    wn++; // number of valid neighbours
    for (s=Sc; s<ns; s++) Sn[s]++; // number of valid neighbours in S-class

  }
  }

  // if no valid neighbour... damn.. use CR
  if (wn == 0){
    for (f=0; f<nb_c; f++) PRED[f][pred_p] = TOA[bands_c[f]][op];
    free((void*)Sclass);
    free((void*)Srecord);
    free_2D((void**)Crecord, nb_c);
    return SUCCESS;
  }

  // determine the spectral similarity cutoff threshold
  for (s=0, Sc=-1; s<ns; s++){
    if (Sn[s] >= mink){ Sc = s; break;}
  }
  if (Sc < 0) Sc = ns-1;


  // compute pixel weight
  for (k=0; k<wn; k++){

    if (Sclass[k] > Sc) continue;

    SS = rescale_weight(Srecord[k], Srange[0][Sc], Srange[1][Sc]);

    W = 1/SS;
    for (f=0; f<nb_c; f++) weightxdata[f] += W*Crecord[f][k];
    weight += W;

  }

  // prediction -> weighted mean
  for (f=0; f<nb_c; f++) PRED[f][pred_p] = (int2s)((weightxdata[f]/weight)*_RS_);


  // tidy up things
  free((void*)Sclass);
  free((void*)Srecord);
  free_2D((void**)Crecord, nb_c);

  return SUCCESS;
}


/** Rescale neighbor pixel weights
+++ This function rescales the neighbor pixel weights (R --> R').
--- weight:    weight R
--- minweight: minimum weight within kernel
--- maxweight: maximum weight within kernel
+++ Return:    rescaled weight R'
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
double rescale_weight(double weight, double minweight, double maxweight){
double normweight;

  if (weight == 0){
    normweight = 1;
  } else if (minweight == maxweight){
    normweight = 1;
  } else {
    normweight = 1+exp(25 * 
      ((weight-minweight)/(maxweight-minweight))-7.5);
  }

  return normweight;
}


/** Compute seasonal averages
+++ This function computes seasonal averages of Level 2 data for the Im-
+++ proPhe code. If one seasonal window has no valid data, the overall 
+++ mean is used instead. The averages are copied into the 3x3 mosaic ima-
+++ ges used in ImproPhe.
--- L2:     Level-2 data
--- n:      Number of Level 2 dates
--- AVG:    Seasonal averages (modified within)
--- x_off:  X-Offset for copying from the current tile to the 3x3 mosaic
--- y_off:  Y-Offset for copying from the current tile to the 3x3 mosaic
--- type:   Aggregation type: mean (0), median (1)
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int average_season(level2_t *L2, int n, float **AVG, int x_off, int y_off, int type){
double **avg = NULL, *k = NULL;
float ***arr = NULL, tmp;
int t, i, j, p, np, b, w, wb;
bool **NODATA = NULL; 
qai_t QAI;
bool ok;


  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  #ifdef DEBUG
  printf("Compute seasonal averages.\n");
  #endif

  
  // compile nodata
  alloc_2D((void***)&NODATA, n, _NC_, sizeof(bool));

  for (t=0; t<n; t++){

    // skip if not in seasonal window
    // skip if sensor is not in use-for-prediction list
    if (L2[t].date.ce < par.cewin[0] || L2[t].date.ce > par.cewin[par.bwin-1] ||
        strcasestr(par.csen_mr, par.l2dict.sensor[L2[t].sid]) == NULL){

      for (p=0; p<_NC_; p++) NODATA[t][p] = true;
      continue;

    }

    // parse Quality Assurance Information
    alloc_qai(&QAI);
    parse_qai(QAI, L2[t].qai);

    // quality check
    for (p=0; p<_NC_; p++){
      if (!use_this_pixel(QAI, p)) NODATA[t][p] = true;
    }

    free_qai(QAI);
  }


  #pragma omp parallel private(avg,arr,k,t,p,np,w,b,wb,ok,tmp) shared(L2,AVG,NODATA,_NB_,_NX_,_NY_,_NC_,n,par,x_off,y_off,type) default(none)
  {

    // allocate memory
    alloc((void**)&k,   par.nwin+1, sizeof(double));

    switch (type){
      case 0:
        alloc_2D((void***)&avg, par.nwin+1, _NB_, sizeof(double));
        break;
      case 1:
        alloc_3D((void****)&arr, par.nwin+1, _NB_, n, sizeof(float));
        break;
    }


    #pragma omp for collapse(2) schedule(dynamic,1)
    for (i=0; i<_NY_; i++){
    for (j=0; j<_NX_; j++){

      // pixel in current tile
      p = i*_NX_+j;

      // initialize
      for (w=0; w<=par.nwin; w++){
        if (type == 0){
          for (b=0; b<_NB_; b++) avg[w][b] = 0.0;
        }
        k[w] = 0.0;
      }

      // compute stats
      for (t=0; t<n; t++){
        
        if (NODATA[t][p]) continue;

        // get seasonal window
        for (w=0, ok=false; w<par.nwin; w++){
          if (L2[t].date.ce >= par.cewin[w] && L2[t].date.ce < par.cewin[w+1]){
            ok = true; break;}
        }

        // not in any window, skip
        if (!ok) continue;

        // sum up for mean
        for (b=0; b<_NB_; b++){
          
          tmp = L2[t].ref[b][p]/_RS_;
          
          switch (type){
            case 0:
              avg[w][b] += tmp;
              avg[par.nwin][b] += tmp;
              break;
            case 1:
              arr[w][b][(int)k[w]] = tmp;
              arr[par.nwin][b][(int)k[par.nwin]] = tmp;
              break;
          }

        }

        k[w]++; k[par.nwin]++;

      }

      // pixel in 3x3 mosaic
      np = (i+y_off)*_NX_*3+(j+x_off);

      // copy average for each window into 3x3 mosaic
      for (w=0; w<par.nwin; w++){

        if (k[w] > 0){

          // seasonal mean
          for (b=0, wb=w*_NB_; b<_NB_; b++, wb++){
            switch (type){
              case 0:
                AVG[wb][np] = (float)(avg[w][b]/k[w]);
                break;
              case 1:
                AVG[wb][np] = quantile(arr[w][b], k[w], 0.5);
                break;
            }
          }
          
        } else if (k[par.nwin] > 0){

          // overall mean
          for (b=0, wb=w*_NB_; b<_NB_; b++, wb++){
            switch (type){
              case 0:
                AVG[wb][np] = (float)(avg[par.nwin][b]/k[par.nwin]);
                break;
              case 1:
                AVG[wb][np] = quantile(arr[par.nwin][b], k[w], 0.5);
                break;
            }
          }

        }

      }

    }
    }

    // clean
    switch (type){
      case 0:
        free_2D((void**)avg, par.nwin+1);
        break;
      case 1:
        free_3D((void***)arr, par.nwin+1, _NB_);
        break;
    }
    free((void*)k);

  }


  // clean
  free_2D((void**)NODATA, n);

  #ifdef CLOCK
  proctime_print("computing seasonal average", TIME);
  #endif

  return SUCCESS;
}


/** Standardize float image
+++ This function computes the spatial mean and standard deviation and 
+++ standardizes the input image.
--- data:   image (modified within)
--- nodata: nodata value
--- nobs:   number of cells
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int standardize_float(float *data, float nodata, int nobs){
int p;
double avg = 0, var = 0, k = 0;

  // calculate stats
  for (p=0; p<nobs; p++){
    if (data[p] == nodata) continue;

    k++;
    if (k == 1){
      avg = data[p];
    } else {
      var_recurrence(data[p], &avg, &var, k);
    }

  }

  // standardize
  for (p=0; p<nobs; p++){
    if (data[p] == nodata) continue;
    data[p] = (data[p]-(float)avg)/(float)standdev(var, k);
  }


  return SUCCESS;
}

