/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
This file contains functions for computing spectral indices
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#include "force.h"


/** Compute spectral index time series
+++ This function computes the requested spectral index for the complete
+++ time series, and appends time information and timestamp.
--- L2:     Level-2 data
--- MASK:   Analysis mask
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int spectral_index(level2_t *L2, bool *MASK){
int t;
float **endmem = NULL;
int n_endmem;

  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  #ifdef DEBUG
  printf("Compute Index. "); fflush(stdout);
  #endif

  // allocate index, date and timestamp
  alloc_2D((void***)&par.m2d[_B_.tss].data, par.m2d[_B_.tss].length, _NC_, sizeof(int2s));
  alloc((void**)&par.m2d[_B_.tss].scale, par.m2d[_B_.tss].length, sizeof(int2s));
  alloc((void**)&par.m2d[_B_.tss].date, par.m2d[_B_.tss].length, sizeof(date_t));
  alloc((void**)&par.m2d[_B_.tss].ts, par.m2d[_B_.tss].length, sizeof(float));
  alloc_2D((void***)&par.m2d[_B_.tss].bn, par.m2d[_B_.tss].length, MAX_STR_LEN, sizeof(char));
  alloc_2D((void***)&par.m2d[_B_.tss].sid, par.m2d[_B_.tss].length, MAX_STR_LEN, sizeof(char));

  if (par.index == 81){
    alloc_2D((void***)&par.m2d[_B_.rms].data, par.m2d[_B_.rms].length, _NC_, sizeof(int2s));
    alloc((void**)&par.m2d[_B_.rms].scale, par.m2d[_B_.rms].length, sizeof(int2s));
    alloc((void**)&par.m2d[_B_.rms].date, par.m2d[_B_.rms].length, sizeof(date_t));
    alloc((void**)&par.m2d[_B_.rms].ts, par.m2d[_B_.rms].length, sizeof(float));
    alloc_2D((void***)&par.m2d[_B_.rms].bn, par.m2d[_B_.rms].length, MAX_STR_LEN, sizeof(char));
    alloc_2D((void***)&par.m2d[_B_.rms].sid, par.m2d[_B_.rms].length, MAX_STR_LEN, sizeof(char));
    // read endmember list
    n_endmem = parse_endmember_list(par.f_endmem, &endmem);
  }

  for (t=0; t<dir.n; t++){

    // scale factor
    par.m2d[_B_.tss].scale[t] = _RS_;

    // compute timestamp (decimal years)
    par.m2d[_B_.tss].ts[t] = L2[t].date.year + L2[t].date.doy/366.0;

    //  append date and sensor ID
    par.m2d[_B_.tss].date[t].ce = L2[t].date.ce;
    par.m2d[_B_.tss].date[t].day = L2[t].date.day;
    par.m2d[_B_.tss].date[t].doy = L2[t].date.doy;
    par.m2d[_B_.tss].date[t].week = L2[t].date.week;
    par.m2d[_B_.tss].date[t].month = L2[t].date.month;
    par.m2d[_B_.tss].date[t].year = L2[t].date.year;
    strcpy(par.m2d[_B_.tss].sid[t], par.l2dict.sensor[L2[t].sid]);

    // band names
    sprintf(par.m2d[_B_.tss].bn[t], "Level-2 %s", par.index_c);
    
    if (par.index == 81){
      par.m2d[_B_.rms].scale[t] = _RS_;
      par.m2d[_B_.rms].ts[t] = par.m2d[_B_.tss].ts[t];
      par.m2d[_B_.rms].date[t].ce = par.m2d[_B_.tss].date[t].ce;
      par.m2d[_B_.rms].date[t].day = par.m2d[_B_.tss].date[t].day;
      par.m2d[_B_.rms].date[t].doy = par.m2d[_B_.tss].date[t].doy;
      par.m2d[_B_.rms].date[t].week = par.m2d[_B_.tss].date[t].week;
      par.m2d[_B_.rms].date[t].month = par.m2d[_B_.tss].date[t].month;
      par.m2d[_B_.rms].date[t].year = par.m2d[_B_.tss].date[t].year;
      strcpy(par.m2d[_B_.rms].sid[t], par.m2d[_B_.tss].sid[t]);
      sprintf(par.m2d[_B_.rms].bn[t], "Level-2 %s RMSE", par.index_c);
    }

    // compute index
    if (par.index >= 0 && par.index < 80){
      if (common_index(L2[t], MASK, par.index, par.m2d[_B_.tss].data[t]) != SUCCESS){
        return FAILURE;}
    } else if (par.index == 80){
      if (disturbance_index(L2[t], MASK, par.m2d[_B_.tss].data[t]) != SUCCESS){
        return FAILURE;}
    } else if (par.index == 81){
      if (sma_index(L2[t], MASK, endmem, n_endmem, par.m2d[_B_.tss].data[t], par.m2d[_B_.rms].data[t]) != SUCCESS){
        return FAILURE;}
    }

  }

  // free endmember
  if (par.index == 81) free_2D((void**)endmem, _NB_);

  #ifdef CLOCK
  proctime_print("computing index", TIME);
  #endif

  return SUCCESS;
}


/** Compute common spectral index
+++ This function computes traditional spectral indices (e.g. NDVI) for 
+++ one date. Quality is checked within.
--- L2:     Level-2 data, single date
--- MASK:   Analysis mask
--- type:   Index ID
--- index:  spectral index (returned)
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int common_index(level2_t L2, bool *MASK, int type, int2s *index){
int p;
float rb, tmp, ind, x[3];
qai_t QAI;


  // parse Quality Assurance Information
  alloc_qai(&QAI);
  parse_qai(QAI, L2.qai);

  #pragma omp parallel private(rb,tmp,ind,x) shared(L2,QAI,MASK,_NC_,type,index,_B_) default(none)
  {

    #pragma omp for
    for (p=0; p<_NC_; p++){

      // quality check
      if (!use_this_pixel(QAI, p) || (MASK != NULL && !MASK[p])){
        index[p] = _TSAOFF_;
        continue;
      }

      switch (type){
        case 0:
          // blue band
          index[p] = L2.ref[_B_.blue][p];
          break;
        case 1:
          // green band
          index[p] = L2.ref[_B_.green][p];
          break;
        case 2:
          // red band
          index[p] = L2.ref[_B_.red][p];
          break;
        case 3:
          // nir band
          index[p] = L2.ref[_B_.nir][p];
          break;
        case 4:
          // swir 1 bands
          index[p] = L2.ref[_B_.swir1][p];
          break;
        case 5:
          // swir 2 band
          index[p] = L2.ref[_B_.swir2][p];
          break;
        case 6:
          // red edge 1 band
          index[p] = L2.ref[_B_.rededge1][p];
          break;
        case 7:
          // red edge 2 band
          index[p] = L2.ref[_B_.rededge2][p];
          break;
        case 8:
          // red edge 3 band
          index[p] = L2.ref[_B_.rededge3][p];
          break;
        case 9:
          // broad nir band
          index[p] = L2.ref[_B_.bnir][p];
          break;
        case 10:
          // NDVI
          if ((tmp = L2.ref[_B_.nir][p]+L2.ref[_B_.red][p]) != 0){
            ind = (L2.ref[_B_.nir][p]-L2.ref[_B_.red][p])/tmp;
            index[p] = (int2s)(ind*_RS_);
          } else {
            index[p] = _TSAOFF_;
          }
          break;
        case 11:
          // EVI
          if ((tmp = L2.ref[_B_.nir][p]+6.0*L2.ref[_B_.red][p]-7.5*L2.ref[_B_.blue][p]+_RS_) != 0){
            ind = 2.5*(L2.ref[_B_.nir][p]-L2.ref[_B_.red][p])/tmp;
            index[p] = (int2s)(ind*_RS_);
          } else {
            index[p] = _TSAOFF_;
          }
          break;
        case 12:
          // NBR
          if ((tmp = L2.ref[_B_.nir][p]+L2.ref[_B_.swir2][p]) != 0){
            ind = (L2.ref[_B_.nir][p]-L2.ref[_B_.swir2][p])/tmp;
            index[p] = (int2s)(ind*_RS_);
          } else {
            index[p] = _TSAOFF_;
          }
          break;
        case 13:
          // ARVI
          rb = L2.ref[_B_.red][p]-1.0*(L2.ref[_B_.blue][p]-L2.ref[_B_.red][p]);
          if ((tmp = L2.ref[_B_.nir][p]+rb) != 0){
            ind = (L2.ref[_B_.nir][p]-rb)/tmp;
            index[p] = (int2s)(ind*_RS_);
          } else {
            index[p] = _TSAOFF_;
          }
          break;
        case 14:
          // SAVI
          if ((tmp = L2.ref[_B_.nir][p]+L2.ref[_B_.red][p]+5000) != 0){
            ind = (1+0.5)*(L2.ref[_B_.nir][p]-L2.ref[_B_.red][p])/tmp;
            index[p] = (int2s)(ind*_RS_);
          } else {
            index[p] = _TSAOFF_;
          }
          break;
        case 15:
          // SARVI
          rb = L2.ref[_B_.red][p]-1.0*(L2.ref[_B_.blue][p]-L2.ref[_B_.red][p]);
          if ((tmp = L2.ref[_B_.nir][p]+rb+5000) != 0){
            ind = (1+0.5)*(L2.ref[_B_.nir][p]-rb)/tmp;
            index[p] = (int2s)(ind*_RS_);
          } else {
            index[p] = _TSAOFF_;
          }
          break;
        case 16:
          // Tasseled Cap Brigtness
          ind =  0.2043*L2.ref[_B_.blue][p]  + 0.4158*L2.ref[_B_.green][p] + 
                 0.5524*L2.ref[_B_.red][p]   + 0.5741*L2.ref[_B_.nir][p]   + 
                 0.3124*L2.ref[_B_.swir1][p] + 0.2303*L2.ref[_B_.swir2][p];
          index[p] = (int2s)ind;
          break;
        case 17:
        // Tasseled Cap Greenness
          ind = -0.1603*L2.ref[_B_.blue][p]  - 0.2819*L2.ref[_B_.green][p] - 
                 0.4934*L2.ref[_B_.red][p]   + 0.7940*L2.ref[_B_.nir][p]   - 
                 0.0002*L2.ref[_B_.swir1][p] - 0.1446*L2.ref[_B_.swir2][p];
          index[p] = (int2s)ind;
          break;
        case 18:
          // Tasseled Cap Wetness
          ind =  0.0315*L2.ref[_B_.blue][p]  +  0.2021*L2.ref[_B_.green][p] + 
                 0.3102*L2.ref[_B_.red][p]   +  0.1594*L2.ref[_B_.nir][p]   - 
                 0.6806*L2.ref[_B_.swir1][p] -  0.6109*L2.ref[_B_.swir2][p];
          index[p] = (int2s)ind;
          break;
        case 19:
          // Tasseled Cap Brigtness/Greenness/Wetness, DI without rescaling
          x[0] =  0.2043*L2.ref[_B_.blue][p]  + 0.4158*L2.ref[_B_.green][p] + 
                  0.5524*L2.ref[_B_.red][p]   + 0.5741*L2.ref[_B_.nir][p]   + 
                  0.3124*L2.ref[_B_.swir1][p] + 0.2303*L2.ref[_B_.swir2][p];
          x[1] = -0.1603*L2.ref[_B_.blue][p]  - 0.2819*L2.ref[_B_.green][p] - 
                  0.4934*L2.ref[_B_.red][p]   + 0.7940*L2.ref[_B_.nir][p]   - 
                  0.0002*L2.ref[_B_.swir1][p] - 0.1446*L2.ref[_B_.swir2][p];
          x[2] =  0.0315*L2.ref[_B_.blue][p]  +  0.2021*L2.ref[_B_.green][p] + 
                  0.3102*L2.ref[_B_.red][p]   +  0.1594*L2.ref[_B_.nir][p]   - 
                  0.6806*L2.ref[_B_.swir1][p] -  0.6109*L2.ref[_B_.swir2][p];
          index[p] = (int2s)(x[0] - (x[1] + x[2]));
          break;
      }

    }

  }

  free_qai(QAI);

  return SUCCESS;
}


/** Compute spectral mixture analysis index
+++ This function computes SMA fractions for one date. Quality is checked 
+++ within. Only the first fraction is stored.
--- L2:          Level-2 data, single date
--- MASK:        Analysis mask
--- endmember:   endmemer matrix
--- n_endmember: number of endmembers
--- index:       spectral index (returned)
--- rmse:        rmse (returned)
+++ Return:      SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int sma_index(level2_t L2, bool *MASK, float **endmember, int n_endmember, int2s *index, int2s *rmse){
int p, i, j, ik, jk;
int it, itmax;
int m, nP, sign;
int M = n_endmember;
int L = _NB_;
double tol = FLT_MIN;
double s_min, alpha;
double f = 1.0;
double rsum, dsum, res;
qai_t QAI;

gsl_matrix *Z   =  NULL;
gsl_vector *x =    NULL;
gsl_matrix *ZtZ =  NULL;
gsl_vector *Ztx =  NULL;
gsl_vector *P   =  NULL;
gsl_vector *R   =  NULL;
gsl_vector *d   =  NULL;
gsl_vector *s   =  NULL;
gsl_vector *w   =  NULL;
gsl_vector *a   =  NULL;


  itmax = 30*M;

  if (par.sma_sto) L++;

  // allocate working variables
  Z   = gsl_matrix_alloc(L, M);
  ZtZ = gsl_matrix_alloc(M, M);

  // parse Quality Assurance Information
  alloc_qai(&QAI);
  parse_qai(QAI, L2.qai);


  // copy endmember to GSL matrix
  for (i=0; i<_NB_; i++){
  for (j=0; j<M; j++){
    gsl_matrix_set(Z, i, j, (double)endmember[i][j]);
  }
  }
  if (par.sma_sto){ // append a row of 1
    for (j=0; j<M; j++){
      gsl_matrix_set(Z, i, j, 1);
    }
  }

  #ifdef DEBUG
  printf("Endmember matrix:\n");
  for (i=0; i<L; i++){
  for (j=0; j<M; j++){
    printf(" %.2f", gsl_matrix_get(Z, i, j));
  }
  printf("\n");
  }
  #endif

  // pre-compute crossproduct of Z: t(Z)Z
  gsl_blas_dgemm(CblasTrans, CblasNoTrans, 1.0, Z, Z, 0.0, ZtZ);


  #pragma omp parallel firstprivate(f) private(i,j,ik,jk,it,x,Ztx,P,R,d,s,w,a,nP,m,alpha,s_min,sign,dsum,rsum,res) shared(L2,QAI,MASK,par,_NC_,index,rmse,itmax,tol,L,M,_NB_,Z,ZtZ,stdout) default(none)
  {

    // allocate working variables
    x   = gsl_vector_alloc(L);
    Ztx = gsl_vector_alloc(M);
    P   = gsl_vector_alloc(M);
    R   = gsl_vector_alloc(M);
    d   = gsl_vector_alloc(M);
    s   = gsl_vector_alloc(M);
    w   = gsl_vector_alloc(M);
    a   = gsl_vector_alloc(M);
    
    // set last element of spectral vector to 1
    if (par.sma_sto) gsl_vector_set(x, L-1, 1);

    #pragma omp for schedule(dynamic,1)
    for (p=0; p<_NC_; p++){

      // quality check
      if (!use_this_pixel(QAI, p) || (MASK != NULL && !MASK[p])){
        index[p] = rmse[p] = _TSAOFF_;
        continue;
      }

      it = 0;

      // copy spectrum to GSL vector
      for (i=0; i<_NB_; i++) gsl_vector_set(x, i, (double)(L2.ref[i][p]/_RS_));

      // compute crossproduct of Z and x: t(Z)x
      gsl_blas_dgemv(CblasTrans, 1.0, Z, x, 0.0, Ztx);

      /** unconstrained inversion (negative values are allowed) **/
      if (!par.sma_pos){

        // allocate working variables
        gsl_matrix *ZtZ_ = gsl_matrix_alloc(M, M);
        gsl_matrix *inv = gsl_matrix_alloc(M, M);
        gsl_permutation *per = gsl_permutation_alloc(M);

        // copy t(Z)Z, as it will be modified by the inversion
        gsl_matrix_memcpy(ZtZ_, ZtZ);
        
        // LU decomposition + matrix inversion
        gsl_linalg_LU_decomp(ZtZ_, per, &sign); 
        gsl_linalg_LU_invert(ZtZ_, per, inv);

        // d = (t(Z)Z)^-1 t(z)x
        gsl_blas_dgemv(CblasNoTrans, 1.0, inv, Ztx, 0.0, d);

        // free working variables
        gsl_matrix_free(ZtZ_);
        gsl_permutation_free(per);
        gsl_matrix_free(inv);

      /** constrained inversion (negative values are not allowed) **/
      } else {

        // A: initialization (A1-A3)
        for (i=0; i<M; i++){
          gsl_vector_set(P, i, 0);
          gsl_vector_set(R, i, 1);
          gsl_vector_set(d, i, 0);
          gsl_vector_set(s, i, 0);
          gsl_vector_set(a, i, INT_MAX);
        }

        // compute w = t(Z)x - t(Z)Zd (A4)
        gsl_blas_dgemv(CblasNoTrans, 1.0, ZtZ, d, 0.0, w);
        for (i=0; i<M; i++) gsl_vector_set(w, i, 
          gsl_vector_get(Ztx, i) - gsl_vector_get(w, i));

        // B: main loop (B1)
        while (gsl_vector_max(R) > 0 && 
               gsl_vector_max(w) > tol){

          // index of maximum w (B2)
          m = gsl_vector_max_index(w);

          // add to passive, remove from active set (B3)
          gsl_vector_set(P, m, 1);
          gsl_vector_set(R, m, 0);

          // number of passive elements
          nP = gsl_blas_dasum(P);

          // allocate subsets for the matrices
          gsl_matrix *ZtZ_ = gsl_matrix_alloc(nP, nP);
          gsl_vector *Ztx_ = gsl_vector_alloc(nP);
          gsl_vector *s_ = gsl_vector_alloc(nP);
          gsl_matrix *inv = gsl_matrix_alloc(nP, nP);
          gsl_permutation *per = gsl_permutation_alloc(nP);          

          // subset the matrices according to the passive set
          for (i=0, ik=0; i<M; i++){
            if (gsl_vector_get(P, i) == 1){
              gsl_vector_set(Ztx_, ik, gsl_vector_get(Ztx, i));
              for (j=0, jk=0; j<M; j++){
                if (gsl_vector_get(P, j) == 1){
                  gsl_matrix_set(ZtZ_, ik, jk, gsl_matrix_get(ZtZ, i, j));
                  jk++;
                }
              }
              ik++;
            }
          }

          // unconstrained inversion of the passive set (B4)

          // LU decomposition + matrix inversion
          gsl_linalg_LU_decomp(ZtZ_, per, &sign); 
          gsl_linalg_LU_invert(ZtZ_, per, inv);

          // compute s = (t(Z)Z)^-1 t(z)x for the passive subset
          gsl_blas_dgemv(CblasNoTrans, 1.0, inv, Ztx_, 0.0, s_);

          // minimum of s in the passive set
          s_min = gsl_vector_min(s_);

          // copy subsetted s to complete s, set active elements to 0
          for (i=0, ik=0; i<M; i++){
            if (gsl_vector_get(P, i) == 1){
              gsl_vector_set(s, i, gsl_vector_get(s_, ik++));
            } else {
              gsl_vector_set(s, i, 0);
            }
          }

          // free working variables
          gsl_matrix_free(ZtZ_);
          gsl_vector_free(Ztx_);
          gsl_vector_free(s_);
          gsl_permutation_free(per);
          gsl_matrix_free(inv);

          // C: inner loop (C1)
          while (s_min <= 0 && it < itmax){

            it++;

            // compute alpha = -min(d/d-s) of passive set (C2)
            for (i=0; i<M; i++){
              if (gsl_vector_get(s, i) <= tol && gsl_vector_get(P, i) == 1){
                gsl_vector_set(a, i, gsl_vector_get(d, i) / 
                    (gsl_vector_get(d, i) - gsl_vector_get(s, i)));
              } else {
                gsl_vector_set(a, i, INT_MAX);
              }
            }
            alpha = gsl_vector_min(a);

            // compute new d = d + alpha(s-d) for all elements (C3)
            for (i=0; i<M; i++) gsl_vector_set(d, i, gsl_vector_get(d, i) + 
              alpha*(gsl_vector_get(s, i)-gsl_vector_get(d, i)));

            // remove from passive and add to active set (C4)
            for (i=0; i<M; i++){
              if (fabs(gsl_vector_get(d, i)) < tol && gsl_vector_get(P, i) == 1){
                gsl_vector_set(P, i, 0);
                gsl_vector_set(R, i, 1);
              }
            }

            // number of passive elements
            nP = gsl_blas_dasum(P);

            // allocate subsets for the matrices
            gsl_matrix *ZtZ_ = gsl_matrix_alloc(nP, nP);
            gsl_vector *Ztx_ = gsl_vector_alloc(nP);
            gsl_vector *s_ = gsl_vector_alloc(nP);
            gsl_matrix *inv = gsl_matrix_alloc(nP, nP);
            gsl_permutation *p = gsl_permutation_alloc(nP);

            // subset the matrices according to the passive set
            for (i=0, ik=0; i<M; i++){
              if (gsl_vector_get(P, i) == 1){
                gsl_vector_set(Ztx_, ik, gsl_vector_get(Ztx, i));
                for (j=0, jk=0; j<M; j++){
                  if (gsl_vector_get(P, j) == 1){
                    gsl_matrix_set(ZtZ_, ik, jk, gsl_matrix_get(ZtZ, i, j));
                    jk++;
                  }
                }
                ik++;
              }
            }

            // unconstrained inversion of the passive set (C5)

            // LU decomposition + matrix inversion
            gsl_linalg_LU_decomp(ZtZ_, p, &sign); 
            gsl_linalg_LU_invert(ZtZ_, p, inv);

            // compute s = (t(Z)Z)^-1 t(z)x for the passive subset
            gsl_blas_dgemv(CblasNoTrans, 1.0, inv, Ztx_, 0.0, s_);

            // minimum of s in the passive set
            s_min = gsl_vector_min(s_);

            // copy subsetted s to complete s, set active elements to 0 (C5-6)
            for (i=0, ik=0; i<M; i++){
              if (gsl_vector_get(P, i) == 1){
                gsl_vector_set(s, i, gsl_vector_get(s_, ik++));
              } else {
                gsl_vector_set(s, i, 0);
              }
            }

            // free working variables
            gsl_matrix_free(ZtZ_);
            gsl_vector_free(Ztx_);
            gsl_vector_free(s_);
            gsl_permutation_free(p);
            gsl_matrix_free(inv);


          }

          // update d with s (B5)
          gsl_vector_memcpy(d, s);

          // compute w = t(Z)x - t(Z)Zd (B6)
          // set passive elements to -1, such that the 
          // looping condition is constrained to the active set
          gsl_blas_dgemv(CblasNoTrans, 1.0, ZtZ, d, 0.0, w);
          for (i=0; i<M; i++){
            if (gsl_vector_get(R, i) == 1){
              gsl_vector_set(w, i, 
                gsl_vector_get(Ztx, i) - gsl_vector_get(w, i));
            } else {
              gsl_vector_set(w, i, -1);
            }
          }

        }

      }


      // RMSE
      rsum = 0;
      for (i=0; i<L; i++){
        dsum = 0;
        for (j=0; j<M; j++){
          dsum += gsl_vector_get(d, j)*gsl_matrix_get(Z, i, j);
        }
        res = gsl_vector_get(x, i)-dsum;
        rsum += res*res;
      }
      rmse[p]  = (int2s)(sqrt(rsum/L)*_RS_);


      // apply optional shade normalization (shade must be the last endmember)
      if (par.sma_shn){
        f = 1.0 / (1.0 - gsl_vector_get(d, M-1));
        for (i=0; i<M-1; i++) gsl_vector_set(d, i, gsl_vector_get(d, i)*f);
        gsl_vector_set(d, i, 0);

      }

      // retain one fraction
      index[p] = (int2s)(gsl_vector_get(d, par.sma_emb-1)*_RS_);



    }

    // free working variables
    gsl_vector_free(x);
    gsl_vector_free(Ztx);
    gsl_vector_free(P);
    gsl_vector_free(R);
    gsl_vector_free(d);
    gsl_vector_free(s);
    gsl_vector_free(w);
    gsl_vector_free(a);

  }

  // free working variables + QAI
  gsl_matrix_free(Z);
  gsl_matrix_free(ZtZ);
  free_qai(QAI);

  return SUCCESS;
}


/** Compute disturbance index
+++ This function computes the disturbance index for one date. Quality
+++ is checked within. The Disturbance Index is based on spatially bench-
+++ marked Tasseled Cap indices.
--- L2:     Level-2 data, single date
--- MASK:   Analysis mask
--- index:  spectral index (returned)
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int disturbance_index(level2_t L2, bool *MASK, int2s *index){
int p, i;
double x[3], xr[3];
double avg[3] = { 0.0, 0.0, 0.0 }, var[3] = { 0.0, 0.0, 0.0 }, k = 0.0;
qai_t QAI;


  // parse Quality Assurance Information
  alloc_qai(&QAI);
  parse_qai(QAI, L2.qai);

  for (p=0; p<_NC_; p++){

    // quality check
    if (!use_this_pixel(QAI, p) || (MASK != NULL && !MASK[p])){
      index[p] = _TSAOFF_;
      continue;
    }

    // Tasseled Cap Brigtness/Greenness/Wetness
    x[0] =  0.2043*L2.ref[_B_.blue][p]  + 0.4158*L2.ref[_B_.green][p] + 
            0.5524*L2.ref[_B_.red][p]   + 0.5741*L2.ref[_B_.nir][p]   + 
            0.3124*L2.ref[_B_.swir1][p] + 0.2303*L2.ref[_B_.swir2][p];
    x[1] = -0.1603*L2.ref[_B_.blue][p]  - 0.2819*L2.ref[_B_.green][p] - 
            0.4934*L2.ref[_B_.red][p]   + 0.7940*L2.ref[_B_.nir][p]   - 
            0.0002*L2.ref[_B_.swir1][p] - 0.1446*L2.ref[_B_.swir2][p];
    x[2] =  0.0315*L2.ref[_B_.blue][p]  + 0.2021*L2.ref[_B_.green][p] + 
            0.3102*L2.ref[_B_.red][p]   + 0.1594*L2.ref[_B_.nir][p]   - 
            0.6806*L2.ref[_B_.swir1][p] - 0.6109*L2.ref[_B_.swir2][p];
    for (i=0; i<3; i++) x[i] /= _RS_;

    // compute mean and variance of B/G/W
    k++;
    if (k == 1){
      for (i=0; i<3; i++) avg[i] = x[i];
    } else {
      for (i=0; i<3; i++) var_recurrence(x[i], &avg[i], &var[i], k);
    }

  }


  if (k > 0){

    for (p=0; p<_NC_; p++){

      if (index[p] == _TSAOFF_) continue;

      // Tasseled Cap Brigtness/Greenness/Wetness
      x[0] =  0.2043*L2.ref[_B_.blue][p]  + 0.4158*L2.ref[_B_.green][p] + 
              0.5524*L2.ref[_B_.red][p]   + 0.5741*L2.ref[_B_.nir][p]   + 
              0.3124*L2.ref[_B_.swir1][p] + 0.2303*L2.ref[_B_.swir2][p];
      x[1] = -0.1603*L2.ref[_B_.blue][p]  - 0.2819*L2.ref[_B_.green][p] - 
              0.4934*L2.ref[_B_.red][p]   + 0.7940*L2.ref[_B_.nir][p]   - 
              0.0002*L2.ref[_B_.swir1][p] - 0.1446*L2.ref[_B_.swir2][p];
      x[2] =  0.0315*L2.ref[_B_.blue][p]  + 0.2021*L2.ref[_B_.green][p] + 
              0.3102*L2.ref[_B_.red][p]   + 0.1594*L2.ref[_B_.nir][p]   - 
              0.6806*L2.ref[_B_.swir1][p] - 0.6109*L2.ref[_B_.swir2][p];
      for (i=0; i<3; i++) x[i] /= _RS_;

      // standardize B/G/W + compute DI
      for (i=0; i<3; i++) xr[i] = (x[i]-avg[i])/standdev(var[i], k);
      index[p] = (int2s)((xr[0] - (xr[1] + xr[2]))*1000);

    }

  }

  free_qai(QAI);

  return SUCCESS;
}

