/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
This file contains functions that handle image input operations
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#include "force.h"


/** Obtain image information with resolution adjustment
+++ This function extracts the projection, upper-left coordinates, number
+++ of pixels, and resolution of the given image. A resolution adjustment
+++ is made with respect to the requested resolution in the parameter file
--- fname:   filename
--- geotran: geotransformation (returned)
+++ Return:  SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int image_info(char *fname, double *geotran){
GDALDatasetH fp;
int nx, ny;

  GDALAllRegister();

  if ((fp = GDALOpen(fname, GA_ReadOnly)) == NULL){
    printf("unable to open %s. ", fname); return FAILURE;}

  nx = GDALGetRasterXSize(fp);
  ny = GDALGetRasterYSize(fp);
  GDALGetGeoTransform(fp, geotran);
  strcpy(par.css_dst, GDALGetProjectionRef(fp));
  GDALClose(fp);

  _NX_ = (int)((float)nx*geotran[1]/par.res);
  _NY_ = (int)((float)ny*geotran[5]/par.res * -1);
  _NC_ = _NX_*_NY_;

  geotran[1] = par.res;
  geotran[5] = par.res * -1;

  return SUCCESS;
}


/** Obtain image information
+++ This function extracts the projection, upper-left coordinates, number
+++ of pixels, and resolution of the given image.
--- fname:   filename
--- geotran: geotransformation (returned)
+++ Return:  SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int image_info_fullres(char *fname, double *geotran){
GDALDatasetH fp;

  GDALAllRegister();

  if ((fp = GDALOpen(fname, GA_ReadOnly)) == NULL){
    printf("unable to open %s. ", fname); return FAILURE;}

  _NX_ = GDALGetRasterXSize(fp);
  _NY_ = GDALGetRasterYSize(fp);
  GDALGetGeoTransform(fp, geotran);
  strcpy(par.css_dst, GDALGetProjectionRef(fp));
  GDALClose(fp);

  _NC_ = _NX_*_NY_;

  return SUCCESS;
}


/** Obtain image information for creating image mosaics
+++ This function extracts the projection, upper-left coordinates, number
+++ of pixels, and resolution of the generated mosaic. A resolution adjus-
+++ tment is made with respect to the requested resolution in the parame-
+++ ter file.
--- geotran: geotransformation (returned)
+++ Return:  SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int mosaic_info(double *geotran){
int tx, ty;
char dname[MAX_STR_LEN];
char fname[MAX_STR_LEN];
bool ok = false;

  for (ty=par.tminy; ty<=par.tmaxy; ty++){
  for (tx=par.tminx; tx<=par.tmaxx; tx++){

    sprintf(dname, "%s/X%04d_Y%04d", par.d_level2, tx, ty);
    if (list_level2(dname) == FAILURE) continue;

    sprintf(fname, "%s/%s", dname, dir.list[0]);

    if (image_info(fname, geotran) != SUCCESS){
      printf("Obtaining image info failed: %s. ", fname); return FAILURE;}

    geotran[0] = geotran[0]-_NX_*geotran[1]*(tx-par.tminx);
    geotran[3] = geotran[3]-_NY_*geotran[5]*(ty-par.tminy);

    free_2D((void**)dir.list, dir.N);
    free_2D((void**)dir.LIST, dir.N);

    ok = true; break;
  }
  }

  if (!ok){ printf("no valid data. "); return FAILURE;}

  #ifdef DEBUG
  print_dvector(geotran, "geotransformation", 6, 1, 2);
  #endif

  return SUCCESS;
}


/** Read single-band image (bool)
+++ This function reads a single-band image. In practice, there is no re-
+++ striction on data format - provided there is a GDAL driver. GDAL also
+++ takes care of image decimation / replication if the image has a diffe-
+++ rent spatial resolution than expected.
--- file:   filename
--- img:    image (returned)
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int read_single_image_bool(char *file, bool **img){
GDALDatasetH dataset;
GDALRasterBandH band;
bool *buf;
int nx, ny;

  GDALAllRegister();
  CPLSetConfigOption("GDAL_PAM_ENABLED", "NO");
  CPLPushErrorHandler(CPLQuietErrorHandler);

  alloc((void**)&buf, _NC_, sizeof(bool));


  /** read data, GDAL takes care of image decimation / replication
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ **/
  if ((dataset = GDALOpen(file, GA_ReadOnly)) == NULL){
    printf("unable to open %s. ", file); return FAILURE;
  } else {
    CPLPopErrorHandler();
  }
  nx = GDALGetRasterXSize(dataset); 
  ny = GDALGetRasterYSize(dataset);
  strcpy(meta.css_map, GDALGetProjectionRef(dataset));
  band = GDALGetRasterBand(dataset, 1);
  if (GDALRasterIO(band, GF_Read, 0, 0, nx, ny, buf, 
    _NX_, _NY_, GDT_Byte, 0, 0) == CE_Failure){
    printf("could not read image. "); return FAILURE;}
  GDALClose(dataset);

  *img = buf;
  return SUCCESS;
}


/** Read single-band image (unsigned 16bit)
+++ This function reads a single-band image. In practice, there is no re-
+++ striction on data format - provided there is a GDAL driver. GDAL also
+++ takes care of image decimation / replication if the image has a diffe-
+++ rent spatial resolution than expected.
--- file:   filename
--- img:    image (returned)
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int read_single_image_2u(char *file, int2u **img){
GDALDatasetH dataset;
GDALRasterBandH band;
int2u *buf;
int nx, ny;

  GDALAllRegister();
  CPLSetConfigOption("GDAL_PAM_ENABLED", "NO");
  CPLPushErrorHandler(CPLQuietErrorHandler);

  alloc((void**)&buf, _NC_, sizeof(int2u));


  /** read data, GDAL takes care of image decimation / replication
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ **/
  if ((dataset = GDALOpen(file, GA_ReadOnly)) == NULL){
    printf("unable to open %s. ", file); return FAILURE;
  } else {
    CPLPopErrorHandler();
  }
  nx = GDALGetRasterXSize(dataset); 
  ny = GDALGetRasterYSize(dataset);
  strcpy(meta.css_map, GDALGetProjectionRef(dataset));
  band = GDALGetRasterBand(dataset, 1);
  if (GDALRasterIO(band, GF_Read, 0, 0, nx, ny, buf, 
    _NX_, _NY_, GDT_UInt16, 0, 0) == CE_Failure){
    printf("could not read image. "); return FAILURE;}
  GDALClose(dataset);

  *img = buf;
  return SUCCESS;
}


/** Reducce spatial resolution using an approximate Point Spread Function
+++ This function will convolve the full-res image with a Gaussian Lowpass
+++ with filter size based upon the two resolutions. Afterward, the image
+++ is reduced using boxcar averaging.
--- hr:     full-res image
--- nx:     number of x-pixels (full-res)
--- ny:     number of y-pixels (full-res)
--- nc:     number of cells (full-res)
--- NX:     number of x-pixels (low-res)
--- NY:     number of y-pixels (low-res)
--- NC:     number of cells (low-res)
--- lr:     low-res image (modified within)
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int reduce_psf(int2s *hr, int nx, int ny, int nc, int NX, int NY, int NC, int2s *lr){
int nk;
int i, j, ii, jj, ni, nj, p, np;
double sum, num;
float **kernel = NULL;
float *GAUSS= NULL;
int2s *test = NULL;
float sigma, r;
float i0, i1, j0, j1, iw, jw, w;

  /** Convolute band with gaussian kernel **/

  r = nx/(float)NX;

  if ((nk = r*2) % 2 == 0) nk++;

  sigma = find_sigma(r);

  if (gauss_kernel(nk, sigma, &kernel) != SUCCESS){
    printf("Could not generate kernel. "); return FAILURE;}

  alloc((void**)&GAUSS, nc, sizeof(float));

  #pragma omp parallel private(p,np,ii,jj,ni,nj,sum,num) shared(nx,ny,GAUSS,hr,kernel,nk) default(none)
  {

    #pragma omp for collapse(2) schedule(dynamic,1)
    for (i=0; i<ny; i++){
    for (j=0; j<nx; j++){

      p = i*nx+j;
      
      GAUSS[p] = _NODATA_;

      if (hr[p] == _NODATA_) continue;

      sum = num = 0;

      for (ii=0; ii<nk; ii++){
      for (jj=0; jj<nk; jj++){

        ni = -(nk-1)/2 + ii + i; nj = -(nk-1)/2 + jj + j;
        if (ni < 0 || ni >= ny || nj < 0 || nj >= nx) continue;
        np = ni*nx+nj;
        
        if (hr[np] == _NODATA_) continue;

        sum += (hr[np]/_RS_)*kernel[ii][jj];
        num += kernel[ii][jj];

      }
      }

      if (num > 0) GAUSS[p] = sum/num;

    }
    }
    
  }

  free_2D((void**)kernel, nk);


  /** Reduce spatial resolution **/

  #pragma omp parallel private(p,np,ii,jj,ni,nj,i0,i1,j0,j1,iw,jw,w,sum,num) shared(nx,ny,NX,NY,GAUSS,lr,r) default(none)
  {

    #pragma omp for collapse(2) schedule(dynamic,1)
    for (i=0; i<NY; i++){
    for (j=0; j<NX; j++){
      
      p = i*NX+j;

      sum = num = 0;

      for (ii=0; ii<r; ii++){
      for (jj=0; jj<r; jj++){

        ni = floor(i*r)+ii;
        nj = floor(j*r)+jj;
        
        if (ni >= ny || nj >= nx) continue;
        np = ni*nx+nj;

        if (GAUSS[np] == _NODATA_) continue;

        // compute contribution of hr pixel to lr pixel
        i1 = (ni+1)/r-i; if (i1 > 1.0) i1 = 1.0;
        j1 = (nj+1)/r-j; if (j1 > 1.0) j1 = 1.0;
        i0 = ni/r-i; if (i0 < 0.0) i0 = 0.0;
        j0 = nj/r-j; if (j0 < 0.0) j0 = 0.0;
        iw = i1-i0;
        jw = j1-j0;
        w = iw*jw;

        // weighted mean
        sum += GAUSS[np]*w;
        num += w;

      }
      }

      if (num > 0) lr[p] = (int2s)(sum/num*_RS_); else lr[p] = _NODATA_;

    }
    }
    
  }

  free((void*)GAUSS);

  return SUCCESS;
}


/** Read single-band image (signed 16bit)
+++ This function reads a single-band image. In practice, there is no re-
+++ striction on data format - provided there is a GDAL driver. GDAL also
+++ takes care of image decimation / replication if the image has a diffe-
+++ rent spatial resolution than expected. A PSF aggregation can also be
+++ used.
--- file:   filename
--- img:    image (returned)
--- fnx_:   requested number of x-pixels and offset
--- fny_:   requested number of y-pixels and offset
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int read_single_image_2s(char *file, int2s **img, int2s nodata, float fnx_, float fny_, bool psf){
GDALDatasetH dataset;
GDALRasterBandH band;
int2s *buf  = NULL;
int2s *buf2 = NULL;
int2s *buf3 = NULL;
int i, j, p;
int nx, ny, nc;
int NX, NY, NC;
double geotran[6];
int nx_, ny_, nc_, offx_, offy_;
int NX_, NY_, NC_, OFFX_, OFFY_;


  GDALAllRegister();
  CPLSetConfigOption("GDAL_PAM_ENABLED", "NO");
  CPLPushErrorHandler(CPLQuietErrorHandler);

  /** read data, GDAL takes care of image decimation / replication
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ **/
  if ((dataset = GDALOpen(file, GA_ReadOnly)) == NULL){
    printf("unable to open %s. ", file); return FAILURE;
  } else {
    CPLPopErrorHandler();
  }

  // number of pixels, resolution and projection in image on disc
  nx = GDALGetRasterXSize(dataset); 
  ny = GDALGetRasterYSize(dataset);
  nc  = nx*ny;
  GDALGetGeoTransform(dataset, geotran); geotran[5] *= -1;
  strcpy(meta.css_map, GDALGetProjectionRef(dataset));

  // requested number of pixels and offset in image on disc and memory
  nx_ = ceil(fabs(fnx_/geotran[1]));
  while (nx_*geotran[1]/par.res != (NX_ = floor(nx_*geotran[1]/par.res))) nx_++;

  ny_ = ceil(fabs(fny_/geotran[5]));
  while (ny_*geotran[5]/par.res != (NY_ = floor(ny_*geotran[5]/par.res))) ny_++;

  nc_ = nx_*ny_; NC_ = NX_*NY_;

  if (fnx_ < 0) offx_ = nx-nx_;   else offx_ = 0;
  if (fny_ < 0) offy_ = ny-ny_;   else offy_ = 0;
  if (fnx_ < 0) OFFX_ = _NX_-NX_; else OFFX_ = 0;
  if (fny_ < 0) OFFY_ = _NY_-NY_; else OFFY_ = 0;


  // depending on PSF flag, read image part at full or reduced resolution
  if (psf && nc > _NC_){
    NX = nx_; NY = ny_; NC = nc_;
    alloc((void**)&buf2, NC_, sizeof(int2s));
    #ifdef DEBUG
    printf("reduce with psf: %s\n", basename(file));
    #endif
  } else {
    NX = ceil(fabs(fnx_/par.res));
    NY = ceil(fabs(fny_/par.res));
    NC = NX*NY;
    #ifdef DEBUG
    printf("input using NN:  %s\n", basename(file));
    #endif
  }


  alloc((void**)&buf, NC, sizeof(int2s));
  for (p=0; p<NC; p++) buf[p] = nodata;

  band = GDALGetRasterBand(dataset, 1);
    if (GDALRasterIO(band, GF_Read, 
      offx_, offy_, nx_, ny_, 
      buf, NX, NY, GDT_Int16, 0, 0) == CE_Failure){
    printf("could not read image %s. ", file); return FAILURE;}

  if (psf && nc > _NC_){
    reduce_psf(buf, NX, NY, NC, NX_, NY_, NC_, buf2);
  }

  GDALClose(dataset);


  if (nc_ < nc){
  
    alloc((void**)&buf3, _NC_, sizeof(int2s));
  
    for (i=0, p=0; i<_NY_; i++){
    for (j=0; j<_NX_; j++, p++){
  
      if (i < OFFY_ || j < OFFX_ ||
         i > OFFY_+NY_ || j > OFFX_+NX_){
       buf3[p] = nodata;
      } else {
        if (psf && nc > _NC_){
          buf3[p] = buf2[(i-OFFY_)*NX_+(j-OFFX_)];
        } else {
          buf3[p] = buf[(i-OFFY_)*NX_+(j-OFFX_)];
        }
      }
  
    }
    }
  
    *img = buf3;
    free((void*)buf);
    if (psf && nc > _NC_) free((void*)buf2);
  
  } else {

    if (psf && nc > _NC_){
      *img = buf2;
      free((void*)buf);
    } else {
      *img = buf;
    }

  }

  return SUCCESS;
}


/** Read multi-band image (signed 16bit)
+++ This function reads a multi-band image. In practice, there is no re-
+++ striction on data format - provided there is a GDAL driver. GDAL also
+++ takes care of image decimation / replication if the image has a diffe-
+++ rent spatial resolution than expected. A PSF aggregation can also be
+++ used. The bands that are to be read, need to be given.
--- file:   filename
--- img:    image (returned)
--- bands:  bands to read
--- fnx_:   requested number of x-pixels and offset
--- fny_:   requested number of y-pixels and offset
--- psf:    use PSF?
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int read_multi_image_2s(char *file, int2s ***img, int2s nodata, int *bands, float fnx_, float fny_, bool psf){
GDALDatasetH dataset;
GDALRasterBandH band;
char **open_options = NULL;
int2s **buf  = NULL;
int2s **buf2 = NULL;
int2s **buf3 = NULL;
char cpu[MAX_STR_LEN];
int b, i, j, p;
int nx, ny, nc;
int NX, NY, NC;
double geotran[6];
int nx_, ny_, nc_, offx_, offy_;
int NX_, NY_, NC_, OFFX_, OFFY_;


  GDALAllRegister();
  sprintf(cpu, "%d", par.cpu);
  CPLSetConfigOption("GDAL_PAM_ENABLED", "NO");
  CPLSetConfigOption("GDAL_NUM_THREADS", cpu);
  open_options = CSLSetNameValue(open_options, "NUM_THREADS", cpu);
  CPLPushErrorHandler(CPLQuietErrorHandler);

  
  /** read data
  +++ case 1: target res = image res:
  +++         read @ target res using NN
  +++ case 2: target res < image res && psf = false:
  +++         read @ target res using NN
  +++ case 3: target res < image res && psf = true:
  +++         read @ image res using NN,
  +++         reduce to target res using PSF
  +++ case 4: target res > image res:
  +++         read @ target res using NN
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

  // decide on reading function depending on GDAL version
  if (GDAL_VERSION_MAJOR  > 2 || 
     (GDAL_VERSION_MAJOR == 2 && GDAL_VERSION_MINOR >= 1)){

    if ((dataset = GDALOpenEx(file, GDAL_OF_READONLY, NULL, open_options, NULL)) == NULL){
      printf("unable to open %s. ", file); return FAILURE;
    } else {
      CPLPopErrorHandler();
    }

  } else {

     if ((dataset = GDALOpen(file, GA_ReadOnly)) == NULL){
      printf("unable to open %s. ", file); return FAILURE;
    } else {
      CPLPopErrorHandler();
    }

  }


  // number of pixels, resolution and projection in image on disc
  nx  = GDALGetRasterXSize(dataset);
  ny  = GDALGetRasterYSize(dataset);
  nc  = nx*ny;
  GDALGetGeoTransform(dataset, geotran); geotran[5] *= -1;
  strcpy(meta.css_map, GDALGetProjectionRef(dataset));

  // requested number of pixels and offset in image on disc and memory
  nx_ = ceil(fabs(fnx_/geotran[1]));
  while (nx_*geotran[1]/par.res != (NX_ = floor(nx_*geotran[1]/par.res))) nx_++;

  ny_ = ceil(fabs(fny_/geotran[5]));
  while (ny_*geotran[5]/par.res != (NY_ = floor(ny_*geotran[5]/par.res))) ny_++;

  nc_ = nx_*ny_; NC_ = NX_*NY_;

  if (fnx_ < 0) offx_ = nx-nx_;   else offx_ = 0;
  if (fny_ < 0) offy_ = ny-ny_;   else offy_ = 0;
  if (fnx_ < 0) OFFX_ = _NX_-NX_; else OFFX_ = 0;
  if (fny_ < 0) OFFY_ = _NY_-NY_; else OFFY_ = 0;
  
  #ifdef DEBUG
  printf("read %d x %d pixels with offset %d / %d at resolution %f\n", nx_, ny_, offx_, offy_, geotran[1]);
  printf("copy to %d x %d pixels with offset %d / %d at resolution %f\n", NX_, NY_, OFFX_, OFFY_, par.res);
  #endif


  // depending on PSF flag, read image part at full or reduced resolution
  if (psf && nc > _NC_){
    NX = nx_; NY = ny_; NC = nc_;
    alloc_2D((void***)&buf2, _NB_, NC_, sizeof(int2s));
    #ifdef DEBUG
    printf("reduce with psf: %s\n", basename(file));
    #endif
  } else {
    NX = NX_; NY = NY_; NC = NX_*NY_;
    #ifdef DEBUG
    printf("input using NN:  %s\n", basename(file));
    #endif
  }


  alloc_2D((void***)&buf, _NB_, NC, sizeof(int2s));
  for (b=0; b<_NB_; b++){
    for (p=0; p<NC; p++) buf[b][p] = nodata;
  }

  for (b=0; b<_NB_; b++){

    if (bands[b] < 0) continue;

    band = GDALGetRasterBand(dataset, bands[b]);
    if (GDALRasterIO(band, GF_Read, 
      offx_, offy_, nx_, ny_, 
      buf[b], NX, NY, GDT_Int16, 0, 0) == CE_Failure){
      printf("could not read image. "); return FAILURE;}

    if (psf && nc > _NC_){
      reduce_psf(buf[b], NX, NY, NC, NX_, NY_, NC_, buf2[b]);
    }

  }

  GDALClose(dataset);

  if (nc_ < nc){
  
    alloc_2D((void***)&buf3, _NB_, _NC_, sizeof(int2s));
    
    for (b=0; b<_NB_; b++){
  
      for (i=0, p=0; i<_NY_; i++){
      for (j=0; j<_NX_; j++, p++){
  
       if (i < OFFY_ || j < OFFX_ ||
           i > OFFY_+NY_ || j > OFFX_+NX_){
         buf3[b][p] = nodata;
       } else {
        if (psf && nc > _NC_){
          buf3[b][p] = buf2[b][(i-OFFY_)*NX_+(j-OFFX_)];
        } else {
          buf3[b][p] = buf[b][(i-OFFY_)*NX_+(j-OFFX_)];
        }
       }
       
  
      }
      }
    
    }
  
    *img = buf3;
    free_2D((void**)buf, _NB_);
    if (psf && nc > _NC_) free_2D((void**)buf2, _NB_);
  
  } else {

    if (psf && nc > _NC_){
      *img = buf2;
      free_2D((void**)buf, _NB_);
    } else {
      *img = buf;
    }

  }

  CSLDestroy(open_options);

  return SUCCESS;
}


/** Read multi-band image (float)
+++ This function reads a multi-band image. In practice, there is no re-
+++ striction on data format - provided there is a GDAL driver. GDAL also
+++ takes care of image decimation / replication if the image has a diffe-
+++ rent spatial resolution than expected.
--- file:   filename
--- img:    image (returned)
--- nb:     number of bands to read
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int read_multi_image_fl(char *file, float ***img, int *nb){
GDALDatasetH dataset;
GDALRasterBandH band;
float **buf = NULL;
int b, nx, ny;


  GDALAllRegister();
  CPLSetConfigOption("GDAL_PAM_ENABLED", "NO");
  CPLPushErrorHandler(CPLQuietErrorHandler);

  /** read data, GDAL takes care of image decimation / replication
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ **/
  if ((dataset = GDALOpen(file, GA_ReadOnly)) == NULL){
    printf("unable to open %s. ", file); return FAILURE;
  } else {
    CPLPopErrorHandler();
  }
  nx  = GDALGetRasterXSize(dataset); 
  ny  = GDALGetRasterYSize(dataset);
  *nb = GDALGetRasterCount(dataset);
  strcpy(meta.css_map, GDALGetProjectionRef(dataset));

  alloc_2D((void***)&buf, *nb, _NC_, sizeof(float));

  for (b=0; b<*nb; b++){
    band = GDALGetRasterBand(dataset, b+1);
    if (GDALRasterIO(band, GF_Read, 0, 0, nx, ny, buf[b], 
      _NX_, _NY_, GDT_Float32, 0, 0) == CE_Failure){
      printf("could not read image %s. ", file); return FAILURE;}
  }
  GDALClose(dataset);

  *img = buf;
  return SUCCESS;
}


/** Read Level 1 data
+++ This function reads all necessary or available Level 1 data, which is
+++ dependent on satellite mission. If available, cirrus, water vapour and
+++ temperature bands will be read separately. The 'usual' reflectance 
+++ bands will be read to an image stack.
--- DN:     Digital Numbers (returned)
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int read_level1(dn_t *DN){
int b;
dn_t dn;

  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif
  
  #ifdef DEV
  printf("need to change par.band_ --> par.use.\n");
  #endif

  // thermal band
  if (par.band_bt){
    if (read_single_image_2u(meta.btcal.fname, &dn.bt) != SUCCESS){
        printf("Error in temperatue. "); return FAILURE;}
  }

  // reflectance bands
  if ((dn.ref = (int2u**)malloc(_NB_*sizeof(int2u*))) == NULL){
    printf("unable to allocate memory. "); return FAILURE;}

  for (b=0; b<_NB_; b++){
    if (read_single_image_2u(meta.refcal[b].fname, &dn.ref[b]) != SUCCESS){
      printf("Error in reflectance. "); return FAILURE;}
  }


  #ifdef CLOCK
  proctime_print("read Level 1", TIME);
  #endif

  #ifdef DEBUG
  printf("\n%s\n", meta.css_map);
  db_write_image_2D((void**)dn.ref, "DN-REF", _NB_, _NX_, _NY_, sizeof(int2u), 12);
  if (par.band_bt)  db_write_image_1D((void**)dn.bt,  "DN-BT",  _NX_, _NY_, sizeof(int2u), 12);
  #endif

  *DN = dn;
  return SUCCESS;
}


/** Read Level 2 data
+++ This function reads all necessary or available Level 2 data, which is
+++ dependent on the generated products. If available, reflectance, cloud
+++ and cloud shadow, aod, hot and view zenith products are input.
--- dname:  Level-2 directory
--- L2:     Level-2 data (returned)
--- fnx_:   requested number of x-pixels and offset
--- fny_:   requested number of y-pixels and offset
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int read_level2(char *dname, level2_t **L2, float fnx_, float fny_){
int t;
char fname[MAX_STR_LEN], *pch = NULL;
char temp[MAX_STR_LEN];
level2_t *l2 = NULL;

  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  #ifdef DEBUG
  printf("Reading. "); fflush(stdout);
  #endif

  alloc((void**)&l2, dir.n, sizeof(level2_t));

  // open threads
  omp_set_num_threads(par.cpu);

  #pragma omp parallel private(fname,temp,pch) shared(l2,dir,par,dname,fnx_,fny_) default(none)
  {

  #pragma omp for
  for (t=0; t<dir.n; t++){

    // information about this dataset
    date_level2(&l2[t].date, dir.list[t]);
    if (sensor_level2(&l2[t].sid, dir.list[t]) != SUCCESS){
    printf("unknown sensor of L2 data\n"); }//return FAILURE;}

    // filename
    sprintf(fname, "%s/%s", dname, dir.list[t]);

    // use improphed product?
    if (par.use.imp){
      
      // backup filename
      strcpy(temp, fname);

      // new filename
      if (strstr(fname, "BOA")  != NULL) pch = strstr(fname, "BOA");
      if (strstr(fname, "TOA")  != NULL) pch = strstr(fname, "TOA");
      strncpy(pch, "IMP", 3);
      
      // if no improphed product exists, use normal one
      if (!fileexist(fname)) strcpy(fname, temp);

    }

    // read BOA or TOA (or IMP) reflectance
    if (par.use.ref){
      if (read_multi_image_2s(fname, &l2[t].ref, _NODATA_, par.l2dict.band[l2[t].sid], fnx_, fny_, par.psf) != SUCCESS){
        printf("Error reading reflectance product. "); }//return FAILURE;}
    } else l2[t].ref = NULL;

    // read Quality Assurance Information
    if (strstr(fname, "BOA")  != NULL) pch = strstr(fname, "BOA");
    if (strstr(fname, "TOA")  != NULL) pch = strstr(fname, "TOA");
    if (strstr(fname, "IMP")  != NULL) pch = strstr(fname, "IMP");
    strncpy(pch, "QAI", 3);

    if (par.use.qai){
      if (read_single_image_2s(fname, &l2[t].qai, _QAIOFF_, fnx_, fny_, false) != SUCCESS){
        printf("Error reading QAI product. "); }//return FAILURE;}
    } else l2[t].qai = NULL;

    // read Cloud distance
    pch = strstr(fname, "QAI"); strncpy(pch, "CLD", 3);
    if (par.use.cld){
      if (read_single_image_2s(fname, &l2[t].cld, _NODATA_, fnx_, fny_, par.psf) != SUCCESS){
        printf("Error reading CLD product. "); }//return FAILURE;}
    } else l2[t].cld = NULL;

    // read Aerosol Optical Depth
    pch = strstr(fname, "CLD"); strncpy(pch, "AOD", 3);
    if (par.use.aod){
      if (read_single_image_2s(fname, &l2[t].aod, _NODATA_, fnx_, fny_, par.psf) != SUCCESS){
        printf("Error reading AOD product. "); }//return FAILURE;}
    } else l2[t].aod = NULL;

    // read Haze Optimized Transformation
    pch = strstr(fname, "AOD"); strncpy(pch, "HOT", 3);
    if (par.use.hot){
      if (read_single_image_2s(fname, &l2[t].hot, _NODATA_, fnx_, fny_, par.psf) != SUCCESS){
        printf("Error reading HOT product. "); }//return FAILURE;}
    } else l2[t].hot = NULL;

    // read View Zenith Angle
    pch = strstr(fname, "HOT"); strncpy(pch, "VZN", 3);
    if (par.use.vzn){
      if (read_single_image_2s(fname, &l2[t].vzn, _NODATA_, fnx_, fny_,  par.psf) != SUCCESS){
        printf("Error reading VZN product. "); }//return FAILURE;}
    } else l2[t].vzn = NULL;

    // read Water Vapor
    pch = strstr(fname, "VZN"); strncpy(pch, "WVP", 3);
    if (par.use.wvp){
      if (read_single_image_2s(fname, &l2[t].wvp, _NODATA_, fnx_, fny_, par.psf) != SUCCESS){
        printf("Error reading WVP product. "); }//return FAILURE;}
    } else l2[t].wvp = NULL;

  }

  }

  #ifdef CLOCK
  proctime_print("read Level 2", TIME);
  #endif

  *L2 = l2;
  return SUCCESS;
}


/** Read analysis mask
+++ This function reads analysis masks. If no mask is available, FAILURE
+++ is returned.
--- dname:  mask directory
--- MSK:    mask data (returned)
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int read_mask(char *dname, bool **MSK){
char *fname = NULL;
bool *msk = NULL;

  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  #ifdef DEBUG
  printf("Reading. "); fflush(stdout);
  #endif

  if ((fname = findfile_filter(dname, "MASK", ".dat")) == NULL){
    if ((fname = findfile_filter(dname, "MASK", ".tif")) == NULL){
      return FAILURE;}
  }

  if (read_single_image_bool(fname, &msk) != SUCCESS){
          printf("Error reading analysis mask %s. ", fname); return FAILURE;}

  free((void*)fname);

  #ifdef CLOCK
  proctime_print("read analysis mask", TIME);
  #endif

  *MSK = msk;
  return SUCCESS;
}


/** Buffer Level 2 data
+++ This function buffers the nodata regions in Level 2 data with one
+++ pixel.
--- L2:     Level-2 data (modified)
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int buffer_level2(level2_t *L2){
int t, p;
bool *TMP;

  alloc((void**)&TMP, _NC_, sizeof(bool));

  for (t=0; t<dir.n; t++){

    for (p=0; p<_NC_; p++){
      if (L2[t].qai[p] == _QAIOFF_) TMP[p] = true; else TMP[p] = false;
    }

    buffer(TMP, 1);

    for (p=0; p<_NC_; p++){
      if (TMP[p]) L2[t].qai[p] = _QAIOFF_;
    }

  }

  free((void*)TMP);

  return SUCCESS;
}


/** Deallocate Level 2 data
+++ This function deallocates the Level 2 data
--- L2:     Level-2 data (deallocated within)
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int free_level2(level2_t *L2){
int t;

  for (t=0; t<dir.n; t++){
    if (par.use.ref) free_2D((void**)L2[t].ref, _NB_);
    if (par.use.qai) free((void*)L2[t].qai);
    if (par.use.cld) free((void*)L2[t].cld);
    if (par.use.aod) free((void*)L2[t].aod);
    if (par.use.hot) free((void*)L2[t].hot);
    if (par.use.vzn) free((void*)L2[t].vzn);
    if (par.use.wvp) free((void*)L2[t].wvp);
  }
  free((void*)L2);

  return SUCCESS;
}


/** Extract date from Level 2 data
+++ This function extracts the date from the level 2 file's basename, i.e.
+++ acquisition year, month, day, day of year, week and days since CE.
--- DATE:   date struct (returned)
--- bname:  basename of Level 2 image
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int date_level2(date_t *DATE, char *bname){
char cy[5], cm[3], cd[3];
date_t date;

  strncpy(cy, bname,   4); cy[4] = '\0';
  strncpy(cm, bname+4, 2); cm[2] = '\0';
  strncpy(cd, bname+6, 2); cd[2] = '\0';

  date.year = atoi(cy);
  date.month = atoi(cm);
  date.day = atoi(cd);
  date.doy = md2doy(date.month, date.day);
  date.week = doy2week(date.doy);
  date.ce  = doy2ce(date.doy, date.year);

  *DATE = date;
  return SUCCESS;
}


/** Extract sensor ID of Level 2 data
+++ This function extracts the sensor from the level 2 file's basename,
+++ e.g. LND08 and relates it to the L2 sensor dictionary.
--- sid:    sensor ID (returned)
--- bname:  basename of Level 2 image
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int sensor_level2(int *sid, char *bname){
int s;
char cs[6];

  strncpy(cs, bname+16, 5); cs[5] = '\0';

  for (s=0; s<par.l2dict.n; s++){
    if (strstr(cs, par.l2dict.sensor[s]) != NULL){
      *sid = s; return SUCCESS;
    }
  }

  return FAILURE;
}


/** List Level 2 data
+++ This function lists all Level 2 reflectance products in a given direc-
+++ tory. If there are none, FAILURE is returned. Only requested sensors 
+++ are listed. Only the requested time frame is listed. The dir_t struct 
+++ must be freed on success.
--- dname:  Level 2 directory
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int list_level2(char *dname){
int t, s;
bool vs;
date_t date;

  // directory listing
  if ((dir.N = scandir(dname, &dir.LIST, 0, alphasort)) < 0){
    return FAILURE;}

  // reflectance products
  alloc_2D((void***)&dir.list, dir.N, MAX_STR_LEN, sizeof(char));
  for (t=0, dir.n=0; t<dir.N; t++){

    if ((strstr(dir.LIST[t]->d_name, ".dat") != NULL ||
         strstr(dir.LIST[t]->d_name, ".tif") != NULL) && 
         strstr(dir.LIST[t]->d_name, ".aux.xml") == NULL &&
        (strstr(dir.LIST[t]->d_name, "BOA")  != NULL ||
         strstr(dir.LIST[t]->d_name, "TOA")  != NULL)){

      // check against sensor list
      for (s=0, vs=false; s<par.l2dict.n; s++){
        if (strstr(dir.LIST[t]->d_name, par.l2dict.sensor[s]) != NULL){
          vs = true; break;
        }
      }

      // extract and filter date
      date_level2(&date, dir.LIST[t]->d_name);
      if (date.year < par.yearmin) vs = false;
      if (date.year > par.yearmax) vs = false;
      if (date.doy < par.doymin) vs = false;
      if (date.doy > par.doymax) vs = false;

      // white-list image
      if (vs) strcpy(dir.list[dir.n++], dir.LIST[t]->d_name);

    }

  }

  if (dir.n<1){
    free_2D((void**)dir.list, dir.N);
    free_2D((void**)dir.LIST, dir.N);
    return FAILURE;
  }

  #ifdef DEBUG
  printf("\n%d reflectance products in tile %s\n", dir.n, basename_with_ext(dname));
  #endif

  return SUCCESS;
}


/** Filter Level 2 list
+++ This function filters the Level 2 list. All datasets that are within
+++ the requested temporal range are retained. The range must be given in
+++ days since CE (no-leap-year approximation). The Level 2 list is modi-
+++ fied (re-ordered and shortened). If there is no dataset left, FAILURE
+++ is returned. The dir_t struct must be freed on success.
--- cemin:  start of range
--- cemax:  end of range
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int list_level2_filter_ce(int cemin, int cemax){
int t;
date_t date;
char **list;
int n;


  alloc_2D((void***)&list, dir.n, MAX_STR_LEN, sizeof(char));

  for (t=0, n=0; t<dir.n; t++){
    date_level2(&date, dir.list[t]);
    if (date.ce < cemin || date.ce > cemax) continue;
    strcpy(list[n++], dir.list[t]);
  }

  for (t=0; t<dir.n; t++){
    if (t >= n){
      strcpy(dir.list[t], "NULL");
    } else {
      strcpy(dir.list[t], list[t]);
    }
  }
  free_2D((void**)list, dir.n);

  if ((dir.n = n)<1){
    free_2D((void**)dir.list, dir.N);
    free_2D((void**)dir.LIST, dir.N);
    return FAILURE;
  }

  #ifdef DEBUG
  printf("\n%d reflectance products within ce limits\n", dir.n);
  #endif

  return SUCCESS;
}


/** Filter Level 2 list
+++ This function filters the Level 2 list. Only multiple datasets per day
+++ are retained. The Level 2 list is modified (re-ordered and shortened). 
+++ If there is no dataset left, FAILURE is returned. The dir_t struct 
+++ must be freed on success.
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int list_level2_filter_multiday(){
int t;
date_t date_0, date_1;
char **list;
int n;
bool ok;

  alloc_2D((void***)&list, dir.n, MAX_STR_LEN, sizeof(char));


  for (t=0, n=0; t<dir.n-1; t++){

    ok = false;
    do {

      date_level2(&date_0, dir.list[t]);
      if (t < dir.n){
        date_level2(&date_1, dir.list[t+1]);
      } else {
        date_1.ce = 0;
      }

      if (date_0.ce == date_1.ce){
        strcpy(list[n++], dir.list[t++]);
        ok = true;
      } else if (ok){
        strcpy(list[n++], dir.list[t]);
        ok = false;
      }

    } while (ok);

  }

  for (t=0; t<dir.n; t++){
    if (t >= n){
      strcpy(dir.list[t], "NULL");
    } else {
      strcpy(dir.list[t], list[t]);
    }
  }
  free_2D((void**)list, dir.n);

  if ((dir.n = n)<1){
    free_2D((void**)dir.list, dir.N);
    free_2D((void**)dir.LIST, dir.N);
    return FAILURE;
  }

  #ifdef DEBUG
  printf("\n%d reflectance products within ce limits\n", dir.n);
  #endif

  return SUCCESS;
}


/** Test if certain band is available
+++ This function tests if a requested band is available, and if so, sets
+++ a bool array (with all available bands) flag to true
--- b:      requested band
--- valid:  available bands, bool array
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int band_available(int b, bool *valid){

  if (b == -1){
    printf("requested band is not available. unable to compute index. ");
    return FAILURE;
  } else {
    valid[b] = true;
    return SUCCESS;
  }
}


/** Level 2 sensor dictionary
+++ This function builds a Level 2 sensor dictionary, which is needed to
+++ generate multi-sensor products. It computes the most restrictive over-
+++ lap between matching bands, determine rules how to read these, and de-
+++ fines commonly used wavelengths.
--- request: requested sensor list
--- l2s:     Level 2 sensor dictionary (returned)
+++ Return:  SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int level2sensor_dict(char *request, level2_sensor_t *l2s){
int s, ss, b, bb, c;
char upper[MAX_STR_LEN] = "UPPER";
const int  ns = 12, nb = 10;
const char sensor[12][MAX_STR_LEN] = { 
                          "LND04", "LND05", "LND07", 
                          "LND08", "SEN2A", "SEN2B",
                          "sen2a", "sen2b", "LNDLG",
                          "SEN2L", "SEN2H", "R-G-B"};
const int  band[12][10] = {{ 1, 2, 3, 0, 0, 0, 0, 4, 5,  6 },  // Landsat 4 TM   (legacy bands)
                           { 1, 2, 3, 0, 0, 0, 0, 4, 5,  6 },  // Landsat 5 TM   (legacy bands)
                           { 1, 2, 3, 0, 0, 0, 0, 4, 5,  6 },  // Landsat 7 ETM+ (legacy bands)
                           { 1, 2, 3, 0, 0, 0, 0, 4, 5,  6 },  // Landsat 8 OLI  (legacy bands)
                           { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 },  // Sentinel-2A MSI (land surface bands)
                           { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 },  // Sentinel-2B MSI (land surface bands)
                           { 1, 2, 3, 0, 0, 0, 7, 0, 0,  0 },  // Sentinel-2A MSI (high-res bands)
                           { 1, 2, 3, 0, 0, 0, 7, 0, 0,  0 },  // Sentinel-2B MSI (high-res bands)
                           { 1, 2, 3, 0, 0, 0, 0, 4, 5,  6 },  // Landsat legacy bands
                           { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 },  // Sentinel-2 land surface bands
                           { 1, 2, 3, 0, 0, 0, 0, 4, 5,  6 },  // Sentinel-2 high-res bands
                           { 1, 2, 3, 0, 0, 0, 0, 0, 0,  0 }}; // VIS bands
bool vs[12], vb[10];

  // match available sensors with requested sensors
  for (s=0, ss=0; s<ns; s++){
    if (strstr(request, sensor[s]) != NULL){
      vs[s] = true;
      ss++;
    } else vs[s] = false;
  }

  if ((l2s->n = ss) == 0){ printf("no sensor match (%s).\n", request);
    return FAILURE;}


  // kick out bands that are incomplete
  for (b=0, bb=0; b<nb; b++){

    for (s=0, vb[b]=true; s<ns; s++){
      if (vs[s] && band[s][b] == 0) vb[b] = false;
    }

    if (vb[b]) bb++;

  }

  if ((_NB_ = bb) == 0){ printf("no band overlap for requested sensors.\n"); return FAILURE;}

  // set target sensor
  if (_NB_ == 6){
    strcpy(l2s->target, "LNDLG");
  } else if (_NB_ == 10){
    strcpy(l2s->target, "SEN2L");
  } else if (_NB_ == 4){
    strcpy(l2s->target, "SEN2H");
  } else if (_NB_ == 3){
    strcpy(l2s->target, "R-G-B");
  } else {
    printf("unknown sensors.\n"); return FAILURE;
  }


  // build sensor struct
  alloc_2D((void***)&l2s->sensor, l2s->n, MAX_STR_LEN, sizeof(char));
  alloc_2D((void***)&l2s->band,   l2s->n, _NB_,        sizeof(int));

  for (b=0, bb=0; b<nb; b++){
    if (!vb[b]) continue;
    for (s=0, ss=0; s<ns; s++){
      if (!vs[s]) continue;
      for (c=0; c<5; c++) upper[c] = toupper(sensor[s][c]);
      strcpy(l2s->sensor[ss], upper);
      l2s->band[ss][bb] = band[s][b];
      ss++;
    }
    bb++;
  }


  // initialize named bands
  init_bands();

  for (b=0, bb=0; b<nb; b++){
    if (!vb[b]) continue;
    if (b == 0) _B_.blue     = bb;
    if (b == 1) _B_.green    = bb;
    if (b == 2) _B_.red      = bb;
    if (b == 3) _B_.rededge1 = bb;
    if (b == 4) _B_.rededge2 = bb;
    if (b == 5) _B_.rededge3 = bb;
    if (b == 6) _B_.bnir     = bb;
    if (b == 7) _B_.nir      = bb;
    if (b == 8) _B_.swir1    = bb;
    if (b == 9) _B_.swir2    = bb;
    bb++;
  }


  #ifdef DEBUG
  printf("waveband mapping:\n");
  for (b=0; b<nb; b++) printf("%d ", vb[b]); printf("\n");
  printf("blue    %02d, green %02d, red   %02d\n", _B_.blue, _B_.green, _B_.red);
  printf("re_1    %02d, re_2  %02d, re_3  %02d\n", _B_.rededge1, _B_.rededge2, _B_.rededge3);
  printf("bnir  %02d, nir   %02d, swir1 %02d\n", _B_.bnir, _B_.nir, _B_.swir1);
  printf("swir2 %02d\n", _B_.swir2);
  #endif

  #ifdef DEBUG
  printf("processing with %d sensors and %d bands\n", l2s->n, _NB_);
  for (s=0; s<l2s->n; s++){
    printf("%s: ", l2s->sensor[s]);
    for (b=0; b<_NB_; b++) printf("%2d ", l2s->band[s][b]); printf("\n");
  }
  #endif

  return SUCCESS;
}


/** Read coordinates from text file
+++ This function reads coordinates from a text file. Put X-coords in 1st
+++ column and Y-coords in 2nd column. Coordinates must be in geographic 
+++ decimal degree notation (South and West coordinates are negative). Do
+++ not use a header.
--- fcoords: text file containing the coordinates
--- COORD:   array for storing the coordinates
+++ Return:  number of coordinate pairs
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int parse_coord_list(char *fcoords, float ***COORD){
FILE *fp;
char  buffer[MAX_STR_LEN] = "\0";
char *tokenptr = NULL;
char *separator = " ";
int c, nc = 0;
float **coord = NULL;


  // open file
  if (!(fp = fopen(fcoords, "r"))){
    printf("unable to open coordinate file. "); return FAILURE;}

  // count lines
  while (fgets(buffer, MAX_STR_LEN, fp) != NULL) nc++;
  fseek(fp, 0, SEEK_SET);

  #ifdef DEBUG
  printf("number of coordinate pairs: %d\n", nc);
  #endif

  alloc_2D((void***)&coord, 2, nc, sizeof(float));

  // read line by line
  for (c=0; c<nc; c++){
  
    if (fgets(buffer, MAX_STR_LEN, fp) == NULL){
      printf("error reading coordinate file.\n"); exit(1); }

    tokenptr = strtok(buffer, separator);
    coord[0][c] = atof(tokenptr); // X
    tokenptr = strtok(NULL, separator);
    coord[1][c] = atof(tokenptr); // Y

  }

  fclose(fp);
  
  *COORD = coord;
  return nc;
}


/** Read endmembers from text file
+++ This function reads endmembers from a text file. Put each endmember in
+++ a separate column, separated by space. Do not use a header. Put prima-
+++ ry endmember into the first column, and if applicable, shade endmember
+++ into the last column. Values are integers, scaled by 10000.
--- fendmem: text file containing the endmembers
--- ENDMEM:  array for storing the endmembers
+++ Return:  number of endmembers
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int parse_endmember_list(char *fendmem, float ***ENDMEM){
FILE *fp;
char  buffer[MAX_STR_LEN] = "\0";
char *ptr = NULL;
char *separator = " ";
int b, n;
float **endmem = NULL;
float tmp;


  // open file
  if (!(fp = fopen(fendmem, "r"))){
    printf("unable to open endmember file. "); return FAILURE;}

  alloc_2D((void***)&endmem, _NB_, MAX_STR_LEN, sizeof(float));

  // read line by line
  for (b=0; b<_NB_; b++){
  
    if (fgets(buffer, MAX_STR_LEN, fp) == NULL){
      printf("error reading endmember file.\n"); exit(1); }

    n = 0;

    ptr = strtok(buffer, separator);

    do {

      tmp = atof(ptr)/_RS_;

      if (tmp < 0 || tmp > 1){
        printf("\nendmember file is invalid.\n");
        printf("valid range: [0,10000].\n");
        exit(1);
      } else {
        endmem[b][n++] = tmp;
      }

    } while ((ptr = strtok(NULL, separator)) != NULL);

  }

  fclose(fp);
  
  *ENDMEM  = endmem;
  return n;
}

