/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
This program is the FORCE Level-2 Processing System (single image)
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#include "force.h"


int main( int argc, char *argv[] ){
dn_t  DN;
toa_t TOA;
top_t TOP;
qai_t QAI;
int2u  *GRD = NULL;
int2s **DAT = NULL;


  if (argc != 3){
    printf("usage: %s image(dir) prm(file)\n", argv[0]);
    printf("version: %s\n", _VERSION_);
    exit(1);
  }


  /** initialization + read metadata, parameter and tile file
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
  time_t TIME; time(&TIME);


  #ifdef DEBUG
  db_check();
  #endif

  // allow a test, if lnd-prepro is compiled in DEBUG mode
  if (strcmp(argv[1], "?") == 0){
    #ifdef DEBUG
    return 1;
    #endif
    return 0;
  }


  // get command line parameters
  par.d_img = argv[1];
  par.f_par = argv[2];

  // parse parameter file
  if (parse_param_l2() != SUCCESS){
    printf("Parsing parameter file failed.\n"); exit(1);}

  // parse metadata
  if (parseMetadata() != SUCCESS){
    printf("Parsing metadata failed.\n"); exit(1);}

  atc_init();
  mem_to_disc_dict_l2();


  /** read Digital Numbers + projection
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
  if (read_level1(&DN) != SUCCESS){
    printf("reading DNs failed.\n"); exit(1);}


  /** allocate Quality Assurance Information
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
  if (alloc_qai(&QAI) != SUCCESS){
    printf("allocating QAI failed.\n"); exit(1);}


  /** no data boundary & saturation & impulse noise in VIS layers
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
  if (dn_extreme(DN, QAI) != SUCCESS){
    printf("extreme value identification failed.\n"); exit(1);}


  /** fine-to-coarse resolution grid + allocate coarse parameters
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
  cggrid(&GRD);


  /** sun-target-view geometry
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
  sun_target_view(QAI);


  /** TOA reflectance + brightness temperature
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
  if (dn_to_toa(GRD, QAI, DN, &TOA) != SUCCESS){
    printf("DN to TOA conversion failed.\n"); exit(1);}


  /** read/reproject/ckeck DEM and compute slope/aspect
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
  if (compile_topo(&TOP, QAI, GRD) != SUCCESS){
    printf("unable to compile topography.\n"); return FAILURE;}


 /** cloud and cloud shadow detection
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
  if (cloud_module(TOA, QAI, TOP) != SUCCESS){
    proctime_print("Processing time", TIME); exit(1);}


  /** resolution merge
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
  if (resolution_merge(TOA.ref, QAI) != SUCCESS){
    printf("unable to merge resolutions.\n"); return FAILURE;}


  /** radiometric correction
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
  if (radiometric_module(TOA.ref, QAI, TOP, GRD, &DAT) != SUCCESS){
    printf("Error in radiometric module.\n"); exit(1);}


  /** reprojection, tiling and output
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
  if (geometric_module(DAT) != SUCCESS){
    printf("Error in geometric module.\n"); exit(1);}

  printf("Success! "); proctime_print("Processing time", TIME);

  return SUCCESS;
}

