/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
This program is the FORCE Level-3 Processing System
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#include "force.h"


int main ( int argc, char *argv[] ){
level2_t *L2 = NULL;
qai_t *QAI = NULL;
target_t tar;
int t, i, j, p, b, l, nt;
int tx, ty, tc = 0, tk = 0;
int *tiles_x = NULL, *tiles_y = NULL, tiles_k;
score_t *score = NULL;
int2s  **range_stats = NULL;
double **recur_stats = NULL;
float **q_array = NULL;
int2s q25, q75;
char dname[MAX_STR_LEN];
double geotran_mosaic[6] = { 0, 0, 0, 0, 0, 0 };
double geotran_tile[6]   = { 0, 0, 0, 0, 0, 0 };
double **cor = NULL;
double hsum, hnum, hmean, hsd;
float maxt;
bool water;
float wk, k;
time_t TIME, TIME_I, TIME_C, TIME_O;
double time_i = 0, time_c = 0, time_o = 0, ctime = 0;
double eta_h, eta_m, eta_s;
off_t hist[12][BIG_STR_LEN];
int h, hbin;

  if (argc != 2){ printf("usage: %s parameter-file\n", argv[0]); return FAILURE;}

/** INITIALIZING
+** read parameter file, get raster properties, prepare output files etc.
+** *********************************************************************/

  time(&TIME);
  printf("Initializing startup"); fflush(stdout);

  par.f_par = argv[1];

  // parse parameter file
  if (parse_param_l3() != SUCCESS){
    printf("Reading parameter file failed!\n"); return FAILURE;}
    
  // copy datacube definition
  if (copy_datacube_def(par.d_level2, par.d_level3) == FAILURE){
    printf("Copying datacube definition failed.\n"); exit(1);}

  // get tile file and read it
  if (tilefileread(&tiles_x, &tiles_y, &tiles_k) != SUCCESS){
    printf("Reading tile file failed!\n"); return FAILURE;}

  // how many tiles to do?
  for (ty=par.tminy; ty<=par.tmaxy; ty++){
  for (tx=par.tminx; tx<=par.tmaxx; tx++){

    // if tile is not whitelisted (if specified), skip
    if (tile_whitelisted(tiles_x, tiles_y, tiles_k, tx, ty) == FAILURE) continue;
    tk++;

  }
  }

  // find UL coordinate of mosaic + resolution
  if (mosaic_info(geotran_mosaic) != SUCCESS){
    printf("Obtaining mosaic info failed.\n"); return FAILURE;}
  geotran_tile[1] = geotran_mosaic[1];
  geotran_tile[5] = geotran_mosaic[5];
  
  // map memory to disc output and allocate memory
  mem_to_disc_dict_l3();

  // initialize mosaic (if requested)
  if (init_products_l3(geotran_mosaic) != SUCCESS){
    printf("Initializing mosaic failed!\n"); return FAILURE;}

  // init stats
  for (h=0; h<12; h++){
    for (hbin=0; hbin<BIG_STR_LEN; hbin++) hist[h][hbin] = 0;
  }
  
  printf("\rL3 processing started, I/C/O TBD, ETA: TBD");
  fflush(stdout);

  // open threads
  omp_set_num_threads(par.cpu);


/** START PROCESSING; LOOP OVER ALL AVAILABLE TILES
+** *********************************************************************/

  for (ty=par.tminy; ty<=par.tmaxy; ty++){
  for (tx=par.tminx; tx<=par.tmaxx; tx++){

    time(&TIME_I);

    // current tile
    sprintf(dname, "%s/X%04d_Y%04d", par.d_level2, tx, ty);
    
    geotran_tile[0] = geotran_mosaic[0]+_NX_*geotran_tile[1]*(tx-par.tminx);
    geotran_tile[3] = geotran_mosaic[3]+_NY_*geotran_tile[5]*(ty-par.tminy);
    
    // if tile is not whitelisted (if specified), skip
    if (tile_whitelisted(tiles_x, tiles_y, tiles_k, tx, ty) == FAILURE) continue;

    // get Level2 file listing, skip if tile is empty
    if (list_level2(dname) == FAILURE){ tc++; continue;}


    /** RESET L3
    +** *************************************************************/
    for (p=0; p<_NC_; p++){
      for (l=0; l<_NP_; l++){
        if (!par.m2d[l].save) continue;
        for (b=0; b<par.m2d[l].length; b++){
          par.m2d[l].data[b][p] = par.m2d[l].nodata;
        }
      }
      if (par.m2d[_B_.inf].save) par.m2d[_B_.inf].data[0][p] = _QAIOFF_;
    }


    /** PROCESS LSP INFORMATION OR MIMICK LSP WITH STATIC INPUT
    +** *************************************************************/
    if (compile_temporal_target(tx, ty, &tar) != SUCCESS){
      #ifdef DEBUG
      printf("compling temporal target failed. continue with next tile.\n");
      #endif
      tc++;
      continue;
    }


    /** READ LANDSAT DATA
    +** *************************************************************/

    if (read_level2(dname, &L2, _NX_*par.res, _NY_*par.res) != SUCCESS){
      printf("error reading L2.\n"); return FAILURE;}
    buffer_level2(L2);
    
    time_i += proctime(TIME_I);
    time(&TIME_C);


    /** PARSE QUALITY
    +** *************************************************************/
    #ifdef DEBUG
    printf("Parsing QAI. "); fflush(stdout);
    #endif

    alloc((void**)&QAI, dir.n, sizeof(qai_t));

    for (t=0; t<dir.n; t++){
      alloc_qai(&QAI[t]);
      parse_qai(QAI[t], L2[t].qai);
    }


    /** COMPOSITE
    +** *************************************************************/
    #ifdef DEBUG
    printf("Compositing. "); fflush(stdout);
    #endif


    #pragma omp parallel private(b, j, p, t, l, range_stats, recur_stats, hbin, q_array, q25, q75, nt, score, maxt, cor, hsum, hnum, hmean, hsd, wk, k, water)
    {

      alloc_2D((void***)&range_stats, _NB_, 3, sizeof(int2s));
      alloc_2D((void***)&recur_stats, _NB_, 5, sizeof(double));

      alloc_2D((void***)&q_array, _NB_, dir.n, sizeof(float));

      alloc_2D((void***)&cor, dir.n, dir.n, sizeof(double));
      alloc((void**)&score, dir.n, sizeof(score_t));

      #pragma omp for
      for (i=0; i<_NY_; i++){
      for (j=0; j<_NX_; j++){

        p = _NX_*i+j;

        wk = k = 0;

        for (t=0; t<dir.n; t++){

          // check quality
          if (!use_this_pixel(QAI[t], p)) continue;

          if (QAI[t].wtr[p]) wk++;
          k++;

        }

        if (k > 0 && wk/k > 0.9) water = true; else water = false;


        // reset stats
        for (b=0; b<_NB_; b++){
          range_stats[b][0] = SHRT_MAX;
          for (l=1; l<3; l++) range_stats[b][l] = 0.0;
          for (l=0; l<5; l++) recur_stats[b][l] = 0.0;
        }
        maxt = -1;

        // compute correlation matrix
        corr_matrix(L2, QAI, p, cor);

        // compute scores
        compositing_scores(L2, QAI, tar, p, cor, score);

        // override total score in case of water
        if (water && _B_.swir2 != -1){
          for (t=0, nt=0; t<dir.n; t++){
            if (!use_this_pixel(QAI[t], p)) continue;
            score[t].t = (10000-L2[t].ref[_B_.swir2][p])/10000.0;
          }
        }

        // mean of haze score
        hsum = hnum = 0;
        for (t=0; t<dir.n; t++){
          if (!use_this_pixel(QAI[t], p) || (!par.offsea && score[t].d < 0.01)) continue;
          if (par.use.cld && score[t].c < 0.01) continue;
          hsum += score[t].h;
          hnum++;
        }
        if (hnum > 0) hmean = hsum/hnum; else hmean = -9999;

        // sd of haze score
        hsum = hnum = 0;
        for (t=0; t<dir.n; t++){
          if (!use_this_pixel(QAI[t], p) || (!par.offsea && score[t].d < 0.01)) continue;
          if (par.use.cld && score[t].c < 0.01) continue;
          hsum += (score[t].h-hmean)*(score[t].h-hmean);
          hnum++;
        }
        if (hnum > 0) hsd = sqrt(hsum/hnum); else hsd = -9999;


        // compositing / temporal statistics
        for (t=0, nt=0; t<dir.n; t++){

          if (!use_this_pixel(QAI[t], p)) continue;
          if (!par.offsea && score[t].d < 0.01) continue;
          if (par.use.hot && score[t].h < 0.01 && hmean > 0.01 && hsd > 0.01) continue;
          if (par.use.cld && score[t].c < 0.01) continue;


          // update if score increased
          if (score[t].t > maxt){

            if (water && _B_.swir2 != -1 && _B_.nir != -1 && 
                L2[t].ref[_B_.nir][p] < L2[t].ref[_B_.swir2][p]) continue;

            maxt = score[t].t;

            // update score
            if (par.m2d[_B_.scr].save){
              par.m2d[_B_.scr].data[0][p] = (int2s) (score[t].t*10000);
              if (!water){
                par.m2d[_B_.scr].data[1][p] = (int2s) (score[t].d*10000);
                par.m2d[_B_.scr].data[2][p] = (int2s) (score[t].y*10000);
                par.m2d[_B_.scr].data[3][p] = (int2s) (score[t].c*10000);
                par.m2d[_B_.scr].data[4][p] = (int2s) (score[t].h*10000);
                par.m2d[_B_.scr].data[5][p] = (int2s) (score[t].r*10000);
                par.m2d[_B_.scr].data[6][p] = (int2s) (score[t].v*10000);
              }
            }

            // update difference to target DOY and YEAR
            if (par.m2d[_B_.inf].save){
              par.m2d[_B_.inf].data[0][p] = L2[t].qai[p];
              par.m2d[_B_.inf].data[2][p] = L2[t].date.doy;
              par.m2d[_B_.inf].data[3][p] = L2[t].date.year;
              par.m2d[_B_.inf].data[4][p] = score[t].dd;
              par.m2d[_B_.inf].data[5][p] = L2[t].sid+1;
            }

            if (par.m2d[_B_.bap].save){
              for (b=0; b<_NB_; b++) par.m2d[_B_.bap].data[b][p] = L2[t].ref[b][p];
            }

          }


          //if (par.use.hot && score[t].h < 0.01 && hmean > 0.01 && hsd > 0.01) continue;
          //if (par.use.cld && score[t].c < 0.01) continue;

          // temporal statistics
          recur_stats[0][0]++;
          for (b=0; b<_NB_; b++){
            if (par.m2d[_B_.avg].save || par.m2d[_B_.std].save || 
                par.m2d[_B_.skw].save || par.m2d[_B_.krt].save){
                  kurt_recurrence(L2[t].ref[b][p], &recur_stats[b][1], 
                               &recur_stats[b][2], &recur_stats[b][3], 
                               &recur_stats[b][4],  recur_stats[0][0]);}
            if ((par.m2d[_B_.min].save || par.m2d[_B_.rng].save) && 
                L2[t].ref[b][p] < range_stats[b][0]){
                  range_stats[b][0] = L2[t].ref[b][p];}
            if ((par.m2d[_B_.max].save || par.m2d[_B_.rng].save) && 
                L2[t].ref[b][p] > range_stats[b][1]){
                  range_stats[b][1] = L2[t].ref[b][p];}
            if (par.m2d[_B_.q25].save || par.m2d[_B_.q50].save || 
                par.m2d[_B_.q75].save || par.m2d[_B_.iqr].save){
                  q_array[b][nt] = L2[t].ref[b][p];}
          }
          nt++;

        }



        /** finalize spectral variability stats **/

        // # of obs
        if (par.m2d[_B_.inf].save) par.m2d[_B_.inf].data[1][p] = recur_stats[0][0];
        if (recur_stats[0][0] == 0) continue;

        // copy to final products
        for (b=0; b<_NB_; b++){
          if (par.m2d[_B_.min].save){
            par.m2d[_B_.min].data[b][p] = 
              range_stats[b][0];
          }
          if (par.m2d[_B_.max].save){
            par.m2d[_B_.max].data[b][p] = 
              range_stats[b][1];
          }
          if (par.m2d[_B_.rng].save){
            par.m2d[_B_.rng].data[b][p] = 
              range_stats[b][1]-range_stats[b][0];
          }
          if (par.m2d[_B_.avg].save){
            par.m2d[_B_.avg].data[b][p] = 
              (int2s) recur_stats[b][1];
          }
          if (par.m2d[_B_.std].save){
            par.m2d[_B_.std].data[b][p] = 
              (int2s) standdev(recur_stats[b][2], recur_stats[0][0]);
          }
          if (par.m2d[_B_.skw].save){
            par.m2d[_B_.skw].data[b][p] = 
              (int2s) (skewness(recur_stats[b][2], recur_stats[b][3], recur_stats[0][0])*10000);
          }
          if (par.m2d[_B_.krt].save){
            par.m2d[_B_.krt].data[b][p] = 
              (int2s) (kurtosis(recur_stats[b][2], recur_stats[b][4], recur_stats[0][0])*10);
          }
          if (par.m2d[_B_.q25].save){
            par.m2d[_B_.q25].data[b][p] = 
              (int2s) quantile(q_array[b], nt, 0.25);
          }
          if (par.m2d[_B_.q50].save){
            par.m2d[_B_.q50].data[b][p] = 
              (int2s) quantile(q_array[b], nt, 0.50);
          }
          if (par.m2d[_B_.q75].save){
            par.m2d[_B_.q75].data[b][p] = 
              (int2s) quantile(q_array[b], nt, 0.75);
          }
          if (par.m2d[_B_.iqr].save){
            if (par.m2d[_B_.q25].save){
              q25 = par.m2d[_B_.q25].data[b][p];
            } else {
              q25 = (int2s) quantile(q_array[b], nt, 0.25);
            }
            if (par.m2d[_B_.q75].save){
              q75 = par.m2d[_B_.q75].data[b][p];
            } else {
              q75 = (int2s) quantile(q_array[b], nt, 0.75);
            }
            par.m2d[_B_.iqr].data[b][p] = 
              q75 - q25;
          }
        }

      }
      }
      

      // clean
      free_2D((void**)range_stats, _NB_);
      free_2D((void**)recur_stats, _NB_);
      free_2D((void**)q_array, _NB_);
      free_2D((void**)cor, dir.n);
      free((void*)score);

    }
    
    // log stats

    for (p=0; p<_NC_; p++){
      if (par.m2d[_B_.inf].save){
        if ((hbin = par.m2d[_B_.inf].data[1][p]) != par.m2d[_B_.inf].nodata){
          hist[0][hbin]++; // # of obs
        }
        if ((hbin = par.m2d[_B_.inf].data[2][p]) != par.m2d[_B_.inf].nodata){
          hist[1][hbin]++; // D
        }
        if ((hbin = par.m2d[_B_.inf].data[3][p]) != par.m2d[_B_.inf].nodata){
          hist[2][hbin]++; // Y
        }
        if ((hbin = par.m2d[_B_.inf].data[4][p]) != par.m2d[_B_.inf].nodata){
          hbin += BIG_STR_LEN/2; hist[3][hbin]++; // dD
        }
        if ((hbin = par.m2d[_B_.inf].data[5][p]) != par.m2d[_B_.inf].nodata){
          hist[4][hbin]++; // Sensor ID
        }
      }
      if (par.m2d[_B_.scr].save){
        if ((hbin = par.m2d[_B_.scr].data[0][p]) != par.m2d[_B_.scr].nodata){
          hbin = ceil(hbin/500.0); hist[5][hbin]++; // ST
        }
        if ((hbin = par.m2d[_B_.scr].data[1][p]) != par.m2d[_B_.scr].nodata){
          hbin = ceil(hbin/500.0); hist[6][hbin]++;  // SD
        }
        if ((hbin = par.m2d[_B_.scr].data[2][p]) != par.m2d[_B_.scr].nodata){
          hbin = ceil(hbin/500.0); hist[7][hbin]++;  // SY
        }
        if ((hbin = par.m2d[_B_.scr].data[3][p]) != par.m2d[_B_.scr].nodata){
          hbin = ceil(hbin/500.0); hist[8][hbin]++;  // SC
        }
        if ((hbin = par.m2d[_B_.scr].data[4][p]) != par.m2d[_B_.scr].nodata){
          hbin = ceil(hbin/500.0); hist[9][hbin]++;  // SH
        }
        if ((hbin = par.m2d[_B_.scr].data[5][p]) != par.m2d[_B_.scr].nodata){
          hbin = ceil(hbin/500.0); hist[10][hbin]++; // SR
        }
        if ((hbin = par.m2d[_B_.scr].data[6][p]) != par.m2d[_B_.scr].nodata){
          hbin = ceil(hbin/500.0); hist[11][hbin]++; // SV
        }
      }
    }
    
    time_c += proctime(TIME_C);

    #ifdef CLOCK
    proctime_print("Compositing", TIME_C);
    #endif


    time(&TIME_O);


    /** write output **/
    write_products_l3(tx, ty, geotran_tile);
    
    /** clean **/
    free_level2(L2);
    for (t=0; t<dir.n; t++) free_qai(QAI[t]); free((void*)QAI);

    free_2D((void**)tar.a, par.Yn);
    free_2D((void**)tar.b, par.Yn);
    free_3D((void***)tar.p, 3, par.Yn);

    free_2D((void**)dir.list, dir.N);
    free_2D((void**)dir.LIST, dir.N);

    time_o += proctime(TIME_O);

    #ifdef CLOCK
    proctime_print("Output", TIME_O);
    #endif


    /** PROGRESS
    +** *****************************************************************/
    tc++; ctime = time_i + time_c + time_o;
    eta_s = (tk-tc)*(ctime/tc);
    if (eta_s >= 3600){
      eta_h = floor(eta_s/3600); eta_s -= eta_h*3600;
    } else eta_h = 0;
    if (eta_s >= 60){
      eta_m = floor(eta_s/60); eta_s -= eta_m*60;
    } else eta_m = 0;

    printf("\rL3 processing %6.2f%%, I/C/O %03.0f/%03.0f/%03.0f, ETA: %02.0fh %02.0fm %02.0fs", 
      (float)tc/tk*100.0, round(time_i/ctime*100), round(time_c/ctime*100), round(time_o/ctime*100), 
       eta_h, eta_m, eta_s);
    fflush(stdout);

  }
  }
  
  printf("\n\n");

  // finalize mosaic (if requested)
  if (final_products_l3(geotran_mosaic) != SUCCESS){
    printf("Finalizing mosaic failed!\n"); return FAILURE;}


  /** CLEAN
  +** *****************************************************************/
  mem_to_disc_free();
  free((void*)tiles_x); free((void*)tiles_y);


  /** PRINT STATS
  +** *****************************************************************/
  print_bap_stats(hist[0], "# of observations", 0, 1);
  print_bap_stats(hist[1], "Acquisition DOY", 0, 1);
  print_bap_stats(hist[2], "Acquisition Year", 0, 1);
  print_bap_stats(hist[3], "Difference between Acquisition DOY and Target DOY", -BIG_STR_LEN/2, 1);
  print_bap_stats(hist[4], "Sensor ID", 0, 1);
  print_bap_stats(hist[5], "Total score", 0, 500);
  print_bap_stats(hist[6], "DOY score", 0, 500);
  print_bap_stats(hist[7], "Year score", 0, 500);
  print_bap_stats(hist[8], "Cloud distance score", 0, 500);
  print_bap_stats(hist[9], "Haze score", 0, 500);
  print_bap_stats(hist[10], "Correlation score", 0, 500);
  print_bap_stats(hist[11], "View angle score", 0, 500);

  proctime_print("Total runtime", TIME);

  return SUCCESS;
}

