/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
This file contains functions that handle LSP-related functionality
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#include "force.h"


/** List LSP files
+++ This function lists the LSP files and finds the filenames of the re-
+++ quested LSP parameters. If the requested files cannot be found, 
+++ FAILURE is returned. Unless compiled in DEBUG mode, no error message
+++ is displayed.
--- tx:     Tile-ID X
--- ty:     Tile-ID Y
+++ Return: SUCCESS or FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int list_lsp(int tx, int ty){
int t, f;
dir_t dlsp;
char dname[MAX_STR_LEN];

  sprintf(dname, "%s/X%04d_Y%04d", par.d_lsp, tx, ty);

  // directory listing
  if ((dlsp.N = scandir(dname, &dlsp.LIST, 0, alphasort)) < 0){
    #ifdef DEBUG
    printf("there is no file in %s. ", dname); 
    #endif
    return FAILURE;
  }

  
  for (f=0; f<3; f++) sprintf(par.f_lsp[f], "NULL");

  // matching LSP pattern
  for (t=0; t<dlsp.N; t++){

    if (strstr(dlsp.LIST[t]->d_name, ".dat") == NULL &&
        strstr(dlsp.LIST[t]->d_name, ".tif") == NULL) continue;

    for (f=0; f<3; f++){
      if (strstr(dlsp.LIST[t]->d_name, par.p_lsp[f]) != NULL){
        sprintf(par.f_lsp[f], "%s/%s", dname, dlsp.LIST[t]->d_name);
      }
    }

  }
  
  free_2D((void**)dlsp.LIST, dlsp.N);

  for (f=0; f<3; f++){
    if (strcmp(par.f_lsp[f], "NULL") == 0){
      #ifdef DEBUG
      printf("could not find pattern %s in %s. ", par.p_lsp[f], dname);
      #endif
      return FAILURE;
    }
  }

  return SUCCESS;
}


/** Offset LSP time series
+++ This function offsets the LSP time series to a common time baseline,
+++ i.e. relative to Year 0 DOY 0.
--- nb:     Number of years (i.e. bands) in full time series
--- LSP0:   The full LSP time series
+++ Return: SUCCESS or FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int offset_lsp(int nb, float **LSP0){
int b, p;

  #pragma omp parallel private(b)
  {

    #pragma omp for
    for (p=0; p<_NC_; p++){

      for (b=0; b<nb; b++){
        if (LSP0[b][p] == par.lsp_nodata) continue;
        LSP0[b][p] += par.lsp_start - 1;
      }

    }

  }

  #ifdef DEBUG
  double *sum = NULL;
  double *num = NULL;
  int y, m, d;
  alloc((void**)&sum, nb, sizeof(double));
  alloc((void**)&num, nb, sizeof(double));
  #pragma omp parallel private(p)
  {
    #pragma omp for
    for (b=0; b<nb; b++){
      for (p=0; p<_NC_; p++){
        if (LSP0[b][p] == par.lsp_nodata) continue;
        sum[b] += LSP0[b][p]/(2000.0*365.0);
        num[b] ++;
      }
    }
  }
  ce2date((int)(sum[b]/num[b]*2000*365), &m, &d, &y);
  for (b=0; b<nb; b++) printf("mean LSP band %02d: %04d-%02d-%02d\n", b, y, m, d);
  free((void*)sum);
  free((void*)num);
  #endif

  return SUCCESS;
}


/** Read and generate annual LSP
+++ This function reads the complete time series of one phenologial para-
+++ meter and generates a clean time series that coincides with the compo-
+++ siting period. 
--- fname:  File name of LSP parameter
--- LSP:    The annual LSP time series (needs to be deallocated)
+++ Return: SUCCESS or FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int read_lsp(char *fname, float ***LSP){
int nb;
float **lsp0, **lsp1;


  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  /** read the LSP time series **/
  if (read_multi_image_fl(fname, &lsp0, &nb) != SUCCESS){
    printf("Error reading LSP product. "); return FAILURE;}

  /** offset the LSP time series **/
  if (offset_lsp(nb, lsp0) != SUCCESS){
      printf("Error offsetting LSP product. "); return FAILURE;}

  /** interpolate, extraolate and clip the time series **/
  if (annual_lsp(nb, lsp0, &lsp1) != SUCCESS){
    printf("Error reading LSP product. "); return FAILURE;}

  #ifdef CLOCK
  proctime_print("read (& clean) LSP", TIME);
  #endif

  *LSP = lsp1;
  return SUCCESS;
}


/** Annual LSP time series
+++ This function generates a clean time series that coincides with the 
+++ compositing period. A linear regression is fit to the time series. No-
+++ data pixels are interpolated. High RMSE pixels can also be replaced by
+++ the interpolation. Eventually, the time series is clipped (or extended
+++ via extrapolation) to the compositing period.
--- nb:     Number of years (i.e. bands) in full time series
--- LSP0:   The full LSP time series (is deallocated within)
--- LSP1:   The annual LSP time series (needs to be deallocated)
+++ Return: SUCCESS or FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int annual_lsp(int nb, float **LSP0, float ***LSP1){
double mx, my, vx, vy, cv, n, intercept, slope, yhat, sum, num, rmse;
int y0 = par.lsp_y0, y1 = par.Yt-par.Yr;
int yr, b, p;
int lsp_yr, lsp_doy;
float **lsp1;
  
  
  alloc_2D((void***)&lsp1, par.Yn, _NC_, sizeof(float));

  #pragma omp parallel private(b, yr, mx, my, vx, vy, cv, n, slope, intercept, yhat, rmse, num, sum, lsp_doy, lsp_yr)
  {

    #pragma omp for
    for (p=0; p<_NC_; p++){


      /** linear regression **/
      mx = my = vx = vy = cv = n = 0.0;

      for (b=0, yr=y0; b<nb; b++, yr++){

        ce2doy(LSP0[b][p], &lsp_doy, &lsp_yr);
        if (lsp_yr < yr-1 || lsp_yr > yr+1) LSP0[b][p] = par.lsp_nodata;
        if (LSP0[b][p] == par.lsp_nodata) continue;

        if (++n == 1){
          mx = (double)yr;
          my = (double)LSP0[b][p];
        } else {
          covar_recurrence((double)yr, (double)LSP0[b][p], &mx, &my,
            &vx, &vy, &cv, n);
        }

      }


      /** root mean squared error **/
      num = sum = 0;
      if (n > 1){
        linreg_coefs(mx, my, cv, vx, &slope, &intercept);
      } else if (n == 1){
        intercept = my-mx*365;
        slope     = 365;
      }

      for (b=0, yr=y0; b<nb; b++, yr++){

        if (LSP0[b][p] == par.lsp_nodata) continue;
        linreg_predict((double)yr, slope, intercept, &yhat);
        sum += (LSP0[b][p] - yhat)*(LSP0[b][p] - yhat);
        num++;

      }
      rmse = sqrt(sum/num);


      /** interpolate nodata and large residuals **/
      for (b=0, yr=y0; b<nb; b++, yr++){

        if (LSP0[b][p] == par.lsp_nodata || rmse > par.lsp_rmse){
          linreg_predict((double)yr, slope, intercept, &yhat);
          if (yhat < 0 || yhat > INT_MAX || isnan(yhat)) yhat = 0;
          LSP0[b][p] = yhat;
        }

      }


      /** clip and/or extend the time series to the compositing period **/
      for (b=0, yr=y1; b<par.Yn; b++, yr++){

        if (yr < y0 || yr >= y0+nb){
          linreg_predict((double)yr, slope, intercept, &yhat);
          if (yhat < 0 || yhat > INT_MAX || isnan(yhat)) yhat = 0;
          lsp1[b][p] = yhat;
        } else {
          lsp1[b][p] = LSP0[b+(y1-y0)][p];
        }

      }

    }

  }


  /** clean the full time series **/
  free_2D((void**)LSP0, nb);

  *LSP1 = lsp1;
  return SUCCESS;
}

