/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
This file contains functions for parsing metadata
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#include "force.h"


/** Initilaize metadata
+++ This function initializes metadata to enable value testing
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int mtlinit(){
int fill = -32767;

  meta.id = fill;
  strcpy(meta.satellite,  "NULL");
  meta.dtype = fill;
  meta.res = fill;
  meta.map_ul.x = meta.map_ul.y = fill;
  strcpy(meta.css_map, "NULL");
  meta.date.ce = meta.date.day = fill;
  meta.date.doy = meta.date.month = fill;
  meta.date.year = fill;
  meta.date.hh = meta.date.mm = fill;
  meta.date.ss = meta.date.tz = fill;
  meta.tier = fill;
  meta.scale = fill;

  meta.s2.szen  = NULL;  meta.s2.sazi = NULL;
  meta.s2.vzen  = NULL;  meta.s2.vazi = NULL;
  meta.s2.nx = meta.s2.ny = fill;

  return SUCCESS;
}


/** Check metadata
+++ This function tests if all metadata are OK
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int mtlcheck(){
int fill = -32767, b;

  if (meta.id == fill){
    printf("error in Landsat ID. "); return FAILURE;}
   if (meta.id != 4 && meta.id != 5 && meta.id != 7 && meta.id != 8 && meta.id != 2){
    printf("error in sensor ID. Unknown sensor. "); return FAILURE;}
  if(strcmp(meta.satellite, "NULL") == 0){
    printf("error in satellite name. "); return FAILURE;}
   if (meta.dtype == fill){
    printf("error in dtype. "); return FAILURE;}
   if (meta.scale == fill){
    printf("error in scale. "); return FAILURE;}
  if (meta.res == fill){
    printf("error in resolution. "); return FAILURE;}
  if (meta.map_ul.x == fill || meta.map_ul.y == fill){
    printf("error in map ul. "); return FAILURE;}
  if (atof(meta.wrs) == 0){
    printf("error in path/row. "); return FAILURE;}
  if (meta.date.day == fill){
    printf("error in date.day. "); return FAILURE;}
  if (meta.date.doy == fill){
    printf("error in date.doy. "); return FAILURE;}
  if (meta.date.month == fill){
    printf("error in date.month. "); return FAILURE;}
  if (meta.date.year == fill){
    printf("error in date.year. "); return FAILURE;}
  if (meta.date.hh == fill){
    printf("error in time.hh. "); return FAILURE;}
  if (meta.date.mm == fill){
    printf("error in time.mm. "); return FAILURE;}
  if (meta.date.ss == fill){
    printf("error in time.ss. "); return FAILURE;}
  if (meta.date.tz == fill){
    printf("error in time.tz. "); return FAILURE;}
  if (meta.tier > par.tier){
    printf("Unacceptable tier.\n"); exit(1);}
  for (b=0; b<_NB_; b++){
    if(strcmp(meta.refcal[b].fname, "NULL") == 0){
      printf("error in fname. "); return FAILURE;}
    if (meta.refcal[b].lmax == fill){
      printf("error in Lmax. "); return FAILURE;}
    if (meta.refcal[b].lmax == fill){
      printf("error in Lmin. "); return FAILURE;}
    if (meta.refcal[b].qmax == fill){
      printf("error in Qcalmax. "); return FAILURE;}
    if (meta.refcal[b].qmin == fill){
      printf("error in Qcalmin. "); return FAILURE;}
    if (meta.refcal[b].rmul == fill){
      printf("error in reflectance scale factor (x). "); return FAILURE;}
    if (meta.refcal[b].radd == fill){
      printf("error in reflectance scale factor (+). "); return FAILURE;}
  }
  if(strcmp(meta.btcal.fname, "NULL") == 0){
    printf("error in fname. "); return FAILURE;}
  if (meta.btcal.lmax == fill){
    printf("error in Lmax. "); return FAILURE;}
  if (meta.btcal.lmax == fill){
    printf("error in Lmin. "); return FAILURE;}
  if (meta.btcal.qmax == fill){
    printf("error in Qcalmax. "); return FAILURE;}
  if (meta.btcal.qmin == fill){
    printf("error in Qcalmin. "); return FAILURE;}
  if (meta.btcal.k1 == fill || meta.btcal.k2 == fill){
    printf("error in k1/k2. "); return FAILURE;}

  return SUCCESS;
}


/** Print metadata
+++ This function prints the metadata
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int mtlprint(){
int b;

  for (b=0; b<_NB_; b++){
    printf("REF: %s\n", basename_with_ext(meta.refcal[b].fname));
    printf(" LMAX/LMIN %.2f/%.2f, QMAX/QMIN %.2f/%.2f, RMULTI/RADD %.2f/%.2f, E0 %.2f\n", 
      meta.refcal[b].lmax, meta.refcal[b].lmin, meta.refcal[b].qmax, meta.refcal[b].qmin, meta.refcal[b].rmul, meta.refcal[b].radd, meta.refcal[b].E0);
  }

  if (par.band_bt){
  printf("BT:  %s\n", basename_with_ext(meta.btcal.fname));
  printf(" LMAX/LMIN %.2f/%.2f, QMAX/QMIN %.2f/%.2f\n", 
    meta.btcal.lmax, meta.btcal.lmin, meta.btcal.qmax, meta.btcal.qmin);
  printf(" K1/K2 %.2f/%.2f\n", meta.btcal.k1, meta.btcal.k2);
  }

  printf("dtype: %d\n", meta.dtype);
  printf("id: %d\n", meta.id);
  printf("wrs   = %s\n", meta.wrs);
  printf("res: %d\n", meta.res);
  printf("yyyy/mm/dd + doy = %04d/%02d/%02d + %03d\n", meta.date.year, meta.date.month, meta.date.day, meta.date.doy);
  printf("hh/mm/ss= %02d/%02d/%02d\n", meta.date.hh, meta.date.mm, meta.date.ss);
  printf("ul: %.1f/%.1f\n", meta.map_ul.x, meta.map_ul.y);
  printf("tier: %d\n", meta.tier);
  printf("satellite ID: %s\n", meta.satellite);
  printf("scale factor: %d\n", meta.scale);

  return SUCCESS;
}


/** Parse Landsat metadata
+++ This function reads the Landsat metadata
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int parseMetadataLandsat(){
FILE *fp = NULL;
char *fname = NULL;
char  buffer[MAX_STR_LEN] = "\0";
char *tag = NULL;
char *tokenptr = NULL;
char *tokenptr2 = NULL;
char *separator = " =\":";
char *separator2 = "-";
int b, k, kk, path = 0, row = 0;


  mtlinit();

  par.band_bt  = true;
  par.owvp = false;

  // scan directory for MTL.txt file
  fname = findfile(par.d_img, "MTL.txt");

  // open MTL file
  if ((fp = fopen(fname, "r")) == NULL){
    printf("Unable to open MTL file! "); return FAILURE;
  } else free((void*)fname);

  // process line by line
  while (fgets(buffer, MAX_STR_LEN, fp) != NULL){

    if (strstr(buffer, "END_GROUP = L1_METADATA_FILE")) break;
    
    // get tag
    tokenptr = strtok(buffer, separator);
    tag=tokenptr;

    // extract parameters by comparing tag
    while (tokenptr != NULL){

      tokenptr = strtok(NULL, separator);

      // Landsat sensor
      if (strcmp(tag, "SPACECRAFT_ID") == 0){

        meta.id = atoi(tokenptr+strlen(tokenptr)-1);
        sprintf(meta.satellite, "LND%02d", meta.id);

        if (meta.id == 8){
          meta.dtype = 16;
          _NB_ = 8;
        } else {
          meta.dtype = 8;
          _NB_ = 6;
        }

        alloc((void**)&meta.refcal, _NB_, sizeof(cal_t));
        for (b=0; b<_NB_; b++) calinit(&meta.refcal[b]);
        calinit(&meta.btcal);
        alloc((void**)&par.use.refbands, _NB_, sizeof(bool));

        k = 0;
        if (meta.id == 8){
          kk = 18;
          strcpy(meta.refcal[k].mtl_band, "1"); meta.refcal[k].rsr_band = kk++; par.use.refbands[k] = false; _B_.ultrablue  = k++;
          strcpy(meta.refcal[k].mtl_band, "2"); meta.refcal[k].rsr_band = kk++; par.use.refbands[k] = true;  _B_.blue  = k++;
          strcpy(meta.refcal[k].mtl_band, "3"); meta.refcal[k].rsr_band = kk++; par.use.refbands[k] = true;  _B_.green = k++;
          strcpy(meta.refcal[k].mtl_band, "4"); meta.refcal[k].rsr_band = kk++; par.use.refbands[k] = true;  _B_.red   = k++;
          strcpy(meta.refcal[k].mtl_band, "5"); meta.refcal[k].rsr_band = kk++; par.use.refbands[k] = true;  _B_.nir   = k++;
          strcpy(meta.refcal[k].mtl_band, "9"); meta.refcal[k].rsr_band = kk++; par.use.refbands[k] = false; _B_.cir   = k++;
          strcpy(meta.refcal[k].mtl_band, "6"); meta.refcal[k].rsr_band = kk++; par.use.refbands[k] = true;  _B_.swir1 = k++;
          strcpy(meta.refcal[k].mtl_band, "7"); meta.refcal[k].rsr_band = kk++; par.use.refbands[k] = true;  _B_.swir2 = k++;
          strcpy(meta.btcal.mtl_band, "10");
        } else if (meta.id == 7 || meta.id == 5 || meta.id == 4){
          if (meta.id == 4) kk = 0; else if (meta.id == 5) kk = 6; else if (meta.id == 7) kk = 12;
          strcpy(meta.refcal[k].mtl_band, "1"); meta.refcal[k].rsr_band = kk++; par.use.refbands[k] = true;  _B_.blue  = k++;
          strcpy(meta.refcal[k].mtl_band, "2"); meta.refcal[k].rsr_band = kk++; par.use.refbands[k] = true;  _B_.green = k++;
          strcpy(meta.refcal[k].mtl_band, "3"); meta.refcal[k].rsr_band = kk++; par.use.refbands[k] = true;  _B_.red   = k++;
          strcpy(meta.refcal[k].mtl_band, "4"); meta.refcal[k].rsr_band = kk++; par.use.refbands[k] = true;  _B_.nir   = k++;
          strcpy(meta.refcal[k].mtl_band, "5"); meta.refcal[k].rsr_band = kk++; par.use.refbands[k] = true;  _B_.swir1 = k++;
          strcpy(meta.refcal[k].mtl_band, "7"); meta.refcal[k].rsr_band = kk++; par.use.refbands[k] = true;  _B_.swir2 = k++;
          strcpy(meta.btcal.mtl_band, "6");
        } else {
          printf("unknown sensor ID! "); return FAILURE;
        }

        _NO_ = bsum(par.use.refbands, _NB_);

        #ifdef DEBUG
        printf("%d input bands, %d output bands.\n", _NB_, _NO_);
        #endif

      // file names
      } else if (strstr(tag, "FILE_NAME") != NULL){
        for (b=0; b<_NB_; b++) parse_metaband(tag, tokenptr, &meta.refcal[b], 0);
        parse_metaband(tag, tokenptr, &meta.btcal,  0);

      // product type
      } else if (strcmp(tag, "DATA_TYPE") == 0 || 
                 strcmp(tag, "PRODUCT_TYPE") == 0){
        if (strstr(tokenptr, "L1T")  != NULL || 
            strstr(tokenptr, "L1TP") != NULL){
          meta.tier = 1;
        } else meta.tier = 2;

      // tier level
      } else if (strcmp(tag, "COLLECTION_CATEGORY") == 0){
        if (strstr(tokenptr, "T2") != NULL) meta.tier = 2;
        if (strstr(tokenptr, "RT") != NULL) meta.tier = 3;

      // dimension variables
      } else if (strcmp(tag, "PRODUCT_SAMPLES_REF") == 0 || 
                 strcmp(tag, "REFLECTIVE_SAMPLES") == 0){
        _NX_ = atoi(tokenptr);
      } else if (strcmp(tag, "PRODUCT_LINES_REF") == 0 || 
                 strcmp(tag, "REFLECTIVE_LINES") == 0){
        _NY_ = atoi(tokenptr);

      // resolution variables
      } else if (strcmp(tag, "GRID_CELL_SIZE_REF") == 0 || 
                 strcmp(tag, "GRID_CELL_SIZE_REFLECTIVE") == 0){
        meta.res = atoi(tokenptr);

      // bounding box variables: map
      } else if (strcmp(tag, "PRODUCT_UL_CORNER_MAPX") == 0 || 
          strcmp(tag, "CORNER_UL_PROJECTION_X_PRODUCT") == 0){
        meta.map_ul.x = atof(tokenptr)-15;
      } else if (strcmp(tag, "PRODUCT_UL_CORNER_MAPY") == 0 || 
          strcmp(tag, "CORNER_UL_PROJECTION_Y_PRODUCT") == 0){
        meta.map_ul.y = atof(tokenptr)+15;

      // acquisition variables
      } else if (strcmp(tag, "WRS_PATH") == 0){
        path = atoi(tokenptr);
      } else if (strcmp(tag, "WRS_ROW") == 0 ||
                 strcmp(tag, "STARTING_ROW") == 0){
        row = atoi(tokenptr);
      } else if (strcmp(tag, "ACQUISITION_DATE") == 0 ||
                 strcmp(tag, "DATE_ACQUIRED") == 0){
        tokenptr2 = strtok(tokenptr, separator2);
        meta.date.year = atoi(tokenptr2);
        tokenptr2 = strtok(NULL, separator2);
        meta.date.month = atoi(tokenptr2);
        tokenptr2 = strtok(NULL, separator2);
        meta.date.day = atoi(tokenptr2);
        meta.date.doy = md2doy(meta.date.month, meta.date.day);
      } else if (strcmp(tag, "SCENE_CENTER_TIME") == 0 ||
                 strcmp(tag, "SCENE_CENTER_SCAN_TIME") == 0){
        meta.date.hh = atoi(tokenptr);
        tokenptr = strtok(NULL, separator);
        meta.date.mm = atoi(tokenptr);
        tokenptr = strtok(NULL, separator);
        meta.date.ss = atoi(tokenptr);
        meta.date.tz = 0;

      // calibration variables
      } else if (strstr(tag, "Q") == NULL &&
                (strstr(tag, "LMAX_") != NULL ||
                 strstr(tag, "RADIANCE_MAXIMUM_") != NULL)){
        for (b=0; b<_NB_; b++) parse_metaband(tag, tokenptr, &meta.refcal[b], 1);
        parse_metaband(tag, tokenptr, &meta.btcal,  1);
      } else if (strstr(tag, "Q") == NULL &&
                (strstr(tag, "LMIN_") != NULL ||
                 strstr(tag, "RADIANCE_MINIMUM_") != NULL)){
        for (b=0; b<_NB_; b++) parse_metaband(tag, tokenptr, &meta.refcal[b], 2);
        parse_metaband(tag, tokenptr, &meta.btcal,  2);
      } else if (strstr(tag, "QCALMAX_") != NULL ||
                 strstr(tag, "QUANTIZE_CAL_MAX_") != NULL){
        for (b=0; b<_NB_; b++) parse_metaband(tag, tokenptr, &meta.refcal[b], 3);
        parse_metaband(tag, tokenptr, &meta.btcal,  3);
      } else if (strstr(tag, "QCALMIN_") != NULL ||
                 strstr(tag, "QUANTIZE_CAL_MIN_") != NULL){
        for (b=0; b<_NB_; b++) parse_metaband(tag, tokenptr, &meta.refcal[b], 4);
        parse_metaband(tag, tokenptr, &meta.btcal,  4);
      } else if (strstr(tag, "REFLECTANCE_MULT_") != NULL){
        for (b=0; b<_NB_; b++) parse_metaband(tag, tokenptr, &meta.refcal[b], 5);
      } else if (strstr(tag, "REFLECTANCE_ADD_") != NULL){
        for (b=0; b<_NB_; b++) parse_metaband(tag, tokenptr, &meta.refcal[b], 6);
      } else if (strcmp(tag, "K1_CONSTANT_BAND_10") == 0){
        meta.btcal.k1 = atof(tokenptr);
      } else if (strcmp(tag, "K2_CONSTANT_BAND_10") == 0){
        meta.btcal.k2 = atof(tokenptr);
      }

      // in case tag (key words) is not the first word in a line
      tag = tokenptr;

    }

  }

  fclose(fp);

  sprintf(meta.wrs, "%03d%03d", path, row);

  _NC_ = _NX_*_NY_;

  par.svgrid = floor(par.svgrid/meta.res);
  
  meta.scale = 1;

  if (!par.doreproj) par.res = meta.res;
  if (fmod(par.tilesize, par.res) != 0){
    printf("TILE_SIZE must be a multiple of RESOLUTION."); return FAILURE;
  } else par.tilesize_px = floor(par.tilesize/par.res);

  if (meta.id == 8){
    // k1 & k2 shoud be given in MTL
  } else if (meta.id == 7){
    meta.btcal.k1 =  666.09;
    meta.btcal.k2 = 1282.71;
  } else if (meta.id == 5){
    meta.btcal.k1 =  607.76;
    meta.btcal.k2 = 1260.56;
  } else if (meta.id == 4){
    meta.btcal.k1 =  671.62;
    meta.btcal.k2 = 1284.30;
  } else {
    printf("unknown sensor: %d\n", meta.id);
    return FAILURE;
  }

  if (E0() == FAILURE){ printf("error in getting E0. "); return FAILURE;}

  #ifdef DEBUG
  mtlprint();
  #endif

  if (mtlcheck() != SUCCESS){
    printf("Unable to read parameter from MTL file!\n"); exit(1);}

  return SUCCESS;
}


/** Parse Sentinel-2 metadata
+++ This function reads the Sentinel-2 metadata
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int parseMetadataSentinel(){
FILE *fp = NULL;
char *fname = NULL;
char  buffer[BIG_STR_LEN] = "\0";
char *tag = NULL;
char *tokenptr = NULL;
char *tokenptr2 = NULL;
char *tokenptr3 = NULL;
char *separator = " =\"<>";
char *separator2 = "-";
char *separator3 = "-T:.";
char *d_top_1 = NULL;
char *d_top_2 = NULL;
int b, k, kk;
int i, j;
int **k_vz = NULL, **k_va = NULL;
bool s = false, v = false, z = false, a = false;

char d_img[MAX_STR_LEN];
char id_img[MAX_STR_LEN];
char *b_img = NULL;

  mtlinit();

  meta.dtype = 16;
  
  par.band_bt  = false;

  /** parse top-level xml **/
  d_top_1 = directoryname(par.d_img);
  d_top_2 = directoryname(d_top_1);

  // scan directory for xml file
  if ((fname = findfile_filter(d_top_2, "S2A", ".xml")) == NULL &&
      (fname = findfile_filter(d_top_2, "MTD", ".xml")) == NULL){
    printf("Finding top-level metadata file failed. "); return FAILURE;
  }
  free((void*)d_top_1); free((void*)d_top_2);


  // open xml
  if ((fp = fopen(fname, "r")) == NULL){
    printf("Unable to open metadata file. "); return FAILURE;
  } else free((void*)fname);

  // process line by line
  while (fgets(buffer, BIG_STR_LEN, fp) != NULL){

    // get tag
    tokenptr = strtok(buffer, separator);
    tag=tokenptr;

    // extract parameters by comparing tag
    while (tokenptr != NULL){

      tokenptr = strtok(NULL, separator);


      // Sentinel ID
      if (strcmp(tag, "SPACECRAFT_NAME") == 0){

        tokenptr2 = strtok(tokenptr, separator2);
        tokenptr2 = strtok(NULL, separator2);

        if ((meta.id = atoi(tokenptr2+0)) != 2){
          printf("unknown/unsupported sensor ID! "); return FAILURE;
        }

        sprintf(meta.satellite, "SEN%s", tokenptr2);

        _NB_ = 13;

        alloc((void**)&meta.refcal, _NB_, sizeof(cal_t));
        for (b=0; b<_NB_; b++) calinit(&meta.refcal[b]);
        alloc((void**)&par.use.refbands, _NB_, sizeof(bool));

        k = 0; 
        if (strcmp(meta.satellite, "SEN2A") == 0){
          kk = 26;
        } else if (strcmp(meta.satellite, "SEN2B") == 0){
          kk = 39;
        } else {
          printf("unknown/unsupported Sentinel-2! "); return FAILURE;
        }
        strcpy(meta.refcal[k].mtl_band, "01"); meta.refcal[k].rsr_band = kk++; par.use.refbands[k] = false; _B_.ultrablue = k++;
        strcpy(meta.refcal[k].mtl_band, "02"); meta.refcal[k].rsr_band = kk++; par.use.refbands[k] = true;  _B_.blue      = k++;
        strcpy(meta.refcal[k].mtl_band, "03"); meta.refcal[k].rsr_band = kk++; par.use.refbands[k] = true;  _B_.green     = k++;
        strcpy(meta.refcal[k].mtl_band, "04"); meta.refcal[k].rsr_band = kk++; par.use.refbands[k] = true;  _B_.red       = k++;
        strcpy(meta.refcal[k].mtl_band, "05"); meta.refcal[k].rsr_band = kk++; par.use.refbands[k] = true;  _B_.rededge1  = k++;
        strcpy(meta.refcal[k].mtl_band, "06"); meta.refcal[k].rsr_band = kk++; par.use.refbands[k] = true;  _B_.rededge2  = k++;
        strcpy(meta.refcal[k].mtl_band, "07"); meta.refcal[k].rsr_band = kk++; par.use.refbands[k] = true;  _B_.rededge3  = k++;
        strcpy(meta.refcal[k].mtl_band, "08"); meta.refcal[k].rsr_band = kk++; par.use.refbands[k] = true;  _B_.bnir      = k++;
        strcpy(meta.refcal[k].mtl_band, "8A"); meta.refcal[k].rsr_band = kk++; par.use.refbands[k] = true;  _B_.nir       = k++;
        strcpy(meta.refcal[k].mtl_band, "09"); meta.refcal[k].rsr_band = kk++; par.use.refbands[k] = false; _B_.w1        = k++;
        strcpy(meta.refcal[k].mtl_band, "10"); meta.refcal[k].rsr_band = kk++; par.use.refbands[k] = false; _B_.cir       = k++;
        strcpy(meta.refcal[k].mtl_band, "11"); meta.refcal[k].rsr_band = kk++; par.use.refbands[k] = true;  _B_.swir1     = k++;
        strcpy(meta.refcal[k].mtl_band, "12"); meta.refcal[k].rsr_band = kk++; par.use.refbands[k] = true;  _B_.swir2     = k++;

        _B_.w0 = _B_.nir;

        sprintf(d_img, "%s/IMG_DATA", par.d_img);

        for (b=0; b<_NB_; b++){

          if (par.use.refbands[b]) _NO_++;

          sprintf(id_img, "_B%s.jp2", meta.refcal[b].mtl_band);

          if ((b_img = findfile(d_img, id_img)) == NULL){
            printf("Unable to find image. "); return FAILURE;
          } else {
            strcpy(meta.refcal[b].fname, b_img);
            free((void*)b_img);
          }
        }
        
        #ifdef DEBUG
        printf("%d input bands, %d output bands.\n", _NB_, _NO_);
        #endif
        

        if (strlen(par.b_img) > 50){ // old, long naming convention
          strncpy(meta.wrs, par.b_img+49, 6);
        } else { // new, compact naming convention
          strncpy(meta.wrs, par.b_img+4, 6);
        }
        meta.wrs[6] = '\0';

      // product type
      } else if (strcmp(tag, "PROCESSING_LEVEL") == 0){
        if (strstr(tokenptr, "1C") == NULL){
          printf("Unknown processing level. "); return FAILURE;
        } else {
          meta.tier = 1;
        }

      // acquisition variables
      } else if (strcmp(tag, "PRODUCT_START_TIME") == 0){
        tokenptr3 = strtok(tokenptr, separator3); 
        meta.date.year = atoi(tokenptr3);
        tokenptr3 = strtok(NULL, separator3);
        meta.date.month = atoi(tokenptr3);
        tokenptr3 = strtok(NULL, separator3);
        meta.date.day = atoi(tokenptr3);
        meta.date.doy = md2doy(meta.date.month, meta.date.day);
        tokenptr3 = strtok(NULL, separator3);
        meta.date.hh = atoi(tokenptr3);
        tokenptr3 = strtok(NULL, separator3);
        meta.date.mm = atoi(tokenptr3);
        tokenptr3 = strtok(NULL, separator3);
        meta.date.ss = atoi(tokenptr3);
        meta.date.tz = 0;
 
      // scaling factor
      } else if (strcmp(tag, "QUANTIFICATION_VALUE") == 0){
        meta.scale = atoi(tokenptr);
        while (strcmp(tokenptr, "/QUANTIFICATION_VALUE") != 0){
          if (atoi(tokenptr) != 0){
            meta.scale = atoi(tokenptr);
          }
          tag = tokenptr;
          tokenptr = strtok(NULL, separator);
        }

      }

      // in case tag (key words) is not the first word in a line
      tag = tokenptr;

    }

  }

  fclose(fp);


  /** parse granule xml **/

  // scan directory for xml file
  if ((fname = findfile_filter(par.d_img, "S2A", ".xml")) == NULL &&
      (fname = findfile_filter(par.d_img, "MTD", ".xml")) == NULL){
    printf("Finding granule metadata file failed. "); return FAILURE;
  }


  // open xml
  if ((fp = fopen(fname, "r")) == NULL){
    printf("Unable to open metadata file. "); return FAILURE;
  } else free((void*)fname);

  // process line by line
  while (fgets(buffer, BIG_STR_LEN, fp) != NULL){

    // get tag
    tokenptr = strtok(buffer, separator);
    tag=tokenptr;

    // extract parameters by comparing tag
    while (tokenptr != NULL){

      tokenptr = strtok(NULL, separator);

      // nx/ny/res of highest resolution bands
      if (strcmp(tag, "resolution") == 0){
        if (meta.res < 0 || atoi(tokenptr) < meta.res) meta.res = atoi(tokenptr);
      } else if (strcmp(tag, "NROWS") == 0){
        if (atoi(tokenptr) > _NY_) _NY_ = atoi(tokenptr);
      } else if (strcmp(tag, "NCOLS") == 0){
        if (atoi(tokenptr) > _NX_) _NX_ = atoi(tokenptr);
      } else if (strcmp(tag, "ULX") == 0){
        meta.map_ul.x = atof(tokenptr);
      } else if (strcmp(tag, "ULY") == 0){
        meta.map_ul.y = atof(tokenptr);

      // sun/view grids
      } else if (strcmp(tag, "COL_STEP") == 0){
        while (atoi(tokenptr) == 0) tokenptr = strtok(NULL, separator);
        par.svgrid = atoi(tokenptr);
        meta.s2.nx = ceil(_NX_*meta.res/(float)par.svgrid);
        meta.s2.ny = ceil(_NY_*meta.res/(float)par.svgrid);
        if (meta.s2.szen == NULL) alloc_2D((void***)&meta.s2.szen, meta.s2.ny, meta.s2.nx, sizeof(float));
        if (meta.s2.vzen == NULL) alloc_2D((void***)&meta.s2.vzen, meta.s2.ny, meta.s2.nx, sizeof(float));
        if (meta.s2.sazi == NULL) alloc_2D((void***)&meta.s2.sazi, meta.s2.ny, meta.s2.nx, sizeof(float));
        if (meta.s2.vazi == NULL) alloc_2D((void***)&meta.s2.vazi, meta.s2.ny, meta.s2.nx, sizeof(float));
        if (k_vz == NULL) alloc_2D((void***)&k_vz, meta.s2.ny, meta.s2.nx, sizeof(int));
        if (k_va == NULL) alloc_2D((void***)&k_va, meta.s2.ny, meta.s2.nx, sizeof(int));
      } else if (strcmp(tag, "Sun_Angles_Grid") == 0){
        s = true;
      } else if (strcmp(tag, "/Sun_Angles_Grid") == 0){
        s = false;
      } else if (strcmp(tag, "Viewing_Incidence_Angles_Grids") == 0){
        v = true;
      } else if (strcmp(tag, "/Viewing_Incidence_Angles_Grids") == 0){
        v = false;
      } else if (strcmp(tag, "Zenith") == 0){
        z = true; i = 0;
      } else if (strcmp(tag, "/Zenith") == 0){
        z = false;
      } else if (strcmp(tag, "Azimuth") == 0){
        a = true; i = 0;
      } else if (strcmp(tag, "/Azimuth") == 0){
        a = false;
      } else if (strcmp(tag, "VALUES") == 0 && i < meta.s2.ny){

        j = 0;

        while (strcmp(tokenptr, "/VALUES") != 0 && j < meta.s2.nx){

          if (s){
            if (z){
              if (strcmp(tokenptr, "NaN") !=0) meta.s2.szen[i][j] = atof(tokenptr);
            } else if (a){
              if (strcmp(tokenptr, "NaN") !=0) meta.s2.sazi[i][j] = atof(tokenptr);
            }
          } else if (v){
            if (z){
              if (strcmp(tokenptr, "NaN") !=0){
                meta.s2.vzen[i][j] += atof(tokenptr);
                k_vz[i][j]++;
              }
            } else if (a){
              if (strcmp(tokenptr, "NaN") !=0){
                meta.s2.vazi[i][j] += atof(tokenptr);
                k_va[i][j]++;
              }
            }
          }

          tokenptr = strtok(NULL, separator);
          j++;

        }

        i++;

      }

      // in case tag (key words) is not the first word in a line
      tag = tokenptr;
    }

  }

  fclose(fp);


  // average of view angles
  for (i=0; i<meta.s2.ny; i++){
  for (j=0; j<meta.s2.nx; j++){
    if (k_vz[i][j] > 0){
      meta.s2.vzen[i][j] /= k_vz[i][j];
    } else {
      meta.s2.vzen[i][j] = _NODATA_;
    }
    if (k_va[i][j] > 0){
      meta.s2.vazi[i][j] /= k_va[i][j];
    } else {
      meta.s2.vazi[i][j] = _NODATA_;
    }
  }
  }

  free_2D((void**)k_vz, meta.s2.ny);
  free_2D((void**)k_va, meta.s2.ny);


  _NC_ = _NX_*_NY_;

  par.svgrid = floor(par.svgrid/meta.res);

  if (!par.doreproj) par.res = meta.res;
  if (fmod(par.tilesize, par.res) != 0){
    printf("TILE_SIZE must be a multiple of RESOLUTION."); return FAILURE;
  } else par.tilesize_px = floor(par.tilesize/par.res);

  if (E0() == FAILURE){ printf("error in getting E0. "); return FAILURE;}

  #ifdef DEBUG
  mtlprint();
  #endif

  #ifdef DEV
  printf("\nwarning: mtlcheck disabled.\n");
  #endif
//  if (mtlcheck() != SUCCESS){
//    printf("Unable to read parameter from MTL file!\n"); exit(1);}


  return SUCCESS;
}


/** Parse band-specific values of Landsat metadata
+++ This function reads band-specific Landsat metadata like filenames and
+++ calibration parameters
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void parse_metaband(char *tag, char *value, cal_t *cal, int type){
char str1[MAX_STR_LEN], add1[MAX_STR_LEN];
char str2[MAX_STR_LEN], add2[MAX_STR_LEN];
int fill = -32767;

  //if (meta.id == 7 && cal->mtl_band == 6){
  if (meta.id == 7 && strcmp(cal->mtl_band, "6") == 0){
    strcpy(add1, "_VCID_1"); strcpy(add2, "1");
  } else {
    strcpy(add1, ""); strcpy(add2, "");
  }

  if (type == 0){
    if (strcmp(cal->fname, "NULL") == 0){
      sprintf(str1, "FILE_NAME_BAND_%s%s", cal->mtl_band, add1);
      sprintf(str2, "BAND%s%s_FILE_NAME",  cal->mtl_band, add2);
      if (strcmp(tag, str1) == 0 || strcmp(tag, str2) == 0){
        sprintf((cal->fname), "%s/%s", par.d_img, value);
      }
    }
  } else if (type == 1){
    if (cal->lmax ==  fill){
      sprintf(str1, "RADIANCE_MAXIMUM_BAND_%s%s", cal->mtl_band, add1);
      sprintf(str2, "LMAX_BAND%s%s",              cal->mtl_band, add2);
      if (strcmp(tag, str1) == 0 || strcmp(tag, str2) == 0){
        cal->lmax = atof(value);
      }
    }
  } else if (type == 2){
    if (cal->lmin == fill){
      sprintf(str1, "RADIANCE_MINIMUM_BAND_%s%s", cal->mtl_band, add1);
      sprintf(str2, "LMIN_BAND%s%s",              cal->mtl_band, add2);
      if (strcmp(tag, str1) == 0 || strcmp(tag, str2) == 0){
        cal->lmin = atof(value);
      }
    }
  } else if (type == 3){
    if (cal->qmax == fill){
      sprintf(str1, "QUANTIZE_CAL_MAX_BAND_%s%s", cal->mtl_band, add1);
      sprintf(str2, "QCALMAX_BAND%s%s",           cal->mtl_band, add2);
      if (strcmp(tag, str1) == 0 || strcmp(tag, str2) == 0){
        cal->qmax = atof(value);
      }
    }
  } else if (type == 4){
    if (cal->qmin == fill){
      sprintf(str1, "QUANTIZE_CAL_MIN_BAND_%s%s", cal->mtl_band, add1);
      sprintf(str2, "QCALMIN_BAND%s%s",           cal->mtl_band, add2);
      if (strcmp(tag, str1) == 0 || strcmp(tag, str2) == 0){
        cal->qmin = atof(value);
      }
    }
  } else if (type == 5){
    if (cal->rmul == fill){
      sprintf(str1, "REFLECTANCE_MULT_BAND_%s%s", cal->mtl_band, add1);
      //sprintf(str2, "QCALMAX_BAND%s%s",           cal->mtl_band, add2);
      if (strcmp(tag, str1) == 0){// || strcmp(tag, str2) == 0){
        cal->rmul = atof(value);
      }
    }
  } else if (type == 6){
    if (cal->radd == fill){
      sprintf(str1, "REFLECTANCE_ADD_BAND_%s%s", cal->mtl_band, add1);
      //sprintf(str2, "QCALMIN_BAND%s%s",           cal->mtl_band, add2);
      if (strcmp(tag, str1) == 0){// || strcmp(tag, str2) == 0){
        cal->radd = atof(value);
      }
    }
  }

  return;
}


/** Identify satellite mission
+++ This function identifies the satellite mission, i.e. Landsat or Senti-
+++ nel-2
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int parseMission(){
char *fname = NULL;

  if ((fname = findfile(par.d_img, "MTL.txt")) != NULL){
    meta.mission = MISSION_LANDSAT;
  } else if ((fname = findfile_filter(par.d_img, "S2A", ".xml")) != NULL ||
             (fname = findfile_filter(par.d_img, "MTD", ".xml")) != NULL){
    meta.mission = MISSION_SENTINEL2;
  } else {
    printf("unknown Satellite Mission. "); return FAILURE;
  }

  free((void*)fname);

  #ifdef DEBUG
  printf("\nMission: %d\n", meta.mission);
  #endif

  return SUCCESS;
}


/** Initialize names bands
+++ This function initializes the named bands to a fill value
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void init_bands(){

  _B_.blue = _B_.green = _B_.red = -1;
  _B_.rededge1 = _B_.rededge2 = _B_.rededge3 = -1;
  _B_.nir = _B_.swir1 = _B_.swir2 = -1;
  _B_.bnir = -1;
  _B_.ultrablue = _B_.cir = -1;
  _B_.w0 = _B_.w1 = -1;
  _B_.ref = _B_.qai = _B_.cld = -1;
  _B_.hot = _B_.aod = _B_.wvp = _B_.vzn = -1;
  _B_.bap = _B_.inf = _B_.scr = _B_.num = -1;
  _B_.avg = _B_.std = _B_.min = _B_.max = -1;
  _B_.rng = _B_.skw = _B_.krt = _B_.q25 = -1;
  _B_.q50 = _B_.q75 = _B_.iqr = -1;
  _B_.tss = _B_.rms = _B_.sta = _B_.tsi = -1;
  _B_.fby = _B_.fbm = _B_.fbw = _B_.fbd = -1;
  _B_.trd = _B_.cat = -1;
  _B_.dem = _B_.dss = _B_.dri = _B_.dps = -1;
  _B_.dfi = _B_.des = _B_.dlm = _B_.lts = -1;
  _B_.lgs = _B_.vem = _B_.vss = _B_.vri = -1;
  _B_.vps = _B_.vfi = _B_.ves = _B_.vlm = -1;
  _B_.vbl = _B_.vsa = _B_.ist = _B_.ibl = -1;
  _B_.ibt = _B_.igs = _B_.rar = _B_.raf = -1;
  _B_.rmr = _B_.rmf = -1;

  return;
}


/** Parse metadata
+++ This function reads the metadata
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int parseMetadata(){
OGRSpatialReference css;
char *wkt;

  css.SetWellKnownGeogCS("WGS84");
  css.exportToWkt(&wkt);
  strcpy(meta.css_geo, wkt);
  free((void*)wkt);

  init_bands();

  if (parseMission() != SUCCESS) return FAILURE;

  switch (meta.mission){
    case MISSION_LANDSAT:
      if (parseMetadataLandsat()  != SUCCESS) return FAILURE;
      break;
    case MISSION_SENTINEL2:
      if (parseMetadataSentinel() != SUCCESS) return FAILURE;
      break;
    default:
      printf("unknown mission. ");
      return FAILURE;
  }

  // base name
  sprintf(par.b_out, "%04d%02d%02d_LEVEL2_%s",
      meta.date.year, meta.date.month, 
      meta.date.day, meta.satellite);

  return SUCCESS;
}

