/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
This file contains functions that handle standard output operations
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#include "force.h"


/** Initialize memory-to-output mapping struct
+++ This function initializes the memory-to-output mapping struct. This has
+++ to be done early on in any higher level FORCE programs.
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int mem_to_disc_dict_init(){
int l;

  for (l=0; l<MAX_STR_LEN; l++){
    par.m2d[l].save       = false;
    par.m2d[l].nodata     = _NODATA_;
    par.m2d[l].start      = 0;
    par.m2d[l].stop       = 0;
    par.m2d[l].length     = 0;
    par.m2d[l].product[0] = '\0';
    par.m2d[l].fname[0]   = '\0';
    par.m2d[l].data       = NULL;
    par.m2d[l].scale      = NULL;
    par.m2d[l].ts         = NULL;
    par.m2d[l].date       = NULL;
    par.m2d[l].bn         = NULL;
    par.m2d[l].sid        = NULL;
  }
  
  return SUCCESS;
}


/** Free and re-initialize memory-to-output mapping struct
+++ This function frees all the memory that was allocated in the memory-
+++ to-output mapping struct. And it re-initializes the struct for repea-
+++ ted use.
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int mem_to_disc_free(){
int l;

  for (l=0; l<MAX_STR_LEN; l++){

    if (par.m2d[l].data != NULL){
      free_2D((void**)par.m2d[l].data, par.m2d[l].length);
      par.m2d[l].data = NULL;
    }
    if (par.m2d[l].scale != NULL){
      free((void*)par.m2d[l].scale);
      par.m2d[l].scale = NULL;
    }
    if (par.m2d[l].ts != NULL){
      free((void*)par.m2d[l].ts);
      par.m2d[l].ts = NULL;
    }
    if (par.m2d[l].date  != NULL){
      free((void*)par.m2d[l].date);
      par.m2d[l].date = NULL;
    }
    if (par.m2d[l].bn != NULL){
      free_2D((void**)par.m2d[l].bn, par.m2d[l].length);
      par.m2d[l].bn = NULL;
    }
    if (par.m2d[l].sid != NULL){
      free_2D((void**)par.m2d[l].sid, par.m2d[l].length);
      par.m2d[l].sid = NULL;
    }

    par.m2d[l].save       = false;
    par.m2d[l].nodata     = _NODATA_;
    par.m2d[l].start      = 0;
    par.m2d[l].stop       = 0;
    par.m2d[l].length     = 0;
    par.m2d[l].product[0] = '\0';
    par.m2d[l].fname[0]   = '\0';

  }

  return SUCCESS;
}


/** Level 2 memory to output dictionary
+++ This function maps the requested Level 2 products in the data stack.
+++ This is done for preparing output.
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int mem_to_disc_dict_l2(){
int l, b, ob;


  // order is mostly irrelevant, 
  // BUT ref and qai need to be the first and last ones
  _B_.ref = _NP_++; _B_.cld = _NP_++; _B_.hot = _NP_++; _B_.vzn = _NP_++; 
  _B_.aod = _NP_++; _B_.wvp = _NP_++; _B_.qai = _NP_++; 

  // output?
  par.m2d[_B_.ref].save = par.oref; par.m2d[_B_.cld].save = par.ocld;
  par.m2d[_B_.hot].save = par.ohot; par.m2d[_B_.vzn].save = par.ovzn;
  par.m2d[_B_.aod].save = par.oaod; par.m2d[_B_.wvp].save = par.owvp;
  par.m2d[_B_.qai].save = par.oqai;

  // product names
  if (par.doatmo){
    strcpy(par.m2d[_B_.ref].product, "BOA"); 
  } else {
    strcpy(par.m2d[_B_.ref].product, "TOA");
  }
  strcpy(par.m2d[_B_.cld].product, "CLD");
  strcpy(par.m2d[_B_.hot].product, "HOT");
  strcpy(par.m2d[_B_.vzn].product, "VZN");
  strcpy(par.m2d[_B_.aod].product, "AOD");
  strcpy(par.m2d[_B_.wvp].product, "WVP");
  strcpy(par.m2d[_B_.qai].product, "QAI");

  // number of layers
  par.m2d[_B_.ref].length = bsum(par.use.refbands, _NB_);
  par.m2d[_B_.cld].length = 1; par.m2d[_B_.hot].length = 1;
  par.m2d[_B_.vzn].length = 1; par.m2d[_B_.aod].length = 1;
  par.m2d[_B_.wvp].length = 1; par.m2d[_B_.qai].length = 1;

  
  // init band names & wavelengths
  for (l=0; l<_NP_; l++){
    par.m2d[l].bn = NULL;
    par.m2d[l].ts = NULL;
    par.m2d[l].scale  = NULL;
  }

  // allocate band names
  for (l=0; l<_NP_; l++){
    alloc_2D((void***)&par.m2d[l].bn, par.m2d[l].length, MAX_STR_LEN, sizeof(char));
  }

  // allocate scale factors
  for (l=0; l<_NP_; l++){
    alloc((void**)&par.m2d[l].scale, par.m2d[l].length, sizeof(int2s));
  }
  
  // allocate wavelengths of reflectance product
  alloc((void**)&par.m2d[_B_.ref].ts, par.m2d[_B_.ref].length, sizeof(float));

  // band names + wavelengths of reflectance product
  for (b=0, ob=0; b<_NB_; b++){
    if (!par.use.refbands[b]) continue;
    if (b == _B_.blue)     sprintf(par.m2d[_B_.ref].bn[ob], "BLUE %s reflect. (%02d)", par.m2d[_B_.ref].product, ob+1);
    if (b == _B_.green)    sprintf(par.m2d[_B_.ref].bn[ob], "GREEN %s reflect. (%02d)", par.m2d[_B_.ref].product, ob+1);
    if (b == _B_.red)      sprintf(par.m2d[_B_.ref].bn[ob], "RED %s reflect. (%02d)", par.m2d[_B_.ref].product, ob+1);
    if (b == _B_.rededge1) sprintf(par.m2d[_B_.ref].bn[ob], "RE1 %s reflect. (%02d)", par.m2d[_B_.ref].product, ob+1);
    if (b == _B_.rededge2) sprintf(par.m2d[_B_.ref].bn[ob], "RE2 %s reflect. (%02d)", par.m2d[_B_.ref].product, ob+1);
    if (b == _B_.rededge3) sprintf(par.m2d[_B_.ref].bn[ob], "RE3 %s reflect. (%02d)", par.m2d[_B_.ref].product, ob+1);
    if (b == _B_.bnir)     sprintf(par.m2d[_B_.ref].bn[ob], "BNIR %s reflect. (%02d)", par.m2d[_B_.ref].product, ob+1);
    if (b == _B_.nir)      sprintf(par.m2d[_B_.ref].bn[ob], "NIR %s reflect. (%02d)", par.m2d[_B_.ref].product, ob+1);
    if (b == _B_.swir1)    sprintf(par.m2d[_B_.ref].bn[ob], "SWIR1 %s reflect. (%02d)", par.m2d[_B_.ref].product, ob+1);
    if (b == _B_.swir2)    sprintf(par.m2d[_B_.ref].bn[ob], "SWIR2 %s reflect. (%02d)", par.m2d[_B_.ref].product, ob+1);
    par.m2d[_B_.ref].ts[ob] = atc.wvl[b];
    ob++;
  }


  // band names of other products
  sprintf(par.m2d[_B_.cld].bn[0], "Cloud/shadow distance");
  sprintf(par.m2d[_B_.hot].bn[0], "Haze optimized transform");
  sprintf(par.m2d[_B_.vzn].bn[0], "View zenith");
  sprintf(par.m2d[_B_.aod].bn[0], "Aerosol optical depth");
  sprintf(par.m2d[_B_.wvp].bn[0], "Water vapor");
  sprintf(par.m2d[_B_.qai].bn[0], "Quality assurance information");
  
  // scale factors
  for (b=0; b<par.m2d[_B_.ref].length; b++) par.m2d[_B_.ref].scale[b] = 10000;
  par.m2d[_B_.cld].scale[0] = 1;    par.m2d[_B_.hot].scale[0] = 10000;
  par.m2d[_B_.vzn].scale[0] = 1000; par.m2d[_B_.aod].scale[0] = 1000;
  par.m2d[_B_.wvp].scale[0] = 1000; par.m2d[_B_.qai].scale[0] = 1;

  // nodata value
  par.m2d[_B_.ref].nodata = par.m2d[_B_.cld].nodata = _NODATA_;
  par.m2d[_B_.hot].nodata = par.m2d[_B_.vzn].nodata = _NODATA_;
  par.m2d[_B_.aod].nodata = par.m2d[_B_.wvp].nodata = _NODATA_;
  par.m2d[_B_.qai].nodata = 1;

  // start / stop band in data array
  for (l=0; l<_NP_; l++){
    if (par.m2d[l].save){
      par.m2d[l].start = _ND_;
      par.m2d[l].stop  = par.m2d[l].start+par.m2d[l].length;
      _ND_ += par.m2d[l].length;
    } else {
      par.m2d[l].start = _ND_;
      par.m2d[l].stop  = _ND_;
    }
  }

  #ifdef DEBUG
  printf("\nmemory-to-output:\n");
  for (l=0; l<_NP_; l++){
    printf("  %s: save %d, start %d, stop %d, length %d\n", par.m2d[l].product,
    par.m2d[l].save, par.m2d[l].start, par.m2d[l].stop, par.m2d[l].length);}
  printf("  # of data layers: %d\n", _ND_);
  #endif


  return SUCCESS;
}


/** Level 3 memory to output dictionary
+++ This function maps the requested Level 3 products. This is done for 
+++ preparing output.
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int mem_to_disc_dict_l3(){
int l, b;
char wavename[MAX_STR_LEN][MAX_STR_LEN];

  _B_.bap = _NP_++; _B_.inf = _NP_++; _B_.scr = _NP_++; _B_.avg = _NP_++;
  _B_.std = _NP_++; _B_.min = _NP_++; _B_.max = _NP_++; _B_.rng = _NP_++;
  _B_.skw = _NP_++; _B_.krt = _NP_++; _B_.q25 = _NP_++; _B_.q50 = _NP_++; 
  _B_.q75 = _NP_++; _B_.iqr = _NP_++;

  // output?
  par.m2d[_B_.bap].save = par.obap; par.m2d[_B_.inf].save = par.oinf;
  par.m2d[_B_.scr].save = par.oscr; par.m2d[_B_.avg].save = par.oavg;
  par.m2d[_B_.std].save = par.ostd; par.m2d[_B_.min].save = par.omin;
  par.m2d[_B_.max].save = par.omax; par.m2d[_B_.rng].save = par.orng;
  par.m2d[_B_.skw].save = par.oskw; par.m2d[_B_.krt].save = par.okrt;
  par.m2d[_B_.q25].save = par.oq25; par.m2d[_B_.q50].save = par.oq50;
  par.m2d[_B_.q75].save = par.oq75; par.m2d[_B_.iqr].save = par.oiqr;

  // product names
  strcpy(par.m2d[_B_.bap].product, "BAP"); strcpy(par.m2d[_B_.inf].product, "INF");
  strcpy(par.m2d[_B_.scr].product, "SCR"); strcpy(par.m2d[_B_.avg].product, "AVG");
  strcpy(par.m2d[_B_.std].product, "STD"); strcpy(par.m2d[_B_.min].product, "MIN");
  strcpy(par.m2d[_B_.max].product, "MAX"); strcpy(par.m2d[_B_.rng].product, "RNG");
  strcpy(par.m2d[_B_.skw].product, "SKW"); strcpy(par.m2d[_B_.krt].product, "KRT");
  strcpy(par.m2d[_B_.q25].product, "Q25"); strcpy(par.m2d[_B_.q50].product, "Q50");
  strcpy(par.m2d[_B_.q75].product, "Q75"); strcpy(par.m2d[_B_.iqr].product, "IQR"); 

  // number of layers
  for (l=0; l<_NP_; l++) par.m2d[l].length = _NB_;
  par.m2d[_B_.inf].length = 6;
  par.m2d[_B_.scr].length = 7;

  // nodata value
  for (l=0; l<_NP_; l++) par.m2d[l].nodata = _NODATA_;

  // allocate Level 3 data, scale factors & band names
  for (l=0; l<_NP_; l++){
    if (par.m2d[l].save){
      alloc_2D((void***)&par.m2d[l].data, par.m2d[l].length, _NC_, sizeof(int2s));
    } else {
      par.m2d[l].data == NULL;
    }
    alloc_2D((void***)&par.m2d[l].bn, par.m2d[l].length, MAX_STR_LEN, sizeof(char));
    alloc((void**)&par.m2d[l].scale, par.m2d[l].length, sizeof(int2s));
  }

  // set scale factors
  for (l=0; l<_NP_; l++){
    for (b=0; b<par.m2d[l].length; b++) par.m2d[l].scale[b] = 10000;
  }
  for (b=0; b<par.m2d[_B_.inf].length; b++) par.m2d[_B_.inf].scale[b] = 1;
  for (b=0; b<par.m2d[_B_.krt].length; b++) par.m2d[_B_.krt].scale[b] = 10;

  // wavelength names
  for (b=0; b<_NB_; b++){
    if (b == _B_.blue)     strcpy(wavename[b], "BLUE");
    if (b == _B_.green)    strcpy(wavename[b], "GREEN");
    if (b == _B_.red)      strcpy(wavename[b], "RED");
    if (b == _B_.rededge1) strcpy(wavename[b], "RE1");
    if (b == _B_.rededge2) strcpy(wavename[b], "RE2");
    if (b == _B_.rededge3) strcpy(wavename[b], "RE3");
    if (b == _B_.bnir)     strcpy(wavename[b], "BNIR");
    if (b == _B_.nir)      strcpy(wavename[b], "NIR");
    if (b == _B_.swir1)    strcpy(wavename[b], "SWIR1");
    if (b == _B_.swir2)    strcpy(wavename[b], "SWIR2");
  }

  // set band names for reflectance products
  for (b=0; b<par.m2d[_B_.bap].length; b++) sprintf(par.m2d[_B_.bap].bn[b], "best available %s reflect. (%02d)", wavename[b], b+1);
  for (b=0; b<par.m2d[_B_.avg].length; b++) sprintf(par.m2d[_B_.avg].bn[b], "average of %s reflect. (%02d)", wavename[b], b+1);
  for (b=0; b<par.m2d[_B_.std].length; b++) sprintf(par.m2d[_B_.std].bn[b], "stand. dev. of %s reflect. (%02d)", wavename[b], b+1);
  for (b=0; b<par.m2d[_B_.min].length; b++) sprintf(par.m2d[_B_.min].bn[b], "min. of %s reflect. (%02d)", wavename[b], b+1);
  for (b=0; b<par.m2d[_B_.max].length; b++) sprintf(par.m2d[_B_.max].bn[b], "max. of %s reflect. (%02d)", wavename[b], b+1);
  for (b=0; b<par.m2d[_B_.rng].length; b++) sprintf(par.m2d[_B_.rng].bn[b], "range of %s reflect. (%02d)", wavename[b], b+1);
  for (b=0; b<par.m2d[_B_.skw].length; b++) sprintf(par.m2d[_B_.skw].bn[b], "skewness of %s reflect. (%02d)", wavename[b], b+1);
  for (b=0; b<par.m2d[_B_.krt].length; b++) sprintf(par.m2d[_B_.krt].bn[b], "kurtosis of %s reflect. (%02d)", wavename[b], b+1);
  for (b=0; b<par.m2d[_B_.q25].length; b++) sprintf(par.m2d[_B_.q25].bn[b], "25%% quantile of %s reflect. (%02d)", wavename[b], b+1);
  for (b=0; b<par.m2d[_B_.q50].length; b++) sprintf(par.m2d[_B_.q50].bn[b], "50%% quantile of %s reflect. (%02d)", wavename[b], b+1);
  for (b=0; b<par.m2d[_B_.q75].length; b++) sprintf(par.m2d[_B_.q75].bn[b], "75%% quantile of %s reflect. (%02d)", wavename[b], b+1);
  for (b=0; b<par.m2d[_B_.iqr].length; b++) sprintf(par.m2d[_B_.iqr].bn[b], "Interquantile range of %s reflect. (%02d)", wavename[b], b+1);

   // set band names for compositing information
  sprintf(par.m2d[_B_.inf].bn[0], "Quality assurance information");
  sprintf(par.m2d[_B_.inf].bn[1], "# of obs.");
  sprintf(par.m2d[_B_.inf].bn[2], "Acquisition DOY");
  sprintf(par.m2d[_B_.inf].bn[3], "Acquisition Year");
  sprintf(par.m2d[_B_.inf].bn[4], "Acquisition - Target DOY");
  sprintf(par.m2d[_B_.inf].bn[5], "Sensor ID");

  // set band names for compositing scores
  sprintf(par.m2d[_B_.scr].bn[0], "Total score");
  sprintf(par.m2d[_B_.scr].bn[1], "DOY score");
  sprintf(par.m2d[_B_.scr].bn[2], "Year score");
  sprintf(par.m2d[_B_.scr].bn[3], "Cloud score");
  sprintf(par.m2d[_B_.scr].bn[4], "Haze score");
  sprintf(par.m2d[_B_.scr].bn[5], "Correlation score");
  sprintf(par.m2d[_B_.scr].bn[6], "View angle score");


  #ifdef DEBUG
  printf("\nmemory-to-output:\n");
  for (l=0; l<_NP_; l++){
    printf("  %s: save %d, length %d\n", par.m2d[l].product,
    par.m2d[l].save,  par.m2d[l].length);}
  #endif


  return SUCCESS;
}


/** Level 4 TSA memory to output dictionary
+++ This function maps the requested Level 4 TSA products. This is done 
+++ for preparing output.
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int mem_to_disc_dict_tsa(){
int l;

  _NP_ = 0;

  _B_.tss = _NP_++; _B_.rms = _NP_++; _B_.sta = _NP_++; _B_.tsi = _NP_++;
  _B_.fby = _NP_++; _B_.fbm = _NP_++; _B_.fbw = _NP_++; _B_.fbd = _NP_++;
  _B_.trd = _NP_++; _B_.cat = _NP_++;
  _B_.dem = _NP_++; _B_.dss = _NP_++; _B_.dri = _NP_++; _B_.dps = _NP_++;
  _B_.dfi = _NP_++; _B_.des = _NP_++; _B_.dlm = _NP_++; _B_.lts = _NP_++;
  _B_.lgs = _NP_++; _B_.vem = _NP_++; _B_.vss = _NP_++; _B_.vri = _NP_++;
  _B_.vps = _NP_++; _B_.vfi = _NP_++; _B_.ves = _NP_++; _B_.vlm = _NP_++;
  _B_.vbl = _NP_++; _B_.vsa = _NP_++; _B_.ist = _NP_++; _B_.ibl = _NP_++;
  _B_.ibt = _NP_++; _B_.igs = _NP_++; _B_.rar = _NP_++; _B_.raf = _NP_++;
  _B_.rmr = _NP_++; _B_.rmf = _NP_++; 


  // output?
  par.m2d[_B_.tss].save = par.otss; par.m2d[_B_.rms].save = par.orms; 
  par.m2d[_B_.sta].save = par.osta; par.m2d[_B_.tsi].save = par.otsi;
  par.m2d[_B_.fby].save = par.ofby; par.m2d[_B_.fbm].save = par.ofbm;
  par.m2d[_B_.fbw].save = par.ofbw; par.m2d[_B_.fbd].save = par.ofbd;
  par.m2d[_B_.trd].save = par.otrd; par.m2d[_B_.cat].save = par.ocat;
  par.m2d[_B_.dem].save = par.olsp; par.m2d[_B_.dss].save = par.olsp; 
  par.m2d[_B_.dri].save = par.olsp; par.m2d[_B_.dps].save = par.olsp; 
  par.m2d[_B_.dfi].save = par.olsp; par.m2d[_B_.des].save = par.olsp; 
  par.m2d[_B_.dlm].save = par.olsp; par.m2d[_B_.lts].save = par.olsp; 
  par.m2d[_B_.lgs].save = par.olsp; par.m2d[_B_.vem].save = par.olsp; 
  par.m2d[_B_.vss].save = par.olsp; par.m2d[_B_.vri].save = par.olsp; 
  par.m2d[_B_.vps].save = par.olsp; par.m2d[_B_.vfi].save = par.olsp; 
  par.m2d[_B_.ves].save = par.olsp; par.m2d[_B_.vlm].save = par.olsp; 
  par.m2d[_B_.vbl].save = par.olsp; par.m2d[_B_.vsa].save = par.olsp; 
  par.m2d[_B_.ist].save = par.olsp; par.m2d[_B_.ibl].save = par.olsp; 
  par.m2d[_B_.ibt].save = par.olsp; par.m2d[_B_.igs].save = par.olsp; 
  par.m2d[_B_.rar].save = par.olsp; par.m2d[_B_.raf].save = par.olsp; 
  par.m2d[_B_.rmr].save = par.olsp; par.m2d[_B_.rmf].save = par.olsp; 


  // product names
  strcpy(par.m2d[_B_.tss].product, "TSS"); strcpy(par.m2d[_B_.rms].product, "RMS"); 
  strcpy(par.m2d[_B_.sta].product, "STA"); strcpy(par.m2d[_B_.tsi].product, "TSI");
  strcpy(par.m2d[_B_.fby].product, "FBY"); strcpy(par.m2d[_B_.fbm].product, "FBM");
  strcpy(par.m2d[_B_.fbw].product, "FBW"); strcpy(par.m2d[_B_.fbd].product, "FBD");
  strcpy(par.m2d[_B_.trd].product, "TRD"); strcpy(par.m2d[_B_.cat].product, "CAT");
  strcpy(par.m2d[_B_.dem].product, "DEM"); strcpy(par.m2d[_B_.dss].product, "DSS"); 
  strcpy(par.m2d[_B_.dri].product, "DRI"); strcpy(par.m2d[_B_.dps].product, "DPS"); 
  strcpy(par.m2d[_B_.dfi].product, "DFI"); strcpy(par.m2d[_B_.des].product, "DES"); 
  strcpy(par.m2d[_B_.dlm].product, "DLM"); strcpy(par.m2d[_B_.lts].product, "LTS"); 
  strcpy(par.m2d[_B_.lgs].product, "LGS"); strcpy(par.m2d[_B_.vem].product, "VEM"); 
  strcpy(par.m2d[_B_.vss].product, "VSS"); strcpy(par.m2d[_B_.vri].product, "VRI"); 
  strcpy(par.m2d[_B_.vps].product, "VPS"); strcpy(par.m2d[_B_.vfi].product, "VFI"); 
  strcpy(par.m2d[_B_.ves].product, "VES"); strcpy(par.m2d[_B_.vlm].product, "VLM"); 
  strcpy(par.m2d[_B_.vbl].product, "VBL"); strcpy(par.m2d[_B_.vsa].product, "VSA"); 
  strcpy(par.m2d[_B_.ist].product, "IST"); strcpy(par.m2d[_B_.ibl].product, "IBL"); 
  strcpy(par.m2d[_B_.ibt].product, "IBT"); strcpy(par.m2d[_B_.igs].product, "IGS"); 
  strcpy(par.m2d[_B_.rar].product, "RAR"); strcpy(par.m2d[_B_.raf].product, "RAF"); 
  strcpy(par.m2d[_B_.rmr].product, "RMR"); strcpy(par.m2d[_B_.rmf].product, "RMF"); 


  // number of layers
  par.m2d[_B_.tss].length = dir.n;  par.m2d[_B_.rms].length = dir.n; 
  par.m2d[_B_.sta].length = 5;      par.m2d[_B_.tsi].length = par.int_n;
  par.m2d[_B_.fby].length = par.yearmax  - par.yearmin  + 1;
  par.m2d[_B_.fbm].length = par.monthmax - par.monthmin + 1;
  par.m2d[_B_.fbw].length = par.weekmax  - par.weekmin  + 1;
  par.m2d[_B_.fbd].length = par.doymax   - par.doymin   + 1;
  par.m2d[_B_.trd].length = 9;      par.m2d[_B_.cat].length = 29;
  par.m2d[_B_.dem].length = par.Yn; par.m2d[_B_.dss].length = par.Yn; 
  par.m2d[_B_.dri].length = par.Yn; par.m2d[_B_.dps].length = par.Yn; 
  par.m2d[_B_.dfi].length = par.Yn; par.m2d[_B_.des].length = par.Yn; 
  par.m2d[_B_.dlm].length = par.Yn; par.m2d[_B_.lts].length = par.Yn; 
  par.m2d[_B_.lgs].length = par.Yn; par.m2d[_B_.vem].length = par.Yn; 
  par.m2d[_B_.vss].length = par.Yn; par.m2d[_B_.vri].length = par.Yn; 
  par.m2d[_B_.vps].length = par.Yn; par.m2d[_B_.vfi].length = par.Yn; 
  par.m2d[_B_.ves].length = par.Yn; par.m2d[_B_.vlm].length = par.Yn; 
  par.m2d[_B_.vbl].length = par.Yn; par.m2d[_B_.vsa].length = par.Yn; 
  par.m2d[_B_.ist].length = par.Yn; par.m2d[_B_.ibl].length = par.Yn; 
  par.m2d[_B_.ibt].length = par.Yn; par.m2d[_B_.igs].length = par.Yn; 
  par.m2d[_B_.rar].length = par.Yn; par.m2d[_B_.raf].length = par.Yn; 
  par.m2d[_B_.rmr].length = par.Yn; par.m2d[_B_.rmf].length = par.Yn; 


  // nodata value
  for (l=0; l<_NP_; l++) par.m2d[l].nodata = _TSAOFF_;


  #ifdef DEBUG
  printf("\nmemory-to-output:\n");
  for (l=0; l<_NP_; l++){
    printf("  %s: save %d, length %d\n", par.m2d[l].product,
    par.m2d[l].save,  par.m2d[l].length);}
  #endif


  return SUCCESS;
}


/** CSO statistics memory to output dictionary
+++ This function maps the requested CSO statistics products. This is done 
+++ for preparing output.
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int mem_to_disc_dict_cso(){
int l;

  _NP_ = 0;

  _B_.num = _NP_++; _B_.avg = _NP_++; _B_.std = _NP_++; _B_.min = _NP_++;
  _B_.max = _NP_++; _B_.rng = _NP_++; _B_.skw = _NP_++; _B_.krt = _NP_++;
  _B_.q25 = _NP_++; _B_.q50 = _NP_++; _B_.q75 = _NP_++; _B_.iqr = _NP_++;


  // output?
  par.m2d[_B_.num].save = par.onum; par.m2d[_B_.avg].save = par.oavg;
  par.m2d[_B_.std].save = par.ostd; par.m2d[_B_.min].save = par.omin;
  par.m2d[_B_.max].save = par.omax; par.m2d[_B_.rng].save = par.orng;
  par.m2d[_B_.skw].save = par.oskw; par.m2d[_B_.krt].save = par.okrt;
  par.m2d[_B_.q25].save = par.oq25; par.m2d[_B_.q50].save = par.oq50;
  par.m2d[_B_.q75].save = par.oq75; par.m2d[_B_.iqr].save = par.oiqr;


  // product names
  strcpy(par.m2d[_B_.num].product, "NUM"); strcpy(par.m2d[_B_.avg].product, "AVG");
  strcpy(par.m2d[_B_.std].product, "STD"); strcpy(par.m2d[_B_.min].product, "MIN");
  strcpy(par.m2d[_B_.max].product, "MAX"); strcpy(par.m2d[_B_.rng].product, "RNG");
  strcpy(par.m2d[_B_.skw].product, "SKW"); strcpy(par.m2d[_B_.krt].product, "KRT");
  strcpy(par.m2d[_B_.q25].product, "Q25"); strcpy(par.m2d[_B_.q50].product, "Q50");
  strcpy(par.m2d[_B_.q75].product, "Q75"); strcpy(par.m2d[_B_.iqr].product, "IQR");


  // number of layers
  for (l=0; l<_NP_; l++) par.m2d[l].length = par.nwin;


  // nodata value
  for (l=0; l<_NP_; l++) par.m2d[l].nodata = 0;


  #ifdef DEBUG
  printf("\nmemory-to-output:\n");
  for (l=0; l<_NP_; l++){
    printf("  %s: save %d, length %d\n", par.m2d[l].product,
    par.m2d[l].save,  par.m2d[l].length);}
  #endif


  return SUCCESS;
}


/** Level 2 ImproPhe memory to output dictionary
+++ This function maps the requested Level 2 ImproPhe products. This is 
+++ done for preparing output.
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int mem_to_disc_dict_il2(){
int l, b;

  _NP_ = 0;

  _B_.ref = _NP_++;

  // output?
  par.m2d[_B_.ref].save = true;

  // product names
  strcpy(par.m2d[_B_.ref].product, "IMP"); 

  // number of layers
  par.m2d[_B_.ref].length = _NB_;
  
  // allocate band names
  alloc_2D((void***)&par.m2d[_B_.ref].bn, 
    par.m2d[_B_.ref].length, MAX_STR_LEN, sizeof(char));

  // allocate scale factors
  alloc((void**)&par.m2d[_B_.ref].scale, 
    par.m2d[_B_.ref].length, sizeof(int2s));
  
  // allocate wavelengths of reflectance product
  alloc((void**)&par.m2d[_B_.ref].ts, 
    par.m2d[_B_.ref].length, sizeof(float));

  // band names + wavelengths of reflectance product
  for (b=0; b<_NB_; b++){
    sprintf(par.m2d[_B_.ref].bn[b], "%s reflect. %02d", par.m2d[_B_.ref].product, b+1);
    par.m2d[_B_.ref].ts[b] = b+1;
  }

  // scale factors
  for (b=0; b<par.m2d[_B_.ref].length; b++) par.m2d[_B_.ref].scale[b] = _RS_;

  // nodata value
  par.m2d[_B_.ref].nodata = _NODATA_;


  #ifdef DEBUG
  printf("\nmemory-to-output:\n");
  for (l=0; l<_NP_; l++){
    printf("  %s: save %d, length %d\n", par.m2d[l].product,
    par.m2d[l].save,  par.m2d[l].length);}
  #endif


  return SUCCESS;
}


/** Initialize Level 3 mosaic
+++ This function initializes the Level 3 mosaics on disc.
--- geotran: geotransformation
+++ Return:  SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int init_products_l3(double geotran[6]){
int l, m, d, b, i, j;
char bname[MAX_STR_LEN];
char fname[MAX_STR_LEN];
char *lock = NULL;
double timeout;
GDALDatasetH fp;
GDALRasterBandH band;
GDALDriverH driver;
char **options = NULL;
int2s *init = NULL;
char ds[MAX_STR_LEN];
char form[MAX_STR_LEN];
char ext[MAX_STR_LEN];


  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  // if single-tile output is requested, exit with SUCCESS
  if (par.dotile) return SUCCESS;

  GDALAllRegister();
  CPLSetConfigOption("GDAL_PAM_ENABLED", "YES");

  // choose between formats
  if (par.format == 0){
    strcpy(form, "ENVI"); strcpy(ext, "dat"); 
  } else if (par.format == 1){
    strcpy(form, "GTiff"); strcpy(ext, "tif"); 
  } else {
    printf("unknown output format. "); return FAILURE;
  }
  driver = GDALGetDriverByName(form);

  // set GeoTiff options
  if (par.format == 1){
    options = CSLSetNameValue(options, "COMPRESS", "LZW");
    options = CSLSetNameValue(options, "PREDICTOR", "2");
    options = CSLSetNameValue(options, "INTERLEAVE", "BAND");
  }

  // get target DOY
  if (par.score_type == SCR_SIG_DES) doy2md(par.Dt[0], &m, &d);
  if (par.score_type == SCR_GAUSS)   doy2md(par.Dt[1], &m, &d);
  if (par.score_type == SCR_SIG_ASC) doy2md(par.Dt[2], &m, &d);

  // allocate one row
  alloc((void**)&init, _NX_*par.tnx, sizeof(int2s));
  for (j=0; j<_NX_*par.tnx; j++) init[j] = -9999;
  
  sprintf(ds, "FORCE v. %s Level 3 processing system", _VERSION_);

  // create output directory
  sprintf(bname, "%s/mosaic", par.d_level3);
  createdir(bname);

  // write empty data for each product
  for (l=0; l<_NP_; l++){

    if (!par.m2d[l].save) continue;

    // filename
    sprintf(fname, "%s/%04d%02d%02d_LEVEL3_%s_%s.%s", 
      bname, par.Yt, m, d, par.l2dict.target, par.m2d[l].product, ext);

    if (_NX_*par.tnx*_NY_*par.tny > 3000000){
      timeout = 60.0*_NX_*par.tnx*_NY_*par.tny/3000000.0; 
    } else { 
      timeout = 60;
    }

    if ((lock = (char*)CPLLockFile(fname, timeout)) == NULL){
      printf("Unable to lock file (timeout: %fs, nx/ny: %d/%d). ", timeout, _NX_*par.tnx, _NY_*par.tny); return FAILURE;}

    // create file
    if ((fp = GDALCreate(driver, fname, _NX_*par.tnx, _NY_*par.tny, par.m2d[l].length, GDT_Int16, options)) == NULL){
      printf("Error creating output file. "); return FAILURE;}

    // write empty data
    for (b=0; b<par.m2d[l].length; b++){
      band = GDALGetRasterBand(fp, b+1);
      for (i=0; i<_NY_*par.tny; i++){
        GDALRasterIO(band, GF_Write, 0, i, _NX_*par.tnx, 1, init, _NX_*par.tnx, 1, GDT_Int16, 0, 0);
      }
    }
    
    // write geometadata
    GDALSetGeoTransform(fp, geotran);
    GDALSetProjection(fp, par.css_dst);

    // in case of ENVI, update description
    if (par.format == 0) GDALSetDescription(fp, "Level 3 data generated by FORCE");


    GDALClose(fp);

    CPLUnlockFile(lock);

  }

  free((void*)init);


  #ifdef CLOCK
  proctime_print("initializing L3 mosaic", TIME);
  #endif

  return SUCCESS;
}


/** Finalize Level 3 mosaic
+++ This function finalizes the Level 3 mosaics on disc.
--- geotran: geotransformation
+++ Return:  SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int final_products_l3(double geotran[6]){
int l, m, d;
char fname[MAX_STR_LEN];
char *lock = NULL;
char aq[MAX_STR_LEN];
int *zlim = NULL;
char **zlab = NULL;
char form[MAX_STR_LEN];
char ext[MAX_STR_LEN];

  // if single-tile output is requested, exit with SUCCESS
  if (par.dotile) return SUCCESS;

  // choose between formats
  if (par.format == 0){
    strcpy(form, "ENVI"); strcpy(ext, "dat"); 
  } else if (par.format == 1){
    strcpy(form, "GTiff"); strcpy(ext, "tif"); 
  } else {
    printf("unknown output format. "); return FAILURE;
  }

  // get target DOY
  if (par.score_type == SCR_SIG_DES) doy2md(par.Dt[0], &m, &d);
  if (par.score_type == SCR_GAUSS)   doy2md(par.Dt[1], &m, &d);
  if (par.score_type == SCR_SIG_ASC) doy2md(par.Dt[2], &m, &d);

  // compile metadata
  sprintf(aq, "%04d-%02d-%02dT12:00:00.0Z", par.Yt, m, d);

  // for every requested output product
  for (l=0; l<_NP_; l++){

    if (!par.m2d[l].save) continue;

    // filename
    sprintf(fname, "%s/mosaic/%04d%02d%02d_LEVEL3_%s_%s.%s", 
      par.d_level3, par.Yt, m, d, par.l2dict.target, par.m2d[l].product, ext);

    // set z-plot metadata (should be fine for most products/indices)
    if (strcmp(par.m2d[l].product, "INF") == 0){

      zlim = NULL;
      zlab = NULL;
      
    } else {

      alloc((void**)&zlim, 2, sizeof(int));
      alloc_2D((void***)&zlab, 2, MAX_STR_LEN, sizeof(char));

      if (strcmp(par.m2d[l].product, "SCR") == 0){
        zlim[0] = 0; zlim[1] = 10000;
        sprintf(zlab[0], "Proxy"); strcpy(zlab[1], "Score");
      } else if (strcmp(par.m2d[l].product, "SKW") == 0){
        zlim[0] = -30000; zlim[1] = 30000;
        sprintf(zlab[0], "Wavelength"); strcpy(zlab[1], "Reflectance");
      } else if (strcmp(par.m2d[l].product, "KRT") == 0){
        zlim[0] = 0; zlim[1] = 1200;
        sprintf(zlab[0], "Wavelength"); strcpy(zlab[1], "Reflectance");
      } else {
        zlim[0] = 0; zlim[1] = 6000;
        sprintf(zlab[0], "Wavelength"); strcpy(zlab[1], "Reflectance");
      }
    }

    if ((lock = (char*)CPLLockFile(fname, 60)) == NULL){
      printf("Unable to lock file. "); return FAILURE;}

    // append/write metadata
    append_envi_header(fname, "Level 3 data generated by FORCE", par.m2d[l].product, 3, par.tnx*_NX_, par.tny*_NY_, par.m2d[l].length, geotran, par.m2d[l].bn, NULL, NULL, aq, par.m2d[l].nodata, -1, -1, -1, -1, -1, -1, -1, par.m2d[l].scale, zlim, zlab, NULL, -1, -1);

    if (zlim != NULL) free((void*)zlim);
    if (zlab != NULL) free_2D((void**)zlab, 2);

    CPLUnlockFile(lock);

  }

  return SUCCESS;
}


/** Write/Update ENVI header
+++ This function writes (if GeoTiff) or appends metadata (if ENVI) to an 
+++ existing ENVI header.
--- fname:   filename of the image (*.dat/*.tif)
--- desc:    description
--- prd:     product type
--- level:   processing level
--- nx:      number of pixels in x-direction
--- ny:      number of pixels in y-direction
--- nb:      number of bands
--- geotran: geotransformation
--- bn:      band names
--- wvl:     wavelengths
--- wvu:     wavelength unit
--- acq:     acquisition date
--- nodata:  nodata value
--- dc:      data cover
--- occ:     opaque cloud cover
--- bcc:     buffered cloud cover
--- ccc:     cirrus cloud cover
--- csc:     cloud shadow cover
--- sc:      snow cover
--- wc:      water cover
--- scale:   scale factor
--- zlim:    z-plot range
--- zlab:    z-plot labels
--- sid:     sensor ID
--- sa:      sun azimuth
--- se:      sun elevation
+++ Return:  SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int append_envi_header(char *fname, char *desc, char *prd, int level, int nx, int ny, int nb, double geotran[6], char **bn, float *wvl, char *wvu, char *acq, int nodata, float dc, float occ, float bcc, float ccc, float csc, float sc, float wc, int2s *scale, int *zlim, char **zlab, char **sid, float sa, float se){
int b;
char hname[MAX_STR_LEN];
char *pch = NULL;
time_t TIME;
struct utsname sys;
char user[MAX_STR_LEN];
FILE *fp = NULL;


  if (par.format != 0 && par.format != 1){
    printf("unknown output format. "); return FAILURE;}

  time(&TIME);

  // header name
  strcpy(hname, fname);
  if (par.format == 0) pch = strstr(hname, ".dat"); else pch = strstr(hname, ".tif");
  strncpy(pch, ".hdr", 4);

  // if GTiff, write from scratch
  if (par.format == 0){

    if (!(fp = fopen(hname, "a"))){
      printf("unable to open ENVI header. "); return FAILURE;}

  } else {

    if (!(fp = fopen(hname, "w"))){
      printf("unable to open ENVI header. "); return FAILURE;}

    // general metadata
    fprintf(fp, "ENVI\ndescription = {\n%s}\n", desc);
    fprintf(fp, "samples = %d\nlines = %d\nbands = %d\n", nx, ny, nb);
    fprintf(fp, "header offset = 0\nfile type = TIFF\n");
    fprintf(fp, "data type = 2\ninterleave = bsq\nbyte order = 0\n");

    // bandnames
    if (bn != NULL){
      fprintf(fp, "band names = {\n%s", bn[0]);
      for (b=1; b<nb; b++) fprintf(fp, ", %s", bn[b]);
      fprintf(fp, "}\n");
    }

  }

  // product type
  if (prd != NULL) fprintf(fp, "product type = %s\n", prd);

  // software version
  fprintf(fp, "FORCE version = %s\n", _VERSION_);

  // acquisition time
  if (acq != NULL) fprintf(fp, "acquisition time = %s\n", acq);

  // scale factors
  if (scale != NULL){
    fprintf(fp, "scale factor = {\n%d", scale[0]);
    for (b=1; b<nb; b++) fprintf(fp, ", %d", scale[b]);
    fprintf(fp, "}\n");
  }

  // nodata / res
  fprintf(fp, "data ignore value = %d\n", nodata);
  fprintf(fp, "pixel size = {%f, %f}\n", geotran[1], -1*geotran[5]);

  // sun angles
  if (sa > -1) fprintf(fp, "sun azimuth = %.2f\n", sa);
  if (se > -1) fprintf(fp, "sun elevation = %.2f\n", se);

  // meteor cover
  if (dc > -1)  fprintf(fp, "data cover = %.2f\n", dc);
  if (occ > -1) fprintf(fp, "opaque cloud cover = %.2f\n", occ);
  if (bcc > -1) fprintf(fp, "buffered cloud cover = %.2f\n", bcc);
  if (ccc > -1) fprintf(fp, "cirrus cloud cover = %.2f\n", ccc);
  if (csc > -1) fprintf(fp, "cloud shadow cover = %.2f\n", csc);
  if (sc > -1)  fprintf(fp, "snow cover = %.2f\n", sc);
  if (wc > -1)  fprintf(fp, "water cover = %.2f\n", wc);

  // wavelength
  if (wvl != NULL){
    fprintf(fp, "wavelength = {\n%.3f", wvl[0]);
    for (b=1; b<nb; b++) fprintf(fp, ", %.3f", wvl[b]);
    fprintf(fp, "}\n");
    if (wvu != NULL) fprintf(fp, "wavelength units = %s\n", wvu);
  }

  // z-plot info
  if (zlim != NULL){
    fprintf(fp, "z plot range = {%d, %d}\n", zlim[0], zlim[1]);
    if (zlab != NULL){
      fprintf(fp, "z plot titles = {%s, %s}\n", zlab[0], zlab[1]);
    }
  }

  // sensor ID
  if (sid != NULL){
    fprintf(fp, "sensor types = {\n%s", sid[0]);
    for (b=1; b<nb; b++) fprintf(fp, ", %s", sid[b]);
    fprintf(fp, "}\n");
  }

  // processing time
  fprintf(fp, "production time = %s", ctime(&TIME));

  // system info
  if(uname(&sys) != -1){
    fprintf(fp, "system name = %s\nhost name = %s\n", 
      sys.sysname, sys.nodename);
    fprintf(fp, "kernel = %s %s\nmachine = %s\n",
      sys.release, sys.version, sys.machine);
  }
  if (getlogin_r(user, MAX_STR_LEN) == 0){
    fprintf(fp, "operator = %s\n", user);
  }


  // append parameterization
  fprintf(fp, "parameterization = {\n");
  switch (level){
    case 0:
      break;
    case 1:
      log_param_cso(fp);
      break;
    case 2:
      log_param_l2(fp);
      break;
    case 3:
      log_param_l3(fp);
      break;
    case 4:
      log_param_tsa(fp);
      break;
    case 5:
      log_param_imp(fp);
      break;
    case 6:
      log_param_il2(fp);
      break;
  }
  fprintf(fp, "}\n");

  fclose(fp);
  
  return SUCCESS;
}


/** Write Level 2 products
+++ This function outputs the requested Level 2 products.
--- DAT:     Data stack
--- nx:      number of pixels in x-direction
--- ny:      number of pixels in y-direction
--- xmid:    x-position of center
--- ymid:    y-position of center
--- geotran: geotransformation
+++ Return:  SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int write_products_l2(int2s **DAT, int nx, int ny, int xmid, int ymid, double geotran[6]){
int l, b, p, c;
char fname[MAX_STR_LEN];
char *lock = NULL;
double timeout;
GDALDatasetH fp;
GDALDatasetH fe;
GDALRasterBandH band;
GDALDriverH driver;
char **options = NULL;
char aq[MAX_STR_LEN];
float sa, se;
float dc, occ, bcc, ccc, csc, sc, wc;
int nbit = 16;
bool bin[16];
int *zlim = NULL;
char **zlab = NULL;
char form[MAX_STR_LEN];
char ext[MAX_STR_LEN];
int ob;
int2s **buf = NULL;


  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  GDALAllRegister();
  CPLSetConfigOption("GDAL_PAM_ENABLED", "YES");

  // choose between formats
  if (par.format == 0){
    strcpy(form, "ENVI"); strcpy(ext, "dat"); 
  } else if (par.format == 1){
    strcpy(form, "GTiff"); strcpy(ext, "tif"); 
  } else {
    printf("unknown output format. "); return FAILURE;
  }
  driver = GDALGetDriverByName(form);

  // set GeoTiff options
  if (par.format == 1){
    options = CSLSetNameValue(options, "COMPRESS", "LZW");
    options = CSLSetNameValue(options, "PREDICTOR", "2");
    options = CSLSetNameValue(options, "INTERLEAVE", "BAND");
  }


  // center of image
  if (xmid < 0) xmid = 0; else if (xmid >= _NX_) xmid = _NX_-1;
  if (ymid < 0) ymid = 0; else if (ymid >= _NY_) ymid = _NY_-1;

  // create output directory
  createdir(par.d_out);

  // compile metadata
  sprintf(aq, "%04d-%02d-%02dT%02d:%02d:%02d.0Z", 
    meta.date.year, meta.date.month, meta.date.day,
    meta.date.hh, meta.date.mm, meta.date.ss);
  sa = rad2deg(atc.cg[tocg(ymid)*atc.ngx+tocg(xmid)].sazi);
  se = 90-rad2deg(atc.cg[tocg(ymid)*atc.ngx+tocg(xmid)].szen);

  // for each requested product, 
  // start with QAI (last product) because of cloud cover etc.
  for (l=(_NP_-1); l>=0; l--){

    if (!par.m2d[l].save) continue;

    // filenames
    sprintf(fname, "%s/%s_%s.%s", par.d_out, par.b_out, par.m2d[l].product, ext);

    // set z-plot metadata
    if (strcmp(par.m2d[l].product, "BOA") == 0 ||
        strcmp(par.m2d[l].product, "TOA") == 0){

      alloc((void**)&zlim, 2, sizeof(int));
      alloc_2D((void***)&zlab, 2, MAX_STR_LEN, sizeof(char));

      zlim[0] = 0; zlim[1] = 6000;
      sprintf(zlab[0], "Wavelength"); sprintf(zlab[1], "Reflectance"); 

    } else {

      zlim = NULL;
      zlab = NULL;

    }

    if (nx*ny > 3000000) timeout = 60.0*nx*ny/3000000.0; else timeout = 60;

    if ((lock = (char*)CPLLockFile(fname, timeout)) == NULL){
      printf("Unable to lock file (timeout: %fs, nx/ny: %d/%d). ", timeout, nx, ny); return FAILURE;}

    // Level 2 mosaicking into existing file
    // read and rewrite image (safer when using compression)
    if (fileexist(fname)){

      // read image

      if ((fe = GDALOpen(fname, GA_ReadOnly)) == NULL){
        printf("Unable to open (r) image. "); return FAILURE;}

      alloc_2D((void***)&buf, par.m2d[l].length, nx*ny, sizeof(int2s));

      for (b=0; b<par.m2d[l].length; b++){

        band = GDALGetRasterBand(fe, b+1);
        
        if (GDALGetRasterXSize(fe) != nx){
          printf("Dimensions (x) are not correct (r/w). "); return FAILURE;}
        if (GDALGetRasterYSize(fe) != ny){
          printf("Dimensions (y) are not correct (r/w). "); return FAILURE;}
        if (GDALGetRasterCount(fe) != par.m2d[l].length){
          printf("Dimensions (z) are not correct (r/w). "); return FAILURE;}

        if (GDALRasterIO(band, GF_Read, 0, 0, nx, ny, buf[b], 
          nx, ny, GDT_Int16, 0, 0) == CE_Failure){
          printf("Unable to read image. "); return FAILURE;}

      }

      GDALClose(fe);


      // update values

      for (b=par.m2d[l].start, ob=0; b<par.m2d[l].stop; b++, ob++){
        
        for (p=0; p<nx*ny; p++){

          // if both old and new are valid
          if (DAT[b][p] != par.m2d[l].nodata && buf[ob][p] != par.m2d[l].nodata){

            // if QAI, take the new value, else compute mean
            if (strcmp(par.m2d[l].product, "QAI") != 0){
              DAT[b][p] = (DAT[b][p] + buf[ob][p])/2;
            }

          // if only old is valid, take old value
          } else if (DAT[b][p] == par.m2d[l].nodata && buf[ob][p] != par.m2d[l].nodata){
            DAT[b][p] = buf[ob][p];
          }
          // if only new is valid, nothing to do

        }

      }

      free_2D((void**)buf, par.m2d[l].length);

    }


    // re-compute meteor cover
    if (strcmp(par.m2d[l].product, "QAI") == 0){

      dc = bcc = occ = ccc = csc = sc = wc = 0.0;

      for (p=0; p<nx*ny; p++){

        int2bit(DAT[par.m2d[_B_.qai].start][p], bin, 0, nbit);

        // count valid and cloudy pixels
        if (bin[0] == 0){
          c = bit2int(bin, 1, 2);
          if (c == 1)  bcc++;
          if (c == 2)  occ++;
          if (c == 3)  ccc++;
          if (bin[3] == 1) csc++;
          if (bin[4] == 1) sc++;
          if (bin[5] == 1) wc++;
          dc++;
        }
        
      }

      if (dc > 0){
        bcc = bcc/dc*100.0; occ = occ/dc*100.0;
        ccc = ccc/dc*100.0; csc = csc/dc*100.0;
        sc  = sc/dc*100.0;  wc  = wc/dc*100.0;
      }
      dc = dc/(nx*ny)*100;

    }


    if ((fp = GDALCreate(driver, fname, nx, ny, par.m2d[l].length, GDT_Int16, options)) == NULL){
      printf("Error creating output file. "); return FAILURE;}

    for (b=par.m2d[l].start, ob=0; b<par.m2d[l].stop; b++, ob++){
      band = GDALGetRasterBand(fp, ob+1);
      GDALRasterIO(band, GF_Write, 0, 0, nx, ny, DAT[b], nx, ny, GDT_Int16, 0, 0);
      if (par.format == 0) GDALSetDescription(band, par.m2d[l].bn[ob]);
      GDALSetRasterNoDataValue(band, _NODATA_);
    }

    // write geo-metadata
    GDALSetGeoTransform(fp, geotran);
    GDALSetProjection(fp, par.css_dst);

    // in case of ENVI, update description
    if (par.format == 0) GDALSetDescription(fp, "Level 2 data generated by FORCE");

    GDALClose(fp);

    // append/write metadata for tiled output
    // if mosaic, it will be done at the end of the program
    append_envi_header(fname, "Level 2 data generated by FORCE", par.m2d[l].product, 2, nx, ny, par.m2d[l].length, geotran, par.m2d[l].bn, par.m2d[l].ts, "Micrometers", aq, par.m2d[l].nodata, dc, occ, bcc, ccc, csc, sc, wc, par.m2d[l].scale, zlim, zlab, NULL, sa, se);
    
    if (zlim != NULL) free((void*)zlim);
    if (zlab != NULL) free_2D((void**)zlab, 2);
   
    CPLUnlockFile(lock);

  }

  if (options != NULL){
     CSLDestroy(options); options = NULL;}

  #ifdef CLOCK
  proctime_print("write data", TIME);
  #endif

  return SUCCESS;
}


/** Write Level 3 products
+++ This function outputs the requested Level 3 products.
--- tx:      tile coordinate ID in x-direction
--- ty:      tile coordinate ID in y-direction
--- geotran: geotransformation
+++ Return:  SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int write_products_l3(int tx, int ty, double geotran[6]){
int l, m, d, b;
int xoff, yoff;
char bname[MAX_STR_LEN];
char fname[MAX_STR_LEN];
char *lock = NULL;
double timeout;
GDALDatasetH fp;
GDALRasterBandH band;
GDALDriverH driver;
char **options = NULL;
char aq[MAX_STR_LEN];
int *zlim = NULL;
char **zlab = NULL;
char form[MAX_STR_LEN];
char ext[MAX_STR_LEN];


  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  GDALAllRegister();
  CPLSetConfigOption("GDAL_PAM_ENABLED", "YES");

  // choose between formats
  if (par.format == 0){
    strcpy(form, "ENVI"); strcpy(ext, "dat"); 
  } else if (par.format == 1){
    strcpy(form, "GTiff"); strcpy(ext, "tif"); 
  } else {
    printf("unknown output format. "); return FAILURE;
  }
  driver = GDALGetDriverByName(form);

  // set GeoTiff options
  if (par.format == 1){
    options = CSLSetNameValue(options, "COMPRESS", "LZW");
    options = CSLSetNameValue(options, "PREDICTOR", "2");
    options = CSLSetNameValue(options, "INTERLEAVE", "BAND");
  }

  // get target DOY
  if (par.score_type == SCR_SIG_DES) doy2md(par.Dt[0], &m, &d);
  if (par.score_type == SCR_GAUSS)   doy2md(par.Dt[1], &m, &d);
  if (par.score_type == SCR_SIG_ASC) doy2md(par.Dt[2], &m, &d);

  // compile metadata
  sprintf(aq, "%04d-%02d-%02dT12:00:00.0Z", par.Yt, m, d);


  // create output directory
  if (par.dotile){
    xoff = 0;
    yoff = 0;
    sprintf(bname, "%s/X%04d_Y%04d", par.d_level3, tx, ty);
  } else {
    xoff = (tx-par.tminx)*_NX_;
    yoff = (ty-par.tminy)*_NY_;
    sprintf(bname, "%s/mosaic", par.d_level3);
  }
  createdir(bname);

  // for every requested output product
  for (l=0; l<_NP_; l++){

    if (!par.m2d[l].save) continue;

    // filename
    sprintf(fname, "%s/%04d%02d%02d_LEVEL3_%s_%s.%s", 
      bname, par.Yt, m, d, par.l2dict.target, par.m2d[l].product, ext);

    // set z-plot metadata (should be fine for most products/indices)
    if (strcmp(par.m2d[l].product, "INF") == 0){

      zlim = NULL;
      zlab = NULL;
      
    } else {

      alloc((void**)&zlim, 2, sizeof(int));
      alloc_2D((void***)&zlab, 2, MAX_STR_LEN, sizeof(char));

      if (strcmp(par.m2d[l].product, "SCR") == 0){
        zlim[0] = 0; zlim[1] = 10000;
        sprintf(zlab[0], "Proxy"); strcpy(zlab[1], "Score");
      } else if (strcmp(par.m2d[l].product, "SKW") == 0){
        zlim[0] = -30000; zlim[1] = 30000;
        sprintf(zlab[0], "Wavelength"); strcpy(zlab[1], "Reflectance");
      } else if (strcmp(par.m2d[l].product, "KRT") == 0){
        zlim[0] = 0; zlim[1] = 1200;
        sprintf(zlab[0], "Wavelength"); strcpy(zlab[1], "Reflectance");
      } else {
        zlim[0] = 0; zlim[1] = 6000;
        sprintf(zlab[0], "Wavelength"); strcpy(zlab[1], "Reflectance");
      }
    }

    if (_NC_ > 3000000) timeout = 60.0*_NC_/3000000.0; else timeout = 60;

    if ((lock = (char*)CPLLockFile(fname, timeout)) == NULL){
      printf("Unable to lock file. "); return FAILURE;}

    // write as single-tile output
    if (par.dotile){

      if ((fp = GDALCreate(driver, fname, _NX_, _NY_, par.m2d[l].length, GDT_Int16, 
                           options)) == NULL){
        printf("Error creating output file. "); return FAILURE;}
      GDALSetGeoTransform(fp, geotran);
      GDALSetProjection(fp, par.css_dst);

    // write into mosaic
    } else {

      if ((fp = GDALOpen(fname, GA_Update)) == NULL){
        printf("Unable to open L3 output. "); return FAILURE;}

    }

    // write data
    for (b=0; b<par.m2d[l].length; b++){
      band = GDALGetRasterBand(fp, b+1);
      GDALRasterIO(band, GF_Write, xoff, yoff, _NX_, _NY_, 
        par.m2d[l].data[b], _NX_, _NY_, GDT_Int16, 0, 0);
      if (par.m2d[l].bn != NULL && par.format == 0) GDALSetDescription(band, par.m2d[l].bn[b]);
      GDALSetRasterNoDataValue(band, par.m2d[l].nodata);
    }

    // in case of ENVI, update description
    if (par.format == 0) GDALSetDescription(fp, "Level 3 data generated by FORCE");

    GDALClose(fp);

    // append/write metadata
    if (par.dotile){
      append_envi_header(fname, "Level 3 data generated by FORCE", par.m2d[l].product, 3, _NX_, _NY_, par.m2d[l].length, geotran, par.m2d[l].bn, NULL, NULL, aq, par.m2d[l].nodata, -1, -1, -1, -1, -1, -1, -1, par.m2d[l].scale, zlim, zlab, NULL, -1, -1);
    }

    if (zlim != NULL) free((void*)zlim);
    if (zlab != NULL) free_2D((void**)zlab, 2);

    CPLUnlockFile(lock);

  }

  if (options != NULL){
     CSLDestroy(options); options = NULL;}

  #ifdef CLOCK
  proctime_print("write data", TIME);
  #endif

  return SUCCESS;
}


/** Write Level 4 TSA products
+++ This function outputs the requested Level 4 TSA products.
--- tx:      tile coordinate ID in x-direction
--- ty:      tile coordinate ID in y-direction
--- m2d:     TSA memory to disc dictionary
--- geotran: geotransformation
+++ Return:  SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int write_products_tsa(int tx, int ty, mem_to_disc_t m2d, double geotran[6]){
int b;
int *zlim = NULL;
char **zlab = NULL;
char bname[MAX_STR_LEN];
char fname[MAX_STR_LEN];
char *lock = NULL;
double timeout;
GDALDatasetH fp;
GDALRasterBandH band;
GDALDriverH driver;
char **options = NULL;
char form[MAX_STR_LEN];
char ext[MAX_STR_LEN];


  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  GDALAllRegister();
  CPLSetConfigOption("GDAL_PAM_ENABLED", "YES");

  // choose between formats
  if (par.format == 0){
    strcpy(form, "ENVI"); strcpy(ext, "dat"); 
  } else if (par.format == 1){
    strcpy(form, "GTiff"); strcpy(ext, "tif"); 
  } else {
    printf("unknown output format. "); return FAILURE;
  }
  driver = GDALGetDriverByName(form);

  // set GeoTiff options
  if (par.format == 1){
    options = CSLSetNameValue(options, "COMPRESS", "LZW");
    options = CSLSetNameValue(options, "PREDICTOR", "2");
    options = CSLSetNameValue(options, "INTERLEAVE", "BAND");
  }


  // create output directory
  sprintf(bname, "%s/X%04d_Y%04d", par.d_tsa, tx, ty);
  createdir(bname);

  // filename
  sprintf(fname, "%s/%04d-%04d_%03d-%03d_LEVEL4_TSA_%s_%s_C%d_S%d_F%s_T%s_C%02.0f%s_%s.%s", 
    bname, par.yearmin, par.yearmax, par.doymin, par.doymax, par.l2dict.target, par.index_c, par.docenter, par.dostandard, 
    par.fold_c, par.trend_c, par.conf*100, par.tail_c, m2d.product, ext);
    
  // set z-plot metadata (should be fine for most products/indices)
  if (strcmp(m2d.product, "STA") != 0 &&
      strcmp(m2d.product, "TRD") != 0 &&
      strcmp(m2d.product, "CAT") != 0){

    alloc((void**)&zlim, 2, sizeof(int));
    alloc_2D((void***)&zlab, 2, MAX_STR_LEN, sizeof(char));

    zlim[0] = 0; zlim[1] = 10000;
    sprintf(zlab[0], "Time"); strcpy(zlab[1], par.index_c); 

  } else {

    zlim = NULL;
    zlab = NULL;

  }

  if (m2d.length > 600) timeout = m2d.length*0.1; else timeout = 60;

  if ((lock = (char*)CPLLockFile(fname, timeout)) == NULL){
    printf("Unable to lock file. "); return FAILURE;}

  // create file
  if ((fp = GDALCreate(driver, fname, _NX_, _NY_, m2d.length, GDT_Int16, 
                       options)) == NULL){
    printf("Error creating output file. "); return FAILURE;}

  // write data
  for (b=0; b<m2d.length; b++){
    band = GDALGetRasterBand(fp, b+1);
    GDALRasterIO(band, GF_Write, 0, 0, _NX_, _NY_, 
      m2d.data[b], _NX_, _NY_, GDT_Int16, 0, 0);
    if (m2d.bn != NULL && par.format == 0) GDALSetDescription(band, m2d.bn[b]);
    GDALSetRasterNoDataValue(band, m2d.nodata);
  }

  // write geo-metadata
  GDALSetGeoTransform(fp, geotran);
  GDALSetProjection(fp, par.css_dst);

  // in case of ENVI, update description
  if (par.format == 0) GDALSetDescription(fp, "TSA data generated by FORCE");

  GDALClose(fp);

  // append/write metadata
  append_envi_header(fname, "TSA data generated by FORCE", m2d.product, 4, _NX_, _NY_, m2d.length, geotran, m2d.bn, m2d.ts, "Decimal Years", NULL, m2d.nodata, -1, -1, -1, -1, -1, -1, -1, m2d.scale, zlim, zlab, m2d.sid, -1, -1);
  
  if (zlim != NULL) free((void*)zlim);
  if (zlab != NULL) free_2D((void**)zlab, 2);

  CPLUnlockFile(lock);

  if (options != NULL){
     CSLDestroy(options); options = NULL;}

  #ifdef CLOCK
  proctime_print("write data", TIME);
  #endif

  return SUCCESS;
}


/** Write CSO products
+++ This function outputs the requested CSO statistics products.
--- tx:      tile coordinate ID in x-direction
--- ty:      tile coordinate ID in y-direction
--- m2d:     CSO memory to disc dictionary
--- geotran: geotransformation
+++ Return:  SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int write_products_cso(int tx, int ty, mem_to_disc_t m2d, double geotran[6]){
int l, m, d, b, p;
int *zlim = NULL;
char **zlab = NULL;
double avg, var, sd, k;
int2s min, max;
char bname[MAX_STR_LEN];
char fname[MAX_STR_LEN];
char hname[MAX_STR_LEN];
char *pch = NULL;
char *lock = NULL;
double timeout;
GDALDatasetH fp;
GDALRasterBandH band;
GDALDriverH driver;
char **options = NULL;
char form[MAX_STR_LEN];
char ext[MAX_STR_LEN];
FILE *fh;


  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  GDALAllRegister();
  CPLSetConfigOption("GDAL_PAM_ENABLED", "YES");

  // choose between formats
  if (par.format == 0){
    strcpy(form, "ENVI"); strcpy(ext, "dat"); 
  } else if (par.format == 1){
    strcpy(form, "GTiff"); strcpy(ext, "tif"); 
  } else {
    printf("unknown output format. "); return FAILURE;
  }
  driver = GDALGetDriverByName(form);

  // set GeoTiff options
  if (par.format == 1){
    options = CSLSetNameValue(options, "COMPRESS", "LZW");
    options = CSLSetNameValue(options, "PREDICTOR", "2");
    options = CSLSetNameValue(options, "INTERLEAVE", "BAND");
  }


  // create output directory
  sprintf(bname, "%s/X%04d_Y%04d", par.d_cso, tx, ty);
  createdir(bname);

  // filename
  sprintf(fname, "%s/%04d-%04d_%02dM_CSO-STATS_%s_%s.%s", 
    bname, par.yearmin, par.yearmax, par.monthstep, par.l2dict.target, m2d.product, ext);


  alloc((void**)&zlim, 2, sizeof(int));
  alloc_2D((void***)&zlab, 2, MAX_STR_LEN, sizeof(char));

  // set z-plot metadata (should be fine for most occurences)
  if (strcmp(m2d.product, "NUM") == 0){

    zlim[0] = 0; zlim[1] = par.monthstep*10;
    sprintf(zlab[0], "Time"); sprintf(zlab[1], "# of obs"); 

  } else if (strcmp(m2d.product, "SKW") == 0){

    zlim[0] = -3*m2d.scale[0]; zlim[1] = 3*m2d.scale[0];
    sprintf(zlab[0], "Time"); sprintf(zlab[1], "%s of dt in days", m2d.product); 

  } else if (strcmp(m2d.product, "KRT") == 0){

    zlim[0] = 0; zlim[1] = 9*m2d.scale[0];
    sprintf(zlab[0], "Time"); sprintf(zlab[1], "%s of dt in days", m2d.product); 

  } else {

    if (par.nwin > 1){
      zlim[0] = 0; zlim[1] = par.cewin[1]-par.cewin[0];
    } else {
      zlim[0] = 0; zlim[1] = 365;
    }
    sprintf(zlab[0], "Time"); sprintf(zlab[1], "%s of dt in days", m2d.product); 

  }

  if (m2d.length > 600) timeout = m2d.length*0.1; else timeout = 60;

  if ((lock = (char*)CPLLockFile(fname, timeout)) == NULL){
    printf("Unable to lock file. "); return FAILURE;}

  // create file
  if ((fp = GDALCreate(driver, fname, _NX_, _NY_, m2d.length, GDT_Int16, 
                       options)) == NULL){
    printf("Error creating output file. "); return FAILURE;}

  // write data
  for (b=0; b<m2d.length; b++){
    band = GDALGetRasterBand(fp, b+1);
    GDALRasterIO(band, GF_Write, 0, 0, _NX_, _NY_, 
      m2d.data[b], _NX_, _NY_, GDT_Int16, 0, 0);
    if (m2d.bn != NULL && par.format == 0) GDALSetDescription(band, m2d.bn[b]);
    GDALSetRasterNoDataValue(band, m2d.nodata);
  }

  // write geo-metadata
  GDALSetGeoTransform(fp, geotran);
  GDALSetProjection(fp, par.css_dst);

  // in case of ENVI, update description
  if (par.format == 0) GDALSetDescription(fp, "CSO data generated by FORCE");

  GDALClose(fp);

  // append/write metadata
  append_envi_header(fname, "CSO data generated by FORCE", m2d.product, 1, _NX_, _NY_, m2d.length, geotran, m2d.bn, m2d.ts, "Decimal Years", NULL, m2d.nodata, -1, -1, -1, -1, -1, -1, -1, m2d.scale, zlim, zlab, NULL, -1, -1);

  free((void*)zlim);
  free_2D((void**)zlab, 2);

  /** compute tile stats and add to metadata **/
  
  // header name
  strcpy(hname, fname);
  if (par.format == 0) pch = strstr(hname, ".dat"); else pch = strstr(hname, ".tif");
  strncpy(pch, ".hdr", 4);

  if (!(fh = fopen(hname, "a"))){
    printf("unable to open ENVI header. "); return FAILURE;}

  fprintf(fh, "cso stats = {\n");
  for (b=0; b<m2d.length; b++){
    
    var = k = 0.0;
    avg = (double)m2d.data[b][0];
    min = m2d.data[b][0];
    max = m2d.data[b][0];

    for (p=1, k=2; p<_NC_; p++, k++){
      var_recurrence((double)m2d.data[b][p], &avg, &var, k);
      if (m2d.data[b][p] < min) min = m2d.data[b][p];
      if (m2d.data[b][p] > max) max = m2d.data[b][p];
    }
    sd = standdev(var, k);
    fprintf(fh, "%05d/%08.2f/%08.2f/%08.2f/%05d", min, avg-sd, avg, avg+sd, max);
    if (b < m2d.length-1) fprintf(fh, ", "); else fprintf(fh, "}\n");

  }

  fclose(fh);

  CPLUnlockFile(lock);
  
  if (options != NULL){
     CSLDestroy(options); options = NULL;}

  #ifdef CLOCK
  proctime_print("write data", TIME);
  #endif

  return SUCCESS;
}


/** Write ImproPhe products
+++ This function outputs the ImproPhe products.
--- tx:      tile coordinate ID in x-direction
--- ty:      tile coordinate ID in y-direction
--- y:       year
--- IMP:     ImproPhe struct
--- geotran: geotransformation
+++ Return:  SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int write_products_imp(int tx, int ty, int y, imp_t IMP, double geotran[6]){
int f;
char bname[MAX_STR_LEN];
char fname[MAX_STR_LEN];
char *lock = NULL;
double timeout;
GDALDatasetH fp;
GDALRasterBandH band;
GDALDriverH driver;
char **options = NULL;
char form[MAX_STR_LEN];
char ext[MAX_STR_LEN];
char **bn = NULL;

  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  GDALAllRegister();
  CPLSetConfigOption("GDAL_PAM_ENABLED", "YES");

  // choose between formats
  if (par.format == 0){
    strcpy(form, "ENVI"); strcpy(ext, "dat"); 
  } else if (par.format == 1){
    strcpy(form, "GTiff"); strcpy(ext, "tif"); 
  } else {
    printf("unknown output format. "); return FAILURE;
  }
  driver = GDALGetDriverByName(form);

  // set GeoTiff options
  if (par.format == 1){
    options = CSLSetNameValue(options, "COMPRESS", "LZW");
    options = CSLSetNameValue(options, "PREDICTOR", "2");
    options = CSLSetNameValue(options, "INTERLEAVE", "BAND");
  }

  // create output directory
  sprintf(bname, "%s/X%04d_Y%04d", par.d_improphe, tx, ty);
  createdir(bname);

  // band name
  alloc_2D((void***)&bn, 1, MAX_STR_LEN, sizeof(char));
  sprintf(bn[0], "ImproPhe'd data");

  for (f=0; f<par.num_cr; f++){
  
    // filename
    sprintf(fname, "%s/%04d_IMPROPHE_%s.%s", 
      bname, y, par.imp_prd[f], ext);

    if (_NC_ > 3000000) timeout = 60.0*_NC_/3000000.0; else timeout = 60;

    if ((lock = (char*)CPLLockFile(fname, timeout)) == NULL){
      printf("Unable to lock file. "); return FAILURE;}

    // create file
    if ((fp = GDALCreate(driver, fname, _NX_, _NY_, 1, GDT_Float32, 
                         options)) == NULL){
      printf("Error creating output file. "); return FAILURE;}

    // write data
    band = GDALGetRasterBand(fp, 1);
    GDALRasterIO(band, GF_Write, 0, 0, _NX_, _NY_, 
      IMP.PRED[f], _NX_, _NY_, GDT_Float32, 0, 0);
    if (par.format == 0) GDALSetDescription(band, "ImproPhe'd data");
    GDALSetRasterNoDataValue(band, par.nodata_cr);

    // write geo-metadata
    GDALSetGeoTransform(fp, geotran);
    GDALSetProjection(fp, par.css_dst);

    // in case of ENVI, update description
    if (par.format == 0) GDALSetDescription(fp, "ImproPhe'd data generated by FORCE");

    GDALClose(fp);

    // append/write metadata
    append_envi_header(fname, "ImproPhe'd data generated by FORCE", NULL, 5, _NX_, _NY_, 1, geotran, bn, NULL, NULL, NULL, par.nodata_cr, -1, -1, -1, -1, -1, -1, -1, NULL, NULL, NULL, NULL, -1, -1);

    CPLUnlockFile(lock);

  }

  free_2D((void**)bn, 1);

  if (options != NULL){
     CSLDestroy(options); options = NULL;}

  #ifdef CLOCK
  proctime_print("write data", TIME);
  #endif

  return SUCCESS;
}


/** Write ImproPhe Level 2 products
+++ This function outputs the ImproPhe products.
--- tx:      tile coordinate ID in x-direction
--- ty:      tile coordinate ID in y-direction
--- m2d:     memory to disc dictionary
--- geotran: geotransformation
+++ Return:  SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int write_products_il2(int tx, int ty, mem_to_disc_t m2d, double geotran[6]){
int b;
char fname[MAX_STR_LEN];
char *lock = NULL;
double timeout;
GDALDatasetH fp;
GDALRasterBandH band;
GDALDriverH driver;
char **options = NULL;
int *zlim = NULL;
char **zlab = NULL;
char form[MAX_STR_LEN];
char ext[MAX_STR_LEN];


  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  GDALAllRegister();
  CPLSetConfigOption("GDAL_PAM_ENABLED", "YES");

  // choose between formats
  if (par.format == 0){
    strcpy(form, "ENVI"); strcpy(ext, "dat"); 
  } else if (par.format == 1){
    strcpy(form, "GTiff"); strcpy(ext, "tif"); 
  } else {
    printf("unknown output format. "); return FAILURE;
  }
  driver = GDALGetDriverByName(form);

  // set GeoTiff options
  if (par.format == 1){
    options = CSLSetNameValue(options, "COMPRESS", "LZW");
    options = CSLSetNameValue(options, "PREDICTOR", "2");
    options = CSLSetNameValue(options, "INTERLEAVE", "BAND");
  }


  // filename
  sprintf(fname, "%s/X%04d_Y%04d/%s.%s", 
    par.d_level2, tx, ty, par.b_out, ext);

  if (fileexist(fname) && !par.owr){
    printf("output file exists. overwrite is disabled."); return FAILURE;}

  alloc((void**)&zlim, 2, sizeof(int));
  alloc_2D((void***)&zlab, 2, MAX_STR_LEN, sizeof(char));

  zlim[0] = 0; zlim[1] = 6000;
  sprintf(zlab[0], "Wavelength"); sprintf(zlab[1], "Reflectance"); 

  if (_NC_ > 3000000) timeout = 60.0*_NC_/3000000.0; else timeout = 60;
  
  if ((lock = (char*)CPLLockFile(fname, timeout)) == NULL){
    printf("Unable to lock file. "); return FAILURE;}

  // create file
  if ((fp = GDALCreate(driver, fname, _NX_, _NY_, _NB_, GDT_Int16, 
                       options)) == NULL){
    printf("Error creating output file. "); return FAILURE;}

  // write data
  for (b=0; b<_NB_; b++){
    band = GDALGetRasterBand(fp, b+1);
    GDALRasterIO(band, GF_Write, 0, 0, _NX_, _NY_, 
      m2d.data[b], _NX_, _NY_, GDT_Int16, 0, 0);
      if (par.format == 0) GDALSetDescription(band, m2d.bn[b]);
      GDALSetRasterNoDataValue(band, m2d.nodata);
      GDALSetDescription(band, "ImproPhe'd Level 2 reflectance");
  }

  // write geo-metadata
  GDALSetGeoTransform(fp, geotran);
  GDALSetProjection(fp, par.css_dst);

  // in case of ENVI, update description
  if (par.format == 0) GDALSetDescription(fp, "ImproPhe'd Level 2 data generated by FORCE");

  GDALClose(fp);

  // append/write metadata
  append_envi_header(fname, "ImproPhe'd Level 2 data generated by FORCE", m2d.product, 6, _NX_, _NY_, m2d.length, geotran, m2d.bn, m2d.ts, "Micrometers", "tbd", m2d.nodata, -1, -1, -1, -1, -1, -1, -1, m2d.scale, zlim, zlab, NULL, 0, 0);
  
  free((void*)zlim);
  free_2D((void**)zlab, 2);

  CPLUnlockFile(lock);

  if (options != NULL){
     CSLDestroy(options); options = NULL;}

  #ifdef CLOCK
  proctime_print("write data", TIME);
  #endif

  return SUCCESS;
}

