/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/** The STARFM implementation included in this file, is adopted from the
+++ version provided provided by the United States Department of Agricul-
+++ ture, Agricultural Research Service:
+++ https://www.ars.usda.gov/research/software/download/?softwareid=432
+++ STARFM Copyright (C) 2005-2014 Feng Gao
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+++ STARFM License:
+++ The package is now being released at no cost to the public for further
+++ development. ARS is releasing the STARFM so that interested parties 
+++ can use the software tool for their own needs and purposes. ARS does 
+++ not foresee providing monetary or technical support to refine, adapt, 
+++ or use this software tool, and provides no warranty for its use for 
+++ any purpose. ARS does not reserve any rights or interests in the work 
+++ that may be performed by others to refine or adapt it. ARS does reser-
+++ ve the right to continue its own refinement of the current version of 
+++ the software package at a later date, should program needs require it.
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
This file contains functions to enhance spatial resolution
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#include "force.h"


/** Sentinel-2 resolution merge controller
+++ This function controlls the Sentinel-2 resolution merging of the 20m 
+++ Sentinel-2 bands. Based on parameterization, different methods are 
+++ used. The function will exit gracefully if not Sentinel-2.
--- TOA:    TOA reflectance (will be altered)
--- QAI:    Quality Assurance Information
+++ Return: SUCCESS / FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int resolution_merge(int2s **TOA, qai_t QAI){

  if (meta.mission != MISSION_SENTINEL2) return SUCCESS;
  
  switch (par.resmerge){
    case NONE:
      return SUCCESS;
    case REGRESSION:
      if (resolution_merge_1(TOA, QAI) != SUCCESS){
        printf("unable to merge resolutions (1). "); return FAILURE;}
      break;
    case IMPROPHE:
      if (resolution_merge_2(TOA, QAI) != SUCCESS){
        printf("unable to merge resolutions (2). "); return FAILURE;}
      break;
    case STARFM:
      if (resolution_merge_3(TOA, QAI) != SUCCESS){
        printf("unable to merge resolutions (3). "); return FAILURE;}
      break;
    default:
      printf("unknown resolution merge option. "); return FAILURE;
  }

  return SUCCESS;
}


/** Sentinel-2 resolution merge option 1
+++ This function enhances the spatial resolution of the 20m Sentinel-2
+++ bands using a multi-parameter regression of the general form: y = c X,
+++ where y is a vector of n observations (20m band), X is an n-by-p mat-
+++ rix of predictor variables (intercept, green, red, NIR bands), c are
+++ p regression coefficients, i.e. y = c0 + c1 GREEN + c2 RED + c3 NIR.
+++ A least squares fit to a linear model is used by minimizing the cost 
+++ function chi^2 (sum of squares of the residuals from the best-fit). 
+++ The best-fit is found by singular value decomposition of the matrix X
+++ using the modified Golub-Reinsch SVD algorithm, with column scaling to
+++ improve the accuracy of the singular values. Any components which have
+++ zero singular value (to machine precision) are discarded from the fit.
+++ A moving kernel of size n = 5*5 is used based on a sensitivity study 
+++ of Haß et al. (in preparation).
--- TOA:    TOA reflectance (will be altered)
--- QAI:    Quality Assurance Information
+++ Return: SUCCESS / FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int resolution_merge_1(int2s **TOA, qai_t QAI){
int b, nb = 6, bands[6] = { 4, 5, 6, 8, 11, 12 };
int i, j, p, ii, jj, ni, nj, np;
int r = 2, w, nw, k, nv = 4;
gsl_matrix *X, **cov;
gsl_vector *x, **y, **c;
gsl_multifit_linear_workspace **work;
double chisq, est, err;

  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif


  /** initialize and allocate
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

  // kernel size
  w = r*2+1; nw = w*w;

  // nw-by-nv predictor variables; kernel + central pixel
  X = gsl_matrix_calloc(nw, nv);
  x = gsl_vector_calloc(nv);
  
  // set first column of X to 1 -> intercept c0
  for (k=0; k<nw; k++) gsl_matrix_set(X, k, 0, 1.0);
  gsl_vector_set(x, 0, 1.0);

  // vector of nw observations
  alloc((void**)&y, nb, sizeof(gsl_vector*));
  for (b=0; b<nb; b++) y[b] = gsl_vector_calloc(nw);

  // nv regression coefficients
  alloc((void**)&c, nb, sizeof(gsl_vector*));
  for (b=0; b<nb; b++) c[b] = gsl_vector_calloc(nv);

  // nv-by-nv covariance matrix
  alloc((void**)&cov, nb, sizeof(gsl_matrix*));
  for (b=0; b<nb; b++) cov[b] = gsl_matrix_calloc(nv, nv);

  // workspace
  alloc((void**)&work, nb, sizeof(gsl_multifit_linear_workspace*));
  for (b=0; b<nb; b++) work[b] = gsl_multifit_linear_alloc(nw, nv);

  #ifdef DEBUG
  int f;
  float ***coef = NULL;
  alloc_3D((void****)&coef, nb, nv+1, _NC_, sizeof(float));
  #endif


  /** do regression for every valid pixel, and for each 20m band
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
  for (i=0, p=0; i<_NY_; i++){
  for (j=0; j<_NY_; j++, p++){

    if (QAI.off[p] || QAI.cld[p] > 0 || QAI.shd[p]) continue;

    // add central pixel
    gsl_vector_set(x, 1, TOA[_B_.green][p]/_RS_);
    gsl_vector_set(x, 2, TOA[_B_.red][p]/_RS_);
    gsl_vector_set(x, 3, TOA[_B_.bnir][p]/_RS_);

    k = 0;

    // add neighboring pixels
    for (ii=-r; ii<=r; ii++){
    for (jj=-r; jj<=r; jj++){

      ni = i+ii; nj = j+jj;
      if (ni < 0 || ni >= _NY_ || nj < 0 || nj >= _NX_) continue;
      np = ni*_NX_+nj;

      if (QAI.off[np]) continue;

      gsl_matrix_set(X, k, 1, TOA[_B_.green][np]/_RS_);
      gsl_matrix_set(X, k, 2, TOA[_B_.red][np]/_RS_);
      gsl_matrix_set(X, k, 3, TOA[_B_.bnir][np]/_RS_);
      
      for (b=0; b<nb; b++) gsl_vector_set(y[b], k, TOA[bands[b]][np]/_RS_);

      k++;

    }
    }

    // append zeros, if less than nw neighboring pixels were added
    while (k < nw){
      gsl_matrix_set(X, k, 1, 0.0);
      gsl_matrix_set(X, k, 2, 0.0);
      gsl_matrix_set(X, k, 3, 0.0);
      for (b=0; b<nb; b++) gsl_vector_set(y[b], k, 0.0);
      k++;
    }

    // solve model, and predict central pixel
    for (b=0; b<nb; b++){

      gsl_multifit_linear(X, y[b], c[b], cov[b], &chisq, work[b]);
      gsl_multifit_linear_est(x, c[b], cov[b], &est, &err);
      TOA[bands[b]][p] = (int2s)(est*_RS_);
      
      #ifdef DEBUG
      // regression coefficients + R^2
      for (f=0; f<nv; f++) coef[b][f][p] = (float)gsl_vector_get(c[b],f);
      coef[b][f][p] = (float)(1.0-chisq/gsl_stats_tss(y[b]->data, y[b]->stride, y[b]->size));
      #endif

    }

  }
  }


  #ifdef DEBUG
  db_write_image_2D((void**)TOA,  "PANSHARP-TOA", _NB_, _NX_, _NY_, sizeof(int2s), 2);
  char fname[MAX_STR_LEN];
  for (b=0; b<nb; b++){
    sprintf(fname, "PANSHARP-COEF-BAND%02d", b);
    db_write_image_2D((void**)coef[b],  fname, nv+1, _NX_, _NY_, sizeof(float), 4);
  }
  free_3D((void***)coef, nb, nv+1);
  #endif


  /** clean
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
  gsl_matrix_free (X); gsl_vector_free (x);
  for (b=0; b<nb; b++) gsl_vector_free(y[b]); free((void*)y);
  for (b=0; b<nb; b++) gsl_vector_free (c[b]); free((void*)c);
  for (b=0; b<nb; b++) gsl_matrix_free (cov[b]); free((void*)cov);
  for (b=0; b<nb; b++) gsl_multifit_linear_free(work[b]); free((void*)work);

  #ifdef CLOCK
  proctime_print("Resolution merge", TIME);
  #endif
  
  return SUCCESS;
}


/** Sentinel-2 resolution merge option 2
+++ This function enhances the spatial resolution of the 20m Sentinel-2
+++ bands. The red, green and BNIR bands are used. In principle, the 
+++ ImproPhe code is used, but is paraparameterized using a spectral in-
+++ stead of a temporal setup. The texture metrics are disabled as they 
+++ work badly with a fine/coarse resolution factor of 2. This version is
+++ faster than resolution_merge_1.
--- TOA:    TOA reflectance (will be altered)
--- QAI:    Quality Assurance Information
+++ Return: SUCCESS / FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int resolution_merge_2(int2s **TOA, qai_t QAI){
int b, nb_m = 3, nb_c = 6;
int bands_m[3] = { _B_.green, _B_.red, _B_.bnir }; 
int bands_c[6] = { _B_.rededge1, _B_.rededge2, _B_.rededge3, _B_.nir, _B_.swir1, _B_.swir2 };
int i, j, p;
float **KDIST = NULL; // kernel distance
int ksize = 5, nk, mink, h; // number of kernel pixels, and minimum number of pixels for good prediction, radius
int2s **PRED = NULL;

  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif
  
  // pre-compute kernel distances
  // determine minimum # of neighbors that make a good prediction
  if (distance_kernel(ksize, &KDIST) != SUCCESS){
    printf("error in generating kernel. "); return FAILURE;}
  nk = ksize*ksize;
  h   = (ksize-1)/2;
  mink = 3;//(int)((nk*M_PI/4.0)*0.5/100.0);
  #ifdef DEBUG
  printf("ksize/nkernel/minnum: %d/%d/%d\n", ksize, nk, mink);
  #endif


  /** Predict at fine resolution **/
  alloc_2D((void***)&PRED, nb_c, _NC_, sizeof(int2s));

  for (i=0; i<_NY_; i++){
  for (j=0; j<_NX_; j++){

    //if (i-h < 0 || i+h >= _NY_ || j-h < 0 || j+h >= _NX_) continue;

    p = i*_NX_+j;

    // if nodata or cloud/shadow: skip
    if (QAI.off[p] || QAI.cld[p] > 0 || QAI.shd[p]) continue;

    // predict fine data
    improphe_sentinel2(TOA, QAI, PRED, KDIST, i, j, p, 
      h, _NX_, _NY_, nb_m, nb_c, bands_m, bands_c, nk, mink);

  }
  }

  // update toa with prediction
  for (p=0; p<_NC_; p++){
    if (QAI.off[p] || QAI.cld[p] > 0 || QAI.shd[p]) continue;
    for (b=0; b<nb_c; b++) TOA[bands_c[b]][p] = PRED[b][p];
  }

  #ifdef DEBUG
  db_write_image_2D((void**)PRED, "IMPROPHE-RES_PRED", nb_c, _NX_, _NY_, sizeof(int2s), 2);
  db_write_image_2D((void**)TOA,  "PANSHARP-ALL", _NB_, _NX_, _NY_, sizeof(int2s), 2);
  #endif


  free_2D((void**)KDIST, ksize);
  free_2D((void**)PRED, nb_c);

  #ifdef CLOCK
  proctime_print("Pansharpening", TIME);
  #endif
  
  return SUCCESS;
}


/** Sentinel-2 resolution merge option 3
+++ This function enhances the spatial resolution of the 20m Sentinel-2
+++ bands. A pseudo-PAN (average of green/red) and the BNIR are used.
+++ Coarse and fine image pairs are generated by spatially degrading the
+++ PANs using an approximated Point Spread Function. In principle, STARFM
+++ is used but is parameterized using a spectral instead of a temporal 
+++ setup. Prediction artefcats were observed at object boundaries, but 
+++ this version is faster than resolution_merge_1.
--- TOA:    TOA reflectance (will be altered)
--- QAI:    Quality Assurance Information
+++ Return: SUCCESS / FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int resolution_merge_3(int2s **TOA, qai_t QAI){
int b, bb, nb = 6, bands[6] = { 4, 5, 6, 8, 11, 12 };
int i, j, p, ii, jj, ni, nj, np;
int nk = 5;
double sum[2], num;
float **coarse = NULL;
float **fine   = NULL;
float  *pred   = NULL;
float **kernel = NULL;

  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif


  alloc_2D((void***)&coarse, 3, _NC_, sizeof(float));
  alloc_2D((void***)&fine,   2, _NC_, sizeof(float));

  for (p=0; p<_NC_; p++){
    fine[0][p] = (TOA[_B_.green][p]/_RS_+TOA[_B_.red][p]/_RS_)/2.0;
    fine[1][p] = TOA[_B_.bnir][p]/_RS_;
  }


  /** Convolute PANs with gaussian kernel **/

  if (gauss_kernel(nk, 0.5, &kernel) != SUCCESS){
    printf("Could not generate kernel. "); return FAILURE;}

  for (i=0, p=0; i<_NY_; i++){
  for (j=0; j<_NX_; j++, p++){
    
    if (QAI.off[p]) continue;

    sum[0] = sum[1] = num = 0;

    for (ii=0; ii<nk; ii++){
    for (jj=0; jj<nk; jj++){

      ni = -(nk-1)/2 + ii + i; nj = -(nk-1)/2 + jj + j;
      if (ni < 0 || ni >= _NY_ || nj < 0 || nj >= _NX_) continue;
      np = ni*_NX_+nj;
      
      if (QAI.off[np]) continue;

      sum[0] += (TOA[_B_.green][np]/_RS_+TOA[_B_.red][np]/_RS_)/2.0*kernel[ii][jj];
      sum[1] += TOA[_B_.bnir][np]/_RS_*kernel[ii][jj];
      num += kernel[ii][jj];

    }
    }

    if (num > 0){
      coarse[0][p] = sum[0]/num;
      coarse[1][p] = sum[1]/num;
    }

  }
  }

  free_2D((void**)kernel, nk);
  
  #ifdef DEBUG
  db_write_image_2D((void**)coarse,  "PANSHARP-GAUSS", 2, _NX_, _NY_, sizeof(float), 4);
  #endif


  /** Reduce spatial resolution, supersample to target resolution **/

  for (i=0, p=0; i<_NY_; i+=2){
  for (j=0; j<_NX_; j+=2, p++){
    
    if (QAI.off[p]) continue;

    sum[0] = sum[1] = num = 0;

    for (ii=0; ii<2; ii++){
    for (jj=0; jj<2; jj++){

      ni = i+ii; nj = j+jj;
      if (ni >= _NY_ || nj >= _NX_) continue;
      np = ni*_NX_+nj;

      if (QAI.off[np]) continue;

      sum[0] += coarse[0][np];
      sum[1] += coarse[1][np];
      num++;

    }
    }

    if (num > 0){

      for (ii=0; ii<2; ii++){
      for (jj=0; jj<2; jj++){

        ni = i+ii; nj = j+jj;
        if (ni >= _NY_ || nj >= _NX_) continue;
        np = ni*_NX_+nj;

        if (QAI.off[np]) continue;

        coarse[0][np] = sum[0]/num;
        coarse[1][np] = sum[1]/num;

      }
      }

    }

  }
  }

  #ifdef DEBUG
  db_write_image_2D((void**)coarse,  "PANSHARP-COARSE", 2, _NX_, _NY_, sizeof(float), 4);
  db_write_image_2D((void**)fine,    "PANSHARP-FINE",   2, _NX_, _NY_, sizeof(float), 4);
  #endif


  /** Predict at fine resolution **/

  for (bb=0; bb<nb; bb++){
    
    b = bands[bb];
    
    for (p=0; p<_NC_; p++) coarse[2][p] = TOA[b][p]/_RS_;

    starfm(QAI, coarse, fine, &pred, 1);

    for (p=0; p<_NC_; p++) TOA[b][p] = (int2s)(pred[p]*_RS_);

    free((void*)pred);

  }

  #ifdef DEBUG
  db_write_image_2D((void**)TOA,  "PANSHARP-ALL", _NB_, _NX_, _NY_, sizeof(int2s), 2);
  #endif

  free_2D((void**)coarse, 3);
  free_2D((void**)fine,   2);

  #ifdef CLOCK
  proctime_print("Pansharpening", TIME);
  #endif
  
  return SUCCESS;
}


/** STARFM
+++ This function is a streamlined version of the STARFM code using two
+++ image pairs of coarse/fine data and one coarse image that should be
+++ prediction at the fine scale.
+++-----------------------------------------------------------------------
+++ Gao, F., Masek, J., Schwaller, M., & Hall, F. (2006). On the Blending 
+++ of the Landsat and MODIS Surface Reflectance: Predicting Daily Landsat
+++ Surface Reflectance. IEEE TGRS, 44, 2207-2218.
+++-----------------------------------------------------------------------
--- OFF:    image boundary
--- coarse: coarse resolution data (3 bands, 1-2 base pairs, 3 prediction)
--- fine:   fine data (2 bands, i.e. base pairs)
--- pred:   prediction (needs to be deallocated)
--- r:      prediction radius in pixels
+++ Return: SUCCESS / FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int starfm(qai_t QAI, float **coarse, float **fine, float **pred, int r){
int i, j, p, ii, jj, ni, nj, np, c, ncan;
float diff;
float d, D;     // spatial distance, relative spatial distance
float T, S, TS; // temporal, spectral, and combined difference
float C_sum;    // sum of combined inversed weight
float W;        // normalized weight
float min_r_coarse, min_r_fine; // min. of rel. differences of central pix
float unc_coarse_1, unc_coarse_2, unc_fine; // uncertainty coarse and fine
float unc_r_coarse, unc_r_fine;             // uncertainty rel. difference
float unc_all;                              // combined uncertainty
starfm_candidate_t *can = NULL; // candidate pixel struct
float *prediction = NULL; // prediction

#ifdef DEBUG
float *percentsample = NULL; // percentage of valid candidates
#endif

int pair, npair = 2;

double sumx, sumx2, num;
float slice_value[2], std, nclass = 40;


  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif


  /** compute uncertainty **/

  unc_coarse_1 = unc_coarse_2 = unc_fine  = 0.002;
  unc_r_coarse = sqrt(unc_coarse_1*unc_coarse_1 + unc_coarse_2*unc_coarse_2);
  unc_r_fine = sqrt(unc_coarse_1*unc_coarse_1 + unc_fine*unc_fine);
  unc_all = sqrt(unc_r_coarse*unc_r_coarse +  unc_r_fine* unc_r_fine);

  #ifdef DEBUG
  printf("uncertainty coarse:   %.4f\n", unc_r_coarse);
  printf("uncertainty fine:     %.4f\n", unc_r_fine);
  printf("uncertainty combined: %.4f\n", unc_all);
  #endif


  /** compute slice values for spectal similarity **/

  for (pair=0; pair<npair; pair++){

    sumx = sumx2 = num = 0.0;

    for (i=0, p=0; i<_NY_; i++){
    for (j=0; j<_NX_; j++, p++){
      
      if (QAI.off[p]) continue;

      sumx  += fine[pair][p];
      sumx2 += fine[pair][p]*fine[pair][p];
      num++;

    }
    }

    std = sqrt(sumx2/num - (sumx/num) * (sumx/num));
    slice_value[pair] = std * 2.0 / nclass;
    
    #ifdef DEBUG
    printf("std %.4f, slice value: %.4f (fine #%d)\n", std, slice_value[pair], pair);
    #endif

  }


  /** predict fine resolution data **/

  alloc((void**)&can, (r*2+1)*(r*2+1)*npair+1, sizeof(starfm_candidate_t));
  alloc((void**)&prediction, _NC_, sizeof(float));
  #ifdef DEBUG
  alloc((void**)&percentsample, _NC_, sizeof(float));
  #endif

  for (i=0, p=0; i<_NY_; i++){
  for (j=0; j<_NX_; j++, p++){

    if (QAI.off[p]) continue;


    /** add original fine observation to candidate list (Index 0) **/

    c = 0;
    can[c].dx = 0;
    can[c].dy = 0;
    can[c].coarse = coarse[npair][p];
    c++;


    /** add base pairs of central pixel to candidates and
    +++ estimate minimum relative differences **/

    min_r_coarse = min_r_fine = SHRT_MAX;

    for (pair=0; pair<npair; pair++){

      can[c].dx = 0;
      can[c].dy = 0;
      can[c].coarse = coarse[pair][p];
      can[c].fine   = fine[pair][p];

      can[c].r_coarse = coarse[pair][p] - coarse[npair][p];
      can[c].r_fine   = coarse[pair][p] - fine[pair][p];

      if (fabs(can[c].r_coarse) < min_r_coarse) min_r_coarse = fabs(can[c].r_coarse);
      if (fabs(can[c].r_fine)   < min_r_fine)   min_r_fine   = fabs(can[c].r_fine);
      c++;

    }


    /** add all base observations within prediction kernel to candidates
    +++ if they are more similar than the minimum relative difference
    +++ at the central pixel. This includes the central pixel (reset index
    +++ to 1). **/

    c = 1;

    for (ii=-r; ii<=r; ii++){
    for (jj=-r; jj<=r; jj++){

      ni = i+ii; nj = j+jj;
      if (ni < 0 || ni >= _NY_ || nj < 0 || nj >= _NX_) continue;
      np = ni*_NX_+nj;

      if (QAI.off[np]) continue;
      
      for (pair=0; pair<npair; pair++){

        diff = fabs(fine[pair][np] - fine[pair][p]);
        if (diff > slice_value[pair]) continue;

        can[c].dx = jj;
        can[c].dy = ii;
        can[c].coarse = coarse[pair][np];
        can[c].fine   = fine[pair][np];

        can[c].r_coarse = coarse[pair][np] - coarse[npair][np];
        can[c].r_fine   = coarse[pair][np] - fine[pair][np];

        if (fabs(can[c].r_coarse) < (min_r_coarse+unc_r_coarse) ||
            fabs(can[c].r_fine)   < (min_r_fine+unc_r_fine)) c++;

      }

    }
    }


    /** predict fine data from candidates based on 
    +++ spatial distance,
    +++ spectral distance (coarse-fine difference of input pairs) and 
    +++ 'temporal distance' (coarse-coarse difference of one base to
    +++ target image).
    +++ If there is no candidate, use coarse data. **/

    C_sum = 0.0;
    
    if ((ncan = c) > 1){

      for (c=1; c<ncan; c++){

          // location distance to central pixel
          d = sqrt(can[c].dx*can[c].dx + can[c].dy*can[c].dy);

          // temporal, spectral and combined distance
          // add a small uncertainty to avoid zero in T or S (avoid a 
          // biased large weight)

          T = fabs(can[c].r_coarse) + 0.0001;
          S = fabs(can[c].r_fine)   + 0.0001;
          TS = T*S;


          // use max weight if very close or compute combined inverted weight
          if (TS < unc_all){
            can[c].C = 1.0;
          } else {
            D = 1.0+d/r;
            can[c].C = 1.0/(TS*D);
          }

          // sum of all inverted weights
          C_sum += can[c].C;

      }

      for (c=1; c<ncan; c++){

        // normalized weight
        W = can[c].C/C_sum;

        // predict fine data using weighting function
        prediction[p] += W * (can[c].fine - can[c].r_coarse);

        #ifdef DEBUG
        percentsample[p] = (ncan/((r*2+1)*(r*2+1)*npair));
        #endif

      }
      
    } else {

      // retain coarse data if no suitable candidate
      prediction[p] = can[0].coarse;

    }


  }
  }

  free((void*)can);

  #ifdef DEBUG
  db_write_image_1D((void**)percentsample,  "PANSHARP-PSAM", _NX_, _NY_, sizeof(float), 4);
  free((void*)percentsample);
  #endif

  #ifdef CLOCK
  proctime_print("STARFM", TIME);
  #endif

  *pred = prediction;
  return SUCCESS;
}

