/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
This file contains functions for parsing parameter files
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#include "force.h"


/** Parse logical value
+++ This function takes a parameter in tag and value notation. It is tes-
+++ ted if the value is TRUE or FALSE. If the test was successful, the pa-
+++ rameter is set and a status flag is returned.
--- tag:   parameter tag
--- value: parameter value to be parsed
--- param: parsed parameter (returned)
+++ Return: true/false
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
bool parse_logical(char *tag, char *value, bool *param){

  if (strcmp(value, "TRUE\n") == 0){
    *param = true;
  } else if (strcmp(value, "FALSE\n") == 0){
    *param = false;
  } else {
    printf("\n%s is invalid.\n", tag);
    printf("valid values: TRUE/FALSE.\n");
    return false;
  }

  return true;
}


/** Parse float value
+++ This function takes a parameter in tag and value notation.  A scale 
+++ factor can be applied to the value. It is tested if the value is with-
+++ in the valid range. If the test was successful, the parameter is set.
--- tag:   parameter tag
--- value: parameter value to be parsed
--- param: parsed parameter (returned)
--- multi: scale factor
--- min:   allowed minimum value
--- max:   allowed maximum value
+++ Return: true/false
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
bool parse_float(char *tag, char *value, float *param, float multi, float min, float max){
float tmp;

  tmp = atof(value)*multi;
  if (tmp < min || tmp > max){
    printf("\n%s is invalid.\n", tag);
    printf("valid range: [%f,%f].\n", min/multi, max/multi);
    return false;
  } else {
    *param = tmp;
  }

  return true;
}


/** Parse integer value
+++ This function takes a parameter in tag and value notation.  A scale 
+++ factor can be applied to the value. It is tested if the value is with-
+++ in the valid range. If the test was successful, the parameter is set.
--- tag:   parameter tag
--- value: parameter value to be parsed
--- param: parsed parameter (returned)
--- multi: scale factor
--- min:   allowed minimum value
--- max:   allowed maximum value
+++ Return: true/false
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
bool parse_int(char *tag, char *value, int *param, int multi, int min, int max){
int tmp;

  tmp = atoi(value)*multi;
  if (tmp < min || tmp > max){
    printf("\n%s is invalid.\n", tag);
    printf("valid range: [%d,%d].\n", min/multi, max/multi);
    return false;
  } else {
    *param = tmp;
  }

  return true;
}


/** Parse sequence of char strings
+++ This function takes a parameter in tag and multi-value string notation
--- tag:   parameter tag
--- str:   tag/values string
--- param: parsed parameters (returned)
--- n:     number of parameters (returned)
+++ Return: true/false
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int parse_char_seq(char *tag, char *str, char param[MAX_STR_LEN][MAX_STR_LEN], int *n){
int num = 0;
char *ptr = NULL;
const char *separator = " =\n";

  ptr = strtok(str, separator);


  while((ptr = strtok(NULL, separator)) != NULL){
    strcpy(param[num++], ptr);
  }

  *n = num;
  return true;
}


/** Parse sequence of float values
+++ This function takes a parameter in tag and multi-value notation.  A 
+++ scale factor can be applied to the values. It is tested if the values
+++ are within the valid range. If the test was successful, the parameters
+++ are set.
--- tag:   parameter tag
--- str:   tag/values string
--- param: parsed parameters (returned)
--- n:     number of parameters (returned)
--- multi: scale factor
--- min:   allowed minimum value
--- max:   allowed maximum value
+++ Return: true/false
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
bool parse_float_seq(char *tag, char *str, float *param, int *n, float multi, float min, float max){
int num = 0;
float tmp;
char *ptr = NULL;
const char *separator = " =";

  ptr = strtok(str, separator);

  while((ptr = strtok(NULL, separator)) != NULL){

    tmp = atof(ptr)*multi;

    if (tmp < min || tmp > max){
      printf("\n%s is invalid.\n", tag);
      printf("valid range: [%f,%f].\n", min/multi, max/multi);
      return false;
    } else {
      param[num++] = tmp;
    }

  }

  *n = num;
  return true;
}


/** Parse sequence of integer values
+++ This function takes a parameter in tag and multi-value notation.  A 
+++ scale factor can be applied to the values. It is tested if the values
+++ are within the valid range. If the test was successful, the parameters
+++ are set.
--- tag:   parameter tag
--- str:   tag/values string
--- param: parsed parameters (returned)
--- n:     number of parameters (returned)
--- multi: scale factor
--- min:   allowed minimum value
--- max:   allowed maximum value
+++ Return: true/false
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
bool parse_int_seq(char *tag, char *str, int *param, int *n, int multi, int min, int max){
int tmp, num = 0;
char *ptr = NULL;
const char *separator = " =";

  ptr = strtok(str, separator);

  while((ptr = strtok(NULL, separator)) != NULL){

    tmp = atoi(ptr)*multi;

    if (tmp < min || tmp > max){
      printf("\n%s is invalid.\n", tag);
      printf("valid range: [%d,%d].\n", min/multi, max/multi);
      return false;
    } else {
      param[num++] = tmp;
    }

  }

  *n = num;
  return true;
}


/** Parse custom Quality Assurance Information rules
+++ This function sets the rules for QAI screening
--- request: requested QAI flags
+++ Return:  void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int parse_quality(char *request){

  if (strstr(request, "NODATA") != NULL) par.qai_off = true; else par.qai_off = false;
  if (strstr(request, "CLOUD_OPAQUE") != NULL) par.qai_cld_opq = true; else par.qai_cld_opq = false;
  if (strstr(request, "CLOUD_BUFFER") != NULL) par.qai_cld_unc = true; else par.qai_cld_unc = false;
  if (strstr(request, "CLOUD_CIRRUS") != NULL) par.qai_cld_cir = true; else par.qai_cld_cir = false;
  if (strstr(request, "CLOUD_SHADOW") != NULL) par.qai_shd = true; else par.qai_shd = false;
  if (strstr(request, "SNOW") != NULL) par.qai_snw = true; else par.qai_snw = false;
  if (strstr(request, "WATER") != NULL) par.qai_wtr = true; else par.qai_wtr = false;
  if (strstr(request, "AOD_FILL") != NULL) par.qai_aod_fill = true; else par.qai_aod_fill = false;
  if (strstr(request, "AOD_HIGH") != NULL) par.qai_aod_high = true; else par.qai_aod_high = false;
  if (strstr(request, "AOD_INT") != NULL) par.qai_aod_int = true; else par.qai_aod_int = false;
  if (strstr(request, "SUBZERO") != NULL) par.qai_sub = true; else par.qai_sub = false;
  if (strstr(request, "SATURATION") != NULL) par.qai_sat = true; else par.qai_sat = false;
  if (strstr(request, "SUN_LOW") != NULL) par.qai_sun = true; else par.qai_sun = false;
  if (strstr(request, "ILLUMIN_NONE") != NULL) par.qai_ill_shd = true; else par.qai_ill_shd = false;
  if (strstr(request, "ILLUMIN_POOR") != NULL) par.qai_ill_poor = true; else par.qai_ill_poor = false;
  if (strstr(request, "ILLUMIN_LOW") != NULL) par.qai_ill_low = true; else par.qai_ill_low = false;
  if (strstr(request, "SLOPED") != NULL) par.qai_slp = true; else par.qai_slp = false;
  if (strstr(request, "WVP_NONE") != NULL) par.qai_wvp = true; else par.qai_wvp = false;

  #ifdef DEBUG
  if (par.qai_off) printf("filter NODATA\n");
  if (par.qai_cld_opq) printf("filter CLOUD_OPAQUE\n");
  if (par.qai_cld_unc) printf("filter CLOUD_BUFFER\n");
  if (par.qai_cld_cir) printf("filter CLOUD_CIRRUS\n");
  if (par.qai_shd) printf("filter CLOUD_SHADOW\n");
  if (par.qai_snw) printf("filter SNOW\n");
  if (par.qai_wtr) printf("filter WATER\n");
  if (par.qai_aod_fill) printf("filter AOD_FILL\n");
  if (par.qai_aod_high) printf("filter AOD_HIGH\n");
  if (par.qai_aod_int) printf("filter AOD_INT\n");
  if (par.qai_sub) printf("filter SUBZERO\n");
  if (par.qai_sat) printf("filter SATURATION\n");
  if (par.qai_sun) printf("filter SUN_LOW\n");
  if (par.qai_ill_shd) printf("filter ILLUMIN_NONE\n");
  if (par.qai_ill_poor) printf("filter ILLUMIN_POOR\n");
  if (par.qai_ill_low) printf("filter ILLUMIN_LOW\n");
  if (par.qai_slp) printf("filter SLOPED\n");
  if (par.qai_wvp) printf("filter WVP_NONE\n");
  #endif

  return SUCCESS;
}


/** Initialize Level 2 parameters
+++ This function initializes all parameters with a fill value in order to
+++ enable a test if all were set by the user.
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void init_param_l2(){
char  cfill[] = "CFILL";
int   ifill   = -1;
float ffill   = -1.0;
bool  bfill   = false;

  strcpy(par.f_queue,  cfill);
  strcpy(par.f_tile,   cfill);
  strcpy(par.d_aod,    cfill);
  strcpy(par.d_level2, cfill);
  strcpy(par.d_tmp,    cfill);
  strcpy(par.d_wvp,    cfill);
  //strcpy(par.d_dob,    cfill);
  strcpy(par.css_dst,  cfill);
  strcpy(par.fdem,     cfill);

  par.tilesize_px = ifill;
  par.resample    = ifill;
  par.resmerge    = ifill;
  par.svgrid      = ifill;
  par.tier        = ifill;
  par.format      = ifill;

  par.vtopo   = bfill;
  par.vatmo   = bfill;
  par.vbrdf   = bfill;
  //par.vdodb   = bfill;
  par.vaod    = bfill;
  par.vreproj = bfill;
  par.vtile   = bfill;
  par.vmulti  = bfill;
  par.venv    = bfill;
  par.vocld   = bfill;
  par.voaod   = bfill;
  par.vowvp   = bfill;
  par.vovzn   = bfill;
  par.vohot   = bfill;

  par.res      = ffill;
  par.tilesize = ffill;
  par.maxcc    = ffill;
  par.maxtc    = ffill;
  par.cldprob  = ffill;
  par.shdprob  = ffill;
  par.wvp      = ffill;

  par.origin_geo.x = ffill*10000.0;
  par.origin_geo.y = ffill*10000.0;
  par.origin_map.x = ffill*10000.0;
  par.origin_map.y = ffill*10000.0;

  return;
}


/** Initialize Level 3 parameters
+++ This function initializes all parameters with a fill value in order to
+++ enable a test if all were set by the user.
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void init_param_l3(){
char  cfill[] = "CFILL";
int   ifill   = -1;
float ffill   = -1.0;
bool  bfill   = false;


  strcpy(par.d_level2, cfill);
  strcpy(par.d_lsp,    cfill);
  strcpy(par.d_level3, cfill);
  strcpy(par.p_lsp[0], cfill);
  strcpy(par.p_lsp[1], cfill);
  strcpy(par.p_lsp[2], cfill);
  strcpy(par.f_tile,   cfill);

  par.l2dict.n    = ifill;
  par.tminx       = ifill;
  par.tminy       = ifill;
  par.tmaxx       = ifill;
  par.tmaxy       = ifill;
  par.Yt          = ifill;
  par.Yr          = ifill;
  par.Dt[0]       = ifill;
  par.Dt[1]       = ifill;
  par.Dt[2]       = ifill;
  par.lsp_y0      = ifill;
  par.lsp_start   = ifill;
  par.cpu         = ifill;
  par.format      = ifill;

  par.res        = ffill;
  par.Yf         = ffill;
  par.s[0]       = ffill;
  par.s[1]       = ffill;
  par.s[2]       = ffill;
  par.lsp_rmse   = ffill;
  par.lsp_nodata = ffill;
  par.w.d        = ffill;
  par.w.y        = ffill;
  par.w.c        = ffill;
  par.w.h        = ffill;
  par.w.r        = ffill;
  par.w.v        = ffill;
  par.dreq       = ffill;
  par.vreq       = ffill;

  par.vqai    = bfill;
  par.vtile   = bfill;
  par.vpsf    = bfill;
  par.vimp    = bfill;
  par.vlsp    = bfill;
  par.voffsea = bfill;
  par.vobap   = bfill;
  par.voinf   = bfill;
  par.voscr   = bfill;
  par.voavg   = bfill;
  par.vostd   = bfill;
  par.vomin   = bfill;
  par.vomax   = bfill;
  par.vorng   = bfill;
  par.voskw   = bfill;
  par.vokrt   = bfill;
  par.voq25   = bfill;
  par.voq50   = bfill;
  par.voq75   = bfill;
  par.voiqr   = bfill;

  return;
}


/** Initialize TSA parameters
+++ This function initializes all parameters with a fill value in order to
+++ enable a test if all were set by the user.
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void init_param_tsa(){
int i;
char  cfill[] = "CFILL";
int   ifill   = -1;
float ffill   = -1.0;
bool  bfill   = false;

  strcpy(par.d_level2, cfill);
  strcpy(par.d_mask,   cfill);
  strcpy(par.d_tsa,    cfill);
  strcpy(par.f_tile,   cfill);
  strcpy(par.f_endmem, cfill);

  par.l2dict.n    = ifill;
  par.tminx       = ifill;
  par.tminy       = ifill;
  par.tmaxx       = ifill;
  par.tmaxy       = ifill;
  par.doymin      = ifill;
  par.doymax      = ifill;
  par.monthmin    = ifill;
  par.monthmax    = ifill;
  par.yearmin     = ifill;
  par.yearmax     = ifill;
  par.index       = ifill;
  par.trend       = ifill;
  par.tail        = ifill;
  par.fold        = ifill;
  par.cpu         = ifill;
  par.interpolate = ifill;
  par.int_nstep   = ifill;
  par.format      = ifill;
  par.mov_max     = ifill;
  par.rbf_nk      = ifill;
  for (i=0; i<MAX_STR_LEN; i++) par.rbf_sigma[i] = ifill;
  par.sma_emb     = ifill;
  par.lsp_dprev   = ifill;
  par.lsp_dnext   = ifill;
  par.lsp_hemi    = ifill;
  par.lsp_nseg    = ifill;

  par.res        = ffill;
  par.conf       = ffill;
  par.despike    = ffill;
  par.rbf_cutoff = ffill;

  par.vqai       = bfill;
  par.vpsf       = bfill;
  par.vimp       = bfill;
  par.vcenter    = bfill;
  par.vostandard = bfill;
  par.vsma_sto   = bfill;
  par.vsma_pos   = bfill;
  par.vsma_shn   = bfill;
  par.votss      = bfill;
  par.vorms      = bfill;
  par.vosta      = bfill;
  par.votsi      = bfill;
  par.vofby      = bfill;
  par.vofbm      = bfill;
  par.vofbw      = bfill;
  par.vofbd      = bfill;
  par.votrd      = bfill;
  par.vocat      = bfill;  
  par.volsp      = bfill;

  return;
}


/** Initialize CSO parameters
+++ This function initializes all parameters with a fill value in order to
+++ enable a test if all were set by the user.
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void init_param_cso(){
char  cfill[] = "CFILL";
int   ifill   = -1;
float ffill   = -1.0;
bool  bfill   = false;

  strcpy(par.d_level2, cfill);
  strcpy(par.d_cso,    cfill);
  strcpy(par.f_tile,   cfill);

  par.l2dict.n  = ifill;
  par.tminx     = ifill;
  par.tminy     = ifill;
  par.tmaxx     = ifill;
  par.tmaxy     = ifill;
  par.doymin    = ifill;
  par.doymax    = ifill;
  par.monthmin  = ifill;
  par.monthmax  = ifill;
  par.yearmin   = ifill;
  par.yearmax   = ifill;
  par.monthstep = ifill;
  par.cpu       = ifill;
  par.format    = ifill;

  par.res   = ffill;

  par.vqai    = bfill;
  par.vonum   = bfill;
  par.voavg   = bfill;
  par.vostd   = bfill;
  par.vomin   = bfill;
  par.vomax   = bfill;
  par.vorng   = bfill;
  par.voskw   = bfill;
  par.vokrt   = bfill;
  par.voq25   = bfill;
  par.voq50   = bfill;
  par.voq75   = bfill;
  par.voiqr   = bfill;

  return;
}


/** Initialize ImproPHE parameters
+++ This function initializes all parameters with a fill value in order to
+++ enable a test if all were set by the user.
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void init_param_imp(){
int i;
char  cfill[] = "CFILL";
int   ifill   = -1;
float ffill   = -1.0;
bool  bfill   = false;

  strcpy(par.d_level2,   cfill);
  strcpy(par.d_improphe, cfill);
  strcpy(par.f_tile,     cfill);
  for (i=0; i<MAX_STR_LEN; i++) strcpy(par.f_cr[i], cfill);
  for (i=0; i<MAX_STR_LEN; i++) strcpy(par.imp_prd[i], cfill);

  par.l2dict.n  = ifill;
  par.tminx     = ifill;
  par.tminy     = ifill;
  par.tmaxx     = ifill;
  par.tmaxy     = ifill;
  par.doymin    = ifill;
  par.doymax    = ifill;
  par.yearmin   = ifill;
  par.yearmax   = ifill;
  par.ymin      = ifill;
  par.ymax      = ifill;
  par.ksize     = ifill;
  par.ksd       = ifill;
  par.ystart_cr = ifill;
  par.nwin      = ifill;
  par.bwin      = ifill;
  par.num_cr    = ifill;
  par.cpu       = ifill;
  par.format   = ifill;
  for (i=0; i<MAX_STR_LEN; i++) par.dwin[i] = ifill;
  for (i=0; i<MAX_STR_LEN; i++) par.ywin[i] = ifill;

  par.res        = ffill;
  par.nodata_cr  = ffill;

  par.vqai    = bfill;
  par.vpsf    = bfill;

  return;
}


/** Initialize Level 2 ImproPhe parameters
+++ This function initializes all parameters with a fill value in order to
+++ enable a test if all were set by the user.
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void init_param_il2(){
int i;
char  cfill[] = "CFILL";
int   ifill   = -1;
float ffill   = -1.0;
bool  bfill   = false;

  strcpy(par.d_level2,   cfill);
  strcpy(par.f_tile,     cfill);

  par.l2dict.n  = ifill;
  par.tminx     = ifill;
  par.tminy     = ifill;
  par.tmaxx     = ifill;
  par.tmaxy     = ifill;
  par.ymin      = ifill;
  par.ymax      = ifill;
  par.doymin    = ifill;
  par.doymax    = ifill;
  par.monthmin  = ifill;
  par.monthmax  = ifill;
  par.yearmin   = ifill;
  par.yearmax   = ifill;
  par.ksize     = ifill;
  par.ksd       = ifill;
  par.nwin      = ifill;
  par.bwin      = ifill;
  par.cpu       = ifill;
  par.format   = ifill;
  for (i=0; i<MAX_STR_LEN; i++) par.dwin[i] = ifill;
  for (i=0; i<MAX_STR_LEN; i++) par.ywin[i] = ifill;

  par.res        = ffill;

  par.vqai    = bfill;
  par.vpsf    = bfill;
  par.vowr    = bfill;

  return;
}


/** Check Level 2 parameters
+++ This function checks if any parameter is still set to the initialized
+++ value. If this is the case, the function exits unsuccessful. In addi-
+++ tion, some test are performed to test for the existence of files and
+++ if incompatible parameters were set.
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int check_param_l2(){
char  cfill[] = "CFILL";
int   ifill   = -1;
float ffill   = -1.0;
bool  bfill   = false;

  if (par.vtopo   == bfill){
    printf("DO_TOPO is not specified."); return FAILURE;}
  if (par.vatmo   == bfill){
    printf("DO_ATMO is not specified."); return FAILURE;}
  if (par.vbrdf   == bfill){
    printf("DO_BRDF is not specified."); return FAILURE;}
  //if (par.vdodb   == bfill){
  //  printf("DO_DODB is not specified."); return FAILURE;}
  if (par.vaod    == bfill){
    printf("DO_AOD is not specified."); return FAILURE;}
  if (par.vreproj == bfill){
    printf("DO_REPROJ is not specified."); return FAILURE;}
  if (par.vtile   == bfill){
    printf("DO_TILE is not specified."); return FAILURE;}
  if (par.vmulti  == bfill){
    printf("MULTI_SCATTERING is not specified."); return FAILURE;}
  if (par.venv    == bfill){
    printf("ADJACENCY_EFFECT is not specified."); return FAILURE;}
  if (par.vocld == bfill){
    printf("OUTPUT_CLD is not specified."); return FAILURE;}
  if (par.voaod == bfill){
    printf("OUTPUT_AOD is not specified."); return FAILURE;}
  if (par.vowvp == bfill){
    printf("OUTPUT_WVP is not specified."); return FAILURE;}
  if (par.vovzn == bfill){
    printf("OUTPUT_VZN is not specified."); return FAILURE;}
  if (par.vohot == bfill){
    printf("OUTPUT_HOT is not specified."); return FAILURE;}

  if (par.resample == ifill){
    printf("RESAMPLING is not specified."); return FAILURE;}
  if (par.resmerge == ifill){
    printf("RES_MERGE is not specified."); return FAILURE;}
  if (par.svgrid == ifill){
    printf("SUN_VIEW_GRID is not specified."); return FAILURE;}
  if (par.tier   == ifill){
    printf("TIER is not specified."); return FAILURE;}
  if (par.format   == ifill){
    printf("OUTPUT_FORMAT is not specified."); return FAILURE;}

  if (par.res     == ffill){
    printf("RESOLUTION is not specified."); return FAILURE;}
  if (par.tilesize == ffill){
    printf("TILE_SIZE is not specified."); return FAILURE;}
  if (par.maxcc   == ffill){
    printf("MAX_CLOUD_COVER_FRAME is not specified."); return FAILURE;}
  if (par.maxcc   == 100) par.maxcc = 99;
  if (par.maxtc   == ffill){
    printf("MAX_CLOUD_COVER_TILE is not specified."); return FAILURE;}
  if (par.maxtc   == 100) par.maxtc = 99;
  if (par.cldprob == ffill){
    printf("CLOUD_THRESHOLD is not specified."); return FAILURE;}
  if (par.shdprob == ffill){
    printf("SHADOW_THRESHOLD is not specified."); return FAILURE;}
  if (par.wvp     == ffill){
    printf("WATER_VAPOR is not specified."); return FAILURE;}

  if (par.origin_geo.x == ffill*10000.0){
    printf("ORIGIN_LON is not specified."); return FAILURE;}
  if (par.origin_geo.y == ffill*10000.0){
    printf("ORIGIN_LAT is not specified."); return FAILURE;}

  if (strcmp(par.d_level2, cfill) == 0){
    printf("DIR_LEVEL2 is not specified."); return FAILURE;}
  if (strcmp(par.d_tmp, cfill) == 0){
    printf("DIR_TEMP is not specified."); return FAILURE;}
  if (strcmp(par.d_wvp, cfill) == 0){
    printf("DIR_WVPLUT is not specified."); return FAILURE;}
  if (strcmp(par.d_aod, cfill) == 0){
    printf("DIR_AOD is not specified."); return FAILURE;}
  //if (strcmp(par.d_dob, cfill) == 0){
  //  printf("DIR_DARKOB is not specified."); return FAILURE;}
  if (strcmp(par.fdem,  cfill) == 0){
    printf("FILE_DEM is not specified."); return FAILURE;}
  if (strcmp(par.f_tile,  cfill) == 0){
    printf("FILE_TILE is not specified."); return FAILURE;}
  if (strcmp(par.f_queue, cfill) == 0){
    printf("FILE_QUEUE is not specified."); return FAILURE;}

  if (!fileexist(par.d_level2)){
    printf("DIR_LEVEL2 does not exist in the file system."); return FAILURE;}
  if (!fileexist(par.d_tmp)){
    printf("DIR_TEMP does not exist in the file system."); return FAILURE;}
  if (par.doatmo){
    if (strcmp(par.d_wvp, "NULL") != 0 && !fileexist(par.d_wvp)){
      printf("DIR_WVPLUT does not exist in the file system. NULL to disable."); return FAILURE;
    }
  }
  if (par.doatmo){
    if (!par.doaod){
      if ((strcmp(par.d_aod, "NULL") == 0)){
        printf("DO_AOD = FALSE. DIR_AOD cannot be NULL."); return FAILURE;}
      if (!fileexist(par.d_aod)){
        printf("DIR_AOD does not exist in the file system."); return FAILURE;}
    } else {
      if ((strcmp(par.d_aod, "NULL") != 0) && !fileexist(par.d_aod)){
        printf("DIR_AOD does not exist in the file system. NULL to disable."); return FAILURE;}
    }
  }
  if (par.dotile){
    if ((strcmp(par.f_tile, "NULL") != 0) && !fileexist(par.f_tile)){
      printf("FILE_TILE does not exist in the file system. NULL to disable."); return FAILURE;}
  }
  if ((strcmp(par.f_queue, "NULL") != 0) && !fileexist(par.f_queue)){
      printf("FILE_QUEUE does not exist in the file system."); return FAILURE;}
  if (!par.doatmo) par.doaod = par.dobrdf = par.dotopo = par.doenv = false;
  //if (!par.doatmo) par.dododb = par.doaod = par.dobrdf = par.dotopo = par.doenv = false;
  //if (par.doatmo && !par.doaod) par.dododb = false;
  //if (par.dododb && !fileexist(par.d_dob)){
  //  printf("DIR_DARKOB does not exist in the file system."); return FAILURE;}

  if (par.dotopo && !fileexist(par.fdem)){
    printf("FILE_DEM does not exist in the file system. Give a DEM, or use DOTOPO = FALSE + FILE_DEM = NULL (surface will be assumed flat, z = 0m)."); return FAILURE;}
  if ((strcmp(par.fdem, "NULL") != 0) && !fileexist(par.fdem)){
    printf("FILE_DEM does not exist in the file system. NULL to disable (surface will be assumed flat, z = 0m)"); return FAILURE;}

  return SUCCESS;
}


/** Check Level 3 parameters
+++ This function checks if any parameter is still set to the initialized
+++ value. If this is the case, the function exits unsuccessful. In addi-
+++ tion, some test are performed to test for the existence of files and
+++ if incompatible parameters were set.
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int check_param_l3(){
char  cfill[] = "CFILL";
int   ifill   = -1;
float ffill   = -1.0;
bool  bfill   = false;

  if (strcmp(par.d_level2, cfill) == 0){
    printf("DIR_LEVEL2 is not specified."); return FAILURE;}
  if (strcmp(par.d_level3, cfill) == 0){
    printf("DIR_LEVEL3 is not specified."); return FAILURE;}
  if (strcmp(par.d_lsp, cfill) == 0){
    printf("DIR_LSP is not specified."); return FAILURE;}
  if (strcmp(par.p_lsp[0], cfill) == 0){
    printf("LSP_PATTERN_PAR_0 is not specified."); return FAILURE;}
  if (strcmp(par.p_lsp[0], cfill) == 0){
    printf("LSP_PATTERN_PAR_1 is not specified."); return FAILURE;}
  if (strcmp(par.p_lsp[0], cfill) == 0){
    printf("LSP_PATTERN_PAR_3 is not specified."); return FAILURE;}
  if (strcmp(par.f_tile,  cfill) == 0){
    printf("FILE_TILE is not specified."); return FAILURE;}

  if (par.l2dict.n == ifill){
    printf("SENSORS is not specified."); return FAILURE;}
  if (par.tminx       == ifill){
    printf("X_TILE_MIN is not specified."); return FAILURE;}
  if (par.tminy       == ifill){
    printf("Y_TILE_MIN is not specified."); return FAILURE;}
  if (par.tmaxx       == ifill){
    printf("X_TILE_MAX is not specified."); return FAILURE;}
  if (par.tmaxy       == ifill){
    printf("Y_TILE_MAX is not specified."); return FAILURE;}
  if (par.Yt       == ifill){
    printf("YEAR_TARGET is not specified."); return FAILURE;}
  if (par.Yr       == ifill){
    printf("YEAR_NUM is not specified."); return FAILURE;}
  if (par.Dt[0]    == ifill){
    printf("DOY_STATIC_0 is not specified."); return FAILURE;}
  if (par.Dt[1]    == ifill){
    printf("DOY_STATIC_1 is not specified."); return FAILURE;}
  if (par.Dt[2]    == ifill){
    printf("DOY_STATIC_2 is not specified."); return FAILURE;}
  if (par.lsp_y0   == ifill){
    printf("LSP_1ST_YEAR is not specified."); return FAILURE;}
  if (par.lsp_start   == ifill){
    printf("LSP_START is not specified."); return FAILURE;}
  if (par.cpu      == ifill){
    printf("NUM_CPU is not specified."); return FAILURE;}
  if (par.format   == ifill){
    printf("OUTPUT_FORMAT is not specified."); return FAILURE;}

  if (par.res     == ffill){
    printf("RESOLUTION is not specified."); return FAILURE;}
  if (par.Yf   == ffill){
    printf("Y_FACTOR is not specified."); return FAILURE;}
  if (par.s[0] == ffill){
    printf("DOY_SCORE_0 is not specified."); return FAILURE;}
  if (par.s[1] == ffill){
    printf("DOY_SCORE_1 is not specified."); return FAILURE;}
  if (par.s[2] == ffill){
    printf("DOY_SCORE_2 is not specified."); return FAILURE;}
  if (par.lsp_rmse == ffill){
    printf("LSP_THRESHOLD is not specified."); return FAILURE;}
  if (par.lsp_nodata == ffill){
    printf("LSP_NODATA is not specified."); return FAILURE;}
  if (par.w.d  == ffill){
    printf("SCORE_DOY_WEIGHT is not specified."); return FAILURE;}
  if (par.w.y  == ffill){
    printf("SCORE_YEAR_WEIGHT is not specified."); return FAILURE;}
  if (par.w.c  == ffill){
    printf("SCORE_CLOUD_WEIGHT is not specified."); return FAILURE;}
  if (par.w.h  == ffill){
    printf("SCORE_HAZE_WEIGHT is not specified."); return FAILURE;}
  if (par.w.r  == ffill){
    printf("SCORE_CORREL_WEIGHT is not specified."); return FAILURE;}
  if (par.w.v  == ffill){
    printf("SCORE_VZEN_WEIGHT is not specified."); return FAILURE;}
  if (par.dreq == ffill){
    printf("DREQ is not specified."); return FAILURE;}
  if (par.vreq == ffill){
    printf("VREQ is not specified."); return FAILURE;}

  if (par.vqai    == bfill){
    printf("SCREEN_QAI is not specified."); return FAILURE;}
  if (par.vpsf    == bfill){
    printf("REDUCE_PSF is not specified."); return FAILURE;}
  if (par.vimp    == bfill){
    printf("USE_IMPROPHE is not specified."); return FAILURE;}
  if (par.vtile   == bfill){
    printf("DO_TILE is not specified."); return FAILURE;}
  if (par.vlsp    == bfill){
    printf("LSP_DO is not specified."); return FAILURE;}
  if (par.voffsea == bfill){
    printf("OFF_SEASON is not specified."); return FAILURE;}
  if (par.vobap   == bfill){
    printf("OUTPUT_BAP is not specified."); return FAILURE;}
  if (par.voinf   == bfill){
    printf("OUTPUT_INF is not specified."); return FAILURE;}
  if (par.voscr   == bfill){
    printf("OUTPUT_SCR is not specified."); return FAILURE;}
  if (par.voavg   == bfill){
    printf("OUTPUT_AVG is not specified."); return FAILURE;}
  if (par.vostd   == bfill){
    printf("OUTPUT_STD is not specified."); return FAILURE;}
  if (par.vomin   == bfill){
    printf("OUTPUT_MIN is not specified."); return FAILURE;}
  if (par.vomax   == bfill){
    printf("OUTPUT_MAX is not specified."); return FAILURE;}
  if (par.vorng   == bfill){
    printf("OUTPUT_RNG is not specified."); return FAILURE;}
  if (par.voskw   == bfill){
    printf("OUTPUT_SKW is not specified."); return FAILURE;}
  if (par.vokrt   == bfill){
    printf("OUTPUT_KRT is not specified."); return FAILURE;}
  if (par.voq25   == bfill){
    printf("OUTPUT_Q25 is not specified."); return FAILURE;}
  if (par.voq50   == bfill){
    printf("OUTPUT_Q50 is not specified."); return FAILURE;}
  if (par.voq75   == bfill){
    printf("OUTPUT_Q75 is not specified."); return FAILURE;}
  if (par.voiqr   == bfill){
    printf("OUTPUT_IQR is not specified."); return FAILURE;}

  if (!fileexist(par.d_level2)){
    printf("DIR_LEVEL2 does not exist in the file system."); return FAILURE;}
  if (!fileexist(par.d_level3)){
    printf("DIR_LEVEL3 does not exist in the file system."); return FAILURE;}
  if ((strcmp(par.f_tile, "NULL") != 0) && !fileexist(par.f_tile)){
    printf("FILE_TILE does not exist in the file system. NULL to disable."); return FAILURE;}


  return SUCCESS;
}


/** Check TSA parameters
+++ This function checks if any parameter is still set to the initialized
+++ value. If this is the case, the function exits unsuccessful. In addi-
+++ tion, some test are performed to test for the existence of files and
+++ if incompatible parameters were set.
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int check_param_tsa(){
int i;
char  cfill[] = "CFILL";
int   ifill   = -1;
float ffill   = -1.0;
bool  bfill   = false;

  if (strcmp(par.d_level2, cfill) == 0){
    printf("DIR_LEVEL2 is not specified."); return FAILURE;}
  if (strcmp(par.d_mask, cfill) == 0){
    printf("DIR_MASK is not specified."); return FAILURE;}
  if (strcmp(par.d_tsa, cfill) == 0){
    printf("DIR_TSA is not specified."); return FAILURE;}
  if (strcmp(par.f_tile, cfill) == 0){
    printf("FILE_TILE is not specified."); return FAILURE;}
  if (strcmp(par.f_endmem, cfill) == 0){
    printf("FILE_ENDMEM is not specified."); return FAILURE;}

  if (par.l2dict.n == ifill){
    printf("SENSORS is not specified."); return FAILURE;}
  if (par.tminx    == ifill){
    printf("X_TILE_MIN is not specified."); return FAILURE;}
  if (par.tminy    == ifill){
    printf("Y_TILE_MIN is not specified."); return FAILURE;}
  if (par.tmaxx    == ifill){
    printf("X_TILE_MAX is not specified."); return FAILURE;}
  if (par.tmaxy    == ifill){
    printf("Y_TILE_MAX is not specified."); return FAILURE;}
  if (par.doymin   == ifill){
    printf("DOY_MIN is not specified."); return FAILURE;}
  if (par.doymax   == ifill){
    printf("DOY_MAX is not specified."); return FAILURE;}
  if (par.monthmin == ifill){
    printf("MONTH_MIN is not specified."); return FAILURE;}
  if (par.monthmax == ifill){
    printf("MONTH_MAX is not specified."); return FAILURE;}
  if (par.yearmin  == ifill){
    printf("YEAR_MIN is not specified."); return FAILURE;}
  if (par.yearmax  == ifill){
    printf("YEAR_MAX is not specified."); return FAILURE;}
  if (par.index    == ifill){
    printf("INDEX is not specified."); return FAILURE;}
  if (par.trend    == ifill){
    printf("TREND is not specified."); return FAILURE;}
  if (par.tail     == ifill){
    printf("TAIL is not specified."); return FAILURE;} else par.tail--;
  if (par.fold     == ifill){
    printf("FOLD is not specified."); return FAILURE;}
  if (par.cpu      == ifill){
    printf("NUM_CPU is not specified."); return FAILURE;}
  if (par.interpolate == ifill){
    printf("INTERPOLATE is not specified."); return FAILURE;}
  if (par.int_nstep == ifill){
    printf("INT_DAY is not specified."); return FAILURE;}
  if (par.format   == ifill){
    printf("OUTPUT_FORMAT is not specified."); return FAILURE;}
  if (par.mov_max == ifill){
    printf("MOVING_MAX is not specified."); return FAILURE;}
  if (par.rbf_nk == ifill){
    printf("RBF_SIGMA is not specified."); return FAILURE;}
  for (i=0; i<par.rbf_nk; i++){
    if (par.rbf_sigma[i] == ifill){
      printf("RBF_SIGMA is not specified."); return FAILURE;}
  }
  if (par.sma_emb == ifill){
    printf("SMA_ENDMEMBER is not specified."); return FAILURE;}
  if (par.lsp_dprev == ifill){
    printf("LSP_DOY_PREV_YEAR is not specified."); return FAILURE;}
  if (par.lsp_dnext == ifill){
    printf("LSP_DOY_NEXT_YEAR is not specified."); return FAILURE;}
  if (par.lsp_hemi == ifill){
    printf("LSP_HEMISPHERE is not specified."); return FAILURE;}
  if (par.lsp_nseg == ifill){
    printf("LSP_N_SEGMENT is not specified."); return FAILURE;}
  
  if (par.res  == ffill){
    printf("RESOLUTION is not specified."); return FAILURE;}
  if (par.conf == ffill){
    printf("CONF is not specified."); return FAILURE;}
  if (par.despike == ffill){
    printf("DESPIKE is not specified."); return FAILURE;}
  if (par.rbf_cutoff == ffill){
    printf("RBF_CUTOFF is not specified."); return FAILURE;}

  if (par.vqai    == bfill){
    printf("SCREEN_QAI is not specified."); return FAILURE;}
  if (par.vpsf    == bfill){
    printf("REDUCE_PSF is not specified."); return FAILURE;}
  if (par.vimp    == bfill){
    printf("USE_IMPROPHE is not specified."); return FAILURE;}
  if (par.vcenter   == bfill){
    printf("CENTER is not specified."); return FAILURE;}
  if (par.vostandard == bfill){
    printf("STANDARD is not specified."); return FAILURE;}
  if (par.vsma_sto == bfill){
    printf("SMA_SUM_TO_ONE is not specified."); return FAILURE;}
  if (par.vsma_pos == bfill){
    printf("SMA_NON_NEG is not specified."); return FAILURE;}
  if (par.vsma_shn == bfill){
    printf("SMA_SHD_NORM is not specified."); return FAILURE;}
  if (par.votss     == bfill){
    printf("OUTPUT_TSS is not specified."); return FAILURE;}
  if (par.vorms     == bfill){
    printf("OUTPUT_RMS is not specified."); return FAILURE;}
  if (par.vosta     == bfill){
    printf("OUTPUT_STA is not specified."); return FAILURE;}
  if (par.votsi     == bfill){
    printf("OUTPUT_TSI is not specified."); return FAILURE;}
  if (par.vofby     == bfill){
    printf("OUTPUT_FBY is not specified."); return FAILURE;}
  if (par.vofbm     == bfill){
    printf("OUTPUT_FBM is not specified."); return FAILURE;}
  if (par.vofbw     == bfill){
    printf("OUTPUT_FBW is not specified."); return FAILURE;}
  if (par.vofbd     == bfill){
    printf("OUTPUT_FBD is not specified."); return FAILURE;}
  if (par.votrd     == bfill){
    printf("OUTPUT_TRD is not specified."); return FAILURE;}
  if (par.vocat     == bfill){
    printf("OUTPUT_CAT is not specified."); return FAILURE;}
  if (par.volsp     == bfill){
    printf("OUTPUT_LSP is not specified."); return FAILURE;}

  if (par.index != 81 && par.orms){
    printf("Warning: Model residuals are only available if using SMA.\n");
    printf("         Setting OUTPUT_RMS = FALSE\n");
    par.orms = false;
  }

  if (par.fold == 3 && par.trend != 0){
    printf("If FOLD = LSP, TREND must be YEAR\n"); return FAILURE;}

  if (par.lsp && par.interpolate == NONE){
    printf("If FOLD = LSP or OUTPUT_LSP = TRUE, INTERPOLATE cannot be NONE\n"); return FAILURE;}

  if (par.index == 81 && !par.sma_pos && !par.sma_sto){
    printf("SMA_NON_NEG and SMA_SUM_TO_ONE cannot both be negative.\n"); return FAILURE;}
    

  #ifndef SPLITS
  if (par.fold == 3 || par.olsp || par.lsp){
    printf("Some option indicated to use SPLITS..\n");
    printf("SPLITS is not available.\n");
    printf("Install SPLITS and re-compile (see user guide)\n");
    printf(" OR use non-SPLITS methods.\n");
    return FAILURE;
  }
  #endif

  if (!fileexist(par.d_level2)){
    printf("DIR_LEVEL2 does not exist in the file system."); return FAILURE;}
  if ((strcmp(par.d_mask, "NULL") != 0) && !fileexist(par.d_mask)){
    printf("DIR_MASK does not exist in the file system."); return FAILURE;}
  if (!fileexist(par.d_tsa)){
    printf("DIR_TSA does not exist in the file system."); return FAILURE;}
  if ((strcmp(par.f_tile, "NULL") != 0) && !fileexist(par.f_tile)){
    printf("FILE_TILE does not exist in the file system. NULL to disable."); return FAILURE;}
  if ((strcmp(par.f_endmem, "NULL") != 0) && !fileexist(par.f_endmem)){
    printf("FILE_ENDMEM does not exist in the file system. NULL to disable."); return FAILURE;}
  if ((strcmp(par.f_endmem, "NULL") == 0) && par.index == 81){
    printf("FILE_ENDMEM cannot be NULL if INDEX = SMA."); return FAILURE;}


  return SUCCESS;
}


/** Check CSO parameters
+++ This function checks if any parameter is still set to the initialized
+++ value. If this is the case, the function exits unsuccessful. In addi-
+++ tion, some test are performed to test for the existence of files and
+++ if incompatible parameters were set.
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int check_param_cso(){
char  cfill[] = "CFILL";
int   ifill   = -1;
float ffill   = -1.0;
bool  bfill   = false;

  if (strcmp(par.d_level2, cfill) == 0){
    printf("DIR_LEVEL2 is not specified."); return FAILURE;}
  if (strcmp(par.d_cso, cfill) == 0){
    printf("DIR_CSO is not specified."); return FAILURE;}
  if (strcmp(par.f_tile,  cfill) == 0){
    printf("FILE_TILE is not specified."); return FAILURE;}

  if (par.l2dict.n == ifill){
    printf("SENSORS is not specified."); return FAILURE;}
  if (par.tminx    == ifill){
    printf("X_TILE_MIN is not specified."); return FAILURE;}
  if (par.tminy    == ifill){
    printf("Y_TILE_MIN is not specified."); return FAILURE;}
  if (par.tmaxx    == ifill){
    printf("X_TILE_MAX is not specified."); return FAILURE;}
  if (par.tmaxy    == ifill){
    printf("Y_TILE_MAX is not specified."); return FAILURE;}
  if (par.yearmin  == ifill){
    printf("YEAR_MIN is not specified."); return FAILURE;}
  if (par.yearmax  == ifill){
    printf("YEAR_MAX is not specified."); return FAILURE;}
  if (par.monthstep == ifill){
    printf("MONTH_STEP is not specified."); return FAILURE;}
  if (par.cpu      == ifill){
    printf("NUM_CPU is not specified."); return FAILURE;}
  if (par.format   == ifill){
    printf("OUTPUT_FORMAT is not specified."); return FAILURE;}

  if (par.res  == ffill){
    printf("RESOLUTION is not specified."); return FAILURE;}

  if (par.vqai    == bfill){
    printf("SCREEN_QAI is not specified."); return FAILURE;}
  if (par.vonum   == bfill){
    printf("OUTPUT_NUM is not specified."); return FAILURE;}
  if (par.voavg   == bfill){
    printf("OUTPUT_AVG is not specified."); return FAILURE;}
  if (par.vostd   == bfill){
    printf("OUTPUT_STD is not specified."); return FAILURE;}
  if (par.vomin   == bfill){
    printf("OUTPUT_MIN is not specified."); return FAILURE;}
  if (par.vomax   == bfill){
    printf("OUTPUT_MAX is not specified."); return FAILURE;}
  if (par.vorng   == bfill){
    printf("OUTPUT_RNG is not specified."); return FAILURE;}
  if (par.voskw   == bfill){
    printf("OUTPUT_SKW is not specified."); return FAILURE;}
  if (par.vokrt   == bfill){
    printf("OUTPUT_KRT is not specified."); return FAILURE;}
  if (par.voq25   == bfill){
    printf("OUTPUT_Q25 is not specified."); return FAILURE;}
  if (par.voq50   == bfill){
    printf("OUTPUT_Q50 is not specified."); return FAILURE;}
  if (par.voq75   == bfill){
    printf("OUTPUT_Q75 is not specified."); return FAILURE;}
  if (par.voiqr   == bfill){
    printf("OUTPUT_IQR is not specified."); return FAILURE;}

  if (!fileexist(par.d_level2)){
    printf("DIR_LEVEL2 does not exist in the file system."); return FAILURE;}
  if (!fileexist(par.d_cso)){
    printf("DIR_CSO does not exist in the file system."); return FAILURE;}
  if ((strcmp(par.f_tile, "NULL") != 0) && !fileexist(par.f_tile)){
    printf("FILE_TILE does not exist in the file system. NULL to disable."); return FAILURE;}


  return SUCCESS;
}


/** Check ImproPhe parameters
+++ This function checks if any parameter is still set to the initialized
+++ value. If this is the case, the function exits unsuccessful. In addi-
+++ tion, some test are performed to test for the existence of files and
+++ if incompatible parameters were set.
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int check_param_imp(){
int i;
char  cfill[] = "CFILL";
int   ifill   = -1;
float ffill   = -1.0;
bool  bfill   = false;


  if (strcmp(par.d_level2, cfill) == 0){
    printf("DIR_LEVEL2 is not specified."); return FAILURE;}
  if (strcmp(par.d_improphe, cfill) == 0){
    printf("DIR_IMPROPHE is not specified."); return FAILURE;}
  if (strcmp(par.f_tile,  cfill) == 0){
    printf("FILE_TILE is not specified."); return FAILURE;}

  if (par.num_cr    == ifill){
    printf("FILE_COARSE is not specified."); return FAILURE;}
  if (par.l2dict.n  == ifill){
    printf("SENSORS is not specified."); return FAILURE;}
  if (par.tminx     == ifill){
    printf("X_TILE_MIN is not specified."); return FAILURE;}
  if (par.tminy     == ifill){
    printf("Y_TILE_MIN is not specified."); return FAILURE;}
  if (par.tmaxx     == ifill){
    printf("X_TILE_MAX is not specified."); return FAILURE;}
  if (par.tmaxy     == ifill){
    printf("Y_TILE_MAX is not specified."); return FAILURE;}
  if (par.ymin      == ifill){
    printf("YEAR_MIN is not specified."); return FAILURE;}
  if (par.ymax      == ifill){
    printf("YEAR_MAX is not specified."); return FAILURE;}
  if (par.nwin      == ifill){
    printf("SEASONAL_WINDOW is not specified."); return FAILURE;}
  if (par.ystart_cr == ifill){
    printf("COARSE_1ST_YEAR is not specified."); return FAILURE;}
  if (par.ksize     == ifill){
    printf("KERNEL_SIZE is not specified."); return FAILURE;}
  if (par.ksd       == ifill){
    printf("KERNEL_TEXT is not specified."); return FAILURE;}
  if (par.cpu       == ifill){
    printf("NUM_CPU is not specified."); return FAILURE;}
  for (i=0; i<par.bwin; i++){
    if (par.dwin[i] == ifill){
      printf("SEASONAL_WINDOW breakpoint is not specified."); return FAILURE;}
  }
  if (par.format   == ifill){
    printf("OUTPUT_FORMAT is not specified."); return FAILURE;}

  if (par.res       == ffill){
    printf("RESOLUTION is not specified."); return FAILURE;}
  if (par.nodata_cr == ffill){
    printf("COARSE_NODATA is not specified."); return FAILURE;}

  if (par.vqai    == bfill){
    printf("SCREEN_QAI is not specified."); return FAILURE;}
  if (par.vpsf    == bfill){
    printf("REDUCE_PSF is not specified."); return FAILURE;}

  if (!fileexist(par.d_level2)){
    printf("DIR_LEVEL2 does not exist in the file system."); return FAILURE;}
  if (!fileexist(par.d_improphe)){
    printf("DIR_IMPROPHE does not exist in the file system."); return FAILURE;}
  if ((strcmp(par.f_tile, "NULL") != 0) && !fileexist(par.f_tile)){
    printf("FILE_TILE does not exist in the file system. NULL to disable."); return FAILURE;}
  for (i=0; i<par.num_cr; i++){
    if (!fileexist(par.f_cr[i])){
      printf("FILE_COARSE (%d) does not exist in the file system.", i+1); return FAILURE;}
    if (strcmp(par.imp_prd[i], cfill) == 0){
      printf("PRODUCT_TAG is not specified."); return FAILURE;}
  }

  #ifdef DEBUG
  printf("DIR_LEVEL2: %s\n", par.d_level2);
  printf("DIR_IMPROPHE: %s\n", par.d_improphe);
  printf("FILE_TILE: %s\n", par.f_tile);
  printf("number of coarse files: %02d\n", par.num_cr);
  for (i=0; i<par.num_cr; i++) printf("FILE_COARSE (%02d): %s\n", i, par.f_cr[i]);
  printf("number of sensors: %02d\n", par.l2dict.n);
  printf("tiles: X %04d-%04d, Y %04d-%04d\n", par.tminx, par.tmaxx, par.tminy, par.tmaxy);
  printf("read data from %04d-%04d %03d-%03d\n", par.yearmin, par.yearmax, par.doymin, par.doymax);
  printf("predict %04d-%04d\n", par.ymin, par.ymax);
  printf("predict at %f spatial resolution\n", par.res);
  printf("use PSF: %d\n", par.psf);
  printf("number of seasonal windows: %02d\n", par.nwin);
  for (i=0; i<par.bwin; i++) printf("DOY (rel. year): %03d (%+d)\n", par.dwin[i], par.ywin[i]);
  printf("kernel sizes: %d %d\n", par.ksize, par.ksd);
  printf("1st year of coarse data: %04d\n", par.ystart_cr);
  printf("nodata of coarse data: %f\n", par.nodata_cr);
  printf("number of cpus: %d\n", par.cpu);
  #endif

  return SUCCESS;
}


/** Check Level 2 ImproPhe parameters
+++ This function checks if any parameter is still set to the initialized
+++ value. If this is the case, the function exits unsuccessful. In addi-
+++ tion, some test are performed to test for the existence of files and
+++ if incompatible parameters were set.
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int check_param_il2(){
int i;
char  cfill[] = "CFILL";
int   ifill   = -1;
float ffill   = -1.0;
bool  bfill   = false;


  if (strcmp(par.d_level2, cfill) == 0){
    printf("DIR_LEVEL2 is not specified."); return FAILURE;}
  if (strcmp(par.f_tile,  cfill) == 0){
    printf("FILE_TILE is not specified."); return FAILURE;}
  if (strcmp(par.csen_mr,  cfill) == 0){
    printf("SENSORS_HR is not specified."); return FAILURE;}
  if (strcmp(par.csen_cr,  cfill) == 0){
    printf("SENSORS_MR is not specified."); return FAILURE;}

  if (par.l2dict.n  == ifill){
    printf("SENSORS_HR and SENSORS_MR are not specified."); return FAILURE;}
  if (par.tminx     == ifill){
    printf("X_TILE_MIN is not specified."); return FAILURE;}
  if (par.tminy     == ifill){
    printf("Y_TILE_MIN is not specified."); return FAILURE;}
  if (par.tmaxx     == ifill){
    printf("X_TILE_MAX is not specified."); return FAILURE;}
  if (par.tmaxy     == ifill){
    printf("Y_TILE_MAX is not specified."); return FAILURE;}
  if (par.ymin      == ifill){
    printf("YEAR is not specified."); return FAILURE;}
  if (par.ymax      == ifill){
    printf("YEAR is not specified."); return FAILURE;}
  if (par.nwin      == ifill){
    printf("SEASONAL_WINDOW is not specified."); return FAILURE;}
  if (par.bwin      == ifill){
    printf("SEASONAL_WINDOW is not specified."); return FAILURE;}
  if (par.ksize     == ifill){
    printf("KERNEL_SIZE is not specified."); return FAILURE;}
  if (par.ksd       == ifill){
    printf("KERNEL_TEXT is not specified."); return FAILURE;}
  if (par.cpu       == ifill){
    printf("NUM_CPU is not specified."); return FAILURE;}
  for (i=0; i<par.bwin; i++){
    if (par.dwin[i] == ifill){
      printf("SEASONAL_WINDOW breakpoint is not specified."); return FAILURE;}
  }
  if (par.format   == ifill){
    printf("OUTPUT_FORMAT is not specified."); return FAILURE;}

  if (par.res       == ffill){
    printf("RESOLUTION is not specified."); return FAILURE;}

  if (par.vqai    == bfill){
    printf("SCREEN_QAI is not specified."); return FAILURE;}
  if (par.vpsf    == bfill){
    printf("REDUCE_PSF is not specified."); return FAILURE;}
  if (par.vowr    == bfill){
    printf("OUTPUT_OVERWRITE is not specified."); return FAILURE;}

  if (!fileexist(par.d_level2)){
    printf("DIR_LEVEL2 does not exist in the file system."); return FAILURE;}
  if ((strcmp(par.f_tile, "NULL") != 0) && !fileexist(par.f_tile)){
    printf("FILE_TILE does not exist in the file system. NULL to disable."); return FAILURE;}


  #ifdef DEBUG
  printf("DIR_LEVEL2: %s\n", par.d_level2);
  printf("FILE_TILE: %s\n", par.f_tile);
  printf("number of sensors: %02d\n", par.l2dict.n);
  printf("MR sensors: %s\n", par.csen_cr);
  printf("HR sensors: %s\n", par.csen_mr);
  printf("tiles: X %04d-%04d, Y %04d-%04d\n", par.tminx, par.tmaxx, par.tminy, par.tmaxy);
  printf("predict data from %04d-%04d and %03d-%03d\n", par.yearmin, par.yearmax, par.doymin, par.doymax);
  printf("predict at %f spatial resolution\n", par.res);
  printf("use PSF: %d\n", par.psf);
  printf("number of seasonal windows: %02d\n", par.nwin);
  for (i=0; i<par.bwin; i++) printf("DOY (rel. year): %03d (%+d)\n", par.dwin[i], par.ywin[i]);
  printf("kernel sizes: %d %d\n", par.ksize, par.ksd);
  printf("output format: %d\n", par.format);
  printf("overwrite existing: %d\n", par.owr);
  printf("number of cpus: %d\n", par.cpu);
  #endif

  return SUCCESS;
}


/** Log custom Quality Assurance Information rules
+++ This function writes the QAI screening parameterization into the ENVI
+++ metadata hdr.
--- fp:     file handle for metadata
+++ Return:  void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int log_quality(FILE *fp){
char delim[MAX_STR_LEN] = "";

  fprintf(fp, "SCREEN_QAI:");
  if (par.qai_off){      fprintf(fp, "%sNODATA", delim);       strcpy(delim, "/");}
  if (par.qai_cld_unc){  fprintf(fp, "%sCLOUD_OPAQUE", delim); strcpy(delim, "/");}
  if (par.qai_cld_opq){  fprintf(fp, "%sCLOUD_BUFFER", delim); strcpy(delim, "/");}
  if (par.qai_cld_cir){  fprintf(fp, "%sCLOUD_CIRRUS", delim); strcpy(delim, "/");}
  if (par.qai_shd){      fprintf(fp, "%sCLOUD_SHADOW", delim); strcpy(delim, "/");}
  if (par.qai_snw){      fprintf(fp, "%sSNOW", delim);         strcpy(delim, "/");}
  if (par.qai_wtr){      fprintf(fp, "%sWATER", delim);        strcpy(delim, "/");}
  if (par.qai_aod_int){  fprintf(fp, "%sAOD_FILL", delim);     strcpy(delim, "/");}
  if (par.qai_aod_high){ fprintf(fp, "%sAOD_HIGH", delim);     strcpy(delim, "/");}
  if (par.qai_aod_fill){ fprintf(fp, "%sAOD_INT", delim);      strcpy(delim, "/");}
  if (par.qai_sub){      fprintf(fp, "%sSUBZERO", delim);      strcpy(delim, "/");}
  if (par.qai_sat){      fprintf(fp, "%sSATURATION", delim);   strcpy(delim, "/");}
  if (par.qai_sun){      fprintf(fp, "%sSUN_LOW", delim);      strcpy(delim, "/");}
  if (par.qai_ill_low){  fprintf(fp, "%sILLUMIN_NONE", delim); strcpy(delim, "/");}
  if (par.qai_ill_poor){ fprintf(fp, "%sILLUMIN_POOR", delim); strcpy(delim, "/");}
  if (par.qai_ill_shd){  fprintf(fp, "%sILLUMIN_LOW", delim);  strcpy(delim, "/");}
  if (par.qai_slp){      fprintf(fp, "%sSLOPED", delim);       strcpy(delim, "/");}
  if (par.qai_wvp){      fprintf(fp, "%sWVP_NONE", delim);     strcpy(delim, "/");}
  fprintf(fp, ", ");
  
  return SUCCESS;
}


/** Log Level 2 parameters
+++ This function writes the parameterization into the ENVI metadata hdr.
--- fp:     file handle for metadata
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int log_param_l2(FILE *fp){

  fprintf(fp, "FILE_QUEUE:%s, ",            par.f_queue);
  fprintf(fp, "DIR_LEVEL2:%s, ",            par.d_level2);
  fprintf(fp, "DIR_TEMP:%s, ",              par.d_tmp);
  fprintf(fp, "DIR_WVPLUT:%s, ",            par.d_wvp);
  fprintf(fp, "DIR_AOD:%s, ",               par.d_aod);
  //fprintf(fp, "DIR_DARKOB:%s, ",            par.d_dob);
  fprintf(fp, "FILE_DEM:%s, ",              par.fdem);
  fprintf(fp, "FILE_TILE:%s, ",             par.f_tile);
  fprintf(fp, "DO_REPROJ:%d, ",             par.doreproj);
  fprintf(fp, "DO_TILE:%d, ",               par.dotile);
  fprintf(fp, "TILE_SIZE:%f, ",             par.tilesize);
  fprintf(fp, "RESOLUTION:%f, ",            par.res);
  fprintf(fp, "ORIGIN_LAT:%f, ",            par.origin_geo.y);
  fprintf(fp, "ORIGIN_LON:%f, ",            par.origin_geo.x);
  fprintf(fp, "PROJECTION:%s, ",            par.css_dst);
  fprintf(fp, "RESAMPLING:%d, ",            par.resample);
  fprintf(fp, "RES_MERGE:%d, ",             par.resmerge);
  fprintf(fp, "TIER:%d, ",                  par.tier);
  fprintf(fp, "DO_TOPO:%d, ",               par.dotopo);
  fprintf(fp, "DO_ATMO:%d, ",               par.doatmo);
  fprintf(fp, "DO_BRDF:%d, ",               par.dobrdf);
  //fprintf(fp, "DO_DODB:%d, ",               par.dododb);
  fprintf(fp, "DO_AOD:%d, ",                par.doaod);
  fprintf(fp, "MULTI_SCATTERING:%d, ",      par.domulti);
  fprintf(fp, "ADJACENCY_EFFECT:%d, ",      par.doenv);
  fprintf(fp, "SUN_VIEW_GRID:%d, ",         par.svgrid);
  fprintf(fp, "WATER_VAPOR:%f, ",           par.wvp);
  fprintf(fp, "MAX_CLOUD_COVER_FRAME:%f, ", par.maxcc);
  fprintf(fp, "MAX_CLOUD_COVER_TILE:%f, ",  par.maxtc);
  fprintf(fp, "CLOUD_THRESHOLD:%f, ",       par.cldprob);
  fprintf(fp, "SHADOW_THRESHOLD:%f, ",      par.shdprob);
  fprintf(fp, "OUTPUT_FORMAT:%d, ",         par.format);
  fprintf(fp, "OUTPUT_CLD:%d, ",            par.ocld);
  fprintf(fp, "OUTPUT_AOD:%d, ",            par.oaod);
  fprintf(fp, "OUTPUT_WVP:%d, ",            par.owvp);
  fprintf(fp, "OUTPUT_VZN:%d, ",            par.ovzn);
  fprintf(fp, "OUTPUT_HOT:%d",              par.ohot);

  return SUCCESS;
}


/** Log Level 3 parameters
+++ This function writes the parameterization into the ENVI metadata hdr.
--- fp:     file handle for metadata
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int log_param_l3(FILE *fp){
int i;

  fprintf(fp, "DIR_LEVEL2:%s, ",          par.d_level2);
  fprintf(fp, "DIR_LSP:%s, ",             par.d_lsp);
  fprintf(fp, "DIR_LEVEL3:%s, ",          par.d_level3);
  fprintf(fp, "FILE_TILE:%s, ",           par.f_tile);
  fprintf(fp, "SENSORS:%s",               par.l2dict.sensor[0]);
  for (i=1; i<par.l2dict.n; i++) fprintf(fp, "/%s", par.l2dict.sensor[i]);
  fprintf(fp, ", ");
  log_quality(fp);
  fprintf(fp, "X_TILE_MIN:%d, ",          par.tminx);
  fprintf(fp, "Y_TILE_MIN:%d, ",          par.tminy);
  fprintf(fp, "X_TILE_MAX:%d, ",          par.tmaxx);
  fprintf(fp, "Y_TILE_MAX:%d, ",          par.tmaxy);
  fprintf(fp, "RESOLUTION:%f, ",          par.res);
  fprintf(fp, "REDUCE_PSF:%d, ",          par.psf);
  fprintf(fp, "USE_IMPROPHE:%d, ",        par.use.imp);
  fprintf(fp, "YEAR_TARGET:%d, ",         par.Yt);
  fprintf(fp, "YEAR_NUM:%d, ",            par.Yr);
  fprintf(fp, "Y_FACTOR:%f, ",            par.Yf);
  fprintf(fp, "DOY_SCORE_0:%f, ",         par.s[0]);
  fprintf(fp, "DOY_SCORE_1:%f, ",         par.s[1]);
  fprintf(fp, "DOY_SCORE_2:%f, ",         par.s[2]);
  fprintf(fp, "DOY_STATIC_0:%d, ",        par.Dt[0]);
  fprintf(fp, "DOY_STATIC_1:%d, ",        par.Dt[1]);
  fprintf(fp, "DOY_STATIC_2:%d, ",        par.Dt[2]);
  fprintf(fp, "OFF_SEASON:%d, ",          par.offsea);
  fprintf(fp, "LSP_DO:%d, ",              par.lsp);
  fprintf(fp, "LSP_PATTERN_PAR_0:%s, ",   par.p_lsp[0]);
  fprintf(fp, "LSP_PATTERN_PAR_1:%s, ",   par.p_lsp[1]);
  fprintf(fp, "LSP_PATTERN_PAR_2:%s, ",   par.p_lsp[2]);
  fprintf(fp, "LSP_1ST_YEAR:%d, ",        par.lsp_y0);
  fprintf(fp, "LSP_START:%d, ",           par.lsp_start);
  fprintf(fp, "LSP_THRESHOLD:%f, ",       par.lsp_rmse);
  fprintf(fp, "LSP_NODATA:%f, ",          par.lsp_nodata);
  fprintf(fp, "SCORE_DOY_WEIGHT:%f, ",    par.w.d);
  fprintf(fp, "SCORE_YEAR_WEIGHT:%f, ",   par.w.y);
  fprintf(fp, "SCORE_CLOUD_WEIGHT:%f, ",  par.w.c);
  fprintf(fp, "SCORE_HAZE_WEIGHT:%f, ",   par.w.h);
  fprintf(fp, "SCORE_CORREL_WEIGHT:%f, ", par.w.r);
  fprintf(fp, "SCORE_VZEN_WEIGHT:%f, ",   par.w.v);
  fprintf(fp, "DREQ:%f, ",                par.dreq);
  fprintf(fp, "VREQ:%f, ",                par.vreq);
  fprintf(fp, "NUM_CPU:%d, ",             par.cpu);
  fprintf(fp, "OUTPUT_FORMAT:%d, ",       par.format);
  fprintf(fp, "OUTPUT_BAP:%d, ",          par.obap);
  fprintf(fp, "OUTPUT_INF:%d, ",          par.oinf);
  fprintf(fp, "OUTPUT_SCR:%d, ",          par.oscr);
  fprintf(fp, "OUTPUT_AVG:%d, ",          par.oavg);
  fprintf(fp, "OUTPUT_STD:%d, ",          par.ostd);
  fprintf(fp, "OUTPUT_MIN:%d, ",          par.omin);
  fprintf(fp, "OUTPUT_MAX:%d, ",          par.omax);
  fprintf(fp, "OUTPUT_RNG:%d, ",          par.orng);
  fprintf(fp, "OUTPUT_SKW:%d, ",          par.oskw);
  fprintf(fp, "OUTPUT_KRT:%d, ",          par.okrt);
  fprintf(fp, "OUTPUT_Q25:%d, ",          par.oq25);
  fprintf(fp, "OUTPUT_Q50:%d, ",          par.oq50);
  fprintf(fp, "OUTPUT_Q75:%d, ",          par.oq75);
  fprintf(fp, "OUTPUT_IQR:%d",            par.oiqr);

  return SUCCESS;
}


/** Log TSA parameters
+++ This function writes the parameterization into the ENVI metadata hdr.
--- fp:     file handle for metadata
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int log_param_tsa(FILE *fp){
int i; 

  fprintf(fp, "DIR_LEVEL2:%s, ",        par.d_level2);
  fprintf(fp, "DIR_MASK:%s, ",          par.d_mask);
  fprintf(fp, "DIR_TSA:%s, ",           par.d_tsa);
  fprintf(fp, "FILE_TILE:%s, ",         par.f_tile);
  fprintf(fp, "FILE_ENDMEM :%s, ",      par.f_endmem);
  fprintf(fp, "SENSORS:%s", par.l2dict.sensor[0]);
  for (i=1; i<par.l2dict.n; i++) fprintf(fp, "/%s", par.l2dict.sensor[i]);
  fprintf(fp, ", ");
  log_quality(fp);
  fprintf(fp, "X_TILE_MIN:%d, ",        par.tminx);
  fprintf(fp, "Y_TILE_MIN:%d, ",        par.tminy);
  fprintf(fp, "X_TILE_MAX:%d, ",        par.tmaxx);
  fprintf(fp, "Y_TILE_MAX:%d, ",        par.tmaxy);
  fprintf(fp, "RESOLUTION:%f, ",        par.res);
  fprintf(fp, "REDUCE_PSF:%d, ",        par.psf);
  fprintf(fp, "USE_IMPROPHE:%d, ",      par.use.imp);
  fprintf(fp, "DOY_MIN:%d, ",           par.doymin);
  fprintf(fp, "DOY_MAX:%d, ",           par.doymax);
  fprintf(fp, "MONTH_MIN:%d, ",         par.doymin);
  fprintf(fp, "MONTH_MAX:%d, ",         par.doymax);
  fprintf(fp, "YEAR_MIN:%d, ",          par.yearmin);
  fprintf(fp, "YEAR_MAX:%d, ",          par.yearmax);
  fprintf(fp, "INDEX:%s, ",             par.index_c);
  fprintf(fp, "CENTER:%d, ",            par.docenter);
  fprintf(fp, "STANDARD:%d, ",          par.dostandard);
  fprintf(fp, "TREND:%s, ",             par.trend_c);
  fprintf(fp, "FOLD:%s, ",              par.fold_c);
  fprintf(fp, "TAIL:%s, ",              par.tail_c);
  fprintf(fp, "CONF:%f, ",              par.conf);
  fprintf(fp, "DESPIKE:%f, ",           par.despike);
  fprintf(fp, "SMA_SUM_TO_ONE:%d, ",    par.sma_sto);
  fprintf(fp, "SMA_NON_NEG:%d, ",       par.sma_pos);
  fprintf(fp, "SMA_SHD_NORM:%d, ",      par.sma_shn);
  fprintf(fp, "SMA_ENDMEMBER:%d, ",     par.sma_emb);
  fprintf(fp, "INTERPOLATE:%d, ",       par.interpolate);
  fprintf(fp, "MOVING_MAX:%d, ",        par.mov_max);
  fprintf(fp, "RBF_SIGMA:%d",           par.rbf_sigma[0]);
  for (i=1; i<par.rbf_nk; i++) fprintf(fp, "/%d", par.rbf_sigma[i]);
  fprintf(fp, ", ");
  fprintf(fp, "RBF_CUTOFF:%f, ",        par.rbf_cutoff);
  fprintf(fp, "INT_DAY:%d, ",           par.int_nstep);
  fprintf(fp, "LSP_DOY_PREV_YEAR:%d, ", par.lsp_dprev);
  fprintf(fp, "LSP_DOY_NEXT_YEAR:%d, ", par.lsp_dnext);
  fprintf(fp, "LSP_HEMISPHERE:%d, ",    par.lsp_hemi);
  fprintf(fp, "LSP_N_SEGMENT:%d, ",     par.lsp_nseg);
  
  fprintf(fp, "NUM_CPU:%d, ",           par.cpu);
  fprintf(fp, "OUTPUT_FORMAT:%d, ",     par.format);
  fprintf(fp, "OUTPUT_TSS:%d, ",        par.otss);
  fprintf(fp, "OUTPUT_RMS:%d, ",        par.orms);
  fprintf(fp, "OUTPUT_STA:%d, ",        par.osta);
  fprintf(fp, "OUTPUT_TSI:%d, ",        par.otsi);
  fprintf(fp, "OUTPUT_FBY:%d, ",        par.ofby);
  fprintf(fp, "OUTPUT_FBM:%d, ",        par.ofbm);
  fprintf(fp, "OUTPUT_FBW:%d, ",        par.ofbw);
  fprintf(fp, "OUTPUT_FBD:%d, ",        par.ofbd);
  fprintf(fp, "OUTPUT_TRD:%d, ",        par.otrd);
  fprintf(fp, "OUTPUT_CAT:%d, ",        par.ocat);
  fprintf(fp, "OUTPUT_LSP:%d",          par.olsp);

  return SUCCESS;
}


/** Log CSO parameters
+++ This function writes the parameterization into the ENVI metadata hdr.
--- fp:     file handle for metadata
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int log_param_cso(FILE *fp){
int i;

  fprintf(fp, "DIR_LEVEL2:%s, ",     par.d_level2);
  fprintf(fp, "DIR_CSO:%s, ",        par.d_cso);
  fprintf(fp, "FILE_TILE:%s, ",      par.f_tile);
  fprintf(fp, "SENSORS:%s", par.l2dict.sensor[0]);
  for (i=1; i<par.l2dict.n; i++) fprintf(fp, "/%s", par.l2dict.sensor[i]);
  fprintf(fp, ", ");
  log_quality(fp);
  fprintf(fp, "X_TILE_MIN:%d, ",     par.tminx);
  fprintf(fp, "Y_TILE_MIN:%d, ",     par.tminy);
  fprintf(fp, "X_TILE_MAX:%d, ",     par.tmaxx);
  fprintf(fp, "Y_TILE_MAX:%d, ",     par.tmaxy);
  fprintf(fp, "RESOLUTION:%f, ",     par.res);
  fprintf(fp, "YEAR_MIN:%d, ",       par.yearmin);
  fprintf(fp, "YEAR_MAX:%d, ",       par.yearmax);
  fprintf(fp, "MONTH_STEP:%d, ",     par.monthstep);
  fprintf(fp, "NUM_CPU:%d, ",        par.cpu);
  fprintf(fp, "OUTPUT_FORMAT:%d, ",  par.format);
  fprintf(fp, "OUTPUT_NUM:%d, ",     par.onum);
  fprintf(fp, "OUTPUT_AVG:%d, ",     par.oavg);
  fprintf(fp, "OUTPUT_STD:%d, ",     par.ostd);
  fprintf(fp, "OUTPUT_MIN:%d, ",     par.omin);
  fprintf(fp, "OUTPUT_MAX:%d, ",     par.omax);
  fprintf(fp, "OUTPUT_RNG:%d, ",     par.orng);
  fprintf(fp, "OUTPUT_SKW:%d, ",     par.oskw);
  fprintf(fp, "OUTPUT_KRT:%d, ",     par.okrt);
  fprintf(fp, "OUTPUT_Q25:%d, ",     par.oq25);
  fprintf(fp, "OUTPUT_Q50:%d, ",     par.oq50);
  fprintf(fp, "OUTPUT_Q75:%d, ",     par.oq75);
  fprintf(fp, "OUTPUT_IQR:%d",       par.oiqr);

  return SUCCESS;
}


/** Log ImproPhe parameters
+++ This function writes the parameterization into the ENVI metadata hdr.
--- fp:     file handle for metadata
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int log_param_imp(FILE *fp){
int i;

  fprintf(fp, "DIR_LEVEL2:%s, ",      par.d_level2);
  fprintf(fp, "FILE_COARSE:%s", par.f_cr[0]);
  for (i=1; i<par.num_cr; i++) fprintf(fp, "/%s", par.f_cr[i]);
  fprintf(fp, ", ");  
  fprintf(fp, "PRODUCT_TAG:%s", par.imp_prd[0]);
  for (i=1; i<par.num_cr; i++) fprintf(fp, "/%s", par.imp_prd[i]);
  fprintf(fp, ", ");  
  fprintf(fp, "DIR_IMPROPHE:%s, ",    par.d_improphe);
  fprintf(fp, "FILE_TILE :%s, ",      par.f_tile);
  fprintf(fp, "SENSORS:%s", par.l2dict.sensor[0]);
  for (i=1; i<par.l2dict.n; i++) fprintf(fp, "/%s", par.l2dict.sensor[i]);
  fprintf(fp, ", ");
  log_quality(fp);
  fprintf(fp, "X_TILE_MIN:%d, ",      par.tminx);
  fprintf(fp, "Y_TILE_MIN:%d, ",      par.tminy);
  fprintf(fp, "X_TILE_MAX:%d, ",      par.tmaxx);
  fprintf(fp, "Y_TILE_MAX:%d, ",      par.tmaxy);
  fprintf(fp, "RESOLUTION:%f, ",      par.res);
  fprintf(fp, "REDUCE_PSF:%d, ",      par.psf);
  fprintf(fp, "YEAR_MIN:%d, ",        par.ymin);
  fprintf(fp, "YEAR_MAX:%d, ",        par.ymax);
  fprintf(fp, "SEASONAL_WINDOW:%d",   par.dwin[0]);
  for (i=1; i<par.bwin; i++) fprintf(fp, "/%d", par.dwin[i]);
  fprintf(fp, ", ");
  fprintf(fp, "KERNEL_SIZE:%d, ",     par.ksize);
  fprintf(fp, "KERNEL_TEXT:%d, ",     par.ksd);
  fprintf(fp, "COARSE_1ST_YEAR:%d, ", par.ystart_cr);
  fprintf(fp, "COARSE_NODATA:%f, ",   par.nodata_cr);
  fprintf(fp, "NUM_CPU:%d, ",         par.cpu);
  fprintf(fp, "OUTPUT_FORMAT:%d",   par.format);

  return SUCCESS;
}


/** Log Level 2 ImproPhe parameters
+++ This function writes the parameterization into the ENVI metadata hdr.
--- fp:     file handle for metadata
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int log_param_il2(FILE *fp){
int i, nmr, ncr;
char str_mr[MAX_STR_LEN];
char str_cr[MAX_STR_LEN];
char mr[MAX_STR_LEN][MAX_STR_LEN];
char cr[MAX_STR_LEN][MAX_STR_LEN];


  strcpy(str_mr, par.csen_mr); parse_char_seq("SENSORS_HR", str_mr, mr, &nmr);
  strcpy(str_cr, par.csen_cr); parse_char_seq("SENSORS_MR", str_cr, cr, &ncr);

  fprintf(fp, "DIR_LEVEL2:%s, ",       par.d_level2);
  fprintf(fp, "FILE_TILE :%s, ",       par.f_tile);
  fprintf(fp, "SENSORS_MR:%s", cr[0]);
  for (i=1; i<ncr; i++) fprintf(fp, "/%s", cr[i]);
  fprintf(fp, ", ");
  fprintf(fp, "SENSORS_HR:%s", mr[0]);
  for (i=1; i<nmr; i++) fprintf(fp, "/%s", mr[i]);
  fprintf(fp, ", ");
  log_quality(fp);
  fprintf(fp, "X_TILE_MIN:%d, ",       par.tminx);
  fprintf(fp, "Y_TILE_MIN:%d, ",       par.tminy);
  fprintf(fp, "X_TILE_MAX:%d, ",       par.tmaxx);
  fprintf(fp, "Y_TILE_MAX:%d, ",       par.tmaxy);
  fprintf(fp, "RESOLUTION:%f, ",       par.res);
  fprintf(fp, "REDUCE_PSF:%d, ",       par.psf);
  fprintf(fp, "YEAR:%d, ",             par.ymin);
  fprintf(fp, "SEASONAL_WINDOW:%d",    par.dwin[0]);
  for (i=1; i<par.bwin; i++) fprintf(fp, "/%d", par.dwin[i]);
  fprintf(fp, ", ");
  fprintf(fp, "KERNEL_SIZE:%d, ",      par.ksize);
  fprintf(fp, "KERNEL_TEXT:%d, ",      par.ksd);
  fprintf(fp, "NUM_CPU:%d, ",          par.cpu);
  fprintf(fp, "OUTPUT_OVERWRITE:%d, ", par.owr);
  fprintf(fp, "OUTPUT_FORMAT:%d",    par.format);

  return SUCCESS;
}


/** Parse Level 2 parameters
+++ This function reads the parameter file.
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int parse_param_l2(){
FILE *fpar = NULL;
char  buffer[MAX_STR_LEN] = "\0";
char *tag = NULL;
char *ptr = NULL;
const char *separator = " =";
bool projfound = false;


  if (par.d_img != NULL){
    par.b_img = basename_without_ext(par.d_img);
    printf("%s: ", par.b_img);
  }

  init_param_l2();

  // open parameter file
  if ((fpar = fopen(par.f_par, "r")) == NULL){
    printf("Unable to open parameter file!\n"); return FAILURE;}

  if (fscanf(fpar, "%s", buffer) < 0){
    printf("No valid parameter file!\n"); return FAILURE;}
  if (strcmp(buffer, "++PARAM_LEVEL2_START++") != 0){
    printf("No valid parameter file!\n"); return FAILURE;}

  // process line by line
  while (fgets(buffer, MAX_STR_LEN, fpar) != NULL){

    if (strcmp(buffer, "++PARAM_LEVEL2_END++\n") == 0) break;

    if (projfound){
      strncpy(par.css_dst, buffer, strlen(buffer)-1);
      par.css_dst[strlen(buffer)-1] = '\0';
      projfound = false;
    }

    // get tag
    ptr = strtok(buffer, separator);
    tag = ptr;

    // continue if tag is comment
    if (strcmp(tag, "#") == 0) continue;

    // extract parameters by comparing tag
    while(ptr != NULL){

      ptr = strtok(NULL, separator);

      // directoriy variables
      if (strcmp(tag, "FILE_QUEUE") == 0){
        strncpy(par.f_queue, ptr, strlen(ptr)-1);
        par.f_queue[strlen(ptr)-1] = '\0';
      } else if (strcmp(tag, "DIR_LEVEL2") == 0){
        strncpy(par.d_level2, ptr, strlen(ptr)-1);
        par.d_level2[strlen(ptr)-1] = '\0';
      } else if (strcmp(tag, "DIR_TEMP") == 0){
        strncpy(par.d_tmp, ptr, strlen(ptr)-1);
        par.d_tmp[strlen(ptr)-1] = '\0';
      } else if (strcmp(tag, "DIR_WVPLUT") == 0){
        strncpy(par.d_wvp, ptr, strlen(ptr)-1);
        par.d_wvp[strlen(ptr)-1] = '\0';
      //} else if (strcmp(tag, "DIR_DARKOB") == 0){
      //  strncpy(par.d_dob, ptr, strlen(ptr)-1);
      //  par.d_dob[strlen(ptr)-1] = '\0';
      } else if (strcmp(tag, "DIR_AOD") == 0){
        strncpy(par.d_aod, ptr, strlen(ptr)-1);
        par.d_aod[strlen(ptr)-1] = '\0';
      } else if (strcmp(tag, "FILE_TILE") == 0){
        strncpy(par.f_tile, ptr, strlen(ptr)-1);
        par.f_tile[strlen(ptr)-1] = '\0';
      } else if (strcmp(tag, "FILE_DEM") == 0){
        strncpy(par.fdem, ptr, strlen(ptr)-1);
        par.fdem[strlen(ptr)-1] = '\0';

      // spatial variables
      } else if (strcmp(tag, "TILE_SIZE") == 0){
        par.tilesize = atof(ptr);
      } else if (strcmp(tag, "RESOLUTION") == 0){
        par.res = atof(ptr);
      } else if (strcmp(tag, "DO_REPROJ") == 0){
        par.vreproj = parse_logical("DO_REPROJ", ptr, &par.doreproj);
      } else if (strcmp(tag, "DO_TILE") == 0){
        par.vtile = parse_logical("DO_TILE", ptr, &par.dotile);
      } else if (strcmp(tag, "ORIGIN_LAT") == 0){
        parse_float("ORIGIN_LAT", ptr, &par.origin_geo.y, 1, -90, 90);
      } else if (strcmp(tag, "ORIGIN_LON") == 0){
        parse_float("ORIGIN_LON", ptr, &par.origin_geo.x, 1, -180, 180);
      } else if (strcmp(tag, "PROJECTION") == 0){
        projfound = true;
        continue;
      } else if (strcmp(tag, "RESAMPLING") == 0){
        if (strcmp(ptr, "NN\n") == 0){
          par.resample = 1;
        } else if (strcmp(ptr, "BL\n") == 0){
          par.resample = 2;
        } else if (strcmp(ptr, "CC\n") == 0){
          par.resample = 3;
        } else {
          printf("RESAMPLING in parameter file is invalid.\n");
          printf("valid values: NN/BL/CC.\n");
          return FAILURE;
        }
      } else if (strcmp(tag, "RES_MERGE") == 0){
        if (strcmp(ptr, "NONE\n") == 0){
          par.resmerge = NONE;
        } else if (strcmp(ptr, "REGRESSION\n") == 0){
          par.resmerge = REGRESSION;
        } else if (strcmp(ptr, "IMPROPHE\n") == 0){
          par.resmerge = IMPROPHE;
        } else if (strcmp(ptr, "STARFM\n") == 0){
          par.resmerge = STARFM;
        } else {
          printf("RES_MERGE in parameter file is invalid.\n");
          printf("valid values: IMPROPHE/REGRESSION/STARFM/NONE.\n");
          return FAILURE;
        }
      } else if (strcmp(tag, "TIER") == 0){
        parse_int("TIER", ptr, &par.tier, 1, 1, 2);

      // radiometric variables
      } else if (strcmp(tag, "DO_TOPO") == 0){
        par.vtopo = parse_logical("DO_TOPO", ptr, &par.dotopo);
      } else if (strcmp(tag, "DO_ATMO") == 0){
        par.vatmo = parse_logical("DO_ATMO", ptr, &par.doatmo);
      //} else if (strcmp(tag, "DO_DODB") == 0){
      //  par.vdodb = parse_logical("DO_DODB", ptr, &par.dododb);
      } else if (strcmp(tag, "DO_AOD") == 0){
        par.vaod = parse_logical("DO_AOD", ptr, &par.doaod);
      } else if (strcmp(tag, "DO_BRDF") == 0){
        par.vbrdf = parse_logical("DO_BRDF", ptr, &par.dobrdf);

      // advanced radiometric variables
      } else if (strcmp(tag, "MULTI_SCATTERING") == 0){
        par.vmulti = parse_logical("MULTI_SCATTERING", ptr, &par.domulti);
      } else if (strcmp(tag, "ADJACENCY_EFFECT") == 0){
        par.venv = parse_logical("ADJACENCY_EFFECT", ptr, &par.doenv);
      } else if (strcmp(tag, "SUN_VIEW_GRID") == 0){
        parse_int("SUN_VIEW_GRID", ptr, &par.svgrid, 1, 1, 200000);
      } else if (strcmp(tag, "WATER_VAPOR") == 0){
        parse_float("WATER_VAPOR", ptr, &par.wvp, 1, 0, 15);

      // cloud variables
      } else if (strcmp(tag, "MAX_CLOUD_COVER_FRAME") == 0){
        parse_float("MAX_CLOUD_COVER_FRAME", ptr, &par.maxcc, 1, 1, 100);
      } else if (strcmp(tag, "MAX_CLOUD_COVER_TILE") == 0){
        parse_float("MAX_CLOUD_COVER_TILE", ptr, &par.maxtc, 1, 1, 100);
      } else if (strcmp(tag, "CLOUD_THRESHOLD") == 0){
        parse_float("CLOUD_THRESHOLD", ptr, &par.cldprob, 1, 0, 1);
      } else if (strcmp(tag, "SHADOW_THRESHOLD") == 0){
        parse_float("SHADOW_THRESHOLD", ptr, &par.shdprob, 1, 0, 1);

      // output options
      } else if (strcmp(tag, "OUTPUT_FORMAT") == 0){
        if (strcmp(ptr, "ENVI\n") == 0){
          par.format = 0;
        } else if (strcmp(ptr, "GTiff\n") == 0){
          par.format = 1;
        } else {
          printf("OUTPUT_FORMAT in parameter file is invalid.\n");
          printf("valid values: ENVI/GTiff.\n");
          return FAILURE;
        }
      } else if (strcmp(tag, "OUTPUT_CLD") == 0){
        par.vocld = parse_logical("OUTPUT_CLD", ptr, &par.ocld);
      } else if (strcmp(tag, "OUTPUT_AOD") == 0){
        par.voaod = parse_logical("OUTPUT_AOD", ptr, &par.oaod);
      } else if (strcmp(tag, "OUTPUT_WVP") == 0){
        par.vowvp = parse_logical("OUTPUT_WVP", ptr, &par.owvp);
      } else if (strcmp(tag, "OUTPUT_VZN") == 0){
        par.vovzn = parse_logical("OUTPUT_VZN", ptr, &par.ovzn);
      } else if (strcmp(tag, "OUTPUT_HOT") == 0){
        par.vohot = parse_logical("OUTPUT_HOT", ptr, &par.ohot);
      }

      // in case tag is not the first word in a line
      tag = ptr;

    }

  }

  fclose(fpar);

  par.oref = par.oqai = true;

  if (check_param_l2() != SUCCESS){
    printf(" Unable to read parameter file!\n"); return FAILURE;}

  return SUCCESS;
}


/** Parse Level 3 parameters
+++ This function reads the parameter file.
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int parse_param_l3(){
FILE *fpar;
char  buffer[MAX_STR_LEN] = "\0";
char *tag = NULL;
char *ptr = NULL;
const char *separator = " =";
char  ctmp[MAX_STR_LEN];


  init_param_l3();

  // open parameter file
  if ((fpar = fopen(par.f_par, "r")) == NULL){
    printf("Unable to open parameter file!\n"); return FAILURE;}

  if (fscanf(fpar, "%s", buffer) < 0){
    printf("No valid parameter file!\n"); return FAILURE;}
  if(strcmp(buffer, "++PARAM_LEVEL3_START++") != 0){
    printf("No valid parameter file!\n"); return FAILURE;}

  // process line by line
  while (fgets(buffer, MAX_STR_LEN, fpar) != NULL){

    if(strcmp(buffer, "++PARAM_LEVEL3_END++\n") == 0) break;
    
    strcpy(ctmp, buffer);

    // get tag
    ptr = strtok(buffer, separator);
    tag = ptr;

    // continue if tag is comment
    if(strcmp(tag,"#") == 0) continue;

    // extract parameters by comparing tag
    while(ptr != NULL){

      ptr = strtok(NULL, separator);

      // directory variables
      if (strcmp(tag, "DIR_LEVEL2") == 0){
        strncpy(par.d_level2, ptr, strlen(ptr)-1);
        par.d_level2[strlen(ptr)-1] = '\0';
      } else if (strcmp(tag, "DIR_LSP") == 0){
        strncpy(par.d_lsp, ptr, strlen(ptr)-1);
        par.d_lsp[strlen(ptr)-1] = '\0';
      } else if (strcmp(tag, "DIR_LEVEL3") == 0){
        strncpy(par.d_level3, ptr, strlen(ptr)-1);
        par.d_level3[strlen(ptr)-1] = '\0';
      } else if (strcmp(tag, "FILE_TILE") == 0){
        strncpy(par.f_tile, ptr, strlen(ptr)-1);
        par.f_tile[strlen(ptr)-1] = '\0';

      // sensors
      } else if (strcmp(tag, "SENSORS") == 0){
        if (level2sensor_dict(ctmp, &par.l2dict) != SUCCESS) exit(1);

      // QAI screening
      } else if (strcmp(tag, "SCREEN_QAI") == 0){
        if (parse_quality(ctmp) != SUCCESS) exit(1);
        par.vqai = true;

      // spatial variables
      } else if (strcmp(tag, "X_TILE_MIN") == 0){
        parse_int("X_TILE_MIN", ptr, &par.tminx, 1, -999, 9999);
      } else if (strcmp(tag, "Y_TILE_MIN") == 0){
        parse_int("Y_TILE_MIN", ptr, &par.tminy, 1, -999, 9999);
      } else if (strcmp(tag, "X_TILE_MAX") == 0){
        parse_int("X_TILE_MAX", ptr, &par.tmaxx, 1, -999, 9999);
      } else if (strcmp(tag, "Y_TILE_MAX") == 0){
        parse_int("Y_TILE_MAX", ptr, &par.tmaxy, 1, -999, 9999);
      } else if (strcmp(tag, "DO_TILE") == 0){
        par.vtile = parse_logical("DO_TILE", ptr, &par.dotile);
      } else if (strcmp(tag, "RESOLUTION") == 0){
        parse_float("RESOLUTION", ptr, &par.res, 1, 0, FLT_MAX);
      } else if (strcmp(tag, "REDUCE_PSF") == 0){
        par.vpsf = parse_logical("REDUCE_PSF", ptr, &par.psf);
      } else if (strcmp(tag, "USE_IMPROPHE") == 0){
        par.vimp = parse_logical("USE_IMPROPHE", ptr, &par.use.imp);

      // temporal variables
      } else if (strcmp(tag, "YEAR_TARGET") == 0){
        parse_int("YEAR_TARGET", ptr, &par.Yt, 1, 1900, 2100);
      } else if (strcmp(tag, "YEAR_NUM") == 0){
        parse_int("YEAR_NUM", ptr, &par.Yr, 1, 0, 100);
      } else if (strcmp(tag, "Y_FACTOR") == 0){
        parse_float("Y_FACTOR", ptr, &par.Yf, 1, 0, FLT_MAX);

      // phenology variables
      } else if (strcmp(tag, "LSP_DO") == 0){
        par.vlsp = parse_logical("LSP_DO", ptr, &par.lsp);
      } else if (strcmp(tag, "LSP_PATTERN_PAR_0") == 0){
        strncpy(par.p_lsp[0], ptr, strlen(ptr)-1);
        par.p_lsp[0][strlen(ptr)-1] = '\0';
      } else if (strcmp(tag, "LSP_PATTERN_PAR_1") == 0){
        strncpy(par.p_lsp[1], ptr, strlen(ptr)-1);
        par.p_lsp[1][strlen(ptr)-1] = '\0';
      } else if (strcmp(tag, "LSP_PATTERN_PAR_2") == 0){
        strncpy(par.p_lsp[2], ptr, strlen(ptr)-1);
        par.p_lsp[2][strlen(ptr)-1] = '\0';
      } else if (strcmp(tag, "LSP_1ST_YEAR") == 0){
        parse_int("LSP_1ST_YEAR", ptr, &par.lsp_y0, 1, 1900, 2100);
      } else if (strcmp(tag, "LSP_START") == 0){
        parse_int("LSP_START", ptr, &par.lsp_start, 1, 1, 2100*365);
      } else if (strcmp(tag, "LSP_THRESHOLD") == 0){
        parse_float("LSP_THRESHOLD", ptr, &par.lsp_rmse, 1, 0, 365);
      } else if (strcmp(tag, "LSP_NODATA") == 0){
        parse_float("LSP_NODATA", ptr, &par.lsp_nodata, 1, LONG_MIN, LONG_MAX);

      // scoring variables
      } else if (strcmp(tag, "DOY_SCORE_0") == 0){
        parse_float("DOY_SCORE_0", ptr, &par.s[0], 1, 0, 1);
      } else if (strcmp(tag, "DOY_SCORE_1") == 0){
        parse_float("DOY_SCORE_1", ptr, &par.s[1], 1, 0, 1);
      } else if (strcmp(tag, "DOY_SCORE_2") == 0){
        parse_float("DOY_SCORE_2", ptr, &par.s[2], 1, 0, 1);
      } else if (strcmp(tag, "DOY_STATIC_0") == 0){
        parse_int("DOY_STATIC_0", ptr, &par.Dt[0], 1, 0, 365);
      } else if (strcmp(tag, "DOY_STATIC_1") == 0){
        parse_int("DOY_STATIC_1", ptr, &par.Dt[1], 1, 0, 365);
      } else if (strcmp(tag, "DOY_STATIC_2") == 0){
        parse_int("DOY_STATIC_2", ptr, &par.Dt[2], 1, 0, 365);
      } else if (strcmp(tag, "OFF_SEASON") == 0){
        par.voffsea = parse_logical("OFF_SEASON", ptr, &par.offsea);
      } else if (strcmp(tag, "DREQ") == 0){
        parse_float("DREQ", ptr, &par.dreq, 1, 1, FLT_MAX);
      } else if (strcmp(tag, "VREQ") == 0){
        parse_float("DREQ", ptr, &par.vreq, 1, 1, 90);

      // compositing scores
      } else if (strcmp(tag, "SCORE_DOY_WEIGHT") == 0){
        parse_float("SCORE_DOY_WEIGHT", ptr, &par.w.d, 1, 0, 1);
      } else if (strcmp(tag, "SCORE_YEAR_WEIGHT") == 0){
        parse_float("SCORE_YEAR_WEIGHT", ptr, &par.w.y, 1, 0, 1);
      } else if (strcmp(tag, "SCORE_CLOUD_WEIGHT") == 0){
        parse_float("SCORE_CLOUD_WEIGHT", ptr, &par.w.c, 1, 0, 1);
      } else if (strcmp(tag, "SCORE_HAZE_WEIGHT") == 0){
        parse_float("SCORE_HAZE_WEIGHT", ptr, &par.w.h, 1, 0, 1);
      } else if (strcmp(tag, "SCORE_CORREL_WEIGHT") == 0){
        parse_float("SCORE_CORREL_WEIGHT", ptr, &par.w.r, 1, 0, 1);
      } else if (strcmp(tag, "SCORE_VZEN_WEIGHT") == 0){
        parse_float("SCORE_VZEN_WEIGHT", ptr, &par.w.v, 1, 0, 1);

      // other variables
      } else if (strcmp(tag, "NUM_CPU") == 0){
        parse_int("NUM_CPU", ptr, &par.cpu, 1, 0, INT_MAX);

      // output options
      } else if (strcmp(tag, "OUTPUT_FORMAT") == 0){
        if (strcmp(ptr, "ENVI\n") == 0){
          par.format = 0;
        } else if (strcmp(ptr, "GTiff\n") == 0){
          par.format = 1;
        } else {
          printf("OUTPUT_FORMAT in parameter file is invalid.\n");
          printf("valid values: ENVI/GTiff.\n");
          return FAILURE;
        }
      } else if (strcmp(tag, "OUTPUT_BAP") == 0){
        par.vobap = parse_logical("OUTPUT_BAP", ptr, &par.obap);
      } else if (strcmp(tag, "OUTPUT_INF") == 0){
        par.voinf = parse_logical("OUTPUT_INF", ptr, &par.oinf);
      } else if (strcmp(tag, "OUTPUT_SCR") == 0){
        par.voscr = parse_logical("OUTPUT_SCR", ptr, &par.oscr);
      } else if (strcmp(tag, "OUTPUT_AVG") == 0){
        par.voavg = parse_logical("OUTPUT_AVG", ptr, &par.oavg);
      } else if (strcmp(tag, "OUTPUT_STD") == 0){
        par.vostd = parse_logical("OUTPUT_STD", ptr, &par.ostd);
      } else if (strcmp(tag, "OUTPUT_MIN") == 0){
        par.vomin = parse_logical("OUTPUT_MIN", ptr, &par.omin);
      } else if (strcmp(tag, "OUTPUT_MAX") == 0){
        par.vomax = parse_logical("OUTPUT_MAX", ptr, &par.omax);
      } else if (strcmp(tag, "OUTPUT_RNG") == 0){
        par.vorng = parse_logical("OUTPUT_RNG", ptr, &par.orng);
      } else if (strcmp(tag, "OUTPUT_SKW") == 0){
        par.voskw = parse_logical("OUTPUT_SKW", ptr, &par.oskw);
      } else if (strcmp(tag, "OUTPUT_KRT") == 0){
        par.vokrt = parse_logical("OUTPUT_KRT", ptr, &par.okrt);
      } else if (strcmp(tag, "OUTPUT_Q25") == 0){
        par.voq25 = parse_logical("OUTPUT_Q25", ptr, &par.oq25);
      } else if (strcmp(tag, "OUTPUT_Q50") == 0){
        par.voq50 = parse_logical("OUTPUT_Q50", ptr, &par.oq50);
      } else if (strcmp(tag, "OUTPUT_Q75") == 0){
        par.voq75 = parse_logical("OUTPUT_Q75", ptr, &par.oq75);
      } else if (strcmp(tag, "OUTPUT_IQR") == 0){
        par.voiqr = parse_logical("OUTPUT_IQR", ptr, &par.oiqr);
      }

      // in case tag is not the first word in a line
      tag = ptr;

    }

  }

  fclose(fpar);

  // total scoring weight
  par.w.t = par.w.d + par.w.y + par.w.c + 
               par.w.h + par.w.r + par.w.v;

  // number of tiles
  par.tnx = par.tmaxx-par.tminx+1;
  par.tny = par.tmaxy-par.tminy+1;

  // number of years
  par.Yn = (par.Yr*2)+1;
  par.yearmin = 1900; par.yearmax = 2100;
  par.doymin = 1; par.doymax = 366;

  // choose type of scoring function
  if (par.s[1] > par.s[0] && 
      par.s[1] > par.s[2]){
    par.score_type = SCR_GAUSS; // gaussian
  } else if (par.s[0] > par.s[2]){
    par.score_type = SCR_SIG_DES; // descending sigmoid
  } else if (par.s[2] > par.s[0]){
    par.score_type = SCR_SIG_ASC; // ascending sigmoid
  } else { printf("DOY scoring function type is undefined!\n"); exit(1);}

  // choose products
  par.use.ref = true;
  par.use.qai = true;
  if (par.w.c > 0) par.use.cld = true; else par.use.cld = false;
  if (par.w.h > 0) par.use.hot = true; else par.use.hot = false;
  if (par.w.v > 0) par.use.vzn = true; else par.use.vzn = false;
  par.use.aod = false;
  par.use.wvp = false;


  if (check_param_l3() != SUCCESS){
    printf(" Unable to read parameter file!\n"); return FAILURE;}

  return SUCCESS;
}


/** Parse TSA parameters
+++ This function reads the parameter file.
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int parse_param_tsa(){
FILE *fpar;
char  buffer[MAX_STR_LEN] = "\0";
char *tag = NULL;
char *ptr = NULL;
const char *separator = " =";
char  ctmp[MAX_STR_LEN];
int m, d, doy;
int b, s;
bool *valid = NULL;


  init_param_tsa();

  // open parameter file
  if ((fpar = fopen(par.f_par, "r")) == NULL){
    printf("Unable to open parameter file!\n"); return FAILURE;}

  if (fscanf(fpar, "%s", buffer) < 0){
    printf("No valid parameter file!\n"); return FAILURE;}
  if(strcmp(buffer, "++PARAM_TSA_START++") != 0){
    printf("No valid parameter file!\n"); return FAILURE;}

  // process line by line
  while (fgets(buffer, MAX_STR_LEN, fpar) != NULL){

    if(strcmp(buffer, "++PARAM_TSA_END++\n") == 0) break;
    
    strcpy(ctmp, buffer);

    // get tag
    ptr = strtok(buffer, separator);
    tag = ptr;

    // continue if tag is comment
    if(strcmp(tag,"#") == 0) continue;

    // extract parameters by comparing tag
    while(ptr != NULL){

      ptr = strtok(NULL, separator);

      // directory variables
      if (strcmp(tag, "DIR_LEVEL2") == 0){
        strncpy(par.d_level2, ptr, strlen(ptr)-1);
        par.d_level2[strlen(ptr)-1] = '\0';
      } else if (strcmp(tag, "DIR_MASK") == 0){
        strncpy(par.d_mask, ptr, strlen(ptr)-1);
        par.d_mask[strlen(ptr)-1] = '\0';
      } else if (strcmp(tag, "DIR_TSA") == 0){
        strncpy(par.d_tsa, ptr, strlen(ptr)-1);
        par.d_tsa[strlen(ptr)-1] = '\0';
      } else if (strcmp(tag, "FILE_TILE") == 0){
        strncpy(par.f_tile, ptr, strlen(ptr)-1);
        par.f_tile[strlen(ptr)-1] = '\0';
      } else if (strcmp(tag, "FILE_ENDMEM") == 0){
        strncpy(par.f_endmem, ptr, strlen(ptr)-1);
        par.f_endmem[strlen(ptr)-1] = '\0';

      // sensors
      } else if (strcmp(tag, "SENSORS") == 0){
        if (level2sensor_dict(ctmp, &par.l2dict) != SUCCESS) exit(1);

      // QAI screening
      } else if (strcmp(tag, "SCREEN_QAI") == 0){
        if (parse_quality(ctmp) != SUCCESS) exit(1);
        par.vqai = true;

      // spatial variables
      } else if (strcmp(tag, "X_TILE_MIN") == 0){
        parse_int("X_TILE_MIN", ptr, &par.tminx, 1, -999, 9999);
      } else if (strcmp(tag, "Y_TILE_MIN") == 0){
        parse_int("Y_TILE_MIN", ptr, &par.tminy, 1, -999, 9999);
      } else if (strcmp(tag, "X_TILE_MAX") == 0){
        parse_int("X_TILE_MAX", ptr, &par.tmaxx, 1, -999, 9999);
      } else if (strcmp(tag, "Y_TILE_MAX") == 0){
        parse_int("Y_TILE_MAX", ptr, &par.tmaxy, 1, -999, 9999);
      } else if (strcmp(tag, "RESOLUTION") == 0){
        parse_float("RESOLUTION", ptr, &par.res, 1, 0, FLT_MAX);
      } else if (strcmp(tag, "REDUCE_PSF") == 0){
        par.vpsf = parse_logical("REDUCE_PSF", ptr, &par.psf);
      } else if (strcmp(tag, "USE_IMPROPHE") == 0){
        par.vimp = parse_logical("USE_IMPROPHE", ptr, &par.use.imp);

      // temporal variables
      } else if (strcmp(tag, "DOY_MIN") == 0){
        parse_int("DOY_MIN", ptr, &par.doymin, 1, 1, 366);
      } else if (strcmp(tag, "DOY_MAX") == 0){
        parse_int("DOY_MAX", ptr, &par.doymax, 1, 1, 366);
      } else if (strcmp(tag, "MONTH_MIN") == 0){
        parse_int("MONTH_MIN", ptr, &par.monthmin, 1, 1, 12);
      } else if (strcmp(tag, "MONTH_MAX") == 0){
        parse_int("MONTH_MAX", ptr, &par.monthmax, 1, 1, 12);
      } else if (strcmp(tag, "YEAR_MIN") == 0){
        parse_int("YEAR_MIN", ptr, &par.yearmin, 1, 1900, 2100);
      } else if (strcmp(tag, "YEAR_MAX") == 0){
        parse_int("YEAR_MAX", ptr, &par.yearmax, 1, 1900, 2100);

      // time series parameters
      } else if (strcmp(tag, "INDEX") == 0){
        if (strcmp(ptr, "BLUE\n") == 0){
          par.index = 0; strcpy(par.index_c, "BLU");
        } else if (strcmp(ptr, "GREEN\n") == 0){
          par.index = 1; strcpy(par.index_c, "GRN");
        } else if (strcmp(ptr, "RED\n") == 0){
          par.index = 2; strcpy(par.index_c, "RED");
        } else if (strcmp(ptr, "NIR\n") == 0){
          par.index = 3; strcpy(par.index_c, "NIR");
        } else if (strcmp(ptr, "SWIR1\n") == 0){
          par.index = 4; strcpy(par.index_c, "SW1");
        } else if (strcmp(ptr, "SWIR2\n") == 0){
          par.index = 5; strcpy(par.index_c, "SW2");
        } else if (strcmp(ptr, "RE1\n") == 0){
          par.index = 6; strcpy(par.index_c, "RE1");
        } else if (strcmp(ptr, "RE2\n") == 0){
          par.index = 7; strcpy(par.index_c, "RE2");
        } else if (strcmp(ptr, "RE3\n") == 0){
          par.index = 8; strcpy(par.index_c, "RE3");
        } else if (strcmp(ptr, "BNIR\n") == 0){
          par.index = 9; strcpy(par.index_c, "BNR");
        } else if (strcmp(ptr, "NDVI\n") == 0){
          par.index = 10; strcpy(par.index_c, "NDV");
        } else if (strcmp(ptr, "EVI\n") == 0){
          par.index = 11; strcpy(par.index_c, "EVI");
        } else if (strcmp(ptr, "NBR\n") == 0){
          par.index = 12; strcpy(par.index_c, "NBR");
        } else if (strcmp(ptr, "ARVI\n") == 0){
          par.index = 13; strcpy(par.index_c, "ARV");
        } else if (strcmp(ptr, "SAVI\n") == 0){
          par.index = 14; strcpy(par.index_c, "SAV");
        } else if (strcmp(ptr, "SARVI\n") == 0){
          par.index = 15; strcpy(par.index_c, "SRV");
        } else if (strcmp(ptr, "TC-BRIGHT\n") == 0){
          par.index = 16; strcpy(par.index_c, "TCB");
        } else if (strcmp(ptr, "TC-GREEN\n") == 0){
          par.index = 17; strcpy(par.index_c, "TCG");
        } else if (strcmp(ptr, "TC-WET\n") == 0){
          par.index = 18; strcpy(par.index_c, "TCW");
        } else if (strcmp(ptr, "TC-DI0\n") == 0){
          par.index = 19; strcpy(par.index_c, "TCD");
        } else if (strcmp(ptr, "TC-DI\n") == 0){
          par.index = 80; strcpy(par.index_c, "TCD");
        } else if (strcmp(ptr, "SMA\n") == 0){
          par.index = 81; strcpy(par.index_c, "SMA");
        } else {
          printf("INDEX in parameter file is invalid.\n");
          printf("valid values: BLUE/GREEN/RED/NIR/SWIR1/SWIR2/NDVI/EVI/NBR/ARVI/SARVI/TC-BRIGHT/TC-GREEN/TC-WET/TC-DI/TC-DI0/SMA.\n");
          return FAILURE;
        }
      } else if (strcmp(tag, "CENTER") == 0){
        par.vcenter = parse_logical("CENTER", ptr, &par.docenter);
      } else if (strcmp(tag, "STANDARD") == 0){
        par.vostandard = parse_logical("STANDARD", ptr, &par.dostandard);
      } else if (strcmp(tag, "TREND") == 0){
        if (strcmp(ptr, "YEAR\n") == 0){
          par.trend = 0; strcpy(par.trend_c, "Y");
        } else if (strcmp(ptr, "MONTH\n") == 0){
          par.trend = 1; strcpy(par.trend_c, "M");
        } else if (strcmp(ptr, "WEEK\n") == 0){
          par.trend = 2; strcpy(par.trend_c, "W");
        } else if (strcmp(ptr, "DAY\n") == 0){
          par.trend = 3; strcpy(par.trend_c, "D");
        } else {
          printf("TREND in parameter file is invalid.\n");
          printf("valid values: YEAR/MONTH/WEEK/DAY.\n");
          return FAILURE;
        }
      } else if (strcmp(tag, "TAIL") == 0){
        if (strcmp(ptr, "LEFT\n") == 0){
          par.tail = 0; strcpy(par.tail_c, "L");
        } else if (strcmp(ptr, "TWO\n") == 0){
          par.tail = 1; strcpy(par.tail_c, "T");
        } else if (strcmp(ptr, "RIGHT\n") == 0){
          par.tail = 2; strcpy(par.tail_c, "R");
        } else {
          printf("TAIL in parameter file is invalid.\n");
          printf("valid values: LEFT/TWO/RIGHT.\n");
          return FAILURE;
        }
      } else if (strcmp(tag, "CONF") == 0){
        parse_float("CONF", ptr, &par.conf, 1, 0, 1);
      } else if (strcmp(tag, "FOLD") == 0){
        if (strcmp(ptr, "AVG\n") == 0){
          par.fold = 0; strcpy(par.fold_c, "AVG");
        } else if (strcmp(ptr, "MIN\n") == 0){
          par.fold = 1; strcpy(par.fold_c, "MIN");
        } else if (strcmp(ptr, "MAX\n") == 0){
          par.fold = 2; strcpy(par.fold_c, "MAX");
        } else if (strcmp(ptr, "LSP\n") == 0){
          par.fold = 3; strcpy(par.fold_c, "LSP");
        } else {
          printf("FOLD in parameter file is invalid.\n");
          printf("valid values: AVG/MIN/MAX/LSP.\n");
          return FAILURE;
        }
      } else if (strcmp(tag, "DESPIKE") == 0){
        parse_float("DESPIKE", ptr, &par.despike, 1, INT_MIN, INT_MAX);

      // interpolation parameters
      } else if (strcmp(tag, "INTERPOLATE") == 0){
        if (strcmp(ptr, "NONE\n") == 0){
          par.interpolate = NONE;
        } else if (strcmp(ptr, "LINEAR\n") == 0){
          par.interpolate = LINEAR;
        } else if (strcmp(ptr, "MOVING\n") == 0){
          par.interpolate = MOVING;
        } else if (strcmp(ptr, "RBF\n") == 0){
          par.interpolate = RBF;
        } else {
          printf("INTERPOLATE in parameter file is invalid.\n");
          printf("valid values: NONE/LINEAR/MOVING/RBF.\n");
          return FAILURE;
        }
      } else if (strcmp(tag, "MOVING_MAX") == 0){
        parse_int("MOVING_MAX", ptr, &par.mov_max, 1, 1, 366);
      } else if (strcmp(tag, "RBF_SIGMA") == 0){
        parse_int_seq("RBF_SIGMA", ctmp, par.rbf_sigma, &par.rbf_nk, 1, 1, 365);
      } else if (strcmp(tag, "RBF_CUTOFF") == 0){
        parse_float("RBF_CUTOFF", ptr, &par.rbf_cutoff, 1, 0, 1);
      } else if (strcmp(tag, "INT_DAY") == 0){
        parse_int("INT_DAY", ptr, &par.int_nstep, 1, 0, INT_MAX);

      // SMA parameters
      } else if (strcmp(tag, "SMA_SUM_TO_ONE") == 0){
        par.vsma_sto = parse_logical("SMA_SUM_TO_ONE", ptr, &par.sma_sto);
      } else if (strcmp(tag, "SMA_NON_NEG") == 0){
        par.vsma_pos = parse_logical("SMA_NON_NEG", ptr, &par.sma_pos);
      } else if (strcmp(tag, "SMA_SHD_NORM") == 0){
        par.vsma_shn = parse_logical("SMA_SHD_NORM", ptr, &par.sma_shn);
      } else if (strcmp(tag, "SMA_ENDMEMBER") == 0){
        parse_int("SMA_ENDMEMBER", ptr, &par.sma_emb, 1, 0, INT_MAX);

      // LSP parameters
      } else if (strcmp(tag, "LSP_DOY_PREV_YEAR") == 0){
        parse_int("LSP_DOY_PREV_YEAR", ptr, &par.lsp_dprev, 1, 1, 365);
      } else if (strcmp(tag, "LSP_DOY_NEXT_YEAR") == 0){
        parse_int("LSP_DOY_NEXT_YEAR", ptr, &par.lsp_dnext, 1, 1, 365);
      } else if (strcmp(tag, "LSP_HEMISPHERE") == 0){
        parse_int("LSP_HEMISPHERE", ptr, &par.lsp_hemi, 1, 0, 2);
      } else if (strcmp(tag, "LSP_N_SEGMENT") == 0){
        parse_int("LSP_N_SEGMENT", ptr, &par.lsp_nseg, 1, 1, INT_MAX);

      // other variables
      } else if (strcmp(tag, "NUM_CPU") == 0){
        parse_int("NUM_CPU", ptr, &par.cpu, 1, 0, INT_MAX);

      // output options
      } else if (strcmp(tag, "OUTPUT_FORMAT") == 0){
        if (strcmp(ptr, "ENVI\n") == 0){
          par.format = 0;
        } else if (strcmp(ptr, "GTiff\n") == 0){
          par.format = 1;
        } else {
          printf("OUTPUT_FORMAT in parameter file is invalid.\n");
          printf("valid values: ENVI/GTiff.\n");
          return FAILURE;
        }
      } else if (strcmp(tag, "OUTPUT_TSS") == 0){
        par.votss = parse_logical("OUTPUT_TSS", ptr, &par.otss);
      } else if (strcmp(tag, "OUTPUT_RMS") == 0){
        par.vorms = parse_logical("OUTPUT_RMS", ptr, &par.orms);
      } else if (strcmp(tag, "OUTPUT_STA") == 0){
        par.vosta = parse_logical("OUTPUT_STA", ptr, &par.osta);
      } else if (strcmp(tag, "OUTPUT_TSI") == 0){
        par.votsi = parse_logical("OUTPUT_TSI", ptr, &par.otsi);
      } else if (strcmp(tag, "OUTPUT_FBY") == 0){
        par.vofby = parse_logical("OUTPUT_FBY", ptr, &par.ofby);
      } else if (strcmp(tag, "OUTPUT_FBM") == 0){
        par.vofbm = parse_logical("OUTPUT_FBM", ptr, &par.ofbm);
      } else if (strcmp(tag, "OUTPUT_FBW") == 0){
        par.vofbw = parse_logical("OUTPUT_FBW", ptr, &par.ofbw);
      } else if (strcmp(tag, "OUTPUT_FBD") == 0){
        par.vofbd = parse_logical("OUTPUT_FBD", ptr, &par.ofbd);
      } else if (strcmp(tag, "OUTPUT_TRD") == 0){
        par.votrd = parse_logical("OUTPUT_TRD", ptr, &par.otrd);
      } else if (strcmp(tag, "OUTPUT_CAT") == 0){
        par.vocat = parse_logical("OUTPUT_CAT", ptr, &par.ocat);
      } else if (strcmp(tag, "OUTPUT_LSP") == 0){
        par.volsp = parse_logical("OUTPUT_LSP", ptr, &par.olsp);
      }

      // in case tag is not the first word in a line
      tag = ptr;

    }

  }

  fclose(fpar);

  // number of tiles
  par.tnx = par.tmaxx-par.tminx+1;
  par.tny = par.tmaxy-par.tminy+1;

  // seasonal subset
  doy2md(par.doymin, &m, &d);
  if (par.monthmin < m) par.monthmin = m;
  doy2md(par.doymax, &m, &d);
  if (par.monthmax > m) par.monthmax = m;
  doy = md2doy(par.monthmin, 1);
  if (par.doymin < doy) par.doymin = doy;
  doy = md2doy(par.monthmax, 31);
  if (par.doymax > doy) par.doymax = doy;
  par.weekmin = doy2week(par.doymin);
  par.weekmax = doy2week(par.doymax);

  // range for interpolation
  par.cemin = par.yearmin*365 + par.doymin;
  par.cemax = par.yearmax*365 + par.doymax;
  par.int_n = floor((par.cemax-par.cemin+1)/par.int_nstep);

  if (par.olsp || par.fold == 3){
    par.lsp = true; 
  } else { 
    par.lsp = false;
  }

  // number of LSP years
  par.Yn = par.yearmax-par.yearmin-1;
  
  #ifdef DEBUG
  printf("year: %04d - %04d\n", par.yearmin, par.yearmax);
  printf("month: %02d - %02d\n", par.monthmin, par.monthmax);
  printf("week: %02d - %02d\n", par.weekmin, par.weekmax);
  printf("day: %03d - %03d\n", par.doymin, par.doymax);
  #endif


  // only input bands that are needed to compute index, 
  // throw error if a band is not available
  alloc((void**)&valid, _NB_, sizeof(bool));

  switch (par.index){
    case 0:
      // blue band
      if (band_available(_B_.blue, valid) == FAILURE) return FAILURE;
      break;
    case 1:
      // green band
      if (band_available(_B_.green, valid) == FAILURE) return FAILURE;
      break;
    case 2:
      // red band
      if (band_available(_B_.red, valid) == FAILURE) return FAILURE;
      break;
    case 3:
      // nir band
      if (band_available(_B_.nir, valid) == FAILURE) return FAILURE;
      break;
    case 4:
      // swir 1 bands
      if (band_available(_B_.swir1, valid) == FAILURE) return FAILURE;
      break;
    case 5:
      // swir 2 band
      if (band_available(_B_.swir2, valid) == FAILURE) return FAILURE;
      break;
    case 6:
      // red edge 1 band
      if (band_available(_B_.rededge1, valid) == FAILURE) return FAILURE;
      break;
    case 7:
      // red edge 2 band
      if (band_available(_B_.rededge2, valid) == FAILURE) return FAILURE;
      break;
    case 8:
      // red edge 3 band
      if (band_available(_B_.rededge3, valid) == FAILURE) return FAILURE;
      break;
    case 9:
      // broad nir band
      if (band_available(_B_.bnir, valid) == FAILURE) return FAILURE;
      break;
    case 10:
      // NDVI
      if (band_available(_B_.red, valid) == FAILURE) return FAILURE;
      if (band_available(_B_.nir, valid) == FAILURE) return FAILURE;
      break;
    case 11:
      // EVI
      if (band_available(_B_.blue, valid) == FAILURE) return FAILURE;
      if (band_available(_B_.red,  valid) == FAILURE) return FAILURE;
      if (band_available(_B_.nir,  valid) == FAILURE) return FAILURE;
      break;
    case 12:
      // NBR
      if (band_available(_B_.nir,   valid) == FAILURE) return FAILURE;
      if (band_available(_B_.swir2, valid) == FAILURE) return FAILURE;
      break;
    case 13:
      // ARVI
      if (band_available(_B_.blue, valid) == FAILURE) return FAILURE;
      if (band_available(_B_.red,  valid) == FAILURE) return FAILURE;
      if (band_available(_B_.nir,  valid) == FAILURE) return FAILURE;
      break;
    case 14:
      // SAVI
      if (band_available(_B_.red,  valid) == FAILURE) return FAILURE;
      if (band_available(_B_.nir,  valid) == FAILURE) return FAILURE;
      break;
    case 15:
      // SARVI
      if (band_available(_B_.blue, valid) == FAILURE) return FAILURE;
      if (band_available(_B_.red,  valid) == FAILURE) return FAILURE;
      if (band_available(_B_.nir,  valid) == FAILURE) return FAILURE;
      break;
    case 16:
      // Tasseled Cap Brigtness
      if (band_available(_B_.blue,  valid) == FAILURE) return FAILURE;
      if (band_available(_B_.green, valid) == FAILURE) return FAILURE;
      if (band_available(_B_.red,   valid) == FAILURE) return FAILURE;
      if (band_available(_B_.nir,   valid) == FAILURE) return FAILURE;
      if (band_available(_B_.swir1, valid) == FAILURE) return FAILURE;
      if (band_available(_B_.swir2, valid) == FAILURE) return FAILURE;
      break;
    case 17:
    // Tasseled Cap Greenness
      if (band_available(_B_.blue,  valid) == FAILURE) return FAILURE;
      if (band_available(_B_.green, valid) == FAILURE) return FAILURE;
      if (band_available(_B_.red,   valid) == FAILURE) return FAILURE;
      if (band_available(_B_.nir,   valid) == FAILURE) return FAILURE;
      if (band_available(_B_.swir1, valid) == FAILURE) return FAILURE;
      if (band_available(_B_.swir2, valid) == FAILURE) return FAILURE;
      break;
    case 18:
      // Tasseled Cap Wetness
      if (band_available(_B_.blue,  valid) == FAILURE) return FAILURE;
      if (band_available(_B_.green, valid) == FAILURE) return FAILURE;
      if (band_available(_B_.red,   valid) == FAILURE) return FAILURE;
      if (band_available(_B_.nir,   valid) == FAILURE) return FAILURE;
      if (band_available(_B_.swir1, valid) == FAILURE) return FAILURE;
      if (band_available(_B_.swir2, valid) == FAILURE) return FAILURE;
      break;
    case 19:
      // Tasseled Cap Brigtness/Greenness/Wetness, DI without rescaling
      if (band_available(_B_.blue,  valid) == FAILURE) return FAILURE;
      if (band_available(_B_.green, valid) == FAILURE) return FAILURE;
      if (band_available(_B_.red,   valid) == FAILURE) return FAILURE;
      if (band_available(_B_.nir,   valid) == FAILURE) return FAILURE;
      if (band_available(_B_.swir1, valid) == FAILURE) return FAILURE;
      if (band_available(_B_.swir2, valid) == FAILURE) return FAILURE;
      break;
    case 80:
      // Tasseled Cap Brigtness/Greenness/Wetness, DI
      if (band_available(_B_.blue,  valid) == FAILURE) return FAILURE;
      if (band_available(_B_.green, valid) == FAILURE) return FAILURE;
      if (band_available(_B_.red,   valid) == FAILURE) return FAILURE;
      if (band_available(_B_.nir,   valid) == FAILURE) return FAILURE;
      if (band_available(_B_.swir1, valid) == FAILURE) return FAILURE;
      if (band_available(_B_.swir2, valid) == FAILURE) return FAILURE;
      break;
  }


  // only input bands that are needed to compute index
  for (b=0; b<_NB_; b++){
    if (!valid[b]){
      #ifdef DEBUG
      printf("band %02d: do not read image band", b);
      #endif
      for (s=0; s<par.l2dict.n; s++){
        #ifdef DEBUG
        printf(" %02d", par.l2dict.band[s][b]);
        #endif
        par.l2dict.band[s][b] = -1;
      }
      #ifdef DEBUG
      printf("\n");
      #endif
    } else {
      #ifdef DEBUG
      printf("band %02d: read image bands\n", b);
      #endif
    }
  }

  free((void*)valid);


  // choose products
  par.use.ref = true;
  par.use.qai = true;
  par.use.cld = false;
  par.use.hot = false;
  par.use.vzn = false;
  par.use.aod = false;
  par.use.wvp = false;


  if (check_param_tsa() != SUCCESS){
    printf(" Unable to read parameter file!\n"); return FAILURE;}

  return SUCCESS;
}


/** Parse CSO parameters
+++ This function reads the parameter file.
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int parse_param_cso(){
FILE *fpar;
char  buffer[MAX_STR_LEN] = "\0";
char *tag = NULL;
char *ptr = NULL;
const char *separator = " =";
char  ctmp[MAX_STR_LEN];
int w, y, m;


  init_param_cso();

  // open parameter file
  if ((fpar = fopen(par.f_par, "r")) == NULL){
    printf("Unable to open parameter file!\n"); return FAILURE;}

  if (fscanf(fpar, "%s", buffer) < 0){
    printf("No valid parameter file!\n"); return FAILURE;}
  if(strcmp(buffer, "++PARAM_CSO_START++") != 0){
    printf("No valid parameter file!\n"); return FAILURE;}

  // process line by line
  while (fgets(buffer, MAX_STR_LEN, fpar) != NULL){

    if(strcmp(buffer, "++PARAM_CSO_END++\n") == 0) break;
    
    strcpy(ctmp, buffer);

    // get tag
    ptr = strtok(buffer, separator);
    tag = ptr;

    // continue if tag is comment
    if(strcmp(tag,"#") == 0) continue;

    // extract parameters by comparing tag
    while(ptr != NULL){

      ptr = strtok(NULL, separator);

      // directory variables
      if (strcmp(tag, "DIR_LEVEL2") == 0){
        strncpy(par.d_level2, ptr, strlen(ptr)-1);
        par.d_level2[strlen(ptr)-1] = '\0';
      } else if (strcmp(tag, "DIR_CSO") == 0){
        strncpy(par.d_cso, ptr, strlen(ptr)-1);
        par.d_cso[strlen(ptr)-1] = '\0';
      } else if (strcmp(tag, "FILE_TILE") == 0){
        strncpy(par.f_tile, ptr, strlen(ptr)-1);
        par.f_tile[strlen(ptr)-1] = '\0';

      // sensors
      } else if (strcmp(tag, "SENSORS") == 0){
        if (level2sensor_dict(ctmp, &par.l2dict) != SUCCESS) exit(1);

      // QAI screening
      } else if (strcmp(tag, "SCREEN_QAI") == 0){
        if (parse_quality(ctmp) != SUCCESS) exit(1);
        par.vqai = true;

      // spatial variables
      } else if (strcmp(tag, "X_TILE_MIN") == 0){
        parse_int("X_TILE_MIN", ptr, &par.tminx, 1, -999, 9999);
      } else if (strcmp(tag, "Y_TILE_MIN") == 0){
        parse_int("Y_TILE_MIN", ptr, &par.tminy, 1, -999, 9999);
      } else if (strcmp(tag, "X_TILE_MAX") == 0){
        parse_int("X_TILE_MAX", ptr, &par.tmaxx, 1, -999, 9999);
      } else if (strcmp(tag, "Y_TILE_MAX") == 0){
        parse_int("Y_TILE_MAX", ptr, &par.tmaxy, 1, -999, 9999);
      } else if (strcmp(tag, "RESOLUTION") == 0){
        parse_float("RESOLUTION", ptr, &par.res, 1, 0, FLT_MAX);

      // temporal variables
      } else if (strcmp(tag, "YEAR_MIN") == 0){
        parse_int("YEAR_MIN", ptr, &par.yearmin, 1, 1900, 2100);
      } else if (strcmp(tag, "YEAR_MAX") == 0){
        parse_int("YEAR_MAX", ptr, &par.yearmax, 1, 1900, 2100);
      } else if (strcmp(tag, "MONTH_STEP") == 0){
        parse_int("MONTH_STEP", ptr, &par.monthstep, 1, 1, 12);

      // other variables
      } else if (strcmp(tag, "NUM_CPU") == 0){
        parse_int("NUM_CPU", ptr, &par.cpu, 1, 0, INT_MAX);

      // output options
      } else if (strcmp(tag, "OUTPUT_FORMAT") == 0){
        if (strcmp(ptr, "ENVI\n") == 0){
          par.format = 0;
        } else if (strcmp(ptr, "GTiff\n") == 0){
          par.format = 1;
        } else {
          printf("OUTPUT_FORMAT in parameter file is invalid.\n");
          printf("valid values: ENVI/GTiff.\n");
          return FAILURE;
        }
      } else if (strcmp(tag, "OUTPUT_NUM") == 0){
        par.vonum = parse_logical("OUTPUT_NUM", ptr, &par.onum);
      } else if (strcmp(tag, "OUTPUT_AVG") == 0){
        par.voavg = parse_logical("OUTPUT_AVG", ptr, &par.oavg);
      } else if (strcmp(tag, "OUTPUT_STD") == 0){
        par.vostd = parse_logical("OUTPUT_STD", ptr, &par.ostd);
      } else if (strcmp(tag, "OUTPUT_MIN") == 0){
        par.vomin = parse_logical("OUTPUT_MIN", ptr, &par.omin);
      } else if (strcmp(tag, "OUTPUT_MAX") == 0){
        par.vomax = parse_logical("OUTPUT_MAX", ptr, &par.omax);
      } else if (strcmp(tag, "OUTPUT_RNG") == 0){
        par.vorng = parse_logical("OUTPUT_RNG", ptr, &par.orng);
      } else if (strcmp(tag, "OUTPUT_SKW") == 0){
        par.voskw = parse_logical("OUTPUT_SKW", ptr, &par.oskw);
      } else if (strcmp(tag, "OUTPUT_KRT") == 0){
        par.vokrt = parse_logical("OUTPUT_KRT", ptr, &par.okrt);
      } else if (strcmp(tag, "OUTPUT_Q25") == 0){
        par.voq25 = parse_logical("OUTPUT_Q25", ptr, &par.oq25);
      } else if (strcmp(tag, "OUTPUT_Q50") == 0){
        par.voq50 = parse_logical("OUTPUT_Q50", ptr, &par.oq50);
      } else if (strcmp(tag, "OUTPUT_Q75") == 0){
        par.voq75 = parse_logical("OUTPUT_Q75", ptr, &par.oq75);
      } else if (strcmp(tag, "OUTPUT_IQR") == 0){
        par.voiqr = parse_logical("OUTPUT_IQR", ptr, &par.oiqr);
      }

      // in case tag is not the first word in a line
      tag = ptr;

    }

  }

  fclose(fpar);

  // number of tiles
  par.tnx = par.tmaxx-par.tminx+1;
  par.tny = par.tmaxy-par.tminy+1;

  // temporal range
  par.doymin = 1;
  par.doymax = 365;
  par.monthmin = 1;
  par.monthmax = 12;
  par.weekmin = doy2week(par.doymin);
  par.weekmax = doy2week(par.doymax);
  
  // seasonal windows
  par.nwin = ((par.yearmax*12+par.monthmax) - (par.yearmin*12+par.monthmin) + 1) / par.monthstep;
  par.bwin = par.nwin + 1;
  for (w=0, y=par.yearmin, m=par.monthmin; w<par.bwin; w++, m+=par.monthstep){
    if (m > 12){ m -= 12; y++;}
    par.cewin[w] = date2ce(m, 1, y);
    par.ywin[w] = y;
    par.mwin[w] = m;
  }

  #ifdef DEBUG
  printf("year: %04d - %04d\n", par.yearmin, par.yearmax);
  printf("month: %02d - %02d\n", par.monthmin, par.monthmax);
  printf("week: %02d - %02d\n", par.weekmin, par.weekmax);
  printf("day: %03d - %03d\n", par.doymin, par.doymax);
  printf("number of seasonal windows: %d\n", par.nwin);
  #endif


  // choose products
  par.use.ref = false;
  par.use.qai = true;
  par.use.cld = false;
  par.use.hot = false;
  par.use.vzn = false;
  par.use.aod = false;
  par.use.wvp = false;
  par.use.imp = false;


  if (check_param_cso() != SUCCESS){
    printf(" Unable to read parameter file!\n"); return FAILURE;}

  return SUCCESS;
}


/** Parse ImproPhe parameters
+++ This function reads the parameter file.
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int parse_param_imp(){
FILE *fpar;
char  buffer[MAX_STR_LEN] = "\0";
char *tag = NULL;
char *ptr = NULL;
const char *separator = " =";
char  ctmp[MAX_STR_LEN];
int w, ntmp;
float ftmp;


  init_param_imp();

  // open parameter file
  if ((fpar = fopen(par.f_par, "r")) == NULL){
    printf("Unable to open parameter file!\n"); return FAILURE;}

  if (fscanf(fpar, "%s", buffer) < 0){
    printf("No valid parameter file!\n"); return FAILURE;}
  if(strcmp(buffer, "++PARAM_IMP_START++") != 0){
    printf("No valid parameter file!\n"); return FAILURE;}

  // process line by line
  while (fgets(buffer, MAX_STR_LEN, fpar) != NULL){

    if(strcmp(buffer, "++PARAM_IMP_END++\n") == 0) break;
    
    strcpy(ctmp, buffer);

    // get tag
    ptr = strtok(buffer, separator);
    tag = ptr;

    // continue if tag is comment
    if(strcmp(tag,"#") == 0) continue;

    // extract parameters by comparing tag
    while(ptr != NULL){

      ptr = strtok(NULL, separator);

      // directory variables
      if (strcmp(tag, "DIR_LEVEL2") == 0){
        strncpy(par.d_level2, ptr, strlen(ptr)-1);
        par.d_level2[strlen(ptr)-1] = '\0';
      } else if (strcmp(tag, "FILE_COARSE") == 0){
        parse_char_seq("FILE_COARSE", ctmp, par.f_cr, &par.num_cr);
      } else if (strcmp(tag, "PRODUCT_TAG") == 0){
        parse_char_seq("PRODUCT_TAG", ctmp, par.imp_prd, &ntmp);
      } else if (strcmp(tag, "DIR_IMPROPHE") == 0){
        strncpy(par.d_improphe, ptr, strlen(ptr)-1);
        par.d_improphe[strlen(ptr)-1] = '\0';
      } else if (strcmp(tag, "FILE_TILE") == 0){
        strncpy(par.f_tile, ptr, strlen(ptr)-1);
        par.f_tile[strlen(ptr)-1] = '\0';

      // sensors
      } else if (strcmp(tag, "SENSORS") == 0){
        if (level2sensor_dict(ctmp, &par.l2dict) != SUCCESS) exit(1);
        strcpy(par.csen_mr, ctmp);

      // QAI screening
      } else if (strcmp(tag, "SCREEN_QAI") == 0){
        if (parse_quality(ctmp) != SUCCESS) exit(1);
        par.vqai = true;

      // spatial variables
      } else if (strcmp(tag, "X_TILE_MIN") == 0){
        parse_int("X_TILE_MIN", ptr, &par.tminx, 1, -999, 9999);
      } else if (strcmp(tag, "Y_TILE_MIN") == 0){
        parse_int("Y_TILE_MIN", ptr, &par.tminy, 1, -999, 9999);
      } else if (strcmp(tag, "X_TILE_MAX") == 0){
        parse_int("X_TILE_MAX", ptr, &par.tmaxx, 1, -999, 9999);
      } else if (strcmp(tag, "Y_TILE_MAX") == 0){
        parse_int("Y_TILE_MAX", ptr, &par.tmaxy, 1, -999, 9999);
      } else if (strcmp(tag, "RESOLUTION") == 0){
        parse_float("RESOLUTION", ptr, &par.res, 1, 0, FLT_MAX);
      } else if (strcmp(tag, "REDUCE_PSF") == 0){
        par.vpsf = parse_logical("REDUCE_PSF", ptr, &par.psf);

      // temporal variables
      } else if (strcmp(tag, "YEAR_MIN") == 0){
        parse_int("YEAR_MIN", ptr, &par.ymin, 1, 1900, 2100);
      } else if (strcmp(tag, "YEAR_MAX") == 0){
        parse_int("YEAR_MAX", ptr, &par.ymax, 1, 1900, 2100);

      // window definition
      } else if (strcmp(tag, "SEASONAL_WINDOW") == 0){
        parse_int_seq("SEASONAL_WINDOW", ctmp, par.dwin, &par.bwin, 1, -364, 730);
        par.nwin = par.bwin-1;
        for (w=0; w<par.bwin; w++){
          if (par.dwin[w] < 1){
            par.dwin[w] = 365+par.dwin[w]; par.ywin[w] = -1;
          } else if (par.dwin[w] > 365){
            par.dwin[w] = par.dwin[w]-365; par.ywin[w] = 1;
          } else {
            par.ywin[w] = 0;
          }
        }

      // prediction parameters
      } else if (strcmp(tag, "KERNEL_SIZE") == 0){
        parse_float("KERNEL_SIZE", ptr, &ftmp, 1, 0, FLT_MAX);
        par.ksize = (int)(ftmp/par.res);
        // kernel size should be given odd, increment if not
        if (par.ksize % 2 == 0) par.ksize++;
      } else if (strcmp(tag, "KERNEL_TEXT") == 0){
        parse_float("KERNEL_TEXT", ptr, &ftmp, 1, 0, FLT_MAX);
        par.ksd = (int)(ftmp/par.res);
        // kernel size should be given odd, increment if not
        if (par.ksd % 2 == 0) par.ksd++;

      // coarse resolution information
      } else if (strcmp(tag, "COARSE_1ST_YEAR") == 0){
        parse_int("COARSE_1ST_YEAR", ptr, &par.ystart_cr, 1, 1900, 2100);
      } else if (strcmp(tag, "COARSE_NODATA") == 0){
        parse_float("COARSE_NODATA", ptr, &par.nodata_cr, 1, INT_MIN, INT_MAX);

      // other variables
      } else if (strcmp(tag, "NUM_CPU") == 0){
        parse_int("NUM_CPU", ptr, &par.cpu, 1, 0, INT_MAX);

      // output options
      } else if (strcmp(tag, "OUTPUT_FORMAT") == 0){
        if (strcmp(ptr, "ENVI\n") == 0){
          par.format = 0;
        } else if (strcmp(ptr, "GTiff\n") == 0){
          par.format = 1;
        } else {
          printf("OUTPUT_FORMAT in parameter file is invalid.\n");
          printf("valid values: ENVI/GTiff.\n");
          return FAILURE;
        }
      
      }

      // in case tag is not the first word in a line
      tag = ptr;

    }

  }

  fclose(fpar);

  // number of tiles
  par.tnx = par.tmaxx-par.tminx+1;
  par.tny = par.tmaxy-par.tminy+1;

  // temporal subset for input
  par.yearmin = par.ymin + par.ywin[0];
  par.yearmax = par.ymax + par.ywin[par.bwin-1];
  if (par.ywin[0] < 0 || par.ywin[par.bwin-1] > 0){
    par.doymin = 1; par.doymax = 365;
  } else {
    par.doymin = par.dwin[0]; par.doymax = par.dwin[par.bwin-1];
  }

  if (ntmp != par.num_cr){
    printf("PRODUCT_TAG and FILE_COARSE do not match.\n"); return FAILURE;}

  #ifdef DEBUG
  printf("year: %04d - %04d\n", par.yearmin, par.yearmax);
  printf("day: %03d - %03d\n", par.doymin, par.doymax);
  #endif

  // choose products
  par.use.ref = true;
  par.use.qai = true;
  par.use.cld = false;
  par.use.hot = false;
  par.use.vzn = false;
  par.use.aod = false;
  par.use.wvp = false;
  par.use.imp = false;


  if (check_param_imp() != SUCCESS){
    printf(" Unable to read parameter file!\n"); return FAILURE;}

  return SUCCESS;
}


/** Parse Level 2 ImproPhe parameters
+++ This function reads the parameter file.
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int parse_param_il2(){
FILE *fpar;
char  buffer[MAX_STR_LEN] = "\0";
char *tag = NULL;
char *ptr = NULL;
const char *separator = " =";
char  ctmp[MAX_STR_LEN];
char  csen[MAX_STR_LEN];
int w, ntmp;
float ftmp;


  init_param_il2();

  // open parameter file
  if ((fpar = fopen(par.f_par, "r")) == NULL){
    printf("Unable to open parameter file!\n"); return FAILURE;}

  if (fscanf(fpar, "%s", buffer) < 0){
    printf("No valid parameter file!\n"); return FAILURE;}
  if(strcmp(buffer, "++PARAM_L2IMP_START++") != 0){
    printf("No valid parameter file!\n"); return FAILURE;}

  // process line by line
  while (fgets(buffer, MAX_STR_LEN, fpar) != NULL){

    if(strcmp(buffer, "++PARAM_L2IMP_END++\n") == 0) break;
    
    strcpy(ctmp, buffer);

    // get tag
    ptr = strtok(buffer, separator);
    tag = ptr;

    // continue if tag is comment
    if(strcmp(tag,"#") == 0) continue;

    // extract parameters by comparing tag
    while(ptr != NULL){

      ptr = strtok(NULL, separator);

      // directory variables
      if (strcmp(tag, "DIR_LEVEL2") == 0){
        strncpy(par.d_level2, ptr, strlen(ptr)-1);
        par.d_level2[strlen(ptr)-1] = '\0';
      } else if (strcmp(tag, "FILE_TILE") == 0){
        strncpy(par.f_tile, ptr, strlen(ptr)-1);
        par.f_tile[strlen(ptr)-1] = '\0';

      // sensors
      } else if (strcmp(tag, "SENSORS_HR") == 0){
        strcpy(par.csen_mr, ctmp);
      } else if (strcmp(tag, "SENSORS_MR") == 0){
        strcpy(par.csen_cr, ctmp);

      // QAI screening
      } else if (strcmp(tag, "SCREEN_QAI") == 0){
        if (parse_quality(ctmp) != SUCCESS) exit(1);
        par.vqai = true;

      // spatial variables
      } else if (strcmp(tag, "X_TILE_MIN") == 0){
        parse_int("X_TILE_MIN", ptr, &par.tminx, 1, -999, 9999);
      } else if (strcmp(tag, "Y_TILE_MIN") == 0){
        parse_int("Y_TILE_MIN", ptr, &par.tminy, 1, -999, 9999);
      } else if (strcmp(tag, "X_TILE_MAX") == 0){
        parse_int("X_TILE_MAX", ptr, &par.tmaxx, 1, -999, 9999);
      } else if (strcmp(tag, "Y_TILE_MAX") == 0){
        parse_int("Y_TILE_MAX", ptr, &par.tmaxy, 1, -999, 9999);
      } else if (strcmp(tag, "RESOLUTION") == 0){
        parse_float("RESOLUTION", ptr, &par.res, 1, 0, FLT_MAX);
      } else if (strcmp(tag, "REDUCE_PSF") == 0){
        par.vpsf = parse_logical("REDUCE_PSF", ptr, &par.psf);

      // temporal variables
      } else if (strcmp(tag, "YEAR") == 0){
        parse_int("YEAR", ptr, &par.ymin, 1, 1900, 2100);
        par.ymax = par.ymin+1;

      // window definition
      } else if (strcmp(tag, "SEASONAL_WINDOW") == 0){
        parse_int_seq("SEASONAL_WINDOW", ctmp, par.dwin, &par.bwin, 1, -364, 730);
        par.nwin = par.bwin-1;
        for (w=0; w<par.bwin; w++){
          if (par.dwin[w] < 1){
            par.dwin[w] = 365+par.dwin[w]; par.ywin[w] = -1;
          } else if (par.dwin[w] > 365){
            par.dwin[w] = par.dwin[w]-365; par.ywin[w] = 1;
          } else {
            par.ywin[w] = 0;
          }
        }

      // prediction parameters
      } else if (strcmp(tag, "KERNEL_SIZE") == 0){
        parse_float("KERNEL_SIZE", ptr, &ftmp, 1, 0, FLT_MAX);
        par.ksize = (int)(ftmp/par.res);
        // kernel size should be given odd, increment if not
        if (par.ksize % 2 == 0) par.ksize++;
      } else if (strcmp(tag, "KERNEL_TEXT") == 0){
        parse_float("KERNEL_TEXT", ptr, &ftmp, 1, 0, FLT_MAX);
        par.ksd = (int)(ftmp/par.res);
        // kernel size should be given odd, increment if not
        if (par.ksd % 2 == 0) par.ksd++;

      // other variables
      } else if (strcmp(tag, "NUM_CPU") == 0){
        parse_int("NUM_CPU", ptr, &par.cpu, 1, 0, INT_MAX);

      // output options
      } else if (strcmp(tag, "OUTPUT_OVERWRITE") == 0){
        par.vowr = parse_logical("OUTPUT_OVERWRITE", ptr, &par.owr);
      } else if (strcmp(tag, "OUTPUT_FORMAT") == 0){
        if (strcmp(ptr, "ENVI\n") == 0){
          par.format = 0;
        } else if (strcmp(ptr, "GTiff\n") == 0){
          par.format = 1;
        } else {
          printf("OUTPUT_FORMAT in parameter file is invalid.\n");
          printf("valid values: ENVI/GTiff.\n");
          return FAILURE;
        }
      
      }

      // in case tag is not the first word in a line
      tag = ptr;

    }

  }

  fclose(fpar);

  // number of tiles
  par.tnx = par.tmaxx-par.tminx+1;
  par.tny = par.tmaxy-par.tminy+1;

  // sensor dictionary
  sprintf(csen, "%s/%s", par.csen_cr, par.csen_mr);
  if (level2sensor_dict(csen, &par.l2dict) != SUCCESS) exit(1);

  // temporal subset for input
  par.yearmin = par.ymin + par.ywin[0];
  par.yearmax = par.ymin + par.ywin[par.bwin-1];
  par.doymin = 1; par.doymax = 365;

  // update window definition relative to prediction year
  for (w=0; w<par.bwin; w++) par.cewin[w] = 
    doy2ce(par.dwin[w], par.ymin+par.ywin[w]);


  #ifdef DEBUG
  printf("year: %04d - %04d\n", par.yearmin, par.yearmax);
  printf("day: %03d - %03d\n", par.doymin, par.doymax);
  #endif

  // choose products
  par.use.ref = true;
  par.use.qai = true;
  par.use.cld = false;
  par.use.hot = false;
  par.use.vzn = false;
  par.use.aod = false;
  par.use.wvp = false;
  par.use.imp = false;


  if (check_param_il2() != SUCCESS){
    printf(" Unable to read parameter file!\n"); return FAILURE;}

  return SUCCESS;
}

