/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
This file contains functions for quality assurance
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#include "force.h"


/** Allocate Quality Assurance Information
+++ This function allocates the QAI status flags.
--- QAI:    Quality Assurance Information (allocated within)
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int alloc_qai(qai_t *QAI){

  alloc((void**)&QAI->off, _NC_, sizeof(bool));
  alloc((void**)&QAI->cld, _NC_, sizeof(int1u));
  alloc((void**)&QAI->shd, _NC_, sizeof(bool));
  alloc((void**)&QAI->snw, _NC_, sizeof(bool));
  alloc((void**)&QAI->wtr, _NC_, sizeof(bool));
  alloc((void**)&QAI->aod, _NC_, sizeof(int1u));
  alloc((void**)&QAI->sub, _NC_, sizeof(bool));
  alloc((void**)&QAI->sat, _NC_, sizeof(bool));
  alloc((void**)&QAI->sun, _NC_, sizeof(bool));
  alloc((void**)&QAI->ill, _NC_, sizeof(int1u));
  alloc((void**)&QAI->slp, _NC_, sizeof(bool));
  alloc((void**)&QAI->wvp, _NC_, sizeof(bool));

  return SUCCESS;
}


/** Deallocate Quality Assurance Information
+++ This function deallocates the QAI status flags.
--- QAI:    Quality Assurance Information (deallocated within)
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int free_qai(qai_t QAI){

  free((void*)QAI.off);
  free((void*)QAI.cld);
  free((void*)QAI.shd);
  free((void*)QAI.snw);
  free((void*)QAI.wtr);
  free((void*)QAI.aod);
  free((void*)QAI.sub);
  free((void*)QAI.sat);
  free((void*)QAI.sun);
  free((void*)QAI.ill);
  free((void*)QAI.slp);
  free((void*)QAI.wvp);

  return SUCCESS;
}


/** Convert integer to bits
+++ This function converts an integer value to bit fields
--- integer: integer to be converted
--- bin:     bit field array (modified)
--- start:   starting bin (usually 0)
--- size:    number of bits (e.g. 16)
+++ Return:  void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void int2bit(int integer, bool *bin, int start, int size){
int x = integer, bit;

  for (bit=0; bit<size; bit++){
    bin[start+bit]= x % 2;
    x /= 2;
  }

  return;
}


/** Convert bits to integer
+++ This function converts bit fields to an integer value
--- bin:     bit field array
--- start:   starting bin (usually 0)
--- size:    number of bits (e.g. 16)
+++ Return:  integer
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int bit2int(bool *bin, int start, int size){
int x = 0, bit;

  for (bit=0; bit<size; bit++) x += bin[start+bit]*pow(2,bit);

  return x;
}


/** Decide whether to use this pixel
+++ This function checks the QAI layer against the user-defined QAI crite-
+++ ria. If all OK, the function returns SUCCESS. If any QAI flag is set,
+++ FAILURE is returned.
--- QAI:     Quality Assurance Information
--- p:       Pixel
+++ Return:  true/false
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
bool use_this_pixel(qai_t QAI, int p){

  if (par.qai_off && QAI.off[p]) return false;
  if (par.qai_cld_unc && QAI.cld[p] == 1) return false;
  if (par.qai_cld_opq && QAI.cld[p] == 2) return false;
  if (par.qai_cld_cir && QAI.cld[p] == 3) return false;
  if (par.qai_shd && QAI.shd[p]) return false;
  if (par.qai_snw && QAI.snw[p]) return false;
  if (par.qai_wtr && QAI.wtr[p]) return false;
  if (par.qai_aod_int  && QAI.aod[p] == 1) return false;
  if (par.qai_aod_high && QAI.aod[p] == 2) return false;
  if (par.qai_aod_fill && QAI.aod[p] == 3) return false;
  if (par.qai_sub && QAI.sub[p]) return false;
  if (par.qai_sat && QAI.sat[p]) return false;
  if (par.qai_sun && QAI.sun[p]) return false;
  if (par.qai_ill_low  && QAI.ill[p] == 1) return false;
  if (par.qai_ill_poor && QAI.ill[p] == 2) return false;
  if (par.qai_ill_shd  && QAI.ill[p] == 3) return false;
  if (par.qai_slp && QAI.slp[p]) return false;
  if (par.qai_wvp && QAI.wvp[p]) return false;

  return true;
}


/** Assemble Quality Assurance Information
+++ This function builds the final QAI bit layer by combining the indivi-
+++ dual QAI status flags. The QAI bit band is written to the data stack
+++ into the specified band.
--- QAI:     Quality Assurance Information
--- DAT:     Data stack
--- qa_band: band in data stack, where QAI should be stored
+++ Return:  SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int assemble_qai(qai_t QAI, int2s **DAT, int qa_band){
int p, bit, nbit = 16;
bool bin[16];


  #ifdef CLOCK 
  time_t TIME; time(&TIME);
  #endif


  for (p=0; p<_NC_; p++){

    // nodata flag
    bin[0] = QAI.off[p];

    if (QAI.off[p]){
      
      for (bit=1; bit<nbit; bit++) bin[bit] = 0;
      
    } else {

      // cloud status
      int2bit(QAI.cld[p], bin, 1, 2);

      // cloud shadow flag
      bin[3] = QAI.shd[p];

      // snow flag
      bin[4] = QAI.snw[p];

      // water flag
      bin[5] = QAI.wtr[p];

      // aerosol state
      int2bit(QAI.aod[p], bin, 6, 2);

      // subzero flag
      bin[8] = QAI.sub[p];

      // saturation flag
      bin[9] = QAI.sat[p];

      // high sun zenith, theta > 75°
      bin[10] = QAI.sun[p];

      // illumination state
      int2bit(QAI.ill[p], bin, 11, 2);
      
      // slope flag
      bin[13] = QAI.slp[p];
      
      // water vapor flag
      bin[14] = QAI.wvp[p];
      
      // Fusion flag, unused
      bin[15] = 0;

    }

    // convert to integer
    DAT[qa_band][p] = bit2int(bin, 0, nbit);

  }


  #ifdef CLOCK
  proctime_print("assemble QAI", TIME);
  #endif

  return SUCCESS;
}


/** Parse Quality Assurance Information
+++ This function parses a QAI integer layer and populates the individual
+++ QAI status flags.
--- DAT:     Quality Assurance Information integer (one-band)
--- QAI:     Quality Assurance Information status flags
+++ Return:  SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int parse_qai(qai_t QAI, int2s *DAT){
int p, nbit = 16;
bool bin[16];


  #pragma omp parallel private(bin) shared(DAT,QAI,_NC_,nbit) default(none)
  {

    #pragma omp for
    for (p=0; p<_NC_; p++){

      int2bit(DAT[p], bin, 0, nbit);
      
      // nodata flag
      QAI.off[p] = bin[0];

      if (QAI.off[p]) continue;

      // cloud status
      QAI.cld[p] = bit2int(bin, 1, 2);

      // cloud shadow flag
      QAI.shd[p] = bin[3];

      // snow flag
      QAI.snw[p] = bin[4];

      // water flag
      QAI.wtr[p] = bin[5];

      // aerosol state
      QAI.aod[p] = bit2int(bin, 6, 2);

      // subzero flag
      QAI.sub[p] = bin[8];

      // saturation flag
      QAI.sat[p] = bin[9];

      // high sun zenith, theta > 75°
      QAI.sun[p] = bin[10];

      // illumination state
      QAI.ill[p] = bit2int(bin, 11, 2);
      
      // slope flag
      QAI.slp[p] = bin[13];
      
      // water vapor flag
      QAI.wvp[p] = bin[14];
      
      // Fusion flag, unused
      // QAI. [p] = bin[15];

    }

  }

  return SUCCESS;
}

