/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
This program contains functions for deriving land surface phenology
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#include "force.h"


#ifdef SPLITS

#include <splits.h>

using namespace splits;


int derive_lsp(int p, int year){
int i, ii, i_, i0 = -1, i1 = -1, ni = par.m2d[_B_.tsi].length;
int yindex = year-par.yearmin-1;
char cdat0[MAX_STR_LEN];
char cdat1[MAX_STR_LEN];
float *y    = NULL;
//float *yhat = NULL;
float *w    = NULL;
float *doy  = NULL;
float dce, ce0;
float ymax = INT_MIN;
int doymax, yoff = 0;
bool southern;
bool valid = false;
float dseg;
int nseg;
float x_left, x_right, x;
float y_left, y_right;
Spline *spl;

  /** allocate **/
  alloc((void**)&y,    ni,  sizeof(float));
//alloc((void**)&yhat, ni,  sizeof(float));
  alloc((void**)&w,    ni,  sizeof(float));
  alloc((void**)&doy,  ni,  sizeof(float));
  
  /** which hemisphere? **/
  switch (par.lsp_hemi){
    case 0:
      southern = false; break; // N
    case 1:
      southern = true;  break; // S
    case 2:
      southern = false; break; // decide later
  }

  /** number of segments **/
  dseg = par.lsp_nseg/365.0;
  nseg = (int)round((365-par.lsp_dprev)*dseg + par.lsp_nseg + par.lsp_dnext*dseg);


  /** copy x/y/w to working variables **/
  for (i=0, ii=0; i<ni; i++){

    // subset by time
    if (par.m2d[_B_.tsi].date[i].year == year ||
       (par.m2d[_B_.tsi].date[i].year == year-1 && par.m2d[_B_.tsi].date[i].doy >= par.lsp_dprev) || 
       (par.m2d[_B_.tsi].date[i].year == year+1 && par.m2d[_B_.tsi].date[i].doy <= par.lsp_dnext)){

      if (i0 < 0) i0 = i;
      i1 = i+1;

      // copy DOY
      doy[ii] = par.m2d[_B_.tsi].date[i].doy;

      // linearly interpolate y-value, give half weight
      // if nothing to interpolate, assign nodata, and give 0 weight
      if (par.m2d[_B_.tsi].data[i][p] == _TSAOFF_){
        
        x_left = x_right = INT_MIN;
        x = par.m2d[_B_.tsi].date[i].ce;
        
        for (i_=i-1; i_>=0; i_--){
          if (par.m2d[_B_.tsi].date[i_].year < year-1 ||
             (par.m2d[_B_.tsi].date[i_].year == year-1 && par.m2d[_B_.tsi].date[i_].doy < par.lsp_dprev)){
            break;
          }
          if (par.m2d[_B_.tsi].data[i_][p] != _TSAOFF_){
            x_left = par.m2d[_B_.tsi].date[i_].ce;
            y_left = par.m2d[_B_.tsi].data[i_][p]/_RS_;
            break;
          }
        }
        for (i_=i+1; i_<ni; i_++){
          if (par.m2d[_B_.tsi].date[i_].year > year+1 ||
             (par.m2d[_B_.tsi].date[i_].year == year+1 && par.m2d[_B_.tsi].date[i_].doy > par.lsp_dnext)){
            break;
          }
          if (par.m2d[_B_.tsi].data[i_][p] != _TSAOFF_){
            x_right = par.m2d[_B_.tsi].date[i_].ce;
            y_right = par.m2d[_B_.tsi].data[i_][p]/_RS_;
            break;
          }
        }
        
        if (x_left > 0 && x_right > 0){
          y[ii] = (y_left*(x_right-x) + y_right*(x-x_left))/(x_right-x_left);
          w[ii] = 0.5;
        } else {
          y[ii] = _TSAOFF_;
          w[ii] = 0.0;
        }

      // copy y-value, determine max y, give max weight
      } else {

        y[ii] = par.m2d[_B_.tsi].data[i][p]/_RS_;
        w[ii] = 1.0;

        if (y[ii] > ymax){
          ymax = y[ii];
          doymax = par.m2d[_B_.tsi].date[i].doy;
        }

      }

      ii++;

    }

  }

  // length of compiled time series
  ni = ii;

  /** derive LSP if values are high enough **/

  if (ymax > 0.05){

    /** time step in days, and first day **/
    dce   = par.m2d[_B_.tsi].date[i1].ce - par.m2d[_B_.tsi].date[i0].ce;
    ce0   = par.m2d[_B_.tsi].date[i0].ce-par.cemin-365;
    //printf("ce0, dce: %f, %f\n", ce0, dce);

    /** allow switching hemispheres **/
    if (par.lsp_hemi == 2 && (doymax < 90 || doymax > 275)){
      southern = true; yoff = -1;}
    //printf("doymax: %d\n", doymax);

    /** compile temporal domaion **/
    sprintf(cdat0, "%d/%d", par.m2d[_B_.tsi].date[i0].doy, par.m2d[_B_.tsi].date[i0].year);
    sprintf(cdat1, "%d/%d", par.m2d[_B_.tsi].date[i1].doy, par.m2d[_B_.tsi].date[i1].year);
    Date dat0(cdat0); Date dat1(cdat1); Domain x(dat0, dat1, doy, ni);
    //printf("Domain: %s - %s\n", cdat0, cdat1);
  
    /** fit spline **/
    if ((spl = create_spline(x, y, w, UNIFORM_BSPLINE, nseg, 3)) == NULL){
      printf("NULL spline returned\n");}

    /** evaluate spline **/
    //evaluate(spl, x, yhat);
    //for (i=0; i<ni; i++) par.m2d[_B_.tsi].data[i+i0][p] = (int2s)(yhat[i]*_RS_);

    /** derive the LSP **/
    std::map<int, Pheno_set> season;
    season = phenology(spl, x, southern, 0.2, PHENOLOGY_RAW);
    Pheno_set ph = season[year+yoff];
  
    /** sanity check **/
    if (ph.peak_val > 0.05 &&
        ph.doy_early_min < ph.doy_start_green && 
        ph.doy_start_green < ph.doy_peak &&
        ph.doy_peak < ph.doy_end_green && 
        ph.doy_end_green < ph.doy_late_min && 
        ph.min_min_integral > 0 &&
        ph.latent_integral  > 0 &&
        ph.total_integral   > 0 &&
        ph.green_integral   > 0) valid = true;

    /** copy LSP if all OK **/
    if (valid){
      par.m2d[_B_.dem].data[yindex][p] = (int2s)(ph.doy_early_min*dce+ce0);   // days since 1st LSP year
      par.m2d[_B_.dss].data[yindex][p] = (int2s)(ph.doy_start_green*dce+ce0); // days since 1st LSP year
      par.m2d[_B_.dri].data[yindex][p] = (int2s)(ph.doy_early_flex*dce+ce0);  // days since 1st LSP year
      par.m2d[_B_.dps].data[yindex][p] = (int2s)(ph.doy_peak*dce+ce0);        // days since 1st LSP year
      par.m2d[_B_.dfi].data[yindex][p] = (int2s)(ph.doy_late_flex*dce+ce0);   // days since 1st LSP year
      par.m2d[_B_.des].data[yindex][p] = (int2s)(ph.doy_end_green*dce+ce0);   // days since 1st LSP year
      par.m2d[_B_.dlm].data[yindex][p] = (int2s)(ph.doy_late_min*dce+ce0);    // days since 1st LSP year
      par.m2d[_B_.lts].data[yindex][p] = (int2s)(ph.min_min_duration*dce);    // days
      par.m2d[_B_.lgs].data[yindex][p] = (int2s)(ph.green_duration*dce);      // days
      par.m2d[_B_.vem].data[yindex][p] = (int2s)(ph.early_min_val*_RS_);      // index value * 10000
      par.m2d[_B_.vss].data[yindex][p] = (int2s)(ph.start_green_val*_RS_);    // index value * 10000
      par.m2d[_B_.vri].data[yindex][p] = (int2s)(ph.early_flex_val*_RS_);     // index value * 10000
      par.m2d[_B_.vps].data[yindex][p] = (int2s)(ph.peak_val*_RS_);           // index value * 10000
      par.m2d[_B_.vfi].data[yindex][p] = (int2s)(ph.late_flex_val*_RS_);      // index value * 10000
      par.m2d[_B_.ves].data[yindex][p] = (int2s)(ph.end_green_val*_RS_);      // index value * 10000
      par.m2d[_B_.vlm].data[yindex][p] = (int2s)(ph.late_min_val*_RS_);       // index value * 10000
      par.m2d[_B_.vbl].data[yindex][p] = (int2s)(ph.latent_val*_RS_);         // index value * 10000
      par.m2d[_B_.vsa].data[yindex][p] = (int2s)(ph.amplitude*_RS_);          // index value * 10000
      par.m2d[_B_.ist].data[yindex][p] = (int2s)(ph.min_min_integral*dce*10); // days * index value * 10
      par.m2d[_B_.ibl].data[yindex][p] = (int2s)(ph.latent_integral*dce*10);  // days * index value * 10
      par.m2d[_B_.ibt].data[yindex][p] = (int2s)(ph.total_integral*dce*10);   // days * index value * 10
      par.m2d[_B_.igs].data[yindex][p] = (int2s)(ph.green_integral*dce*10);   // days * index value * 10
      par.m2d[_B_.rar].data[yindex][p] = (int2s)(ph.greenup_rate/dce*_RS_);    // index value *10000 / days
      par.m2d[_B_.raf].data[yindex][p] = (int2s)(ph.senescence_rate/dce*_RS_); // index value *10000 / days
      par.m2d[_B_.rmr].data[yindex][p] = (int2s)(ph.early_flex_rate/dce*_RS_); // index value *10000 / days
      par.m2d[_B_.rmf].data[yindex][p] = (int2s)(ph.late_flex_rate/dce*_RS_);  // index value *10000 / days
    }

    destroy_spline(spl);

  }

  /** nodata if deriving LSP failed **/
  if (!valid){
    par.m2d[_B_.dem].data[yindex][p] = par.m2d[_B_.dss].data[yindex][p] = _TSAOFF_;
    par.m2d[_B_.dri].data[yindex][p] = par.m2d[_B_.dps].data[yindex][p] = _TSAOFF_;
    par.m2d[_B_.dfi].data[yindex][p] = par.m2d[_B_.des].data[yindex][p] = _TSAOFF_;
    par.m2d[_B_.dlm].data[yindex][p] = par.m2d[_B_.lts].data[yindex][p] = _TSAOFF_;
    par.m2d[_B_.lgs].data[yindex][p] = par.m2d[_B_.vem].data[yindex][p] = _TSAOFF_;
    par.m2d[_B_.vss].data[yindex][p] = par.m2d[_B_.vri].data[yindex][p] = _TSAOFF_;
    par.m2d[_B_.vps].data[yindex][p] = par.m2d[_B_.vfi].data[yindex][p] = _TSAOFF_;
    par.m2d[_B_.ves].data[yindex][p] = par.m2d[_B_.vlm].data[yindex][p] = _TSAOFF_;
    par.m2d[_B_.vbl].data[yindex][p] = par.m2d[_B_.vsa].data[yindex][p] = _TSAOFF_;
    par.m2d[_B_.ist].data[yindex][p] = par.m2d[_B_.ibl].data[yindex][p] = _TSAOFF_;
    par.m2d[_B_.ibt].data[yindex][p] = par.m2d[_B_.igs].data[yindex][p] = _TSAOFF_;
    par.m2d[_B_.rar].data[yindex][p] = par.m2d[_B_.raf].data[yindex][p] = _TSAOFF_;
    par.m2d[_B_.rmr].data[yindex][p] = par.m2d[_B_.rmf].data[yindex][p] = _TSAOFF_;
  }

 
  /** clean **/
  free((void*)y); //free((void*)yhat);
  free((void*)w);  free((void*)doy);

  return SUCCESS;
}

#endif

