/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#include "force.h"


/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
This program creates a shapefile containing the tile IDs of the prepro-
cessing grid + the bounding box of each tile in projected coordinates
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


int main( int argc, char *argv[] ){
char *dir = NULL;
char pname[MAX_STR_LEN];
char *pext = NULL;
char tile[MAX_STR_LEN];
int nx_grid, ny_grid;
int px, py, x0, y0, x1, y1, i, j, c;
float left, right, top, bottom;
float startx, starty;
char *wkt = par.css_dst;
OGRSFDriverH driver;
OGRSpatialReferenceH srs;
GDALDatasetH fp;
OGRLayerH layer;
OGRFieldDefnH field;
OGRFeatureH feature;
OGRGeometryH poly, ring;
coord_t geo[4], map[4], min, max;



  if (argc != 6){ printf("usage: %s datacube bottom top left right\n", argv[0]); exit(1);}


  OGRRegisterAll();

  // get command line parameters
  dir      = argv[1];
  geo[0].x = atof(argv[4]); geo[0].y = atof(argv[2]); // LL
  geo[1].x = atof(argv[5]); geo[1].y = atof(argv[2]); // LR
  geo[2].x = atof(argv[4]); geo[2].y = atof(argv[3]); // UL
  geo[3].x = atof(argv[5]); geo[3].y = atof(argv[3]); // UR
  min.x = INT_MAX; min.y = INT_MAX; max.x = INT_MIN; max.x = INT_MIN; 
  par.d_img   = NULL;

  // get driver
  if ((driver = OGRGetDriverByName("ESRI Shapefile")) == NULL){
    printf("ESRI Shapefile driver is not available\n"); exit(1);}

  // read datacube definition
  if (read_datacube_def(dir) == FAILURE){
    printf("Reading datacube definition failed.\n"); exit(1);}

  // Output name
  sprintf(pname, "%s/datacube-grid.shp", dir);
  printf("%s\n", pname);
  
  if (fileexist(pname)){
    printf("Shapefile already exists.\n"); exit(1);}

  // get border coordinates in target css coordinates
  for (c=0; c<4; c++){
    if ((geo_to_map(geo[c].x,  geo[c].y, &map[c].x, &map[c].y, 
                     par.css_dst)) == FAILURE){
      printf("Computing bbox coordinates in dst_srs failed!\n"); exit(1);}
    if (map[c].x < min.x) min.x = map[c].x;
    if (map[c].y < min.y) min.y = map[c].y;
    if (map[c].x > max.x) max.x = map[c].x;
    if (map[c].y > max.y) max.y = map[c].y;
  }

  // find the UL and LR tile
  tile_find(min.x, max.y, &startx, &starty, &x0, &y0);
  tile_find(max.x, min.y, &startx, &starty, &x1, &y1);


  // create file
  if ((fp = OGR_Dr_CreateDataSource(driver, pname, NULL)) == NULL){
    printf("Error creating output file.\n"); exit(1);}
    

  // set output projection
  srs = OSRNewSpatialReference(NULL);
  OSRImportFromWkt(srs, &wkt);

  // create layer
  if ((layer = OGR_DS_CreateLayer(fp, "processing grid", srs, 
        wkbPolygon, NULL)) == NULL){
    printf("Error creating layer.\n"); exit(1);}

  // add field
  field = OGR_Fld_Create("Tile_ID", OFTString);
  OGR_Fld_SetWidth(field, 12);
  if (OGR_L_CreateField(layer, field, TRUE) != OGRERR_NONE){
      printf("Error creating field.\n"); exit(1);}
  OGR_Fld_Destroy(field);

  // add field
  field = OGR_Fld_Create("Tile_X", OFTInteger );
  OGR_Fld_SetWidth(field, 4);
  if (OGR_L_CreateField(layer, field, TRUE) != OGRERR_NONE){
      printf("Error creating field.\n"); exit(1);}
  OGR_Fld_Destroy(field);

  // add field
  field = OGR_Fld_Create("Tile_Y", OFTInteger );
  OGR_Fld_SetWidth(field, 4);
  if (OGR_L_CreateField(layer, field, TRUE) != OGRERR_NONE){
      printf("Error creating field.\n"); exit(1);}
  OGR_Fld_Destroy(field);


  for (py=y0; py<=y1; py++){
  for (px=x0; px<=x1; px++){

    // bounding box
    left   = par.origin_map.x + par.tilesize*px;
    right  = par.origin_map.x + par.tilesize*(px+1);
    top    = par.origin_map.y - par.tilesize*py;
    bottom = par.origin_map.y - par.tilesize*(py+1);

    // tile ID
    sprintf(tile, "X%04d_Y%04d", px, py);

    // create feature
    feature = OGR_F_Create(OGR_L_GetLayerDefn(layer));

    // set fields
    OGR_F_SetFieldString(feature, OGR_F_GetFieldIndex(feature, "Tile_ID"), tile);
    OGR_F_SetFieldInteger(feature, OGR_F_GetFieldIndex(feature, "Tile_X"), px);
    OGR_F_SetFieldInteger(feature, OGR_F_GetFieldIndex(feature, "Tile_Y"), py);

    // create local geometry
    poly = OGR_G_CreateGeometry(wkbPolygon);
    ring = OGR_G_CreateGeometry(wkbLinearRing );
    OGR_G_SetPointCount(ring, 5);
    OGR_G_SetPoint_2D(ring, 0, left,  top);
    OGR_G_SetPoint_2D(ring, 1, right, top);
    OGR_G_SetPoint_2D(ring, 2, right, bottom);
    OGR_G_SetPoint_2D(ring, 3, left,  bottom);
    OGR_G_SetPoint_2D(ring, 4, left,  top);
    OGR_G_AddGeometry(poly, ring);
    OGR_F_SetGeometry(feature, poly);
    OGR_G_DestroyGeometry(poly);

    // create feature in the file
    if (OGR_L_CreateFeature(layer, feature) != OGRERR_NONE){
      printf("Error creating feature in file.\n"); exit(1);}
    OGR_F_Destroy(feature);

  }
  }

  //GDALClose(fp);
  OGR_DS_Destroy(fp);

}

