/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
This file contains functions that handle tiling operations
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#include "force.h"


/** Read tile white-list
+++ This function reads the tile white-list. If none is given, it is gra-
+++ cefully ignored. The X/Y IDs and the number of entries is returned.
--- X:      X tile coordinate IDs (returned)
--- Y:      Y tile coordinate IDs (returned)
--- k:      number of tiles       (returned)
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int tilefileread(int **X, int **Y, int *k){
FILE *fp = NULL;
char  buffer[MAX_STR_LEN] = "\0";
int i, *x = NULL, *y = NULL;
char cx[5], cy[5];


  // if no file is specified, make sure that processing can continue
  if ((strcmp(par.f_tile, "NULL") == 0)){
    *k = 0;
    alloc((void**)&x, 1, sizeof(int)); *X = x;
    alloc((void**)&y, 1, sizeof(int)); *Y = y;
    return SUCCESS;
  }


  if ((fp = fopen(par.f_tile, "r")) == NULL){
    printf("Unable to open tile file!\n"); return FAILURE;}

  if (fgets(buffer, MAX_STR_LEN, fp) == NULL){
    printf("Error while reading from tile file!\n"); return FAILURE;}

  // number of positive tiles should be in 1st line
  if ((*k = atoi(buffer)) <= 0){
    printf("Error retrieving number of lines in tile file!\n"); return FAILURE;}

  alloc((void**)&x, *k, sizeof(int)); 
  alloc((void**)&y, *k, sizeof(int));

  // read line by line, one Tile ID should be in each line
  for (i=0; i<*k; i++){

    if (fgets(buffer, MAX_STR_LEN, fp) == NULL){
      printf("Error reading line %d in tile file\n", i+2); return FAILURE;}

    strncpy(cx, buffer+1, 4);
    strncpy(cy, buffer+7, 4);
    cx[4] = '\0'; x[i] = atoi(cx);
    cy[4] = '\0'; y[i] = atoi(cy);

  }

  fclose(fp);
  
  #ifdef DEBUG
  printf("positive tile list contains %d tiles\n", *k);
  #endif

  *X = x;
  *Y = y;
  return SUCCESS;
}


/** Test if tile is white-listed
+++ This function tests whether a given tile is white-listed or not.
--- white_x: white-listed X tile coordinate IDs
--- white_y: white-listed Y tile coordinate IDs
--- white_k: number of white-listed tiles
--- x:       x tile coordinate ID
--- y:       y tile coordinate ID
+++ Return:  SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int tile_whitelisted(int *white_x, int *white_y, int white_k, int x, int y){
int k;
bool white = false;

  if (white_k > 0){
    for (k=0; k<white_k; k++){
      if (x == white_x[k] && y == white_y[k]) white = true;
    }
  } else {
    white = true;
  }
  
  if (white) return SUCCESS; else return FAILURE;
}


/** Tile finder
+++ This function returns the tile ID and tile upper left coordinate of
+++ the requestedcoordinate.
--- x:       x coordinate
--- y:       y coordinate
--- tilex:   upper left x coordinate of tile (returned)
--- tiley:   upper left y coordinate of tile (returned)
--- idx:     x tile coordinate ID (returned)
--- idy:     y tile coordinate ID (returned)
+++ Return:  SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int tile_find(float x, float y, float *tilex, float *tiley, int *idx, int *idy){
float tx = par.origin_map.x, ty = par.origin_map.y;
int px = 0, py = 0;

  // if x is right of tile-origin
  if (x > tx){

    // shift tile-x to the right until x is left of tile-x
    while (x > tx){
      tx += par.tilesize;
      px++;
    }
    // step back one tile; x should be right of tile-x
    tx -= par.tilesize;
    px--;

  // if x is left of tile-origin
  } else if (x < tx){

    // shift tile-x to the left until x is right of tile-x
    while (x < tx){
      tx -= par.tilesize;
      px--;
    }
  }
  // if x is tile-origin, nothing to do

  // if y is below tile-origin
  if (y < ty){

    // shift tile-y to the bottom until y is above tile-y
    while (y < ty){
      ty -= par.tilesize;
      py++;
    }
    // step back one tile; y should be below tile-y
    ty += par.tilesize;
    py--;
 
  // if y is above tile-origin
  } else if (y > ty){

    // shift tile-y to the top until y is below tile-y
    while (y > ty){
      ty += par.tilesize;
      py--;
    }
  }
  // if y is tile-origin, nothing to do

  #ifdef DEBUG
  printf("ul x/y of first tile (X%04d_Y%04d): %.2f/%.2f\n", px, py, tx, ty);
  #endif

  *idx   = px; // tile id x
  *idy   = py; // tile id y
  *tilex = tx; // ulx of tile
  *tiley = ty; // uly of tile

  return SUCCESS;
}


/** Tile counter
+++ This function returns the number of tiles in x- and y-direction that
+++ cover the original image.
--- nx:     number of pixels in x-direction
--- ny:     number of pixels in y-direction
--- ulx:    upper-left x-coordinate
--- uly:    upper-left y-coordinate
--- t_ulx:  upper left x coordinate of top-left tile
--- t_uly:  upper left y coordinate of top-left tile
--- ntx:    number of tiles in x-direction (returned)
--- nty:    number of tiles in y-direction (returned)
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int tile_count(int nx, int ny, float ulx, float uly, float t_ulx, float t_uly, int *ntx, int *nty){
float lrx, lry;

  // compute LR coordinate
  lrx = ulx+par.res*nx;
  lry = uly-par.res*ny;

  // number of tiles
  *ntx = ceil((lrx-t_ulx)/(par.tilesize));
  *nty = ceil((t_uly-lry)/(par.tilesize));
  
  #ifdef DEBUG
  printf("number of tiles: %d/%d\n", *ntx, *nty);
  #endif

  return SUCCESS;
}


/** Output function for gridded images
+++ This function tiles the image, computes tile cloud coverage and writes
+++ gridded images to disc.
--- DAT:    Data stack
--- nx:     number of pixels in x-direction
--- ny:     number of pixels in y-direction
--- ulx:    upper-left x-coordinate
--- uly:    upper-left y-coordinate
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int tile_image(int2s **DAT, int nx, int ny, float ulx, float uly){
int2s **SUB = NULL;
float t0x, t0y; // ul x/y of first   tile
float tx,  ty;  // ul x/y of current tile
int idx, idy; // id of first tile
int ntx, nty; // number of tiles in x- and y-direction
int istart, jstart; // image-to-chip offset
int ntile = 0; // number of written tiles
int nbit = 16;
bool bin[16];
int *tiles_x = NULL, *tiles_y = NULL, tiles_k;
int x, y, i, j, k, b, p, pp;
bool empty;
double geotransform[6];
double ncld, ndata; // cloud cover


  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  // get tile file and read it
  if (tilefileread(&tiles_x, &tiles_y, &tiles_k) != SUCCESS){
    printf("Reading tile file failed! "); return FAILURE;}

  geotransform[1] = par.res;
  geotransform[2] = 0;
  geotransform[4] = 0;
  geotransform[5] = -1*par.res;

  // find the top-left tile (id + tile ul x/y)
  tile_find(ulx, uly, &t0x, &t0y, &idx, &idy);
  // compute number of tiles in x- and y-direction
  tile_count(nx, ny, ulx, uly, t0x, t0y, &ntx, &nty);

  alloc_2D((void***)&SUB, _ND_, par.tilesize_px*par.tilesize_px, sizeof(int2s));

  // intersect the image with the tile grid
  for (y=0, ty=t0y; y<nty; y++, ty-=par.tilesize){
  for (x=0, tx=t0x; x<ntx; x++, tx+=par.tilesize){
  
    // if tile is not whitelisted (if specified), skip
    if (tile_whitelisted(tiles_x, tiles_y, tiles_k, idx+x, idy+y) == FAILURE) continue;

    empty = true;
    ndata = ncld = 0.0;

    // image offset relative to current tile
    jstart = floor((tx-ulx)/par.res); // changed from round to floor
    istart = floor((uly-ty)/par.res); // changed from round to floor

    // copy image to chip + compute tile cloud cover
    for (i=0; i<par.tilesize_px; i++){
    for (j=0; j<par.tilesize_px; j++){

      p  = nx*(i+istart) + j+jstart; // image
      pp = par.tilesize_px*i+j;      // chip

      if (i+istart < 0 || i+istart >= ny || j+jstart < 0 || j+jstart >= nx){

        // fill chip with nodata if out-of-image
        for (b=0; b<_ND_; b++){
          if (b == par.m2d[_B_.qai].start){
            SUB[b][pp] = par.m2d[_B_.qai].nodata;
          } else {
            SUB[b][pp] = _NODATA_;
          }
        }

      } else {

        for (b=0; b<_ND_; b++) SUB[b][pp] = DAT[b][p];

        int2bit(SUB[par.m2d[_B_.qai].start][pp], bin, 0, nbit);

        // count valid and cloudy pixels
        if (bin[0] == 0){
          if (bit2int(bin, 1, 2) > 0 || bin[3] == 1 || bin[4] == 1) ncld++;
          ndata++;
          empty = false;
        }

      }

    }
    }


    // meteor cover of tile
    if (ndata > 0) ncld = ncld/ndata*100.0;


    #ifdef DEBUG
    printf("tile X%04d_Y%04d: empty: %d, cloud cover: %03.0f%%\n", 
      idx+x, idy+y, empty, ncld);
    #endif

    if (ncld > par.maxtc) empty = true;


    // if there are data in tile -> output
    if (!empty){

      geotransform[0] = tx;
      geotransform[3] = ty;

      sprintf(par.d_out, "%s/X%04d_Y%04d", par.d_level2, idx+x, idy+y);

      if (write_products_l2(SUB, par.tilesize_px, par.tilesize_px, 
          jstart+par.tilesize_px/2, istart+par.tilesize_px/2, 
          geotransform) != SUCCESS){
        printf("Error in writing products! "); return FAILURE;}

      ntile++;

    }

  }
  }

  // print to stdout for logfile
  printf("%2d product(s) written. ", ntile);
  
  // clean
  free_2D((void**)SUB, _ND_);
  free((void*)tiles_x); free((void*)tiles_y);

  #ifdef CLOCK
  proctime_print("tiling output stack", TIME);
  #endif

  return SUCCESS;
}


/** Output function for full-frame images
+++ This function is the alternative to tile_image and simply flushes the
+++ image to dics as it is.
--- DAT:    Data stack
--- nx:     number of pixels in x-direction
--- ny:     number of pixels in y-direction
--- ulx:    upper-left x-coordinate
--- uly:    upper-left y-coordinate
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int flush_image(int2s **DAT, int nx, int ny, float ulx, float uly){
double geotransform[6];


  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif
  
  #ifdef DEBUG
  db_write_image_2D((void**)DAT, "REF4", _ND_, _NX_, _NY_, sizeof(int2s), 2);
  #endif


  geotransform[0] = ulx;
  geotransform[1] = par.res;
  geotransform[2] = 0;
  geotransform[3] = uly;
  geotransform[4] = 0;
  geotransform[5] = -1*par.res;

  sprintf(par.d_out, "%s/%s", par.d_level2, meta.wrs);

  if (write_products_l2(DAT, nx, ny, nx/2, ny/2, geotransform) != SUCCESS){
     printf("Error in writing products! "); return FAILURE;}

  printf(" 1 product(s) written. ");

  #ifdef CLOCK
  proctime_print("flushing output stack", TIME);
  #endif


  return SUCCESS;
}


/** Write spatial definition of data cube
+++ This function writes a text file with all spatial specs of the data
+++ cube. The file contains (1) projection as WKT string, (2) origin of 
+++ the tile system as geographic Longitude, (3) origin of the tile system
+++ as geographic Latitude, (4) width of the tiles in projection units.
--- d_write: Output directory
+++ Return:  SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int write_datacube_def(char *d_write){
char fname[MAX_STR_LEN];
char *lock = NULL;
FILE *fp = NULL;

  sprintf(fname, "%s/datacube-definition.prj", d_write);

  if (!fileexist(fname)){
    
    if ((lock = (char*)CPLLockFile(fname, 60)) == NULL){
      printf("Unable to lock file (timeout: 60s). "); return FAILURE;}

    if ((fp = fopen(fname, "w")) == NULL){
      printf("Unable to open %s. ", fname); return FAILURE;}

    fprintf(fp, "%s\n", par.css_dst);
    fprintf(fp, "%f\n", par.origin_geo.x);
    fprintf(fp, "%f\n", par.origin_geo.y);
    fprintf(fp, "%f\n", par.origin_map.x);
    fprintf(fp, "%f\n", par.origin_map.y);
    fprintf(fp, "%f\n", par.tilesize);

    fclose(fp);

    CPLUnlockFile(lock);

  }

  return SUCCESS;
}


/** Read spatial definition of data cube
+++ This function read the text file with all spatial specs of the data
+++ cube.
--- d_read: Input directory
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int read_datacube_def(char *d_read){
char fname[MAX_STR_LEN];
char buffer[MAX_STR_LEN] = "\0";
FILE *fp = NULL;

  sprintf(fname, "%s/datacube-definition.prj", d_read);

  if ((fp = fopen(fname, "r")) == NULL){
    printf("Unable to open %s. ", fname); return FAILURE;}

  if (fgets(buffer, MAX_STR_LEN, fp) == NULL){
    printf("Unable to read projection. "); return FAILURE;
  } else {
    strncpy(par.css_dst, buffer, strlen(buffer)-1);
    par.css_dst[strlen(buffer)-1] = '\0';
  }

  if (fgets(buffer, MAX_STR_LEN, fp) == NULL){
    printf("Unable to read X-origin (geo). "); return FAILURE;
  } else {
    par.origin_geo.x = atof(buffer);
  }

  if (fgets(buffer, MAX_STR_LEN, fp) == NULL){
    printf("Unable to read Y-origin (geo). "); return FAILURE;
  } else {
    par.origin_geo.y = atof(buffer);
  }

  if (fgets(buffer, MAX_STR_LEN, fp) == NULL){
    printf("Unable to read X-origin (map). "); return FAILURE;
  } else {
    par.origin_map.x = atof(buffer);
  }

  if (fgets(buffer, MAX_STR_LEN, fp) == NULL){
    printf("Unable to read Y-origin (map). "); return FAILURE;
  } else {
    par.origin_map.y = atof(buffer);
  }

  if (fgets(buffer, MAX_STR_LEN, fp) == NULL){
    printf("Unable to read tile size. "); return FAILURE;
  } else {
    par.tilesize = atof(buffer);
  }

  fclose(fp);

  return SUCCESS;
}


/** Copy spatial definition of data cube
+++ This function copies the text file with all spatial specs of the data
+++ cube.
--- d_read:  Input directory
--- d_write: Output directory
+++ Return:  SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int copy_datacube_def(char *d_read, char *d_write){

  if (read_datacube_def(d_read) == FAILURE){
    printf("Reading datacube definition failed. "); return FAILURE;}
  if (write_datacube_def(d_write) == FAILURE){
    printf("Writing datacube definition failed. "); return FAILURE;}

  return SUCCESS;
}

/** Geometric module
+++ The geometric module is the 3rd core module of FORCE L2PS, and handles
+++ the geometric homogenization useful for higher-level analyses and pro-
+++ cessing. Reprojection, tiling, and final output is done here.
--- DAT0:   Data stack
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int geometric_module(int2s **DAT0){
int l;
int nx, ny;     // number of lines/samples of reprojected image
float ulx, uly; // ul-corner of reprojected image
int2s **DAT = NULL;  // array 4 reprojected data


  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif
  
  if (!par.doreproj) strcpy(par.css_dst, meta.css_map);
  
  if (par.dotile){

    // get origin of arbitrary grid in target css coordinates
    if ((geo_to_map(par.origin_geo.x,  par.origin_geo.y, 
            &par.origin_map.x, &par.origin_map.y, 
             par.css_dst)) == FAILURE){
      printf("Computing tile origin in dst_srs failed. "); return FAILURE;}

    // write data cube definition
    if (write_datacube_def(par.d_level2) == FAILURE){
      printf("Writing datacube definition failed. "); return FAILURE;}

  }

  /** reproject data to target srs - or leave it in original srs
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

  if (par.doreproj){
    if ((warp_image(DAT0, &DAT, meta.map_ul.x, meta.map_ul.y,
        meta.css_map, &ulx, &uly, &nx, &ny)) != SUCCESS){
      printf("Reprojecting DAT failed! "); return FAILURE; }
    if ((warp_qai(DAT0, DAT, meta.map_ul.x, meta.map_ul.y,
        meta.css_map, ulx, uly, nx, ny)) != SUCCESS){
      printf("Reprojecting QAI failed! "); return FAILURE; }
    free_2D((void**)DAT0, _ND_);
  } else {
    DAT = DAT0;
    ulx = meta.map_ul.x; 
    uly = meta.map_ul.y;
    nx = _NX_; ny = _NY_;
  }


  /** tile or flush the data to disc
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

  if (par.dotile){

    // tile the images
    if ((tile_image(DAT, nx, ny, ulx, uly)) != SUCCESS){
      printf("Tiling images failed! "); return FAILURE; }

  } else {

    // write images as they are
    if ((flush_image(DAT, nx, ny, ulx, uly)) != SUCCESS){
      printf("Flushing images to disc failed! "); return FAILURE; }

  }


  /** clean
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

  if (par.doreproj){
    free((void*)DAT[0]); free((void*)DAT);
  } else {
    free_2D((void**)DAT, _ND_);
  }
  free((void*)par.use.refbands);
  free((void*)par.b_img);
  atc_free();
  cgfree();

  for (l=0; l<_NP_; l++){
    if (par.m2d[l].ts != NULL){
      free((void*)par.m2d[l].ts);
    }
    if (par.m2d[l].bn  != NULL){
      free_2D((void**)par.m2d[l].bn, par.m2d[l].length);
    }
    if (par.m2d[l].scale != NULL){
      free((void*)par.m2d[l].scale);
    }
  }

  #ifdef CLOCK
  proctime_print("geometric module", TIME);
  #endif

  return SUCCESS;
}

