/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
This file contains functions for time series preparation and analysis
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#include "force.h"


/** Despike time series
+++ This function detects spikes (positive outliers) and removes them
--- MASK:   Analysis mask
--- thr:    Threshold above which an outlier is considered a spike
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int despike_timeseries(bool *MASK, int thr){
int t, tt, p;
float x_left, x_right, x;
float y_left, y_right, y, y_;


  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  #ifdef DEBUG
  printf("Despiking time series. "); fflush(stdout);
  #endif


  #pragma omp parallel private(x_left,x_right,x,y_left,y_right,y,y_,t,tt) shared(MASK,par,_NC_,_B_,dir,thr) default(none)
  {

    #pragma omp for schedule(guided)
    for (p=0; p<_NC_; p++){
      
      if (MASK != NULL && !MASK[p]) continue;

      x_left = y_left = INT_MAX;

      for (t=0; t<dir.n-1; t++){

        if (par.m2d[_B_.tss].data[t][p] == _TSAOFF_) continue;

        x = par.m2d[_B_.tss].date[t].ce;
        y = par.m2d[_B_.tss].data[t][p];

        if (x_left < x){

          x_right = INT_MAX;
          y_right = INT_MAX;

          for (tt=t+1; tt<dir.n; tt++){
            if (par.m2d[_B_.tss].data[tt][p] == _TSAOFF_) continue;
            x_right = par.m2d[_B_.tss].date[tt].ce;
            y_right = par.m2d[_B_.tss].data[tt][p];
            break;
          }

          if (x_right == INT_MAX) continue;

          y_ = (y_left*(x_right-x) + y_right*(x-x_left))/(x_right-x_left);

          //printf("ce: %d %d %d\n", x_left, par.m2d[_B_.tss].date[t].ce, x_right);
          //printf("y: %f %f %f\n", y_left, par.m2d[_B_.tss].data[t][p]/_RS_, y_right);
          //printf("y_: %f, dy: %f\n", y_, fabs(y_-par.m2d[_B_.tss].data[t][p]/_RS_));

          if (fabs(y_-y) > thr){
            par.m2d[_B_.tss].data[t][p] = _TSAOFF_;
            continue;
          }

        }

        x_left = x;
        y_left = y;

      }

    }
    
  }


  #ifdef CLOCK
  proctime_print("despiking time series", TIME);
  #endif

  return SUCCESS;
}


/** Interpolate time series
+++ This function interpolates a time series at given steps. Currently im-
+++ plemented are linear, moving mean/min/max, RBF ensembles
--- MASK:   Analysis mask
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int interpolate_timeseries(bool *MASK){
int t, i;

  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  #ifdef DEBUG
  printf("Interpolating time series. "); fflush(stdout);
  #endif


  if (par.otsi || par.lsp){
    alloc_2D((void***)&par.m2d[_B_.tsi].data,  par.m2d[_B_.tsi].length, _NC_, sizeof(int2s));
    alloc((void**)&par.m2d[_B_.tsi].scale, par.m2d[_B_.tsi].length, sizeof(int2s));
    alloc((void**)&par.m2d[_B_.tsi].date, par.m2d[_B_.tsi].length,       sizeof(date_t));
    alloc((void**)&par.m2d[_B_.tsi].ts,   par.m2d[_B_.tsi].length,       sizeof(float));
    alloc_2D((void***)&par.m2d[_B_.tsi].bn, par.m2d[_B_.tsi].length, MAX_STR_LEN, sizeof(char));
    for (i=0; i<par.int_n; i++){
      par.m2d[_B_.tsi].scale[i] = _RS_;
      par.m2d[_B_.tsi].date[i].ce = par.cemin + i*par.int_nstep;
      ce2date(par.m2d[_B_.tsi].date[i].ce, &par.m2d[_B_.tsi].date[i].month, &par.m2d[_B_.tsi].date[i].day, &par.m2d[_B_.tsi].date[i].year);
      par.m2d[_B_.tsi].date[i].doy   = md2doy(par.m2d[_B_.tsi].date[i].month, par.m2d[_B_.tsi].date[i].day);
      par.m2d[_B_.tsi].date[i].week  = doy2week(par.m2d[_B_.tsi].date[i].doy);
      par.m2d[_B_.tsi].ts[i] =  par.m2d[_B_.tsi].date[i].year + par.m2d[_B_.tsi].date[i].doy/366.0;
      sprintf(par.m2d[_B_.tsi].bn[i], "Interpolated %s", par.index_c);
    }
  }


  switch (par.interpolate){
    case LINEAR:
      interpolate_linear(MASK);
      break;
    case MOVING:
      interpolate_moving(MASK);
      break;
    case RBF:
      interpolate_rbf(MASK);
      break;
  }


  #ifdef CLOCK
  proctime_print("interpolating time series", TIME);
  #endif

  return SUCCESS;
}


/** Derive Land Surface Phenology from time series
+++ This function derives phenometrics from an interpolated time series.
+++ The function will fail if FORCE was not compiled with SPLITS (see in-
+++ stallation instructions). Do not expect this function to work if you
+++ have fairly sparse time series. In addition, the time series needs to
+++ be long enough (at least extend into the previous and next year). Phe-
+++ nometrics are derived for each given year.
--- MASK:   Analysis mask
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int phenology_timeseries(bool *MASK){
int l, t, p, y;
bool valid;


  #ifndef SPLITS
  printf("SpliTS is not available.\n");
  printf("Install SpliTS and re-compile (see user guide)\n");
  printf(" OR use another interpolation method.\n");
  return FAILURE;
  #endif

  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  #ifdef DEBUG
  printf("Phenology from time series. "); fflush(stdout);
  #endif


  // allocate data
  for (l=_B_.dem; l<=_B_.rmf; l++){
    alloc_2D((void***)&par.m2d[l].data, par.m2d[l].length, _NC_, sizeof(int2s));
    alloc((void**)&par.m2d[l].scale, par.m2d[l].length, sizeof(int2s));
    alloc((void**)&par.m2d[l].ts, par.m2d[l].length, sizeof(float));
    alloc_2D((void***)&par.m2d[l].bn, par.m2d[l].length, MAX_STR_LEN, sizeof(char));
  }


  for (y=0; y<par.Yn; y++){

    // set timestamp
    for (l=_B_.dem; l<=_B_.rmf; l++) par.m2d[l].ts[y] = par.yearmin+y+1;

    // set scale factor
    par.m2d[_B_.dem].scale[y] = 1;    par.m2d[_B_.dss].scale[y] = 1;
    par.m2d[_B_.dri].scale[y] = 1;    par.m2d[_B_.dps].scale[y] = 1;
    par.m2d[_B_.dfi].scale[y] = 1;    par.m2d[_B_.des].scale[y] = 1;
    par.m2d[_B_.dlm].scale[y] = 1;    par.m2d[_B_.lts].scale[y] = 1;
    par.m2d[_B_.lgs].scale[y] = 1;    par.m2d[_B_.vem].scale[y] = _RS_;
    par.m2d[_B_.vss].scale[y] = _RS_; par.m2d[_B_.vri].scale[y] = _RS_;
    par.m2d[_B_.vps].scale[y] = _RS_; par.m2d[_B_.vfi].scale[y] = _RS_;
    par.m2d[_B_.ves].scale[y] = _RS_; par.m2d[_B_.vlm].scale[y] = _RS_;
    par.m2d[_B_.vbl].scale[y] = _RS_; par.m2d[_B_.vsa].scale[y] = _RS_;
    par.m2d[_B_.ist].scale[y] = 10;   par.m2d[_B_.ibl].scale[y] = 10;
    par.m2d[_B_.ibt].scale[y] = 10;   par.m2d[_B_.igs].scale[y] = 10;
    par.m2d[_B_.rar].scale[y] = _RS_; par.m2d[_B_.raf].scale[y] = _RS_;
    par.m2d[_B_.rmr].scale[y] = _RS_; par.m2d[_B_.rmf].scale[y] = _RS_;

    // set band names
    sprintf(par.m2d[_B_.dem].bn[y], "%s: Date of Early Minimum",       par.index_c);
    sprintf(par.m2d[_B_.dss].bn[y], "%s: Date of Start of Season",     par.index_c);
    sprintf(par.m2d[_B_.dri].bn[y], "%s: Date of Rising Inflection",   par.index_c);
    sprintf(par.m2d[_B_.dps].bn[y], "%s: Date of Peak of Season",      par.index_c);
    sprintf(par.m2d[_B_.dfi].bn[y], "%s: Date of Falling Inflection",  par.index_c);
    sprintf(par.m2d[_B_.des].bn[y], "%s: Date of End of Season",       par.index_c);
    sprintf(par.m2d[_B_.dlm].bn[y], "%s: Date of Late Minimum",        par.index_c);
    sprintf(par.m2d[_B_.lts].bn[y], "%s: Length of Total Season",      par.index_c);
    sprintf(par.m2d[_B_.lgs].bn[y], "%s: Length of Green Season",      par.index_c);
    sprintf(par.m2d[_B_.vem].bn[y], "%s: Value of Early Minimum",      par.index_c);
    sprintf(par.m2d[_B_.vss].bn[y], "%s: Value of Start of Season",    par.index_c);
    sprintf(par.m2d[_B_.vri].bn[y], "%s: Value of Rising Inflection",  par.index_c);
    sprintf(par.m2d[_B_.vps].bn[y], "%s: Value of Peak of Season",     par.index_c);
    sprintf(par.m2d[_B_.vfi].bn[y], "%s: Value of Falling Inflection", par.index_c);
    sprintf(par.m2d[_B_.ves].bn[y], "%s: Value of End of Season",      par.index_c);
    sprintf(par.m2d[_B_.vlm].bn[y], "%s: Value of Late Minimum",       par.index_c);
    sprintf(par.m2d[_B_.vbl].bn[y], "%s: Value of Base Level",         par.index_c);
    sprintf(par.m2d[_B_.vsa].bn[y], "%s: Value of Seasonal Amplitude", par.index_c);
    sprintf(par.m2d[_B_.ist].bn[y], "%s: Integral of Total Season",    par.index_c);
    sprintf(par.m2d[_B_.ibl].bn[y], "%s: Integral of Base Level",      par.index_c);
    sprintf(par.m2d[_B_.ibt].bn[y], "%s: Integral of Base+Total",      par.index_c);
    sprintf(par.m2d[_B_.igs].bn[y], "%s: Integral of Green Season",    par.index_c);
    sprintf(par.m2d[_B_.rar].bn[y], "%s: Rate of Average Rising",      par.index_c);
    sprintf(par.m2d[_B_.raf].bn[y], "%s: Rate of Average Falling",     par.index_c);
    sprintf(par.m2d[_B_.rmr].bn[y], "%s: Rate of Maximum Rising",      par.index_c);
    sprintf(par.m2d[_B_.rmf].bn[y], "%s: Rate of Maximum Falling",     par.index_c);

  }


  #ifdef SPLITS

  #pragma omp parallel private(y,l) shared(MASK,_NC_,_B_,par) default(none)
  {

    #pragma omp for schedule(guided)
    for (p=0; p<_NC_; p++){

      if (MASK != NULL && !MASK[p]){
        for (l=_B_.dem; l<=_B_.rmf; l++){
          for (y=0; y<par.m2d[l].length; y++) par.m2d[l].data[y][p] = _TSAOFF_;
        }
        continue;
      }

      for (y=par.yearmin+1; y<par.yearmax; y++) derive_lsp(p, y);

    }

  }

  #endif

 
  #ifdef CLOCK
  proctime_print("phenology from time series", TIME);
  #endif

  return SUCCESS;
}


/** Basic temporal statistics
+++ This function computes basic temporal statistics for each pixel. Cur-
+++ rently implemented are mean, sd, min, max, # of obs.
--- MASK:   Analysis mask
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int basic_timeseries_stats(bool *MASK){
double avg, var, k;
int tss, min, max;
int t, p, b, nb;


  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  #ifdef DEBUG
  printf("Basic time series stats. "); fflush(stdout);
  #endif
  
  nb = par.m2d[_B_.sta].length;

  // allocate statistics
  alloc_2D((void***)&par.m2d[_B_.sta].data, nb, _NC_, sizeof(int2s));
  alloc((void**)&par.m2d[_B_.sta].scale, nb, sizeof(int2s));
  alloc_2D((void***)&par.m2d[_B_.sta].bn, nb, MAX_STR_LEN, sizeof(char));
  
  // scale factors
  par.m2d[_B_.sta].scale[0] = _RS_;
  par.m2d[_B_.sta].scale[1] = _RS_;
  par.m2d[_B_.sta].scale[2] = _RS_;
  par.m2d[_B_.sta].scale[3] = _RS_;
  par.m2d[_B_.sta].scale[4] = 1;

  // band names
  sprintf(par.m2d[_B_.sta].bn[0], "Average");
  sprintf(par.m2d[_B_.sta].bn[1], "Std. dev.");
  sprintf(par.m2d[_B_.sta].bn[2], "Minimum");
  sprintf(par.m2d[_B_.sta].bn[3], "Maximum");
  sprintf(par.m2d[_B_.sta].bn[4], "# of obs.");

  #pragma omp parallel private(t,b,tss,avg,var,k,min,max) shared(MASK,nb,dir,_NC_,_B_,par) default(none)
  {

    #pragma omp for
    for (p=0; p<_NC_; p++){
      
      if (MASK != NULL && !MASK[p]){
        for (b=0; b<nb; b++) par.m2d[_B_.sta].data[b][p] = _TSAOFF_;
        continue;
      }

      avg = var = k = 0.0;
      min = INT_MAX;
      max = INT_MIN;

      // compute stats
      for (t=0; t<dir.n; t++){
        
        if ((tss = par.m2d[_B_.tss].data[t][p]) == _TSAOFF_) continue;
        
        k++;
        if (k == 1){
          avg = tss/_RS_;
        } else {
          var_recurrence(tss/_RS_, &avg, &var, k);
        }

        if (tss < min) min = tss;
        if (tss > max) max = tss;

      }

      // copy to image
      if (k > 0){
        par.m2d[_B_.sta].data[0][p] = (int2s)(avg*_RS_);
      } else {
        par.m2d[_B_.sta].data[0][p] = _TSAOFF_;
      }
      if (k > 1){
        par.m2d[_B_.sta].data[1][p] = (int2s)(standdev(var, k)*_RS_);
      } else {
        par.m2d[_B_.sta].data[1][p] = _TSAOFF_;
      }
      if (min < INT_MAX){
        par.m2d[_B_.sta].data[2][p] = (int2s)min;
      } else {
        par.m2d[_B_.sta].data[2][p] = _TSAOFF_;
      }
      if (max > INT_MIN){
        par.m2d[_B_.sta].data[3][p] = (int2s)max;
      } else {
        par.m2d[_B_.sta].data[3][p] = _TSAOFF_;
      }
      par.m2d[_B_.sta].data[4][p] = (int2s)k;

    }

  }

  #ifdef CLOCK
  proctime_print("computing basic stats", TIME);
  #endif

  return SUCCESS;
}


/** Standardize time series
+++ This function centers or standardizes any time series to the pixel's
+++ mean (and sd). Note that the result is limited to values between 
+++ -32767 & 32767 (scale 10000).
--- TS:     time series (modified)
--- MASK:   Analysis mask
--- n:      length of time series
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int standardize_timeseries(int2s **TS, bool *MASK, int n){
double avg, var, std, k, tmp;
int t, p;


  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  #ifdef DEBUG
  printf("Standardize time series. "); fflush(stdout);
  #endif


  #pragma omp parallel private(t,avg,var,k,std,tmp) shared(MASK,TS,n,_NC_,par) default(none)
  {

    #pragma omp for
    for (p=0; p<_NC_; p++){

      if (MASK != NULL && !MASK[p]) continue;

      avg = var = k = 0.0;

      // compute stats
      for (t=0; t<n; t++){
        
        if (TS[t][p] == _TSAOFF_) continue;
        
        k++;
        if (k == 1){
          avg = TS[t][p]/_RS_;
        } else {
          var_recurrence(TS[t][p]/_RS_, &avg, &var, k);
        }

      }

      if (k < 2) continue;


      // set center or standardize option
      if (par.dostandard){
        std = standdev(var, k);
      } else {
        std = 1.0;
      }


      // center or standardize time series
      for (t=0; t<n; t++){
        
        if (TS[t][p] == _TSAOFF_) continue;

        tmp = (TS[t][p]/_RS_ - avg) / std * _RS_;
        if (tmp > INT_MAX) tmp = INT_MAX;
        if (tmp < INT_MIN) tmp = INT_MIN;
        TS[t][p] = (int2s)tmp;;

      }

    }

  }

  #ifdef CLOCK
  proctime_print("standardizing time series", TIME);
  #endif

  return SUCCESS;
}


/** Trend analysis of time series
+++ This function computes a trend analysis for any time series. Currently
+++ implemented trend parameters are mean, intercept, slope, R-squared, 
+++ significance of slope, RMSE, MAE, max. absolute residual and # of obs.
--- TS:        time series (modified)
--- timestamp: timestamp of the time series
--- MASK:      Analysis mask
--- n:         length of time series
--- days:      are the values given in continuous days?
+++ Return:    SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int trend_timeseries(int2s **TS, float *timestamp, bool *MASK, int n, bool days){
double mx, my, vx, vy, cv, k;
double off, slp, rsq, e, ssqe, sae, sxsq, maxe, seb, yhat;
int t, p, b, nb;


  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  #ifdef DEBUG
  printf("Trend of time series. "); fflush(stdout);
  #endif
  
  nb = par.m2d[_B_.trd].length;

  // allocate trend parameters
  alloc_2D((void***)&par.m2d[_B_.trd].data, nb, _NC_, sizeof(int2s));
  alloc((void**)&par.m2d[_B_.trd].scale, nb, sizeof(int2s));
  alloc_2D((void***)&par.m2d[_B_.trd].bn, nb, MAX_STR_LEN, sizeof(char));
  
  // scale factors
  par.m2d[_B_.trd].scale[0] = _RS_; par.m2d[_B_.trd].scale[1] = _RS_;
  par.m2d[_B_.trd].scale[2] = _RS_; par.m2d[_B_.trd].scale[3] = _RS_;
  par.m2d[_B_.trd].scale[4] = 1;    par.m2d[_B_.trd].scale[5] = _RS_;
  par.m2d[_B_.trd].scale[6] = _RS_; par.m2d[_B_.trd].scale[7] = _RS_;
  par.m2d[_B_.trd].scale[8] = 1;

  // band names
  sprintf(par.m2d[_B_.trd].bn[0], "Average");
  sprintf(par.m2d[_B_.trd].bn[1], "Intercept");
  sprintf(par.m2d[_B_.trd].bn[2], "Trend");
  sprintf(par.m2d[_B_.trd].bn[3], "R-squared");
  sprintf(par.m2d[_B_.trd].bn[4], "Significance");
  sprintf(par.m2d[_B_.trd].bn[5], "RMSE");
  sprintf(par.m2d[_B_.trd].bn[6], "MAE");
  sprintf(par.m2d[_B_.trd].bn[7], "Max. abs. E");
  sprintf(par.m2d[_B_.trd].bn[8], "# of obs.");

  #pragma omp parallel private(b,t,mx,my,vx,vy,cv,k,slp,off,rsq,ssqe,sae,sxsq,maxe,yhat,e,seb) shared(MASK,nb,TS,timestamp,n,_NC_,_B_,par,days) default(none)
  {

    #pragma omp for
    for (p=0; p<_NC_; p++){
      
      if (MASK != NULL && !MASK[p]){
        for (b=0; b<nb; b++) par.m2d[_B_.trd].data[b][p] = _TSAOFF_;
        continue;
      }

      mx = my = vx = vy = cv = k = 0.0;

      // compute stats
      for (t=0; t<n; t++){

        if (TS[t][p] == _TSAOFF_) continue;

        k++;
        if (k == 1){
          mx = timestamp[t]-timestamp[0];
          my = TS[t][p]/_RS_;
        } else {
          covar_recurrence(timestamp[t]-timestamp[0], TS[t][p]/_RS_, &mx, &my, &vx, &vy, &cv, k);
        }

      }

      if (k < 3){
        for (b=0; b<nb; b++) par.m2d[_B_.trd].data[b][p] = _TSAOFF_;
        continue;
      }


      // compute trend coefficients and R-squared
      cv = covariance(cv, k);
      vx = variance(vx, k);
      vy = variance(vy, k);
      linreg_coefs(mx, my, cv, vx, &slp, &off);
      linreg_rsquared(cv, vx, vy, &rsq);


      // compute residual metrics
      ssqe = sae = sxsq = 0; maxe = INT_MIN;

      for (t=0; t<n; t++){

        if (TS[t][p] == _TSAOFF_) continue;

        linreg_predict(timestamp[t]-timestamp[0], slp, off, &yhat);
        e = (TS[t][p]/_RS_-yhat);
        ssqe += e*e;
        sae += fabs(e);
        sxsq += (timestamp[t]-timestamp[0]-mx)*(timestamp[t]-timestamp[0]-mx);
        if (fabs(e) > maxe) maxe = fabs(e);

      }

      // account for values given in continuous days
      if (days){ my  -= 365*(n-1)/(2*_RS_); slp -= 365/_RS_;}

      // mean
      par.m2d[_B_.trd].data[0][p] = (int2s)(my*_RS_);

      // intercept
      par.m2d[_B_.trd].data[1][p] = (int2s)(off*_RS_);

      // slope
      par.m2d[_B_.trd].data[2][p] = (int2s)(slp*_RS_);

      // R-squared
      par.m2d[_B_.trd].data[3][p] = (int2s)(rsq*_RS_);

      // standard error of slope + significance of slope
      seb = sqrt(1.0/(k-2)*ssqe)/sqrt(sxsq);
      par.m2d[_B_.trd].data[4][p] = (int2s)slope_significant(1-par.conf, par.tail, k, slp, 0.0, seb);

      // RMSE
      par.m2d[_B_.trd].data[5][p] = (int2s)(sqrt(ssqe/k)*_RS_);

      // MAE
      par.m2d[_B_.trd].data[6][p] = (int2s)(sae/k*_RS_);

      // max. absolute residual
      par.m2d[_B_.trd].data[7][p] = (int2s)(maxe*_RS_);

      // nobs
      par.m2d[_B_.trd].data[8][p] = k;

    }

  }

  #ifdef CLOCK
  proctime_print("trending time series", TIME);
  #endif

  return SUCCESS;
}


/** Change, Aftereffect and Trend analysis of time series
+++ This function computes a CAT analysis for any time series. Each time 
+++ series is split into three parts; the splitting point is the largest 
+++ change. The largest change and its timestamp, as well as trend para-
+++ meters for the three parts are computed. Currently implemented trend 
+++ parameters are mean, intercept, slope, R-squared, significance of slo-
+++ pe, RMSE, MAE, max. absolute residual and # of obs.
--- TS:        time series (modified)
--- timestamp: timestamp of the time series
--- MASK:      Analysis mask
--- n:         length of time series
--- days:      are the values given in continuous days?
+++ Return:    SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int cat_timeseries(int2s **TS, float *timestamp, bool *MASK, int n, bool days){
double mx, my, vx, vy, cv, k;
double off, slp, rsq, e, ssqe, sae, sxsq, maxe, seb, yhat;
int p, b, nb, part;
int t, t_pre, t_change, t_min[3], t_max[3];

float change;


  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  #ifdef DEBUG
  printf("Trend of time series. "); fflush(stdout);
  #endif

  nb = par.m2d[_B_.cat].length;

  // allocate cat parameters
  alloc_2D((void***)&par.m2d[_B_.cat].data, nb, _NC_, sizeof(int2s));
  alloc((void**)&par.m2d[_B_.cat].scale, nb, sizeof(int2s));
  alloc_2D((void***)&par.m2d[_B_.cat].bn, nb, MAX_STR_LEN, sizeof(char));
  
  // scale factors
  par.m2d[_B_.cat].scale[0] = _RS_;  par.m2d[_B_.cat].scale[1] = 1;
  par.m2d[_B_.cat].scale[2] = _RS_;  par.m2d[_B_.cat].scale[3] = _RS_;
  par.m2d[_B_.cat].scale[4] = _RS_;  par.m2d[_B_.cat].scale[5] = _RS_;
  par.m2d[_B_.cat].scale[6] = 1;     par.m2d[_B_.cat].scale[7] = _RS_;
  par.m2d[_B_.cat].scale[8] = _RS_;  par.m2d[_B_.cat].scale[9] = _RS_;
  par.m2d[_B_.cat].scale[10] = 1;   
  par.m2d[_B_.cat].scale[11] = _RS_; par.m2d[_B_.cat].scale[12] = _RS_;
  par.m2d[_B_.cat].scale[13] = _RS_; par.m2d[_B_.cat].scale[14] = _RS_;
  par.m2d[_B_.cat].scale[15] = 1;    par.m2d[_B_.cat].scale[16] = _RS_;
  par.m2d[_B_.cat].scale[17] = _RS_; par.m2d[_B_.cat].scale[18] = _RS_;
  par.m2d[_B_.cat].scale[19] = 1;
  par.m2d[_B_.cat].scale[20] = _RS_; par.m2d[_B_.cat].scale[21] = _RS_;
  par.m2d[_B_.cat].scale[22] = _RS_; par.m2d[_B_.cat].scale[23] = _RS_;
  par.m2d[_B_.cat].scale[24] = 1;    par.m2d[_B_.cat].scale[25] = _RS_;
  par.m2d[_B_.cat].scale[26] = _RS_; par.m2d[_B_.cat].scale[27] = _RS_;
  par.m2d[_B_.cat].scale[28] = 1;

  // band names
  sprintf(par.m2d[_B_.cat].bn[0], "Change");
  sprintf(par.m2d[_B_.cat].bn[1], "Time of change");
  sprintf(par.m2d[_B_.cat].bn[2], "Complete: Average");
  sprintf(par.m2d[_B_.cat].bn[3], "Complete: Intercept");
  sprintf(par.m2d[_B_.cat].bn[4], "Complete: Trend");
  sprintf(par.m2d[_B_.cat].bn[5], "Complete: R-squared");
  sprintf(par.m2d[_B_.cat].bn[6], "Complete: Significance");
  sprintf(par.m2d[_B_.cat].bn[7], "Complete: RMSE");
  sprintf(par.m2d[_B_.cat].bn[8], "Complete: MAE");
  sprintf(par.m2d[_B_.cat].bn[9], "Complete: Max. abs. E");
  sprintf(par.m2d[_B_.cat].bn[10], "Complete: # of obs.");
  sprintf(par.m2d[_B_.cat].bn[11], "Before: Average");
  sprintf(par.m2d[_B_.cat].bn[12], "Before: Intercept");
  sprintf(par.m2d[_B_.cat].bn[13], "Before: Trend");
  sprintf(par.m2d[_B_.cat].bn[14], "Before: R-squared");
  sprintf(par.m2d[_B_.cat].bn[15], "Before: Significance");
  sprintf(par.m2d[_B_.cat].bn[16], "Before: RMSE");
  sprintf(par.m2d[_B_.cat].bn[17], "Before: MAE");
  sprintf(par.m2d[_B_.cat].bn[18], "Before: Max. abs. E");
  sprintf(par.m2d[_B_.cat].bn[19], "Before: # of obs.");
  sprintf(par.m2d[_B_.cat].bn[20], "After: Average");
  sprintf(par.m2d[_B_.cat].bn[21], "After: Intercept");
  sprintf(par.m2d[_B_.cat].bn[22], "After: Trend");
  sprintf(par.m2d[_B_.cat].bn[23], "After: R-squared");
  sprintf(par.m2d[_B_.cat].bn[24], "After: Significance");
  sprintf(par.m2d[_B_.cat].bn[25], "After: RMSE");
  sprintf(par.m2d[_B_.cat].bn[26], "After: MAE");
  sprintf(par.m2d[_B_.cat].bn[27], "After: Max. abs. E");
  sprintf(par.m2d[_B_.cat].bn[28], "After: # of obs.");


  #pragma omp parallel private(b,part,t,t_pre,t_change,t_min,t_max,change,mx,my,vx,vy,cv,k,slp,off,rsq,ssqe,sae,sxsq,maxe,yhat,e,seb) shared(MASK,nb,TS,timestamp,n,_NC_,_B_,par,days) default(none)
  {

    #pragma omp for
    for (p=0; p<_NC_; p++){
      
      if (MASK != NULL && !MASK[p]){
        for (b=0; b<nb; b++) par.m2d[_B_.cat].data[b][p] = _TSAOFF_;
        continue;
      }

      change = INT_MIN;
      b = 0;

      // find largest change
      for (t=1; t<n; t++){

        if (TS[t][p] == _TSAOFF_) continue;

        t_pre = t-1;
        while (t_pre >= 0 && TS[t_pre][p] == _TSAOFF_) t_pre--;

        if (t_pre < 0 || TS[t_pre][p] == _TSAOFF_) continue;

        if (TS[t_pre][p]-TS[t][p] > change){
          change = TS[t_pre][p]-TS[t][p];
          t_change = t;
        }

      }

      if (change == INT_MIN){
        for (b=0; b<nb; b++) par.m2d[_B_.cat].data[b][p] = _TSAOFF_;
        continue;
      }

      // largest change + timestamp
      par.m2d[_B_.cat].data[b++][p] = (int2s)change;
      par.m2d[_B_.cat].data[b++][p] = (int2s)timestamp[t_change];

   
      // split the time series into three parts
      t_min[0] = 0;        t_max[0] = n;          // complete
      t_min[1] = 0;        t_max[1] = t_change-1; // before
      t_min[2] = t_change; t_max[2] = n;          // after

      // compute trend parameters for each part
      for (part=0; part<3; part++){

        mx = my = vx = vy = cv = k = 0.0;

        // compute stats
        for (t=t_min[part]; t<t_max[part]; t++){

          if (TS[t][p] == _TSAOFF_) continue;

          k++;
          if (k == 1){
            mx = timestamp[t]-timestamp[0];
            my = TS[t][p]/_RS_;
          } else {
            covar_recurrence(timestamp[t]-timestamp[0], TS[t][p]/_RS_, &mx, &my, &vx, &vy, &cv, k);
          }

        }

        if (k < 3){

          // mean
          par.m2d[_B_.cat].data[b++][p] = _TSAOFF_;
          // intercept
          par.m2d[_B_.cat].data[b++][p] = _TSAOFF_;
          // slope
          par.m2d[_B_.cat].data[b++][p] = _TSAOFF_;
          // R-squared
          par.m2d[_B_.cat].data[b++][p] = _TSAOFF_;
          // standard error of slope + significance of slope
          par.m2d[_B_.cat].data[b++][p] = _TSAOFF_;
          // RMSE
          par.m2d[_B_.cat].data[b++][p] = _TSAOFF_;
          // MAE
          par.m2d[_B_.cat].data[b++][p] = _TSAOFF_;
          // max. absolute residual
          par.m2d[_B_.cat].data[b++][p] = _TSAOFF_;
          // nobs
          par.m2d[_B_.cat].data[b++][p] = k;

          continue;
        }


        // compute trend coefficients and R-squared
        cv = covariance(cv, k);
        vx = variance(vx, k);
        vy = variance(vy, k);
        linreg_coefs(mx, my, cv, vx, &slp, &off);
        linreg_rsquared(cv, vx, vy, &rsq);


        // compute residual metrics
        ssqe = sae = sxsq = 0; maxe = INT_MIN;

        for (t=t_min[part]; t<t_max[part]; t++){

          if (TS[t][p] == _TSAOFF_) continue;

          linreg_predict(timestamp[t]-timestamp[0], slp, off, &yhat);
          e = (TS[t][p]/_RS_-yhat);
          ssqe += e*e;
          sae += fabs(e);
          sxsq += (timestamp[t]-timestamp[0]-mx)*(timestamp[t]-timestamp[0]-mx);
          if (fabs(e) > maxe) maxe = fabs(e);

        }

        // account for values given in continuous days
        if (days){ my  -= 365*(n-1)/(2*_RS_); slp -= 365/_RS_;}

        // mean
        par.m2d[_B_.cat].data[b++][p] = (int2s)(my*_RS_);

        // intercept
        par.m2d[_B_.cat].data[b++][p] = (int2s)(off*_RS_);

        // slope
        par.m2d[_B_.cat].data[b++][p] = (int2s)(slp*_RS_);

        // R-squared
        par.m2d[_B_.cat].data[b++][p] = (int2s)(rsq*_RS_);

        // standard error of slope + significance of slope
        seb = sqrt(1.0/(k-2)*ssqe)/sqrt(sxsq);
        par.m2d[_B_.cat].data[b++][p] = (int2s)slope_significant(1-par.conf, par.tail, k, slp, 0.0, seb);

        // RMSE
        par.m2d[_B_.cat].data[b++][p] = (int2s)(sqrt(ssqe/k)*_RS_);

        // MAE
        par.m2d[_B_.cat].data[b++][p] = (int2s)(sae/k*_RS_);

        // max. absolute residual
        par.m2d[_B_.cat].data[b++][p] = (int2s)(maxe*_RS_);

        // nobs
        par.m2d[_B_.cat].data[b++][p] = k;

      }

    }

  }

  #ifdef CLOCK
  proctime_print("trending time series", TIME);
  #endif

  return SUCCESS;
}


/** Fold time series
+++ This aggregation function folds the full time series by year, month, 
+++ week, or day to create a weekly time series. Folding can be done using
+++ mean/min/max statistics.
--- MASK:   Analysis mask
--- by:     Folding type: year (0), month (1), week (2), day (3)
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int fold_timeseries(bool *MASK, int by){
int t, f, fmin, n, p, l;
int tss, min, max;
double sum, num;
char bn_[MAX_STR_LEN];


  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  #ifdef DEBUG
  printf("Fold time series. "); fflush(stdout);
  #endif

  switch (by){
    case 0:
      l = _B_.fby;
      fmin = par.yearmin;
      strcpy(bn_, "Yearly");
      break;
    case 1:
      l = _B_.fbm;
      fmin = par.monthmin;
      strcpy(bn_, "Monthly");
      break;
    case 2:
      l = _B_.fbw;
      fmin = par.weekmin;
      strcpy(bn_, "Weekly");
      break;
    case 3:
      l = _B_.fbd;
      fmin = par.doymin;
      strcpy(bn_, "Daily");
      break;
  }

  // number of bins
  n = par.m2d[l].length;

  // allocate folded time series and timestamp
  alloc_2D((void***)&par.m2d[l].data, n, _NC_, sizeof(int2s));
  alloc((void**)&par.m2d[l].scale, n, sizeof(int2s));
  for (f=0; f<n; f++) par.m2d[l].scale[f] = _RS_;
  alloc((void**)&par.m2d[l].ts, n, sizeof(float));
  for (f=0; f<n; f++) par.m2d[l].ts[f] = fmin + f;
  alloc_2D((void***)&par.m2d[l].bn, n, MAX_STR_LEN, sizeof(char));
  for (f=0; f<n; f++) sprintf(par.m2d[l].bn[f], "%s %s", bn_, par.index_c);


  #pragma omp parallel private(t,tss,min,max,sum,num) shared(MASK,dir,n,_NC_,_B_,l,par,fmin,by) default(none)
  {

    #pragma omp for collapse(2) schedule(dynamic,1)
    for (f=0; f<n; f++){
    for (p=0; p<_NC_; p++){
      
      if (MASK != NULL && !MASK[p]){
        par.m2d[l].data[f][p] = _TSAOFF_;
        continue;
      }

      min = INT_MAX;
      max = INT_MIN;
      sum = 0.0;
      num = 0.0;

      // compute stats
      for (t=0; t<dir.n; t++){

        switch (by){
          case 0:
            if ((par.m2d[_B_.tss].date[t].year-fmin)  != f) continue;
            break;
          case 1:
            if ((par.m2d[_B_.tss].date[t].month-fmin) != f) continue;
            break;
          case 2:
            if ((par.m2d[_B_.tss].date[t].week-fmin)  != f) continue;
            break;
          case 3:
            if ((par.m2d[_B_.tss].date[t].doy-fmin)   != f) continue;
            break;
        }

        if ((tss = par.m2d[_B_.tss].data[t][p]) == _TSAOFF_) continue;

        sum += tss; num++;
        if (tss < min) min = tss;
        if (tss > max) max = tss;

      }

      // fold by mean (0), min (1), max (2)
      switch (par.fold){
        case 0:
          if (num > 0){
            par.m2d[l].data[f][p] = (int2s)(sum/num);
          } else {
            par.m2d[l].data[f][p] = _TSAOFF_;
          }
          break;
        case 1:
          if (min < INT_MAX){
            par.m2d[l].data[f][p] = (int2s)min;
          } else {
            par.m2d[l].data[f][p] = _TSAOFF_;
          }
          break;
        case 2:
          if (max > INT_MIN){
            par.m2d[l].data[f][p] = (int2s)max;
          } else {
            par.m2d[l].data[f][p] = _TSAOFF_;
          }
          break;
      }

    }
    }

  }


  #ifdef CLOCK
  proctime_print("folding time series", TIME);
  #endif

  return SUCCESS;
}


/** Clear Sky Observation statistics
+++ This function summarizes Level 2 data availability for given windows.
+++ This is the main functionality of the CSO module, and can count the 
+++ number of observations and computete statistics regarding differences
+++ between observations
--- L2:     Level-2 data
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int clear_sky_stats(level2_t *L2){
qai_t *QAI = NULL;
double d_ce;
double mx, vx, sx, kx, k, n;
double min, max;
float *arr = NULL;
int l, t, t0[MAX_WIN_LEN], t1[MAX_WIN_LEN], ce_left, w, p;
int2s q25, q75;


  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif


  for (l=0; l<_NP_; l++){

    if (!par.m2d[l].save) continue;
      
    alloc_2D((void***)&par.m2d[l].data,  par.m2d[l].length, _NC_, sizeof(int2s));
    alloc((void**)&par.m2d[l].ts, par.m2d[l].length, sizeof(float));
    alloc_2D((void***)&par.m2d[l].bn, par.m2d[l].length, MAX_STR_LEN, sizeof(char));
    alloc((void**)&par.m2d[l].scale, par.m2d[l].length, sizeof(int2s));

    // timestamps
    for (w=0; w<par.nwin; w++) par.m2d[l].ts[w] = par.ywin[w] + (par.mwin[w]-1)/12.0;

  }


  for (w=0; w<par.nwin; w++){

    // scale factors
    if (par.m2d[_B_.num].save) par.m2d[_B_.num].scale[w] = 1;
    if (par.m2d[_B_.avg].save) par.m2d[_B_.avg].scale[w] = 1;
    if (par.m2d[_B_.std].save) par.m2d[_B_.std].scale[w] = 1;
    if (par.m2d[_B_.min].save) par.m2d[_B_.min].scale[w] = 1;
    if (par.m2d[_B_.max].save) par.m2d[_B_.max].scale[w] = 1;
    if (par.m2d[_B_.rng].save) par.m2d[_B_.rng].scale[w] = 1;
    if (par.m2d[_B_.skw].save) par.m2d[_B_.skw].scale[w] = _RS_;
    if (par.m2d[_B_.krt].save) par.m2d[_B_.krt].scale[w] = 1000;
    if (par.m2d[_B_.q25].save) par.m2d[_B_.q25].scale[w] = 1;
    if (par.m2d[_B_.q50].save) par.m2d[_B_.q50].scale[w] = 1;
    if (par.m2d[_B_.q75].save) par.m2d[_B_.q75].scale[w] = 1;
    if (par.m2d[_B_.iqr].save) par.m2d[_B_.iqr].scale[w] = 1;

    // band names
    if (par.m2d[_B_.num].save) sprintf(par.m2d[_B_.num].bn[w], "Clear-sky N %03d", w+1);
    if (par.m2d[_B_.avg].save) sprintf(par.m2d[_B_.avg].bn[w], "Clear-sky average of dt %03d", w+1);
    if (par.m2d[_B_.std].save) sprintf(par.m2d[_B_.std].bn[w], "Clear-sky stand. dev. of dt %03d", w+1);
    if (par.m2d[_B_.min].save) sprintf(par.m2d[_B_.min].bn[w], "Clear-sky min. of dt %03d", w+1);
    if (par.m2d[_B_.max].save) sprintf(par.m2d[_B_.max].bn[w], "Clear-sky max. of dt %03d", w+1);
    if (par.m2d[_B_.rng].save) sprintf(par.m2d[_B_.rng].bn[w], "Clear-sky range of dt %03d", w+1);
    if (par.m2d[_B_.skw].save) sprintf(par.m2d[_B_.skw].bn[w], "Clear-sky skewness of dt %03d", w+1);
    if (par.m2d[_B_.krt].save) sprintf(par.m2d[_B_.krt].bn[w], "Clear-sky kurtosis of dt %03d", w+1);
    if (par.m2d[_B_.q25].save) sprintf(par.m2d[_B_.q25].bn[w], "Clear-sky 25%% quantile of dt %03d", w+1);
    if (par.m2d[_B_.q50].save) sprintf(par.m2d[_B_.q50].bn[w], "Clear-sky 50%% quantile of dt %03d", w+1);
    if (par.m2d[_B_.q75].save) sprintf(par.m2d[_B_.q75].bn[w], "Clear-sky 75%% quantile of dt %03d", w+1);
    if (par.m2d[_B_.iqr].save) sprintf(par.m2d[_B_.iqr].bn[w], "Clear-sky interquantile range of dt %03d", w+1);

  }



  // first and last t per window
  for (w=0; w<par.nwin; w++){
    t0[w] = t1[w] = -1;
    for (t=0; t<dir.n; t++){
      if (L2[t].date.ce >= par.cewin[w] && L2[t].date.ce < par.cewin[w+1]){
        if (t0[w] < 0) t0[w] = t;
        t1[w] = t;
      }
    }
  }

    // parse quality
    alloc((void**)&QAI, dir.n, sizeof(qai_t));

    for (t=0; t<dir.n; t++){
      alloc_qai(&QAI[t]);
      parse_qai(QAI[t], L2[t].qai);
    }


  #pragma omp parallel private(t,w,mx,vx,sx,kx,k,n,min,max,d_ce,ce_left,arr,q25,q75) shared(dir,_NC_,L2,QAI,_B_,t0,t1,par) default(none)
  {

    alloc((void**)&arr, dir.n+1, sizeof(float));

    #pragma omp for
    for (p=0; p<_NC_; p++){

      for (w=0; w<par.nwin; w++){

        mx = vx = sx = kx = k = n = 0.0;
        min = INT_MAX;
        max = INT_MIN;

        if (t0[w] > -1){

          ce_left = par.cewin[w];

          for (t=t0[w]; t<=t1[w]; t++){

            // check nodata
            if (!use_this_pixel(QAI[t], p)) continue;

            // if current date is larger than previous date (incl. left window boundary),
            // include dt in stats
            if (L2[t].date.ce > ce_left){
              d_ce = L2[t].date.ce - ce_left;
              arr[(int)k] = (float)d_ce;
              kurt_recurrence(d_ce, &mx, &vx, &sx, &kx, ++k);
              if (d_ce < min) min = d_ce;
              if (d_ce > max) max = d_ce;
              ce_left = L2[t].date.ce;
            }

            n++;

          }

          // if current date is smaller than right window boundary,
          // include dt in stats
          if (ce_left < par.cewin[w+1]-1){
            d_ce = (par.cewin[w+1]-1) - ce_left;
            arr[(int)k] = (float)d_ce;
            kurt_recurrence(d_ce, &mx, &vx, &sx, &kx, ++k);
            if (d_ce < min) min = d_ce;
            if (d_ce > max) max = d_ce;
          }

        } else {

          min = max = mx = par.cewin[w+1]-par.cewin[w];

        }

        // add to output layers
        if (par.m2d[_B_.num].save) par.m2d[_B_.num].data[w][p] = (int2s)n;
        if (par.m2d[_B_.avg].save) par.m2d[_B_.avg].data[w][p] = (int2s)mx;
        if (par.m2d[_B_.std].save){
          if (k > 1) par.m2d[_B_.std].data[w][p] = (int2s)standdev(vx, k);
        }
        if (par.m2d[_B_.min].save) par.m2d[_B_.min].data[w][p] = (int2s)min;
        if (par.m2d[_B_.max].save) par.m2d[_B_.max].data[w][p] = (int2s)max;
        if (par.m2d[_B_.rng].save) par.m2d[_B_.rng].data[w][p] = (int2s)(max-min);
        if (par.m2d[_B_.skw].save){
          if (k > 2) par.m2d[_B_.skw].data[w][p] = (int2s)(skewness(vx, sx, k)*_RS_);
        }
        if (par.m2d[_B_.krt].save){
          if (k > 3) par.m2d[_B_.krt].data[w][p] = (int2s)(kurtosis(vx, kx, k)*1000.0);
        }
        if (par.m2d[_B_.q25].save){
          if (k > 0) par.m2d[_B_.q25].data[w][p] = (int2s)quantile(arr, k, 0.25); else par.m2d[_B_.q25].data[w][p] = (int2s)mx;
        }
        if (par.m2d[_B_.q50].save){
          if (k > 0) par.m2d[_B_.q50].data[w][p] = (int2s)quantile(arr, k, 0.50); else par.m2d[_B_.q50].data[w][p] = (int2s)mx;
        }
        if (par.m2d[_B_.q75].save){
          if (k > 0) par.m2d[_B_.q75].data[w][p] = (int2s)quantile(arr, k, 0.75); else par.m2d[_B_.q75].data[w][p] = (int2s)mx;
        }
        if (par.m2d[_B_.iqr].save){
          if (par.m2d[_B_.q25].save){
            q25 = par.m2d[_B_.q25].data[w][p];
          } else {
            if (k > 0) q25 = (int2s)quantile(arr, k, 0.25); else q25 = (int2s)mx;
          }
          if (par.m2d[_B_.q75].save){
            par.m2d[_B_.q75].data[w][p];
          } else {
            if (k > 0) q75 = (int2s)quantile(arr, k, 0.75); else q75 = (int2s)mx;
          }
          par.m2d[_B_.iqr].data[w][p] = q75 - q25;
        }

      }

    }
    
    free((void*)arr);

  }

  // clean
  for (t=0; t<dir.n; t++) free_qai(QAI[t]); free((void*)QAI);

  #ifdef CLOCK
  proctime_print("computing clear sky stats", TIME);
  #endif

  return SUCCESS;
}

