/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
This file contains functions for handling topographic effects
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#include "force.h"


/** Compile topography
+++ This function compiles the basic set of topographic derivatives. This 
+++ includes reprojecting the DEM to the image extent/projection and com-
+++ puting slope and aspect
--- TOP:    Topographic Derivatives (returned)
--- QAI:    Quality Assurance Information
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int compile_topo(top_t *TOP, qai_t QAI, int2u *GRD){
float *DEM = NULL;


  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  TOP->dem = NULL; TOP->slp = NULL; TOP->asp = NULL; 
  TOP->sky = NULL; TOP->ill = NULL; TOP->cor = NULL;

  // reproject topography
  if ((reproj_topo(1, &DEM)) != SUCCESS){
    printf("Reprojecting of DEM failed! "); return FAILURE;}

  // detect oceans and set to 0m a.s.l
  if ((ocean_topo(DEM)) != SUCCESS){
    printf("Compiling ocean DEM failed! "); return FAILURE;}

  // smooth DEM
  if ((smooth_topo(&DEM)) != SUCCESS){
    printf("Smoothing of DEM failed! "); return FAILURE;}

  // slope and aspect
  if ((slope_and_aspect(QAI, DEM, &TOP->slp, &TOP->asp, -32767)) != SUCCESS){
    printf("Slope/aspect failed! "); return FAILURE;}

  // calculate stats
  if ((elevation_stats(DEM, QAI, &TOP->dem)) != SUCCESS){
    printf("Elevation statistics failed! "); return FAILURE;}

  // illumination angle and sky view factor
  if (illumination(TOP, QAI, GRD) != SUCCESS){
    printf("error in topographic correction. "); return FAILURE;}

  #ifdef DEBUG
  db_write_image_1D((void**)QAI.off,  "QAI-OFF",   _NX_, _NY_, sizeof(bool),   1);
  db_write_image_1D((void**)QAI.slp,  "QAI-SLP",   _NX_, _NY_, sizeof(bool),   1);
  db_write_image_1D((void*)QAI.ill,   "QAI-ILL",   _NX_, _NY_, sizeof(bool),   1);
  db_write_image_1D((void**)DEM,      "TOP-DEM-F", _NX_, _NY_, sizeof(float),  4);
  db_write_image_1D((void**)TOP->dem, "TOP-DEM-I", _NX_, _NY_, sizeof(int1u),  1);
  db_write_image_1D((void**)TOP->slp, "TOP-SLP",   _NX_, _NY_, sizeof(int2u), 12);
  db_write_image_1D((void**)TOP->asp, "TOP-ASP",   _NX_, _NY_, sizeof(int2u), 12);
  db_write_image_1D((void*)TOP->ill,  "TOP-ILL",   _NX_, _NY_, sizeof(int2s),  2);
  db_write_image_1D((void*)TOP->sky,  "TOP-SKY",   _NX_, _NY_, sizeof(int2u), 12);
  #endif

  free((void*)DEM);

  #ifdef CLOCK
  proctime_print("compiled topography", TIME);
  #endif

  return SUCCESS;
}


/** Detect Ocean in DEM 
+++ This function detects oceans and sets these pixels to 0m a.s.l. as 
+++ ocean is often masked out in DEMs (but is valid data)
--- DEM:    Digital Elevation Model
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int ocean_topo(float *DEM){
int i, j, ii, jj, p, np;

  for (j=0; j<_NX_; j++){
  for (i=0, ii=1; ii<_NY_; i++, ii++){
    p = i*_NX_+j; np = ii*_NX_+j;
    if (DEM[p] == 0 && DEM[np] == -32767) DEM[np] = 0;
  }
  }

  for (j=0; j<_NX_; j++){
  for (i=(_NY_-1), ii=(_NY_-2); ii>=0; i--, ii--){
    p = i*_NX_+j; np = ii*_NX_+j;
    if (DEM[p] == 0 && DEM[np] == -32767) DEM[np] = 0;
  }
  }

  for (i=0; i<_NY_; i++){
  for (j=0, jj=1; jj<_NX_; j++, jj++){
    p = i*_NX_+j; np = i*_NX_+jj;
    if (DEM[p] == 0 && DEM[np] == -32767) DEM[np] = 0;
  }
  }

  for (i=0; i<_NY_; i++){
  for (j=(_NX_-1), jj=(_NX_-2); jj>=0; j--, jj--){
    p = i*_NX_+j; np = i*_NX_+jj;
    if (DEM[p] == 0 && DEM[np] == -32767) DEM[np] = 0;
  }
  }

  return SUCCESS;
}


/** Smooth DEM 
+++ This function smooths the DEM with a lowpass filter as there are often
+++ undesired effects in high resolution DEMs
--- DEM:    Digital Elevation Model
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int smooth_topo(float **DEM){
int i, j, ii, jj, ni, nj, p, np, k;
float *buf;
double sum, num;

  if (meta.res >= 30) return SUCCESS;
  
  #ifdef DEV
  printf("smooth_topo should consider actual pixel size of DEM...\n");
  #endif

  k = 30/meta.res;

  alloc((void**)&buf, _NC_, sizeof(float));
  for (p=0; p<_NC_; p++) buf[p] = -32767;

  for (i=0, p=0; i<_NY_; i++){
  for (j=0; j<_NX_; j++, p++){

    if ((*DEM)[p] == -32767) continue;

    sum = num = 0;

    for (ii=-1*k; ii<=k; ii++){
    for (jj=-1*k; jj<=k; jj++){

      ni = i+ii; nj = j+jj;
      if (ni > _NY_-1 || ni < 0 || nj > _NX_-1 || nj < 0) continue;

      np = ni*_NX_+nj;
      if ((*DEM)[np] == -32767) continue;

      sum += (*DEM)[np];
      num++;

    }
    }

    if (num > 0) buf[p] = (float) (sum/num);

  }
  }

  free((void*)*DEM);
  *DEM=buf;

  return SUCCESS;
}


/** Compute slope and aspect
+++ This function computes slope and aspect with the Horn (1981) method
--- QAI:    Quality Assurance Information
--- DEM:    Digital Elevation Model
--- SLP:    slope  (returned)
--- ASP:    aspect (returned)
--- nodata: DEM nodata value
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int slope_and_aspect(qai_t QAI, float *DEM, int2u **SLP, int2u **ASP, float nodata){
int i, j, ii, jj, p;
float devx, devy;
float dem[3][3];
int2u *SLOPE = NULL;
int2u *ASPECT = NULL;
float slope, aspect;
bool valid;


  alloc((void**)&SLOPE,  _NC_, sizeof(int2u));
  alloc((void**)&ASPECT, _NC_, sizeof(int2u));

  for (i=1; i<(_NY_-1); i++){
  for (j=1; j<(_NX_-1); j++){

    p = i*_NX_+j;

    if (QAI.off[p]) continue;

    valid = true;

    for (ii=-1; ii<=1; ii++){
    for (jj=-1; jj<=1; jj++){
      if ((dem[ii+1][jj+1] = DEM[(i+ii)*_NX_+(j+jj)]) == nodata){
        valid = false;
      }
    }
    }


    if (valid){

      devx = ((dem[0][2]+2*dem[1][2]+dem[2][2]) - 
              (dem[0][0]+2*dem[1][0]+dem[2][0])) / (8*meta.res);
      devy = ((dem[2][0]+2*dem[2][1]+dem[2][2]) - 
              (dem[0][0]+2*dem[0][1]+dem[0][2])) / (8*meta.res);

      slope = rad2deg(atan(sqrt(devx*devx+devy*devy)));
      if (slope > 2) QAI.slp[p] = true;

      if (slope > 0){
        aspect = rad2deg(atan2(devy, -1*devx));
        if (aspect < 0){
          aspect = 90-aspect;
        } else if (aspect > 90){
          aspect = 360-aspect+90;
        } else {
          aspect = 90-aspect;
        }
      } else {
        slope = aspect = 0;
      }

      SLOPE[p]  = (int2u)(deg2rad(slope)*10000);
      ASPECT[p] = (int2u)(deg2rad(aspect)*10000);

    } else {

      QAI.off[p]  = true;

    }

  }
  }


  for (p=0; p<_NC_; p+=_NX_){

    if (QAI.off[p]) continue;
    
    if (QAI.off[p+1]){
      QAI.off[p] = true;
    } else {
      SLOPE[p]  = SLOPE[p+1];
      ASPECT[p] = ASPECT[p+1];
    }
  }

  for (p=_NX_-1; p<_NC_; p+=_NX_){

    if (QAI.off[p]) continue;

    if (QAI.off[p-1]){
      QAI.off[p] = true;
    } else {
      SLOPE[p]  = SLOPE[p-1];
      ASPECT[p] = ASPECT[p-1];
    }
  }

  for (p=0; p<_NX_; p++){

    if (QAI.off[p]) continue;

    if (QAI.off[p+_NX_]){
      QAI.off[p] = true;
    } else {
      SLOPE[p]  = SLOPE[p+_NX_];
      ASPECT[p] = ASPECT[p+_NX_];
    }
  }

  for (p=(_NY_-1)*_NX_; p<_NC_; p++){

    if (QAI.off[p]) continue;

    if (QAI.off[p-_NX_]){
      QAI.off[p] = true;
    } else {
      SLOPE[p]  = SLOPE[p-_NX_];
      ASPECT[p] = ASPECT[p-_NX_];
    }
  }


  *SLP = SLOPE;
  *ASP = ASPECT;

  return SUCCESS;
}


/** DEM statistics
+++ This function computes elevation statistics (in km) and computes the
+++ binned DEM (100m classes)
--- DEM:    Digital Elevation Model
--- QAI:    Quality Assurance Information
--- CDEM:   Binned Digital Elevation Model (returned)
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int elevation_stats(float *DEM, qai_t QAI, int1u **CDEM){
int p;
float dem;
double sum = 0, num = 0;
int1u *cdem;

  alloc((void**)&cdem, _NC_, sizeof(int1u));
  
  atc.dem.min = SHRT_MAX;
  atc.dem.max = SHRT_MIN;

  for (p=0; p<_NC_; p++){
    if (QAI.off[p]) continue;
    dem = DEM[p]/1000; // -> kilometer
    if (dem < atc.dem.min) atc.dem.min = dem;
    if (dem > atc.dem.max) atc.dem.max = dem;
    sum += dem; // -> kilometer
    num++;
  }

  atc.dem.avg = (float)(sum/num);
  atc.dem.max += 0.001; // add 1m

  atc.dem.cnum = 255;
  atc.dem.step = (atc.dem.max-atc.dem.min)/atc.dem.cnum;
  
  for (p=0; p<_NC_; p++){
    if (QAI.off[p]) continue;
    dem = DEM[p]/1000; // -> kilometer
    cdem[p] = floor((dem-atc.dem.min)/atc.dem.step);
  }

  #ifdef DEBUG
  printf("elevation stats:\n");
  printf(" avg: %+6.3f, min: %+6.3f, max: %+6.3f, step: %+6.3f, cnum: %d\n",
  atc.dem.avg, atc.dem.min, atc.dem.max, atc.dem.step, atc.dem.cnum);
  #endif

  if (atc.dem.min < -0.5 || atc.dem.max > 9){
    printf("DEM out of bounds: min %f max %f. ", 
    atc.dem.min, atc.dem.max); return FAILURE;
  }

  *CDEM = cdem;
  return SUCCESS;
}


/** Average elevation of coarse grid cell
+++ This function computes the average binned elevation in the given
+++ coarse grid cell
--- TOP:    Topographic Derivatives
--- QAI:    Quality Assurance Information
--- gres:   resolution of coarse grid
--- i:      image y-coordinate of upper-left corner of coarse grid cell
--- j:      image x-coordinate of upper-left corner of coarse grid cell
+++ Return: average binned elevation
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int average_elevation_cg(top_t TOP, qai_t QAI, int gres, int j, int i){
int p, ii, jj, ipos, jpos;
double sum = 0, num = 0;

  for (ii=0; ii<gres; ii++){
  for (jj=0; jj<gres; jj++){

    ipos = i+ii; jpos = j+jj;
    if (ipos > _NY_-1) continue;
    if (jpos > _NX_-1) continue;
    p = ipos*_NX_+jpos;
    if (QAI.off[p]) continue;

    sum += TOP.dem[p];
    num++;

  }
  }

  if (num>0){
    return round(sum/num);
  } else {
    return 0;
  }

}


/** Illumination angle and sky view factor
+++ This function computes the illumination angle and a simple sky view 
+++ factor
--- TOP:    Topographic Derivatives
--- QAI:    Quality Assurance Information
--- GRD:    Coarse resolution grid
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int illumination(top_t *TOP, qai_t QAI, int2u *GRD){
int p, g;
float slp, asp;

  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  alloc((void**)&TOP->ill, _NC_, sizeof(int2s));
  alloc((void**)&TOP->sky, _NC_, sizeof(int2u));

  //if (!par.doatmo) return SUCCESS;


  for (p=0; p<_NC_; p++){

    if (QAI.off[p]){
      TOP->ill[p] = -10000;
      continue;
    }

    g = GRD[p];

    slp = TOP->slp[p]/10000.0;
    asp = TOP->asp[p]/10000.0;

    // illumination angle
    if (TOP->slp[p] == 0){
      TOP->ill[p] = (int2s)(atc.cg[g].cosszen*10000);
    } else {
      TOP->ill[p] = (int2s)(illumin(atc.cg[g].cosszen, atc.cg[g].sinszen, 
                            cos(slp), sin(slp), 
                            atc.cg[g].sazi, asp)*10000);
    }

    // set illumination QAI
    if (TOP->ill[p] < 0){
      QAI.ill[p] = 3; // deep shadow
    } else if (TOP->ill[p] < 1736.482){
      QAI.ill[p] = 2; // poor
    } else if (TOP->ill[p] < 5735.764){
      QAI.ill[p] = 1; // moderate
    }

    // sky view factor (portion of the sky dome diffusing on to a tilted surface)
    TOP->sky[p] = (int2u)(10000 - (slp/M_PI)*10000);

  }


  #ifdef CLOCK
  proctime_print("illumination condition", TIME);
  #endif

  return SUCCESS;
}


/** Compute topographic correction factor
+++ This function computes the C-factor used for topographic correction. C
+++ is derived for the SWIR2 band and is then propagated through the spec-
+++ trum in the actual radiometric correction. C is derived for each pixel
+++ with sufficient illumination.
--- TOA:    TOA reflectance
--- QAI:    Quality Assurance Information
--- GRD:    Coarse resolution grid
--- TOP:    Topographic Derivatives
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int topographic_factor(int2s **TOA, qai_t QAI, int2u *GRD, top_t *TOP){
int i, j, p, ii, jj, ip, jp, np, g, z;
int2u s_min  = 350; // 2° slope
double mx, my;
double cov, varx, vary, num;
double offset, gain;
float c_, h0, f;
float *cor = NULL;
float tmp;
float *swir = NULL;
int k, nk = 0, *K = NULL;

  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  // just allocate and return
  if (!par.dotopo){
    alloc((void**)&TOP->cor, _NC_, sizeof(int2u));
    return SUCCESS;
  }

  /** kernel for sampling neighborhood **/
  alloc((void**)&K, 3000/meta.res*2+1, sizeof(int));

  K[nk] = 3000/meta.res; nk++;
  while (K[nk-1] > 100/meta.res){
    K[nk] = K[nk-1]/sqrt(2);
    nk++;
  }

  if (K[nk-1] != 0) nk++;

  for (k=nk-2; k>=0; k--){
    K[nk] = -1*K[k];
    nk++;
  }

  #ifdef DEBUG
  printf("sampling for estimating C:\n");
  for (k=0; k<nk; k++) printf("%d ", K[k]); printf("\n");
  #endif


  /** allocate memory **/
  alloc((void**)&cor,  _NC_, sizeof(float));
  alloc((void**)&swir, _NC_, sizeof(float));

  #ifdef DEBUG
  double *RSQ = NULL;  alloc((void**)&RSQ,  _NC_, sizeof(double));
  double *INTC = NULL; alloc((void**)&INTC, _NC_, sizeof(double));
  double *GAIN = NULL; alloc((void**)&GAIN, _NC_, sizeof(double));
  int *NUM = NULL;     alloc((void**)&NUM,  _NC_, sizeof(int));
  #endif


  /** compute SWIR index **/
  for (p=0; p<_NC_; p++){
    
    if (QAI.off[p]) continue;

    tmp = TOA[_B_.swir1][p]/_RS_+TOA[_B_.swir2][p]/_RS_;
    if (tmp == 0) swir[p] = 0; else swir[p] = (TOA[_B_.swir1][p]/_RS_-TOA[_B_.swir2][p]/_RS_)/tmp;

  }


  /** estimate C for every pixel **/
  for (i=0, p=0; i<_NY_; i++){
  for (j=0; j<_NX_; j++, p++){
    
    // only do for illuminated pixels
    if (QAI.off[p] || TOP->ill[p] < 0) continue;

    g = GRD[p];
    z = TOP->dem[p];

    // f-factor, h0-factor, theoretical C-factor
    f = f_factor(atc.cg[g].z[z].tss[_B_.swir2], atc.cg[g].z[z].tsd[_B_.swir2]);
    h0 = (M_PI+2*atc.cg[g].szen)/_M_PI2_;
    c_ = c_factor_com(h0, f, atc.cg[g].cosszen);

    // only do for sloped pixels > 2°
    if (TOP->slp[p] > s_min){

      num = mx = my = varx = vary = cov = 0.0;

      // sample neighborhood
      for (ii=0; ii<nk; ii++){
      for (jj=0; jj<nk; jj++){
        
        ip = i+K[ii]; jp = j+K[jj];
        if (ip < 0 || jp < 0 || ip > _NY_-1 || jp > _NX_-1) continue;
        np = ip*_NX_+jp;

        // only use illuminated and sloped pixels > 2°
        if (QAI.off[np] || TOP->ill[np] < 0 || TOP->slp[np] < s_min) continue;

        // only do for same land cover
        if (fabs(swir[p]-swir[np]) > 0.025) continue;

        num++;

        // linear regression
        if (num == 1){
          mx = TOP->ill[np]/10000.0; my = TOA[_B_.swir2][np]/_RS_;
        } else {
          covar_recurrence(TOP->ill[np]/10000.0, TOA[_B_.swir2][np]/_RS_,
          &mx, &my, &varx, &vary, &cov, num);
        }

      }
      }

      #ifdef DEBUG
      NUM[p]   = (int)num;
      #endif


      if (num > 2){

        // regression parameters + c-factor
        linreg_coefs(mx, my, covariance(cov, num), 
          variance(varx, num), &gain, &offset);

        #ifdef DEBUG
        linreg_rsquared(covariance(cov, num), 
         variance(varx, num), variance(vary, num), &RSQ[p]);
        #endif

        // if offset < path reflectance, 
        // use path reflectance as offset and recompute gain
        if (offset < atc.cg[g].z[z].rho_p[_B_.swir2]){
          offset = atc.cg[g].z[z].rho_p[_B_.swir2];
          gain = (my-offset)/mx;
          #ifdef DEBUG
          RSQ[p] = 0.0;
          NUM[p] = 0;
          #endif
        }

        #ifdef DEBUG
        INTC[p] = offset;
        GAIN[p] = gain;
        #endif

        //if (gain < 0) continue;

        // shield against extreme C-values
        if (offset < 10*gain){
          cor[p] = c_factor_emp(offset, gain);
        } else {
          cor[p] = 10.0;
          #ifdef DEBUG
          RSQ[p] = 0.0;
          NUM[p] = 0;
          #endif
        }

      }

    }

    // use computed C if estimated C < 0
    if (cor[p] < 0 || cor[p] < c_) cor[p] = c_;
    if (cor[p] > USHRT_MAX/10000.0) cor[p] = USHRT_MAX/10000.0;

  }
  }

  free((void*)swir);


  /** smooth C-factor with lowpass **/

  alloc((void**)&TOP->cor, _NC_, sizeof(int2u));

  for (i=0, p=0; i<_NY_; i++){
  for (j=0; j<_NX_; j++, p++){
    
    if (QAI.off[p] || TOP->ill[p] < 0) continue;

      num = mx = 0.0;

      for (ip=(i-1); ip<(i+1); ip++){
      for (jp=(j-1); jp<(j+1); jp++){

        if (ip < 0 || jp < 0 || ip > _NY_-1 || jp > _NX_-1) continue;
        np = ip*_NX_+jp;

        if (QAI.off[np] || TOP->ill[np] < 0) continue;

        mx += cor[np];
        num++;

      }
      }

      if (num > 0) TOP->cor[p] = (int2u)(mx/num*10000); else TOP->cor[p] = (int2u)(cor[p]*10000);

  }
  }


  #ifdef DEBUG
  db_write_image_1D((void**)TOP->cor, "TOP-CFACTOR",  _NX_, _NY_, sizeof(int2u), 2);
  db_write_image_1D((void**)RSQ, "TOP-C-RSQ",  _NX_, _NY_, sizeof(double), 5);
  db_write_image_1D((void**)NUM, "TOP-C-NUM",  _NX_, _NY_, sizeof(int), 3);
  db_write_image_1D((void**)GAIN, "TOP-C-SLP",  _NX_, _NY_, sizeof(double), 5);
  db_write_image_1D((void**)INTC, "TOP-C-OFF",  _NX_, _NY_, sizeof(double), 5);
  free((void*)NUM);
  free((void*)RSQ);
  #endif
  
  free((void*)cor);
  free((void*)K);

  #ifdef CLOCK
  proctime_print("topographic correction factors", TIME);
  #endif

  return SUCCESS;
}

