/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/** The RBF interpolation was adpoted from the python routine 
+++ 'RBF_fitting_large_scale.py', Copyright (C) 2018 Andreas Rabe
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
This file contains functions for time series interpolation
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#include "force.h"


/** Linear interpolation
+++ This function linearly interpolates a time series
--- MASK:   Analysis mask (optional)
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int interpolate_linear(bool *MASK){
int p, t, t_left, i;
float x_left, x_right, x;
float y_left, y_right, y;


  #pragma omp parallel private(i,t,t_left,x,y,x_left,x_right,y_left,y_right) shared(MASK,dir,_NC_,par,_B_) default(none)
  {

    #pragma omp for
    for (p=0; p<_NC_; p++){

      // if masked, use nodata
      if (MASK != NULL && !MASK[p]){
        for (i=0; i<par.m2d[_B_.tsi].length; i++) par.m2d[_B_.tsi].data[i][p] = _TSAOFF_;
        continue;
      }

      // interpolate for each equidistant timestep
      for (i=0, t_left=0; i<par.m2d[_B_.tsi].length; i++){

        // current time
        x = par.m2d[_B_.tsi].date[i].ce;

        x_left = x_right = INT_MIN;

        // find previous and next point
        for (t=t_left; t<dir.n; t++){

          if (par.m2d[_B_.tss].data[t][p] == _TSAOFF_) continue;

          if (par.m2d[_B_.tss].date[t].ce < x){
            x_left = par.m2d[_B_.tss].date[t].ce;
            y_left = par.m2d[_B_.tss].data[t][p];
            t_left = t;
          } else if (par.m2d[_B_.tss].date[t].ce == x){
            x_left = x_right = x;
            y_left = y_right = par.m2d[_B_.tss].data[t][p];
            t_left = t;
            break;
          } else if (par.m2d[_B_.tss].date[t].ce > x){
            x_right = par.m2d[_B_.tss].date[t].ce;
            y_right = par.m2d[_B_.tss].data[t][p];
            break;
          }

        }

        // set nodata, copy value or interpolate
        if (x_left < 0 || x_right < 0){
          y = _TSAOFF_;
        } else if (x_left == x_right){
          y = y_left;
        } else {
          y = (y_left*(x_right-x) + y_right*(x-x_left))/(x_right-x_left);
        }

        par.m2d[_B_.tsi].data[i][p] = (int2s)y;

      }
      
    }

  }


  return SUCCESS;
}


/** Moving statistics interpolation
+++ This function interpolates and smoothes a time series using a moving
+++ window average filter.
--- MASK:   Analysis mask (optional)
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int interpolate_moving(bool *MASK){
int p, t, t_left, i;
float x, x_;
float max_x = par.mov_max;
double sum, num;


  #pragma omp parallel private(i,t,t_left,x,x_,sum,num) shared(MASK,dir,_NC_,par,_B_,max_x) default(none)
  {

    #pragma omp for
    for (p=0; p<_NC_; p++){

      // if masked, use nodata
      if (MASK != NULL && !MASK[p]){
        for (i=0; i<par.m2d[_B_.tsi].length; i++) par.m2d[_B_.tsi].data[i][p] = _TSAOFF_;
        continue;
      }

      // interpolate for each equidistant timestep
      for (i=0, t_left=0; i<par.m2d[_B_.tsi].length; i++){

        // current time
        x = par.m2d[_B_.tsi].date[i].ce;

        sum = num = 0.0;

        // use all points within temporal window
        for (t=t_left; t<dir.n; t++){

          if (par.m2d[_B_.tss].data[t][p] == _TSAOFF_) continue;

          x_ = par.m2d[_B_.tss].date[t].ce;

          if (x-x_ > max_x){ // earlier than window
            t_left = t;
            continue;
          } else if (x_-x > max_x){ // later than window
            break;
          } else { // in window
            sum += par.m2d[_B_.tss].data[t][p]; num++;
          }

        }

        // interpolate with moving mean, or use nodata
        if (num > 0){
           par.m2d[_B_.tsi].data[i][p] = (int2s)(sum/num);
        } else {
           par.m2d[_B_.tsi].data[i][p] = _TSAOFF_;
        }

      }

    }

  }


  return SUCCESS;
}


/** RBF interpolation
+++ This function interpolates and smoothes a time series using ensembles
+++ of radial basis functions. The ensembles are weighted according to da-
+++ ta availability.
+++ The RBF interpolation was adpoted from the python routine 
+++ 'RBF_fitting_large_scale.py', Copyright (C) 2018 Andreas Rabe
--- MASK:   Analysis mask (optional)
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int interpolate_rbf(bool *MASK){
int p, t, *t_left = NULL, i, k, nk = par.rbf_nk, bin, nbin;
int x, x_, *max_x = NULL, maxmax_x;
float y;
double *sum_yw = NULL, *sum_w = NULL;
double sum_, sum_kd, sum_d;
float sigma;
float **kernel = NULL;
float *max_w = NULL;


  // number of maximum bins in kernel
  if ((nbin = 8*par.rbf_sigma[nk-1]) % 2 == 0) nbin++;

  // allocate kernels etc
  alloc_2D((void***)&kernel, nk, _NC_, sizeof(float));
  alloc((void**)&max_x,  nk, sizeof(int));
  alloc((void**)&max_w,  nk, sizeof(double));


  // compile kernels
  for (k=0; k<nk; k++){

    // max temporal distance to be considered
    maxmax_x = (nbin-1)/2;
    sum_ = 0.0;
    
    sigma = par.rbf_sigma[k];

    // compute kernel weights, cutoff (alpha=0.1) + sum of weights
    for (bin=0; bin<nbin; bin++){
      kernel[k][bin] = 1/(sqrt(2*M_PI) * 
        sigma)*exp(-0.5*(bin-maxmax_x)*(bin-maxmax_x)/(sigma*sigma));
      if (bin-maxmax_x >= 0){
        sum_ += kernel[k][bin];
        if (sum_ <= par.rbf_cutoff/2.0){
          max_x[k] = bin-maxmax_x;
          max_w[k] += kernel[k][bin];
        }
      }
    }
    
    max_w[k] *= 2;

    #ifdef DEBUG
    printf("sigma: %.0f, kernel width: %d, max. distance: %d, max sign. distance: %d, max weight: %.2f\n", sigma, nbin, maxmax_x, max_x[k], max_w[k]);
    #endif

  }


  #pragma omp parallel private(k,i,t,t_left,x,x_,y,sum_yw,sum_w,sum_kd,sum_d) shared(MASK,dir,_NC_,par,_B_,max_x,maxmax_x,max_w,kernel,nk) default(none)
  {

    // allocate thread-specific memory
    alloc((void**)&t_left, nk, sizeof(int));
    alloc((void**)&sum_yw, nk, sizeof(double));
    alloc((void**)&sum_w,  nk, sizeof(double));

    #pragma omp for
    for (p=0; p<_NC_; p++){

      // if masked, use nodata
      if (MASK != NULL && !MASK[p]){
        for (i=0; i<par.m2d[_B_.tsi].length; i++) par.m2d[_B_.tsi].data[i][p] = _TSAOFF_;
        continue;
      }

      for (k=0; k<nk; k++) t_left[k] = 0;

      // interpolate for each equidistant timestep
      for (i=0; i<par.m2d[_B_.tsi].length; i++){

        // current time
        x = par.m2d[_B_.tsi].date[i].ce;

        for (k=0; k<nk; k++){

          sum_yw[k] = sum_w[k] = 0.0;

          // use all points within temporal window
          for (t=t_left[k]; t<dir.n; t++){

            x_ = par.m2d[_B_.tss].date[t].ce;

            if (x-x_ > max_x[k]){ // earlier than window
              t_left[k] = t;
              continue;
            } else if (x_-x > max_x[k]){ // later than window
              break;
            } else { // in window
              if (par.m2d[_B_.tss].data[t][p] == _TSAOFF_) continue;
              sum_yw[k] += par.m2d[_B_.tss].data[t][p]*kernel[k][x_-x+maxmax_x];
              sum_w[k]  += kernel[k][x_-x+maxmax_x];
            }

          }

        }

        // compute weighted average + 
        // weight the kernels with their data availability (weighted with 
        //        the same kernel as the time series)
        for (k=0, sum_kd=0, sum_d=0; k<nk; k++){
          if (sum_w[k] > 0){
            // weighted average * weighted data availability (weighted fractions of days)
            // sum_kd += sum_yw[k]/sum_w[k] * sum_w[k]/max_w[k];
            // this reduces to:
            sum_kd += sum_yw[k]/max_w[k];
            sum_d  += sum_w[k]/max_w[k];
          }
        }

        // ensemble fit
        if (sum_d > 0){
          y = sum_kd/sum_d;
        } else {
          y = _TSAOFF_;
        }

        par.m2d[_B_.tsi].data[i][p] = (int2s)y;

      }

    }

    // clean
    free((void*)t_left);
    free((void*)sum_yw);
    free((void*)sum_w);

  }

  // clean
  free_2D((void**)kernel, nk);
  free((void*)max_x);
  free((void*)max_w);

  return SUCCESS;
}

