/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
This program generates time series and provides functionality for time 
series analysis
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#include "force.h"


int main ( int argc, char *argv[] ){
level2_t *L2 = NULL;
bool *MASK = NULL;
int l;
int tx, ty, tc = 0, tk = 0;
int *tiles_x, *tiles_y, tiles_k;
char dname[MAX_STR_LEN];
char fname[MAX_STR_LEN];
double geotran[6]   = { 0, 0, 0, 0, 0, 0 };
char tmp_c[4];
bool daytrend;
time_t TIME, TIME_I, TIME_C, TIME_O;
double time_i = 0, time_c = 0, time_o = 0, time_to, ctime = 0;
double eta_h, eta_m, eta_s;

  if (argc != 2){ printf("usage: %s parameter-file\n", argv[0]); exit(1);}


  /** INITIALIZING
  +** *******************************************************************/

  time(&TIME);
  printf("Initializing startup"); fflush(stdout);


  par.f_par = argv[1];

  // parse parameter file
  if (parse_param_tsa() != SUCCESS){
    printf("Reading parameter file failed!\n"); exit(1);}

  // copy datacube definition
  if (copy_datacube_def(par.d_level2, par.d_tsa) == FAILURE){
    printf("Copying datacube definition failed.\n"); exit(1);}

  // get tile file and read it
  if (tilefileread(&tiles_x, &tiles_y, &tiles_k) != SUCCESS){
    printf("Reading tile file failed!\n"); exit(1);}

  // how many tiles to do?
  for (ty=par.tminy; ty<=par.tmaxy; ty++){
  for (tx=par.tminx; tx<=par.tmaxx; tx++){

    // if tile is not whitelisted (if specified), skip
    if (tile_whitelisted(tiles_x, tiles_y, tiles_k, tx, ty) == FAILURE) continue;
    tk++;

  }
  }
   
  mem_to_disc_dict_init();

  printf("\rTSA processing started, I/C/O TBD, ETA: TBD");
  fflush(stdout);

  // open threads
  omp_set_num_threads(par.cpu);


  /** START PROCESSING; LOOP OVER ALL AVAILABLE TILES
  +** *******************************************************************/

  for (ty=par.tminy; ty<=par.tmaxy; ty++){
  for (tx=par.tminx; tx<=par.tmaxx; tx++){

    time(&TIME_I);

    // current tile
    sprintf(dname, "%s/X%04d_Y%04d", par.d_level2, tx, ty);

    // if tile is not whitelisted (if specified), skip
    if (tile_whitelisted(tiles_x, tiles_y, tiles_k, tx, ty) == FAILURE) continue;

    // get Level2 file listing, skip if tile is empty
    if (list_level2(dname) == FAILURE){ tc++; continue;}

    sprintf(fname, "%s/%s", dname, dir.list[0]);
    image_info(fname, geotran);


    /** READ LEVEL2 DATA
    +** *****************************************************************/
    if (read_level2(dname, &L2, _NX_*par.res, _NY_*par.res) != SUCCESS){
      printf("error reading L2.\n"); exit(1);}
    //buffer_level2(L2);


    /** READ MASK DATA
    +** *****************************************************************/
    if (strcmp(par.d_mask, "NULL") != 0){

      sprintf(dname, "%s/X%04d_Y%04d", par.d_mask, tx, ty);
      if (read_mask(dname, &MASK) == FAILURE){
        #ifdef DEBUG
        printf("there is no mask in %s. ", dname); 
        #endif
        tc++;
        continue;
      }

    }

    mem_to_disc_dict_tsa();

    time_i += proctime(TIME_I);
    time(&TIME_C); time_to = 0;

    /** COMPUTE SPECTRAL INDEX TIME SERIES
    +** *****************************************************************/
    if (spectral_index(L2, MASK) != SUCCESS){
      printf("error computing index.\n"); exit(1);}
    free_level2(L2);


    /** OUTPUT MODEL RESIDUALS
    +** *****************************************************************/
    if (par.orms){
      time(&TIME_O);
      if (par.m2d[_B_.rms].save) write_products_tsa(tx, ty, par.m2d[_B_.rms], geotran);
      time_to += proctime(TIME_O);
    }


    /** REMOVE SPIKES IN TIME SERIES
    +** *****************************************************************/
    if (par.despike > 0){
      if (despike_timeseries(MASK, par.despike*_RS_) != SUCCESS){
        printf("despiking time series failed.\n"); exit(1);}
    }


    /** INTERPOLATE TIME SERIES
    +** *****************************************************************/
    if (par.interpolate != NONE && (par.otsi || par.lsp)){
      if (interpolate_timeseries(MASK) != SUCCESS){
        printf("interpolating time series failed.\n"); exit(1);}
    }


    /** LSP FROM TIME SERIES
    +** *****************************************************************/
    if (par.lsp){
      if (phenology_timeseries(MASK) != SUCCESS){
        printf("phenology from time series failed.\n"); exit(1);}
    }


    /** COMPUTE BASIC STATISTICS
    +** *****************************************************************/
    if (par.m2d[_B_.sta].save){
      basic_timeseries_stats(MASK);
      time(&TIME_O);
      write_products_tsa(tx, ty, par.m2d[_B_.sta], geotran);
      time_to += proctime(TIME_O);
    }


    /** FOLD TIME SERIES BY YEAR
    +** *****************************************************************/
    if (par.m2d[_B_.fby].save || 
       (par.trend == 0 && (par.m2d[_B_.trd].save || par.m2d[_B_.cat].save))){
      fold_timeseries(MASK, 0);
    }


    /** FOLD TIME SERIES BY MONTH
    +** *****************************************************************/
    if (par.m2d[_B_.fbm].save || 
       (par.trend == 1 && (par.m2d[_B_.trd].save || par.m2d[_B_.cat].save))){
      fold_timeseries(MASK, 1);
    }


    /** FOLD TIME SERIES BY WEEK
    +** *****************************************************************/
    if (par.m2d[_B_.fbw].save || 
       (par.trend == 2 && (par.m2d[_B_.trd].save || par.m2d[_B_.cat].save))){
      fold_timeseries(MASK, 2);
    }


    /** FOLD TIME SERIES BY DAY
    +** *****************************************************************/
    if (par.m2d[_B_.fbd].save || 
       (par.trend == 3 && (par.m2d[_B_.trd].save || par.m2d[_B_.cat].save))){
      fold_timeseries(MASK, 3);
    }


    /** COMPUTE TREND ANALYSIS ON FOLDED TIME SERIES
    +** *****************************************************************/
    if (par.m2d[_B_.trd].save && par.fold != 3){
      if (par.trend == 0){
        l = _B_.fby;
      } else if (par.trend == 1){
        l = _B_.fbm;
      } else if (par.trend == 2){
        l = _B_.fbw;
      } else if (par.trend == 3){
        l = _B_.fbd;
      }
      trend_timeseries(par.m2d[l].data, par.m2d[l].ts, MASK, par.m2d[l].length, false);
      time(&TIME_O);
      write_products_tsa(tx, ty, par.m2d[_B_.trd], geotran);
      time_to += proctime(TIME_O);
    }


    /** COMPUTE TREND ANALYSIS ON LSP
    +** *****************************************************************/
    if (par.m2d[_B_.trd].save && par.fold == 3){

      for (l=_B_.dem; l<=_B_.rmf; l++){
        if (l > _B_.dlm) daytrend = false; else daytrend = true;
        trend_timeseries(par.m2d[l].data, par.m2d[l].ts, MASK, par.m2d[l].length, daytrend);
        time(&TIME_O);
        strcpy(tmp_c, par.fold_c);
        strcpy(par.fold_c,  par.m2d[l].product);
        write_products_tsa(tx, ty, par.m2d[_B_.trd], geotran);
        free_2D((void**)par.m2d[_B_.trd].data,    par.m2d[_B_.trd].length); par.m2d[_B_.trd].data    = NULL;
        free((void*)par.m2d[_B_.trd].scale); par.m2d[_B_.trd].scale = NULL;
        free_2D((void**)par.m2d[_B_.trd].bn, par.m2d[_B_.trd].length); par.m2d[_B_.trd].bn = NULL;
        strcpy(par.fold_c, tmp_c);
        time_to += proctime(TIME_O);
      }

    }


    /** COMPUTE CHANGE, AFTEREFFECT AND TREND ANALYSIS ON FOLDED TIME SERIES
    +** *****************************************************************/
    if (par.m2d[_B_.cat].save){
      if (par.trend == 0){
        l = _B_.fby;
      } else if (par.trend == 1){
        l = _B_.fbm;
      } else if (par.trend == 2){
        l = _B_.fbw;
      } else if (par.trend == 3){
        l = _B_.fbd;
      }
      cat_timeseries(par.m2d[l].data, par.m2d[l].ts, MASK, par.m2d[l].length, false);
      time(&TIME_O);
      write_products_tsa(tx, ty, par.m2d[_B_.cat], geotran);
      time_to += proctime(TIME_O);
    }


    /** COMPUTE CHANGE, AFTEREFFECT AND TREND ANALYSIS ON LSP
    +** *****************************************************************/
    if (par.m2d[_B_.cat].save && par.fold == 3){

      for (l=_B_.dem; l<=_B_.rmf; l++){
        if (l > _B_.dlm) daytrend = false; else daytrend = true;
        cat_timeseries(par.m2d[l].data, par.m2d[l].ts, MASK, par.m2d[l].length, daytrend);
        time(&TIME_O);
        strcpy(tmp_c, par.fold_c);
        strcpy(par.fold_c,  par.m2d[l].product);
        write_products_tsa(tx, ty, par.m2d[_B_.cat], geotran);
        free_2D((void**)par.m2d[_B_.cat].data,    par.m2d[_B_.cat].length); par.m2d[_B_.cat].data    = NULL;
        free((void*)par.m2d[_B_.cat].scale); par.m2d[_B_.cat].scale = NULL;
        free_2D((void**)par.m2d[_B_.cat].bn, par.m2d[_B_.cat].length); par.m2d[_B_.cat].bn = NULL;
        strcpy(par.fold_c, tmp_c);
        time_to += proctime(TIME_O);
      }

    }


    /** CENTER/STANDARDIZE FULL TIME SERIES
    +** *****************************************************************/
    if (par.m2d[_B_.tss].save){
      if (par.docenter || par.dostandard) standardize_timeseries(par.m2d[_B_.tss].data, MASK, par.m2d[_B_.tss].length);
      time(&TIME_O);
      write_products_tsa(tx, ty, par.m2d[_B_.tss], geotran);
      time_to += proctime(TIME_O);
    }


    /** CENTER/STANDARDIZE INTERPOLATED TIME SERIES
    +** *****************************************************************/
    if (par.interpolate != NONE && par.otsi){
      if (par.docenter || par.dostandard) standardize_timeseries(par.m2d[_B_.tsi].data, MASK, par.m2d[_B_.tsi].length);
      time(&TIME_O);
      if (par.m2d[_B_.tsi].save) write_products_tsa(tx, ty, par.m2d[_B_.tsi], geotran);
      time_to += proctime(TIME_O);
    }


    /** CENTER/STANDARDIZE LSP TIME SERIES
    +** *****************************************************************/
    if (par.lsp){
      for (l=_B_.dem; l<=_B_.rmf; l++){
        if (par.docenter || par.dostandard) standardize_timeseries(par.m2d[l].data, MASK, par.m2d[l].length);
        time(&TIME_O);
        if (par.m2d[l].save) write_products_tsa(tx, ty, par.m2d[l], geotran);
        time_to += proctime(TIME_O);
      }
    }


    /** CENTER/STANDARDIZE FOLD-BY-YEAR TIME SERIES
    +** *****************************************************************/
    if (par.m2d[_B_.fby].save){
      if (par.docenter || par.dostandard) standardize_timeseries(par.m2d[_B_.fby].data, MASK, par.m2d[_B_.fby].length);
      time(&TIME_O);
      if (par.m2d[_B_.fby].save) write_products_tsa(tx, ty, par.m2d[_B_.fby], geotran);
      time_to += proctime(TIME_O);
    }


    /** CENTER/STANDARDIZE FOLD-BY-MONTH TIME SERIES
    +** *****************************************************************/
    if (par.m2d[_B_.fbm].save){
      if (par.docenter || par.dostandard) standardize_timeseries(par.m2d[_B_.fbm].data, MASK, par.m2d[_B_.fbm].length);
      time(&TIME_O);
      if (par.m2d[_B_.fbm].save) write_products_tsa(tx, ty, par.m2d[_B_.fbm], geotran);
      time_to += proctime(TIME_O);
    }


    /** CENTER/STANDARDIZE FOLD-BY-WEEK TIME SERIES
    +** *****************************************************************/
    if (par.m2d[_B_.fbw].save){
      if (par.docenter || par.dostandard) standardize_timeseries(par.m2d[_B_.fbw].data, MASK, par.m2d[_B_.fbw].length);
      time(&TIME_O);
      if (par.m2d[_B_.fbw].save) write_products_tsa(tx, ty, par.m2d[_B_.fbw], geotran);
      time_to += proctime(TIME_O);
    }


    /** CENTER/STANDARDIZE FOLD-BY-DAY TIME SERIES
    +** *****************************************************************/
    if (par.m2d[_B_.fbd].save){
      if (par.docenter || par.dostandard) standardize_timeseries(par.m2d[_B_.fbd].data, MASK, par.m2d[_B_.fbd].length);
      time(&TIME_O);
      if (par.m2d[_B_.fbd].save) write_products_tsa(tx, ty, par.m2d[_B_.fbd], geotran);
      time_to += proctime(TIME_O);
    }


    /** CLEAN
    +** *****************************************************************/
    mem_to_disc_free();
    free((void*)MASK); MASK = NULL;
    free_2D((void**)dir.list, dir.N);
    free_2D((void**)dir.LIST, dir.N);


    /** PROGRESS
    +** *****************************************************************/
    time_c += (proctime(TIME_C)-time_to);
    time_o += time_to;

    tc++; ctime = time_i + time_c + time_o;
    eta_s = (tk-tc)*(ctime/tc);
    if (eta_s >= 3600){
      eta_h = floor(eta_s/3600); eta_s -= eta_h*3600;
    } else eta_h = 0;
    if (eta_s >= 60){
      eta_m = floor(eta_s/60); eta_s -= eta_m*60;
    } else eta_m = 0;

    printf("\rTSA processing %6.2f%%, I/C/O %03.0f/%03.0f/%03.0f, ETA: %02.0fh %02.0fm %02.0fs", 
      (float)tc/tk*100.0, round(time_i/ctime*100), round(time_c/ctime*100), round(time_o/ctime*100), 
       eta_h, eta_m, eta_s);
    fflush(stdout);

  }
  }

  printf("\n\n");

  proctime_print("Total runtime", TIME);

  return SUCCESS;
}

