/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
This file contains functions that handle warping and reprojection requests
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#include "force.h"


/** Reproject point from geographic to any projection
+++ This function reprojects a coordinate.
--- fname:  filename
--- srs_x:    x-coordinate in source projection
--- srs_y:    y-coordinate in source projection
--- dst_x:    x-coordinate in target projection (returned)
--- dst_y:    y-coordinate in target projection (returned)
--- dst_wkt:  target projection
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int geo_to_map(float  srs_x, float  srs_y, 
               float *dst_x, float *dst_y, char *dst_wkt){
OGRSpatialReference oSrcSRS, oDstSRS;
OGRCoordinateTransformation *poCT;
char *wkt = dst_wkt;
double x, y;

  if (srs_x < -180 || srs_x > 180){
    printf("Longitude is out of bounds.\n"); return FAILURE;}
  if (srs_y <  -90 || srs_y >  90){
    printf("Latitude  is out of bounds.\n"); return FAILURE;}

  x = srs_x; y = srs_y;

  // set coordinate systems
  oSrcSRS.SetWellKnownGeogCS("WGS84");
  oDstSRS.importFromWkt(&wkt);

  // create transformation
  poCT = OGRCreateCoordinateTransformation(&oSrcSRS, &oDstSRS);

  // transform
  if (poCT == NULL || !poCT->Transform(1, &x, &y)){
    printf( "Transformation failed.\n" ); return FAILURE;
  } else delete poCT;

  *dst_x = (float)x; *dst_y = (float)y;
  return SUCCESS;
}


/** Reproject point from any to any other projection
+++ This function reprojects a coordinate.
--- srs_x:    x-coordinate in source projection
--- srs_y:    y-coordinate in source projection
--- dst_x:    x-coordinate in target projection (returned)
--- dst_y:    y-coordinate in target projection (returned)
--- src_wkt:  source projection
--- dst_wkt:  target projection
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int reproj_point(float  srs_x, float  srs_y,
                 float *dst_x, float *dst_y,
                 char *src_wkt, char *dst_wkt){
OGRSpatialReference oSrcSRS, oDstSRS;
OGRCoordinateTransformation *poCT;
char *src_css = src_wkt;
char *dst_css = dst_wkt;
double x, y;

  x = srs_x; y = srs_y;

  // set coordinate systems
  oSrcSRS.importFromWkt(&src_css);
  oDstSRS.importFromWkt(&dst_css);

  // create transformation
  poCT = OGRCreateCoordinateTransformation(&oSrcSRS, &oDstSRS);

  // transform
  if (poCT == NULL || !poCT->Transform(1, &x, &y)){
    printf( "Transformation failed.\n" ); return FAILURE;
  } else delete poCT;

  *dst_x = (float)x; *dst_y = (float)y;
  return SUCCESS;
}


/** Reproject Digital Elevation Model
+++ This function reprojects the Digital Elevation Model into the working 
+++ projection. The reprojection is performed from disc to memory. The re-
+++ projection might be performed in chunks if the number of pixels is too
+++ large for GDAL.
--- rsm:    resampling method
--- DEM:    reprojected Digital Elevation Model (returned)
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int reproj_topo(int rsm, float **DEM){
GDALDatasetH src_dataset, dst_dataset;
GDALDriverH driver;
GDALDataType dt = GDT_Float32;
GDALWarpOptions *wopt;
GDALWarpOperation woper;
GDALResampleAlg resample;
CPLErr eErr = CE_Failure;
void *transformer = NULL;
double dst_geotran[6];
float *buf = NULL, *dem = NULL;
const char *src_wkt;
int i, j, p, np, k, k_do;
int dst_nx, dst_ny;
int dst_xoff, dst_yoff;
float tmp;
int nc = 0;

/** for future development, use only reproj_image **/

  // if no DEM was given, use flat DEM, z = 0m
  if (strcmp(par.fdem, "NULL") == 0){
    alloc((void**)&dem, _NC_, sizeof(float));
    *DEM=dem;
    return SUCCESS;
  }

  // set resampling method
  if (rsm == 0){
    resample = GRA_NearestNeighbour;
  } else if (rsm == 1){
    resample = GRA_Bilinear;
  } else if (rsm == 2){
    resample = GRA_Cubic;
  }

  // georeferencing transform:
  dst_geotran[0] = meta.map_ul.x;
  dst_geotran[1] = meta.res;
  dst_geotran[2] = 0;
  dst_geotran[3] = meta.map_ul.y; 
  dst_geotran[4] = 0;
  dst_geotran[5] = -1*meta.res;

  #ifdef DEBUG
  printf("warp DEM to %.0f / %.0f @ %.0f m\n", dst_geotran[0], dst_geotran[3], dst_geotran[1]);
  #endif

  GDALAllRegister();

  // open input dataset
  if ((src_dataset = GDALOpen(par.fdem, GA_ReadOnly)) == NULL){
    printf("unable to open image\n"); exit(1);}

  src_wkt = GDALGetProjectionRef(src_dataset);
  CPLAssert(src_wkt != NULL && strlen(src_wkt) > 0);
  
  #ifdef DEBUG
  printf("WKT of DEM: %s\n", src_wkt);
  #endif

  // Create the "output" file.  
  driver = GDALGetDriverByName("MEM");
  dst_dataset = GDALCreate(driver, "mem", _NX_, _NY_, 1, dt, NULL);
  CPLAssert(dst_dataset != NULL);
  GDALSetProjection(dst_dataset, meta.css_map);
  GDALSetGeoTransform(dst_dataset, dst_geotran);


  // get transformation beween coordindate systems
  transformer = 
    GDALCreateGenImgProjTransformer(src_dataset, src_wkt, dst_dataset, meta.css_map, 
                      FALSE, 0, 2);
  CPLAssert(transformer != NULL);
  

  // set warping options
  wopt = GDALCreateWarpOptions();
  wopt->hSrcDS = src_dataset;
  wopt->hDstDS = NULL;
  wopt->eResampleAlg = resample;
  wopt->nBandCount = 1;
  wopt->panSrcBands = 
    (int *) CPLMalloc(sizeof(int)*wopt->nBandCount);
  wopt->panSrcBands[0] = 1;
  wopt->panDstBands = 
    (int *) CPLMalloc(sizeof(int)*wopt->nBandCount);
  wopt->panDstBands[0] = 1;
  wopt->pTransformerArg = transformer;
  wopt->pfnTransformer = GDALGenImgProjTransform;
  wopt->eWorkingDataType = dt;

  wopt->padfSrcNoDataReal =
    (double*) CPLMalloc(sizeof(double)*wopt->nBandCount);
  wopt->padfSrcNoDataReal[0] = -32767.0;
  wopt->padfSrcNoDataImag =
    (double*) CPLMalloc(sizeof(double)*wopt->nBandCount);
  wopt->padfSrcNoDataImag[0] = 0.0;

  wopt->padfDstNoDataReal =
    (double*) CPLMalloc(sizeof(double)*wopt->nBandCount);
  wopt->padfDstNoDataReal[0] = wopt->padfSrcNoDataReal[0];
  wopt->padfDstNoDataImag =
    (double*) CPLMalloc(sizeof(double)*wopt->nBandCount);
  wopt->padfDstNoDataImag[0] = wopt->padfSrcNoDataImag[0];


  // allocate memory for buffer
  alloc((void**)&dem, _NC_, sizeof(float));
  for (p=0; p<_NX_*_NY_; p++) dem[p] = -32767;

  // warp
  woper.Initialize(wopt);

  dst_xoff = 0; dst_yoff = 0;
  dst_nx = _NX_; dst_ny = _NY_;
  k = 0, k_do = 0;

  // start with whole image, if unsuccessful, warp in chunks
  while (eErr != CE_None && dst_ny > 1){

    #ifdef DEBUG
    printf("try to warp %d x %d pix, x / y offset %d / %d. ", 
      dst_nx, dst_ny, dst_xoff, dst_yoff);
    #endif

    alloc((void**)&buf, dst_nx*dst_ny, sizeof(float));
    for (np=0; np<dst_nx*dst_ny; np++) buf[np] = -32767;

    // warp
    eErr = woper.WarpRegionToBuffer(dst_xoff, dst_yoff, dst_nx, dst_ny, buf, dt, 0, 0, 0, 0);
    
    if (eErr == CE_Failure){

      #ifdef DEBUG
      printf("Failed. Try smaller chunks.\n");
      #endif

      // decrease size
      if (k++ % 2 == 0){
        tmp = ceil(dst_ny/2.0); dst_ny = tmp;
      } else {
        tmp = ceil(dst_nx/2.0); dst_nx = tmp;
      }

    } else {

      #ifdef DEBUG
      printf("OK.\n");
      #endif

      // copy buffer to DEM
      for (i=dst_yoff, np=0; i<dst_yoff+dst_ny; i++){
      for (j=dst_xoff; j<dst_xoff+dst_nx; j++, np++){
        if (i >= _NY_ || j >= _NX_) continue;
        dem[p = i*_NX_+j] = buf[np];
        buf[np] = -32767;
        nc++;
      }
      }

      // if part of image is still missing, increase offsets
      if (nc < _NC_){
        if (dst_yoff < _NY_ && dst_ny < _NY_){
          dst_yoff+=dst_ny;
        } else if (dst_xoff < _NX_ && dst_nx < _NX_){
          dst_xoff+=dst_nx;
        }
        eErr = CE_Failure;
      } else {
        eErr = CE_None;
      }

      k_do++;

    }

    free((void*)buf);

  }

  #ifdef DEBUG
  printf("\nDEM was warped in %d chunks.\n", k_do);
  #endif

  GDALDestroyGenImgProjTransformer(wopt->pTransformerArg);
  GDALDestroyWarpOptions(wopt);


  // close dataset
  GDALClose(src_dataset);
  GDALClose(dst_dataset);

  *DEM=dem;
  return SUCCESS;
}


/** Reproject Image
+++ This function reprojects an image into the working projection. The re-
+++ projection is performed from disc to memory. The reprojection might be
+++ performed in chunks if the number of pixels is too large for GDAL.
--- fname:       filename
--- rsm:         resampling method
--- b:           band
--- nx:          number of pixels (x) in target
--- ny:          number of pixels (y) in target
--- nodata:      nodata value of source
--- dst_geotran: target geotransformation
--- IMG:         reprojected Image (returned)
+++ Return:      SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int reproj_image(char *fname, int rsm, int b, int nx, int ny, float nodata, double dst_geotran[6], float **IMG){
GDALDatasetH src_dataset, dst_dataset;
GDALDriverH driver;
GDALDataType dt = GDT_Float32;
GDALWarpOptions *wopt;
GDALWarpOperation woper;
GDALResampleAlg resample;
CPLErr eErr = CE_Failure;
void *transformer = NULL;

float *buf = NULL, *img = NULL;
const char *src_wkt;
int i, j, p, np, k, k_do;
int dst_nx, dst_ny;
int dst_xoff, dst_yoff;
float tmp;
int nc = 0, nb;


  // set resampling method
  if (rsm == 0){
    resample = GRA_NearestNeighbour;
  } else if (rsm == 1){
    resample = GRA_Bilinear;
  } else if (rsm == 2){
    resample = GRA_Cubic;
  }

  #ifdef DEBUG
  printf("warp coarse to %.0f / %.0f @ %.0f m\n", dst_geotran[0], dst_geotran[3], dst_geotran[1]);
  #endif

  GDALAllRegister();

  // open input dataset
  if ((src_dataset = GDALOpen(fname, GA_ReadOnly)) == NULL){
    printf("unable to open image\n"); exit(1);}

  src_wkt = GDALGetProjectionRef(src_dataset);
  CPLAssert(src_wkt != NULL && strlen(src_wkt) > 0);
  nb = GDALGetRasterCount(src_dataset);
  
  if (b > nb){
    printf("Requested band is out of bounds! "); return FAILURE;}
  
  #ifdef DEBUG
  printf("WKT of DEM: %s\n", src_wkt);
  #endif

  // Create the "output" file.  
  driver = GDALGetDriverByName("MEM");
  dst_dataset = GDALCreate(driver, "mem", nx, ny, 1, dt, NULL);
  CPLAssert(dst_dataset != NULL);
  GDALSetProjection(dst_dataset, meta.css_map);
  GDALSetGeoTransform(dst_dataset, dst_geotran);


  // get transformation beween coordindate systems
  transformer = 
    GDALCreateGenImgProjTransformer(src_dataset, src_wkt, dst_dataset, meta.css_map, 
                      FALSE, 0, 2);
  CPLAssert(transformer != NULL);
  

  // set warping options
  wopt = GDALCreateWarpOptions();
  wopt->hSrcDS = src_dataset;
  wopt->hDstDS = NULL;
  wopt->eResampleAlg = resample;
  wopt->nBandCount = 1;
  wopt->panSrcBands = 
    (int *) CPLMalloc(sizeof(int)*wopt->nBandCount);
  wopt->panSrcBands[0] = b;
  wopt->panDstBands = 
    (int *) CPLMalloc(sizeof(int)*wopt->nBandCount);
  wopt->panDstBands[0] = b;
  wopt->pTransformerArg = transformer;
  wopt->pfnTransformer = GDALGenImgProjTransform;
  wopt->eWorkingDataType = dt;

  wopt->padfSrcNoDataReal =
    (double*) CPLMalloc(sizeof(double)*wopt->nBandCount);
  wopt->padfSrcNoDataReal[0] = nodata;
  wopt->padfSrcNoDataImag =
    (double*) CPLMalloc(sizeof(double)*wopt->nBandCount);
  wopt->padfSrcNoDataImag[0] = 0.0;

  wopt->padfDstNoDataReal =
    (double*) CPLMalloc(sizeof(double)*wopt->nBandCount);
  wopt->padfDstNoDataReal[0] = wopt->padfSrcNoDataReal[0];
  wopt->padfDstNoDataImag =
    (double*) CPLMalloc(sizeof(double)*wopt->nBandCount);
  wopt->padfDstNoDataImag[0] = wopt->padfSrcNoDataImag[0];


  // allocate memory for buffer
  alloc((void**)&img, nx*ny, sizeof(float));
  for (p=0; p<nx*ny; p++) img[p] = nodata;

  // warp
  woper.Initialize(wopt);

  dst_xoff = 0; dst_yoff = 0;
  dst_nx = nx; dst_ny = ny;
  k = 0, k_do = 0;

  // start with whole image, if unsuccessful, warp in chunks
  while (eErr != CE_None && dst_ny > 1){

    #ifdef DEBUG
    printf("try to warp %d x %d pix, x / y offset %d / %d. ", 
      dst_nx, dst_ny, dst_xoff, dst_yoff);
    #endif

    alloc((void**)&buf, dst_nx*dst_ny, sizeof(float));
    for (np=0; np<dst_nx*dst_ny; np++) buf[np] = nodata;

    // warp
    eErr = woper.WarpRegionToBuffer(dst_xoff, dst_yoff, dst_nx, dst_ny, buf, dt, 0, 0, 0, 0);
    
    if (eErr == CE_Failure){

      #ifdef DEBUG
      printf("Failed. Try smaller chunks.\n");
      #endif

      // decrease size
      if (k++ % 2 == 0){
        tmp = ceil(dst_ny/2.0); dst_ny = tmp;
      } else {
        tmp = ceil(dst_nx/2.0); dst_nx = tmp;
      }

    } else {

      #ifdef DEBUG
      printf("OK.\n");
      #endif

      // copy buffer to image
      for (i=dst_yoff, np=0; i<dst_yoff+dst_ny; i++){
      for (j=dst_xoff; j<dst_xoff+dst_nx; j++, np++){
        if (i >= ny || j >= nx) continue;
        img[p = i*nx+j] = buf[np];
        buf[np] = nodata;
        nc++;
      }
      }

      // if part of image is still missing, increase offsets
      if (nc < nx*ny){
        if (dst_yoff < ny && dst_ny < ny){
          dst_yoff+=dst_ny;
        } else if (dst_xoff < nx && dst_nx < nx){
          dst_xoff+=dst_nx;
        }
        eErr = CE_Failure;
      } else {
        eErr = CE_None;
      }

      k_do++;

    }

    free((void*)buf);

  }

  #ifdef DEBUG
  printf("\nImage was warped in %d chunks.\n", k_do);
  #endif

  GDALDestroyGenImgProjTransformer(wopt->pTransformerArg);
  GDALDestroyWarpOptions(wopt);


  // close dataset
  GDALClose(src_dataset);
  GDALClose(dst_dataset);

  *IMG=img;
  return SUCCESS;
}


/** Warp the internal data stack to target projection / resolution
+++ This function warps the data stack (including reflectance and all re-
+++ quested products) to the target projection and resolution. The warping
+++ operation is performed in memory. Depending on the size of the data
+++ stack, the operation is performed in one or multiple chunks.
--- SRC:     Data stack
--- DST:     Reprojected data stack (returned)
--- src_ulx: upper-left x-coordinate
--- src_uly: upper-left y-coordinate
--- src_wkt: source projection
--- dst_ulx: upper-left x-coordinate after reprojection (returned)
--- dst_uly: upper-left y-coordinate after reprojection (returned)
--- dst_nx:  number of pixels in x-direction after reprojection (returned)
--- dst_ny:  number of pixels in y-direction after reprojection (returned)
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int warp_image(int2s **SRC, int2s ***DST, float src_ulx, float src_uly, char *src_wkt, float *dst_ulx, float *dst_uly, int *dst_nx, int *dst_ny){
int i, j, p, b, k, nd = _ND_-1;
size_t src_nx, src_ny;
size_t src_dim, dst_dim;
double src_geotran[6], dst_geotran[6];
float t0x, t0y; // ul x/y of first tile
int idx, idy;   // id of first tile
int2s **buf = NULL;
GDALDatasetH src_dataset, dst_dataset;
GDALRasterBandH src_band;
GDALDriverH driver;
GDALDataType dt = GDT_Int16;
GDALWarpOptions *wopt;
GDALWarpOperation woper;
void *transformer;


  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  // register drivers and fetch in-memory driver
  GDALAllRegister();
  if ((driver = GDALGetDriverByName("MEM")) == NULL){
    printf("could not fetch in-memory driver. "); return FAILURE;}

  /** "create" source dataset
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ **/

  // georeferencing transform:
  src_geotran[0] = src_ulx;
  src_geotran[1] = meta.res;
  src_geotran[2] = 0;
  src_geotran[3] = src_uly;
  src_geotran[4] = 0;
  src_geotran[5] = -1*meta.res;
  
  // create in-memory image
  if ((src_dataset = GDALCreate(driver, "mem", _NX_, _NY_, nd, dt, NULL)) == NULL){
    printf("could not create src image. "); return FAILURE;}
  if (GDALSetProjection(src_dataset, src_wkt) == CE_Failure){
    printf("could not set src projection. "); return FAILURE;}
  if (GDALSetGeoTransform(src_dataset, src_geotran) == CE_Failure){
    printf("could not set src geotransformation. "); return FAILURE;}
  
  // "copy" data to source image
  for (b=0; b<nd; b++){
    src_band = GDALGetRasterBand(src_dataset, b+1);
    if (GDALRasterIO(src_band, GF_Write, 0, 0, _NX_, _NY_, 
      SRC[b], _NX_, _NY_, dt, 0, 0 ) == CE_Failure){
      printf("could not 'copy' src data. "); return FAILURE;}
  }


  /** approx. extent of destination dataset
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ **/

  // create transformer between source and destination
  if ((transformer = GDALCreateGenImgProjTransformer(src_dataset, src_wkt,
    NULL, par.css_dst, FALSE, 0, 2)) == NULL){
    printf("could not create image to image transformer. "); return FAILURE;}

  // estimate approximate extent
  if (GDALSuggestedWarpOutput(src_dataset, GDALGenImgProjTransform, 
    transformer, dst_geotran, dst_nx, dst_ny) == CE_Failure){
    printf("could not suggest dst extent. "); return FAILURE;}

  // align with output grid of data cube    
  if (par.dotile){

    // ul coordinate of 1st tile: t0x, t0y
    tile_find(dst_geotran[0], dst_geotran[3], &t0x, &t0y, &idx, &idy);
    
    // shift t0x to the right until it is right of image-ul, 
    // then step back one pixel and assign as new image-ul
    while (t0x < dst_geotran[0]) t0x += par.res;
    t0x -= par.res;
    dst_geotran[0] = t0x;

    // shift t0y to the bottom until it is below image-ul, 
    // then step back one pixel and assign as new image-ul
    while (t0y > dst_geotran[3]) t0y -= par.res;
    t0y += par.res;
    dst_geotran[3] = t0y;

  }
  
  // convert computed resolution to actual dst resolution
  // add some rows and columns, just to be sure
  *dst_nx = *dst_nx * dst_geotran[1] / par.res + 10;
  *dst_ny = *dst_ny * dst_geotran[5] / par.res*-1 + 10;
  dst_geotran[1] = par.res; dst_geotran[5] = -1*par.res; 

  // destroy transformer
  GDALDestroyGenImgProjTransformer(transformer);

  #ifdef DEBUG
  printf("dst nx/ny: %d/%d\n", *dst_nx, *dst_ny);
  print_dvector(dst_geotran, "dst geotransf.", 6, 1, 2);
  #endif


  /** "create" the destination dataset
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ **/
  if ((dst_dataset = GDALCreate(driver, "mem", *dst_nx, *dst_ny, nd, dt, NULL)) == NULL){
    printf("could not create dst image. "); return FAILURE;}
  if (GDALSetProjection(dst_dataset, par.css_dst) == CE_Failure){
    printf("could not set dst projection. "); return FAILURE;}
  if (GDALSetGeoTransform(dst_dataset, dst_geotran) == CE_Failure){
    printf("could not set dst geotransformation. "); return FAILURE;}


  /** create accurate transformer between source and destination
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ **/
  if ((transformer = GDALCreateGenImgProjTransformer(src_dataset, src_wkt,
    dst_dataset, par.css_dst, FALSE, 0, 2)) == NULL){
    printf("could not create image to image transformer. "); return FAILURE;}


  /** set warping options
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ **/
  wopt = GDALCreateWarpOptions();
  wopt->hSrcDS = src_dataset;
  wopt->hDstDS = NULL;
  if (par.resample == 1) wopt->eResampleAlg = GRA_NearestNeighbour;
  if (par.resample == 2) wopt->eResampleAlg = GRA_Bilinear;
  if (par.resample == 3) wopt->eResampleAlg = GRA_Cubic;
  wopt->nBandCount = nd;
  wopt->panSrcBands = (int *) CPLMalloc(sizeof(int)*nd);
  for (b=0; b<nd; b++) wopt->panSrcBands[b] = b+1;
  wopt->panDstBands = (int *) CPLMalloc(sizeof(int)*nd);
  for (b=0; b<nd; b++) wopt->panDstBands[b] = b+1;
  wopt->pTransformerArg = transformer;
  wopt->pfnTransformer = GDALGenImgProjTransform;
  wopt->eWorkingDataType = dt;
  wopt->padfSrcNoDataReal = (double*) CPLMalloc(sizeof(double)*nd);
  for (b=0; b<nd; b++) wopt->padfSrcNoDataReal[b] = _NODATA_;
  wopt->padfSrcNoDataImag = (double*) CPLMalloc(sizeof(double)*nd);
  //for (b=0; b<nd; b++) wopt->padfSrcNoDataImag[b] = _NODATA_;
  for (b=0; b<nd; b++) wopt->padfSrcNoDataImag[b] = 0;


  /** allocate long memory block for data, and set pointers to bands
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ **/
  // note that this buffer also includes a band for QAI, which will be
  // warped hereafter
  dst_dim = ((size_t)*dst_nx)*((size_t)*dst_ny)*(size_t)_ND_*sizeof(int2s);
  if ((buf = (int2s**) malloc(_ND_*sizeof(int2s*))) == NULL){
    printf("unable to allocate memory!\n"); return FAILURE;}
  if ((buf[0] = (int2s*) malloc(dst_dim)) == NULL){
    printf("unable to allocate memory!\n"); return FAILURE;}
  for (b=0; b<_ND_; b++) buf[b] = (*buf + (*dst_nx) * (*dst_ny)*b);

  /** initialize with no-data
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ **/
  for (b=0; b<nd; b++){
    for (p=0; p<(((size_t)*dst_nx)*((size_t)*dst_ny)); p++){
      buf[b][p] = _NODATA_;
    }
  }

  /** check if we can warp the image in one operation or need chunks
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ **/
  src_nx = _NX_; src_ny = _NY_; k = 0;
  while ((src_dim = src_nx*src_ny*(size_t)nd*sizeof(int2s)) > INT_MAX){
    if (k % 2 == 0) src_nx/=2; else src_ny/=2;
    k++;}
  
  #ifdef DEBUG
  printf("\nImage stack is warped in %d chunks.\n", k+1);
  #endif


  /** warp the image, use chunks if necessary
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ **/
  if (woper.Initialize(wopt) == CE_Failure){
      printf("could not initialize warper. "); return FAILURE;}

  for (i=0; i<_NY_; i+=src_ny){
  for (j=0; j<_NX_; j+=src_nx){
    if (woper.WarpRegionToBuffer(0, 0, *dst_nx, *dst_ny, buf[0],
      dt, j, i, src_nx, src_ny) == CE_Failure){
      printf("could not warp. "); return FAILURE;}
  }
  }

  /** close & clean
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ **/
  GDALDestroyGenImgProjTransformer(wopt->pTransformerArg);
  GDALDestroyWarpOptions(wopt);
  GDALClose(src_dataset);
  GDALClose(dst_dataset);

  #ifdef DEBUG
  char fname[MAX_STR_LEN];
  FILE *fp = NULL;
  sprintf(fname, "%s/WARP.dat", DEBUGPATH);
  fp = fopen(fname, "wb");
  fwrite((const void*)  buf[0], sizeof(int2s), dst_dim/sizeof(int2s), fp);
  fclose(fp);
  sprintf(fname, "%s/WARP.hdr", DEBUGPATH);
  db_write_header(fname, *dst_nx, *dst_ny, _ND_, 2, (char*) "");
  #endif

  #ifdef CLOCK
  proctime_print("warping output stack", TIME);
  #endif


  *DST=buf;
  *dst_ulx = dst_geotran[0];
  *dst_uly = dst_geotran[3];

  return SUCCESS;
}


/** Warp the QAI data stack to target projection / resolution
+++ This function warps the QAI stack to the target projection and reso-
+++ lution. QAI cannot be warped with the rest of the data stack because
+++ it needs nearest neighbor resampling. The warping operation is perfor-
+++ med in memory. Depending on the size of the data stack, the operation
+++ is performed in one or multiple chunks.
--- SRC:     Data stack, QAI
--- DST:     Reprojected data stack, QAI (returned)
--- src_ulx: upper-left x-coordinate
--- src_uly: upper-left y-coordinate
--- src_wkt: source projection
--- dst_ulx: upper-left x-coordinate after reprojection
--- dst_uly: upper-left y-coordinate after reprojection
--- dst_nx:  number of pixels in x-direction after reprojection
--- dst_ny:  number of pixels in y-direction after reprojection
+++ Return: SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int warp_qai(int2s **SRC, int2s **DST, float src_ulx, float src_uly, char *src_wkt, float dst_ulx, float dst_uly, int dst_nx, int dst_ny){

int i, j, p, b = par.m2d[_B_.qai].start, k, nodata = par.m2d[_B_.qai].nodata;
size_t src_nx, src_ny;
size_t src_dim;
double src_geotran[6], dst_geotran[6];

GDALDatasetH src_dataset, dst_dataset;
GDALRasterBandH src_band;
GDALDriverH driver;
GDALDataType dt = GDT_Int16;
GDALWarpOptions *wopt;
GDALWarpOperation woper;
void *transformer;

  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  // register drivers and fetch in-memory driver
  GDALAllRegister();
  if ((driver = GDALGetDriverByName("MEM")) == NULL){
    printf("could not fetch in-memory driver. "); return FAILURE;}

  /** "create" source dataset
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ **/

  // georeferencing transform:
  src_geotran[0] = src_ulx;
  src_geotran[1] = meta.res;
  src_geotran[2] = 0;
  src_geotran[3] = src_uly;
  src_geotran[4] = 0;
  src_geotran[5] = -1*meta.res;
  
  // create in-memory image
  if ((src_dataset = GDALCreate(driver, "mem", _NX_, _NY_, 1, dt, NULL)) == NULL){
    printf("could not create src image. "); return FAILURE;}
  if (GDALSetProjection(src_dataset, src_wkt) == CE_Failure){
    printf("could not set src projection. "); return FAILURE;}
  if (GDALSetGeoTransform(src_dataset, src_geotran) == CE_Failure){
    printf("could not set src geotransformation. "); return FAILURE;}
  
  // "copy" data to source image
  src_band = GDALGetRasterBand(src_dataset, 1);
  if (GDALRasterIO(src_band, GF_Write, 0, 0, _NX_, _NY_, 
    SRC[b], _NX_, _NY_, dt, 0, 0 ) == CE_Failure){
    printf("could not 'copy' src data. "); return FAILURE;}


  /** destination dataset
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ **/

  // georeferencing transform:
  dst_geotran[0] = dst_ulx;
  dst_geotran[1] = par.res;
  dst_geotran[2] = 0;
  dst_geotran[3] = dst_uly;
  dst_geotran[4] = 0;
  dst_geotran[5] = -1*par.res;


  /** "create" the destination dataset
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ **/
  if ((dst_dataset = GDALCreate(driver, "mem", dst_nx, dst_ny, _ND_, dt, NULL)) == NULL){
    printf("could not create dst image. "); return FAILURE;}
  if (GDALSetProjection(dst_dataset, par.css_dst) == CE_Failure){
    printf("could not set dst projection. "); return FAILURE;}
  if (GDALSetGeoTransform(dst_dataset, dst_geotran) == CE_Failure){
    printf("could not set dst geotransformation. "); return FAILURE;}


  /** create accurate transformer between source and destination
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ **/
  if ((transformer = GDALCreateGenImgProjTransformer(src_dataset, src_wkt,
    dst_dataset, par.css_dst, FALSE, 0, 2)) == NULL){
    printf("could not create image to image transformer. "); return FAILURE;}


  /** set warping options
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ **/
  wopt = GDALCreateWarpOptions();
  wopt->hSrcDS = src_dataset;
  wopt->hDstDS = NULL;
  wopt->eResampleAlg = GRA_NearestNeighbour;
  wopt->nBandCount = 1;
  wopt->panSrcBands = (int *) CPLMalloc(sizeof(int));
  wopt->panSrcBands[0] = 1;
  wopt->panDstBands = (int *) CPLMalloc(sizeof(int));
   wopt->panDstBands[0] = 1;
  wopt->pTransformerArg = transformer;
  wopt->pfnTransformer = GDALGenImgProjTransform;
  wopt->eWorkingDataType = dt;
  wopt->padfSrcNoDataReal = (double*) CPLMalloc(sizeof(double));
  wopt->padfSrcNoDataReal[0] = nodata;
  wopt->padfSrcNoDataImag = (double*) CPLMalloc(sizeof(double));
  wopt->padfSrcNoDataImag[0] = 0;


  /** initialize with no-data
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ **/
  for (p=0; p<dst_nx*dst_ny; p++) DST[b][p] = nodata;


  /** check if we can warp the image in one operation or need chunks
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ **/
  src_nx = _NX_; src_ny = _NY_; k = 0;
  while ((src_dim = src_nx*src_ny*sizeof(int2s)) > INT_MAX){
    if (k % 2 == 0) src_nx/=2; else src_ny/=2;
    k++;}
  
  #ifdef DEBUG
  printf("\nQAI is warped in %d chunks.\n", k+1);
  #endif


  /** warp the image, use chunks if necessary
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ **/
  if (woper.Initialize(wopt) == CE_Failure){
      printf("could not initialize warper. "); return FAILURE;}

  for (i=0; i<_NY_; i+=src_ny){
  for (j=0; j<_NX_; j+=src_nx){
    if (woper.WarpRegionToBuffer(0, 0, dst_nx, dst_ny, DST[b],
      dt, j, i, src_nx, src_ny) == CE_Failure){
      printf("could not warp. "); return FAILURE;}
  }
  }

  /** close & clean
  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ **/
  GDALDestroyGenImgProjTransformer(wopt->pTransformerArg);
  GDALDestroyWarpOptions(wopt);
  GDALClose(src_dataset);
  GDALClose(dst_dataset);

  #ifdef DEBUG
  char fname[MAX_STR_LEN];
  FILE *fp = NULL;

  sprintf(fname, "%s/WARP-QAI.dat", DEBUGPATH);
  fp = fopen(fname, "wb");
  fwrite((const void*)  DST[b], sizeof(int2s), dst_nx*dst_ny, fp);
  fclose(fp);
  sprintf(fname, "%s/WARP-QAI.hdr", DEBUGPATH);
  db_write_header(fname, dst_nx, dst_ny, 1, 2, (char*) "");
  #endif

  #ifdef CLOCK
  proctime_print("warping QAI", TIME);
  #endif

  return SUCCESS;
}

