/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

This file is part of FORCE - Framework for Operational Radiometric 
Correction for Environmental monitoring.

Copyright (C) 2013-2018 David Frantz

FORCE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FORCE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FORCE.  If not, see <http://www.gnu.org/licenses/>.

+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

/**+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
This file contains functions for handling Atmospheric Water Vapor
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/


#include "force.h"


/** Water vapor transmittance LUT
+++ This function computes a water vapor transmittace look-up-table, which
+++ is used for fast estimation of Sentinel-2 water vapor. The function 
+++ will exit successfully if atmospheric correction is disabled or if not
+++ Sentinel-2
+++ Return: SUCCESS / FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int wvp_transmitt_lut(){
int b, km, kw;
float w, m, m_min, m_max;


  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif

  if (!par.doatmo || meta.mission != MISSION_SENTINEL2) return SUCCESS;
  
  alloc_3D((void****)&atc.Tw_lut, _NB_, 701, 101, sizeof(float));

  if (atc.cosszen[0] < atc.cosvzen[0]){
    m_min = atc.cosszen[0]; 
  } else {
    m_min = atc.cosvzen[0];
  }
  m_min = floor(m_min*100.0)/100.0;

  if (atc.cosszen[1] > atc.cosvzen[1]){
    m_max = atc.cosszen[1];
  } else {
    m_max = atc.cosvzen[1];
  }
  m_max = ceil(m_max*100.0)/100.0;

  #ifdef DEBUG
  printf("m min/max: %.2f %.2f\n", m_min, m_max);
  #endif


  for (km=0; km<101; km++){
    
    m = km*0.01;
    if (m < m_min || m > m_max) continue;


    for (kw=0; kw<701; kw++){
      w = kw*0.01;
  
      for (b=0; b<_NB_; b++){
        atc.Tw_lut[b][kw][km] = wvp_transmitt(b, w, m);
      }

    }
  }


  #ifdef CLOCK
  proctime_print("water vapor LUT", TIME);
  #endif

  return SUCCESS;
}


/** Water vapor minimizer function
+++ This minimizer function computes BOA reflectance for the reference and
+++ measurement channels and returns the absolute residual.
+++ Return: Absolute residual of BOA reflectance (reference - measurement)
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
double fun_wvp(const gsl_vector *v, void *params){
float *p = (float *)params;
float w, Tsw_nir, Tvw_nir, Tsw_wvp, Tvw_wvp, Tg_nir, Tg_wvp;
float tmp, sr_reference, sr_measure;
int kms, kmv, kw, b_nir, b_wvp;


  w = gsl_vector_get(v, 0);

  // negative water vapor values are not allowed
  if (w < 0.0) w = 0.0;

  // if water vapor > tabulated values, compute
  if (w > 7.0){
    
    Tsw_nir = wvp_transmitt(p[4], w, p[2]);
    Tvw_nir = wvp_transmitt(p[4], w, p[3]);
    Tsw_wvp = wvp_transmitt(p[5], w, p[2]);
    Tvw_wvp = wvp_transmitt(p[5], w, p[3]);

  // use tabulated values
  } else {

    b_nir = (int)p[4];
    b_wvp = (int)p[5];
    kw =  (int)floor(w/0.01);
    kms = (int)floor(p[2]/0.01);
    kmv = (int)floor(p[3]/0.01);

    Tsw_nir = atc.Tw_lut[b_nir][kw][kms];
    Tvw_nir = atc.Tw_lut[b_nir][kw][kmv];
    Tsw_wvp = atc.Tw_lut[b_wvp][kw][kms];
    Tvw_wvp = atc.Tw_lut[b_wvp][kw][kmv];
    
  } 


  Tg_nir  = Tsw_nir*Tvw_nir*p[6];
  Tg_wvp  = Tsw_wvp*Tvw_wvp*p[7];

  tmp = (p[0]-p[8])/Tg_nir;
  sr_reference = tmp / (p[10] + p[12]*tmp);

  tmp = (p[1]-p[9])/Tg_wvp;
  sr_measure = tmp / (p[11] + p[13]*tmp);

  return fabs(sr_reference-sr_measure);
}


/** Water vapor estimation
+++ This function estimates water vapor in Sentinel-2 images. The function
+++ will only return SUCCESS for Landsat. Note that WVP and Tg won't be 
+++ allocated in this case. Water vapor is estimated for each 60m pixel
+++ using the complete radiative transfer assuming that BOA reflectance
+++ of the NIR reference channel @ 0.865µm and the NIR water vapor channel
+++ @ 0.945 should be equal. Nelder-Mead Simplex optimization is used.
+++ Water and shadow pixels will be set to the scene average, and a QAI
+++ flag is set in this case. Gaseous Transmittance is also computed.
+++ Water vapor is scaled by 1000, Gaseous Transmittance by 10000.
--- TOA:    TOA reflectance
--- QAI:    Quality Assurance Information
--- GRD:    Coarse Grid 
--- TOP:    Topograpy layers
--- WVP:    Water vapor (needs to be deallocated)
--- WVP:    Gaseous transmittance (needs to be deallocated)
+++ Return: SUCCESS / FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int water_vapor(int2s **TOA, qai_t QAI, int2u *GRD, top_t TOP, int2s **WVP, int2s ***Tg){
int i, j, ii, jj, p, g, z, b, bb, k;
int kw, kms, kmv;
float reference, measure, dem;
float w = 3, w_avg;
float *tg = NULL, *tg_avg = NULL;
float Tsw, Tvw;
double w_sum = 0, *tg_sum = NULL, num = 0;
int2s *W   = NULL;
int2s **TG = NULL;
float param[14];
const gsl_multimin_fminimizer_type *T = gsl_multimin_fminimizer_nmsimplex2;
gsl_multimin_fminimizer *s = NULL;
gsl_vector *ss = NULL, *x = NULL;
gsl_multimin_function minex_func;
size_t iter = 0;
int status;
double size;


  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif


  if (!par.doatmo || meta.mission != MISSION_SENTINEL2) return SUCCESS;


  /** allocate data and initialize optimizer
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

  alloc((void**)&W,            _NC_, sizeof(int2s));
  alloc_2D((void***)&TG, _NO_, _NC_, sizeof(int2s));
  alloc((void**)&tg,     _NO_, sizeof(float));
  alloc((void**)&tg_sum, _NO_, sizeof(double));
  alloc((void**)&tg_avg, _NO_, sizeof(float));
  
  // start value, is set to last value in loop
  x = gsl_vector_alloc(1);

  // initial step size of simplex
  ss = gsl_vector_alloc(1);
  gsl_vector_set(ss, 0, 0.25);

  // initialize method and iterate
  minex_func.n = 1;
  minex_func.f = fun_wvp;
  minex_func.params = param;
  s = gsl_multimin_fminimizer_alloc(T, 1);


  /**estimate water vapor for each 60m pixel
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

  for (i=0; i<_NY_; i+=6){
  for (j=0; j<_NX_; j+=6){


    /** aggregate reference channel to 60m
    +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

    reference = dem = k = 0;

    for (ii=0; ii<6; ii++){
    for (jj=0; jj<6; jj++){

      p = (i+ii)*_NY_+j+jj;

      if (QAI.off[p] || QAI.shd[p] || 
          QAI.wtr[p] || TOA[_B_.w0][p] < 1500) continue;

      z = TOP.dem[p];
      
      reference += TOA[_B_.w0][p]/_RS_;
      dem       += TOP.dem[p];
      k++;

    }
    }

    if (k == 0) continue;

    reference /= k;
    z = (int)roundf(dem/k);

    p = i*_NY_+j;

    measure = TOA[_B_.w1][p]/_RS_;
    g = GRD[p];
    if (atc.cg[g].vzen == _NODATA_) continue;


    /** copy variables to param, and initialize minimizer
    +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

    param[0]  = reference;
    param[1]  = measure;
    param[2]  = atc.cg[g].ms;
    param[3]  = atc.cg[g].mv;
    param[4]  = _B_.w0;
    param[5]  = _B_.w1;
    param[6]  = atc.cg[g].Tso[_B_.w0]*atc.cg[g].Tvo[_B_.w0];
    param[7]  = atc.cg[g].Tso[_B_.w1]*atc.cg[g].Tvo[_B_.w1];
    param[8]  = atc.cg[g].z[z].rho_p[_B_.w0];
    param[9]  = atc.cg[g].z[z].rho_p[_B_.w1];
    param[10] = atc.cg[g].z[z].T[_B_.w0];
    param[11] = atc.cg[g].z[z].T[_B_.w1];
    param[12] = atc.cg[g].z[z].s[_B_.w0];
    param[13] = atc.cg[g].z[z].s[_B_.w1];

    gsl_vector_set(x, 0, w);
    gsl_multimin_fminimizer_set(s, &minex_func, x, ss);


    /** optimize radiative transfer, estimate water vapor
    +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/

    iter = 0;

    do {
      iter++;
      status = gsl_multimin_fminimizer_iterate(s);

      if (status) break;

      size = gsl_multimin_fminimizer_size(s);
      status = gsl_multimin_test_size(size, 1e-2);

    } while (status == GSL_CONTINUE && iter < 100);

    if ((w = (float)gsl_vector_get(s->x, 0)) > 7 || w <= 0){
      w = 3; continue;}

    w_sum += w;
    num++;


    /** gaseous transmittance
    +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
    
     kw = (int)floor(w/0.01);
     kms = (int)floor(atc.cg[g].ms/0.01);
     kmv = (int)floor(atc.cg[g].mv/0.01);

    for (b=0, bb=0; b<_NB_; b++){
      if (!par.use.refbands[b]) continue;
      Tsw = atc.Tw_lut[b][kw][kms];
      Tvw = atc.Tw_lut[b][kw][kmv];
      tg[bb]  = gas_transmitt(Tsw, Tvw, atc.cg[g].Tso[b], atc.cg[g].Tvo[b]);
      tg_sum[bb] += tg[bb];
      bb++;
    }


    /** replicate values at original resolution
    +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
    for (ii=0; ii<6; ii++){
    for (jj=0; jj<6; jj++){

      p = (i+ii)*_NY_+j+jj;

      if (QAI.off[p]) continue;

      W[p] = (int2s)(w*1000);
      for (bb=0; bb<_NO_; bb++) TG[bb][p]  = (int2s)(tg[bb]*10000);

    }
    }

  }
  }


  /** fill water and shadow with mean water vapor, set QAI flag
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
  if (num > 0){

    w_avg = w_sum/num;
    for (bb=0; bb<_NO_; bb++) tg_avg[bb] = tg_sum[bb]/num*10000;

    for (p=0; p<_NC_; p++){

      if (QAI.off[p] || W[p] > 0) continue;

      QAI.wvp[p] = true;

      W[p] = (int2s)(w_avg*1000);
      for (bb=0; bb<_NO_; bb++) TG[bb][p] = (int2s)tg_avg[bb];

    }

  }


  /** clean
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
  free((void*)tg); free((void*)tg_avg); free((void*)tg_sum);
  gsl_vector_free(x); gsl_vector_free(ss); gsl_multimin_fminimizer_free(s);
  free_3D((void***)atc.Tw_lut, _NB_, 701);

  #ifdef DEBUG
  db_write_image_1D((void**)QAI.wvp, "QAI-WVP", _NX_, _NY_, sizeof(bool), 1);
  db_write_image_1D((void**)W, "WVP", _NX_, _NY_, sizeof(int2s), 2);
  db_write_image_2D((void**)TG, "GAS-TRANSMITT", _NO_, _NX_, _NY_, sizeof(int2s), 2);
  #endif

  #ifdef CLOCK
  proctime_print("water vapor map", TIME);
  #endif

  *Tg  = TG;
  *WVP = W;

  return SUCCESS;
}


/** Estimate ozone amount
+++ This function computes the atmospheric ozone amount as a function of 
+++ gaeographic position and DOY. Note that this is a very rough estimate.
--- lon:    Longitude
--- lat:    Latitude
--- doy:    Day-of-Year
+++ Return: ozone amount
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
float ozone_amount(float lon, float lat, int doy){
float D, G, A, H, beta, C, F, I, DOY;
float o, omin;
D = 0.9856;
G = 20;
omin = 235;

  DOY = (float)doy;

  if (lat > 0) A = 150; else A = 100;
  if (lat > 0) H = 3; else H = 2;
  if (lat > 0) beta = 1.28; else beta = 1.5;
  if (lat > 0) C = 40; else C = 30;
  if (lat > 0) F = -30; else F = 152.625;

  if (lat > 0 && lon > 0) I = 20; else if (lat > 0 && lon < 0) I = 0; else I = -75;

  o = omin + (A+C*sin(deg2rad(D*(DOY+F))) + G*sin(deg2rad(H*(lon+I))))*sin(deg2rad(beta*lat))*sin(deg2rad(beta*lat));

  return(o/1000);
}


/** Water vapor from look up table
+++ This function reads water vapor estimates from a pre-compiled LUT or
+++ a constant value from the parameter file (see parameter file descrip-
+++ tion). This function is intended for sensors, which are not equipped
+++ with a water vapor sensitive band, e.g. Landsat. Two LUTs should be
+++ present. (1) daily LUTs (one per day), and (2) climatologies (one per
+++ month, i.e. 12 LUTs). Naming convention: (1) WVP_YYYY-MM-DD.txt, (2)
+++ WVP_0000-MM-00.txt. The files need to be plain text files, 4 columns 
+++ are separated by a blank space, no header. Columns are as follows: 
+++ Longitude, Latitude, Water vapor, Source/Number. Coordinates must be 
+++ given in geographic decimal degrees with negative values for South and
+++ West. Water vapor values are in cm, nodata value is 9999. The fourth
+++ column is either a three digit code indicating the source (e.g. MOD 
+++ or MYD for MODIS Terra or Aqua, respectively, TBD if nodata) for the
+++ daily LUTs (1) or the number of valid observations that were used to 
+++ compute the monthly climatology (2). The value of the line, which is
+++ closest to the image scene center is selected.
+++ Return: SUCCESS / FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int water_vapor_from_lut(){
char fname[2][MAX_STR_LEN];
char buffer[MAX_STR_LEN] = "\0";
char *tokenptr = NULL;
const char *separator = " ";
char *source = NULL;
FILE *fp = NULL;
coord_t center, site, diff;
float avg, dist, min_dist;
int k;

  #ifdef CLOCK
  time_t TIME; time(&TIME);
  #endif


  if (!par.doatmo) return SUCCESS;

  // skip if Sentinel-2
  if (meta.mission == MISSION_SENTINEL2){ atc.wvp = 0; return SUCCESS;}


  // scene center
  center.x = atc.cg[(atc.ngy/2)*atc.ngx+(atc.ngx/2)].geox;
  center.y = atc.cg[(atc.ngy/2)*atc.ngx+(atc.ngx/2)].geoy;

  // initialize
  atc.wvp = avg = -1;
  alloc((void**)&source, 4, sizeof(char));


  // use wvp from parameter file
  if (strcmp(par.d_wvp, "NULL") == 0){

    atc.wvp = avg = par.wvp;
    strncpy(source, "PRM", 3); source[3] = '\0';

    #ifdef DEBUG
    printf("Use wvp from parameter file: %.2f\n", atc.wvp);
    #endif

  // read wvp from LUT
  } else {

    // climatology LUT
    sprintf(fname[0], "%s/WVP_%04d-%02d-%02d.txt", par.d_wvp,
      0, meta.date.month, 0);

      // daily LUT
    sprintf(fname[1], "%s/WVP_%04d-%02d-%02d.txt", par.d_wvp,
      meta.date.year, meta.date.month, meta.date.day);

    for (k=0; k<2; k++){

      if (fileexist(fname[k])){
   
        if ((fp = fopen(fname[k], "r")) == NULL){
          printf("Unable to open LUT. "); return FAILURE;}

        min_dist = LONG_MAX;

        // read line by line and find closest coordinate
        while (fgets(buffer, MAX_STR_LEN, fp) != NULL){

          tokenptr = strtok(buffer, separator);

          site.x = atof(tokenptr); tokenptr = strtok(NULL, separator);
          site.y = atof(tokenptr); tokenptr = strtok(NULL, separator);
          diff.x = center.x-site.x; diff.y = center.y-site.y;

          dist = sqrt(diff.x*diff.x+diff.y*diff.y);
          if (dist < min_dist){
            min_dist = dist;
            atc.wvp = atof(tokenptr);
            if (k == 0){
              avg = atc.wvp;
              strncpy(source, "AVG", 3); source[3] = '\0';
            } else {
              tokenptr = strtok(NULL, separator);
              strncpy(source, tokenptr, 3); source[3] = '\0';
            }
          }

        }

        fclose(fp);

      }

    }

    // if daily value has nodata, use average value
    if (atc.wvp == 9999) atc.wvp = avg;

    #ifdef DEBUG
    printf("Use wvp from LUT: %.2f, avg: %.2f, src: %s, diff to scene center: %.2f\n", 
      atc.wvp, avg, source, min_dist);
    #endif

  }

  free((void*)source);

  #ifdef CLOCK
  proctime_print("water vapor from LUT", TIME);
  #endif

  // if no value was found, return FAILRE
  if (atc.wvp < 0) return FAILURE; else return SUCCESS;

}


/** MODIS / Terra data outages
+++ This function contains a listing of day-long sensor outages of the MO-
+++ DIS sensor aboard Terra. If a specified date matches one of the listed
+++ dates, true is returned, otherwise false.
+++-----------------------------------------------------------------------
+++ http://modaps.nascom.nasa.gov/services/production/outages_terra.html
+++-----------------------------------------------------------------------
--- d:      any date
+++ Return: true/false
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int failure_terra(date_t d){
int i;
int yv[47] = {  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  1,  1,
                1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  2,
                2,  2,  2,  2,  2,  2,  2,  2,  3,  3,  3,  3,  3,  3,  3,
                8,  8 };
int mv[47] = {  8,  8,  8,  8,  8,  8,  8,  8,  8,  8,  8,  8,  6,  6,  6,
                6,  6,  6,  6,  6,  6,  6,  6,  6,  6,  6,  6,  7,  7,  3,
                3,  3,  3,  3,  3,  3,  3,  4, 12, 12, 12, 12, 12, 12, 12,
               12, 12 };
int dv[47] = {  6,  7,  8,  9, 10, 11, 12, 13, 14, 15, 16, 17, 16, 17, 18,
               19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30,  1,  2, 20,
               21, 22, 23, 24, 25, 26, 27, 15, 17, 18, 19, 20, 21, 22, 23,
               21, 22 };

  for (i=0; i<47; i++){

    if (d.year == 2000+yv[i] && d.month == mv[i] && d.day == dv[i]){
      #ifdef DEBUG
      printf("TERRA outage. ");
      #endif
      return true;
    }
  }

  return false;
}


/** MODIS / Aqua data outages
+++ This function contains a listing of day-long sensor outages of the MO-
+++ DIS sensor aboard Aqua. If a specified date matches one of the listed
+++ dates, true is returned, otherwise false.
+++-----------------------------------------------------------------------
+++ http://modaps.nascom.nasa.gov/services/production/outages_aqua.html
+++-----------------------------------------------------------------------
--- d:      any date
+++ Return: true/false
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int failure_aqua(date_t d){
int i;
int yv[10] = {  2,  2,  2,  2,  2,  2,  2,  2,  2,  2 };
int mv[10] = {  7,  7,  8,  8,  8,  8,  8,  8,  8,  9 };
int dv[10] = { 30, 31,  1,  2,  3,  4,  5,  6,  7, 13 };

  for (i=0; i<10; i++){

    if (d.year == 2000+yv[i] && d.month == mv[i] && d.day == dv[i]){
      #ifdef DEBUG
      printf("AQUA outage. ");
      #endif
      return true;
    }
  }

  return false;
}


/** Read MODIS geometa table
+++ This function reads MODIS geometa tables (MOD03/MYD03) and performs an
+++ initial selection. Night granules are discarded (flagged invalid). 
+++ Granules at high latitudes are also discarded. Gring coordinates are
+++ sorted if not in UL-UR-LR-LL order. Gring coordinates of granules that
+++ extend between the Datum boundary are also modified.
--- fname: filename of geometa table
--- fid:   MODIS file IDs (returned)
--- gr:    MODIS Gring coordinates (returned)
--- v:     valid/invalid state (returned)
--- nl:    number of files (returned)
--- nv:    number of valid files (returned)
+++ Return:  SUCCESS/FAILURE
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int read_modis_geometa(char *fname, char ***fid, float ****gr, bool **v, int *nl, int *nv){
char buffer[MAX_STR_LEN] = "\0";
char *tokenptr = NULL;
char *separator = ",";
FILE *fp = NULL;
int l, g, k, nline = 0, nvalid = 0;
char **id = NULL;
bool *valid = NULL;
float ***bbox = NULL;
float ***gring = NULL;
float tmp, min[3];


  // allocate data
  alloc_2D((void***)&id, MAX_STR_LEN, MAX_STR_LEN, sizeof(char));
  alloc_3D((void****)&bbox,  MAX_STR_LEN, 2, 2, sizeof(float));
  alloc_3D((void****)&gring, MAX_STR_LEN, 2, 4, sizeof(float));
  alloc((void**)&valid, MAX_STR_LEN, sizeof(bool));


  // open file
  if (!(fp = fopen(fname, "r"))){
    printf("unable to open geometa file. "); return FAILURE;}


  // read line by line
  while (fgets(buffer, MAX_STR_LEN, fp) != NULL){

    // skip header
    if (buffer[0] == '#') continue;

    k = 0;

    tokenptr = strtok(buffer, separator);

    while(tokenptr != NULL){

      if (k == 0){ 
        strcpy(id[nline], tokenptr);
      } else if (k == 4){
        if (strcmp(tokenptr, "D") == 0 || strcmp(tokenptr, "B") == 0){
          valid[nline] = true;  // day image
          nvalid++;
        } else {
          valid[nline] = false; // night image
        }
      } else if (k ==  5){ bbox[nline][0][1]  = atof(tokenptr); // E
      } else if (k ==  6){ bbox[nline][1][0]  = atof(tokenptr); // N
      } else if (k ==  7){ bbox[nline][1][1]  = atof(tokenptr); // S
      } else if (k ==  8){ bbox[nline][0][0]  = atof(tokenptr); // W
      } else if (k ==  9){ gring[nline][0][0] = atof(tokenptr); // Gring X 1
      } else if (k == 10){ gring[nline][0][1] = atof(tokenptr); // Gring X 2
      } else if (k == 11){ gring[nline][0][2] = atof(tokenptr); // Gring X 3
      } else if (k == 12){ gring[nline][0][3] = atof(tokenptr); // Gring X 4
      } else if (k == 13){ gring[nline][1][0] = atof(tokenptr); // Gring Y 1
      } else if (k == 14){ gring[nline][1][1] = atof(tokenptr); // Gring Y 2
      } else if (k == 15){ gring[nline][1][2] = atof(tokenptr); // Gring Y 3
      } else if (k == 16){ gring[nline][1][3] = atof(tokenptr); // Gring Y 4
      }

      tokenptr = strtok(NULL, separator);
      k++;
    }

    nline++;

  }

  // swap coordinates if in reverse order, 
  // Aqua GRING coords are in reverse order because of S-N orbit
  for (l=0; l<nline; l++){

    if (!valid[l]) continue;

    // Granules at high latitudes are problematic.. Skip them
    if ((bbox[l][0][0] == -180 && bbox[l][0][1] == 180) || 
            (fabs(bbox[l][1][0]) > 89 || fabs(bbox[l][1][1]) > 89)){
      valid[l] = false;
      nvalid--;
      continue;
    }

    // Granules that extend between the Datum boundary are problematic
    // Correct for this; note that all coordinates are stored relative to GRing1
    // GRing x-coords can be > 180° or < -180° after this
    for (g=1; g<4; g++){

      min[0] = fabs(gring[l][0][0]-(gring[l][0][g]));
      min[1] = fabs(gring[l][0][0]-(gring[l][0][g]+360));
      min[2] = fabs(gring[l][0][0]-(gring[l][0][g]-360));

      if (min[1] < min[0] && min[1] < min[2]) gring[l][0][g] += 360;
      if (min[2] < min[0] && min[2] < min[1]) gring[l][0][g] -= 360;

    }

    // re-order GRing coordinates, such that they all are in UL-UR-LR-LL order
    if (gring[l][0][2] < gring[l][0][0] && gring[l][1][2] > gring[l][1][0]){
      tmp = gring[l][0][0]; gring[l][0][0] = gring[l][0][2]; gring[l][0][2] = tmp;
      tmp = gring[l][1][0]; gring[l][1][0] = gring[l][1][2]; gring[l][1][2] = tmp;
      tmp = gring[l][0][1]; gring[l][0][1] = gring[l][0][3]; gring[l][0][3] = tmp;
      tmp = gring[l][1][1]; gring[l][1][1] = gring[l][1][3]; gring[l][1][3] = tmp;
    }
    
    //#ifdef DEBUG
    //printf("(reordered) GRING:\n");
    //for (g=0; g<4; g++) printf("GRING %d: X/Y %+.2f/%+.2f\n", g, gring[l][0][g], gring[l][1][g]);
    //#endif

  }

  #ifdef DEBUG
  printf("%d lines, %d valid.\n", nline, nvalid);
  #endif

  fclose(fp);

  free_3D((void***)bbox, MAX_STR_LEN, 2);

  *nl  = nline;
  *nv  = nvalid;
  *gr  = gring;
  *v   = valid;
  *fid = id;

  return SUCCESS;
}


/** Identify intersecting MODIS granules
+++ This function identifies MODIS granules that intersect with a pair of
+++ coordinates (+- 0.75°). The spatial selection is based on the Gring 
+++ coordinates of the MODIS granules. The function returns the number of
+++ intersecting granules, as well as a pointer to these elements.
--- lat:    latitude of input coordinate
--- lon:    longitude of input coordinate
--- gr:     MODIS Gring coordinates
--- v:      valid/invalid state
--- nl:     number of granules
--- nv:     number of valid granules
--- ptr:    pointer to intersecting elements (returned)
+++ Return: number of intersecting elements
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int modis_intersect(float lat, float lon, float ***gr, bool *v, int nl, int nv, int **ptr){
int *p = NULL;
int l, j, k = 0;
float wlon, elon, nlat, slat, dboff[3] = { 0, 360, -360 };


  alloc((void**)&p, nv, sizeof(int));

  for (l=0; l<nl; l++){

    if (!v[l]) continue;

    for (j=0; j<3; j++){

      wlon = (lon-0.75)+dboff[j]; elon = (lon+0.75)+dboff[j];
      nlat = lat+0.75; slat = lat-0.75;

      if ((wlon < (gr[l][0][1]*(gr[l][1][2]-nlat)+
          gr[l][0][2]*(nlat-gr[l][1][1]))/(gr[l][1][2]-gr[l][1][1]) || 
         wlon < (gr[l][0][1]*(gr[l][1][2]-slat)+
          gr[l][0][2]*(slat-gr[l][1][1]))/(gr[l][1][2]-gr[l][1][1]))
          &&
        (elon > (gr[l][0][0]*(gr[l][1][3]-nlat)+
          gr[l][0][3]*(nlat-gr[l][1][0]))/(gr[l][1][3]-gr[l][1][0]) || 
         elon > (gr[l][0][0]*(gr[l][1][3]-slat)+
          gr[l][0][3]*(slat-gr[l][1][0]))/(gr[l][1][3]-gr[l][1][0]))
          &&
        (nlat > (gr[l][1][3]*(gr[l][0][2]-wlon)+
          gr[l][1][2]*(wlon-gr[l][0][3]))/(gr[l][0][2]-gr[l][0][3]) || 
         nlat > (gr[l][1][3]*(gr[l][0][2]-elon)+
          gr[l][1][2]*(elon-gr[l][0][3]))/(gr[l][0][2]-gr[l][0][3]))
          &&
        (slat < (gr[l][1][0]*(gr[l][0][1]-wlon)+
          gr[l][1][1]*(wlon-gr[l][0][0]))/(gr[l][0][1]-gr[l][0][0]) || 
         slat < (gr[l][1][0]*(gr[l][0][1]-elon)+
          gr[l][1][1]*(elon-gr[l][0][0]))/(gr[l][0][1]-gr[l][0][0]))){
        p[k] = l;
        k++;
      }
    }
  }

  *ptr = p;

  return k;
}


/** Identify enclosing MODIS granules
+++ This function identifies an intersecting MODIS granule that completely
+++ contains a pair of coordinates (+- 0.75°). The spatial selection is 
+++ based on the Gring coordinates of the MODIS granules. If no such gra-
+++ nule exists, -1 is returned, otherwise the function returns this ele-
+++ ment.
--- lat:    latitude of input coordinate
--- lon:    longitude of input coordinate
--- gr:     MODIS Gring coordinates
--- ni:     number of intersecting granules
--- ptr:    pointer to intersecting elements
+++ Return: number of intersecting elements
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
int modis_inside(float lat, float lon, float ***gr, int nintersect, int *ptr){
int l, j;
float wlon, elon, nlat, slat, dboff[3] = { 0, 360, -360 };


  // return the first granule that completely contains the frame
  for (l=0; l<nintersect; l++){

    for (j=0; j<3; j++){

      wlon = (lon-0.75)+dboff[j]; elon = (lon+0.75)+dboff[j];
      nlat = lat+0.75; slat = lat-0.75;

      if ((wlon > (gr[ptr[l]][0][0]*(gr[ptr[l]][1][3]-nlat)+
          gr[ptr[l]][0][3]*(nlat-gr[ptr[l]][1][0]))/(gr[ptr[l]][1][3]-gr[ptr[l]][1][0])
          && 
        wlon > (gr[ptr[l]][0][0]*(gr[ptr[l]][1][3]-slat)+
          gr[ptr[l]][0][3]*(slat-gr[ptr[l]][1][0]))/(gr[ptr[l]][1][3]-gr[ptr[l]][1][0])
          && 
        elon < (gr[ptr[l]][0][1]*(gr[ptr[l]][1][2]-nlat)+
          gr[ptr[l]][0][2]*(nlat-gr[ptr[l]][1][1]))/(gr[ptr[l]][1][2]-gr[ptr[l]][1][1])
          && 
        elon < (gr[ptr[l]][0][1]*(gr[ptr[l]][1][2]-slat)+
          gr[ptr[l]][0][2]*(slat-gr[ptr[l]][1][1]))/(gr[ptr[l]][1][2]-gr[ptr[l]][1][1])
          && 
        nlat < (gr[ptr[l]][1][0]*(gr[ptr[l]][0][1]-wlon)+
          gr[ptr[l]][1][1]*(wlon-gr[ptr[l]][0][0]))/(gr[ptr[l]][0][1]-gr[ptr[l]][0][0])
          && 
        slat > (gr[ptr[l]][1][3]*(gr[ptr[l]][0][2]-wlon)+
          gr[ptr[l]][1][2]*(wlon-gr[ptr[l]][0][3]))/(gr[ptr[l]][0][2]-gr[ptr[l]][0][3])
          && 
        nlat < (gr[ptr[l]][1][0]*(gr[ptr[l]][0][1]-elon)+
          gr[ptr[l]][1][1]*(elon-gr[ptr[l]][0][0]))/(gr[ptr[l]][0][1]-gr[ptr[l]][0][0])
          && 
        slat > (gr[ptr[l]][1][3]*(gr[ptr[l]][0][2]-elon)+
          gr[ptr[l]][1][2]*(elon-gr[ptr[l]][0][3]))/(gr[ptr[l]][0][2]-gr[ptr[l]][0][3]))){

        return ptr[l];

      }

    }
  }

  // if no granule completely contains the frame: return -1
  return -1;
}


/** Compile MODIS water vapor
+++ This function identifies the MODIS data that should be processed, then
+++ downloads them, reads them, and computes the average for each reques-
+++ ted coordinate.
--- dir_geo:   Directory of MODIS geometa tables
--- dir_hdf:   Directory of MODIS water vapor data
--- d_now:     Date
--- sen:       Sensor (TERRA/AQUA MODIS)
--- nc:        Number of coordinates
--- COO:       Coordinate array
--- avg:       Water vapor averages
--- count:     Number of pixels for averaging
+++ Return:    void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void compile_modis_wvp(char *dir_geo, char *dir_hdf, date_t d_now, char *sen, int nc, float **COO, double **avg, double **count){
char geoname[MAX_STR_LEN];
//char *hdfname = NULL;
char pattern[MAX_STR_LEN], ftp_pattern[MAX_STR_LEN], doy[4];
char httplist[MAX_STR_LEN];
char loclist[MAX_STR_LEN];
char basename[MAX_STR_LEN];
char httpname[MAX_STR_LEN];
char fullname[MAX_STR_LEN];
char buffer[MAX_STR_LEN];
int curl;
float ***gr = NULL; // gring coords: (0/0-3): Lon/ul-ur-lr-ll, (1/0-3): Lat/ul-ur-lr-ll
char **id = NULL;
bool *v = NULL;
int *ptr = NULL;
char *str = NULL;
int nl, nv, ni;
int i, c, p;
int y, yc, yct, x, xc, xct, pos, posc;
int try__;
GDALDatasetH hdfDS, subDS;
GDALRasterBandH band;
char **sds = NULL, **metadata = NULL;
char KeyName[MAX_STR_LEN];
char *sdsname = NULL;
int nx, ny, nxc, nyc;
int yoff, xoff;
float voff, vscl, val;
int fill;
int2s *PW = NULL;
int1u *CLD = NULL;
float *LAT = NULL;
float *LON = NULL;
int pcld = 6, pwvp = 7, plat = 12, plon = 13;
bool bit[8];
FILE *fp = NULL;
double sumwvp, ctrwvp, *average = NULL, *ctrall = NULL;
float wlon, elon, nlat, slat;
bool ok;
const char *separator = ",";


  // allocate and initialize wvp averages and pixel counts
  alloc((void**)&average, nc, sizeof(double));
  alloc((void**)&ctrall,  nc, sizeof(double));
  for (c=0; c<nc; c++) average[c] = 9999;

  // geomate filename
  sprintf(geoname, "%s/%s03_%4d-%02d-%02d.txt", dir_geo, sen, d_now.year, d_now.month, d_now.day);

  // if geometa file doesn't exist, return
  if (!fileexist(geoname)){ *avg = average; *count = ctrall; return;}


  GDALAllRegister();

  // read the geometa table
  if (read_modis_geometa(geoname, &id, &gr, &v, &nl, &nv) != SUCCESS){
    printf("compiling wvp failed.\n"); exit(1);}

  #ifdef DEBUG
  printf("%d valid lines in geometa\n", nv);
  #endif

  // if no granule is valid, return
  if (nv == 0){ *avg = average; *count = ctrall; return;}


  // do for each requested coordinate
  for (c=0; c<nc; c++){

    sumwvp = ctrwvp = 0;

    // estimate average of wvp in a 1.5° x 1.5° box
    wlon = COO[0][c]-0.75; elon = COO[0][c]+0.75;
    nlat = COO[1][c]+0.75; slat = COO[1][c]-0.75;

    // get intersecting granules
    if ((ni = modis_intersect(COO[1][c], COO[0][c], gr, v, 
                                  nl, nv, &ptr)) < 1) continue;

    #ifdef DEBUG
    printf("requested box: UL %.2f/%.2f, LR %.2f/%.2f\n", wlon, nlat, elon, slat);
    printf("%d intersecting granules\n", ni);
    #endif

    // if more than one was found, check if one is completely inside, then use only this one
    if (ni > 1){
      if ((p = modis_inside(COO[1][c], COO[0][c], gr, 
                         ni, ptr)) > -1){
        ptr[0] = p;
        ni = 1;

        #ifdef DEBUG
        printf("one granule contains the box completely.\n");
        #endif

      }
    }


    // for every remaining granule: get correct HDF name, download if not there and process
    for (i=0; i<ni; i++){

      // appr. HDF name with wildcards
      strncpy(pattern, sen, 3);
      strncpy(pattern+3, "05_L2", 5);
      strncpy(pattern+8, id[ptr[i]]+5, 14);
      pattern[22] = '\0';
      strncpy(doy, pattern+14, 3); doy[3] = '\0';
      sprintf(ftp_pattern, "https://ladsweb.modaps.eosdis.nasa.gov/archive/allData/61/%s05_L2/%4d/%s/%s*", 
        sen, d_now.year, doy, pattern);


      // download file listing
      try__ = 0;
      sprintf(httplist, "https://ladsweb.modaps.eosdis.nasa.gov/archive/allData/61/%s05_L2/%4d/%s.csv",
        sen, d_now.year, doy);
      sprintf(loclist, "%s/%s05_L2-%04d-%s.csv", dir_hdf, sen, d_now.year, doy);

      while (try__ < 10 && !fileexist(loclist)){
        curl = download_file(httplist, loclist);
        printf("\b\b%2d", try__+1);
        if (curl != 0) printf("; curl: %02d. fail. ", curl);
        if (curl == 78) try__ = 99;
        try__++; fflush(stdout);
      }

      if (!fileexist(loclist)) continue;


      if ((fp = fopen(loclist, "r")) == NULL){
        printf("Unable to open file list!\n"); continue;}
      ok = false;

      while (fgets(buffer, MAX_STR_LEN, fp) != NULL){
        if (strstr(buffer, pattern) != NULL){
          //puts(buffer);
          str = strtok(buffer, separator);
          strncpy(basename, str, strlen(str));
          basename[strlen(str)] = '\0';
          //printf("%s\n", basename);
          ok = true;
        }
      }

      fclose(fp);

      if (!ok){
        printf("Unable to locate name in filelist!\n"); continue;}

      sprintf(fullname, "%s/%s", dir_hdf, basename);
      sprintf(httpname, "https://ladsweb.modaps.eosdis.nasa.gov/archive/allData/61/%s05_L2/%4d/%s/%s", sen, d_now.year, doy, basename);

      //printf("%s\n", fullname);
      //printf("%s\n", httpname);

      for (try__=0; try__ < 10; try__++){

        //if ((hdfname = findfile(dir_hdf, pattern)) != NULL) break;
        if (fileexist(fullname)) break;

        #ifdef DEBUG
        printf("Download file (%s). Try # %2d\n", pattern, try__);
        #endif

        //download_pattern(dir_hdf, ftp_pattern);
        curl = download_file(httpname, fullname);
        printf("\b\b%2d", try__+1);
        if (curl != 0) printf("; curl: %02d. fail. ", curl);
        if (curl == 78) try__ = 99;

        //if ((hdfname = findfile(dir_hdf, pattern)) != NULL) break;
        if (fileexist(fullname)) break;

      }


      //if (hdfname == NULL){
        
      //  alloc((void**)&hdfname, 1024, sizeof(char));
      //  sprintf(hdfname, "%s/%s.fail", dir_hdf, pattern);
      //  fp = fopen(hdfname, "w"); fclose(fp);

      //  #ifdef DEBUG
      //  printf("Failed to download. Touch ghost file.\n");
      //  #endif

      //} else {
      if (!fileexist(fullname)) continue;

      //  #ifdef DEBUG
      //  printf("File (%s) is there, process.\n", pattern);
      //  #endif

      //}

      //if (strstr(hdfname, ".fail") != NULL){
      //  #ifdef DEBUG
      //  printf("File is ghost, skip.\n");
      //  #endif
      //  free((void*)hdfname); hdfname = NULL; continue;
      //}


      // open input dataset
      if ((hdfDS = GDALOpen(fullname, GA_ReadOnly)) == NULL){
        printf("unable to open image\n"); exit(1);
      } else {
        //free((void*)hdfname); hdfname = NULL;
      }

      // get SDS listing
      sds = GDALGetMetadata(hdfDS, "SUBDATASETS");
      if (CSLCount(sds) == 0){
        printf("unable to retrieve SDS list.\n"); exit(1);}



      // read NIR water vapour retrieval
      sprintf(KeyName, "SUBDATASET_%d_NAME", pwvp);
      sdsname = CPLStrdup(CSLFetchNameValue(sds, KeyName));
      subDS = GDALOpen(sdsname, GA_ReadOnly);
      nx = GDALGetRasterXSize(subDS); 
      ny = GDALGetRasterYSize(subDS);
      alloc((void**)&PW, nx*ny, sizeof(int2s));
      band = GDALGetRasterBand(subDS, 1);
      GDALRasterIO(band, GF_Read,  0, 0, nx, ny, 
            PW, nx, ny, GDT_Int16, 0, 0);
      CPLFree(sdsname);

      // read some metadata
      metadata = GDALGetMetadata(subDS, "GEOLOCATION");
      yoff = atoi(CSLFetchNameValue(metadata,"LINE_OFFSET"));
      xoff = atoi(CSLFetchNameValue(metadata,"PIXEL_OFFSET"));
      //CSLDestroy(metadata);
      
      metadata = GDALGetMetadata(subDS, NULL);
      voff = atof(CSLFetchNameValue(metadata,"add_offset"));
      vscl = atof(CSLFetchNameValue(metadata,"scale_factor"));
      fill = atoi(CSLFetchNameValue(metadata,"_FillValue"));
      //CSLDestroy(metadata);
      GDALClose(subDS);

      // read latitude
      sprintf(KeyName, "SUBDATASET_%d_NAME", plat);
      sdsname = CPLStrdup(CSLFetchNameValue(sds, KeyName));
      subDS = GDALOpen(sdsname, GA_ReadOnly);
      nxc = GDALGetRasterXSize(subDS); 
      nyc = GDALGetRasterYSize(subDS);
      alloc((void**)&LAT, nxc*nyc, sizeof(float));
      band = GDALGetRasterBand(subDS, 1);
      GDALRasterIO(band, GF_Read,  0, 0, nxc, nyc, 
            LAT, nxc, nyc, GDT_Float32, 0, 0);
      GDALClose(subDS);
      CPLFree(sdsname);

      // read longitude
      sprintf(KeyName, "SUBDATASET_%d_NAME", plon);
      sdsname = CPLStrdup(CSLFetchNameValue(sds, KeyName));
      subDS = GDALOpen(sdsname, GA_ReadOnly);
      alloc((void**)&LON, nxc*nyc, sizeof(float));
      band = GDALGetRasterBand(subDS, 1);
      GDALRasterIO(band, GF_Read,  0, 0, nxc, nyc, 
            LON, nxc, nyc, GDT_Float32, 0, 0);
      GDALClose(subDS);
      CPLFree(sdsname);

      // read cloud state
      sprintf(KeyName, "SUBDATASET_%d_NAME", pcld);
      sdsname = CPLStrdup(CSLFetchNameValue(sds, KeyName));
      subDS = GDALOpen(sdsname, GA_ReadOnly);
      alloc((void**)&CLD, nx*ny, sizeof(int1u));
      band = GDALGetRasterBand(subDS, 1);
      GDALRasterIO(band, GF_Read,  0, 0, nx, ny, 
            CLD, nx, ny, GDT_Byte, 0, 0);
      GDALClose(subDS);
      CPLFree(sdsname);

      //CSLDestroy(sds);
      GDALClose(hdfDS);


      // sum up precipitable water for mean
      for (y=yoff, yc=0, yct=1; y<ny; y++){
      for (x=xoff, xc=0, xct=1; x<nx; x++){
       
        pos  = nx*y+x;
        posc = nxc*yc+xc;

        //if in frame
        if (LON[posc] >= wlon && LON[posc] <= elon &&
          LAT[posc] <= nlat && LAT[posc] >= slat){

          if (PW[pos] != fill){

            int2bit(CLD[pos], bit, 0, 8);

            if (bit[2] && bit[4]){
              val = vscl*(PW[pos]-voff);
              sumwvp += val;
              ctrwvp++;
            }

          }
          ctrall[c]++;
        }

        if ((xct++) >= 5){ xct = 1; xc++;}
      }
      if ((yct++) >= 5){ yct = 1; yc++;}
      }

      free((void*)PW); free((void*)CLD);
      free((void*)LAT); free((void*)LON);

    }

    // only calculate average if at least 10% of in-frame pixels were valid
    if (ctrall[c] > 0 && ctrwvp >= ctrall[c]/10){
      average[c] = sumwvp/ctrwvp;
    } else average[c] = 9999;

    free((void*)ptr);
  }

  // free memory
  free_3D((void***)gr, MAX_STR_LEN, 2);
  free_2D((void**)id, MAX_STR_LEN);
  free((void*)v);

  *avg   = average;
  *count = ctrall;
  
  return;
}


/** Choose between AQUA and TERRA water vapor estimate
+++ This function chooses, which water vapor estimate to use. This is 
+++ based on data availability and the number of pixels that were used
+++ for averaging.
--- aqua:   AQUA commissioning flag
--- nc:     Number of coordinates
--- WVP:    Water vapor array
--- SEN:    Sensor source array
--- modavg: TERRA averages
--- mydavg: AQUA  averages
--- modctr: TERRA number of pixels
--- mydctr: AQUA  number of pixels
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void choose_modis_wvp(bool aqua, int nc, float *WVP, char **SEN, double *modavg, double *mydavg, double *modctr, double *mydctr){
int c;
float ctr = 0;


  for (c=0; c<nc; c++){
    if (!aqua){
      WVP[c] = modavg[c];
      if (WVP[c] < 9999) strcpy(SEN[c], "MOD");
    } else {
      if (modavg[c] < 9999 && mydavg[c] < 9999){
        if (modctr[c] >= mydctr[c]){
          WVP[c] = modavg[c];
          if (WVP[c] < 9999) strcpy(SEN[c], "MOD");
        } else {
          WVP[c] = mydavg[c];
          if (WVP[c] < 9999) strcpy(SEN[c], "MYD");
        }
      } else if (modavg[c] < 9999 && mydavg[c] >= 9999){
        WVP[c] = modavg[c];
        if (WVP[c] < 9999) strcpy(SEN[c], "MOD");
      } else if (modavg[c] >= 9999 && mydavg[c] < 9999){
        WVP[c] = mydavg[c];
        if (WVP[c] < 9999) strcpy(SEN[c], "MYD");
      }
    }
    if (WVP[c] < 9999) ctr++;
  }

  printf("%5.1f%% of useful frames.\n", ctr/nc*100.0);

  return;
}


/** Write daily water vapor LUT
+++ This function writes a daily water vapor Look-Up-Table.
--- fname:  LUT filename
--- nc:     Number of coordinates
--- COO:    Coordinate array
--- WVP:    Water vapor array
--- SEN:    Sensor source array
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void write_wvp_lut(char *fname, int nc, float **COO, float *WVP, char **SEN){
FILE *fp;
int c;

  if ((fp = fopen(fname, "w")) == NULL){
    printf("couldn't open file for writing.\n"); exit(1);}

  for (c=0; c<nc; c++){
    fprintf(fp, "%.4f %.4f %f %s\n", COO[0][c], COO[1][c], WVP[c], SEN[c]);
  }

  fclose(fp);

  return;
}


/** Write climatology water vapor LUT
+++ This function writes a climatology water vapor Look-Up-Table.
--- dir_wvp: Output directory
--- nc:      Number of coordinates
--- COO:     Coordinate array
--- AVG:     Water vapor average array
+++ Return:  void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void write_avg_table(char *dir_wvp, int nc, float **COO, double ***AVG){
FILE *fp;
int c, m;
char fname[MAX_STR_LEN];

  for (m=0; m<12; m++){
    sprintf(fname, "%s/WVP_0000-%02d-00.txt", dir_wvp, m+1);
    if ((fp = fopen(fname, "w")) == NULL){
      printf("couldn't open file for writing.\n"); exit(1);}
    for (c=0; c<nc; c++){
      fprintf(fp, "%.4f %.4f %f %f %.0f\n", COO[0][c], COO[1][c], 
        AVG[0][m][c], AVG[1][m][c], AVG[2][m][c]);
    }
    fclose(fp);
  }

  return;
}


/** Read water vapor LUT
+++ This function reads a water vapor Look-Up-Table (daily or climatology)
--- fname:  LUT filename
--- nc:     Number of coordinates
--- COO:    Coordinate array
--- WVP:    Water vapor array
+++ Return: void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void read_wvp_lut(char *fname, int nc, float **COO, float *WVP){
FILE *fp;
char  buffer[MAX_STR_LEN] = "\0";
char *tokenptr = NULL;
char *separator = " ";
int c;
float x, y;

  fp = fopen(fname, "r");

  // process line by line
  for (c=0; c<nc; c++){

    if (fgets(buffer, MAX_STR_LEN, fp) == NULL){
      printf("invalid wvp table!\n"); exit(1);}

    tokenptr = strtok(buffer, separator);
    x = atof(tokenptr); tokenptr = strtok(NULL, separator);
    y = atof(tokenptr); tokenptr = strtok(NULL, separator);

    if (fabs(x-COO[0][c]) > 0.0001 || fabs(y-COO[1][c]) > 0.0001){
      printf("Invalid wvp table! Coordinates are messed up..\n"); exit(1);}

    WVP[c] = atof(tokenptr);

  }

  fclose(fp);

  return;
}


/** Create daily water vapor LUT
+++ This function creates a daily water vapor Look-Up-Table for the given
+++ date from MODIS data, which are downloaded from LAADS.
--- dir_geo:   Directory of MODIS geometa tables
--- dir_hdf:   Directory of MODIS water vapor data
--- tablename: Name of LUT
--- d_now:     Date
--- nc:        Number of coordinates
--- COO:       Coordinate array
--- WVP:       Water vapor array
--- SEN:       Sensor source array
+++ Return:    void
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++**/
void create_wvp_lut(char *dir_geo, char *dir_hdf, char *tablename, date_t d_now, int nc, float **COO, float *WVP, char **SEN){
char ftpname[MAX_STR_LEN], locname[MAX_STR_LEN];
int c, try__, nvalid, curl;
bool aqua = false;
double *modavg, *mydavg, *modctr, *mydctr;


  // test if AQUA was commissioned
  if (d_now.year*10000+d_now.month*100+d_now.day  >= 20020703) aqua = true;

  // initialize precipitable water with fill
  for (c=0; c<nc; c++){
    WVP[c] = 9999;
    strcpy(SEN[c], "TBD");
  }

  // if TERRA geometa doesn't exist: download
  if (!failure_terra(d_now)) try__ = 0; else try__ = 99;

  sprintf(ftpname, "https://ladsweb.modaps.eosdis.nasa.gov/archive/geoMeta/6/TERRA/%4d/MOD03_%4d-%02d-%02d.txt", 
    d_now.year, d_now.year, d_now.month, d_now.day);
  sprintf(locname, "%s/MOD03_%4d-%02d-%02d.txt", dir_geo, d_now.year, d_now.month, d_now.day);

  if (try__ < 1){
    if (!fileexist(locname)){ printf("Download TERRA geometa. Try #   ");
    } else printf("TERRA geometa exists.");
  }

  while (try__ < 10 && !fileexist(locname)){
    curl = download_file(ftpname, locname);
    printf("\b\b%2d", try__+1);
    if (curl != 0) printf("; curl: %02d. fail. ", curl);
    if (curl == 78) try__ = 99;
    try__++; fflush(stdout);
  }
  printf("\n");

  // compile water vapour
  compile_modis_wvp(dir_geo, dir_hdf, d_now, "MOD", nc, COO, &modavg, &modctr);

  if (aqua){

    // if AQUA geometa doesn't exist: download
    if (!failure_aqua(d_now)) try__ = 0; else try__ = 99;

    sprintf(ftpname, "https://ladsweb.modaps.eosdis.nasa.gov/archive/geoMeta/6/AQUA/%4d/MYD03_%4d-%02d-%02d.txt", 
    d_now.year, d_now.year, d_now.month, d_now.day);
    sprintf(locname, "%s/MYD03_%4d-%02d-%02d.txt", dir_geo, d_now.year, d_now.month, d_now.day);

    if (try__ < 1){
      if (!fileexist(locname)){ printf("Download AQUA geometa. Try #   ");
      } else printf("AQUA geometa exists.");
    }

    while (try__ < 10 && !fileexist(locname)){
      curl = download_file(ftpname, locname);
      printf("\b\b%2d", try__+1);
      if (curl != 0) printf("; curl: %02d. fail. ", curl);
      if (curl == 78) try__ = 99;
      try__++; fflush(stdout);
    }
  }
  printf("\n");

  // compile water vapour
  if (aqua) compile_modis_wvp(dir_geo, dir_hdf, d_now, "MYD", nc, COO, &mydavg, &mydctr);

  // choose between TERRA and AQUA
  choose_modis_wvp(aqua, nc, WVP, SEN, modavg, mydavg, modctr, mydctr);

  // clean
  free((void*)modavg); free((void*)modctr);
  if (aqua){ free((void*)mydavg); free((void*)mydctr);}

  // write table
  for (c=0, nvalid=0; c<nc; c++) if (WVP[c] != 9999) nvalid++;
  if (nvalid > 0) write_wvp_lut(tablename, nc, COO, WVP, SEN);
  
  return;
}

